// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/Balancer_skel.h>
#include <OB/Balancer_init.h>
#include <OB/BalancerUtil_init.h>

#include <Hello_impl.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

//
// LoadBalancing::LoadAlert servant class.
//
class LoadAlert_impl : virtual public POA_LoadBalancing::LoadAlert,
		       virtual public PortableServer::RefCountServantBase
{
    PortableServer::POA_var poa_;
    LoadBalancing::Util::LoadAlert_var alert_;

public:
    LoadAlert_impl(PortableServer::POA_ptr poa,
		   LoadBalancing::Util::LoadAlert_ptr alert)
	: poa_(PortableServer::POA::_duplicate(poa)),
	  alert_(LoadBalancing::Util::LoadAlert::_duplicate(alert))
    {
    }

    virtual PortableServer::POA_ptr _default_POA()
    {
	return PortableServer::POA::_duplicate(poa_);
    }

    virtual void alert()
	throw(CORBA::SystemException)
    {
	alert_ -> alert();
    }
};

//
// Usage message.
//
static void
usage(const char* progName)
{
    cerr << "Usage: " << progName << " MEMBER_ID\n";
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Check arguments
    //
    if(argc != 2)
    {
	usage(argv[0]);
	return EXIT_FAILURE;
    }
    const char* memberId = argv[1];

    //
    // Resolve Root POA.
    //
    CORBA::Object_var poaObj = orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);
    
    //
    // Get a reference to the POA manager.
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

    //
    // Create the member policy.
    //
    LoadBalancing::MemberPolicyValue_var value =
	new LoadBalancing::MemberPolicyValue();
    value -> group_id = CORBA::string_dup("Hello");
    value -> member_id = CORBA::string_dup(memberId);
    CORBA::Any any;
    any <<= value._retn();
    CORBA::Policy_var memberPolicy =
	orb -> create_policy(LoadBalancing::MEMBER_POLICY_ID, any);

    //
    // Create a POA for the hello replica.
    //
    CORBA::PolicyList pl(3);
    pl.length(3);
    pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
    pl[1] = rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
    pl[2] = memberPolicy;
    PortableServer::POA_var helloPOA =
        rootPOA -> create_POA("helloPOA", manager, pl);

    //
    // Create implementation object.
    //
    Hello_impl* helloImpl = new Hello_impl(helloPOA, memberId);
    PortableServer::ServantBase_var helloServant = helloImpl;
    PortableServer::ObjectId_var oid =
	PortableServer::string_to_ObjectId("Hello");
    helloPOA -> activate_object_with_id(oid.in(), helloImpl);
    Hello_var hello = helloImpl -> _this();

    //
    // Get the GroupFactory, LoadUpdater and LoadAlert objects.
    //
    LoadBalancing::GroupFactory_var factory;
    LoadBalancing::Util::LoadUpdater_var updater;
    LoadBalancing::Util::LoadAlert_var alert;
    try
    {
	CORBA::Object_var obj =
	    orb -> resolve_initial_references("Balancer");
	factory = LoadBalancing::GroupFactory::_narrow(obj);

	obj = orb -> resolve_initial_references("LoadUpdater");
	updater = LoadBalancing::Util::LoadUpdater::_narrow(obj);

	obj = orb -> resolve_initial_references("LoadAlert");
	alert = LoadBalancing::Util::LoadAlert::_narrow(obj);
    }
    catch(const CORBA::ORB::InvalidName&)
    {
    }
    assert(!CORBA::is_nil(factory));
    assert(!CORBA::is_nil(updater));
    assert(!CORBA::is_nil(alert));

    //
    // Create LoadAlert implementation object.
    //
    LoadAlert_impl* loadAlertImpl = new LoadAlert_impl(rootPOA, alert);
    PortableServer::ServantBase_var alertServant = loadAlertImpl;
    LoadBalancing::LoadAlert_var loadAlert = loadAlertImpl -> _this();

    //
    // Activate the POA manager.
    //
    manager -> activate();

    //
    // Set the LoadAlert for the member.
    //
    try
    {
	LoadBalancing::Group_var group = factory -> get("Hello");
	group -> set_load_alert(memberId, loadAlert);
    }
    catch(const LoadBalancing::GroupNotFound&)
    {
	assert(false); // should be caught during POA creation
    }
    catch(const LoadBalancing::MemberNotFound&)
    {
	assert(false); // should be caught during POA creation
    }

    //
    // Register the group with the LoadUpdater.
    //
    updater -> register_member(memberId, "Hello");

    //
    // Start event loop.
    //
    orb -> run();

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	//
	// Initialize the server to be load balanced.
	//
	LoadBalancing::LB_init();
	LoadBalancing::Util::LBUtil_init();

	//
	// Initialize the ORB
	//
	orb = CORBA::ORB_init(argc, argv);

	//
	// Start the server
	//
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
