// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/Logger.h>
#include <OB/Balancer_init.h>

#include <Library_impl.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <errno.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

//
// Usage message.
//
static void
usage(const char* progName)
{
    cerr << "Usage: " << progName << " MEMBER_ID\n";
}

//
// The library and book servant locator
//
class LibraryLocator_impl : virtual public POA_PortableServer::ServantLocator,
			    virtual public PortableServer::RefCountServantBase
{
    //
    // The POA
    //
    PortableServer::POA_var poa_;

    //
    // The Library servant
    //
    PortableServer::ServantBase_var library_;

    //
    // The Book servant
    //
    PortableServer::ServantBase_var book_;

public:

    LibraryLocator_impl(PortableServer::POA_ptr poa,
		        Library_impl* library,
		        Book_impl* book)
	: poa_(PortableServer::POA::_duplicate(poa))
    {
	library -> _add_ref();
	library_ = library;
	book -> _add_ref();
	book_ = book;
    }
    ~LibraryLocator_impl()
    {
    }

    virtual PortableServer::POA_ptr _default_POA()
    {
	return PortableServer::POA::_duplicate(poa_);
    }

    PortableServer::Servant
    preinvoke(const PortableServer::ObjectId& oid,
	      PortableServer::POA_ptr /*poa*/,
	      const char* /*operation*/,
	      PortableServer::ServantLocator::Cookie& /*cookie*/)
	throw(PortableServer::ForwardRequest, CORBA::SystemException)
    {
	CORBA::String_var id;
	try
	{
	    id = PortableServer::ObjectId_to_string(oid);
	}
	catch(const CORBA::BAD_PARAM&)
	{
	    throw CORBA::OBJECT_NOT_EXIST();
	}

	PortableServer::Servant servant = 0;
	if(strcmp(id, "Library") == 0)
	    servant = library_.in();
	else
	    servant = book_.in();

	servant -> _add_ref();
	return servant;
    }

    void
    postinvoke(const PortableServer::ObjectId& /*oid*/,
	       PortableServer::POA_ptr /*poa*/,
	       const char* /*operation*/,
	       PortableServer::ServantLocator::Cookie /*cookie*/,
	       PortableServer::Servant servant)
	throw(CORBA::SystemException)
    {
	servant -> _remove_ref();
    }
};

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Check arguments
    //
    if(argc != 2)
    {
	usage(argv[0]);
	return EXIT_FAILURE;
    }
    const char* memberId = argv[1];

    //
    // Resolve Root POA.
    //
    CORBA::Object_var poaObj = orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);
    
    //
    // Get a reference to the POA manager.
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();
    
    //
    // Create the member policy.
    //
    LoadBalancing::MemberPolicyValue_var value =
	new LoadBalancing::MemberPolicyValue();
    value -> group_id = CORBA::string_dup("Library");
    value -> member_id = CORBA::string_dup(memberId);
    CORBA::Any any;
    any <<= value._retn();
    CORBA::Policy_var memberPolicy =
	orb -> create_policy(LoadBalancing::MEMBER_POLICY_ID, any);

    //
    // Create a POA for the replicas of the "Library" load balanced group.
    //
    CORBA::PolicyList pl(5);
    pl.length(5);

    pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
    pl[1] = rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
    pl[2] = rootPOA -> create_request_processing_policy(
	PortableServer::USE_SERVANT_MANAGER);
    pl[3] = rootPOA -> create_servant_retention_policy(
	PortableServer::NON_RETAIN);
    pl[4] = memberPolicy;

    PortableServer::POA_var libraryPOA =
        rootPOA -> create_POA("libraryPOA", manager, pl);

    //
    // Create some books.
    //
    BookInfoSeq bookInfos(4);
    bookInfos.length(4);
    
    bookInfos[0].title = (const char*) "Advanced CORBA Programming with C++";
    bookInfos[0].author = (const char*) "Michi Henning, Steve Vinoski";
    bookInfos[0].isbn = (const char*) "0-201-37927-9";
    bookInfos[1].title = (const char*) "The C++ Programming Language";
    bookInfos[1].author = (const char*) "Bjarne Stroustrup";
    bookInfos[1].isbn = (const char*) "0-201-88954-4";
    bookInfos[2].title = (const char*) "JAVA Language Reference";
    bookInfos[2].author = (const char*) "Mark Grand";
    bookInfos[2].isbn = (const char*) "1-56592-326-X";
    bookInfos[3].title = (const char*) "Design Patterns";
    bookInfos[3].author = (const char*) "Gamma, Helm, Johnson, Vlissides";
    bookInfos[3].isbn = (const char*) "0-201-63361-2";

    //
    // Create the library servant.
    //
    Library_impl* libraryImpl = new Library_impl(libraryPOA, bookInfos);
    PortableServer::ServantBase_var libraryServant = libraryImpl;

    //
    // Get the POACurrent object.
    //
    CORBA::Object_var obj = orb -> resolve_initial_references("POACurrent");
    PortableServer::Current_var current =
	PortableServer::Current::_narrow(obj);

    //
    // Create book servant.
    //
    Book_impl* bookImpl = new Book_impl(libraryPOA, current, bookInfos);
    PortableServer::ServantBase_var bookServant = bookImpl;

    //
    // Create and activate the library locator.
    //
    LibraryLocator_impl* locatorImpl =
	new LibraryLocator_impl(rootPOA, libraryImpl, bookImpl);
    PortableServer::ServantBase_var locatorServant = locatorImpl;
    rootPOA -> activate_object(locatorImpl);

    //
    // Register the library locator.
    //
    PortableServer::ServantLocator_var locator = locatorImpl -> _this();
    libraryPOA -> set_servant_manager(locator);

    //
    // Create a reference for the library object.
    //
    PortableServer::ObjectId_var oid =
	PortableServer::string_to_ObjectId("Library");
    CORBA::Object_var libraryObj =
	libraryPOA -> create_reference_with_id(oid.in(), "IDL:Library:1.0");

    //
    // Write IOR of the library object to a file.
    //
    ofstream out("Library.ref");
    if(out.fail())
    {
        cerr << "Can't open `" << "Library.ref" << "': "
             << strerror(errno) << endl;
        return EXIT_FAILURE;
    }
    CORBA::String_var s = orb -> object_to_string(libraryObj);
    out << s << endl;
    out.close();

    //
    // Activate the POA manager.
    //
    manager -> activate();

    //
    // Start event loop.
    //
    orb -> run();

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	//
	// Initialize the server to be load balanced.
	//
	LoadBalancing::LB_init();

	//
	// Initialize the ORB
	//
	orb = CORBA::ORB_init(argc, argv);

	//
	// Start the server
	//
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
