// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/BootManager.h>
#include <OB/Properties.h>

#include <BalancerServer.h>

using namespace CORBA;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options]\n"
        "\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show Load Balancer version.\n"
	"-i, --ior              Print IOR on standard output.\n"
	"-a, --administrative   Run in administrative mode.\n"
	"-d, --database         Set database directory path.\n"
	"-A, --admin-endpoint   Set the admin endpoint information.\n"
	"-F, --forward-endpoint Set the forward endpoint information.\n";
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    OB::Logger_var logger = oborb -> logger();
    OB::Properties_var props = oborb -> properties();

    //
    // Scan arguments.
    //
    bool printIOR = false;
    for(int arg = 1 ; arg < argc && *argv[arg] == '-' ; arg++)
    {
	if(strcmp(argv[arg], "--ior") == 0 ||
		  strcmp(argv[arg], "-i") == 0)
	{
	    printIOR = true;
	}
	else if(strcmp(argv[arg], "-a") == 0 ||
		strcmp(argv[arg], "--administrative") == 0)
	{
	    props -> setProperty("ooc.balancer.administrative", "true");
	}
	else if(strcmp(argv[arg], "-d") == 0 ||
		strcmp(argv[arg], "--database") == 0)
	{
	    ++arg;
	    if(arg >= argc)
	    {
		OB_CERR << argv[0] << ": " << argv[arg-1]
		     << " expects argument" << OB_ENDL;
		usage(argv[0]);
		return EXIT_FAILURE;
	    }
	    props -> setProperty("ooc.balancer.dbdir", argv[arg]);
	}
	else if(strcmp(argv[arg], "-A") == 0 ||
		strcmp(argv[arg], "--admin-endpoint") == 0)
	{
	    ++arg;
	    if(arg >= argc)
	    {
		OB_CERR << argv[0] << ": " << argv[arg-1]
		     << " expects argument" << OB_ENDL;
		usage(argv[0]);
		return EXIT_FAILURE;
	    }
	    props -> setProperty("ooc.balancer.admin_endpoint", argv[arg]);
	}
	else if(strcmp(argv[arg], "-F") == 0 ||
		strcmp(argv[arg], "--forward-endpoint") == 0)
	{
	    ++arg;
	    if(arg >= argc)
	    {
		OB_CERR << argv[0] << ": " << argv[arg-1]
		     << " expects argument" << OB_ENDL;
		usage(argv[0]);
		return EXIT_FAILURE;
	    }
	    props -> setProperty("ooc.balancer.forward_endpoint", argv[arg]);
	}
    }

    //
    // Set the port for the initial POA Manager.  If the
    // ooc.iiop.port has been set by the user print a message
    // indicating it will be ignored.
    //
    // XXX - Should warning be printed for ooc.orb.oa.endpoint?
    //
    const char* value = props -> getProperty("ooc.balancer.admin_endpoint");
    if(value != 0)
    {
	String_var endpoint(value);
	value = props -> getProperty("ooc.iiop.port");
	if(value != 0)
	    logger -> warning("Ignoring ooc.iiop.port setting");

	props -> setProperty("ooc.orb.oa.endpoint", endpoint);
    }

    //
    // Resolve Root POA.
    //
    CORBA::Object_var poaObj =
        orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var root = PortableServer::POA::_narrow(poaObj);
    PortableServer::POAManager_var manager = root -> the_POAManager();

    //
    // Resolve the Boot Manager.
    //
    CORBA::Object_var bmgrObj =
        orb -> resolve_initial_references("BootManager");
    OB::BootManager_var bootManager = OB::BootManager::_narrow(bmgrObj);

    //
    // Activate the POA Manager.  This must be done before creating the
    // Balancer.
    //
    manager -> activate();

    //
    // Create the Balancer.
    //
    LoadBalancing::BalancerServer balancer(orb);

    //
    // Add the GroupFactory bootstrap reference.
    //
    LoadBalancing::GroupFactory_var factory = balancer.groupFactory();
    PortableServer::ObjectId_var oid =
	PortableServer::string_to_ObjectId("Balancer");
    bootManager -> add_binding(oid.in(), factory);

    //
    // Print IOR on standard output.
    //
    if(printIOR)
    {
	String_var s = orb -> object_to_string(factory);
	OB_COUT << s << OB_ENDL;
    }

    //
    // Run the ORB.
    //
    orb -> run();

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
        //
        // Create a set of ORB level custom properties.
        //
        OB::Properties_var dflt = OB::Properties::getDefaultProperties();
	OB::Properties_var props = new OB::Properties(dflt);
        OB::ParseArgs(argc, argv, props);

    	//
    	// Scan arguments for help or version.
    	//
    	for(int arg = 1 ; arg < argc && *argv[arg] == '-' ; arg++)
    	{
            if(strcmp(argv[arg], "--help") == 0 || 
	       strcmp(argv[arg], "-h") == 0)
            {
                usage(argv[0]);
                return EXIT_SUCCESS;
            }
            else if(strcmp(argv[arg], "--version") == 0 || 
                    strcmp(argv[arg], "-v") == 0)
            {
                OB_CERR << "ORBacus Load Balancer " 
    	                << OB::Version << OB_ENDL;
                return EXIT_SUCCESS;
            }
	}

	//
	// Set the concurrency model.
	//
        props -> setProperty("ooc.orb.conc_model", "threaded");
        props -> setProperty("ooc.orb.oa.conc_model", "thread_per_client");

	orb = OBCORBA::ORB_init(argc, argv, props);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
