// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/BootManager.h>

#include <BalancerLogger.h>
#include <BalancerDatabase.h>
#include <StrategyFactory.h>
#include <GroupFactory_impl.h>
#include <ReplicaLocator.h>
#include <BalancerServer.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// BalancerServer constructor/destructor
// ----------------------------------------------------------------------

LoadBalancing::BalancerServer::BalancerServer(CORBA::ORB_ptr orb)
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);

    OB::Logger_var messageLogger = oborb -> logger();
    OB::Properties_var props = oborb -> properties();

    //
    // Create the balancer logger.
    //
    Logger_var logger = new Logger(messageLogger, props);

    //
    // Create the balancer database.
    //
    Database_var db = new Database(orb, props, logger);

    //
    // Check properties.
    //
    bool administrative = false;
    OB::Properties::KeySeq keys = props -> getKeys("ooc.balancer.");
    {
	for(ULong i = 0; i < keys.length(); ++i)
	{
	    const char* key = keys[i];
	    const char* value = props -> getProperty(key);
	    assert(value != 0);
	    
	    if(strcmp(key, "ooc.balancer.administrative") == 0)
	    {
		//
		// Should we run in administrative mode?
		//
		if(strcmp(value, "true") == 0 || strcmp(value, "TRUE") == 0)
		{
		    administrative = true;
		}
		else if(strcmp(value, "false") == 0 ||
			strcmp(value, "FALSE") == 0)
		{
		    administrative = false;
		}
		else
		{
		    String_var msg = string_dup("invalid value for `");
		    msg += key;
		    msg += "' -- starting in non-administrative mode";

		    logger -> warning(msg);

		    administrative = false;
		}
	    }
	    else if(strcmp(key, "ooc.balancer.dbdir") == 0)
	    {
		//
		// Used by the database.
		//
	    }
	    else if(strcmp(key, "ooc.balancer.trace.lifecycle") == 0 ||
		    strcmp(key, "ooc.balancer.trace.load_balance") == 0 ||
		    strcmp(key, "ooc.balancer.trace.database") == 0)
	    {
		//
		// Used by the Logger.
		//
	    }
	    else if(strcmp(key, "ooc.balancer.admin_endpoint") == 0)
	    {
		//
		// Ignore here.
		//
	    }
	    else if(strcmp(key, "ooc.balancer.forward_endpoint") == 0)
	    {
		//
		// Set the endpoint info for the forward POA.
		//
		props -> setProperty("ooc.orb.poamanager.locator.endpoint",
				     value);
	    }
	    else if(strncmp(key, "ooc.balancer.util", 17) == 0)
	    {
		//
		// Used by the utilities.
		//
	    }
	    else
	    {
		String_var msg = string_dup("`");
		msg += key;
		msg += "': unknown property";

		logger -> warning(msg);
	    }
	}
    }

    //
    // Resolve Root POA.
    //
    CORBA::Object_var poaObj = orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

    //
    // Create policies for a persistent POA with SYSTEM_ID.
    //
    CORBA::PolicyList pl(2);
    pl.length(2);
    pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
    pl[1] = rootPOA -> create_id_assignment_policy(PortableServer::SYSTEM_ID);

    //
    // Create the strategy POA.
    //
    PortableServer::POA_var strategyPOA =
	rootPOA -> create_POA("strategy", manager, pl);

    //
    // Create policies for a persistent POA with USER_ID.
    //
    pl[1] = rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);

    //
    // Create the factory POA.
    //
    PortableServer::POA_var factoryPOA =
	rootPOA -> create_POA("group_factory", manager, pl);

    //
    // Create the group POA.
    //
    PortableServer::POA_var groupPOA =
	rootPOA -> create_POA("group", manager, pl);

    //
    // Create policies for servant locator POA.
    //
    pl.length(4);
    pl[2] = rootPOA -> create_request_processing_policy(
	PortableServer::USE_SERVANT_MANAGER);
    pl[3] = rootPOA -> create_servant_retention_policy(
	PortableServer::NON_RETAIN);

    //
    // Create the locator POA Manager.
    //
    Object_var facObj = orb -> resolve_initial_references("POAManagerFactory");
    OBPortableServer::POAManagerFactory_var poaMgrFactory =
	OBPortableServer::POAManagerFactory::_narrow(facObj);
    PortableServer::POAManager_var locatorMgr =
	poaMgrFactory -> create_poa_manager("locator");

    //
    // Create the locator POA.
    //
    PortableServer::POA_var locatorPOA =
	rootPOA -> create_POA("locator", locatorMgr, pl);

    //
    // Get the object reference template for the locator POA.
    //
    OBPortableServer::POA_var obLocatorPOA =
	OBPortableServer::POA::_narrow(locatorPOA);
    assert(!is_nil(obLocatorPOA));
    PortableInterceptor::ObjectReferenceTemplate_var locatorORT =
        PortableInterceptor::ObjectReferenceTemplate::_downcast(
	    obLocatorPOA -> current_factory());
    if(locatorORT.in() == 0)
	locatorORT = obLocatorPOA -> adapter_template();

    //
    // Create the StrategyFactory.
    //
    StrategyFactory_var strategyFactory =
	new StrategyFactory(strategyPOA, logger, db);

    //
    // Create and activate the GroupFactory.
    //
    factoryImpl_ = new GroupFactory_impl(orb, factoryPOA, groupPOA,
					 strategyPOA, logger, db,
					 strategyFactory, locatorORT,
					 administrative);
    PortableServer::ObjectId_var factoryOid =
	PortableServer::string_to_ObjectId("GroupFactory");
    factoryPOA -> activate_object_with_id(factoryOid.in(), factoryImpl_);

    //
    // Create and activate and the ReplicaLocator.
    //
    GroupFactory_var factory = factoryImpl_ -> _this();
    locatorImpl_ = new ReplicaLocator(rootPOA, factory);
    PortableServer::ServantLocator_var locator = locatorImpl_ -> _this();

    //
    // Register the ReplicaLocator.
    //
    locatorPOA -> set_servant_manager(locator);

    //
    // Activate the locator POA Manager.
    //
    locatorMgr -> activate();
}

LoadBalancing::BalancerServer::~BalancerServer()
{
    locatorImpl_ -> _remove_ref();
    factoryImpl_ -> _remove_ref();
}

// ----------------------------------------------------------------------
// BalancerServer public member implementation
// ----------------------------------------------------------------------

LoadBalancing::GroupFactory_ptr
LoadBalancing::BalancerServer::groupFactory()
{
    return factoryImpl_ -> _this();
}
