// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Balancer.h>
#include <OB/Net.h> // For ntohl()

#include <BalancerInternal.h>

#include <ReplicaLocator.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// ReplicaLocator private member implementation
// ----------------------------------------------------------------------

bool
LoadBalancing::ReplicaLocator::extractInfo(
    const PortableServer::ObjectId& oid,
    GroupId_out groupId,
    String_out repositoryId,
    PortableInterceptor::ObjectId_out realOid)
{
    ULong remainingBytes = oid.length();
    const Octet* ptr = oid.get_buffer();

    //
    // Extract the number of bytes that make up the group id.
    //
    if(remainingBytes < 4)
	return false;
    Long tmp = 0;
    memcpy(&tmp, ptr, 4);
    ULong groupIdLen = (ULong)ntohl(tmp);
    remainingBytes -= 4;
    ptr += 4;

    //
    // Extract the group id.
    //
    if(remainingBytes < groupIdLen)
	return false;
    PortableServer::ObjectId groupIdBytes(groupIdLen);
    groupIdBytes.length(groupIdLen);
    memcpy(groupIdBytes.get_buffer(), ptr, groupIdLen);
    try
    {
	groupId = PortableServer::ObjectId_to_string(groupIdBytes);
    }
    catch(const BAD_PARAM&)
    {
	return false;
    }
    remainingBytes -= groupIdLen;
    ptr += groupIdLen;
    
    //
    // Extract the number of bytes that make up the repository id.
    //
    if(remainingBytes < 4)
	return false;
    tmp = 0;
    memcpy(&tmp, ptr, 4);
    ULong repositoryIdLen = (ULong)ntohl(tmp);
    remainingBytes -= 4;
    ptr += 4;

    //
    // Extract the repository id.
    //
    if(remainingBytes < repositoryIdLen)
	return false;
    PortableServer::ObjectId repositoryIdBytes(repositoryIdLen);
    repositoryIdBytes.length(repositoryIdLen);
    memcpy(repositoryIdBytes.get_buffer(), ptr, repositoryIdLen);
    try
    {
	repositoryId =
	    PortableServer::ObjectId_to_string(repositoryIdBytes);
    }
    catch(const BAD_PARAM&)
    {
	return false;
    }
    remainingBytes -= repositoryIdLen;
    ptr += repositoryIdLen;

    //
    // Check the repository id.
    //
    char* p = strchr(repositoryId.ptr(), ':');
    if(p == 0 || p - repositoryId.ptr() == 0)
	return false;
    
    //
    // Extract the real object id.
    //
    realOid = new PortableInterceptor::ObjectId(remainingBytes);
    realOid -> length(remainingBytes);
    memcpy(realOid -> get_buffer(), ptr, remainingBytes);

    return true;
}
 

// ----------------------------------------------------------------------
// ReplicaLocator constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::ReplicaLocator::ReplicaLocator(PortableServer::POA_ptr poa,
					      GroupFactory_ptr factory)
    : poa_(PortableServer::POA::_duplicate(poa)),
      factory_(GroupFactory::_duplicate(factory))
{
#ifdef LB_DEBUG
    OB_CERR << "ReplicaLocator::ReplicaLocator()" << OB_ENDL;
#endif
}

LoadBalancing::ReplicaLocator::~ReplicaLocator()
{
#ifdef LB_DEBUG
    OB_CERR << "ReplicaLocator::~ReplicaLocator()" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// ReplicaLocator public member implementation
// ----------------------------------------------------------------------

PortableServer::POA_ptr
LoadBalancing::ReplicaLocator::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

PortableServer::Servant
LoadBalancing::ReplicaLocator::preinvoke(
    const PortableServer::ObjectId& oid, PortableServer::POA_ptr /* poa */,
    const char* /* operation */,
    PortableServer::ServantLocator::Cookie& /* the_cookie */)
    throw(PortableServer::ForwardRequest, CORBA::SystemException)
{
    //
    // Extract the replica info.
    //
    GroupId_var groupId;
    String_var repositoryId;
    PortableInterceptor::ObjectId_var realOid;
    if(!extractInfo(oid, groupId, repositoryId, realOid))
	throw OBJECT_NOT_EXIST();
    
    try
    {
	//
	// Get the group.
	//
	Group_var grp = factory_ -> get(groupId);
	InternalGroup_var group = InternalGroup::_narrow(grp);

	//
	// Get the replica of the least loaded member.
	//
	Object_var obj = group -> get_next(repositoryId, realOid);

	if(!is_nil(obj))
	    throw PortableServer::ForwardRequest(obj);
    }
    catch(const GroupNotFound&)
    {
    }
    catch(const MemberNotFound&)
    {
    }

    throw OBJECT_NOT_EXIST();
    return 0; // Some compilers need this
}

void
LoadBalancing::ReplicaLocator::postinvoke(
    const PortableServer::ObjectId& /*oid*/, PortableServer::POA_ptr /*poa*/,
    const char* /*operation*/,
    PortableServer::ServantLocator::Cookie /*the_cookie*/,
    PortableServer::Servant /*the_servant*/)
    throw(CORBA::SystemException)
{
    //
    // Nothing to do here
    //
}

