// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/BalancerProperties.h>

#include <BalancerDatabase.h>
#include <StrategyMinDisp_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// StrategyMinDisp_impl private member implementation
// ----------------------------------------------------------------------

CORBA::Float
LoadBalancing::StrategyMinDisp_impl::loadThreshold()
{
    //
    // If there are no elements below the average (minus the threshold),
    // then elements with average loads (plus or minus the threshold)
    // will be chosen.
    //
    Float ave = loadSum() / memberCount();
    Float threshold = (minLoad() < ave - tolerance_)
	? ave - tolerance_ : ave + tolerance_;
    assert(minLoad() <= threshold);

    return threshold;
}

bool
LoadBalancing::StrategyMinDisp_impl::requireAlert(CORBA::Float load)
{
    if(load > loadSum() / memberCount() + tolerance_)
	return true;

    return false;
}

// ----------------------------------------------------------------------
// StrategyMinDisp_impl constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::StrategyMinDisp_impl::StrategyMinDisp_impl(
    PortableServer::POA_ptr poa,
    Database_ptr db,
    const PropertySeq& properties)
    : StrategyAdaptiveBase(poa, db)
{
#ifdef LB_DEBUG
    OB_CERR << "StrategyMinDisp_impl::StrategyMinDisp_impl()" << OB_ENDL;
#endif

    //
    // Set the base properties.
    //
    PropertySeq remaining;
    PropertyErrorSeq errors;
    setProperties(properties, remaining, errors);

    //
    // Set the properties specific to the min-dispersion strategy
    // (currently none).
    //
    for(ULong i = 0 ; i < remaining.length() ; i++)
    {
	ULong len = errors.length();
	errors.length(len + 1);
	errors[len].name = remaining[i].name;
	errors[len].code = BAD_PROPERTY;
    }

    if(errors.length() > 0)
    {
	_OB_setRef(0);
	throw InvalidProperties(errors);
    }

#ifdef LB_DEBUG
    OB_CERR << "strategy properties:" << OB_ENDL
	    << "    tolerance = " << tolerance_ << OB_ENDL
	    << "    reject-load = " << rejectLoad_ << OB_ENDL
	    << "    load-per-client = " << loadPerClient_ << OB_ENDL
	    << "    dampening-multiplier = " << dampeningMultiplier_ << OB_ENDL;
#endif
}

LoadBalancing::StrategyMinDisp_impl::~StrategyMinDisp_impl()
{
#ifdef LB_DEBUG
    OB_CERR << "StrategyMinDisp_impl::~StrategyMinDisp_impl()" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// StrategyMinDisp_impl public member implementation
// ----------------------------------------------------------------------

//
// IDL:ooc.com/LoadBalancing/Strategy/get_name:1.0
//
char*
LoadBalancing::StrategyMinDisp_impl::get_name()
    throw(CORBA::SystemException)
{
    return string_dup("min-dispersion");
}

//
// IDL:ooc.com/LoadBalancing/Strategy/get_properties:1.0
//
LoadBalancing::PropertySeq*
LoadBalancing::StrategyMinDisp_impl::get_properties()
    throw(CORBA::SystemException)
{
    //
    // Add the base properties.
    //
    PropertySeq_var props = getProperties();

    return props._retn();
}

//
// IDL:ooc.com/LoadBalancing/Strategy/adjust:1.0
//
void
LoadBalancing::StrategyMinDisp_impl::adjust(
    const LoadBalancing::MemberDataSeq& members)
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    adjustMembers(members);
}

//
// IDL:ooc.com/LoadBalancing/Strategy/get_next:1.0
//
LoadBalancing::MemberId
LoadBalancing::StrategyMinDisp_impl::get_next()
    throw(LoadBalancing::MemberNotFound,
	  CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    return getNext();
}

//
// IDL:ooc.com/LoadBalancing/Strategy/push_load:1.0
//
void
LoadBalancing::StrategyMinDisp_impl::push_load(const char* member_id,
					       CORBA::ULong load)
    throw(LoadBalancing::MemberNotFound, CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    pushLoad(member_id, load);
}

