// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Balancer.h>
#include <OB/BalancerProperties.h>

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#   include <stdlib.h>
#endif

using namespace CORBA;

enum Command
{
    LIST_GROUPS,
    CREATE_GROUP,
    DESTROY_GROUP,

    GET_GROUP_INFO,
    GET_GROUP_IOR,

    SET_STRATEGY,
    SET_CUSTOM_STRATEGY,

    LIST_MEMBERS,
    ADD_MEMBER,
    REMOVE_MEMBER,

    SHUTDOWN
};

static void
usage(const char* progName)
{
    OB_CERR << "Usage: " << progName << " command\n"
"\n"
"Commands:\n"
"-h, --help                             Display this information.\n"

"--list-groups                          List the load balanced groups.\n"
"--create-group GROUP                   Create a load balanced group.\n"
"--destroy-group GROUP                  Destroy a load balanced group.\n"

"--get-group-info GROUP                 Get the attributes of a group.\n"
"--get-group-ior GROUP REPOID OBJECTID  Get the IOR for use by a client.\n"

"--set-strategy GROUP <STRATEGY>        Use the specified built-in strategy.\n"
"--set-custom-strategy GROUP IOR        Use the given custom strategy.\n"

"--list-members GROUP                   Enumerate the members.\n"
"--add-member GROUP MEMBER              Add a member.\n"
"--remove-member GROUP MEMBER           Remove a member.\n"

"--shutdown                             Shutdown the Balancer.\n"
"\n"
"Where <STRATEGY> is one of \"round-robin\", \"random\", \"least-load\"\n"
"or \"min-dispersion\".\n"
"\n"
"The \"least-load\" strategy has the options:\n"
"--tolerance TOLERANCE\n"
"--reject-load REJECT_LOAD\n"
"--critical-load CRITICAL_LOAD\n"
"--load-per-client LOAD_PER_CLIENT\n"
"--dampening-multiplier DAMPENING_MULTIPLIER\n"
"\n"
"The \"min-dispersion\" strategy has the options:\n"
"--tolerance TOLERANCE\n"
"--reject-load REJECT_LOAD\n"
"--load-per-client LOAD_PER_CLIENT\n"
"--dampening-multiplier DAMPENING_MULTIPLIER\n"
"\n"
;
}

static void
printException(UserException& uex, const char* name = "")
{
    LoadBalancing::MemberExists* memberExists =
        LoadBalancing::MemberExists::_downcast(&uex);
    if(memberExists)
    {
	OB_CERR << "Member `" << name << "' already exists" << OB_ENDL;
        exit(1);
    }

    LoadBalancing::MemberNotFound* memberNotFound =
        LoadBalancing::MemberNotFound::_downcast(&uex);
    if(memberNotFound)
    {
	OB_CERR << "Member `" << name << "' does not exist" << OB_ENDL;
        exit(1);
    }

    LoadBalancing::StrategyNotFound* strategyNotFound =
        LoadBalancing::StrategyNotFound::_downcast(&uex);
    if(strategyNotFound)
    {
	OB_CERR << "Cannot create Strategy `" << name << "'" << OB_ENDL;
        exit(1);
    }

    LoadBalancing::StrategyNotAdaptive* strategyNotAdaptive =
        LoadBalancing::StrategyNotAdaptive::_downcast(&uex);
    if(strategyNotAdaptive)
    {
	OB_CERR << "Strategy `" << name
		<< "' is not an adaptive strategy" << OB_ENDL;
        exit(1);
    }

    LoadBalancing::InvalidProperties* invalidProperties =
        LoadBalancing::InvalidProperties::_downcast(&uex);
    if(invalidProperties)
    {
	OB_CERR << "Invalid properties for Strategy `" << name
		<< "'" << OB_ENDL;
        exit(1);
    }

    LoadBalancing::GroupExists* groupExists =
        LoadBalancing::GroupExists::_downcast(&uex);
    if(groupExists)
    {
	OB_CERR << "Group `" << name << "' already exists" << OB_ENDL;
        exit(1);
    }

    LoadBalancing::GroupNotFound* groupNotFound =
        LoadBalancing::GroupNotFound::_downcast(&uex);
    if(groupNotFound)
    {
	OB_CERR << "Group `" << name << "' does not exist" << OB_ENDL;
        exit(1);
    }

    OB_CERR << "Unknown user exception" << OB_ENDL;
    exit(1);
}

static Object_ptr
stringToObject(ORB_ptr orb, const char* str)
{
    Object_var obj;

    try
    {
        obj = orb -> string_to_object(str);
    }
    catch(const BAD_PARAM&)
    {
        OB_CERR << "Invalid object reference" << OB_ENDL;
        exit(1);
    }

    return obj._retn();
}

static LoadBalancing::Strategy_ptr
stringToStrategy(ORB_ptr orb, const char* str)
{
    Object_var obj = stringToObject(orb, str);
    LoadBalancing::Strategy_var strategy;

    try
    {
        strategy = LoadBalancing::Strategy::_narrow(obj);
        if(is_nil(strategy))
        {
            OB_CERR << "IOR does not denote a LoadBalancing::Strategy"
		    << OB_ENDL;
            exit(1);
        }
    }
    catch(const SystemException& ex)
    {
        OB_CERR << "A system exception occurred while narrowing the "
		<< "given object reference" << OB_ENDL
		<< "to LoadBalancing::Strategy" << OB_ENDL;
        OB_CERR << OB_ENDL;
        OB_CERR << ex << OB_ENDL;
        exit(1);
    }

    return strategy._retn();
}

static void
listGroups(ORB_ptr /* orb */, LoadBalancing::GroupFactory_ptr factory,
	   int argc, char* argv[])
{
    if(argc != 2)
    {
        usage(argv[0]);
        exit(1);
    }

    LoadBalancing::GroupIdSeq_var groups = factory -> list();
    for(ULong i = 0 ; i < groups -> length() ; i++)
	OB_COUT << groups[i] << OB_ENDL;
}

static void
createGroup(ORB_ptr /* orb */, LoadBalancing::GroupFactory_ptr factory,
	    int argc, char* argv[])
{
    if(argc != 3)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* id = argv[2];
    try
    {
	LoadBalancing::Group_var group = factory -> create(id);
    }
    catch(UserException& ex)
    {
        printException(ex, id);
    }
}

static void
destroyGroup(ORB_ptr /* orb */, LoadBalancing::GroupFactory_ptr factory,
	     int argc, char* argv[])
{
    if(argc != 3)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* id = argv[2];
    try
    {
	LoadBalancing::Group_var group = factory -> get(id);
	group -> destroy();
    }
    catch(UserException& ex)
    {
        printException(ex, id);
    }
}

static void
getGroupInfo(ORB_ptr /* orb */, LoadBalancing::GroupFactory_ptr factory,
	     int argc, char* argv[])
{
    if(argc != 3)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* id = argv[2];
    try
    {
	LoadBalancing::Group_var group = factory -> get(id);
	LoadBalancing::StrategyProxy_var proxy =
	    group -> get_strategy_proxy();
	LoadBalancing::MemberIdSeq_var members = group -> list_members();
	OB_COUT << "Name:          " << String_var(group -> get_id())
		<< OB_ENDL;
	OB_COUT << "Strategy:     " << String_var(proxy -> get_name())
		<< OB_ENDL;
	OB_COUT << "Member Count:  " << members -> length()
		<< OB_ENDL;
    }
    catch(UserException& ex)
    {
        printException(ex, id);
    }
}

static void
getGroupIor(ORB_ptr orb, LoadBalancing::GroupFactory_ptr factory,
	    int argc, char* argv[])
{
    if(argc != 5)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* id = argv[2];
    const char* repoid = argv[3];
    const char* oidstr = argv[4];
    try
    {
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId(oidstr);
	PortableInterceptor::ObjectId_var pioid =
	    new PortableInterceptor::ObjectId(oid -> length(),
					      oid -> length(),
					      oid -> get_buffer());
	LoadBalancing::Group_var group = factory -> get(id);
	Object_var obj = group -> get_ior(repoid, pioid);
	String_var s = orb -> object_to_string(obj);
	OB_COUT << s << OB_ENDL;
    }
    catch(UserException& ex)
    {
        printException(ex, id);
    }
}

static void
setStrategy(ORB_ptr /* orb */, LoadBalancing::GroupFactory_ptr factory,
	    int argc, char* argv[])
{
    if(argc < 4)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* id = argv[2];
    const char* strategy = argv[3];
    LoadBalancing::PropertySeq props;

    if(strcmp(strategy, "least-load") == 0 ||
       strcmp(strategy, "min-dispersion") == 0)
    {
	for(int i = 4 ; i < argc ; i++)
	{
	    if(strcmp(argv[i], "--reject-load") == 0)
	    {
		if(++i == argc)
		{
		    usage(argv[0]);
		    exit(1);
		}

		CORBA::ULong len = props.length();
		props.length(len + 1);

		CORBA::ULong rejectLoad = atoi(argv[i]);
		props[len].name = CORBA::string_dup(
				LoadBalancing::RejectLoadPropertyName);
		props[len].value <<= rejectLoad;
	    }
	    else if(strcmp(argv[i], "--load-per-client") == 0)
	    {
		if(++i == argc)
		{
		    usage(argv[0]);
		    exit(1);
		}

		CORBA::ULong len = props.length();
		props.length(len + 1);

		CORBA::ULong loadPerClient = atoi(argv[i]);
		props[len].name = CORBA::string_dup(
				LoadBalancing::LoadPerClientPropertyName);
		props[len].value <<= loadPerClient;
	    }
	    else if(strcmp(argv[i], "--dampening-multiplier") == 0)
	    {
		if(++i == argc)
		{
		    usage(argv[0]);
		    exit(1);
		}

		CORBA::ULong len = props.length();
		props.length(len + 1);

		CORBA::Float dampeningMultiplier = atof(argv[i]);
		props[len].name = CORBA::string_dup(
				LoadBalancing::DampeningMultiplierPropertyName);
		props[len].value <<= dampeningMultiplier;
	    }
	    else if(strcmp(argv[i], "--tolerance") == 0)
	    {
		if(++i == argc)
		{
		    usage(argv[0]);
		    exit(1);
		}

		CORBA::ULong len = props.length();
		props.length(len + 1);

		CORBA::ULong tolerance = atoi(argv[i]);
		props[len].name = CORBA::string_dup(
				LoadBalancing::TolerancePropertyName);
		props[len].value <<= tolerance;
	    }
	    else if(strcmp(argv[i], "--critical-load") == 0)
	    {
		if(++i == argc || strcmp(strategy, "least-load") != 0)
		{
		    usage(argv[0]);
		    exit(1);
		}

		CORBA::ULong len = props.length();
		props.length(len + 1);

		CORBA::ULong criticalLoad = atoi(argv[i]);
		props[len].name = CORBA::string_dup(
				LoadBalancing::CriticalLoadPropertyName);
		props[len].value <<= criticalLoad;
	    }
	    else
	    {
		usage(argv[0]);
		exit(1);
	    }
	}
    }
    else if(strcmp(strategy, "random") != 0 &&
	    strcmp(strategy, "round-robin") != 0)
    {   
	usage(argv[0]);
	exit(1);
    }

    LoadBalancing::Group_var group;
    try
    {
	group = factory -> get(id);
    }
    catch(UserException& ex)
    {
        printException(ex, id);
    }
    try
    {
	group -> set_strategy(strategy, props);
    }
    catch(UserException& ex)
    {
        printException(ex, strategy);
    }
}

static void
setCustomStrategy(ORB_ptr orb, LoadBalancing::GroupFactory_ptr factory,
		  int argc, char* argv[])
{
    if(argc != 4)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* id = argv[2];
    const char* iorString = argv[3];
    try
    {
	LoadBalancing::Group_var group = factory -> get(id);
	LoadBalancing::Strategy_var strategy = stringToStrategy(orb, iorString);
	group -> set_custom_strategy(strategy);
    }
    catch(UserException& ex)
    {
        printException(ex, id);
    }
}

static void
listMembers(ORB_ptr /* orb */, LoadBalancing::GroupFactory_ptr factory,
	    int argc, char* argv[])
{
    if(argc != 3)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* id = argv[2];
    try
    {
	LoadBalancing::Group_var group = factory -> get(id);
	LoadBalancing::MemberIdSeq_var members = group -> list_members();
	for(ULong i = 0 ; i < members -> length() ; i++)
	    OB_COUT << members[i] << OB_ENDL;
    }
    catch(UserException& ex)
    {
        printException(ex, id);
    }
}

static void
addMember(ORB_ptr /* orb */, LoadBalancing::GroupFactory_ptr factory,
	  int argc, char* argv[])
{
    if(argc != 4)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* id = argv[2];
    const char* member = argv[3];
    LoadBalancing::Group_var group;
    try
    {
	group = factory -> get(id);
    }
    catch(UserException& ex)
    {
        printException(ex, id);
    }
    try
    {
	group -> add_member(member);
    }
    catch(UserException& ex)
    {
        printException(ex, member);
    }
}

static void
removeMember(ORB_ptr /* orb */, LoadBalancing::GroupFactory_ptr factory,
	     int argc, char* argv[])
{
    if(argc != 4)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* id = argv[2];
    const char* member = argv[3];
    LoadBalancing::Group_var group;
    try
    {
	group = factory -> get(id);
    }
    catch(UserException& ex)
    {
        printException(ex, id);
    }
    try
    {
	group -> remove_member(member);
    }
    catch(UserException& ex)
    {
        printException(ex, member);
    }
}

static void
shutdown(ORB_ptr /* orb */, LoadBalancing::GroupFactory_ptr factory,
	 int argc, char* argv[])
{
    if(argc != 2)
    {
        usage(argv[0]);
        exit(1);
    }

    factory -> shutdown();
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    if(argc == 1)
    {
        usage(argv[0]);
        return EXIT_FAILURE;
    }

    //
    // Get the command to execute
    //
    Command cmd;
    if(strcmp(argv[1], "--list-groups") == 0)
	cmd = LIST_GROUPS;
    else if(strcmp(argv[1], "--create-group") == 0)
	cmd = CREATE_GROUP;
    else if(strcmp(argv[1], "--destroy-group") == 0)
	cmd = DESTROY_GROUP;
    else if(strcmp(argv[1], "--get-group-info") == 0)
	cmd = GET_GROUP_INFO;
    else if(strcmp(argv[1], "--get-group-ior") == 0)
	cmd = GET_GROUP_IOR;
    else if(strcmp(argv[1], "--set-strategy") == 0)
	cmd = SET_STRATEGY;
    else if(strcmp(argv[1], "--set-custom-strategy") == 0)
	cmd = SET_CUSTOM_STRATEGY;
    else if(strcmp(argv[1], "--list-members") == 0)
	cmd = LIST_MEMBERS;
    else if(strcmp(argv[1], "--add-member") == 0)
	cmd = ADD_MEMBER;
    else if(strcmp(argv[1], "--remove-member") == 0)
	cmd = REMOVE_MEMBER;
    else if(strcmp(argv[1], "--shutdown") == 0)
	cmd = SHUTDOWN;
    else
    {
        usage(argv[0]);
        return EXIT_FAILURE;
    }

    //
    // Resolve the Balancer
    //
    Object_var obj;
    try
    {
        obj = orb -> resolve_initial_references("Balancer");
    }
    catch(ORB::InvalidName&)
    {
        OB_CERR << "Unable to resolve `Balancer' initial reference"
		<< OB_ENDL;
        return EXIT_FAILURE;
    }

    LoadBalancing::GroupFactory_var factory =
	LoadBalancing::GroupFactory::_narrow(obj);
    if(is_nil(factory))
    {
        OB_CERR << "`Balancer' initial reference is not a "
             << "LoadBalancing::GroupFactory object." << OB_ENDL;
        return EXIT_FAILURE;
    }

    //
    // Execute command
    //
    switch(cmd)
    {
    case LIST_GROUPS:
	listGroups(orb, factory, argc, argv);
	break;

    case CREATE_GROUP:
	createGroup(orb, factory, argc, argv);
	break;

    case DESTROY_GROUP:
	destroyGroup(orb, factory, argc, argv);
	break;

    case GET_GROUP_INFO:
	getGroupInfo(orb, factory, argc, argv);
	break;

    case GET_GROUP_IOR:
	getGroupIor(orb, factory, argc, argv);
	break;

    case SET_STRATEGY:
	setStrategy(orb, factory, argc, argv);
	break;

    case SET_CUSTOM_STRATEGY:
	setCustomStrategy(orb, factory, argc, argv);
	break;

    case LIST_MEMBERS:
	listMembers(orb, factory, argc, argv);
	break;

    case ADD_MEMBER:
	addMember(orb, factory, argc, argv);
	break;

    case REMOVE_MEMBER:
	removeMember(orb, factory, argc, argv);
	break;

    case SHUTDOWN:
	shutdown(orb, factory, argc, argv);
	break;

    default:
	assert(false);
    };

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
