// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OBORBInit.h>
#include <OB/Properties.h>
#include <OB/Balancer.h>
#include <OB/BalancerUtil.h>
#include <OB/BalancerUtil_init.h>

#include <UtilLogger.h>
#include <LoadAlert_impl.h>
#include <LoadCalculator_impl.h>
#include <LoadUpdater_impl.h>
#include <LoadAlertInterceptor.h>
#include <UtilInitializer.h>

using namespace CORBA;

//
// IDL:ooc.com/LoadBalancing:1.0
//
namespace LoadBalancing
{

//
// IDL:ooc.com/LoadBalancing/Util:1.0
//
namespace Util
{

//
// Server Request Interceptor which computes load based on the number of
// active requests.
//
class LoadInterceptor :
    public PortableInterceptor::ServerRequestInterceptor,
    public OBCORBA::RefCountLocalObject
{
    //
    // The LoadCalculator
    //
    LoadCalculator_impl* calculator_;

public:

    LoadInterceptor(LoadCalculator_impl* calc)
	: calculator_(calc)
    {
#ifdef LB_DEBUG
	OB_CERR << "LoadInterceptor::LoadInterceptor()" << OB_ENDL;
#endif

	calculator_ -> _OB_incRef();
    }

    ~LoadInterceptor()
    {
#ifdef LB_DEBUG
	OB_CERR << "LoadInterceptor::~LoadInterceptor()" << OB_ENDL;
#endif

	calculator_ -> _OB_decRef();
    }

    //
    // IDL to C++ Mapping
    //

    virtual char*
    name()
    {
	return CORBA::string_dup("LoadInterceptor");
    }

    virtual void
    destroy()
	throw(CORBA::SystemException)
    {
    }

    virtual void
    receive_request_service_contexts(
    	PortableInterceptor::ServerRequestInfo_ptr)
	throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
    {
	calculator_ -> newRequest();
    }

    virtual void
    receive_request(PortableInterceptor::ServerRequestInfo_ptr)
	throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
    {
    }
    
    virtual void
    send_reply(PortableInterceptor::ServerRequestInfo_ptr)
	throw(CORBA::SystemException)
    {
	calculator_ -> requestCompleted();
    }
    
    virtual void
    send_exception(PortableInterceptor::ServerRequestInfo_ptr)
	throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
    {
	calculator_ -> requestCompleted();
    }

    virtual void
    send_other(PortableInterceptor::ServerRequestInfo_ptr)
	throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
    {
	calculator_ -> requestCompleted();
    }
};

} // End of namespace Util

} // End of namespace LoadBalancing

// ----------------------------------------------------------------------
// UtilInitializer constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::Util::UtilInitializer::UtilInitializer()
    : updateFreq_(1000), // 1 second is the default
      alertExpire_(1000), // 1 second is the default
      createAlert_(true),
      createUpdater_(true),
      createCalculator_(true),
      initialized_(false)
{
#ifdef LB_DEBUG
    OB_CERR << "UtilInitializer::UtilInitializer()" << OB_ENDL;
#endif
}

LoadBalancing::Util::UtilInitializer::~UtilInitializer()
{
#ifdef LB_DEBUG
    OB_CERR << "UtilInitializer::~UtilInitializer()" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// UtilInitializer public member implementation
// ----------------------------------------------------------------------

void
LoadBalancing::Util::UtilInitializer::post_init(
    PortableInterceptor::ORBInitInfo_ptr info)
{
    JTCSynchronized sync(*this);

    if(!initialized_)
    {
	OBPortableInterceptor::ORBInitInfo_var obInfo =
	    OBPortableInterceptor::ORBInitInfo::_narrow(info);
	OBCORBA::ORB_var orb = OBCORBA::ORB::_narrow(ORB_var(obInfo -> orb()));
	OB::Logger_var messageLogger = orb -> logger();
	OB::Properties_var props = orb -> properties();

	//
	// Create the balancer logger.
	//
	logger_ = new Logger(messageLogger, props);

	//
	// Get the utility property values.
	//
	OB::Properties::KeySeq keys = props -> getKeys("ooc.balancer.util.");
	for(CORBA::ULong i = 0; i < keys.length(); ++i)
	{
	    const char* key = keys[i];
	    const char* value = props -> getProperty(key);
	
	    if(strcmp(key, "ooc.balancer.util.create_alert") == 0)
	    {
		createAlert_ =
		    strcmp(value, "true") == 0 || strcmp(value, "TRUE") == 0;
	    }
	    else if(strcmp(key, "ooc.balancer.util.create_updater") == 0)
	    {
		createUpdater_ =
		    strcmp(value, "true") == 0 || strcmp(value, "TRUE") == 0;
	    }
	    else if(strcmp(key, "ooc.balancer.util.create_calculator") == 0)
	    {
		createCalculator_ =
		    strcmp(value, "true") == 0 || strcmp(value, "TRUE") == 0;
	    }
	    else if(strcmp(key, "ooc.balancer.util.update_freq") == 0)
	    {
		updateFreq_ = atol(value);
	    }
	    else if(strcmp(key, "ooc.balancer.util.alert_expire") == 0)
	    {
		alertExpire_ = atol(value);
	    }
	    else if(strcmp(key, "ooc.balancer.util.trace_level") == 0)
	    {
		//
		// Used by the Logger.
		//
	    }
	    else
	    {
		String_var msg = string_dup("`");
		msg += key;
		msg += "': unknown property";

		logger_ -> warning(msg);
	    }
	}

	if(createAlert_)
	{
	    //
	    // Create the LoadAlertInterceptor.
	    //
	    loadAlertInterceptor_ = new LoadAlertInterceptor(orb, logger_,
							     alertExpire_);
	    loadAlertDummy_ = loadAlertInterceptor_;

	    //
	    // Create the LoadAlert object.
	    //
	    LoadAlert_var loadAlert = new LoadAlert_impl(loadAlertInterceptor_);

	    //
	    // Register the LoadAlert initial reference.
	    //
	    try
	    {
		info -> register_initial_reference("LoadAlert", loadAlert);
	    }
	    catch(const PortableInterceptor::ORBInitInfo::InvalidName&)
	    {
		assert(false); // Should not occur
	    }
	}

	LoadCalculator_impl* calc = 0;
	if(createCalculator_)
	{
	    //
	    // Create the LoadCalculator local object.
	    //
	    calc = new LoadCalculator_impl();
	    LoadCalculator_var dummyCalc = calc;

	    //
	    // Create the LoadInterceptor.
	    //
	    loadInterceptor_ = new LoadInterceptor(calc);
	    loadDummy_ = loadInterceptor_;

	    //
	    // Register the LoadCalculator initial reference.
	    //
	    try
	    {
		info -> register_initial_reference("LoadCalculator", calc);
	    }
	    catch(const PortableInterceptor::ORBInitInfo::InvalidName&)
	    {
		assert(false); // Should not occur
	    }
	}

	if(createUpdater_)
	{
	    //
	    // Get the GroupFactory.
	    //
	    LoadBalancing::GroupFactory_var factory;
	    try
	    {
		Object_var obj =
		    orb -> resolve_initial_references("Balancer");
		factory = LoadBalancing::GroupFactory::_narrow(obj);
	    }
	    catch(const ORB::InvalidName&)
	    {
	    }
	    catch(const SystemException&)
	    {
	    }
	    if(is_nil(factory))
	    {
		String_var msg =
		    string_dup("unable to locate the Balancer");
		throw INITIALIZE(msg);
	    }

	    //
	    // Create the LoadUpdater.
	    //
	    LoadUpdater_var updater =
		new LoadUpdater_impl(logger_, factory, updateFreq_);

	    //
	    // Register the LoadUpdater initial reference.
	    //
	    try
	    {
		info -> register_initial_reference("LoadUpdater", updater);
	    }
	    catch(const PortableInterceptor::ORBInitInfo::InvalidName&)
	    {
		assert(false); // Should not occur
	    }

	    //
	    // If the LoadCalculator was also created, then pass it to the
	    // LoadUpdater.
	    //
	    if(createCalculator_)
		updater -> set_load_calculator(calc);
	}

	initialized_ = true;
    }

    //
    // Register the LoadAlertInterceptor.
    //
    if(createAlert_)
    {
	try
	{
	    info -> add_server_request_interceptor(loadAlertInterceptor_);
	}
	catch(const PortableInterceptor::ORBInitInfo::DuplicateName&)
	{
	    assert(false); // Should not occur
	}
    }

    //
    // Register the LoadInterceptor.
    //
    if(createCalculator_)
    {
	try
	{
	    info -> add_server_request_interceptor(loadInterceptor_);
	}
	catch(const PortableInterceptor::ORBInitInfo::DuplicateName&)
	{
	    assert(false); // Should not occur
	}
    }
}

// ----------------------------------------------------------------------
// LBUtil_init() implementation (declared in OB/BalancerUtil_init.h)
// ----------------------------------------------------------------------

void
LoadBalancing::Util::LBUtil_init()
{
    //
    // Register the ORB initializer.
    //
    PortableInterceptor::ORBInitializer_var initializer =
	new UtilInitializer();
    PortableInterceptor::register_orb_initializer(initializer);
}
