// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <CustomStrategy_impl.h>
#include <MemberObject_impl.h>
#include <Tester_impl.h>

using namespace CORBA;
using namespace LoadBalancing;

//
// IDL:Tester/test_CustomStrategy:1.0
//
void
Tester_impl::test_CustomStrategy()
    throw(CORBA::SystemException)
{
    try
    {
	MemberIdSeq ids;

	PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(
	    Object_var(orb_ -> resolve_initial_references("RootPOA")));

	//
	// Create the oid for the replicas
	//
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("MemberObject");
	PortableInterceptor::ObjectId piOid(oid -> length(),
					    oid -> length(),
					    oid -> get_buffer());

	//
	// Create a new group
	//
	Group_var group = factory_ -> create("g1");

	//
	// Get the load balancing proxy
	//
	StrategyProxy_var proxy = group -> get_strategy_proxy();

	//
	// Get the ior that when narrowed will get the replica of the least
	// loaded member
	//
	Object_var obj = group -> get_ior("IDL:MemberObject:1.0", piOid);

	//
	// Add the member `m1'
	//
	group -> add_member("m1");

	//
	// Create the replica POA for member `m1'
	//
	PortableServer::POA_var p1 = createReplicaPOA("p1", "g1", "m1");

	//
	// Create the servant for the replica of member `m1'
	//
	MemberObject_impl* m1Impl = new MemberObject_impl(p1, "m1");
	PortableServer::ServantBase_var m1Servant = m1Impl;
	p1 -> activate_object_with_id(oid.in(), m1Impl);

	//
	// Activate the custom strategy
	//
	CustomStrategy_impl* strategyImpl = new CustomStrategy_impl(rootPOA);
	PortableServer::ServantBase_var strategyServant = strategyImpl;
	Strategy_var strategy = strategyImpl -> _this();

	//
	// Use the custom strategy
	//
	group -> set_custom_strategy(strategy);

	//
	// Verify the strategy
	//
	TEST(checkStrategy(proxy, "custom"));

	//
	// Should now get the replica of member `m1' when the ior returned
	// by _this() is used
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroupReplicas(m1Impl, ids));

	//
	// Should now get the replica of member `m1' when the client ior is
	// used
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Create the replica POAs for members `m2' and `m3'
	//
	PortableServer::POA_var p2 = createReplicaPOA("p2", "g1", "m2");
	PortableServer::POA_var p3 = createReplicaPOA("p3", "g1", "m3");

	//
	// Create the servants for the replicas of members `m2' and `m3'
	//
	MemberObject_impl* m2Impl = new MemberObject_impl(p2, "m2");
	PortableServer::ServantBase_var m2Servant = m2Impl;
	p2 -> activate_object_with_id(oid.in(), m2Impl);

	MemberObject_impl* m3Impl = new MemberObject_impl(p3, "m3");
	PortableServer::ServantBase_var m3Servant = m3Impl;
	p3 -> activate_object_with_id(oid.in(), m3Impl);

	//
	// Should now get the replicas of members `m1', `m2' and `m3' in
	// round-robin order
	//
	ids.length(3);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	ids[2] = string_dup("m3");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Remove member `m1' and deactivate the object
	//
	group -> remove_member("m1");
	p1 -> deactivate_object(oid.in());

	//
	// Should now get the replicas of members `m2' and `m3' in
	// round-robin order
	//
	ids.length(2);
	ids[0] = string_dup("m2");
	ids[1] = string_dup("m3");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Remove members `m2' and `m3' and deactivate the objects
	//
	group -> remove_member("m2");
	p2 -> deactivate_object(oid.in());
	group -> remove_member("m3");
	p3 -> deactivate_object(oid.in());

	//
	// Should not get any replicas since there are no members
	//
	ids.length(0);
	TEST(checkGroupReplicas(obj, ids));

	//
	// Destroy the POAs
	//
	p1 -> destroy(true, false);
	p2 -> destroy(true, false);
	p3 -> destroy(true, false);

	//
	// Destroy the group
	//
	group -> destroy();

	//
	// When the group is destroyed, its strategy should also be
	// destroyed
	//
	TEST(strategyImpl -> destroyed());

	//
	// Deactivate the custom strategy object
	//
	oid = rootPOA -> servant_to_id(strategyImpl);
	rootPOA -> deactivate_object(oid);
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	TEST(false);
    }
}
