// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/BalancerPolicyTypes.h>
#include <OB/BalancerProperties.h>

#include <CustomStrategy_impl.h>
#include <MemberObject_impl.h>
#include <Tester_impl.h>

#include <stdlib.h> // For abort()

using namespace CORBA;
using namespace LoadBalancing;

typedef OB::Hashtable< String_var, short, OB::StringHasher > StringTable;
typedef OB::Hashtable< String_var, Any, OB::StringHasher > AnyMap;

void
TestMacro(const char* ex, const char* file, unsigned int line)
{
    OB_COUT << "  Failed!" << OB_ENDL;
    OB_COUT << "  Reason: ( " << ex << " ) == false" << OB_ENDL;
    OB_COUT << "  File: " << file << OB_ENDL;
    OB_COUT << "  Line: " << line << OB_ENDL;

    abort();
}

// ----------------------------------------------------------------------
// TestInfo constructor/destructor
// ----------------------------------------------------------------------

TestInfo::TestInfo()
{
}

TestInfo::~TestInfo()
{
    if(!is_nil(group))
	group -> destroy();

    if(!is_nil(replicaPOA1))
	replicaPOA1 -> destroy(true, false);
    if(!is_nil(replicaPOA2))
	replicaPOA2 -> destroy(true, false);
    if(!is_nil(replicaPOA3))
	replicaPOA3 -> destroy(true, false);
    if(!is_nil(auxPOA))
	auxPOA -> destroy(true, false);
}

// ----------------------------------------------------------------------
// Tester_impl private member implementation
// ----------------------------------------------------------------------

//
// Methods for create test groups (used by the persistence tests)
//
// TestInfo* createDefaultGroup();
// TestInfo* createAdaptiveGroup();
// TestInfo* createCustomGroup();
//
// Defined in TestPersistence.cpp
//

//
// Methods for checking test groups (used by the persistence tests)
//
// bool checkDefaultGroup(TestInfo*);
// bool checkAdaptiveGroup(TestInfo*);
// bool checkCustomGroup(TestInfo*);
//
// Defined in TestPersistence.cpp
//

bool
Tester_impl::checkGroupFactory(const GroupIdSeq& groupIds)
{
    ULong i;
    StringTable table;

    //
    // Insert the returned ids into a table
    //
    GroupIdSeq_var ids = factory_ -> list();
    for(i = 0 ; i < ids -> length() ; i++)
	table.put(ids[i], 0);

    //
    // Verify that the correct number of ids were returned
    //
    if(table.size() != groupIds.length())
	return false;

    //
    // Verify the correct ids were returned
    //
    short dummy;;
    for(i = 0 ; i < groupIds.length() ; i++)
	if(!table.get(groupIds[i], dummy))
	    return false;

    return true;
}

bool
Tester_impl::checkGroup(Group_ptr group, const char* groupId,
			const MemberIdSeq& memberIds)
{
    GroupId_var id = group -> get_id();

    //
    // Verify the group id
    //
    if(strcmp(groupId, id.in()) != 0)
	return false;

    ULong i;
    StringTable table;

    //
    // Insert returned ids into a table
    //
    MemberIdSeq_var ids = group -> list_members();
    for(i = 0 ; i < ids -> length() ; i++)
	table.put(ids[i], 0);

    //
    // Verify that the correct number of ids were returned
    //
    if(table.size() != memberIds.length())
	return false;

    //
    // Verify the correct ids were returned
    //
    short dummy;;
    for(i = 0 ; i < memberIds.length() ; i++)
	if(!table.get(memberIds[i], dummy))
	    return false;

    return true;
}

bool
Tester_impl::checkStrategy(StrategyProxy_ptr proxy, const char* strategyName)
{
    //
    // Verify the strategy name
    //
    String_var name = proxy -> get_name();
    if(strcmp(strategyName, name.in()) != 0)
	return false;

    return true;
}

bool
Tester_impl::checkLeastLoadStrategy(StrategyProxy_ptr proxy,
				    ULong tolerance,
				    ULong loadPerClient,
				    ULong rejectLoad,
				    Float dampeningMultiplier,
				    ULong criticalLoad)
{
    //
    // Verify the strategy name
    //
    String_var name = proxy -> get_name();
    if(strcmp("least-load", name.in()) != 0)
	return false;

    //
    // Verify the properties
    //
    PropertySeq_var props = proxy -> get_properties();

    if(props -> length() != 5)
	return false;

    AnyMap map;
    for(ULong i = 0 ; i < props -> length() ; i++)
	map.put(props[i].name.in(), props[i].value);

    Any any;
    ULong ulong_val;
    if(!map.get((const char*)TolerancePropertyName, any) ||
       !(any >>= ulong_val) || ulong_val != tolerance)
	return false;

    if(!map.get((const char*)LoadPerClientPropertyName, any) ||
       !(any >>= ulong_val) || ulong_val != loadPerClient)
	return false;

    if(!map.get((const char*)RejectLoadPropertyName, any) ||
       !(any >>= ulong_val) || ulong_val != rejectLoad)
	return false;

    Float float_val;
    if(!map.get((const char*)DampeningMultiplierPropertyName, any) ||
       !(any >>= float_val) || float_val != dampeningMultiplier)
	return false;

    if(!map.get((const char*)CriticalLoadPropertyName, any) ||
       !(any >>= ulong_val) || ulong_val != criticalLoad)
	return false;

    return true;
}

bool
Tester_impl::checkMinDispStrategy(StrategyProxy_ptr proxy,
				  ULong tolerance,
				  ULong loadPerClient,
				  ULong rejectLoad,
				  Float dampeningMultiplier)
{
    //
    // Verify the strategy name
    //
    String_var name = proxy -> get_name();
    if(strcmp("min-dispersion", name.in()) != 0)
	return false;

    //
    // Verify the properties
    //
    PropertySeq_var props = proxy -> get_properties();

    if(props -> length() != 4)
	return false;

    AnyMap map;
    for(ULong i = 0 ; i < props -> length() ; i++)
	map.put(props[i].name.in(), props[i].value);

    Any any;
    ULong ulong_val;
    if(!map.get((const char*)TolerancePropertyName, any) ||
       !(any >>= ulong_val) || ulong_val != tolerance)
	return false;

    if(!map.get((const char*)LoadPerClientPropertyName, any) ||
       !(any >>= ulong_val) || ulong_val != loadPerClient)
	return false;

    if(!map.get((const char*)RejectLoadPropertyName, any) ||
       !(any >>= ulong_val) || ulong_val != rejectLoad)
	return false;

    Float float_val;
    if(!map.get((const char*)DampeningMultiplierPropertyName, any) ||
       !(any >>= float_val) || float_val != dampeningMultiplier)
	return false;

    return true;
}

bool
Tester_impl::checkGroupReplicas(MemberObject_impl* impl,
				const MemberIdSeq& memberIds)
{
    MemberObject_var mobj = impl -> _this();
    return checkGroupReplicas(mobj, memberIds);
}

bool
Tester_impl::checkGroupReplicas(Object_ptr obj,
				const MemberIdSeq& memberIds)
{
    if(memberIds.length() == 0)
    {
	//
	// Trying to get a replica should raise an OBJECT_NOT_EXIST
	// exception since there are no replicas available
	//
	try
	{
	    MemberObject_var mobj = MemberObject::_narrow(obj);
	    MemberId_var id = mobj -> get_id();
	    return false;
	}
	catch(const OBJECT_NOT_EXIST&)
	{
	}

	return true;
    }

    ULong i;
    StringTable table;
    MemberIdSeq list;
    ULong len = memberIds.length();

    //
    // Insert the returned ids into a table and a list
    //  - the table is used to insure that the correct replicas and only
    //    the correct replicas were returned
    //  - the list is used to verify that the replicas were returned in
    //    round-robin order
    //
    list.length(len * 2);
    for(i = 0 ; i < len * 2 ; i++)
    {
	MemberObject_var mobj = MemberObject::_narrow(obj);
	MemberId_var id = mobj -> get_id();
	list[i] = id;
	table.put(id.in(), 0);
    }

    //
    // Verify that the correct number of replicas were returned
    //
    if(table.size() != len)
	return false;

    //
    // Verify that the correct replicas were returned in round-robin
    // order
    //
    short dummy;;
    for(i = 0 ; i < len ; i++)
    {
	//
	// Make sure each replica was returned
	//
	if(!table.get(memberIds[i], dummy))
	    return false;
	
	//
	// Make sure the replicas were returned in round-robin order
	//
	if(strcmp(list[i], list[i + len]) != 0)
	    return false;
    }
	
    return true;
}

bool
Tester_impl::checkNextReplica(Object_ptr obj, const char* memberId)
{
    MemberObject_var mobj = MemberObject::_narrow(obj);
    MemberId_var id = mobj -> get_id();
    return strcmp(memberId, id.in()) == 0;
}

PortableServer::POA_ptr
Tester_impl::createReplicaPOA(const char* poaName,
			      const char* groupId,
			      const char* memberId)
{
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(
	Object_var(orb_ -> resolve_initial_references("RootPOA")));
    PortableServer::POAManager_var poaManager = rootPOA -> the_POAManager();

    MemberPolicyValue* value = new MemberPolicyValue();
    value -> group_id = string_dup(groupId);
    value -> member_id = string_dup(memberId);
    Any any;
    any <<= value;

    PolicyList pl(3);
    pl.length(3);
    pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
    pl[1] = rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
    pl[2] = orb_ -> create_policy(LoadBalancing::MEMBER_POLICY_ID, any);
    PortableServer::POA_var poa =
        rootPOA -> create_POA(poaName, poaManager, pl);

    return poa._retn();
}

// ----------------------------------------------------------------------
// Tester_impl constructor/destructor
// ----------------------------------------------------------------------

Tester_impl::Tester_impl(
    ORB_ptr orb,
    PortableServer::POA_ptr poa,
    GroupFactory_ptr factory)
    : orb_(ORB::_duplicate(orb)),
      poa_(PortableServer::POA::_duplicate(poa)),
      factory_(GroupFactory::_duplicate(factory)),
      info_(0),
      stage_(1),
      alert_(0)
{
}

Tester_impl::~Tester_impl()
{
    if(info_ != 0)
    {
	for(int i = 0 ; i < 3 ; i++)
	    delete info_[i];
	delete [] info_;
    }
}

// ----------------------------------------------------------------------
// Tester_impl public member implementation
// ----------------------------------------------------------------------

PortableServer::POA_ptr
Tester_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

//
// IDL:Tester/test_GroupFactory:1.0
//
// Defined in TestGroupFactory.cpp
//

//
// IDL:Tester/test_Group:1.0
//
// Defined in TestGroup.cpp
//

//
// IDL:Tester/test_RandomStrategy:1.0
//
// Defined in TestRandom.cpp
//

//
// IDL:Tester/test_RoundRobinStrategy:1.0
//
// Defined in TestRoundRobin.cpp
//

//
// IDL:Tester/test_LeastLoadStrategy:1.0
//
// Defined in TestLeastLoad.cpp
//

//
// IDL:Tester/test_MinDispStrategy:1.0
//
// Defined in TestMinDisp.cpp
//

//
// IDL:Tester/test_CustomStrategy:1.0
//
// Defined in TestCustom.cpp
//

//
// IDL:Tester/test_persistence_stage1:1.0
//
// Defined in TestPersistence.cpp
//

//
// IDL:Tester/test_persistence_stage2:1.0
//
// Defined in TestPersistence.cpp
//

//
// IDL:Tester/test_administrative_stage1:1.0
//
// Defined in TestAdministrative.cpp
//

//
// IDL:Tester/test_administrative_stage2:1.0
//
// Defined in TestAdministrative.cpp
//

//
// IDL:Tester/test_administrative_stage3:1.0
//
// Defined in TestAdministrative.cpp
//

//
// IDL:ooc.com/shutdown:1.0
//
void
Tester_impl::shutdown()
    throw(CORBA::SystemException)
{
    if(is_nil(orb_))
	return;
		        
    orb_ -> shutdown(false);
    orb_ = ORB::_nil();
}

