// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Timer.h>
#include <ActiveTransport.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OCI::BiDir::Message_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OCI::BiDir::Message_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

void
OBDuplicate(OCI::BiDir::ActiveTransport_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OCI::BiDir::ActiveTransport_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

void
OBDuplicate(OCI::BiDir::ActiveTransportWrapper_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OCI::BiDir::ActiveTransportWrapper_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Message constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::Message::Message(OCI::Buffer_ptr header, OCI::Buffer_ptr body)
    : header_(OCI::Buffer::_duplicate(header)),
      body_(OCI::Buffer::_duplicate(body)),
      state_(Unsent), ex_(0)
{
}

OCI::BiDir::Message::~Message()
{
    delete ex_;
}

// ----------------------------------------------------------------------
// Message public member implementations
// ----------------------------------------------------------------------

OCI::Buffer_ptr
OCI::BiDir::Message::header()
{
    return OCI::Buffer::_duplicate(header_);
}

OCI::Buffer_ptr
OCI::BiDir::Message::body()
{
    return OCI::Buffer::_duplicate(body_);
}

void
OCI::BiDir::Message::waitUntilSent(bool block)
{
    if(!block)
        JTCThread::yield();

    JTCSynchronized synchronized(stateMonitor_);

    while(state_ == Unsent)
    {
        if(!block)
            return;

        try
        {
            stateMonitor_.wait();
        }
        catch(const JTCInterruptedException&)
        {
        }
    }

    if(state_ == Exception)
        ex_ -> _raise();
}

void
OCI::BiDir::Message::waitUntilSent(ULong t)
{
    if(t == 0)
        JTCThread::yield();

    JTCSynchronized synchronized(stateMonitor_);

    while(state_ == Unsent)
    {
        try
        {
            State oldState = state_;

            stateMonitor_.wait(t);

            if(state_ == oldState)
                throw NO_RESPONSE();
        }
        catch(const JTCInterruptedException&)
        {
        }
    }

    if(state_ == Exception)
        ex_ -> _raise();
}

void
OCI::BiDir::Message::sent()
{
    JTCSynchronized synchronized(stateMonitor_);

    assert(state_ == Unsent);
    state_ = Sent;
    stateMonitor_.notify();
}

void
OCI::BiDir::Message::exception(const SystemException& ex)
{
    JTCSynchronized synchronized(stateMonitor_);

    assert(state_ == Unsent);
    assert(ex_ == 0);
    state_ = Exception;
    ex_ = SystemException::_downcast(ex._OB_clone());
    stateMonitor_.notify();
}

// ----------------------------------------------------------------------
// ActiveTransport static members
// ----------------------------------------------------------------------

Octet OCI::BiDir::ActiveTransport::magic_[] =
    { (Octet)'O', (Octet)'O', (Octet)'C', (Octet)'B' };

// ----------------------------------------------------------------------
// ActiveTransport::ReceiverThread public member implementations
// ----------------------------------------------------------------------

OCI::BiDir::ActiveTransport::ReceiverThread::ReceiverThread(
    ActiveTransport_ptr t)
{
    activeTransport_ = ActiveTransport::_duplicate(t);
}

void
OCI::BiDir::ActiveTransport::ReceiverThread::run()
{
    activeTransport_ -> receiverRun();

    //
    // Break cyclic object dependency
    //
    activeTransport_ = ActiveTransport::_nil();
}

// ----------------------------------------------------------------------
// ActiveTransport::SenderThread public member implementations
// ----------------------------------------------------------------------

OCI::BiDir::ActiveTransport::SenderThread::SenderThread(
    ActiveTransport_ptr t)
{
    activeTransport_ = ActiveTransport::_duplicate(t);
}

void
OCI::BiDir::ActiveTransport::SenderThread::run()
{
    activeTransport_ -> senderRun();

    //
    // Break cyclic object dependency
    //
    activeTransport_ = ActiveTransport::_nil();
}

// ----------------------------------------------------------------------
// ActiveTransport private and protected member implementations
// ----------------------------------------------------------------------

//
// Create a buffer containing the BiDir message header
//
OCI::Buffer_ptr
OCI::BiDir::ActiveTransport::createHeader(bool callback, CORBA::ULong size)
{
    BiDirIOP::Header h;
    h.magic[0] = magic_[0];
    h.magic[1] = magic_[1];
    h.magic[2] = magic_[2];
    h.magic[3] = magic_[3];
    h.bidiriop_version.major = (Octet)1;
    h.bidiriop_version.minor = (Octet)0;
    h.flags = (Endian ? 1 : 0);
    if(callback)
        h.flags |= 0x2;
    h.message_size = size;

    OCI::Buffer_var buf = new OCI::Buffer();
    OutputStreamImpl out(buf);
    h._OB_marshal(&out);
    buf -> pos(0);

    return buf._retn();
}

//
// Extract a BiDir message header from the buffer
//
void
OCI::BiDir::ActiveTransport::extractHeader(OCI::Buffer_ptr buf,
                                           BiDirIOP::Header& h)
{
    InputStreamImpl in(buf, 0, false);
    BiDirIOP::Header::_OB_unmarshal(h, &in); // TODO: Unmarshal manually/

    //
    // Check byte order and re-read header if we need to swap
    //
    if(((h.flags & 0x1) == 0x1) != Endian)
    {
        InputStreamImpl in(buf, 0, true);
        BiDirIOP::Header::_OB_unmarshal(h, &in);
    }

    if(h.magic[0] != magic_[0] || h.magic[1] != magic_[1] ||
       h.magic[2] != magic_[2] || h.magic[3] != magic_[3])
        throw COMM_FAILURE();
}

//
// Fill as much of the destination buffer as possible
// (not mutex protected)
//
void
OCI::BiDir::ActiveTransport::fillBuffer(OCI::Buffer_ptr dst, bool callback)
{
    OB_STL::list< OCI::Buffer_var >* l;
    if(callback)
        l = &incomingCallback_;
    else
        l = &incoming_;

    //
    // Continue until we've filled the buffer or until there are no
    // more incoming buffers available
    //
    while(!dst -> is_full() && !l -> empty())
    {
        OCI::Buffer_var src = l -> front();
        ULong dstLen = dst -> rest_length();
        ULong srcLen = src -> rest_length();
        ULong len = (srcLen > dstLen ? dstLen : srcLen);
        memcpy(dst -> rest(), src -> rest(), len);
        dst -> advance(len);
        src -> advance(len);
        if(src -> rest_length() == 0)
            l -> pop_front();
    }
}

//
// Add an outgoing header/data buffer pair
//
void
OCI::BiDir::ActiveTransport::add(Message_ptr msg)
{
    JTCSynchronized synchronized(outgoingMonitor_);

    outgoing_.push_back(Message::_duplicate(msg));
    outgoingMonitor_.notify();
}

// ----------------------------------------------------------------------
// ActiveTransport constructor and destructor
// ----------------------------------------------------------------------

//
// Constructor for an outgoing connection. Both the local and remote
// peers are known.
//
OCI::BiDir::ActiveTransport::ActiveTransport(const char* localPeer,
                                             const char* remotePeer,
                                             TransportObserver* observer,
                                             OCI::Transport_ptr delegate)
    : localPeer_(localPeer),
      remotePeer_(remotePeer),
      observer_(observer),
      delegate_(OCI::Transport::_duplicate(delegate)),
      callbackMessage_(false),
      receiveEx_(0),
      shutdown_(false),
      type_(Outgoing)
{
    //
    // Start receiver thread
    //
    receiverThread_ = new ReceiverThread(this);
    receiverThread_ -> start();

    //
    // Start sender thread
    //
    senderThread_ = new SenderThread(this);
    senderThread_ -> start();
}

//
// Constructor for an incoming connection. Only the local peer
// is known at construction.
//
OCI::BiDir::ActiveTransport::ActiveTransport(const char* localPeer,
                                             TransportObserver* observer,
                                             OCI::Transport_ptr delegate)
    : localPeer_(localPeer),
      observer_(observer),
      delegate_(OCI::Transport::_duplicate(delegate)),
      callbackMessage_(false),
      receiveEx_(0),
      shutdown_(false),
      type_(Incoming)
{
    //
    // Start receiver thread
    //
    receiverThread_ = new ReceiverThread(this);
    receiverThread_ -> start();

    //
    // Start sender thread
    //
    senderThread_ = new SenderThread(this);
    senderThread_ -> start();
}

OCI::BiDir::ActiveTransport::~ActiveTransport()
{
    delete receiveEx_;
}

// ----------------------------------------------------------------------
// ActiveTransport public member implementations
// ----------------------------------------------------------------------

OCI::Transport_ptr
OCI::BiDir::ActiveTransport::delegate()
{
    return OCI::Transport::_duplicate(delegate_);
}

OCI::TransportInfo_ptr
OCI::BiDir::ActiveTransport::info()
{
    return delegate_ -> get_info();
}

void
OCI::BiDir::ActiveTransport::close(bool callback)
{
    if(callback)
    {
        JTCSynchronized synchronized(incomingCallbackCond_);

        //
        // Notify any threads waiting to receive from the
        // callback direction, so that connection loss can
        // be detected
        //
        incomingCallbackCond_.notifyAll();

        //
        // If no more messages have been received on the
        // callback connection, then reset callbackMessage_.
        // Otherwise, inform the observer that we have
        // more callback messages waiting.
        //
        if(incomingCallback_.empty())
            callbackMessage_ = false;
        else
            observer_ -> callbackConnection(this);
    }
    else
    {
        observer_ -> shutdown(this);

        {
            JTCSynchronized synchronized(outgoingMonitor_);
            if(!shutdown_)
            {
                shutdown_ = true;
                outgoingMonitor_.notify();
            }
        }

        delegate_ -> close();
    }
}

void
OCI::BiDir::ActiveTransport::shutdown(bool callback)
{
    if(callback)
    {
        //
        // Notify any threads waiting to receive from the
        // callback direction, so that connection loss can
        // be detected
        //
        JTCSynchronized synchronized(incomingCallbackCond_);

        incomingCallbackCond_.notifyAll();
    }
    else
    {
        observer_ -> shutdown(this);

        {
            JTCSynchronized synchronized(outgoingMonitor_);
            shutdown_ = true;
            outgoingMonitor_.notify();
        }

        delegate_ -> shutdown();
    }
}

void
OCI::BiDir::ActiveTransport::receive(OCI::Buffer_ptr buf, bool block,
                                     bool callback, ShutdownStatus* status)
{
    JTCMonitor* cond;
    if(callback)
        cond = &incomingCallbackCond_;
    else
        cond = &incomingCond_;

    JTCSynchronized synchronized(*cond);

    fillBuffer(buf, callback);

    if(!buf -> is_full() && receiveEx_ != 0)
        receiveEx_ -> _raise();

    if(block)
    {
        while(!buf -> is_full())
        {
            try
            {
                cond -> wait();
            }
            catch(const JTCInterruptedException&)
            {
            }

            fillBuffer(buf, callback);

            if(!buf -> is_full())
            {
                if(receiveEx_ != 0)
                    receiveEx_ -> _raise();

                if(status != 0 && status -> isShutdown())
                    throw COMM_FAILURE("", MinorRecv, COMPLETED_NO);
            }
        }
    }
}

bool
OCI::BiDir::ActiveTransport::receiveDetect(OCI::Buffer_ptr buf,
                                           bool block, bool callback,
                                           ShutdownStatus* status)
{
    JTCMonitor* cond;
    if(callback)
        cond = &incomingCallbackCond_;
    else
        cond = &incomingCond_;

    JTCSynchronized synchronized(*cond);

    fillBuffer(buf, callback);

    if(!buf -> is_full() && receiveEx_ != 0)
        return false;

    if(block)
    {
        while(!buf -> is_full())
        {
            try
            {
                cond -> wait();
            }
            catch(const JTCInterruptedException&)
            {
            }

            fillBuffer(buf, callback);

            if(!buf -> is_full())
            {
                if(receiveEx_ != 0)
                    return false;

                if(status != 0 && status -> isShutdown())
                    return false;
            }
        }
    }

    return true;
}

void
OCI::BiDir::ActiveTransport::receiveTimeout(OCI::Buffer_ptr buf,
                                            CORBA::ULong t,
                                            bool callback,
                                            ShutdownStatus* status)
{
    JTCMonitor* cond;
    if(callback)
        cond = &incomingCallbackCond_;
    else
        cond = &incomingCond_;

    JTCSynchronized synchronized(*cond);

    fillBuffer(buf, callback);

    if(!buf -> is_full() && receiveEx_ != 0)
        receiveEx_ -> _raise();

    timeval len;
    len.tv_sec = t/1000;
    len.tv_usec = (t % 1000) * 1000;
    timeval timeDone = TimeAdd(TimeNow(), len);
    while(!buf -> is_full())
    {
        //
        // Have we timed out?
        //
        timeval now = TimeNow();
        if(TimeCmp(timeDone, now) <= 0)
            break;

        //
        // No, calculate how long we have to wait now
        //
        timeval diff = TimeSub(timeDone, now);
        ULong interval = diff.tv_sec * 1000 + diff.tv_usec / 1000;

        //
        // Block until we are notified, or until we timeout
        //
        try
        {
            cond -> wait(interval);
        }
        catch(const JTCInterruptedException&)
        {
        }

        fillBuffer(buf, callback);

        if(!buf -> is_full())
        {
            if(receiveEx_ != 0)
                receiveEx_ -> _raise();

            if(status != 0 && status -> isShutdown())
                throw COMM_FAILURE("", MinorRecv, COMPLETED_NO);
        }
    }
}

bool
OCI::BiDir::ActiveTransport::receiveTimeoutDetect(OCI::Buffer_ptr buf,
                                                  CORBA::ULong t,
                                                  bool callback,
                                                  ShutdownStatus* status)
{
    JTCMonitor* cond;
    if(callback)
        cond = &incomingCallbackCond_;
    else
        cond = &incomingCond_;

    JTCSynchronized synchronized(*cond);

    fillBuffer(buf, callback);

    if(!buf -> is_full() && receiveEx_ != 0)
        return false;

    timeval len;
    len.tv_sec = t/1000;
    len.tv_usec = (t % 1000) * 1000;
    timeval timeDone = TimeAdd(TimeNow(), len);
    while(!buf -> is_full())
    {
        //
        // Have we timed out?
        //
        timeval now = TimeNow();
        if(TimeCmp(timeDone, now) <= 0)
            break;

        //
        // No, calculate how long we have to wait now
        //
        timeval diff = TimeSub(timeDone, now);
        ULong interval = diff.tv_sec * 1000 + diff.tv_usec / 1000;

        //
        // Block until we are notified, or until we timeout
        //
        try
        {
            cond -> wait(interval);
        }
        catch(const JTCInterruptedException&)
        {
        }

        fillBuffer(buf, callback);

        if(!buf -> is_full())
        {
            if(receiveEx_ != 0)
                return false;

            if(status != 0 && status -> isShutdown())
                return false;
        }
    }

    return true;
}

void
OCI::BiDir::ActiveTransport::send(OCI::Buffer_ptr buf, bool block,
                                  bool callback)
{
    if(shutdown_)
        throw COMM_FAILURE();

    //
    // Create BiDir message and add to outgoing list
    //
    OCI::Buffer_var hbuf = createHeader(callback, buf -> rest_length());
    Message_var msg = new Message(hbuf, buf);
    {
        JTCSynchronized synchronized(outgoingMonitor_);

        if(shutdown_)
            throw COMM_FAILURE();

        outgoing_.push_back(Message::_duplicate(msg));
        outgoingMonitor_.notify();
    }

    msg -> waitUntilSent(block);
}

bool
OCI::BiDir::ActiveTransport::sendDetect(OCI::Buffer_ptr buf, bool block,
                                        bool callback)
{
    if(shutdown_)
        return false;

    //
    // Create BiDir message and add to outgoing list
    //
    OCI::Buffer_var hbuf = createHeader(callback, buf -> rest_length());
    Message_var msg = new Message(hbuf, buf);
    {
        JTCSynchronized synchronized(outgoingMonitor_);

        if(shutdown_)
            return false;

        outgoing_.push_back(Message::_duplicate(msg));
        outgoingMonitor_.notify();
    }

    try
    {
        msg -> waitUntilSent(block);
    }
    catch(const SystemException&)
    {
        return false;
    }

    return true;
}

void
OCI::BiDir::ActiveTransport::sendTimeout(OCI::Buffer_ptr buf, CORBA::ULong t,
                                         bool callback)
{
    if(shutdown_)
        throw COMM_FAILURE();

    //
    // Create BiDir message and add to outgoing list
    //
    OCI::Buffer_var hbuf = createHeader(callback, buf -> rest_length());
    Message_var msg = new Message(hbuf, buf);
    {
        JTCSynchronized synchronized(outgoingMonitor_);

        if(shutdown_)
            throw COMM_FAILURE();

        outgoing_.push_back(Message::_duplicate(msg));
        outgoingMonitor_.notify();
    }

    msg -> waitUntilSent(t);
}

bool
OCI::BiDir::ActiveTransport::sendTimeoutDetect(OCI::Buffer_ptr buf,
                                               CORBA::ULong t,
                                               bool callback)
{
    if(shutdown_)
        return false;

    //
    // Create BiDir message and add to outgoing list
    //
    OCI::Buffer_var hbuf = createHeader(callback, buf -> rest_length());
    Message_var msg = new Message(hbuf, buf);
    {
        JTCSynchronized synchronized(outgoingMonitor_);

        if(shutdown_)
            return false;

        outgoing_.push_back(Message::_duplicate(msg));
        outgoingMonitor_.notify();
    }

    try
    {
        msg -> waitUntilSent(t);
    }
    catch(const SystemException&)
    {
        return false;
    }

    return true;
}

void
OCI::BiDir::ActiveTransport::sendLocalPeerId()
{
    assert(type_ == Outgoing);

    //
    // Transmit the local peer id. The message format is
    //
    //   boolean endian
    //   sequence<octet> id
    //
    OCI::Buffer_var buf = new OCI::Buffer();
    OutputStreamImpl out(buf);
    out._OB_writeEndian();
    ULong len = strlen(localPeer_.in());
    out.write_ulong(len);
    out.write_octet_array((Octet*)localPeer_.in(), len);
    buf -> pos(0);
    send(buf, true, false);
}

void
OCI::BiDir::ActiveTransport::receiveRemotePeerId()
{
    assert(type_ == Incoming);

    //
    // Receive the remote peer id. The message format is
    //
    //   boolean endian
    //   sequence<octet> id
    //
    OCI::Buffer_var buf = new OCI::Buffer(8);
    receive(buf, true, false, 0);
    assert(buf -> is_full());
    InputStreamImpl in(buf, 0, false);
    in._OB_readEndian();
    ULong len = in.read_ulong();
    buf -> alloc(len);
    receive(buf, true, false, 0);
    assert(buf -> is_full());
    remotePeer_ = string_alloc(buf -> length());
    memcpy(remotePeer_.inout(), buf -> data(), len);
    remotePeer_[len] = '\0';
}

void
OCI::BiDir::ActiveTransport::receiverRun()
{
    try
    {
        OCI::Buffer_var buf = new OCI::Buffer;
        while(true)
        {
            //
            // Read header
            //
            buf -> alloc(16);
            delegate_ -> receive(buf, true);
            assert(buf -> is_full());
            BiDirIOP::Header hdr;
            extractHeader(buf, hdr);

            //
            // Read body
            //
            buf -> alloc(hdr.message_size);
            delegate_ -> receive(buf, true);
            assert(buf -> is_full());

            //
            // Consume buffer
            //
            OCI::Buffer_var b = new OCI::Buffer;
            b -> consume(buf);
            b -> pos(0);

            //
            // Add the buffer to the appropriate list and notify any
            // waiters
            //
            if((hdr.flags & 0x2) == 0x2)
            {
                //
                // Callback message
                //
                JTCSynchronized synchronized(incomingCallbackCond_);

                //
                // Is this the first callback message?
                //
                if(!callbackMessage_)
                {
                    callbackMessage_ = true;
                    observer_ -> callbackConnection(this);
                }

                //
                // When the condition variable is signaled, another
                // thread may acquire the buffer. We must release
                // the buffer now, since it is not safe to modify the
                // reference count of OCI::Buffer from multiple threads.
                //
                incomingCallback_.push_back(b._retn());
                incomingCallbackCond_.notifyAll();
            }
            else
            {
                JTCSynchronized synchronized(incomingCond_);

                //
                // When the condition variable is signaled, another
                // thread may acquire the buffer. We must release
                // the buffer now, since it is not safe to modify the
                // reference count of OCI::Buffer from multiple threads.
                //
                incoming_.push_back(b._retn());
                incomingCond_.notifyAll();
            }
        }
    }
    catch(const SystemException& ex)
    {
        receiveEx_ = SystemException::_downcast(ex._OB_clone());

        //
        // Notify any waiters that an exception occurred
        //
        {
            JTCSynchronized synchronized(incomingCond_);
            incomingCond_.notifyAll();
        }
        {
            JTCSynchronized synchronized(incomingCallbackCond_);
            incomingCallbackCond_.notifyAll();
        }
    }
}

void
OCI::BiDir::ActiveTransport::senderRun()
{
    try
    {
        while(!shutdown_)
        {
            Message_var msg;

            {
                JTCSynchronized synchronized(outgoingMonitor_);

                while(!shutdown_ && outgoing_.empty())
                {
                    try
                    {
                        outgoingMonitor_.wait();
                    }
                    catch(const JTCInterruptedException&)
                    {
                    }
                }

                if(!shutdown_)
                {
                    msg = outgoing_.front();
                    outgoing_.pop_front();
                }
            }

            if(msg.in() != 0)
            {
                try
                {
                    OCI::Buffer_var header = msg -> header();
                    delegate_ -> send(header, true);
                    assert(header -> is_full());

                    OCI::Buffer_var body = msg -> body();
                    delegate_ -> send(body, true);
                    assert(body -> is_full());

                    msg -> sent();
                }
                catch(const SystemException& ex)
                {
                    msg -> exception(ex);
                    throw;
                }
            }
        }

        //
        // We're shutdown
        //
        {
            JTCSynchronized synchronized(outgoingMonitor_);

            for(OB_STL::list< Message_var >::iterator iter = outgoing_.begin();
                iter != outgoing_.end() ;
                iter++)
            {
                (*iter) -> exception(COMM_FAILURE());
            }
        }
    }
    catch(const SystemException& ex)
    {
        JTCSynchronized synchronized(outgoingMonitor_);

        for(OB_STL::list< Message_var >::iterator iter = outgoing_.begin() ;
            iter != outgoing_.end() ;
            iter++)
        {
            (*iter) -> exception(ex);
        }
    }
}

OCI::BiDir::ActiveTransportWrapper::ActiveTransportWrapper(
    ActiveTransport_var atVar)
    : at_(atVar), 
      atRefcount_(1)
{
}

int
OCI::BiDir::ActiveTransportWrapper::getRefcount()
{
    return atRefcount_;
}

void
OCI::BiDir::ActiveTransportWrapper::incrRefcount()
{
    atRefcount_++;
}

void
OCI::BiDir::ActiveTransportWrapper::decrRefcount()
{
    atRefcount_--;
}

OCI::BiDir::ActiveTransport_var
OCI::BiDir::ActiveTransportWrapper::getAt()
{
    return at_;
}
