// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OBPolicies.h>
#include <OB/IORUtil.h>
#include <OB/seq_converter.h>
#include <OB/Properties.h>

#include <OCI_BiDir_impl.h>
#include <TransportManager.h>

//
// Undefine minor (this is needed for SunOS 4.1 and some other Unix
// flavors)
//
#undef minor

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// CorbalocProtocol_impl constructor and destructor
// ----------------------------------------------------------------------
        
OCI::BiDir::CorbalocProtocol_impl::CorbalocProtocol_impl(
    const char* id,
    OCI::ProfileId tag,
    OB::CorbalocProtocol_ptr delegate)
    : id_(id), tag_(tag),
      delegate_(OB::CorbalocProtocol::_duplicate(delegate))
{
}

OCI::BiDir::CorbalocProtocol_impl::~CorbalocProtocol_impl()
{
}

char*
OCI::BiDir::CorbalocProtocol_impl::name()
    throw(SystemException)
{
    return string_dup(id_);
}

IOP::TaggedProfile*
OCI::BiDir::CorbalocProtocol_impl::parse_address(
    const char* addr, const CorbalocProtocol::ObjectKey& key)
    throw(SystemException)
{
    //
    // The format of a bidir URL address is one of the following:
    //
    // <name>:<peer>
    // <name>:<peer>:<prot_addr>
    //
    // The first form indicates a callback-only profile, whereas
    // the second form includes address information for the
    // encapsulated protocol. The <peer> field represents the peer
    // ID of the server.
    //

    // 
    // Empty peer ID is illegal
    // 
    if(*addr == '\0' || *addr == ':')
    {
        String_var err = id_;
        err += ": peer ID must be specified in URL";
        throw BAD_PARAM(err, MinorBadAddress, COMPLETED_NO);
    }

    //
    // Peer ID is terminated by ':', or by end of string
    //
    String_var peer;
    const char* colon = strchr(addr, ':');
    if(colon == 0)
        peer = addr;
    else
    {
        ULong len = colon - addr;
        peer = string_alloc(len);
        strncpy(peer.inout(), addr, len);
        peer[len] = '\0';
    }

    //
    // Get nested profile, if necessary
    //
    IOP::TaggedProfile_var nested;
    if(colon != 0)
        nested = delegate_ -> parse_address(colon + 1, key);

    //
    // Create profile
    //
    IOP::TaggedProfile_var profile = new IOP::TaggedProfile;
    profile -> tag = tag_;

    ::BiDirIOP::ProfileBody body;
    body.bidiriop_version.major = 1;
    body.bidiriop_version.minor = 0;
    body.giop_version.major = 1;
    body.giop_version.minor = 0;
    body.peer = peer;
    body.object_key.length(key.length());
    memcpy(body.object_key.get_buffer(), key.get_buffer(), key.length());
    if(colon != 0)
    {
        body.profiles.length(1);
        body.profiles[0] = nested;
    }

    OCI::BufferImpl buf;
    OutputStreamImpl out(&buf);
    out._OB_writeEndian();
    body._OB_marshal(&out);
    ULong count = out._OB_pos();
    profile -> profile_data.length(count);
    Octet* oct = profile -> profile_data.get_buffer();
    memcpy(oct, buf.data(), count);

    return profile._retn();
}

void
OCI::BiDir::CorbalocProtocol_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// Plugin_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::Plugin_impl::Plugin_impl(ORB_ptr orb, const char* peer)
    : orb_(ORB::_duplicate(orb)), peer_(peer)
{
}

OCI::BiDir::Plugin_impl::~Plugin_impl()
{
}

// ----------------------------------------------------------------------
// Plugin_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::BiDir::Plugin_impl::id()
    throw(CORBA::SystemException)
{
    return string_dup("bidir");
}

OCI::ProfileId
OCI::BiDir::Plugin_impl::tag()
    throw(CORBA::SystemException)
{
    throw NO_IMPLEMENT(); // TODO: change this?
    return 0; // some compilers need this
}

void
OCI::BiDir::Plugin_impl::init_client(const OCI::ParamSeq& params)
    throw(SystemException)
{
    const char* protocol = 0;

    ULong i = 0;
    while(i < params.length())
    {
        if(strcmp(params[i], "--protocol") == 0)
        {
            if(i + 1 == params.length())
                throw INITIALIZE("bidir: argument expected for --protocol");

            protocol = params[i + 1];

            i += 2;
        }
        else
        {
            String_var err = string_dup("bidir: unknown client parameter `");
            err += params[i];
            err += "'";
            throw INITIALIZE(err);
        }
    }

    if(protocol == 0)
        throw INITIALIZE("bidir: no protocol specified");

    String_var id = string_dup("bidir_");
    id += protocol;

    try
    {
        Object_var obj;
        TransportManager_var transportManager;

        //
        // See if an acceptor factory already exists for this protocol.
        // If so, get the transport manager from it.
        //
        {
            obj = orb_ -> resolve_initial_references("OCIAccFactoryRegistry");
            OCI::AccFactoryRegistry_var registry =
                OCI::AccFactoryRegistry::_narrow(obj);
            assert(!is_nil(registry));
            OCI::AccFactorySeq_var factories = registry -> get_factories();
            for(i = 0 ; i < factories -> length() ; i++)
            {
                String_var str = factories[i] -> id();
                if(strcmp(str, id) == 0)
                {
                    AccFactory_impl* impl =
                        dynamic_cast<AccFactory_impl*>(factories[i].in());
                    assert(impl != 0);
                    transportManager = impl -> _OB_transportManager();
                    break;
                }
            }
        }

        obj = orb_ -> resolve_initial_references("OCIConFactoryRegistry");
        OCI::ConFactoryRegistry_var registry =
            OCI::ConFactoryRegistry::_narrow(obj);
        assert(!is_nil(registry));
        OCI::ConFactorySeq_var factories = registry -> get_factories();
        for(i = 0 ; i < factories -> length() ; i++)
        {
            String_var str = factories[i] -> id();
            if(strcmp(str, protocol) == 0)
                break;
        }
        if(i == factories -> length())
        {
            String_var err = string_dup("bidir: protocol `");
            err += protocol;
            err += "' not found";
            throw INITIALIZE(err);
        }
        ProfileId tag;
        if(is_nil(transportManager))
        {
            tag = factories[i] -> tag() | 1024;
            transportManager = new TransportManager(id, tag, peer_);
        }
        else
            tag = transportManager -> tag();
        OCI::ConFactory_var factory = new ConFactory_impl(factories[i],
                                                          transportManager);
        registry -> add_factory(factory);

        obj = orb_ -> resolve_initial_references("URLRegistry");
        URLRegistry_var urlRegistry = URLRegistry::_narrow(obj);
        assert(!is_nil(registry));
        URLScheme_var scheme = urlRegistry -> find_scheme("corbaloc");
        assert(!is_nil(scheme));
        CorbalocURLScheme_var corbaloc = CorbalocURLScheme::_narrow(scheme);
        assert(!is_nil(corbaloc));
        CorbalocProtocol_var p = corbaloc -> find_protocol(protocol);
        if(!is_nil(p))
        {
            CorbalocProtocol_var prot = new CorbalocProtocol_impl(id, tag, p);
            corbaloc -> add_protocol(prot);
        }
    }
    catch(const ORB::InvalidName&)
    {
        assert(false);
    }
    catch(const OCI::FactoryAlreadyExists&)
    {
        String_var err = string_dup("bidir: plug-in already installed "
                                    "for protocol `");
        err += protocol;
        err += "'";
        throw INITIALIZE(err);
    }
    catch(const CorbalocURLScheme::ProtocolAlreadyExists&)
    {
        assert(false);
    }
}

void
OCI::BiDir::Plugin_impl::init_server(const OCI::ParamSeq& params)
    throw(SystemException)
{
    const char* protocol = 0;

    ULong i = 0;
    while(i < params.length())
    {
        if(strcmp(params[i], "--protocol") == 0)
        {
            if(i + 1 == params.length())
                throw INITIALIZE("bidir: argument expected for --protocol");

            protocol = params[i + 1];

            i += 2;
        }
        else
        {
            String_var err = string_dup("bidir: unknown server parameter `");
            err += params[i];
            err += "'";
            throw INITIALIZE(err);
        }
    }

    if(protocol == 0)
        throw INITIALIZE("bidir: no protocol specified");

    String_var id = string_dup("bidir_");
    id += protocol;

    try
    {
        Object_var obj;
        TransportManager_var transportManager;

        //
        // See if a connector factory already exists for this bi-dir
        // protocol. If so, get the transport manager from it.
        //
        {
            obj = orb_ -> resolve_initial_references("OCIConFactoryRegistry");
            OCI::ConFactoryRegistry_var registry =
                OCI::ConFactoryRegistry::_narrow(obj);
            assert(!is_nil(registry));
            OCI::ConFactorySeq_var factories = registry -> get_factories();
            for(i = 0 ; i < factories -> length() ; i++)
            {
                String_var str = factories[i] -> id();
                if(strcmp(str, id) == 0)
                {
                    ConFactory_impl* impl =
                        dynamic_cast<ConFactory_impl*>(factories[i].in());
                    assert(impl != 0);
                    transportManager = impl -> _OB_transportManager();
                    break;
                }
            }
        }

        obj = orb_ -> resolve_initial_references("OCIAccFactoryRegistry");
        OCI::AccFactoryRegistry_var registry =
            OCI::AccFactoryRegistry::_narrow(obj);
        assert(!is_nil(registry));
        OCI::AccFactorySeq_var factories = registry -> get_factories();
        for(i = 0 ; i < factories -> length() ; i++)
        {
            String_var str = factories[i] -> id();
            if(strcmp(str, protocol) == 0)
                break;
        }
        if(i == factories -> length())
        {
            String_var err = string_dup("bidir: protocol `");
            err += protocol;
            err += "' not found";
            throw INITIALIZE(err);
        }
        if(is_nil(transportManager))
        {
            ProfileId tag = factories[i] -> tag() | 1024;
            transportManager = new TransportManager(id, tag, peer_);
        }
        OCI::AccFactory_var factory = new AccFactory_impl(factories[i],
                                                          transportManager);
        registry -> add_factory(factory);
    }
    catch(const ORB::InvalidName&)
    {
        assert(false);
    }
    catch(const OCI::FactoryAlreadyExists&)
    {
        String_var err = string_dup("bidir: plug-in already installed "
                                    "for protocol `");
        err += protocol;
        err += "'";
        throw INITIALIZE(err);
    }
}

// ----------------------------------------------------------------------
// AccFactoryInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::AccFactoryInfo_impl::AccFactoryInfo_impl(
    const char* id,
    OCI::ProfileId tag,
    OCI::AccFactoryInfo_ptr delegate)
    : id_(id), tag_(tag), delegate_(OCI::AccFactoryInfo::_duplicate(delegate))
{
}

OCI::BiDir::AccFactoryInfo_impl::~AccFactoryInfo_impl()
{
}

// ----------------------------------------------------------------------
// AccFactoryInfo_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::AccFactoryInfo_impl::id()
    throw(SystemException)
{
    return string_dup(id_);
}

OCI::ProfileId
OCI::BiDir::AccFactoryInfo_impl::tag()
    throw(SystemException)
{
    return tag_;
}

char*
OCI::BiDir::AccFactoryInfo_impl::describe()
    throw(SystemException)
{
    String_var result = id_;
    result += " acceptor factory";
    result += '\n';
    String_var s = delegate_ -> describe();
    result += s;
    return result._retn();
}

// ----------------------------------------------------------------------
// AccFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::AccFactory_impl::AccFactory_impl(
    OCI::AccFactory_ptr delegate,
    TransportManager_ptr transportManager)
    : delegate_(OCI::AccFactory::_duplicate(delegate)),
      transportManager_(TransportManager::_duplicate(transportManager))
{
    OCI::AccFactoryInfo_var info = delegate -> get_info();
    info_ = new AccFactoryInfo_impl(transportManager_ -> id(),
                                    transportManager_ -> tag(),
                                    info);
}

OCI::BiDir::AccFactory_impl::~AccFactory_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// AccFactory_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::AccFactory_impl::id()
    throw(SystemException)
{
    return string_dup(transportManager_ -> id());
}

OCI::ProfileId
OCI::BiDir::AccFactory_impl::tag()
    throw(SystemException)
{
    return transportManager_ -> tag();
}

OCI::Acceptor_ptr
OCI::BiDir::AccFactory_impl::create_acceptor(const OCI::ParamSeq& params)
    throw(OCI::InvalidParam,
          SystemException)
{
    OCI::Acceptor_var result;
    if(params.length() > 0 && strcmp(params[0], "--callback") == 0)
    {
        if(params.length() > 1)
            throw OCI::InvalidParam("extra arguments for callback acceptor");
        if(transportManager_ -> hasCallbackAcceptor())
            throw OCI::InvalidParam("at most one callback acceptor may be "
                                    "created");
        result = new CallbackAcceptor_impl(transportManager_);
    }
    else
    {
        //
        // Pass all parameters to delegate factory
        //
        OCI::Acceptor_var acc = delegate_ -> create_acceptor(params);
        result = new Acceptor_impl(transportManager_, acc);
    }

    return result._retn();
}

void
OCI::BiDir::AccFactory_impl::change_key(IOP::IOR& ior, const ObjectKey& key)
    throw(SystemException)
{
    //
    // Extract the BiDirIOP profile information from the provided IOR
    //
    for(ULong profile = 0 ; profile < ior.profiles.length() ; profile++)
    {
        if(ior.profiles[profile].tag == tag())
        {
            //
            // Get ProfileBody
            //
            const Octet* coct =
                ior.profiles[profile].profile_data.get_buffer();
            ULong len = ior.profiles[profile].profile_data.length();
            OCI::BufferImpl buf((Octet*)coct, len);
            InputStreamImpl in(&buf, 0, false, 0, false);
            in._OB_readEndian();
            BiDirIOP::ProfileBody body;
            BiDirIOP::ProfileBody::_OB_unmarshal(body, &in);

            //
            // Fill in the new object-key
            //
            body.object_key.length(key.length());
            memcpy(body.object_key.get_buffer(), key.get_buffer(),
                   key.length());

            //
            // Update the nested profiles
            //
            IOP::IOR dummy;
            ULong i;
            dummy.type_id = ior.type_id;
            dummy.profiles.length(body.profiles.length());
            for(i = 0 ; i < body.profiles.length() ; i++)
                dummy.profiles[i] = body.profiles[i];
            delegate_ -> change_key(dummy, key);
            for(i = 0 ; i < dummy.profiles.length() ; i++)
                body.profiles[i] = dummy.profiles[i];

            //
            // Remarshal the new body
            //
            OCI::BufferImpl buf2;
            OutputStreamImpl out(&buf2);
            out._OB_writeEndian();
            body._OB_marshal(&out);
            ULong count = out._OB_pos();
            ior.profiles[profile].profile_data.length(count);
            Octet* oct = ior.profiles[profile].profile_data.get_buffer();
            memcpy(oct, buf2.data(), count);
        }
    }
}

OCI::AccFactoryInfo_ptr
OCI::BiDir::AccFactory_impl::get_info()
    throw(SystemException)
{
    return OCI::AccFactoryInfo::_duplicate(info_);
}

OCI::BiDir::TransportManager_ptr
OCI::BiDir::AccFactory_impl::_OB_transportManager()
{
    return TransportManager::_duplicate(transportManager_);
}

// ----------------------------------------------------------------------
// AcceptorInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::AcceptorInfo_impl::AcceptorInfo_impl(
    const char* id,
    OCI::ProfileId tag,
    const char* localPeer,
    OCI::AcceptorInfo_ptr delegate)
    : id_(id), tag_(tag), localPeer_(localPeer),
      delegate_(OCI::AcceptorInfo::_duplicate(delegate))
{
}

OCI::BiDir::AcceptorInfo_impl::~AcceptorInfo_impl()
{
}

// ----------------------------------------------------------------------
// AcceptorInfo_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::AcceptorInfo_impl::id()
    throw(SystemException)
{
    return string_dup(id_);
}

OCI::ProfileId
OCI::BiDir::AcceptorInfo_impl::tag()
    throw(SystemException)
{
    return tag_;
}

char*
OCI::BiDir::AcceptorInfo_impl::describe()
    throw(SystemException)
{
    String_var result = id_;
    if(is_nil(delegate_))
        result += " callback endpoint";
    else
        result += " endpoint";
    result += "\nlocal peer: ";
    result += localPeer_;
    if(!is_nil(delegate_))
    {
        String_var s = delegate_ -> describe();
        result += '\n';
        result += s;
    }
    return result._retn();
}

void
OCI::BiDir::AcceptorInfo_impl::add_accept_cb(OCI::AcceptCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<AcceptCB_var>::iterator iter = acceptCBs_.begin() ;
        iter != acceptCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    acceptCBs_.push_back(OCI::AcceptCB::_duplicate(cb));
}

void
OCI::BiDir::AcceptorInfo_impl::remove_accept_cb(OCI::AcceptCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<AcceptCB_var>::iterator iter = acceptCBs_.begin() ;
        iter != acceptCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
        {
            acceptCBs_.erase(iter);
            return;
        }
    }
}

void
OCI::BiDir::AcceptorInfo_impl::_OB_callAcceptCB(OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<AcceptCB_var>::iterator iter = acceptCBs_.begin() ;
        iter != acceptCBs_.end() ;
        iter++)
    {
        (*iter) -> accept_cb(info);
    }
}

// ----------------------------------------------------------------------
// Acceptor_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::Acceptor_impl::Acceptor_impl(TransportManager_ptr transportManager,
                                         OCI::Acceptor_ptr delegate)
    : transportManager_(TransportManager::_duplicate(transportManager)),
      delegate_(OCI::Acceptor::_duplicate(delegate))
{
    OCI::AcceptorInfo_var info = delegate -> get_info();
    info_ = new AcceptorInfo_impl(transportManager -> id(),
                                  transportManager -> tag(),
                                  transportManager -> peer(),
                                  info);
}

OCI::BiDir::Acceptor_impl::~Acceptor_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// Acceptor_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::Acceptor_impl::id()
    throw(SystemException)
{
    return string_dup(transportManager_ -> id());
}

OCI::ProfileId
OCI::BiDir::Acceptor_impl::tag()
    throw(SystemException)
{
    return transportManager_ -> tag();
}

OCI::Handle
OCI::BiDir::Acceptor_impl::handle()
    throw(SystemException)
{
    return -1;
}

void
OCI::BiDir::Acceptor_impl::close()
    throw(SystemException)
{
    delegate_ -> close();
}

void
OCI::BiDir::Acceptor_impl::shutdown()
    throw(SystemException)
{
    delegate_ -> shutdown();
}

void
OCI::BiDir::Acceptor_impl::listen()
    throw(SystemException)
{
    delegate_ -> listen();
}

OCI::Transport_ptr
OCI::BiDir::Acceptor_impl::accept(Boolean block)
    throw(SystemException)
{
    OCI::Transport_var delegateTransport = delegate_ -> accept(block);

    OCI::Transport_var t;
    if(!is_nil(delegateTransport))
    {
        t = transportManager_ -> addIncomingTransport(this, delegateTransport);

        //
        // Call callbacks
        //
        OCI::TransportInfo_var info = t -> get_info();
        try
        {
            info_ -> _OB_callAcceptCB(info);
        }
        catch(const SystemException&)
        {
            t -> close();
            throw;
        }
    }

    //
    // Return new transport
    //
    return t._retn();
}

OCI::Transport_ptr
OCI::BiDir::Acceptor_impl::connect_self()
    throw(SystemException)
{
    return delegate_ -> connect_self();
}

void
OCI::BiDir::Acceptor_impl::add_profiles(
    const OCI::ProfileInfo& profile_info, OCI::IOR& ref)
    throw(SystemException)
{
    //
    // Create a ProfileInfo without components, and an IOR with
    // no profiles, and call add_profile on the delegate
    //
    OCI::IOR dummyIOR;
    dummyIOR.type_id = ref.type_id;
    delegate_ -> add_profiles(profile_info, dummyIOR);

    //
    // Create BiDirIOP profile
    //
    BiDirIOP::ProfileBody body;
    body.bidiriop_version.major = 1;
    body.bidiriop_version.minor = 0;
    body.giop_version.major = profile_info.major;
    body.giop_version.minor = profile_info.minor;
    body.peer = transportManager_ -> peer();
    body.object_key.length(profile_info.key.length());
    memcpy(body.object_key.get_buffer(), profile_info.key.get_buffer(),
           profile_info.key.length());
    body.components.length(0);
    body.profiles.length(dummyIOR.profiles.length());
    for(ULong i = 0 ; i < dummyIOR.profiles.length() ; i++)
        body.profiles[i] = dummyIOR.profiles[i];

    //
    // Append the BiDirIOP profile to the IOR
    //
    OCI::BufferImpl buf;
    OutputStreamImpl out(&buf);
    out._OB_writeEndian();
    body._OB_marshal(&out);
    ULong len = ref.profiles.length();
    ref.profiles.length(len + 1);
    ref.profiles[len].tag = tag();
    ref.profiles[len].profile_data.length(buf.length());
    memcpy(ref.profiles[len].profile_data.get_buffer(), buf.data(),
           buf.length());
}

OCI::ProfileInfoSeq*
OCI::BiDir::Acceptor_impl::get_local_profiles(const OCI::IOR& ref)
    throw(SystemException)
{
    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;
    for(ULong i = 0 ; i < ref.profiles.length() ; i++)
    {
        if(ref.profiles[i].tag == tag())
        {
            //
            // Get ProfileBody
            //
            const Octet* coct = ref.profiles[i].profile_data.get_buffer();
            ULong len = ref.profiles[i].profile_data.length();
            OCI::BufferImpl buf((Octet*)coct, len);
            InputStreamImpl in(&buf, 0, false, 0, false);
            in._OB_readEndian();
            BiDirIOP::ProfileBody body;
            BiDirIOP::ProfileBody::_OB_unmarshal(body, &in);

            if(strcmp(body.peer, transportManager_ -> peer()) == 0 &&
               body.profiles.length() > 0)
            {
                OCI::IOR dummyIOR;
                dummyIOR.type_id = ref.type_id;
                dummyIOR.profiles.length(body.profiles.length());
                for(ULong j = 0 ; j < body.profiles.length() ; j++)
                    dummyIOR.profiles[j] = body.profiles[j];
                OCI::ProfileInfoSeq_var seq2 =
                    delegate_ -> get_local_profiles(dummyIOR);
                if(seq2 -> length() > 0)
                {
                    //
                    // Collect all components from delegate's profiles
                    //
                    OCI::TaggedComponentSeq components;
                    for(ULong j = 0 ; j < seq2 -> length() ; j++)
                        for(ULong k = 0 ; k < seq2[j].components.length() ;
                            k++)
                        {
                            ULong len = components.length() + 1;
                            components.length(len);
                            components[len - 1] = seq2[j].components[k];
                        }

                    ULong len = seq -> length();
                    seq -> length(len + 1);
                    seq[len].major = body.giop_version.major;
                    seq[len].minor = body.giop_version.minor;
                    seq[len].key.length(body.object_key.length());
                    memcpy(seq[len].key.get_buffer(),
                           body.object_key.get_buffer(),
                           body.object_key.length());
                    seq[len].components = components;
                }
            }
        }
    }

    return seq._retn();
}

OCI::AcceptorInfo_ptr
OCI::BiDir::Acceptor_impl::get_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// CallbackAcceptor_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::CallbackAcceptor_impl::CallbackAcceptor_impl(
    TransportManager_ptr transportManager)
    : transportManager_(TransportManager::_duplicate(transportManager))
{
    info_ = new AcceptorInfo_impl(transportManager -> id(),
                                  transportManager -> tag(),
                                  transportManager -> peer(),
                                  OCI::AcceptorInfo::_nil());
    transportManager_ -> setCallbackAcceptor(this);
}

OCI::BiDir::CallbackAcceptor_impl::~CallbackAcceptor_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// CallbackAcceptor_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::CallbackAcceptor_impl::id()
    throw(SystemException)
{
    return string_dup(transportManager_ -> id());
}

OCI::ProfileId
OCI::BiDir::CallbackAcceptor_impl::tag()
    throw(SystemException)
{
    return transportManager_ -> tag();
}

OCI::Handle
OCI::BiDir::CallbackAcceptor_impl::handle()
    throw(SystemException)
{
    return -1;
}

void
OCI::BiDir::CallbackAcceptor_impl::close()
    throw(SystemException)
{
    transportManager_ -> setCallbackAcceptor(0);
}

void
OCI::BiDir::CallbackAcceptor_impl::shutdown()
    throw(SystemException)
{
    //
    // Nothing to do
    //
}

void
OCI::BiDir::CallbackAcceptor_impl::listen()
    throw(SystemException)
{
    //
    // Nothing to do
    //
}

OCI::Transport_ptr
OCI::BiDir::CallbackAcceptor_impl::accept(Boolean block)
    throw(SystemException)
{
    OCI::Transport_var t;

    {
        JTCSynchronized synchronized(transportsMonitor_);

        //
        // Are there transports waiting to be accepted?
        //
        if(transports_.size() > 0)
        {
            t = transports_.front();
            transports_.pop_front();
        }
        else if(block)
        {
            while(transports_.empty())
            {
                try
                {
                    transportsMonitor_.wait();
                }
                catch(const JTCInterruptedException&)
                {
                }
            }

            t = transports_.front();
            transports_.pop_front();
        }
    }

    //
    // Call callbacks
    //
    if(!is_nil(t))
    {
        OCI::TransportInfo_var info = t -> get_info();
        try
        {
            info_ -> _OB_callAcceptCB(info);
        }
        catch(const SystemException&)
        {
            t -> close();
            throw;
        }
    }

    //
    // Return new transport
    //
    return t._retn();
}

OCI::Transport_ptr
OCI::BiDir::CallbackAcceptor_impl::connect_self()
    throw(SystemException)
{
    OCI::Transport_var t = new SelfTransport_impl();

    {
        JTCSynchronized synchronized(transportsMonitor_);

        transports_.push_back(t);
        transportsMonitor_.notifyAll();
    }

    return t._retn();
}

void
OCI::BiDir::CallbackAcceptor_impl::add_profiles(
    const OCI::ProfileInfo& profile_info, OCI::IOR& ref)
    throw(SystemException)
{
    ULong i;

    //
    // Filter components according to BiDirIOP version
    //
    // We should really delegate this to something else
    //
    OCI::TaggedComponentSeq components;
    if(profile_info.major == 1 && profile_info.minor == 0)
    {
        //
        // No components
        //
    }
    else if(profile_info.major == 1 && profile_info.minor == 1)
    {
        //
        // No code set component
        //
        for(i = 0 ; i < profile_info.components.length() ; i++)
            if(profile_info.components[i].tag != IOP::TAG_CODE_SETS)
            {
                ULong len = components.length() + 1;
                components.length(len);
                components[len - 1] = profile_info.components[i];
            }
    }
    else
    {
        components.length(profile_info.components.length());
        for(i = 0 ; i < profile_info.components.length() ; i++)
            components[i] = profile_info.components[i];
    }

    //
    // Create BiDirIOP profile
    //
    BiDirIOP::ProfileBody body;
    body.bidiriop_version.major = 1;
    body.bidiriop_version.minor = 0;
    body.giop_version.major = profile_info.major;
    body.giop_version.minor = profile_info.minor;
    body.peer = transportManager_ -> peer();
    body.object_key.length(profile_info.key.length());
    memcpy(body.object_key.get_buffer(), profile_info.key.get_buffer(),
           profile_info.key.length());
    body.components.length(components.length());
    for(i = 0 ; i < components.length() ; i++)
        body.components[i] = components[i];
    body.profiles.length(0);

    //
    // Append the BiDirIOP profile to the IOR
    //
    OCI::BufferImpl buf;
    OutputStreamImpl out(&buf);
    out._OB_writeEndian();
    body._OB_marshal(&out);
    ULong len = ref.profiles.length();
    ref.profiles.length(len + 1);
    ref.profiles[len].tag = tag();
    ref.profiles[len].profile_data.length(buf.length());
    memcpy(ref.profiles[len].profile_data.get_buffer(), buf.data(),
           buf.length());
}

OCI::ProfileInfoSeq*
OCI::BiDir::CallbackAcceptor_impl::get_local_profiles(const OCI::IOR& ref)
    throw(SystemException)
{
    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;
    for(ULong i = 0 ; i < ref.profiles.length() ; i++)
    {
        if(ref.profiles[i].tag == tag())
        {
            //
            // Get ProfileBody
            //
            const Octet* coct = ref.profiles[i].profile_data.get_buffer();
            ULong len = ref.profiles[i].profile_data.length();
            OCI::BufferImpl buf((Octet*)coct, len);
            InputStreamImpl in(&buf, 0, false, 0, false);
            in._OB_readEndian();
            BiDirIOP::ProfileBody body;
            BiDirIOP::ProfileBody::_OB_unmarshal(body, &in);

            if(strcmp(body.peer, transportManager_ -> peer()) == 0 &&
               body.profiles.length() == 0)
            {
                ULong len = seq -> length();
                seq -> length(len + 1);
                seq[len].major = body.giop_version.major;
                seq[len].minor = body.giop_version.minor;
                seq[len].key.length(body.object_key.length());
                memcpy(seq[len].key.get_buffer(),
                       body.object_key.get_buffer(),
                       body.object_key.length());
                seq[len].components.length(body.components.length());
                for(ULong j = 0 ; j < body.components.length() ; j++)
                    seq[len].components[j] = body.components[j];
            }
        }
    }

    return seq._retn();
}

OCI::AcceptorInfo_ptr
OCI::BiDir::CallbackAcceptor_impl::get_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(info_);
}

void
OCI::BiDir::CallbackAcceptor_impl::_OB_connect(OCI::Transport_ptr t)
{
    JTCSynchronized synchronized(transportsMonitor_);

    transports_.push_back(t);
    transportsMonitor_.notifyAll();
}

// ----------------------------------------------------------------------
// CallbackTransportInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::CallbackTransportInfo_impl::CallbackTransportInfo_impl(
    const char* id,
    OCI::ProfileId tag,
    const char* localPeer,
    const char* remotePeer,
    OCI::Connector_ptr connector,
    OCI::TransportInfo_ptr delegate)
    : id_(id), tag_(tag), localPeer_(localPeer), remotePeer_(remotePeer),
      connectorInfo_(connector -> get_info()),
      delegate_(OCI::TransportInfo::_duplicate(delegate))
{
}

OCI::BiDir::CallbackTransportInfo_impl::CallbackTransportInfo_impl(
    const char* id,
    OCI::ProfileId tag,
    const char* localPeer,
    const char* remotePeer,
    OCI::Acceptor_ptr acceptor,
    OCI::TransportInfo_ptr delegate)
    : id_(id), tag_(tag), localPeer_(localPeer), remotePeer_(remotePeer),
      acceptorInfo_(acceptor -> get_info()),
      delegate_(OCI::TransportInfo::_duplicate(delegate))
{
}

OCI::BiDir::CallbackTransportInfo_impl::~CallbackTransportInfo_impl()
{
}

// ----------------------------------------------------------------------
// CallbackTransportInfo_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::CallbackTransportInfo_impl::id()
    throw(SystemException)
{
    return string_dup(id_);
}

OCI::ProfileId
OCI::BiDir::CallbackTransportInfo_impl::tag()
    throw(SystemException)
{
    return tag_;
}

OCI::ConnectorInfo_ptr
OCI::BiDir::CallbackTransportInfo_impl::connector_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_duplicate(connectorInfo_);
}

OCI::AcceptorInfo_ptr
OCI::BiDir::CallbackTransportInfo_impl::acceptor_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(acceptorInfo_);
}

char*
OCI::BiDir::CallbackTransportInfo_impl::describe()
    throw(SystemException)
{
    String_var result = id_;
    result += " callback transport";
    result += "\nlocal peer: ";
    result += localPeer_;
    result += "\nremote peer: ";
    result += remotePeer_;
    try
    {
        String_var s = delegate_ -> describe();
        result += '\n';
        result += s;
    }
    catch(const NO_RESOURCES&)
    {
        // Ignore: delegate's transport may have been closed

        // TODO: Another alternative is to add a close callback to
        // the delegate info object, and set delegate_ to null
        // when the close callback occurs
    }
    return result._retn();
}

void
OCI::BiDir::CallbackTransportInfo_impl::add_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    closeCBs_.push_back(OCI::CloseCB::_duplicate(cb));
}

void
OCI::BiDir::CallbackTransportInfo_impl::remove_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
        {
            closeCBs_.erase(iter);
            return;
        }
    }
}

void
OCI::BiDir::CallbackTransportInfo_impl::_OB_callCloseCB(
    OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        (*iter) -> close_cb(info);
    }
}

// ----------------------------------------------------------------------
// CallbackTransport_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::CallbackTransport_impl::CallbackTransport_impl(
    const char* id,
    OCI::ProfileId tag,
    ActiveTransport_ptr t,
    OCI::Connector_ptr connector)
    : id_(id), tag_(tag), transport_(ActiveTransport::_duplicate(t)),
      shutdown_(false)
{
    OCI::TransportInfo_var info = t -> info();
    info_ = new CallbackTransportInfo_impl(id_, tag_, t -> localPeer(),
                                           t -> remotePeer(), connector, info);
}

OCI::BiDir::CallbackTransport_impl::CallbackTransport_impl(
    const char* id,
    OCI::ProfileId tag,
    ActiveTransport_ptr t,
    OCI::Acceptor_ptr acceptor)
    : id_(id), tag_(tag), transport_(ActiveTransport::_duplicate(t)),
      shutdown_(false)
{
    OCI::TransportInfo_var info = t -> info();
    info_ = new CallbackTransportInfo_impl(id_, tag_, t -> localPeer(),
                                           t -> remotePeer(), acceptor, info);
}

OCI::BiDir::CallbackTransport_impl::~CallbackTransport_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// CallbackTransport_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::CallbackTransport_impl::id()
    throw(SystemException)
{
    return string_dup(id_);
}

OCI::ProfileId
OCI::BiDir::CallbackTransport_impl::tag()
    throw(SystemException)
{
    return tag_;
}

OCI::SendReceiveMode
OCI::BiDir::CallbackTransport_impl::mode()
    throw(SystemException)
{
    return SendReceive;
}

OCI::Handle
OCI::BiDir::CallbackTransport_impl::handle()
    throw(SystemException)
{
    return -1;
}

void
OCI::BiDir::CallbackTransport_impl::close()
    throw(SystemException)
{
    shutdown_ = true;

    //
    // Call callbacks
    //
    info_ -> _OB_callCloseCB(info_);

    transport_ -> close(true);
}

void
OCI::BiDir::CallbackTransport_impl::shutdown()
    throw(SystemException)
{
    shutdown_ = true;

    transport_ -> shutdown(true);
}

void
OCI::BiDir::CallbackTransport_impl::receive(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    //
    // There is no need to invoke the ActiveTransport if we
    // have been shutdown
    //
    if(shutdown_)
        throw COMM_FAILURE("", MinorRecv, COMPLETED_NO);

    transport_ -> receive(buf, block, true, this);
}

Boolean
OCI::BiDir::CallbackTransport_impl::receive_detect(OCI::Buffer_ptr buf,
                                                   Boolean block)
    throw(SystemException)
{
    //
    // There is no need to invoke the ActiveTransport if we
    // have been shutdown
    //
    if(shutdown_)
        return false;

    return transport_ -> receiveDetect(buf, block, true, this);
}

void
OCI::BiDir::CallbackTransport_impl::receive_timeout(OCI::Buffer_ptr buf,
                                                    ULong timeout)
    throw(SystemException)
{
    //
    // There is no need to invoke the ActiveTransport if we
    // have been shutdown
    //
    if(shutdown_)
        return;

    transport_ -> receiveTimeout(buf, timeout, true, this);
}

Boolean
OCI::BiDir::CallbackTransport_impl::receive_timeout_detect(OCI::Buffer_ptr buf,
                                                           ULong timeout)
    throw(SystemException)
{
    //
    // There is no need to invoke the ActiveTransport if we
    // have been shutdown
    //
    if(shutdown_)
        return false;

    return transport_ -> receiveTimeoutDetect(buf, timeout, true, this);
}

void
OCI::BiDir::CallbackTransport_impl::send(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    transport_ -> send(buf, block, true);
}

Boolean
OCI::BiDir::CallbackTransport_impl::send_detect(OCI::Buffer_ptr buf,
                                                Boolean block)
    throw(SystemException)
{
    return transport_ -> sendDetect(buf, block, true);
}

void
OCI::BiDir::CallbackTransport_impl::send_timeout(OCI::Buffer_ptr buf,
                                                 ULong timeout)
    throw(SystemException)
{
    transport_ -> sendTimeout(buf, timeout, true);
}

Boolean
OCI::BiDir::CallbackTransport_impl::send_timeout_detect(OCI::Buffer_ptr buf,
                                                        ULong timeout)
    throw(SystemException)
{
    return transport_ -> sendTimeoutDetect(buf, timeout, true);
}

OCI::TransportInfo_ptr
OCI::BiDir::CallbackTransport_impl::get_info()
    throw(SystemException)
{
    return OCI::TransportInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// ConFactoryInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::ConFactoryInfo_impl::ConFactoryInfo_impl(
    const char* id,
    OCI::ProfileId tag,
    const char* localPeer,
    OCI::ConFactoryInfo_ptr delegate)
    : id_(id), tag_(tag), localPeer_(localPeer),
      delegate_(OCI::ConFactoryInfo::_duplicate(delegate))
{
}

OCI::BiDir::ConFactoryInfo_impl::~ConFactoryInfo_impl()
{
}

// ----------------------------------------------------------------------
// ConFactoryInfo_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::ConFactoryInfo_impl::id()
    throw(SystemException)
{
    return string_dup(id_);
}

OCI::ProfileId
OCI::BiDir::ConFactoryInfo_impl::tag()
    throw(SystemException)
{
    return tag_;
}

char*
OCI::BiDir::ConFactoryInfo_impl::describe()
    throw(SystemException)
{
    String_var result = id_;
    result += " connector factory";
    result += "\nlocal peer: ";
    result += localPeer_;
    String_var s = delegate_ -> describe();
    result += '\n';
    result += s;
    return result._retn();
}

void
OCI::BiDir::ConFactoryInfo_impl::add_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin();
        iter != connectCBs_.end();
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    connectCBs_.push_back(OCI::ConnectCB::_duplicate(cb));
}

void
OCI::BiDir::ConFactoryInfo_impl::remove_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin();
        iter != connectCBs_.end();
        iter++)
    {
        if((*iter).in() == cb)
        {
            connectCBs_.erase(iter);
            return;
        }
    }
}

OCI::ConnectCBSeq*
OCI::BiDir::ConFactoryInfo_impl::_OB_getConnectCBSeq()
{
    JTCSynchronized synchronized(*this);

    OCI::ConnectCBSeq* seq = new OCI::ConnectCBSeq;
    seq -> length(connectCBs_.size());
    copy_to_seq(connectCBs_.begin(), connectCBs_.end(), *seq);
    return seq;
}

// ----------------------------------------------------------------------
// ConFactory_impl private and protected member implementations
// ----------------------------------------------------------------------

bool
OCI::BiDir::ConFactory_impl::compareBodies(const BiDirIOP::ProfileBody& body1,
                                           const BiDirIOP::ProfileBody& body2)
{
    //
    // Compare versions
    //
    if(body1.giop_version.major != body2.giop_version.major ||
       body1.giop_version.minor != body2.giop_version.minor)
        return false;

    //
    // Compare peer IDs
    //
    if(strcmp(body1.peer, body2.peer) != 0)
        return false;

    //
    // Compare object keys
    //
    if(body1.object_key.length() != body2.object_key.length())
        return false;

    if(memcmp(body1.object_key.get_buffer(), body2.object_key.get_buffer(),
              body1.object_key.length()) != 0)
        return false;

    //
    // Compare nested profiles
    //
    if(body1.profiles.length() != body2.profiles.length())
        return false;
    if(body1.profiles.length() > 0)
    {
        IOP::IOR ior1;
        ior1.type_id = string_dup("");
        ior1.profiles.length(body1.profiles.length());
        ULong i;
        for(i = 0 ; i < body1.profiles.length() ; i++)
            ior1.profiles[i] = body1.profiles[i];
        IOP::IOR ior2;
        ior2.type_id = string_dup("");
        ior2.profiles.length(body2.profiles.length());
        for(i = 0 ; i < body2.profiles.length() ; i++)
            ior2.profiles[i] = body2.profiles[i];
        if(!delegate_ -> equivalent(ior1, ior2))
            return false;
    }

    //
    // OK, found a match
    //
    return true;
}

// ----------------------------------------------------------------------
// ConFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::ConFactory_impl::ConFactory_impl(OCI::ConFactory_ptr delegate,
                                             TransportManager_ptr tm)
    : delegate_(OCI::ConFactory::_duplicate(delegate)),
      transportManager_(TransportManager::_duplicate(tm))
{
    OCI::ConFactoryInfo_var info = delegate -> get_info();
    info_ = new ConFactoryInfo_impl(tm -> id(), tm -> tag(), tm -> peer(),
                                    info);
}

OCI::BiDir::ConFactory_impl::~ConFactory_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// ConFactory_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::ConFactory_impl::id()
    throw(SystemException)
{
    return string_dup(transportManager_ -> id());
}

OCI::ProfileId
OCI::BiDir::ConFactory_impl::tag()
    throw(SystemException)
{
    return transportManager_ -> tag();
}

char*
OCI::BiDir::ConFactory_impl::describe_profile(
    const IOP::TaggedProfile& profile)
    throw(CORBA::SystemException)
{
    assert(profile.tag == tag());

    //
    // Get the profile body
    //
    const Octet* coct = profile.profile_data.get_buffer();
    ULong len = profile.profile_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    ::BiDirIOP::ProfileBody body;
    ::BiDirIOP::ProfileBody::_OB_unmarshal(body, &in);

    String_var result = string_dup("");

    //
    // General info
    //
    result += "bidiriop_version: ";
    result += int(body.bidiriop_version.major);
    result += '.';
    result += int(body.bidiriop_version.minor);
    result += '\n';
    result += "giop_version: ";
    result += int(body.giop_version.major);
    result += '.';
    result += int(body.giop_version.minor);
    result += '\n';
    result += "peer: ";
    result += body.peer;
    result += '\n';
    result += "object_key: (";
    result += body.object_key.length();
    result += ")\n";
    String_var tmp = dump_octets(body.object_key.get_buffer(),
                                 body.object_key.length());
    result += tmp;

    //
    // Components
    //
    ULong i;
    for(i = 0 ; i < body.components.length() ; i++)
    {
        String_var desc = describe_component(body.components[i]);
        result += desc;
    }

    //
    // Nested profiles
    //
    for(i = 0 ; i < body.profiles.length() ; i++)
    {
        result += "Encapsulated profile #";
        result += (i + 1);
        result += ":\n";
        String_var desc = delegate_ -> describe_profile(body.profiles[i]);
        result += desc;
    }


    return result._retn();
}

OCI::ConnectorSeq*
OCI::BiDir::ConFactory_impl::create_connectors(const OCI::IOR& ref,
                                               const PolicyList& policies)
    throw(SystemException)
{
    ULong i;

    //
    // Check whether policies are satisfied
    //
    for(i = 0 ; i < policies.length() ; i++)
    {
        if(policies[i] -> policy_type() == PROTOCOL_POLICY_ID)
        {
            ProtocolPolicy_var p = ProtocolPolicy::_narrow(policies[i]);
            if(!p -> contains(transportManager_ -> id()))
                return new OCI::ConnectorSeq;
        }
    }

    //
    // Create connectors from profiles
    //
    OCI::ConnectCBSeq_var cbs = info_ -> _OB_getConnectCBSeq();
    OCI::ConnectorSeq_var seq = new OCI::ConnectorSeq;
    for(i = 0 ; i < ref.profiles.length() ; i++)
    {
        if(ref.profiles[i].tag == tag())
        {
            //
            // Get ProfileBody
            //
            const Octet* coct = ref.profiles[i].profile_data.get_buffer();
            ULong len = ref.profiles[i].profile_data.length();
            OCI::BufferImpl buf((Octet*)coct, len);
            InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
            BiDirIOP::ProfileBody profile;
            BiDirIOP::ProfileBody::_OB_unmarshal(profile, &in);

            //
            // Create a BiDir connector for each connector returned
            // by our delegate
            //
            if(profile.profiles.length() > 0)
            {
                ULong j;
                IOP::IOR dummy;
                dummy.type_id = ref.type_id;
                dummy.profiles.length(profile.profiles.length());
                for(j = 0 ; j < profile.profiles.length() ; j++)
                    dummy.profiles[j] = profile.profiles[j];
                OCI::ConnectorSeq_var delegates =
                    delegate_ -> create_connectors(dummy, policies);
                for(j = 0 ; j < delegates -> length() ; j++)
                {
                    ULong len = seq -> length();
                    seq -> length(len + 1);
                    seq[len] = new Connector_impl(transportManager_, profile,
                                                  cbs, delegates[j]);
                }
            }
            else
            {
                //
                // If the BiDir profile contains no encapsulated
                // profiles, then we must use an existing transport,
                // so we create a connector with no delegate
                //
                ULong len = seq -> length();
                seq -> length(len + 1);
                seq[len] = new Connector_impl(transportManager_, profile,
                                              cbs, OCI::Connector::_nil());
            }
        }
    }

    return seq._retn();
}

Boolean
OCI::BiDir::ConFactory_impl::equivalent(const OCI::IOR& ior1,
                                        const OCI::IOR& ior2)
    throw(SystemException)
{
    ULong p1, p2, b1, b2;
    ULong cnt1 = 0, cnt2 = 0;
    BiDirIOP::ProfileBody** bodies1 = 0;
    BiDirIOP::ProfileBody** bodies2 = 0;

    //
    // Calculate number of BiDirIOP profiles in ior1
    //
    for(p1 = 0 ; p1 < ior1.profiles.length() ; p1++)
        if(ior1.profiles[p1].tag == tag())
            cnt1++;

    //
    // Calculate number of BiDirIOP profiles in ior2
    //
    for(p2 = 0 ; p2 < ior2.profiles.length() ; p2++)
        if(ior2.profiles[p2].tag == tag())
            cnt2++;

    //
    // Return false now if the number of BiDirIOP profile bodies
    // do not match
    //
    if(cnt1 != cnt2)
        return false;

    //
    // Create an array with all BiDirIOP profile bodies of ior1
    //
    bodies1 = new BiDirIOP::ProfileBody*[cnt1];
    for(p1 = 0, b1 = 0 ; p1 < ior1.profiles.length() ; p1++)
        if(ior1.profiles[p1].tag == tag())
        {
            const Octet* coct = ior1.profiles[p1].profile_data.get_buffer();
            ULong len = ior1.profiles[p1].profile_data.length();
            OCI::BufferImpl buf((Octet*)coct, len);
            InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
            BiDirIOP::ProfileBody* p = new BiDirIOP::ProfileBody;
            BiDirIOP::ProfileBody::_OB_unmarshal(*p, &in);
            bodies1[b1++] = p;
        }

    assert(b1 == cnt1);

    //
    // Create an array with all BiDirIOP profile bodies of ior2
    //
    bodies2 = new BiDirIOP::ProfileBody*[cnt2];
    for(p2 = 0, b2 = 0 ; p2 < ior2.profiles.length() ; p2++)
        if(ior2.profiles[p2].tag == tag())
        {
            const Octet* coct = ior2.profiles[p2].profile_data.get_buffer();
            ULong len = ior2.profiles[p2].profile_data.length();
            OCI::BufferImpl buf((Octet*)coct, len);
            InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
            BiDirIOP::ProfileBody* p = new BiDirIOP::ProfileBody;
            BiDirIOP::ProfileBody::_OB_unmarshal(*p, &in);
            bodies2[b2++] = p;
        }

    assert(b2 == cnt2);

    //
    // Check for profile body matches
    //
    for(b1 = 0 ; b1 < cnt1 ; b1++)
    {
        bool match = false;
        for(b2 = 0 ; b2 < cnt2 ; b2++)
        {
            if(!bodies2[b2])
                continue;
            
            //
            // Compare profile bodies
            //
            if(!compareBodies(*bodies1[b1], *bodies2[b2]))
                continue;
            
            //
            // OK, found a match
            //
            delete bodies2[b2];
            bodies2[b2] = 0;
            match = true;
        }

        if(match)
        {
            delete bodies1[b1];
            bodies1[b1] = 0;
        }
    }

    //
    // Check whether there are any unmatched BiDirIOP profile bodies
    //
    ULong unmatched1 = 0;
    for(b1 = 0 ; b1 < cnt1 ; b1++)
        if(bodies1[b1])
        {
            unmatched1++;
            delete bodies1[b1];
        }

    delete [] bodies1;

    ULong unmatched2 = 0;
    for(b2 = 0 ; b2 < cnt2 ; b2++)
        if(bodies2[b2])
        {
            unmatched2++;
            delete bodies2[b2];
        }

    delete [] bodies2;

    if(unmatched1 != 0 || unmatched2 != 0)
        return false;

    return true;
}

ULong
OCI::BiDir::ConFactory_impl::hash(const OCI::IOR& ref, ULong maximum)
    throw(SystemException)
{
    return delegate_ -> hash(ref, maximum);
}

OCI::ConFactoryInfo_ptr
OCI::BiDir::ConFactory_impl::get_info()
    throw(SystemException)
{
    return OCI::ConFactoryInfo::_duplicate(info_);
}

OCI::BiDir::TransportManager_ptr
OCI::BiDir::ConFactory_impl::_OB_transportManager()
{
    return TransportManager::_duplicate(transportManager_);
}

// ----------------------------------------------------------------------
// ConnectorInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::ConnectorInfo_impl::ConnectorInfo_impl(
    const char* id,
    OCI::ProfileId tag,
    const char* localPeer,
    const char* remotePeer,
    const OCI::ConnectCBSeq& callbacks,
    OCI::ConnectorInfo_ptr delegate)
    : id_(id), tag_(tag), localPeer_(localPeer), remotePeer_(remotePeer),
      delegate_(OCI::ConnectorInfo::_duplicate(delegate))
{
    copy_from_seq(callbacks, OB_STL::back_inserter(connectCBs_));
}

OCI::BiDir::ConnectorInfo_impl::~ConnectorInfo_impl()
{
}

// ----------------------------------------------------------------------
// ConnectorInfo_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::ConnectorInfo_impl::id()
    throw(SystemException)
{
    return string_dup(id_);
}

OCI::ProfileId
OCI::BiDir::ConnectorInfo_impl::tag()
    throw(SystemException)
{
    return tag_;
}

char*
OCI::BiDir::ConnectorInfo_impl::describe()
    throw(SystemException)
{
    String_var result = id_;
    if(is_nil(delegate_))
    {
        result += " callback connector";
        result += "\nlocal peer: ";
        result += localPeer_;
        result += "\nremote peer: ";
        result += remotePeer_;
    }
    else
    {
        result += " connector";
        result += "\nlocal peer: ";
        result += localPeer_;
        result += "\nremote peer: ";
        result += remotePeer_;
        String_var s = delegate_ -> describe();
        result += '\n';
        result += s;
    }
    return result._retn();
}

void
OCI::BiDir::ConnectorInfo_impl::add_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin() ;
        iter != connectCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    connectCBs_.push_back(OCI::ConnectCB::_duplicate(cb));
}

void
OCI::BiDir::ConnectorInfo_impl::remove_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin() ;
        iter != connectCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
        {
            connectCBs_.erase(iter);
            return;
        }
    }
}

void
OCI::BiDir::ConnectorInfo_impl::_OB_callConnectCB(OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin() ;
        iter != connectCBs_.end() ;
        iter++)
    {
        (*iter) -> connect_cb(info);
    }
}

// ----------------------------------------------------------------------
// Connector_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::Connector_impl::Connector_impl(
    TransportManager_ptr transportManager,
    const BiDirIOP::ProfileBody& profile,
    const OCI::ConnectCBSeq& callbacks,
    OCI::Connector_ptr delegate)
    : transportManager_(TransportManager::_duplicate(transportManager)),
      profile_(profile),
      delegate_(OCI::Connector::_duplicate(delegate))
{
    OCI::ConnectorInfo_var info;
    if(!is_nil(delegate_))
        info = delegate_ -> get_info();
    info_ = new ConnectorInfo_impl(transportManager -> id(),
                                   transportManager -> tag(),
                                   transportManager -> peer(),
                                   profile.peer, callbacks, info);
}

OCI::BiDir::Connector_impl::~Connector_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// Connector_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::Connector_impl::id()
    throw(SystemException)
{
    return string_dup(transportManager_ -> id());
}

OCI::ProfileId
OCI::BiDir::Connector_impl::tag()
    throw(SystemException)
{
    return transportManager_ -> tag();
}

OCI::Transport_ptr
OCI::BiDir::Connector_impl::connect()
    throw(SystemException)
{
    //
    // Delegate to TransportManager
    //
    OCI::Transport_var tr =
        transportManager_ -> connect(profile_.peer, this, delegate_);

    if(is_nil(tr))
        throw TRANSIENT("", MinorConnectFailed, COMPLETED_NO);

    //
    // Call callbacks
    //
    OCI::TransportInfo_var info = tr -> get_info();
    try
    {
        info_ -> _OB_callConnectCB(info);
    }
    catch(const SystemException&)
    {
        tr -> close();
        throw;
    }

    return tr._retn();
}

OCI::Transport_ptr
OCI::BiDir::Connector_impl::connect_timeout(ULong timeout)
    throw(SystemException)
{
    //
    // Delegate to TransportManager
    //
    OCI::Transport_var tr =
        transportManager_ -> connectTimeout(profile_.peer, this, delegate_,
                                            timeout);

    //
    // Call callbacks
    //
    if(!is_nil(tr))
    {
        OCI::TransportInfo_var info = tr -> get_info();
        try
        {
            info_ -> _OB_callConnectCB(info);
        }
        catch(const SystemException&)
        {
            tr -> close();
            throw;
        }
    }

    return tr._retn();
}

OCI::ProfileInfoSeq*
OCI::BiDir::Connector_impl::get_usable_profiles(
    const OCI::IOR& ref, const PolicyList& policies)
    throw(SystemException)
{
    ULong i;

    //
    // Make sure that the set of policies is met
    //
    for(i = 0 ; i < policies.length() ; i++)
    {
        if(policies[i] -> policy_type() == PROTOCOL_POLICY_ID)
        {
            ProtocolPolicy_var p = ProtocolPolicy::_narrow(policies[i]);
            if(!p -> contains(transportManager_ -> id()))
                return new OCI::ProfileInfoSeq;
        }
    }

    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;
    for(i = 0 ; i < ref.profiles.length() ; i++)
    {
        if(ref.profiles[i].tag == tag())
        {
            //
            // Get ProfileBody
            //
            const Octet* coct = ref.profiles[i].profile_data.get_buffer();
            ULong len = ref.profiles[i].profile_data.length();
            OCI::BufferImpl buf((Octet*)coct, len);
            InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
            BiDirIOP::ProfileBody body;
            BiDirIOP::ProfileBody::_OB_unmarshal(body, &in);

            if(strcmp(body.peer, profile_.peer) == 0)
            {
                ULong j;

                ULong len = seq -> length();
                seq -> length(len + 1);
                seq[len].major = body.giop_version.major;
                seq[len].minor = body.giop_version.minor;
                seq[len].key.length(body.object_key.length());
                memcpy(seq[len].key.get_buffer(), body.object_key.get_buffer(),
                       body.object_key.length());
                seq[len].id = ref.profiles[i].tag;
                seq[len].index = i;

                //
                // Compose set of components, including any returned
                // by delegate
                //
                seq[len].components.length(body.components.length());
                for(j = 0 ; j < body.components.length() ; j++)
                    seq[len].components[j] = body.components[j];

                if(!is_nil(delegate_))
                {
                    ULong k;

                    OCI::IOR dummyIOR;
                    dummyIOR.type_id = ref.type_id;
                    dummyIOR.profiles.length(body.profiles.length());
                    for(j = 0 ; j < body.profiles.length() ; j++)
                        dummyIOR.profiles[j] = body.profiles[j];
                    OCI::ProfileInfoSeq_var seq2 =
                        delegate_ -> get_usable_profiles(dummyIOR, policies);
                    for(j = 0 ; j < seq2 -> length() ; j++)
                        for(k = 0 ; k < seq2[j].components.length() ; k++)
                        {
                            ULong clen = seq[len].components.length();
                            seq[len].components.length(clen + 1);
                            seq[len].components[clen] = seq2[j].components[k];
                        }
                }
            }
        }
    }

    return seq._retn();
}

Boolean
OCI::BiDir::Connector_impl::equal(OCI::Connector_ptr con)
    throw(SystemException)
{
    Connector_impl* impl = dynamic_cast<Connector_impl*>(con);
    if(impl == 0)
        return false;

    //
    // Compare peer IDs
    //
    if(strcmp(profile_.peer, impl -> profile_.peer) != 0)
        return false;

    //
    // Compare encapsulated profiles
    //
    if(profile_.profiles.length() != impl -> profile_.profiles.length())
        return false;

    //
    // Compare delegates
    //
    if((is_nil(delegate_) && !is_nil(impl -> delegate_)) ||
       (!is_nil(delegate_) && is_nil(impl -> delegate_)))
        return false;
    if(!is_nil(delegate_) && !delegate_ -> equal(impl -> delegate_))
        return false;

    //
    // OK, connectors are the same
    //
    return true;
}

OCI::ConnectorInfo_ptr
OCI::BiDir::Connector_impl::get_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// SelfTransportInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::SelfTransportInfo_impl::SelfTransportInfo_impl()
{
}

OCI::BiDir::SelfTransportInfo_impl::~SelfTransportInfo_impl()
{
}

// ----------------------------------------------------------------------
// SelfTransportInfo_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::SelfTransportInfo_impl::id()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

OCI::ProfileId
OCI::BiDir::SelfTransportInfo_impl::tag()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

OCI::ConnectorInfo_ptr
OCI::BiDir::SelfTransportInfo_impl::connector_info()
    throw(SystemException)
{
    assert(false);
    return OCI::ConnectorInfo::_nil(); // to satisfy the compiler
}

OCI::AcceptorInfo_ptr
OCI::BiDir::SelfTransportInfo_impl::acceptor_info()
    throw(SystemException)
{
    assert(false);
    return OCI::AcceptorInfo::_nil(); // to satisfy the compiler
}

char*
OCI::BiDir::SelfTransportInfo_impl::describe()
    throw(SystemException)
{
    return string_dup("bidir shutdown transport");
}

void
OCI::BiDir::SelfTransportInfo_impl::add_close_cb(OCI::CloseCB_ptr)
    throw(SystemException)
{
    assert(false);
}

void
OCI::BiDir::SelfTransportInfo_impl::remove_close_cb(OCI::CloseCB_ptr)
    throw(SystemException)
{
    assert(false);
}

// ----------------------------------------------------------------------
// SelfTransport_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::SelfTransport_impl::SelfTransport_impl()
    : info_(new SelfTransportInfo_impl), closed_(false)
{
}

OCI::BiDir::SelfTransport_impl::~SelfTransport_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// SelfTransport_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::SelfTransport_impl::id()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

OCI::ProfileId
OCI::BiDir::SelfTransport_impl::tag()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

OCI::SendReceiveMode
OCI::BiDir::SelfTransport_impl::mode()
    throw(SystemException)
{
    return SendReceive;
}

OCI::Handle
OCI::BiDir::SelfTransport_impl::handle()
    throw(SystemException)
{
    throw NO_IMPLEMENT();
    return -1; // to satisfy the compiler
}

void
OCI::BiDir::SelfTransport_impl::close()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    closed_ = true;
    notifyAll();
}

void
OCI::BiDir::SelfTransport_impl::shutdown()
    throw(SystemException)
{
    //
    // Nothing to do
    //
}

void
OCI::BiDir::SelfTransport_impl::receive(OCI::Buffer_ptr /* buf */,
                                        Boolean block)
    throw(SystemException)
{
    if(block)
    {
        while(!closed_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }

    if(closed_)
        throw COMM_FAILURE();
}

Boolean
OCI::BiDir::SelfTransport_impl::receive_detect(OCI::Buffer_ptr /* buf */,
                                               Boolean block)
    throw(SystemException)
{
    if(block)
    {
        while(!closed_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }

    return !closed_;
}

void
OCI::BiDir::SelfTransport_impl::receive_timeout(OCI::Buffer_ptr /* buf */,
                                                ULong timeout)
    throw(SystemException)
{
    try
    {
        wait(timeout);
    }
    catch(const JTCInterruptedException&)
    {
    }

    if(closed_)
        throw COMM_FAILURE();
}

Boolean
OCI::BiDir::SelfTransport_impl::receive_timeout_detect(OCI::Buffer_ptr /* buf */,
                                                       ULong timeout)
    throw(SystemException)
{
    try
    {
        wait(timeout);
    }
    catch(const JTCInterruptedException&)
    {
    }

    return !closed_;
}

void
OCI::BiDir::SelfTransport_impl::send(OCI::Buffer_ptr /* buf */, Boolean block)
    throw(SystemException)
{
    if(block)
    {
        while(!closed_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }
}

Boolean
OCI::BiDir::SelfTransport_impl::send_detect(OCI::Buffer_ptr /* buf */,
                                            Boolean block)
    throw(SystemException)
{
    if(block)
    {
        while(!closed_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }

    return !closed_;
}

void
OCI::BiDir::SelfTransport_impl::send_timeout(OCI::Buffer_ptr /* buf */,
                                             ULong timeout)
    throw(SystemException)
{
    try
    {
        wait(timeout);
    }
    catch(const JTCInterruptedException&)
    {
    }

    if(closed_)
        throw COMM_FAILURE();
}

Boolean
OCI::BiDir::SelfTransport_impl::send_timeout_detect(OCI::Buffer_ptr /* buf */,
                                                    ULong timeout)
    throw(SystemException)
{
    try
    {
        wait(timeout);
    }
    catch(const JTCInterruptedException&)
    {
    }

    return !closed_;
}

OCI::TransportInfo_ptr
OCI::BiDir::SelfTransport_impl::get_info()
    throw(SystemException)
{
    return OCI::TransportInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// TransportInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::TransportInfo_impl::TransportInfo_impl(
    const char* id,
    OCI::ProfileId tag,
    const char* localPeer,
    const char* remotePeer,
    OCI::Connector_ptr connector,
    OCI::TransportInfo_ptr delegate)
    : id_(id), tag_(tag), localPeer_(localPeer), remotePeer_(remotePeer),
      connectorInfo_(connector -> get_info()),
      delegate_(OCI::TransportInfo::_duplicate(delegate))
{
}

OCI::BiDir::TransportInfo_impl::TransportInfo_impl(
    const char* id,
    OCI::ProfileId tag,
    const char* localPeer,
    const char* remotePeer,
    OCI::Acceptor_ptr acceptor,
    OCI::TransportInfo_ptr delegate)
    : id_(id), tag_(tag), localPeer_(localPeer), remotePeer_(remotePeer),
      acceptorInfo_(acceptor -> get_info()),
      delegate_(OCI::TransportInfo::_duplicate(delegate))
{
}

OCI::BiDir::TransportInfo_impl::~TransportInfo_impl()
{
}

// ----------------------------------------------------------------------
// TransportInfo_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::TransportInfo_impl::id()
    throw(SystemException)
{
    return string_dup(id_);
}

OCI::ProfileId
OCI::BiDir::TransportInfo_impl::tag()
    throw(SystemException)
{
    return tag_;
}

OCI::ConnectorInfo_ptr
OCI::BiDir::TransportInfo_impl::connector_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_duplicate(connectorInfo_);
}

OCI::AcceptorInfo_ptr
OCI::BiDir::TransportInfo_impl::acceptor_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(acceptorInfo_);
}

char*
OCI::BiDir::TransportInfo_impl::describe()
    throw(SystemException)
{
    String_var result = id_;
    result += " transport";
    result += "\nlocal peer: ";
    result += localPeer_;
    result += "\nremote peer: ";
    result += remotePeer_;
    String_var s = delegate_ -> describe();
    result += '\n';
    result += s;
    return result._retn();
}

void
OCI::BiDir::TransportInfo_impl::add_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    closeCBs_.push_back(OCI::CloseCB::_duplicate(cb));
}

void
OCI::BiDir::TransportInfo_impl::remove_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
        {
            closeCBs_.erase(iter);
            return;
        }
    }
}

void
OCI::BiDir::TransportInfo_impl::_OB_callCloseCB(
    OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        (*iter) -> close_cb(info);
    }
}

// ----------------------------------------------------------------------
// Transport_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::Transport_impl::Transport_impl(
    const char* id,
    OCI::ProfileId tag,
    ActiveTransport_ptr t,
    OCI::Connector_ptr connector)
    : id_(id), tag_(tag), transport_(ActiveTransport::_duplicate(t))
{
    OCI::TransportInfo_var info = t -> info();
    info_ = new TransportInfo_impl(id_, tag_, t -> localPeer(),
                                   t -> remotePeer(), connector, info);
}

OCI::BiDir::Transport_impl::Transport_impl(
    const char* id,
    OCI::ProfileId tag,
    ActiveTransport_ptr t,
    OCI::Acceptor_ptr acceptor)
    : id_(id), tag_(tag), transport_(ActiveTransport::_duplicate(t))
{
    OCI::TransportInfo_var info = t -> info();
    info_ = new TransportInfo_impl(id_, tag_, t -> localPeer(),
                                   t -> remotePeer(), acceptor, info);
}

OCI::BiDir::Transport_impl::~Transport_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// Transport_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::BiDir::Transport_impl::id()
    throw(SystemException)
{
    return string_dup(id_);
}

OCI::ProfileId
OCI::BiDir::Transport_impl::tag()
    throw(SystemException)
{
    return tag_;
}

OCI::SendReceiveMode
OCI::BiDir::Transport_impl::mode()
    throw(SystemException)
{
    return SendReceive;
}

OCI::Handle
OCI::BiDir::Transport_impl::handle()
    throw(SystemException)
{
    return -1;
}

void
OCI::BiDir::Transport_impl::close()
    throw(SystemException)
{
    //
    // Call callbacks
    //
    info_ -> _OB_callCloseCB(info_);

    transport_ -> close(false);
}

void
OCI::BiDir::Transport_impl::shutdown()
    throw(SystemException)
{
    transport_ -> shutdown(false);
}

void
OCI::BiDir::Transport_impl::receive(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    transport_ -> receive(buf, block, false, 0);
}

Boolean
OCI::BiDir::Transport_impl::receive_detect(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    return transport_ -> receiveDetect(buf, block, false, 0);
}

void
OCI::BiDir::Transport_impl::receive_timeout(OCI::Buffer_ptr buf, ULong timeout)
    throw(SystemException)
{
    transport_ -> receiveTimeout(buf, timeout, false, 0);
}

Boolean
OCI::BiDir::Transport_impl::receive_timeout_detect(OCI::Buffer_ptr buf,
                                                   ULong timeout)
    throw(SystemException)
{
    return transport_ -> receiveTimeoutDetect(buf, timeout, false, 0);
}

void
OCI::BiDir::Transport_impl::send(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    transport_ -> send(buf, block, false);
}

Boolean
OCI::BiDir::Transport_impl::send_detect(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    return transport_ -> sendDetect(buf, block, false);
}

void
OCI::BiDir::Transport_impl::send_timeout(OCI::Buffer_ptr buf, ULong timeout)
    throw(SystemException)
{
    transport_ -> sendTimeout(buf, timeout, false);
}

Boolean
OCI::BiDir::Transport_impl::send_timeout_detect(OCI::Buffer_ptr buf,
                                                ULong timeout)
    throw(SystemException)
{
    return transport_ -> sendTimeoutDetect(buf, timeout, false);
}

OCI::TransportInfo_ptr
OCI::BiDir::Transport_impl::get_info()
    throw(SystemException)
{
    return OCI::TransportInfo::_duplicate(info_);
}
