// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Timer.h>
#include <TransportManager.h>
#include <OCI_BiDir_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OCI::BiDir::TransportManager_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OCI::BiDir::TransportManager_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// TransportManager private and protected member implementations
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// TransportManager constructor and destructor
// ----------------------------------------------------------------------

OCI::BiDir::TransportManager::TransportManager(const char* id,
                                               IOP::ProfileId tag,
                                               const char* localPeer)
    : id_(id), tag_(tag), localPeer_(localPeer), transportMap_(7),
      callbackAcceptor_(0)
{
}

OCI::BiDir::TransportManager::~TransportManager()
{
}

// ----------------------------------------------------------------------
// TransportManager public member implementations
// ----------------------------------------------------------------------

//
// A new connection has been accepted
//
OCI::Transport_ptr
OCI::BiDir::TransportManager::addIncomingTransport(OCI::Acceptor_ptr acceptor,
                                                   OCI::Transport_ptr delegate)
{
    ActiveTransport_var at = new ActiveTransport(localPeer_, this, delegate);

    //
    // For newly created transports, we must first read
    // the remote peer's id. Then we register the
    // transport in our table, so that it is available
    // for callback connections to the same remote peer.
    //
    try
    {
        at -> receiveRemotePeerId();
    }
    catch(const SystemException&)
    {
        at -> close(false);
        throw;
    }

    {
        JTCSynchronized synchronized(transportMapMonitor_);
	
        ActiveTransportWrapper_var atWrapper = new ActiveTransportWrapper(at);
	
        if( transportMap_.containsKey(at -> remotePeer()) )
	{
	    atWrapper -> incrRefcount();
	}
        transportMap_.put(at -> remotePeer(), atWrapper);
	
        transportMapMonitor_.notifyAll();
    }

    return new Transport_impl(id_, tag_, at, acceptor);
}

OCI::Transport_ptr
OCI::BiDir::TransportManager::connect(const char* remotePeer,
                                      OCI::Connector_ptr connector,
                                      OCI::Connector_ptr delegate)
{
    ActiveTransport_var at;
    bool newTransport = false;

    //
    // Check if there is an existing incoming connection that
    // we can use
    //
    {
        JTCSynchronized synchronized(transportMapMonitor_);

        ActiveTransportWrapper_var atWrapper;
	
	transportMap_.get(remotePeer, atWrapper);
	if( !is_nil(atWrapper) )
	    at = atWrapper -> getAt();
    }

    if(is_nil(at) && !is_nil(delegate))
    {
        //
        // Create a new ActiveTransport if we were given a connector
        // delegate
        //
        OCI::Transport_var t = delegate -> connect();
        assert(!is_nil(t));
        at = new ActiveTransport(localPeer_, remotePeer, this, t);
        newTransport = true;
    }

    OCI::Transport_var result;

    if(!is_nil(at))
    {
        if(newTransport)
        {
            //
            // For newly created connections, we must first send our
            // local peer id so that the remote end can make a
            // callback connection.
            //
            at -> sendLocalPeerId();
            result = new Transport_impl(id_, tag_, at, connector);
        }
        else
            result = new CallbackTransport_impl(id_, tag_, at, connector);
    }

    return result._retn();
}

OCI::Transport_ptr
OCI::BiDir::TransportManager::connectTimeout(const char* remotePeer,
                                             OCI::Connector_ptr connector,
                                             OCI::Connector_ptr delegate,
                                             CORBA::ULong timeout)
{
    ActiveTransportWrapper_var atWrapper;
    ActiveTransport_var at;
    bool newTransport = false;

    //
    // Check if there is an existing incoming connection that
    // we can use
    //
    {
        JTCSynchronized synchronized(transportMapMonitor_);
	
	transportMap_.get(remotePeer, atWrapper);
	if( !is_nil(atWrapper) )
	    at = atWrapper -> getAt();
    }

    if(is_nil(at))
    {
        //
        // If there is no existing incoming connection, and we have
        // a connector delegate, then create a new ActiveTransport
        //
        if(!is_nil(delegate))
        {
            OCI::Transport_var t = delegate -> connect_timeout(timeout);
            if(!is_nil(t))
            {
                at = new ActiveTransport(localPeer_, remotePeer, this, t);
                newTransport = true;
            }
        }
        else
        {
            //
            // Wait for an incoming connection until timeout
            //
            JTCSynchronized synchronized(transportMapMonitor_);

            timeval len;
            len.tv_sec = timeout/1000;
            len.tv_usec = (timeout % 1000) * 1000;
            timeval timeDone = OB::TimeAdd(OB::TimeNow(), len);
            while(is_nil(at))
            {
                //
                // Have we timed out?
                //
                timeval now = OB::TimeNow();
                if(OB::TimeCmp(timeDone, now) <= 0)
                    break;

                //
                // No, calculate how long we have to wait now
                //
                timeval diff = OB::TimeSub(timeDone, now);
                ULong interval = diff.tv_sec * 1000 + diff.tv_usec / 1000;

                try
                {
                    transportMapMonitor_.wait(interval);
                }
                catch(const JTCInterruptedException&)
                {
                }

	        transportMap_.get(remotePeer, atWrapper);
	        if( !is_nil(atWrapper) )
	            at = atWrapper -> getAt();
            }
        }
    }

    OCI::Transport_var result;

    if(!is_nil(at))
    {
        if(newTransport)
        {
            //
            // For newly created connections, we must first send our
            // local peer id so that the remote end can make a
            // callback connection.
            //
            at -> sendLocalPeerId();
            result = new Transport_impl(id_, tag_, at, connector);
        }
        else
            result = new CallbackTransport_impl(id_, tag_, at, connector);
    }

    return result._retn();
}

void
OCI::BiDir::TransportManager::setCallbackAcceptor(
    CallbackAcceptor_impl* acceptor)
{
    JTCSynchronized synchronized(callbackAcceptorMonitor_);

    assert((callbackAcceptor_ == 0 && acceptor != 0) ||
           (callbackAcceptor_ != 0 && acceptor == 0));
    callbackAcceptor_ = acceptor;
}

bool
OCI::BiDir::TransportManager::hasCallbackAcceptor()
{
    JTCSynchronized synchronized(callbackAcceptorMonitor_);

    return (callbackAcceptor_ != 0);
}

//
// ActiveTransport received an initial message on the callback
// connection
//
void
OCI::BiDir::TransportManager::callbackConnection(ActiveTransport_ptr t)
{
    //
    // If the transport is outgoing, then we need to notify the
    // acceptor for the callback connection
    //
    if(t -> type() == ActiveTransport::Outgoing)
    {
        JTCSynchronized synchronized(callbackAcceptorMonitor_);

        if(callbackAcceptor_ != 0)
        {
            CallbackTransport_impl* transport =
                new CallbackTransport_impl(id_, tag_, t, callbackAcceptor_);
            callbackAcceptor_ -> _OB_connect(transport);
        }
        else
        {
            //
            // TODO: What if there is no acceptor? The remote end should
            // really receive a TRANSIENT exception.
            //
        }
    }
}

//
// The forward connection of the ActiveTransport has been shutdown
//
void
OCI::BiDir::TransportManager::shutdown(ActiveTransport_ptr t)
{
    if(t -> type() == ActiveTransport::Incoming)
    {
        //
        // Remove the incoming transport from the transport map
        //
        String_var remotePeer = t -> remotePeer();
        if(remotePeer.in() != 0)
        {
            JTCSynchronized synchronized(transportMapMonitor_);
	    
            ActiveTransportWrapper_var atWrapper;

	    transportMap_.get(remotePeer, atWrapper);

	    if( !is_nil(atWrapper) )
	    {
	        atWrapper -> decrRefcount();

	        if(atWrapper -> getRefcount() == 0)
	        {
	            transportMap_.remove(remotePeer);
	        }
	        else
	        {
	            transportMap_.put(remotePeer, atWrapper);
	        }
	    }
        }
    }
}
