// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <OB/CosEventChannelAdmin.h>
#include <OB/CosEventComm_skel.h>

//
// For ORBacus proprietary EventChannelFactory.
//
#include <OB/OBEventChannelFactory.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#include <stdlib.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

//
// Usage message
//
static void
usage(const char* progName)
{
    cerr << "Usage:\n";
    cerr << progName << " [options] [channel-name]\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
    ;
}

//
// Get the appropriate event channel.
//
CosEventChannelAdmin::EventChannel_ptr
GetEventChannel(const char* prog, CORBA::ORB_ptr orb, const char* name)
{
    CosEventChannelAdmin::EventChannel_var e;
    if(name == 0)
    {
	//
	// Get event channel from the initial reference.
	//
	try
	{
	    CORBA::Object_var obj =
		orb -> resolve_initial_references("EventService");
	    e = CosEventChannelAdmin::EventChannel::_narrow(obj);
	}
	catch(const CORBA::ORB::InvalidName&)
	{
	    cerr << prog << ": can't resolve `EventService'" << endl;
	    exit(1);
	}
	
	if(CORBA::is_nil(e))
	{
	    cerr << prog
		 << ": `EventService' is not an EventChannel object reference"
		 << endl;
	    exit(1);
	}
    }
    else
    {
	//
	// Get event channel factory. Note that this uses ORBacus
	// proprietary methods.
	//
	OBEventChannelFactory::EventChannelFactory_var f;
	try
	{
	    CORBA::Object_var obj =
		orb -> resolve_initial_references("EventChannelFactory");
	    f = OBEventChannelFactory::EventChannelFactory::_narrow(obj);
	}
	catch(const CORBA::ORB::InvalidName&)
	{
	    cerr << prog << ": can't resolve `EventChannelFactory'" << endl;
	    exit(1);
	}
	
	if(CORBA::is_nil(f))
	{
	    cerr << prog
		 << ": `EventChannelFactory' is not an "
		 << "EventChannelFactory object reference"
		 << endl;
	    exit(1);
	}
	
	//
	// Get event channel.
	//
	try
	{
	    e = f -> get_channel_by_id(name);
	}
	catch(const OBEventChannelFactory::ChannelNotAvailable&)
	{
	    try
	    {
		e = f -> create_channel(name);
		cerr << prog << ": `Channel " << name << "' created"
		     << endl;
	    }
	    catch(const OBEventChannelFactory::ChannelAlreadyExists&)
	    {
		cerr << prog << ": Channel exists: try again" << endl;
		exit(1);
	    }
	}
    }
    
    return e._retn();
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Parse command line arguments.
    //
    const char* channelName = 0;
    for(int i = 1; i < argc; ++i)
    {
        if(strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0)
	{
	    usage(argv[0]);
	    return 0;
        }
        else if(*argv[i] != '-')
        {
    	    channelName = argv[i];
        }
        else
        {
    	    cerr << argv[0] << ": unknown option `"
    	         << argv[i] << "'\n" << endl;
    	    usage(argv[0]);
    	    return 1;
        }
    }

    //
    // Get event channel.
    //
    CosEventChannelAdmin::EventChannel_var e =
        GetEventChannel(argv[0], orb, channelName);

    //
    // Get ProxyPullSupplier.
    //
    CosEventChannelAdmin::ConsumerAdmin_var consumerAdmin =
        e -> for_consumers();
    CosEventChannelAdmin::ProxyPullSupplier_var supplier =
        consumerAdmin -> obtain_pull_supplier();
	
    //
    // Since callbacks are not needed a PushSupplier_impl is not
    // created.
    //
    supplier -> connect_pull_consumer(
        CosEventComm::PullConsumer::_nil());

    //
    // Main loop
    //
    while(true)
    {
        CORBA::Any_var any;
        CORBA::Boolean hasEvent;
        try
        {
    	    //
    	    // We use try_pull so that the application can be
    	    // stopped.  The other approach would be to use DII,
    	    // and send_deferred() and then poll for the result.
    	    //
    	    //any = supplier -> try_pull(hasEvent);
    	    any = supplier -> pull();
    	    hasEvent = true;
        }
        catch(const CORBA::SystemException& ex)
        {
    	    cerr << ex << endl;
    	    break;
        }
        catch(const CosEventComm::Disconnected&)
        {
    	    break;
        }
        
        if(hasEvent)
        {
    	    const char* s;
    	    if(any >>= s)
    	        cout << s << endl;
        }
    }
    
    try
    {
        supplier -> disconnect_pull_supplier();
    }
    catch(const CORBA::TRANSIENT&)
    {
        // Expected if event service has gone away.
    }
 
    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
