// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <OB/Timer.h> // For OB::TimerList.

#include <OB/CosEventChannelAdmin.h>
#include <OB/CosEventComm_skel.h>

//
// For ORBacus proprietary EventChannelFactory.
//
#include <OB/OBEventChannelFactory.h>

#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

// ----------------------------------------------------------------------
// PushConsumer_impl implementation
// ----------------------------------------------------------------------

class PullSupplier_impl : virtual public POA_CosEventComm::PullSupplier,
                          virtual public PortableServer::RefCountServantBase
{
    CORBA::ORB_var orb_;
    PortableServer::POA_var poa_;
    bool slow_;

public:

    PullSupplier_impl(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa,
                      bool s)
	: orb_(CORBA::ORB::_duplicate(orb)),
	  poa_(PortableServer::POA::_duplicate(poa)),
	  slow_(s)
    {
    }

    ~PullSupplier_impl()
    {
    }

    //
    // IDL/C++ mapping
    //
    virtual void
    disconnect_pull_supplier()
	throw(CORBA::SystemException)
    {
	PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
	poa_ -> deactivate_object(oid.in());
	
	orb_ -> shutdown(false);
    }

    virtual CORBA::Any*
    pull()
	throw(CORBA::SystemException)
    {
	char str[256];
	time_t clock;

	clock = time(0);
	sprintf(str, "PullSupplier says: %s", ctime(&clock));
	str[strlen(str) - 1] = 0;

	CORBA::Any* any = new CORBA::Any;
	*any <<= str;

	if(slow_)
	{
#ifdef WIN32
	    Sleep(2000);
#else
	    sleep(2);
#endif        
	}

	return any;
    }

    virtual CORBA::Any*
    try_pull(CORBA::Boolean& hasEvent)
	throw(CORBA::SystemException)
    {
	hasEvent = true;
	return pull();
    }

    virtual PortableServer::POA_ptr
    _default_POA()
    {
	return PortableServer::POA::_duplicate(poa_);
    }
};

// ----------------------------------------------------------------------
// ShutdownThread implementation
// ----------------------------------------------------------------------

class ShutdownThread : public JTCThread, public JTCMonitor
{
    bool destroy_;
    long timeout_;
    CORBA::ORB_var orb_; // The ORB.

public:

    //
    // Constructor.
    //
    ShutdownThread(CORBA::ORB_ptr orb, long timeout)
    	: destroy_(false),
	  timeout_(timeout),
	  orb_(CORBA::ORB::_duplicate(orb))
    {
    }

    //
    // Called when the thread is started
    //
    virtual void
    run()
    {
	JTCSynchronized sync(*this);
    	try
	{
	    wait(timeout_ * 1000);
	}
	catch(const JTCInterruptedException&)
	{
	}

	if(!destroy_)
	    orb_ -> shutdown(false);
    }

    void
    destroy()
    {
	JTCSynchronized sync(*this);
	destroy_ = true;
	notifyAll();
    }
};
typedef JTCHandleT<ShutdownThread> ShutdownThreadHandle;

//
// Usage message
//
static void
usage(const char* progName)
{
    cerr << "Usage:\n";
    cerr << progName << " [options]\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-s, --slow             Run in `slow' mode.\n"
        "-t sec, --timeout sec  Run for [sec] seconds, -1 forever.\n"
        "-n, --no-register      Don't register with the event service.\n"
	;
}

//
// Get the appropriate event channel.
//
CosEventChannelAdmin::EventChannel_ptr
GetEventChannel(const char* prog, CORBA::ORB_ptr orb, const char* name)
{
    CosEventChannelAdmin::EventChannel_var e;
    if(name == 0)
    {
	//
	// Get event channel from the initial reference.
	//
	try
	{
	    CORBA::Object_var obj =
		orb -> resolve_initial_references("EventService");
	    e = CosEventChannelAdmin::EventChannel::_narrow(obj);
	}
	catch(const CORBA::ORB::InvalidName&)
	{
	    cerr << prog << ": can't resolve `EventService'" << endl;
	    exit(1);
	}
	
	if(CORBA::is_nil(e))
	{
	    cerr << prog
		 << ": `EventService' is not an EventChannel object reference"
		 << endl;
	    exit(1);
	}
    }
    else
    {
	//
	// Get event channel factory. Note that this uses ORBacus
	// proprietary methods.
	//
	OBEventChannelFactory::EventChannelFactory_var f;
	try
	{
	    CORBA::Object_var obj =
		orb -> resolve_initial_references("EventChannelFactory");
	    f = OBEventChannelFactory::EventChannelFactory::_narrow(obj);
	}
	catch(const CORBA::ORB::InvalidName&)
	{
	    cerr << prog << ": can't resolve `EventChannelFactory'" << endl;
	    exit(1);
	}
	
	if(CORBA::is_nil(f))
	{
	    cerr << prog
		 << ": `EventChannelFactory' is not an "
		 << "EventChannelFactory object reference"
		 << endl;
	    exit(1);
	}
	
	//
	// Get event channel.
	//
	try
	{
	    e = f -> get_channel_by_id(name);
	}
	catch(const OBEventChannelFactory::ChannelNotAvailable&)
	{
	    try
	    {
		e = f -> create_channel(name);
		cerr << prog << ": `Channel " << name << "' created"
		     << endl;
	    }
	    catch(const OBEventChannelFactory::ChannelAlreadyExists&)
	    {
		cerr << prog << ": Channel exists: try again" << endl;
		exit(1);
	    }
	}
    }
    
    return e._retn();
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Resolve Root POA.
    //
    PortableServer::POA_var rootPOA =
	PortableServer::POA::_narrow(CORBA::Object_var(
	    orb -> resolve_initial_references("RootPOA")));

    //
    // Get a reference to the POA manager.
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

    //
    // Parse command line arguments.
    //
    bool slow = false;
    CORBA::Long timeout = 5;
    bool reg = true;
    const char* channelName = 0;
    for(int i = 1 ; i < argc ; i++)
    {
        if(strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0)
        {
    	    usage(argv[0]);
    	    return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "-s") == 0 || strcmp(argv[i], "--slow") == 0)
        {
    	    slow = true;
        }
        else if(strcmp(argv[i], "-n") == 0 ||
	        strcmp(argv[i], "--no-register") == 0)
	{
	    reg = false;
	}
	else if(strcmp(argv[i], "-t") == 0 || strcmp(argv[i], "--timeout") == 0)
	{
            if(i + 1 >= argc)
            {
                cerr << argv[0] << ": argument expected for " << argv[i]
                     << endl;
                exit(1);
            }
            timeout = atoi(argv[i+1]);
            ++i;
        }
        else if(*argv[i] != '-')
        {
    	    channelName = argv[i];
        }
        else
        {
    	    cerr << argv[0] << ": unknown option `"
    	         << argv[i] << "'\n" << endl;
    	    usage(argv[0]);
    	    return EXIT_FAILURE;
        }
    }

    //
    // Get event channel.
    //
    CosEventChannelAdmin::EventChannel_var e =
        GetEventChannel(argv[0], orb, channelName);	
	
    //
    // Create a persistent POA.
    //
    CORBA::PolicyList pl(2);
    pl.length(2);
    pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
    pl[1] = rootPOA -> create_id_assignment_policy(
        PortableServer::USER_ID);

    PortableServer::POA_var supplierPOA =
        rootPOA -> create_POA("supplier", manager, pl);

    //
    // Create the servant.
    //
    PullSupplier_impl* impl =
        new PullSupplier_impl(orb, supplierPOA, slow);

    //
    // This will cause the servant to be destroyed when `dummy'
    // goes out of scope.
    //
    PortableServer::ServantBase_var dummy = impl;

    //
    // Connect the supplier to the POA with the same name each
    // time.
    //
    PortableServer::ObjectId_var oid =
        PortableServer::string_to_ObjectId("DefaultSupplier");
    supplierPOA -> activate_object_with_id(oid.in(), impl);
    CosEventComm::PullSupplier_var supplier = impl -> _this();
        
    CosEventChannelAdmin::ProxyPullConsumer_var consumer;        
    if(reg)
    {
	//
	// Get ProxyPullConsumer.
	//
	CosEventChannelAdmin::SupplierAdmin_var supplierAdmin =
	    e -> for_suppliers();
	consumer = supplierAdmin -> obtain_pull_consumer();
        consumer -> connect_pull_supplier(supplier);
    }

    //
    // Activate the POAManager.
    //	      
    manager -> activate();

    ShutdownThreadHandle t;

    if(timeout > 0)
    {
	//
	// Create an instance of the ShutdownThread
	//
        t = new ShutdownThread(orb, timeout);

	//
	// Start a thread which will shutdown the ORB after timeout
	// seconds
	//
	t -> start();
    }
                
    //
    // Run implementation.
    //
    orb -> run();

    //
    // If the shutdown thread exists then tell it to go away
    //
    if(t)
    {
        t -> destroy();
	//
	// Wait for the thread to terminate
	//
	while(t -> isAlive())
	{
	    try
	    {
	        t -> join();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }

    try
    {
	//
	// Disconnect the pull consumer here to stop the event
	// service from pushing further events.
	//
        if(reg)
    	    consumer -> disconnect_pull_consumer();

	orb -> shutdown(true);
    }
    //
    // Expected if ORB/Client has already been shutdown.
    //
    catch(const CORBA::TRANSIENT&)
    {
    }
    catch(const CORBA::BAD_INV_ORDER&)
    {
    }
    catch(const CORBA::INITIALIZE&)
    {
    }
    catch(const CORBA::NO_RESPONSE&)
    {
    }

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);

	//
	// Calling disconnect_* can cause the event service to call
	// back to inform the proxy that it is disconnecting
	// it. Therefore it's necessary to allow nested method
	// invocations.
	//
	properties -> setProperty("ooc.orb.conc_model", "reactive");
	properties -> setProperty("ooc.orb.oa.conc_model", "reactive");

        //
        // Create ORB.
        //
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
