// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <OB/Timer.h> // For OB::TimerList.

#include <OB/CosTypedEventChannelAdmin.h>
#include <OB/CosEventComm_skel.h>

//
// For the ORBacus proprietary TypedEventChannelFactory.
//
#include <OB/OBTypedEventChannelFactory.h>

#include <TimerUpdate_skel.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#include <stdlib.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

// ----------------------------------------------------------------------
// TimerUpdate_impl implementation
// ----------------------------------------------------------------------

class TimerUpdate_impl : virtual public POA_TypedTimerUpdate,
                         virtual public PortableServer::RefCountServantBase
{
    CORBA::ORB_var orb_; // The ORB.
    PortableServer::POA_var poa_; // My POA.
    bool slow_; // Is this a slow consumer?

public:
    
    TimerUpdate_impl(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa, bool s)
        : orb_(CORBA::ORB::_duplicate(orb)),
	  poa_ (PortableServer::POA::_duplicate(poa)),
	  slow_(s)
    {
    }
    
    //
    // IDL to C++ mapping
    //
    virtual void
    time_update(const char* str)
	throw(CORBA::SystemException)
    {
        cout << str << endl;
        
        if(slow_)
        {
#ifdef WIN32
            Sleep(2000);
#else
            sleep(2);
#endif        
        }
    }

    virtual void
    push(const CORBA::Any& /* data */)
	throw(CORBA::SystemException)
    {
    }

    virtual void
    disconnect_push_consumer()
	throw(CORBA::SystemException)
    {
        PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
        poa_ -> deactivate_object(oid.in());

	orb_ -> shutdown(false);
    }

    virtual PortableServer::POA_ptr
    _default_POA()
    {
        return PortableServer::POA::_duplicate(poa_);
    }    
};

// ----------------------------------------------------------------------
// ShutdownThread implementation
// ----------------------------------------------------------------------

class ShutdownThread : public JTCThread, public JTCMonitor
{
    bool destroy_;
    long timeout_;
    CORBA::ORB_var orb_; // The ORB.

public:

    //
    // Constructor.
    //
    ShutdownThread(CORBA::ORB_ptr orb, long timeout)
    	: destroy_(false),
	  timeout_(timeout),
	  orb_(CORBA::ORB::_duplicate(orb))
    {
    }

    //
    // Called when the thread is started
    //
    virtual void
    run()
    {
	JTCSynchronized sync(*this);
    	try
	{
	    wait(timeout_ * 1000);
	}
	catch(const JTCInterruptedException&)
	{
	}

	if(!destroy_)
	    orb_ -> shutdown(false);
    }

    void
    destroy()
    {
	JTCSynchronized sync(*this);
	destroy_ = true;
	notifyAll();
    }
};
typedef JTCHandleT<ShutdownThread> ShutdownThreadHandle;

//
// Usage message.
//
static void
usage(const char* progName)
{
    cerr << "Usage:\n";
    cerr << progName << " [options] [channel-name]\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-s, --slow             Run in `slow' mode.\n"
        "-t sec, --time sec     Run for [sec] seconds, -1 forever.\n"
	"-n, --no-register      Don't register with the event service.\n"
	;
}

//
// Get the appropriate event channel.
//
CosTypedEventChannelAdmin::TypedEventChannel_ptr
GetTypedEventChannel(const char* prog, CORBA::ORB_ptr orb, const char* name)
{
    CosTypedEventChannelAdmin::TypedEventChannel_var e;
    if(name == 0)
    {
	//
	// Get typed event channel from the initial reference.
	//
	try
	{
	    CORBA::Object_var obj =
		orb -> resolve_initial_references("TypedEventService");
	    e = CosTypedEventChannelAdmin::TypedEventChannel::_narrow(obj);
	}
	catch(const CORBA::ORB::InvalidName&)
	{
	    cerr << prog << ": can't resolve `TypedEventService'" << endl;
	    exit(1);
	}
	
	if(CORBA::is_nil(e))
	{
	    cerr << prog
		 << ": `TypedEventService' is not an TypedEventChannel "
		 << "object reference"
		 << endl;
	    exit(1);
	}
    }
    else
    {
	//
	// Get typed event channel factory. Note that this uses
	// ORBacus proprietary methods.
	//
	OBTypedEventChannelFactory::TypedEventChannelFactory_var f;
	try
	{
	    CORBA::Object_var obj =
		orb -> resolve_initial_references(
		    "TypedEventChannelFactory");
	    f = OBTypedEventChannelFactory::
		TypedEventChannelFactory::_narrow(obj);
	}
	catch(const CORBA::ORB::InvalidName&)
	{
	    cerr << prog << ": can't resolve `TypedEventChannelFactory'"
		 << endl;
	    exit(1);
	}
	
	if(CORBA::is_nil(f))
	{
	    cerr << prog
		 << ": `TypedEventChannelFactory' is not an "
		 << "TypedEventChannelFactory object reference"
		 << endl;
	    exit(1);
	}
	
	//
	// Get event channel.
	//
	try
	{
	    e = f -> get_channel_by_id(name);
	}
	catch(const OBEventChannelFactory::ChannelNotAvailable&)
	{
	    try
	    {
		e = f -> create_channel(name);
		cerr << prog << ": `Channel " << name << "' created"
		     << endl;
	    }
	    catch(const OBEventChannelFactory::ChannelAlreadyExists&)
	    {
		cerr << prog << ": Channel exists: try again" << endl;
		exit(1);
	    }
	}
    }
    
    return e._retn();
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Resolve Root POA.
    //
    PortableServer::POA_var rootPOA =
	PortableServer::POA::_narrow(CORBA::Object_var(
	    orb -> resolve_initial_references("RootPOA")));

    //
    // Get a reference to the POA manager.
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

    //
    // Parse command line arguments.
    //
    bool slow = false;
    CORBA::Long timeout = 5;
    bool reg = true;
    const char* channelName = 0;
    for(int i = 1 ; i < argc ; i++)
    {
        if(strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0)
        {
            usage(argv[0]);
            return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "-s") == 0 || strcmp(argv[i], "--slow") == 0)
        {
            slow = true;
        }
        else if(strcmp(argv[i], "-n") == 0 ||
                strcmp(argv[i], "--no-register") == 0)
        {
            reg = false;
        }
        else if(strcmp(argv[i], "-t") == 0 || strcmp(argv[i], "--timeout") == 0)
        {
            if(i + 1 >= argc)
            {
                cerr << argv[0] << ": argument expected for " << argv[i]
                     << endl;
                return EXIT_FAILURE;
            }
            timeout = atoi(argv[i+1]);
            ++i;
        }
	else if(*argv[i] != '-')
	{
	    channelName = argv[i];
	}
        else
        {
            cerr << argv[0] << ": unknown option `" << argv[i] << "'\n" << endl;
            usage(argv[0]);
            return EXIT_FAILURE;
        }
    }

    //
    // Get event channel.
    //
    CosTypedEventChannelAdmin::TypedEventChannel_var e =
        GetTypedEventChannel(argv[0], orb, channelName);

    //
    // Create a persistent POA.
    //
    CORBA::PolicyList pl(2);
    pl.length(2);
    pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
    pl[1] = rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);

    PortableServer::POA_var consumerPOA =
        rootPOA -> create_POA("consumer", manager, pl);

    //
    // Create the servant.
    //
    TimerUpdate_impl* impl = new TimerUpdate_impl(orb, consumerPOA, slow);

    //
    // This will cause the servant to be destroyed when `dummy'
    // goes out of scope.
    //
    PortableServer::ServantBase_var dummy = impl;
        
    //
    // Connect the consumer to the POA with the same name each
    // time.
    //
    PortableServer::ObjectId_var oid =
        PortableServer::string_to_ObjectId("DefaultConsumer");
    consumerPOA -> activate_object_with_id(oid.in(), impl);
    CosEventComm::PushConsumer_var consumer = impl -> _this();

    //
    // Activate the POAManager.
    //
    manager -> activate();

    CosEventChannelAdmin::ProxyPushSupplier_var supplier;
    if(reg)
    {
        //
        // Connect implementation to TypedProxyPushSupplier
        //
        CosTypedEventChannelAdmin::TypedConsumerAdmin_var consumerAdmin =
            e -> for_consumers();
	    
        try
        {
    	    supplier = consumerAdmin ->
    	        obtain_typed_push_supplier("IDL:event/TimerUpdate:1.0");
        }
        catch(const CosTypedEventChannelAdmin::NoSuchImplementation&)
        {
	    cerr << "NoSuchImplementation" << endl;
	    return EXIT_FAILURE;
        }
        supplier -> connect_push_consumer(consumer);
    }

    ShutdownThreadHandle t;

    if(timeout > 0)
    {
	//
	// Create an instance of the ShutdownThread
	//
    	t = new ShutdownThread(orb, timeout);	

	//
	// Start a thread which will shutdown the ORB after timeout
	// seconds
	//
	t -> start();
    }
	    
    //
    // Run implementation.
    //
    orb -> run();

    //
    // If the shutdown thread exists then tell it to go away
    //
    if(t)
    {
	t -> destroy();
	//
	// Wait for the thread to terminate
	//
	while(t -> isAlive())
	{
	    try
	    {
		t -> join();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }

    try
    {
	//
	// Disconnect the push supplier here to stop the event
	// service from pushing further events.
	//
	if(reg)
	    supplier -> disconnect_push_supplier();

	orb -> shutdown(true);
    }
    //
    // Expected if ORB/Client has already been shutdown.
    //
    catch(const CORBA::TRANSIENT&)
    {
    }
    catch(const CORBA::BAD_INV_ORDER&)
    {
    }
    catch(const CORBA::INITIALIZE&)
    {
    }
    catch(const CORBA::NO_RESPONSE&)
    {
    }
    
    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);

	//
	// Calling disconnect_* can cause the event service to call
	// back to inform the proxy that it is disconnecting
	// it. Therefore it's necessary to allow nested method
	// invocations.
	//
	properties -> setProperty("ooc.orb.conc_model", "reactive");
	properties -> setProperty("ooc.orb.oa.conc_model", "reactive");

	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
