// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <OB/CosEventChannelAdmin.h>

#if !defined(WIN32)
#    include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_SSTREAM
#   include <sstream>
#else
#   if defined(HAVE_STRSTREA_H)
#       include <strstrea.h>
#   else
#       include <strstream.h>
#   endif
#endif

#ifdef HAVE_FSTREAM
#    include <fstream>
#else
#    include <fstream.h>
#endif

#ifdef HAVE_IOMANIP
#    include<iomanip>
#else
#    include<iomanip.h>
#endif

#ifdef WIN32
#   include <sys/timeb.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

static CORBA::Any*
CreateNewEvent(const char* eventDataFile)
{
    const long MaxArraySize = 5000;
    char s[MaxArraySize] = "";

    if(eventDataFile)
    {
	OB_IFSTREAM in;
	in.open(eventDataFile);
	in.read(s, MaxArraySize - 1);
	in.close();
    }

    CORBA::Any_var any = new CORBA::Any;
    any <<= s;

    return any._retn();
}

static CosEventChannelAdmin::EventChannel_ptr
GetEventChannel(CORBA::ORB_ptr orb, const char* channelRef)
{
    //
    // Get a reference to the Event Service EventChannel
    //
    CosEventChannelAdmin::EventChannel_var eventChannel;
    try
    {
	CORBA::Object_var obj;

	if(channelRef)
	    obj = orb -> string_to_object(channelRef);
	else
	    obj = orb -> resolve_initial_references("EventService");

	//
	// Narrow the CORBA::Object reference to an EventChannel
	// reference so we can invoke its methods
	//
	eventChannel= CosEventChannelAdmin::EventChannel::_narrow(obj);

    }
    catch(const CORBA::ORB::InvalidName&)
    {
	OB_CERR << "Cannot resolve `EventService'" << OB_ENDL;
	exit(1);
    }

    //
    // Make sure the CORBA object was really an EventChannelFactory
    //
    if(CORBA::is_nil(eventChannel))
    {
	OB_CERR << "`EventService' initial reference is not "
	     << "a CosEventChannelAdmin::EventChannel"
	     << OB_ENDL;
	exit(1);
    }

    return eventChannel._retn();
}

int
run(CORBA::ORB_ptr orb, int argc, char** argv)
{
    char* eventDataFile = 0;
    long totalEvents = 0;
    char* outputFile = 0;
    char* channelRef = 0;
    bool verbose = false;

    //
    // Get command line options
    //
    for(int i = 0; i < argc; ++i)
    {
	if(strcmp(argv[i], "--total-events") == 0)
	    totalEvents = atol(argv[++i]);

	else if(strcmp(argv[i], "--event-data-file") == 0)
	    eventDataFile = strdup(argv[++i]);

	else if(strcmp(argv[i], "--output-file") == 0)
	    outputFile = strdup(argv[++i]);

	else if(strcmp(argv[i], "--channel-ref") == 0)
	    channelRef = strdup(argv[++i]);

	else if(strcmp(argv[i], "--verbose") == 0)
	    verbose = true;
    }

    try
    {
	//
	// Get Event Channel
	//
	CosEventChannelAdmin::EventChannel_var eventChannel =
	    GetEventChannel(orb, channelRef);

        //
        // Obtain a Supplier Admin
        //
        CosEventChannelAdmin::SupplierAdmin_var supplierAdmin = 
            eventChannel -> for_suppliers();

        //
        // Obtain a Proxy Consumer -- we are using the "Push" model
        //
        CosEventChannelAdmin::ProxyPushConsumer_var
            proxyPushConsumer = supplierAdmin -> obtain_push_consumer();

        //
        // Connect to the Proxy Consumer
        //
        proxyPushConsumer ->
	    connect_push_supplier(CosEventComm::PushSupplier::_nil());

	//
	// Create an event
        //
	CORBA::Any_var event = CreateNewEvent(eventDataFile);

	//
	// Let things settle down before we start pushing events
	//
#ifdef WIN32
	Sleep(1000);
#else
	sleep(10);
#endif

        //
        // Push events
        //
	long i = 0;
#ifdef HAVE_SSTREAM
	OB_STD(stringstream) sout;
#else
	OB_STD(strstream) sout;
#endif
        while(i < totalEvents)
        {
	    if (i == 0)
	    {
                timeval tv;
#ifdef WIN32
                struct _timeb timebuffer;
                _ftime(&timebuffer);
                tv.tv_sec = timebuffer.time;
                tv.tv_usec = timebuffer.millitm * 1000;
#else
                gettimeofday(&tv, 0);
#endif
		sout << "First event sent by supplier: "
		     << setiosflags(ios::fixed) << setprecision(6)
		     << (double)tv.tv_sec + (double)tv.tv_usec / 1000000.0
		     << OB_ENDL;
	    }

            try
            {
                //
                // Push the event
                //
                proxyPushConsumer -> push(event.in());
		if(verbose && (i % 10) == 0)
		    OB_CERR << "-s-" << i << ' ' << OB_FLUSH;
            }
            catch(const CosEventComm::Disconnected&)
            {
		OB_CERR << "Caught Disconnected Exception in the "
		        << "Supplier while pushing events.  At event " 
			<< i << " of " << totalEvents << OB_ENDL; 
                break;
            }
            catch(const CORBA::SystemException& ex)
            {
		OB_CERR << "Caught System Exception in the Supplier while "
		        << "pushing events.  At event " << i << " of "
		        << totalEvents << OB_ENDL; 
                OB_CERR << ex << OB_ENDL;
                break;
            }
	    ++i;
        }
	if(verbose)
	    OB_CERR << "-s-Disconnect " << OB_FLUSH;

        //
        // Remove the Proxy Push Consumer
        //
        try
        {
            proxyPushConsumer -> disconnect_push_consumer();
        }
        catch(const CORBA::TRANSIENT&)
        {
        }
        catch(const CORBA::COMM_FAILURE&)
        {
        }

	//
	// Display our results
	//
        sout << OB_STD(ends);
	if(outputFile)
	{
	    OB_OFSTREAM out(outputFile);
	    out << sout.str();
	    out.close();
	}
	else
	    OB_COUT << sout.str();
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_CERR << "Supplier" << OB_ENDL;
	OB_CERR << ex << OB_ENDL;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

int
main(int argc, char** argv)
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
        orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_CERR << ex << OB_ENDL;
	return EXIT_FAILURE;
    }

    if (!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
