// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
 
#include <ChannelType.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// ChannelType private member implementation
// ----------------------------------------------------------------------

void
ChannelType::typeIds(InterfaceDef_ptr iface)
{
    typeIds_.push_back(iface -> id());

    InterfaceDefSeq_var base = iface -> base_interfaces();
    for(ULong i = 0; i < base -> length(); ++i)
    {
        //
        // Get the ID.
        //
        RepositoryId_var id = base[i] -> id();

        //
        // Find out whether this ID is already in the set of IDs.
        //
	StringList::iterator iter;
	for(iter = typeIds_.begin() ; iter != typeIds_.end() ; ++iter)
            if(strcmp(*iter, id) == 0)
                break;

        //
        // Nope, then add all the typeIds for this interface.
        //
        if(iter == typeIds_.end())
            typeIds(base[i]);
    }
}

// ----------------------------------------------------------------------
// ChannelType constructor/destructor
// ----------------------------------------------------------------------

ChannelType::ChannelType(ORB_ptr orb)
    : orb_(ORB::_duplicate(orb))
{
}

// ----------------------------------------------------------------------
// ChannelType public member implementation
// ----------------------------------------------------------------------

bool
ChannelType::isCompatible(const char* id)
{
    StringList::iterator iter;
    for(iter = typeIds_.begin() ; iter != typeIds_.end() ; ++iter)
        if(strcmp(*iter, id) == 0)
            return true;
    return false;
}

bool
ChannelType::assignTypeId(const char* id)
{
    //
    // Do we already have a type-id assigned?
    //
    if(typeIds_.size() != 0)
        return isCompatible(id);

    Repository_var ir;
    try
    {
        Object_var irObj =
            orb_ -> resolve_initial_references("InterfaceRepository");
        ir = Repository::_narrow(irObj);
    }
    catch(const ORB::InvalidName&)
    {
	//
	// This is really an internal error since we ensure that there
	// is a InterfaceRepository at process startup.
	//
    }
    //
    // We want COMM_FAILURE and TRANSIENT exceptions to cause
    // INTF_REPOS exceptions to be raised. Another valid choice would
    // be InterfaceNotSupported. However, I have chosen to raise
    // INTF_REPOS in the case that the IR isn't available so that the
    // user can differentiate this case from the ID not present in the
    // IR.
    //
    catch(const COMM_FAILURE&)
    {
    }
    catch(const TRANSIENT&)
    {
    }
    if(is_nil(ir))
        throw INTF_REPOS(OB::MinorNoIntfRepos, COMPLETED_NO);
    
    Contained_var contained = ir -> lookup_id(id);
    
    //
    // Not in the interface repository? Not supported.
    //
    if(is_nil(contained))
        return false;

    InterfaceDef_var iface = InterfaceDef::_narrow(contained);
    if(is_nil(iface))
	return false;

    //
    // Retrieve the type-ids.
    //
    typeIds(iface);

    //
    // Query off operation descriptions for all supported operations.
    //
    ContainedSeq_var ops = iface -> contents(dk_Operation, false);
    
    operations_.length(ops -> length());
    
    for(ULong i = 0; i < ops -> length(); i++)
    {
	Contained::Description_var description = ops[i] -> describe();

	const OperationDescription* opDesc;
	bool rc = ((description -> value) >>= opDesc) ? true : false;
	assert(rc);

	operations_[i] = *opDesc;
    }

    interfaceId_ = iface -> id();

    return true;
}

OpDescriptionSeq* 
ChannelType::getOperations()
{
    return &operations_;
}

RepositoryId
ChannelType::getInterfaceId()
{
    return string_dup(interfaceId_);
}
