// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <OB/CosEventChannelAdmin_skel.h>

#include <EventLogger.h>
#include <Event_impl.h>
#include <EventUtil.h>
#include <PropagatorBaseThread.h>
#include <PipeSegment.h>
#include <ProxyBase.h>
#include <QoS.h>

#include <EC_ProxyPullSupplier_impl.h>

#include <stdlib.h> // for atoi, ...

using namespace CORBA;

// ----------------------------------------------------------------------
// EC_PullPipeSegment implementation
// ----------------------------------------------------------------------

typedef OB_STL::list<Event_impl*> EventList;

class EC_PullPipeSegment : public PipeSegment, public JTCMonitor
{
    OB::EventLogger_var logger_; // The message logger
    EventList events_; // Queue of pending events
    ULong maxEvents_; // Maximum number of events to queue
    ULong id_; // My ID

public:

    EC_PullPipeSegment(OB::EventLogger_ptr logger, ULong maxEvents, ULong id)
	: logger_(OB::EventLogger::_duplicate(logger)),
	  maxEvents_(maxEvents), id_(id)
    {
    }

    ~EC_PullPipeSegment()
    {
	assert(events_.size() == 0);
    }

    virtual bool
    receive(Event_impl* event)
    {
	if(destroy_)
	    return false;
	
	JTCSynchronized sync(*this);
	
	//
	// If we queue more than maxEvents_ items then we throw oldest
	// items away first.
	//
	events_.push_front(event);
	event -> acquire();
    
	if(events_.size() > maxEvents_)
	{

#if defined(_AIX)
            //
            // COMPILERFIX: This needs to be done in order for the
            // code to work on AIX VA C++ 5.0 with optimization.
            //
            EventList::iterator it = events_.end();
            --it;
            Event_impl* e = *it;
#else
            Event_impl* e = events_.back();
#endif
	    if(logger_ -> traceEvents() >= 2)
	    {
		String_var msg = string_dup("ProxyPullSupplier(");
		msg += id_;
		msg += ") Discarding event (";
		msg += e -> eventId();
		msg += ')';
		logger_ ->trace("event", msg);
	    }
	    e -> discard();
	    events_.pop_back();
	}

	//
	// Wake any waiting pull() events.
	//
	notify();

	return true;
    }

    virtual void
    destroy()
    {
	JTCSynchronized sync(*this);

	destroy_ = true;
	notify();

	//
	// Discard each event in the queue
	//
	EventList::iterator iter;
	for(iter = events_.begin() ; iter != events_.end() ; ++iter)
	    (*iter) -> discard();
	events_.clear();
    }

    Event_impl*
    getFirstEvent()
    {
	JTCSynchronized sync(*this);

	if(destroy_ || events_.size() == 0)
	    return 0;

#if defined(_AIX)
        //
        // COMPILERFIX: This needs to be done in order for the
        // code to work on AIX VA C++ 5.0 with optimization.
        //
        EventList::iterator it = events_.end();
        --it;
        Event_impl* event = *it;
#else
        Event_impl* event = events_.back();
#endif
	events_.pop_back();
	return event;
    }

    Event_impl*
    getFirstEventWait()
    {
	JTCSynchronized sync(*this);

	//
	// Block until there is an event, and we're not destroyed
	//
	while(events_.size() == 0 && !destroy_)
	{
	    try
	    {
		wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}

	if(destroy_)
	    return 0;

	assert(events_.size() > 0);
	return getFirstEvent();
    }
};

// ----------------------------------------------------------------------
// EC_ProxyPullSupplier_impl constructor/destructor
// ----------------------------------------------------------------------

EC_ProxyPullSupplier_impl::EC_ProxyPullSupplier_impl(
    ProxyConstructorInfo& info)
{
    info.base = "ProxyPullSupplier";
    ProxyBase::init(info);

    //
    // Resolve the QoS params
    //
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(info.orb);
    OB::Properties_var properties = oborb -> properties();
    maxEvents_ = atoi(properties -> getProperty(EventQoS::MAX_EVENTS));
}

EC_ProxyPullSupplier_impl::~EC_ProxyPullSupplier_impl()
{
}

// ----------------------------------------------------------------------
// Standard IDL to C++ Mapping
// ----------------------------------------------------------------------

void
EC_ProxyPullSupplier_impl::connect_pull_consumer(
    CosEventComm::PullConsumer_ptr consumer)
    throw(CosEventChannelAdmin::AlreadyConnected,
    	  SystemException)
{
    JTCSynchronized sync(*this);
    
    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    if(state_ == Connected)
	throw CosEventChannelAdmin::AlreadyConnected();

    //
    // The consumer is permitted to be nil
    //
    if(!is_nil(consumer))
    {
	QoSParameters params;
	resolveQoSParams(params);
	
	consumer_ = CosEventComm::PullConsumer::_narrow(
	    CORBA::Object_var(applyRequestTimeout(consumer, params))
	);
    }

    stampRequest();
    state_ = Connected;

    pullPipe_ = new EC_PullPipeSegment(logger_, maxEvents_, id_);

    //
    // Ensure that the reference counting is correct
    //
    myPipe_ = pullPipe_;

    //
    // Connect my pipe segment to my parent segment
    //
    parentPipe_ -> connect(pullPipe_);
}

Any*
EC_ProxyPullSupplier_impl::pull()
    throw(CosEventComm::Disconnected,
    	  SystemException)
{
    {
	JTCSynchronized sync(*this);
	canReap_++;
    }

    //
    // Get the first event, and return the data. If myPipe_ is not
    // initialized or the PipeSegment has been destroyed
    // getFirstEventWait() returns 0.
    //
    Event_impl* event = 0;
    if(myPipe_)
	event = pullPipe_ -> getFirstEventWait();

    JTCSynchronized sync(*this);

    canReap_--;

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    if(state_ != Connected)
	throw CosEventComm::Disconnected();

    stampRequest();

    assert(event != 0);
    Any* any = new Any(event -> any());
    event -> discard();

    assert(any != 0);
    return any;
}

Any*
EC_ProxyPullSupplier_impl::try_pull(Boolean& hasEvent)
    throw(CosEventComm::Disconnected,
    	  SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    if(state_ != Connected)
        throw CosEventComm::Disconnected();

    stampRequest();

    Event_impl* event = pullPipe_ -> getFirstEvent();

    //
    // Determine if we have an event.
    //
    hasEvent = (event != 0);

    Any* any;
    if(hasEvent)
    {
        //
        // Yes, then get the first event. Return the data.
        //
        any = new Any(event -> any());
        event -> discard();
    }
    else
    {
        any = new Any;
    }
    return any;
}

void
EC_ProxyPullSupplier_impl::disconnect_pull_supplier()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    disconnectImpl();
}

// ----------------------------------------------------------------------
// ProxyBase implementation
// ----------------------------------------------------------------------

void
EC_ProxyPullSupplier_impl::deactivateObject()
{
    PortableServer::POA_var poa = _default_POA();
    PortableServer::ObjectId_var oid = poa -> servant_to_id(this);
    poa -> deactivate_object(oid.in());
}

void
EC_ProxyPullSupplier_impl::disconnectPeer()
{
    //
    // If we're connected then tell the consumer.
    //
    if(!is_nil(consumer_))
    {
        try
        {
            consumer_ -> disconnect_pull_consumer();
        }
        catch(const SystemException&)
        {
        }
	consumer_ = CosEventComm::PullConsumer::_nil();
    }
}

bool
EC_ProxyPullSupplier_impl::canReap()
{
    JTCSynchronized sync(*this);

    return canReap_ == 0;
}
