// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <OB/CosEventChannelAdmin_skel.h>

#include <EventLogger.h>
#include <Event_impl.h>
#include <PropagatorBaseThread.h>
#include <PusherBaseThread.h>
#include <PipeSegment.h>
#include <ProxyBase.h>

#include <EC_ProxyPushSupplier_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// AnyPusherThread implementation
// ----------------------------------------------------------------------

class AnyPusherThread : public PusherBaseThread
{
    OB::EventLogger_var logger_; // The message logger

    PushResult pushEvent(Event_impl* event)
    {
        assert(event != 0);
        assert(!is_nil(consumer_));

        PushResult rc;

        try
        {
            consumer_ -> push(event -> any());
            rc = Success;
        }
        catch(const COMM_FAILURE&)
        {
            rc = Retry;
        }
	catch(const TRANSIENT&)
	{
	    rc = Retry;
	}
	catch(const NO_RESPONSE&)
	{
	    rc = Retry;
	}
        catch(...)
        {
            rc = Failed;
        }
        return rc;
    }

public:

    AnyPusherThread(CosEventComm::PushConsumer_ptr consumer,
                    CosEventChannelAdmin::ProxyPushSupplier_ptr proxy,
		    OB::EventLogger_ptr logger,
                    ULong id,
                    ULong maxEvents,
                    ULong maxRetries,
                    ULong retryTimeout,
                    Double retryMultiplier)
        : PusherBaseThread(consumer, proxy, logger, id, maxEvents, maxRetries,
                           retryTimeout, retryMultiplier),
	  logger_(OB::EventLogger::_duplicate(logger))
    {
    }
};

//
// PipeSegment for the ProxyPushSupplier
//
class EC_PushPipeSegment : public PipeSegment
{
    PusherBaseThreadHandle pusher_; // The event pusher.

public:
    
    EC_PushPipeSegment(PusherBaseThreadHandle& pusher)
	: pusher_(pusher)
    {
    }
    
    virtual bool
    receive(Event_impl* event)
    {
	if(destroy_)
	    return false;

	pusher_ -> newEvent(event);

	return true;
    }
    
    virtual void
    destroy()
    {
	if(destroy_)
	    return;

	assert(pusher_);
	pusher_ -> halt();
	PipeSegment::destroy();
    }
};

// ----------------------------------------------------------------------
// EC_ProxyPushSupplier_impl constructor/destructor
// ----------------------------------------------------------------------

EC_ProxyPushSupplier_impl::EC_ProxyPushSupplier_impl(
    ProxyConstructorInfo& info)
{
    info.base = "ProxyPushSupplier";
    ProxyBase::init(info);
}

EC_ProxyPushSupplier_impl::~EC_ProxyPushSupplier_impl()
{
}

// ----------------------------------------------------------------------
// Standard IDL to C++ Mapping
// ----------------------------------------------------------------------

void
EC_ProxyPushSupplier_impl::connect_push_consumer(
    CosEventComm::PushConsumer_ptr consumer)
    throw(CosEventChannelAdmin::AlreadyConnected,
	  CosEventChannelAdmin::TypeError,
	  SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
        throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();
    
    if(state_ == Connected)
	throw CosEventChannelAdmin::AlreadyConnected();

    if(is_nil(consumer))
        throw BAD_PARAM();
    
    QoSParameters params;
    resolveQoSParams(params);
    
    consumer_ = CosEventComm::PushConsumer::_narrow(
	CORBA::Object_var(applyRequestTimeout(consumer, params))
    );
    
    state_ = Connected;

    CosEventChannelAdmin::ProxyPushSupplier_var s = _this();

    PusherBaseThreadHandle pusher =
	new AnyPusherThread(consumer_, s, logger_, id_, params.maxEvents,
			    params.maxRetries, params.retryTimeout,
			    params.retryMultiplier);
    pusher -> start();

    assert(!myPipe_);
    myPipe_ = new EC_PushPipeSegment(pusher);
    parentPipe_ -> connect(myPipe_);
}

void
EC_ProxyPushSupplier_impl::disconnect_push_supplier()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    disconnectImpl();
}

// ----------------------------------------------------------------------
// ProxyBase implementation
// ----------------------------------------------------------------------

void
EC_ProxyPushSupplier_impl::deactivateObject()
{
    PortableServer::POA_var poa = _default_POA();
    PortableServer::ObjectId_var oid = poa -> servant_to_id(this);
    poa -> deactivate_object(oid.in());
}

void
EC_ProxyPushSupplier_impl::disconnectPeer()
{
    //
    // If we're connected then tell the consumer.
    //
    if(!is_nil(consumer_))
    {
        try
        {
            consumer_ -> disconnect_push_consumer();
        }                                              
        catch(const SystemException&)
        {
        }
	consumer_ = CosEventComm::PushConsumer::_nil();
    }
}

bool
EC_ProxyPushSupplier_impl::canReap()
{
    return false;
}
