// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/Logger.h>
#include <OB/Hashtable.h>
#include <OB/BootManager.h>
#include <OB/NTService.h>

#include <OB/EventService.h>
#include <OB/TypedEventService.h>

#ifndef MAKEDEPEND
#    include <EventMsg.h>
#endif

#include <NTEventService.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

using namespace CORBA;

// ----------------------------------------------------------------------
// NTEventService constructor
// ----------------------------------------------------------------------

NTEventService::NTEventService()
    : OB::NTService("EventService", "ORBacus Event Service", EVENT_MSG),
      stopped_(false)
{
}

// ----------------------------------------------------------------------
// NTEventService public member implementation
// ----------------------------------------------------------------------

void
NTEventService::orb_run(bool typedService)
{
    //
    // Get the logger
    //
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb_);
    OB::Logger_var logger = oborb -> logger();

    //
    // Resolve Root POA
    //
    Object_var poaObj = orb_ -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj.in());

    //
    // Get a reference to the POA manager
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();
    OBPortableServer::POAManager_var obManager =
        OBPortableServer::POAManager::_narrow(manager);
        
    //
    // Create POA for use by factory
    //
    CORBA::PolicyList policies;
    policies.length(2);
    policies[0] =
        rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] =
        rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);

    PortableServer::POA_var poa;
    try
    {
        poa = rootPOA -> create_POA("EventServicePOA", manager, policies);
    }
    catch(const PortableServer::POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(poa));

    //
    // Resolve the BootManager
    //
    Object_var bmgrObj = orb_ -> resolve_initial_references("BootManager");
    OB::BootManager_var bootManager = OB::BootManager::_narrow(bmgrObj);
        
    OB::EventService untyped(orb_, poa);
    OB::TypedEventService typed(orb_, poa);

    if(!typedService)
    {
	//
	// Initialize the event service
	//
	untyped.initialize();
	untypedFactory_ = untyped.eventChannelFactory();

	//
	// Activate the POAManager
	//
	manager -> activate();

        //
        // Create the default event channel
        // 
	CosEventChannelAdmin::EventChannel_var untypedChannel =
	    untypedFactory_ -> create_channel("DefaultEventChannel");

        //
        // Register the default event channel factory with the
        // BootManager
        //
        PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("DefaultEventChannelFactory");
	bootManager -> add_binding(oid, untypedFactory_);
            
        //
        // Register the default event channel with the BootManager
        //
	oid = PortableServer::string_to_ObjectId("DefaultEventChannel");
	bootManager -> add_binding(oid, untypedChannel);
    }
    else
    {
	try
	{
	    Object_var ir =
	        orb_ -> resolve_initial_references("InterfaceRepository");
	}
	catch(const ORB::InvalidName&)
	{
	    logger -> error("No `InterfaceRepository' initial reference");
	    statusUpdate(SERVICE_STOPPED);
	    return;
	}

	//
	// Initialize the typed event service
	//
	typed.initialize();
	typedFactory_ = typed.typedEventChannelFactory();

	//
	// Activate the POAManager
	//
	manager -> activate();

        //
        // Create the default typed event channel
        // 
	CosTypedEventChannelAdmin::TypedEventChannel_var typedChannel =
	    typedFactory_ -> create_channel("DefaultTypedEventChannel");

        //
        // Register the default event channel factory with the BootManager
        //
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId(
	        "DefaultTypedEventChannelFactory");
	bootManager -> add_binding(oid, typedFactory_);

        //
        // Register the default typed channel with the BootManager.
        //
	oid = PortableServer::string_to_ObjectId("DefaultTypedEventChannel");
	bootManager -> add_binding(oid, typedChannel);
    }

    //
    // Run implementation.
    //
    if(!stopped_)
    {
        //
        // Inform the status manager that the service is running.
        //
        statusUpdate(SERVICE_RUNNING);
        orb_ -> run();
    }
}

void
NTEventService::start(int argc, char** argv, OB::Logger_ptr logger)
{
    try
    {
        bool typedService = false;

	//
	// Setup properties
	//
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);

	//
	// We run the server in threaded/thread-pool by default.
	//
        properties -> setProperty("ooc.orb.conc_model", "threaded");
	const char* oaModel =
	    properties -> getProperty("ooc.orb.oa.conc_model");

	//
	// Set the concurrency model to thread_pool if necessary.
	//
	if(oaModel == 0 ||
	   (strcmp(oaModel, "thread_pool") != 0 &&
	    strcmp(oaModel, "thread_per_request") != 0 &&
	    strcmp(oaModel, "thread_per_client") != 0))
	    properties -> setProperty("ooc.orb.oa.conc_model",
				      "thread_pool");

	OB::Properties::KeySeq keys = properties -> getKeys("ooc.event.");
	ULong i;
	for(i = 0; i < keys.length(); ++i)
	{
	    const char* key = keys[i];
	    const char* value = properties -> getProperty(key);
	    assert(value != 0);
		
            if(strcmp(key, "ooc.event.endpoint") == 0)
            {
                //
                // Set the ooc.orb.oa.port property from this
                // value if not already set
                //
		const char* oaPort = 
		    properties -> getProperty("ooc.orb.oa.endpoint");
		if(oaPort == 0)
                    properties -> setProperty("ooc.orb.oa.endpoint", value);
            }
	    else if(strcmp(key, "ooc.event.typed_service") == 0)
	    {
	        //
	        // Check the ooc.event.typed_service
	        // property to determine if we're running the
	        // typed event service or the untyped.
	        //
	        if(strcmp(value, "true") == 0)
		    typedService = true;
	    }
	}

	orb_ = OBCORBA::ORB_init(argc, argv, OB::Properties::_nil(), logger);
	orb_run(typedService);
    }
    catch(const CORBA::Exception& ex)
    {
        CORBA::String_var err = ex._to_string();
        logger -> error(err);
    }

    if(!CORBA::is_nil(orb_))
    {
	try
	{
	    orb_ -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
            CORBA::String_var err = ex._to_string();
            logger -> error(err);
	}
    }
}

void
NTEventService::stop()
{
    //
    // Mark the implementation as stopped.
    //
    stopped_ = true;

    //
    // Destroy factories, if necessary.
    //
    if(!is_nil(untypedFactory_))
    {
	untypedFactory_ -> shutdown();
	untypedFactory_ = OBEventChannelFactory::EventChannelFactory::_nil();
    }
    if(!is_nil(typedFactory_))
    {
	typedFactory_ -> shutdown();
	typedFactory_ =
	    OBTypedEventChannelFactory::TypedEventChannelFactory::_nil();
    }
}
