// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// *********************************************************************

#include <OB/CORBA.h>

#include <OB/CosEventChannelAdmin.h>

#include <EventLogger.h>
#include <Event_impl.h>
#include <PropagatorBaseThread.h>

#include <PusherBaseThread.h>

using namespace CORBA;
using namespace CosEventComm;
using namespace CosEventChannelAdmin;

// ----------------------------------------------------------------------
// PusherBaseThread private member implementation
// ----------------------------------------------------------------------

//
// Wait until we are terminated or for the next event to
// arrive. Return the event if available, 0 if not.
//
Event_impl*
PusherBaseThread::waitEvent()
{
    JTCSynchronized sync(*this);
    
    while(!done_ && events_.size() == 0)
    {
        try
        {
            wait();
        }
        catch(const JTCInterruptedException&)
        {
        }
    }
    
    Event_impl* event;
    if(!done_)
    {
        assert(events_.size() > 0);
#if defined(_AIX)
        //
        // COMPILERFIX: This needs to be done in order for the
        // code to work on AIX VA C++ 5.0 with optimization.
        //
        EventList::iterator it = events_.end();
        --it;
        event = *it;
#else
        event = events_.back();
#endif

        //
        // Acquire a reference to the event. This is so that if the
        // event is dropped from the event queue, then this event is
        // still valid.
        //
        event -> acquire();
    }
    else
    {
        event = 0;
    }
    return event;
}

void
PusherBaseThread::removeEvent(Event_impl* event)
{
    JTCSynchronized sync(*this);
    
    //
    // We know that either this event is either the last event, or not
    // in the event list at all.
    //
#if defined(_AIX)
    //
    // COMPILERFIX: This needs to be done in order for the
    // code to work on AIX VA C++ 5.0 with optimization.
    //
    EventList::iterator it = events_.end();
    --it;
    if(*it == event)
#else
    if(events_.back() == event)
#endif
    {
        events_.pop_back();
        event -> discard();
    }
}

// ----------------------------------------------------------------------
// PusherBaseThread constructor/destructor
// ----------------------------------------------------------------------

PusherBaseThread::PusherBaseThread(
    PushConsumer_ptr consumer,
    ProxyPushSupplier_ptr proxy,
    OB::EventLogger_ptr logger,
    ULong id,
    ULong maxEvents,
    ULong maxRetries,
    ULong retryTimeout,
    Double retryMultiplier)
    
    : PropagatorBaseThread(logger, id, maxRetries, retryTimeout,
    		   	   retryMultiplier),
      maxEvents_(maxEvents),
      consumer_(PushConsumer::_duplicate(consumer)),
      proxy_(ProxyPushSupplier::_duplicate(proxy))
{
}

PusherBaseThread::~PusherBaseThread()
{
    //
    // Discard each event in the queue
    //
    EventList::iterator iter;
    for(iter = events_.begin() ; iter != events_.end() ; ++iter)
        (*iter) -> discard();
}

// ----------------------------------------------------------------------
// PusherBaseThread public member implementation
// ----------------------------------------------------------------------

//
// Notify the pusher that there is a new event
//
void
PusherBaseThread::newEvent(Event_impl* event)
{
    JTCSynchronized sync(*this);
    
    events_.push_front(event);
    event -> acquire();

    //
    // Too many events in the queue? Drop the last event.
    //
    if(events_.size() > maxEvents_)
    {
#if defined(_AIX)
        //
        // COMPILERFIX: This needs to be done in order for the
        // code to work on AIX VA C++ 5.0 with optimization.
        //
        EventList::iterator it = events_.end();
        --it;
        Event_impl* e = *it;
#else
        Event_impl* e = events_.back();
#endif
	events_.pop_back();
        if(logger_ -> traceEvents() >= 1)
        {
            String_var msg = string_dup("PusherBaseThread(");
            msg += id_;
            msg += ") discarding event (";
            msg += e -> eventId();
            msg += ')';
            logger_ -> trace("event", msg);
        }

        //
        // Discard the event.
        //
        e -> discard();
    }
    
    notify();
}

//
// Run the pusher. Wait for an event, then ask the supplier to
// push the first event.
//
void
PusherBaseThread::run()
{
    Event_impl* event;
    
    while((event = waitEvent()) != 0)
    {
        switch(pushEvent(event))
        {
        case Success:
            resetRetry();
            removeEvent(event);
            event -> discard();
            break;

        case Retry:
            //
            // First, discard the event. Set the event to 0 so we can
            // fall through to the Failed case if we're not to retry
            // again.
            //
            event -> discard();
            event = 0;
            
            //
            // If retry returns true, then we should send the event at
            // the front of the event queue.
            //
            if(retry())
                continue;
            
            if(logger_ -> traceEvents() >= 1)
            {
                String_var msg =
		    string_dup("retries exceeded: disconnecting ");
                msg += id_;
                logger_ -> trace("event", msg);
            }
            //
            // fall through
            //
            
        case Failed:
            if(event != 0)
                event -> discard();
            try
            {
                proxy_ -> disconnect_push_supplier();
            }
            catch(...)
            {
            }
            break;
        }
    }
    
    //
    // We're no longer interested in the ProxyPushSupplier. This is
    // needed to break the circular reference.
    //
    proxy_ = ProxyPushSupplier::_nil();

    //
    // The consumer is no longer of any interest either. We need to
    // clear this reference also.
    //
    consumer_ = PushConsumer::_nil();
}
