// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <EventLogger.h>
#include <ProxyBase.h>

#include <Reaper.h>

#include <stdlib.h>

#ifdef WIN32
#   include <sys/timeb.h>
#endif

#include <time.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// Reaper private member implementation
// ----------------------------------------------------------------------

void
Reaper::reap()
{
    if(logger_ -> traceLifecycle() >= 2)
	logger_ -> trace("event", "Reaping");

    time_t now = time(0);

    //
    // Find the element, remove it from the sequence.
    //
    ProxyPtrList::iterator iter;
    for(iter = proxies_.begin() ; iter != proxies_.end() ; iter++)
    {
	if((*iter) -> destroyed())
	{
            (*iter) -> _remove_ref();
	    proxies_.erase(iter);
	    continue;
        }
	if((*iter) -> canReap())
	{
	    time_t last = (*iter) -> lastRequestTime();
	    if(now - last >= reapTime_)
	    {
		if(logger_ -> traceLifecycle() >= 2)
		{
		    String_var msg = string_dup("Reaping: ");
		    msg += (*iter) -> getName();
		    logger_ -> trace("event", msg);
		}

		(*iter) -> disconnect();
		(*iter) -> _remove_ref();
		proxies_.erase(iter);
		continue;
	    }
	}
    }
}

// ----------------------------------------------------------------------
// Reaper constructor/destructor
// ----------------------------------------------------------------------

Reaper::Reaper(ORB_ptr orb, OB::EventLogger_ptr logger)
    : logger_(OB::EventLogger::_duplicate(logger)),
      done_(false),
      sleepTime_(30*60), // 30 minutes
      reapTime_(60*60 * 4)  // 4 hours
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    OB::Properties_var props = oborb -> properties();

    const char* val = props -> getProperty("ooc.event.reap_frequency");
    if(val != 0)
	sleepTime_ = atol(val);

    val = props -> getProperty("ooc.event.inactivity_timeout");
    if(val != 0)
	reapTime_ = atol(val);
}

Reaper::~Reaper()
{
    ProxyPtrList::iterator iter;
    for(iter = proxies_.begin() ; iter != proxies_.end() ; iter++)
	(*iter) -> _remove_ref();
    proxies_.clear();
}

// ----------------------------------------------------------------------
// Reaper public member implementation
// ----------------------------------------------------------------------

void
Reaper::add(ProxyBase* proxy)
{
    JTCSynchronized sync(*this);

    //
    // When this method is called, canReap() will only return
    // false for proxies that can never be reaped.
    //
    if(proxy -> canReap())
    {
	proxies_.push_back(proxy);
	proxy -> _add_ref();
    }
}

void
Reaper::destroy()
{
    JTCSynchronized sync(*this);
    done_ = true;
    notify();
}

void
Reaper::run()
{
    JTCSynchronized sync(*this);
    if(sleepTime_ == 0)
    {
	if(logger_ -> traceLifecycle() >= 2)
	    logger_ -> trace("event", "Reaping disabled");
	return;
    }

    while(!done_)
    {
	try
	{
	    wait(sleepTime_ * 1000);
	}
	catch(const JTCInterruptedException&)
	{
	}
	if(!done_)
	    reap();
    }
}
