// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <OB/CosTypedEventChannelAdmin_skel.h>

#include <EventLogger.h>
#include <Event_impl.h>
#include <EventUtil.h>
#include <PipeSegment.h>
#include <PropagatorBaseThread.h>
#include <PusherBaseThread.h>
#include <ProxyBase.h>

#include <TEC_ProxyPushSupplier_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// TypedPusherThread implementation
// ----------------------------------------------------------------------

class TypedPusherThread : public PusherBaseThread
{
    OB::EventLogger_var logger_; // The message logger

    PushResult pushEvent(Event_impl* event)
    {
        assert(event != 0);
        assert(!is_nil(consumer_));

        //
        // Create a DII request.
        //
        Request_var request = consumer_ -> _request(event -> op());
        
        //
        // Insert each of the arguments.
        //
        for(ULong i = 0; i < event -> args() -> count(); ++i)
        {
            NamedValue_ptr nv = event -> args() -> item(i);
            Any& arg = request -> add_in_arg(nv -> name());
            arg = *nv -> value();
        }

        PushResult rc;

        try
        {
            //
            // Invoke the request.
            //
            request -> invoke();

            //
            // Did we receive an exception?
            //
            Exception* e = request -> env() -> exception();
            if(e != 0)
                e -> _raise();

            rc = Success;
        }
	//
	// Some exceptions are transient and should cause a retry.
	//
        catch(const COMM_FAILURE&)
        {
            rc = Retry;
        }
	catch(const TRANSIENT&)
	{
	    rc = Retry;
	}
	catch(const NO_RESPONSE&)
	{
	    rc = Retry;
	}
        catch(const Exception&)
        {
            rc = Failed;
        }

        return rc;
    }

public:

    TypedPusherThread(CosEventComm::PushConsumer_ptr consumer,
                      CosEventChannelAdmin::ProxyPushSupplier_ptr proxy,
		      OB::EventLogger_ptr logger,
                      ULong id,
                      ULong maxEvents,
                      ULong maxRetries,
                      ULong retryTimeout,
                      Double retryMultiplier)
        : PusherBaseThread(consumer, proxy, logger, id, maxEvents, maxRetries,
                           retryTimeout, retryMultiplier),
          logger_(OB::EventLogger::_duplicate(logger))
    {
    }
};

//
// PipeSegment for the Typed ProxyPushSupplier
//
class TEC_PushPipeSegment : public PipeSegment
{
    PusherBaseThreadHandle pusher_; // The event pusher.

public:
    
    TEC_PushPipeSegment(PusherBaseThreadHandle& pusher)
	: pusher_(pusher)
    {
    }
    
    virtual bool
    receive(Event_impl* event)
    {
	if(destroy_)
	    return false;

	pusher_ -> newEvent(event);

	return true;
    }
    
    virtual void
    destroy()
    {
	if(destroy_)
	    return;

	assert(pusher_);
	pusher_ -> halt();
	PipeSegment::destroy();
    }
};

// ----------------------------------------------------------------------
// TEC_ProxyPushSupplier_impl constructor/destructor
// ----------------------------------------------------------------------

TEC_ProxyPushSupplier_impl::TEC_ProxyPushSupplier_impl(
    ProxyConstructorInfo& info)
{
    info.base = "TypedProxyPushSupplier";
    ProxyBase::init(info);
}

TEC_ProxyPushSupplier_impl::~TEC_ProxyPushSupplier_impl()
{
}

// ----------------------------------------------------------------------
// Standard IDL to C++ Mapping
// ----------------------------------------------------------------------

void
TEC_ProxyPushSupplier_impl::connect_push_consumer(
    CosEventComm::PushConsumer_ptr consumer)
    throw(CosEventChannelAdmin::AlreadyConnected, SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
        throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    if(state_ == Connected)
	throw CosEventChannelAdmin::AlreadyConnected();

    if(is_nil(consumer))
	throw BAD_PARAM();
    
    QoSParameters params;
    resolveQoSParams(params);
    
    consumer_ = CosEventComm::PushConsumer::_narrow(
	CORBA::Object_var(applyRequestTimeout(consumer, params))
    );

    state_ = Connected;

    CosEventChannelAdmin::ProxyPushSupplier_var self = _this();

    PusherBaseThreadHandle pusher =
	new TypedPusherThread(consumer_, self, logger_, id_, params.maxEvents,
                              params.maxRetries, params.retryTimeout,
			      params.retryMultiplier);
    pusher -> start();

    assert(!myPipe_);
    myPipe_ = new TEC_PushPipeSegment(pusher);
    parentPipe_ -> connect(myPipe_);
}

void
TEC_ProxyPushSupplier_impl::disconnect_push_supplier()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    disconnectImpl();
}

// ----------------------------------------------------------------------
// ProxyBase implementation
// ----------------------------------------------------------------------

void
TEC_ProxyPushSupplier_impl::deactivateObject()
{
    PortableServer::POA_var poa = _default_POA();
    PortableServer::ObjectId_var oid = poa -> servant_to_id(this);
    poa -> deactivate_object(oid.in());
}

void
TEC_ProxyPushSupplier_impl::disconnectPeer()
{
    //
    // If we're connected then tell the consumer.
    //
    if(!is_nil(consumer_))
    {
        try
        {
            consumer_ -> disconnect_push_consumer();
        }                                              
        catch(const SystemException&)
        {
        }
	consumer_ = CosEventComm::PushConsumer::_nil();
    }
}

bool
TEC_ProxyPushSupplier_impl::canReap()
{
    return false;
}
