// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>

#include <OB/CosTypedEventChannelAdmin_skel.h>

#include <EventLogger.h>
#include <Event_impl.h>
#include <EventUtil.h>
#include <PipeSegment.h>
#include <ProxyBase.h>
#include <ChannelType.h>

#include <TEC_TypedProxyPushConsumer_impl.h>

using namespace CORBA;

//
// This class receives calls for a interface, and forwards the
// event-request to the event channel
//
class TypedProxyPushConsumerDSI_impl
    : public PortableServer::DynamicImplementation,
      public ProxyBase
{
    ChannelType* type_; // The channel type information

    ProxyBase* proxy_; // The associated proxy

public:
    
    // -----------------------------------------------------------------
    // TypedProxyPushConsumerDIS_impl constructor/destructor
    // -----------------------------------------------------------------

    TypedProxyPushConsumerDSI_impl(
	ProxyConstructorInfo&,
	ChannelType*,
	ProxyBase*);
    ~TypedProxyPushConsumerDSI_impl();

    // -----------------------------------------------------------------
    // Standard IDL to C++ Mapping
    // -----------------------------------------------------------------

    virtual Boolean _is_a(const char*)
	throw(SystemException);

    virtual RepositoryId _primary_interface(
	const PortableServer::ObjectId&, PortableServer::POA_ptr)
	throw(SystemException);

    virtual void invoke(ServerRequest_ptr)
	throw();

    // -----------------------------------------------------------------
    // ProxyBase implementation
    // -----------------------------------------------------------------

    virtual void deactivateObject();
    virtual void disconnectPeer() { proxy_ = 0; }
    virtual bool canReap() { return false; }
};

// ----------------------------------------------------------------------
// TypedProxyPushConsumerDSI_impl constructor and destructor
// ----------------------------------------------------------------------

TypedProxyPushConsumerDSI_impl::TypedProxyPushConsumerDSI_impl(
    ProxyConstructorInfo& info,
    ChannelType* type,
    ProxyBase* proxy)
    : type_(type),
      proxy_(proxy)
{
    info.base = "TypedProxyPushConsumerDSI";
    ProxyBase::init(info);
    state_ = Connected;
}

TypedProxyPushConsumerDSI_impl::~TypedProxyPushConsumerDSI_impl()
{
}

// ----------------------------------------------------------------------
// TypedProxyPushConsumerDSI_impl standard IDL to C++ Mapping
// ----------------------------------------------------------------------

Boolean
TypedProxyPushConsumerDSI_impl::_is_a(const char* name)
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    proxy_ -> stampRequest();

    return type_ -> isCompatible(name);
}

RepositoryId
TypedProxyPushConsumerDSI_impl::_primary_interface(
    const PortableServer::ObjectId&, PortableServer::POA_ptr)
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    proxy_ -> stampRequest();

    return type_ -> getInterfaceId();
}

void
TypedProxyPushConsumerDSI_impl::invoke(ServerRequest_ptr req)
    throw()
{
    {
	JTCSynchronized sync(*this);
	//
	// If the proxy is dead, then don't accept any more method
	// invocations
	//
	if(state_ == Destroyed)
	{
	    Any any;
	    any <<= new OBJECT_NOT_EXIST();
	    req -> set_exception(any);
	    return;
	}

	if(state_ == Disconnected)
	{
	    Any any;
	    any <<= new TRANSIENT();
	    req -> set_exception(any);
	    return;
	}
    }

    proxy_ -> stampRequest();
   
    //
    // Rip out the operation name
    //
    String_var opName = req -> operation();

    OpDescriptionSeq* operations = type_ -> getOperations();

    //
    // Look up the operation in the IR.
    //
    for(ULong i = 0; i < operations -> length(); i++)
    {
        if(strcmp(opName, (*operations)[i].name) == 0)
        {
            //
            // Validate the parameters. All in parameters, and no
            // return value, no exceptions.
            //
            if((*operations)[i].result -> kind() != tk_void)
            {
                Any any;
                any <<= new BAD_OPERATION();
                req -> set_exception(any);
                return;
            }

            ULong j;
            for(j = 0 ; j < (*operations)[i].parameters.length() ; j++)
                if((*operations)[i].parameters[j].mode != PARAM_IN)
                {
                    Any any;
                    any <<= new BAD_OPERATION();
                    req -> set_exception(any);
                    return;
                }
    
            if((*operations)[i].exceptions.length() > 0)
            {
                Any any;
                any <<= new BAD_OPERATION();
                req -> set_exception(any);
                return;
            }

            //
            // Create an NVList to contain the parameters for this
            // method. Firstly create the NVList.
            //
            NVList_var params;
            orb_ -> create_list(0, params.out());

            //
            // Now process each parameter.
            //
            for(j = 0 ; j < (*operations)[i].parameters.length() ; j++)
            {
                const ParameterDescription& par = 
		    (*operations)[i].parameters[j];

                //
                // Get the name.
                //
                char* name = string_dup(par.name);

                //
                // Create an any with the appropriate typecode.
                //
                Any* any = new Any;
                any -> replace(par.type, 0);

                //
                // We've already verified that all parameters are IN.
                //
                assert(par.mode == PARAM_IN);

                //
                // Add this parameter.
                //
                Flags flags = ARG_IN;
                params -> add_value_consume(name, any, flags);
            }

	    if(logger_ -> traceEvents() >= 1)
            {
                String_var msg = string_dup("invoking `");
                msg += opName;
                msg += "'\nnumber of parameters: ";
                msg += params -> count();
                for(ULong i = 0; i < params -> count(); ++i)
                {
                    NamedValue_ptr nv = params -> item(i);
                    msg += "\nname: ";
                    msg += nv -> name();
                }
                logger_ -> trace("event", msg);
            }
            NVList_ptr p = params._retn();
            req -> arguments(p);

            //
            // Create a new Event_impl, and forward this to the event
            // channel.
            //
            Event_impl* event = new Event_impl(logger_, p, opName, id_);
            parentPipe_ -> receive(event);
	    event -> discard();
            return;
        }
    }

    //
    // No such operation.
    //
    Any any;
    any <<= new BAD_OPERATION();
    req -> set_exception(any);
}

// ----------------------------------------------------------------------
// TypedProxyPushConsumerDSI_impl ProxyBase implementation
// ----------------------------------------------------------------------

void
TypedProxyPushConsumerDSI_impl::deactivateObject()
{
    PortableServer::POA_var poa = _default_POA();
    PortableServer::ObjectId_var oid = poa -> servant_to_id(this);
    poa -> deactivate_object(oid.in());
}

// ----------------------------------------------------------------------
// TEC_TypedProxyPushConsumer_impl constructor/destructor
// ----------------------------------------------------------------------

TEC_TypedProxyPushConsumer_impl::TEC_TypedProxyPushConsumer_impl(
    ProxyConstructorInfo& info,
    ChannelType* type)
{
    info.base = "TypedProxyPushConsumer";
    ProxyBase::init(info);
    dsiImpl_ = new TypedProxyPushConsumerDSI_impl(info, type, this);

    PortableServer::ObjectId_var oid = poa_ -> activate_object(dsiImpl_);
    String_var repId = type -> getInterfaceId();
    dsiRef_ = poa_ -> create_reference_with_id(oid, repId);
}

TEC_TypedProxyPushConsumer_impl::~TEC_TypedProxyPushConsumer_impl()
{
    dsiImpl_ -> _OB_decRef();
}

// ----------------------------------------------------------------------
// TEC_TypedProxyPushConsumer_impl standard IDL to C++ Mapping
// ----------------------------------------------------------------------

Object_ptr
TEC_TypedProxyPushConsumer_impl::get_typed_consumer()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
        throw OBJECT_NOT_EXIST();
  
    return Object::_duplicate(dsiRef_);
}

void
TEC_TypedProxyPushConsumer_impl::connect_push_supplier(
    CosEventComm::PushSupplier_ptr supplier)
    throw(CosEventChannelAdmin::AlreadyConnected,
	  SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    if(state_ == Connected)
	throw CosEventChannelAdmin::AlreadyConnected();

    stampRequest();

    //
    // We don't necessarily have a supplier.
    //
    if(!is_nil(supplier))
    {
	QoSParameters params;
	resolveQoSParams(params);
	
	supplier_ = CosEventComm::PushSupplier::_narrow(
	    CORBA::Object_var(applyRequestTimeout(supplier, params))
	);
    }

    //
    // Technically it's not permitted to "push" events until the
    // channel is connected. However, the spec doesn't say what do to
    // in the typed channel case.
    //
    state_ = Connected;
}

void
TEC_TypedProxyPushConsumer_impl::disconnect_push_consumer()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    disconnectImpl();
}

void
TEC_TypedProxyPushConsumer_impl::push(const Any&)
    throw(CosEventComm::Disconnected,
	  SystemException)
{
    throw NO_IMPLEMENT();
}

// ----------------------------------------------------------------------
// TypedProxyPushConsumer_impl ProxyBase implementation
// ----------------------------------------------------------------------

void
TEC_TypedProxyPushConsumer_impl::deactivateObject()
{
    //
    // Ask the DSI implementation to go away.
    //
    dsiImpl_ -> disconnect();

    PortableServer::POA_var poa = _default_POA();
    PortableServer::ObjectId_var oid = poa -> servant_to_id(this);
    poa -> deactivate_object(oid.in());
}

void
TEC_TypedProxyPushConsumer_impl::disconnectPeer()
{
    //
    // If we're connected then tell the supplier.
    //
    if(!is_nil(supplier_))
    {
        try
        {
            supplier_ -> disconnect_push_supplier();
        }
        catch(const SystemException&)
        {
        }
	supplier_ = CosEventComm::PushSupplier::_nil();
    }
}

void
TEC_TypedProxyPushConsumer_impl::discardRequestsImpl()
{
    ProxyBase::discardRequestsImpl();

    //
    // Ask the DSI implementation to go away.
    //
    dsiImpl_ -> discardRequests();
}

bool
TEC_TypedProxyPushConsumer_impl::canReap()
{
    return true;
}
