// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <OB/OBEventChannelFactory.h>
#include <OB/CosEventChannelAdmin.h>
#include <OB/CosEventComm_skel.h>

#include <TestSuiteUtilities.h>
#include <stdio.h>
#include <stdlib.h>

//
// Global variable used for tracing.
//
int TraceLevel = 0;

static 
void Usage()
{
    OB_ERROR("pusheventserver <proxy-ior> <EventList>\n"
	     "  EventList:  'auto', or a list of events separated by spaces");
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{

    if(argc < 3)
    {
	Usage();
	return EXIT_FAILURE;
    }
    
    //
    // Get the event descriptions from argv[2]+
    //
    CORBA::StringSeq eventDescriptions;
    eventDescriptions.length(argc - 2);
    for(int arg = 2 ; arg < argc ; arg++)
    {
	eventDescriptions[arg - 2] = CORBA::string_dup(argv[arg]);
    }
    argc -= eventDescriptions.length();

    PrintStatusMsg(TraceLevel, "PushEventServer", 1, "Starting");

    try
    {
	//
	// Initialize the proxy.
	//
	CORBA::Object_var proxyObj = orb -> string_to_object(argv[1]);
	CosEventChannelAdmin::ProxyPushConsumer_var proxyConsumer = 
	    CosEventChannelAdmin::ProxyPushConsumer::_narrow(proxyObj);

	if(CORBA::is_nil(proxyConsumer))
	{
	    OB_ERROR("Proxy is nil in PushEventServer");
	    OB_INFO_LN("ERROR");
	    return EXIT_FAILURE;
	}

	proxyConsumer -> connect_push_supplier(
	    CosEventComm::PushSupplier::_nil());
	
	//
	// Put streaming event generation and pushing here.
	//
	if(strcmp(eventDescriptions[0], "auto") == 0)
	{
	    while(1)
	    {
		CORBA::Any e;
		e <<= "AUTO";
		proxyConsumer -> push(e);
		PrintStatusMsg(TraceLevel, "PushEventServer", 3, 
			       "Pushed Any AUTO Event");
	    }
	}
	
	//
	// Put 'manual' event generation and pushing here.
	//
	for(CORBA::ULong j = 0 ; j < eventDescriptions.length() ; j++)
	{
	    CORBA::Any e;
	    e <<= eventDescriptions[j];
	    proxyConsumer -> push(e);
	    PrintStatusMsg(TraceLevel, "PushEventServer", 3, 
			   "Pushed Any Event");
	}

	try
	{
	    proxyConsumer -> disconnect_push_consumer();
	    PrintStatusMsg(TraceLevel, "PushEventServer", 3, 
			   "Disconnected Any Push Consumer");
	}
        catch(const CORBA::TRANSIENT&)
        {
            // Expected if event service has gone away.
        }
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR("System exception in PushEventServer");
	OB_ERROR(ex);
	orb -> destroy();
	return 	EXIT_FAILURE;
    }
    OB_INFO(OB_FLUSH);
    return EXIT_SUCCESS;
}

int
main(int argc, char** argv)
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    { 
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);
        OB::Properties::KeySeq keys = 
	    properties -> getKeys("ooc.testsuite.event.");

        for(CORBA::ULong i = 0 ; i < keys.length() ; ++i)
        {
            const char* key = keys[i];
            if(strcmp(key, "ooc.testsuite.event.trace") == 0)
            {
                TraceLevel = atoi(properties -> getProperty(key));
		break;
	    }
	    else 
	    {
		OB_ERROR("Warning: Unknown property - " << key);
	    }
        }
	
	//
        // Create the ORB.
        //
        orb = OBCORBA::ORB_init(argc, argv, properties);
	status = run(orb, argc, argv);
    } 
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR(ex);
        OB_INFO_LN("ERROR");
	status = EXIT_FAILURE;
    }

    if (!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_ERROR(ex);
	    status = EXIT_FAILURE;
	}
    }
    return status;
}
