// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/CosTypedEventChannelAdmin.h>
#include <OB/OBTypedEventChannelFactory.h>
#include <OB/Properties.h>

#include <Echo.h>

#include <TestSuiteUtilities.h>
#include <stdio.h>
#include <stdlib.h>

//
// Global variable used for tracing.
//
int TraceLevel = 0;

static 
void Usage()
{
    OB_ERROR("pusheventserver <proxy-ior> <EventList> "
	     "  EventList:  'auto', or a list of events seperated by spaces");
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    if(argc < 3)
    {
	Usage();
	return EXIT_FAILURE;
    }

    //
    // Get the event descriptions from argv[4]+
    //
    CORBA::StringSeq eventDescriptions;
    eventDescriptions.length(argc - 2);
    for(int arg = 2 ; arg < argc ; arg++)
    {
	eventDescriptions[arg - 2] = CORBA::string_dup(argv[arg]);
    }
    argc -= eventDescriptions.length();

    PrintStatusMsg(TraceLevel, "TypedPushEventServer", 1, "Starting");

    try
    {
	//
	// Initialize the server and get the proxy object.
	//
	CORBA::Object_var obj = orb -> string_to_object(argv[1]);
	CosTypedEventChannelAdmin::TypedProxyPushConsumer_var proxyConsumer = 
	    CosTypedEventChannelAdmin::TypedProxyPushConsumer::_narrow(obj);

	if(CORBA::is_nil(proxyConsumer))
	{
	    OB_ERROR("Proxy is nil in TypedPushEventServer");
	    OB_INFO_LN("ERROR");
	    return EXIT_FAILURE;
	}

	proxyConsumer -> connect_push_supplier(
	    CosEventComm::PushSupplier::_nil());

        CORBA::Object_ptr typedConsumer = 
	    proxyConsumer -> get_typed_consumer();

        Echo_var timer = Echo::_narrow(typedConsumer);
        assert(!CORBA::is_nil(timer));

	//
	// Put streaming event generation and pushing here.
	//
	if (strcmp(eventDescriptions[0], "auto") == 0)
	{
	    PrintStatusMsg(TraceLevel, "PushEventServer", 1, 
			   "Pushing Typed AUTO Events");
	    CORBA::String_var str = CORBA::string_dup("AUTO");
	    while(1)
	    {
	        timer -> echo_string(str);
	    }
	}
	
	//
	// Put 'manual' event generation and pushing here.
	//
	for (CORBA::ULong j = 0 ; j < eventDescriptions.length() ; j++)
	{
	    timer -> echo_string(eventDescriptions[j]);
	    PrintStatusMsg(TraceLevel, "PushEventServer", 1, 
			   "Pushed Any Event");
	}
	
	try
	{
	    proxyConsumer -> disconnect_push_consumer();
	    PrintStatusMsg(TraceLevel, "PushEventServer", 1, 
			   "Disconnected Any Push Consumer");
	}
	catch(const CORBA::TRANSIENT&)
	{
	}
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR("System exception in TypedPushEventServer");
	OB_ERROR(ex);
	return EXIT_FAILURE; 
    }
    return EXIT_SUCCESS; 
}

int
main(int argc, char** argv)
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	//
	// Determine the test suite trace level.
	//
        OB::Properties_var properties = 
	    OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);
        OB::Properties::KeySeq keys = 
	    properties -> getKeys("ooc.testsuite.event.");

        for(CORBA::ULong i = 0; i < keys.length(); ++i) 
        {
            const char* key = keys[i];
            if(strcmp(key, "ooc.testsuite.event.trace") == 0)
            {
                TraceLevel = atoi(properties -> getProperty(key));
                break;
            }
	    else
	    {
		OB_ERROR("Warning: Unknown property - " << key);
	    }
        }

        orb = OBCORBA::ORB_init(argc, argv, properties);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR(ex);
        OB_INFO_LN("ERROR");
	return EXIT_FAILURE;
    }

    if (!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_ERROR(ex);
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
