// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/BootManager.h>
#include <OB/NTService.h>

#include <NamingService.h>

#include <stdio.h>
#include <stdlib.h>

#include <NTNamingService.h>

#ifndef MAKEDEPEND
#    include <NamingMsg.h>
#endif

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

using namespace CORBA;
using namespace CosNaming;
using namespace PortableServer;

// ----------------------------------------------------------------------
// NTNamingService constructor
// ----------------------------------------------------------------------

NTNamingService::NTNamingService()
    : OB::NTService("NamingService", "ORBacus Naming Service", NAMING_MSG),
      stopped_(false)
{
}

// ----------------------------------------------------------------------
// NTNamingService public member implementation
// ----------------------------------------------------------------------

void
NTNamingService::orb_run(OB::Logger_ptr logger)
{
    //
    // Get options.
    //
    OB::NamingServiceOptions options = OB::NamingService::defaultOptions();
    int defaultTimeout = options.databaseTimeout;
    options.databaseMode = OB::NamingServiceOptions::OpenOrCreate;

    //
    // Check properties first.
    //
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb_);
    OB::Properties_var properties = oborb -> properties();
    OB::Properties::KeySeq keys = properties -> getKeys("ooc.naming.");
    {
        for(ULong i = 0; i < keys.length(); ++i)
        {
	    const char* key = keys[i];
	    const char* value = properties -> getProperty(key);
	    assert(value != 0);
	    
	    if(strcmp(key, "ooc.naming.database") == 0)
	    {
	        options.databaseName = string_dup(value);
	    }
	    else if(strcmp(key, "ooc.naming.timeout") == 0)
	    {
	        options.databaseTimeout = atoi(value);
	    }
	    else if(strcmp(key, "ooc.naming.callback_timeout") == 0)
	    {
	        options.callbackTimeout = atoi(value);
	    }
	    else if(strcmp(key, "ooc.naming.no_updates") == 0)
	    {
	        options.noUpdates = (strcmp(value, "true") == 0) ||
				    (strcmp(value, "TRUE") == 0) ||
				    (strcmp(value, "1") == 0);
	    }
	    else if(strcmp(key, "ooc.naming.trace_level") == 0)
	    {
	        options.traceLevel = atol(value);
	    }
            else if(strcmp(key, "ooc.naming.endpoint") == 0)
            {
		// ignore here
            }
	    else
	    {
	        String_var msg = string_dup("`");
	        msg += key;
	        msg += "': unknown property";
	        logger -> warning(msg);
	    }
        }
    }
	
    //
    // Resolve Root POA
    //
    Object_var poaObj = orb_ -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);

    //
    // Get a reference to the POA manager
    //
    POAManager_var manager = rootPOA -> the_POAManager();

    //
    // Create poa
    //
    POA_var rootContextPOA;
    POA_var contextPOA;
    try
    {
        PolicyList policies;
        policies.length(2);
        policies[0] =
            rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
        policies[1] =
            rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
        rootContextPOA = 
	    rootPOA -> create_POA("RootContextPOA", manager, policies);

	policies[1] =
	    rootPOA -> create_id_assignment_policy(PortableServer::SYSTEM_ID);
	contextPOA =
	    rootPOA -> create_POA("ContextPOA", manager, policies);
    }
    catch(const POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(rootContextPOA));
    assert(!is_nil(contextPOA));
	
    if(options.databaseTimeout <= 0)
    {
        logger -> warning("timeout value must be > 0. Using default value");
        options.databaseTimeout = defaultTimeout;
    }

    if(!options.databaseName.in())
        logger -> warning("Running in non-persistent mode.");

    //
    // Initialize the naming service.
    //
    OB::NamingService namingService(orb_, rootContextPOA, contextPOA, options);
    namingService.initialize();

    //
    // Get the root naming context.
    //
    NamingContext_var root = namingService.rootNamingContext();

    //
    // Add reference to boot manager
    //
    ObjectId_var nsObjId = PortableServer::string_to_ObjectId("NameService");
    Object_var bmgrObj = orb_ -> resolve_initial_references("BootManager");
    OB::BootManager_var bootManager = OB::BootManager::_narrow(bmgrObj);
    bootManager -> add_binding(nsObjId.in(), root);

    //
    // Inform the status manager that the service is running.
    //
    statusUpdate(SERVICE_RUNNING);
        
    //
    // Run implementation.
    //
    if(!stopped_)
    {
        manager -> activate();
        orb_ -> run();
    }
}

void
NTNamingService::start(int argc, char** argv, OB::Logger_ptr logger)
{
    try
    {
    	//
	// Setup properties
	//
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);

        //
        // Check to see if the concurrency model is usable for nested
        // method invocations. Either reactive/reactive or threaded/
        // thread_per_request/thread_pool is needed.
        //
        const char* oaModel =
	    properties -> getProperty("ooc.orb.oa.conc_model");
        const char* orbModel =
	    properties -> getProperty("ooc.orb.conc_model");

	bool valid;
	if(oaModel != 0 && orbModel != 0)
	{
	    valid = ((strcmp(orbModel, "reactive") == 0 &&
		      strcmp(oaModel, "reactive") == 0) ||
	             (strcmp(orbModel, "threaded") == 0 &&
		      (strcmp(oaModel, "thread_per_request") == 0 ||
		       strcmp(oaModel, "thread_pool") == 0)));
	}
	else
	    valid = false;

	//
	// Set the concurrency model to threaded/threaed_per_request if 
	// a vaild concurrency model wasn't selected and compiled with
	// JThreads/C++ or reactive/reactive otherwise
	//
        if(!valid)
        {
	    properties -> setProperty("ooc.orb.conc_model", "threaded");
	    properties -> setProperty("ooc.orb.oa.conc_model", 
	    			      "thread_per_request");
        }

	const char* port = properties -> getProperty("ooc.naming.endpoint");
	if(port != 0)
	{
	    //
            // Set the ooc.orb.oa.endpoint property from this value
	    // if not already set.
            //
	    const char* oaPort =
	        properties -> getProperty("ooc.orb.oa.endpoint");
	    if(oaPort == 0)
                properties -> setProperty("ooc.orb.oa.endpoint", port);
	}
	
	orb_ = OBCORBA::ORB_init(argc, argv, OB::Properties::_nil(), logger);
	orb_run(logger);
    }
    catch(const CORBA::Exception& ex)
    {
        CORBA::String_var err = ex._to_string();
        logger -> error(err);
    }

    if(!CORBA::is_nil(orb_))
    {
	try
	{
	    orb_ -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
            CORBA::String_var err = ex._to_string();
            logger -> error(err);
	}
    }
}

void
NTNamingService::stop()
{
    //
    // The service manager calls this in a seperate thread. So first
    // we have inform JTC to adopt this thread.
    //
    JTCAdoptCurrentThread adopt;
    stopped_ = true;
    if(!CORBA::is_nil(orb_))
	orb_ -> shutdown(false);
}
