// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OCI_IIOP.h>
#include <OB/Hashtable.h>
#include <OB/ObjectIdHasher.h>
#include <OB/Logger.h>
#include <OB/Codec.h>
#include <OB/BootManager.h>
#include <OB/Properties.h>

#include <OB/OBNaming_skel.h>

#include <Types.h>
#include <Compactor.h>
#include <OBNaming_impl.h>

#include <stdio.h>

#include <NamingDatabase.h>

#include <sys/types.h>
#include <sys/stat.h>

#ifdef WIN32
#include <time.h>
#include <io.h>
#else
#include <unistd.h>
#endif

#include <fcntl.h>
#include <errno.h>

#include <stdlib.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace CosNaming;

// ----------------------------------------------------------------------
// NamingDatabase private member implementation
// ----------------------------------------------------------------------

//
// Dump the contents of a naming context. This is the set of bindings
// in the context.
//
void
NamingDatabase::dumpContext(const PortableServer::ObjectId& key,
			    NamingContext_ptr nc)
{
    //
    // Retrieve all bindings.
    //
    OBNamingContext_var obnc =
	OBNamingContext::_narrow(nc);

    OBNamingContext::ExtendedBindingList_var eb =
	obnc -> list_extended();

    for(ULong i = 0 ; i < eb -> length() ; i++)
	bind(key, eb[i].binding_name, eb[i].bound_obj, eb[i].timestamp, false);
}

void
NamingDatabase::writeAny(const Any& any)
{
    //
    // Serialize the Any
    //
    // TODO: we should cache the cdrCodec
    //
    OctetSeq_var data = cdrCodec_ -> encode(any);

    String_var s = OB::OctetsToAscii(data -> get_buffer(), data -> length());

    assert(os_ != 0);

    bool err = false;
    err = (::fwrite(s.inout(), ::strlen(s), 1, os_) != 1);
    err = err || (::fputc('\n', os_) == EOF);
    if(!noflush_)
        err = err || (fflush(os_) == EOF);
    
    if(err)
    {
	logger_ -> error("Write failed!");
        throw PERSIST_STORE();
    }
}

//
// Rename the file identified by `from' to the new name identified by
// `to'.
//
bool
NamingDatabase::doRename(const char* from, const char* to)
{
    bool result = (::rename(from, to) == 0);
    if(!result)
    {
	const char* err = ::strerror(errno);
	String_var msg = string_dup("Rename: ");
	msg += from;
	msg += " to ";
	msg += to;
	msg += " failed. Error: ";
	msg += err;
	logger_ -> error(msg);
        return false;
    }
    return true;
}

//
// Check to see if the file identified by `file' exists.
//
bool
NamingDatabase::exists(const char* file)
{
    struct stat sbuf;
    return (::stat(file, &sbuf) == 0);
}

//
// Open the logfile.
//
void
NamingDatabase::openLog(bool create)
{
    //
    // Open the logFile_.
    //
    int mode = O_WRONLY|O_APPEND;
#if !defined(WIN32) && !defined(__FreeBSD__)
    mode |= O_SYNC;
#endif
    if(create)
        mode |= O_CREAT;
    int fd = ::open(logFile_, mode, 0666);
    if(fd < 0)
    {
	const char* err = strerror(errno);
	String_var msg = string_dup("Cannot open `");
	msg += logFile_;
	msg += "': ";
	msg += err;
	logger_ -> error(msg);
        throw PERSIST_STORE();
    }
    os_ = ::fdopen(fd, "a");
}

//
// Convert CosNaming::ContextKey to PortableServer::ObjectId
//
PortableServer::ObjectId*
NamingDatabase::key_to_ObjectId(const ContextKey& key)
{
    PortableServer::ObjectId_var oid = 
        new PortableServer::ObjectId(key.length());
    oid -> length(key.length());
    memcpy(oid -> get_buffer(), key.get_buffer(), key.length());
    return oid._retn();
}

//
// Convert CosNaming::ContextKey to PortableServer::Servant
//
PortableServer::Servant
NamingDatabase::key_to_servant(const ContextKey& key)
{
    PortableServer::ObjectId_var oid = key_to_ObjectId(key);
    String_var name = string_dup("");
    try
    {
        name = PortableServer::ObjectId_to_string(oid);
    }
    catch(const SystemException&)
    {
    }
    if(strcmp(name, "NameService") == 0)
        return userPOA_ -> id_to_servant(oid);
    else
        return systemPOA_ -> id_to_servant(oid);

}

// ----------------------------------------------------------------------
// NamingDatabase constructor and destructor
// ----------------------------------------------------------------------

NamingDatabase::NamingDatabase(PortableServer::POA_ptr userPOA,
			       PortableServer::POA_ptr systemPOA,
                               ORB_ptr orb,
			       const char* log,
			       NamingContextSet* ncs,
			       const PolicyList& timeoutPolicy,
			       bool noUpdates,
			       bool& start,
			       long timeout,
			       CORBA::ULong maxBindings,
			       CORBA::ULong traceLevel,
			       bool service)
    : userPOA_(PortableServer::POA::_duplicate(userPOA)),
      systemPOA_(PortableServer::POA::_duplicate(systemPOA)),
      orb_(ORB::_duplicate(orb)),
      ncs_(ncs),
      os_(0),
      timeoutPolicy_(timeoutPolicy),
      noUpdates_(noUpdates),
      initializing_(false),
      update_(false),
      noflush_(false),
      maxBindings_(maxBindings),
      traceLevel_(traceLevel)
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb_);
    logger_ = oborb -> logger();

    //
    // Are we using the IMR
    //
    OB::Properties_var properties = oborb -> properties();
    const char* value = properties -> getProperty("ooc.orb.server_name");
    if(value != 0)
       usingIMR_ = true;
    else
       usingIMR_ = false;

    //
    // Grab a copy of the CDR Codec
    //
    IOP::CodecFactory_var factory = IOP::CodecFactory::_narrow(
	    Object_var(orb -> resolve_initial_references("CodecFactory")));

    IOP::Encoding how;
    how.major_version = 1;
    how.minor_version = 2;
    how.format = IOP::ENCODING_CDR_ENCAPS;

    cdrCodec_ = factory -> create_codec(how);

    //
    // Determine host and port.
    //
    PortableServer::POAManager_var manager = userPOA_ -> the_POAManager();
    OBPortableServer::POAManager_var obManager =
	OBPortableServer::POAManager::_narrow(manager);

    OCI::AcceptorSeq_var acceptors = obManager -> get_acceptors();

    ULong i;
    for(i = 0 ; i < acceptors -> length() ; i++)
    {
	OCI::AcceptorInfo_var accInfo = acceptors[i] -> get_info();
	OCI::IIOP::AcceptorInfo_var iiopInfo =
	    OCI::IIOP::AcceptorInfo::_narrow(accInfo);
	if(!is_nil(iiopInfo))
	{
            StringSeq_var hosts = iiopInfo -> hosts();
	    host_ = hosts[0];
	    port_ = iiopInfo -> port();

	    break;
	}
    }

    //
    // Create the database compactor
    //
    compactor_ = new Compactor(this, timeout);

    //
    // Setup the filenames.
    //
    logFile_ = string_dup(log);
    logFileNew_ = string_dup(log);
    logFileNew_ += ".new";
    logFileBack_ = string_dup(log);
    logFileBack_ += ".bak";

    //
    // If we're running for the first time, then we need to verify
    // that the logfile is not already present.
    //
    if(start)
    {
        if(exists(logFile_))
        {
	    String_var msg = string_dup("Cannot specify --start option. ");
	    msg += "There is an existing log file: `";
	    msg += logFile_;
	    msg += "'.";
	    logger_ -> error(msg);
            throw PERSIST_STORE();
        }
        openLog(true);
	compactor_ -> start();
        return;
    }

    //
    // If the logfile doesn't exist, and the backup file does then
    // this is probably indicative of a crash.
    //
    if(!exists(logFile_) && exists(logFileBack_))
    {
	String_var msg = string_dup("NamingDatabase file is not present: ");
	msg += logFile_;
	msg += ". Using backup file: ";
	msg += logFileBack_;
	logger_ -> warning(msg);
        if(!doRename(logFileBack_, logFile_))
        {
            logger_ -> warning("Rename of backup file failed.");
            throw PERSIST_STORE();
        }
    }

    //
    // Does the logfile exist?
    //
    if(!exists(logFile_))
    {
	//
	// If we're running as a service, then we create the logfile
	// now.
	//
	if(service)
	{
	    String_var msg = string_dup("create database file: ");
	    msg += logFile_;
	    if(traceLevel >= 1)
	        logger_ -> trace("naming", msg);
	    start = true;
	    openLog(true);
	    return;
	}
	String_var msg = string_dup("NamingDatabase file is not present: `");
	msg += logFile_;
	msg += "'\nrun with --start option if running for the first time";
	logger_ -> error(msg);
        throw PERSIST_STORE();
    }

    //
    // Ok, let's get down to business. Read the contents of the
    // logfile.  This is a set of strings, each of which is an octet
    // encoding of a marshaled any.
    //
    OB_IFSTREAM in; // Must use open(name), not ifstream in(name) (VC++ bug)
    in.open(logFile_);
    if(in.fail())
    {
	String_var msg = string_dup("Cannot open store file: ");
	msg += logFile_;
        logger_ -> error(msg);
        throw PERSIST_STORE();
    }

    initializing_ = true;
    int rec = 0;
    String_var cmd = string_alloc(8192);

    try
    {
        try
        {
	    while(!in.eof())
	    {
		cmd[0] = '\0';
		//
		// Read the line in 8k chunks for efficiency.
		//
		do
		{
		    char str[8192];
		    str[0] = 0; // At least required for HP/UX
		    in.get(str, sizeof(str));
		    cmd += str;
		}
		while(!in.eof() && in.peek() != '\n');

		//
		// Skip the newline if necessary.
		//
		if(!in.eof())
		    in.get();

		//
		// If no data is available we're done.
		//
                ULong len = ::strlen(cmd.in()) / 2;
		if(len == 0)
		    continue;

                //
                // Convert the octet character string to an octet sequence.
                //
                Octet* buf = OB::AsciiToOctets(cmd.in());
                OctetSeq data;
                data.length(len);
                memcpy(data.get_buffer(), buf, len);
                delete[] buf;

                //
                // Unserialize the Any
                //
                Any_var any = cdrCodec_ -> decode(data);

                //
                // Next extract the instruction.
                //
		const Database::Version* version;
		const Database::Info* inf;
                const Database::Create* create;
                const Database::Bind* bind;
                const Database::Destroy* destroy;
                const Database::Unbind* unbind;

		if(any >>= version)
		{
		    // No action has to be taken here
		}
		else if(any >>= create)
                {
                    //
                    // Create a new naming context with the provided key.
                    //
		    PortableServer::ObjectId_var oid =
		        key_to_ObjectId(create -> key);

                    (void) new OBNamingContext_impl(userPOA_.in(),
		    				    systemPOA_.in(),
						    orb_.in(), this, ncs_, oid,
						    timeoutPolicy_, noUpdates_,
						    maxBindings_, traceLevel_);
                }
                else if(any >>= bind)
                {
                    PortableServer::ServantBase_var servant =
		        key_to_servant(bind -> key);
                    OBNamingContext_impl* nc =
                        dynamic_cast<OBNamingContext_impl*>(servant.in());
                    assert(nc);

                    //
                    // Create a new binding.
                    //
                    if(bind -> type._d() == ncontext)
                    {
                        if(bind -> rebind)
                            nc -> rebind_context(bind -> name,
						 bind -> type.nc());
                        else
			{
			    OBNamingContext::BindingUnion boundObj;
			    boundObj.nc(bind -> type.nc());
			    nc -> _OB_bindWithTimestamp(bind -> name,
							boundObj,
							bind -> timestamp);
			}
                    }
                    else if(bind -> rebind)
                    {
                        nc -> rebind(bind -> name, bind -> type.obj());
                    }
                    else
                    {
			OBNamingContext::BindingUnion boundObj;
			boundObj.obj(bind -> type.obj());
			nc -> _OB_bindWithTimestamp(bind -> name,
						    boundObj,
						    bind -> timestamp);
		    }
                }
                else if(any >>= destroy)
                {
                    PortableServer::ServantBase_var servant =
		        key_to_servant(destroy -> key);
                    OBNamingContext_impl* nc =
                        dynamic_cast<OBNamingContext_impl*>(servant.in());
                    assert(nc);

                    //
                    // Destroy a context.
                    //
                    nc -> destroy();
                }
                else if(any >>= unbind)
                {
                    PortableServer::ServantBase_var servant =
		        key_to_servant(unbind -> key);
                    OBNamingContext_impl* nc =
                        dynamic_cast<OBNamingContext_impl*>(servant.in());
                    assert(nc);

                    //
                    // Do an unbind.
                    //
                    nc -> unbind(unbind -> name);
                }
		else if(any >>= inf)
		{
		    String_var msg = string_dup("");
		    if(inf -> imr && !usingIMR_)
		    {
		         msg += "Current database must be used with";
			 msg += " the Implementation Repository";
			logger_ -> error(msg);
			throw PERSIST_STORE();
		    }
		    else if(!inf -> imr)
		    {
		        if(usingIMR_)
			{
		            msg += "Current database cannot be used with";
			    msg += " the Implementation Repository";
			    logger_ -> error(msg);
			    throw PERSIST_STORE();
			}
			else
			{
		            if(strcmp(host_, inf -> host) ||
		               (Short)port_ != inf -> port)
		            {
			        String_var msg =
				    string_dup("Wrong host or port: ");
			        msg += host_;
			        msg += ":";
			        msg += (UShort)port_;
			        msg += ". Expected host and port: ";
			        msg += inf -> host;
			        msg += ":";
			        msg += (UShort)inf -> port;
			        logger_ -> error(msg);
			        throw PERSIST_STORE();
		            }
		        }
		    }
		}
                else
                {
		    logger_ -> error("Unknown any in database.");
                    throw PERSIST_STORE();
                }
                rec++;
            }
        }
        catch(const NamingContext::NotFound& e)
        {
	    String_var msg = string_dup("Restore failed: NotFound: ");
            msg += "Reason: ";
            if(e.why == NamingContext::missing_node)
                msg += "missing_node";
            else if(e.why == NamingContext::not_context)
                msg += "not_context";
            else
                msg +=  "not_object";
            msg += ". Rest of name: ";
            for(ULong i = 0; i < e.rest_of_name.length(); ++i)
            {
                msg += '(';
		msg += e.rest_of_name[i].id;
		msg += '.';
		msg += e.rest_of_name[i].kind;
		msg += ") ";
            }
	    logger_ -> error(msg);
            throw PERSIST_STORE();
        }
        catch(const NamingContext::CannotProceed& e)
        {
	    String_var msg = string_dup("Restore failed: CannotProceed");
	    msg += ". Rest of name: ";
            for(ULong i = 0; i < e.rest_of_name.length(); ++i)
            {
                msg += '(';
		msg += e.rest_of_name[i].id;
		msg += '.';
		msg += e.rest_of_name[i].kind;
		msg += ") ";
            }
	    logger_ -> error(msg);
            throw PERSIST_STORE();
        }
        catch(const NamingContext::InvalidName&)
        {
            logger_ -> error("Restore failed: InvalidName");
            throw PERSIST_STORE();
        }
        catch(const NamingContext::AlreadyBound&)
        {
	    logger_ -> error("Restore failed: AlreadyBound");
            throw PERSIST_STORE();
        }
        catch(const NamingContext::NotEmpty&)
        {
            logger_ -> error("Restore failed: NotEmpty");
            throw PERSIST_STORE();
        }
    }
    catch(const PERSIST_STORE&)
    {
	String_var msg = string_dup("Error processing record: ");
	msg += rec;
	logger_ -> error(msg);
	in.close();
        throw;
    }

    in.close();
    initializing_ = false;

    //
    // Re-open the logfile for new journal entries.
    //
    openLog(false);
    update_ = true;

    //
    // Start Compactor
    //
    compactor_ -> start();
}

NamingDatabase::~NamingDatabase()
{
    compactor_ -> halt();
    compactor_ = 0;

    if(os_ != 0)
        ::fclose(os_);
}

// ----------------------------------------------------------------------
// NamingDatabase public member implementation
// ----------------------------------------------------------------------

void
NamingDatabase::updateJournal()
{
    JTCSynchronized synchronized(*this);

    if(!update_)
	return;

    logger_ -> trace("naming", "compacting database");

    //
    // Close old journal.
    //
    ::fclose(os_);
    os_ = 0;
    
    int fd = open(logFileNew_, O_WRONLY|O_CREAT|O_TRUNC, 0666);
    if(fd < 0)
    {
	const char* err = strerror(errno);
	String_var msg = string_dup("Cannot open `");
	msg += logFileNew_;
	msg += "': ";
	msg += err;
	logger_ -> error(msg);
        throw PERSIST_STORE();;
    }

    //
    // Write 
    //
    os_ = fdopen(fd, "a");
    noflush_ = true;

    //
    // Write version, host and port
    //
    info();

    //
    // We write the entire contents of our local naming tree to the
    // journal.  We do this in two steps. First we write the create
    // instructions.  Next we write the bindings. It's important that
    // this order is preserved else we can have a binding to a local
    // naming context *before* that context exists.
    //
    try
    {
        try
        {
	    JTCSynchronized sync(*ncs_);
	    NamingContextSet::Enumerator keys = ncs_ -> keys();

	    ULong i;
	    for(i = 0; i < keys.length(); i++)
	    {
		create(keys[i], 0);
	    }
    
	    for(i = 0; i < keys.length(); i++)
	    {
		NamingContext_ptr nc;
		ncs_ -> get(keys[i], nc);
                try
                {
                    dumpContext(keys[i], nc);
                }
                catch(const OBJECT_NOT_EXIST&)
                {
                    //
                    // Ignore this. If this occurs then the context is
                    // in the middle of being destroyed and the
                    // destroy record will be added to the store once
                    // we've finished dumping the journal.
                    //
                }
	    }
        }
        catch(const NamingContext::CannotProceed& e)
        {
	    String_var msg = string_dup("Compact failed: CannotProceed");
	    msg += ". Rest of name: ";
            for(ULong i = 0; i < e.rest_of_name.length(); ++i)
            {
                msg += '(';
		msg += e.rest_of_name[i].id;
		msg += '.';
		msg += e.rest_of_name[i].kind;
		msg += ") ";
            }
	    logger_ -> error(msg);
            throw PERSIST_STORE();
        }
        catch(const NamingContext::InvalidName&)
        {
            logger_ -> error("Compact failed: InvalidName");
            throw PERSIST_STORE();
        }
        catch(const NamingContext::AlreadyBound&)
        {
	    logger_ -> error("Compact failed: AlreadyBound");
            throw PERSIST_STORE();
        }
        catch(const UserException&)
        {
            logger_ -> error("Compact failed: UserException");
            throw PERSIST_STORE();
        }
        catch(const SystemException& ex)
        {
            String_var msg = string_dup("Compact failed: SystemException: ");
            String_var err = ex._to_string();
            msg += err;
            logger_ -> error(msg);
            throw PERSIST_STORE();
        }

        if(ferror(os_) != 0)
        {
	    String_var msg = string_dup("Write to: ");
	    msg += logFileNew_;
	    msg += " failed.";
	    logger_ -> error(msg);
            throw PERSIST_STORE();
        }
    }
    catch(const PERSIST_STORE&)
    {
	noflush_ = false;
        ::fclose(os_);
        os_ = 0;

	//
	// Re-open old log file.
	//
	openLog(false);

        throw;
    }

    ::fclose(os_);
    os_ = 0;

    noflush_ = false;

    //
    // Move the current to backup. If this fails, it not a big deal.
    //
    doRename(logFile_, logFileBack_);

    //
    // Move the new logfile to current. This failing IS a big deal!
    //
    if(!doRename(logFileNew_, logFile_))
        throw PERSIST_STORE();

    ::unlink(logFileBack_);

    //
    // Open new.
    //
    openLog(false);
    update_ = false;
}

//
// Add a create instruction to the journal.
//
void
NamingDatabase::create(const PortableServer::ObjectId& oid, long timestamp)
{
    JTCSynchronized synchronized(*this);

    if(initializing_)
        return;

    Database::Create create;
    create.key.length(oid.length());
    memcpy(create.key.get_buffer(), oid.get_buffer(), oid.length());
    create.timestamp = timestamp;
    Any any;
    any <<= create;
    
    writeAny(any);

    update_ = true;
}

//
// Add a destroy instruction to the journal.
//
void
NamingDatabase::destroy(const PortableServer::ObjectId& oid)
{
    JTCSynchronized synchronized(*this);

    if(initializing_)
        return;
    
    Database::Destroy destroy;
    destroy.key.length(oid.length());
    memcpy(destroy.key.get_buffer(), oid.get_buffer(), oid.length());

    Any any;
    any <<= destroy;
    writeAny(any);

    update_ = true;
}

//
// Add a bind instruction to the journal.
//
void
NamingDatabase::bind(const PortableServer::ObjectId& oid,
                     const Name& name,
		     const OBNamingContext::BindingUnion& boundObj,
		     long timestamp,
                     bool rebind)
{
    JTCSynchronized synchronized(*this);

    if(initializing_)
        return;
    
    Database::Bind bind;
    bind.rebind = rebind;
    bind.key.length(oid.length());
    memcpy(bind.key.get_buffer(), oid.get_buffer(), oid.length());
    bind.name = name;
    bind.type = boundObj;
    bind.timestamp = timestamp;

    Any any;
    any <<= bind;
    writeAny(any);

    update_ = true;
}

//
// Add an unbind instruction to the journal.
//
void
NamingDatabase::unbind(const PortableServer::ObjectId& oid,
		       const Name& name)
{
    JTCSynchronized synchronized(*this);

    if(initializing_)
        return;
    
    Database::Unbind unbind;
    unbind.key.length(oid.length());
    memcpy(unbind.key.get_buffer(), oid.get_buffer(), oid.length());
    unbind.name = name;

    Any any;
    any <<= unbind;
    writeAny(any);

    update_ = true;
}

//
// Add version, host and port to the journal.
//
void
NamingDatabase::info()
{
    JTCSynchronized synchronized(*this);

    if(initializing_)
        return;

    //
    // Write version.
    //
    Database::Version version;
    version.ver = string_dup(OB::Version);

    Any any;
    any <<= version;
    writeAny(any);

    //
    // Write host and port.
    //
    Database::Info inf;
    inf.host = string_dup(host_);
    inf.port = port_;
    inf.imr = usingIMR_;

    any <<= inf;
    writeAny(any);

    update_ = true;
}
