// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/Hashtable.h>
#include <OB/ObjectIdHasher.h>
#include <OB/Codec.h>
#include <OB/OBPolicies.h>

#include <OB/OBNaming_skel.h>

#include <Types.h>
#include <OBNaming_impl.h>

#include <stdio.h>
#include <stdlib.h>

#include <Compactor.h>
#include <NamingDatabase.h>

#include <NamingService.h>

using namespace CORBA;

namespace OB
{

// ----------------------------------------------------------------------
// NamingService constructor/destructor
// ----------------------------------------------------------------------

NamingService::NamingService(ORB_ptr orb,
			     PortableServer::POA_ptr rootContextPOA,
			     PortableServer::POA_ptr contextPOA,
			     const NamingServiceOptions& options)
    : orb_(ORB::_duplicate(orb)),
      rootContextPOA_(PortableServer::POA::_duplicate(rootContextPOA)),
      contextPOA_(PortableServer::POA::_duplicate(contextPOA)),
      ncs_(0),
      store_(0),
      timeoutPolicy_(0),
      options_(options),
      initialized_(false)
{
}

NamingService::~NamingService()
{
    //
    // Release the root naming context
    //
    root_ = CosNaming::NamingContext::_nil();

    //
    // Clean up all contexts
    //
    if(ncs_)
	delete ncs_;

    //
    // Delete the naming service database
    //
    if(store_)
	delete store_;

    //
    // Delete the timeout policy
    //
    if(timeoutPolicy_)
	delete timeoutPolicy_;
}

// ----------------------------------------------------------------------
// NamingService public member implementation
// ----------------------------------------------------------------------

//
// Get the default options
//
NamingServiceOptions
NamingService::defaultOptions()
{
    NamingServiceOptions options;
    options.maxBindings = 0x7fffffff;
    options.traceLevel = 0;
    options.noUpdates = false;
    options.callbackTimeout = 5;
    options.databaseTimeout = 5;
    options.databaseMode = NamingServiceOptions::OpenExisting;

    return options;
}

//
// Initialize the naming service
//
void
NamingService::initialize()
{
    if(initialized_)
	return;

    initialized_ = true;

    //
    // Create the set of naming contexts
    //
    ncs_ = new CosNaming::NamingContextSet(63);
	
    //
    // Create the timeout policy
    //
    Any timeoutAny;
    timeoutAny <<= (ULong)(1000 * options_.callbackTimeout);
    timeoutPolicy_ = new PolicyList(1);
    timeoutPolicy_ -> length(1);
    (*timeoutPolicy_)[0] = orb_ -> create_policy(TIMEOUT_POLICY_ID,
    						 timeoutAny);

    bool start = false;
    bool service = false;
    if(options_.databaseName.in())
    {
	//
	// Create a new naming service database
	//
	start = options_.databaseMode == NamingServiceOptions::CreateNew;
	service = options_.databaseMode == NamingServiceOptions::OpenOrCreate;

	store_ = new CosNaming::NamingDatabase(rootContextPOA_, contextPOA_,
					       orb_, options_.databaseName,
					       ncs_, *timeoutPolicy_,
					       options_.noUpdates, start,
					       60 * options_.databaseTimeout,
					       options_.traceLevel, service);

	//
	// Add version, host and port to the journal
	//
	store_ -> info();
    }
        
    //
    // Get the root naming context
    //
    PortableServer::ObjectId_var nsObjId =
	PortableServer::string_to_ObjectId("NameService");
    if(start || !options_.databaseName.in())
    {
	CosNaming::OBNamingContext_impl* impl = 
	    new CosNaming::OBNamingContext_impl(rootContextPOA_, contextPOA_,
	    					orb_, store_, ncs_, nsObjId,
					        *timeoutPolicy_,
					        options_.noUpdates,
						options_.maxBindings,
					        options_.traceLevel);
	root_ = impl -> _this();
    }
    else
    {
	CosNaming::NamingContext_ptr root = CosNaming::NamingContext::_nil();
	if(ncs_ -> get(nsObjId, root))
	    root_ = CosNaming::NamingContext::_duplicate(root);
    }
}

//
// Get root naming context
//
CosNaming::NamingContext_ptr
NamingService::rootNamingContext()
{
    return CosNaming::NamingContext::_duplicate(root_);
}

} // End of namespace OB
