// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_NAMING_IMPL_H
#define OB_NAMING_IMPL_H

namespace CosNaming
{

//
// Forward declarations
//
class NamingDatabase;

class InternalBinding;
typedef InternalBinding* InternalBinding_ptr;

} // End of namespace CosNaming

void OBDuplicate(CosNaming::InternalBinding_ptr);
void OBRelease(CosNaming::InternalBinding_ptr);

namespace CosNaming
{

typedef OB::ObjVar< InternalBinding > InternalBinding_var;

//
// Declaration of NamingContextSet
//
typedef OB::SynchronizedHashtable< PortableServer::ObjectId,
		                   NamingContext_ptr,
    		                   OBPortableServer::ObjectIdHasher 
                                 > NamingContextSet;

typedef OB_STL::list<BindingListener_var> BindingListenerList;

//
// Implementation of CosNaming::OBNamingContext
//
class OBNamingContext_impl
    : virtual public POA_CosNaming::OBNamingContext,
      virtual public PortableServer::RefCountServantBase
{
    PortableServer::POA_var userPOA_;
    PortableServer::POA_var systemPOA_;
    PortableServer::POA_var rootPOA_;
    CORBA::ORB_var orb_;
    PortableServer::ObjectId_var key_;

    struct NameComponentHasher
    {
	static CORBA::ULong hash(const NameComponent& k)
	{
	    return OB::StringHasher::hash(k.id) ^
		OB::StringHasher::hash(k.kind);
	}

	static bool comp(const NameComponent& k1, const NameComponent& k2)
	{
	    return (strcmp(k1.id, k2.id) == 0) &&
		(strcmp(k1.kind, k2.kind) == 0);
	}
    };
    typedef OB::SynchronizedHashtable< CosNaming::NameComponent, 
	                               InternalBinding_var,
	                               NameComponentHasher > BindingsHash;

private:

    BindingsHash* bindings_; // Set of bindings in this context

    NamingContextSet* ncs_; // Global set of naming contexts
    NamingDatabase* store_; // The database

    const CORBA::PolicyList& timeoutPolicy_;  // Timeout for callbacks

    bool noUpdates_; // No automatic updates?
    bool destroyed_; // Has the context been destroyed?

    //
    // Clients to which this service reports events
    //
    BindingListenerList listeners_;
    JTCMutex listenersMutex_;

    NamingContext_var myRef_; // Object-reference to myself

    OB::Logger_var logger_; // Tracing information
    CORBA::ULong maxBindings_;
    CORBA::ULong traceLevel_;

    bool isRoot_; // Is this the Root Context?

    void cleanup();

    void doBind(const CosNaming::Name& n,
		const CosNaming::OBNamingContext::BindingUnion&, bool, long);

    //
    // Support for BindingListener interface
    //
    void bindingAdded(const CosNaming::OBNamingContext::BindingUnion&,
                      const CosNaming::Name&);
    void bindingRemoved(const CosNaming::Name&, bool isContext);
    void bindingChanged(const CosNaming::OBNamingContext::BindingUnion&,
                        const CosNaming::Name&,
			const CosNaming::Name&);

    //
    // Add escapes for INS
    //
    static char* addEscapes(const char*);
    static char* addURLEscapes(const char*);

    OBNamingContext_impl* createNewContext();

public:

    OBNamingContext_impl(PortableServer::POA_ptr, PortableServer::POA_ptr,
    			 CORBA::ORB_ptr, NamingDatabase*, NamingContextSet*,
			 const PortableServer::ObjectId&,
			 const CORBA::PolicyList&, bool,
			 CORBA::ULong, CORBA::ULong);
    ~OBNamingContext_impl();

    PortableServer::POA_ptr _default_POA();

    //
    // Operations from CosNaming::NamingContext
    //
    virtual void
    bind(const CosNaming::Name&, CORBA::Object_ptr)
        throw(CosNaming::NamingContext::NotFound,
              CosNaming::NamingContext::CannotProceed,
              CosNaming::NamingContext::InvalidName,
              CosNaming::NamingContext::AlreadyBound,
              CORBA::SystemException);

    virtual void
    rebind(const CosNaming::Name&, CORBA::Object_ptr)
        throw(CosNaming::NamingContext::NotFound,
              CosNaming::NamingContext::CannotProceed,
              CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    virtual void
    bind_context(const CosNaming::Name&, CosNaming::NamingContext_ptr)
        throw(CosNaming::NamingContext::NotFound,
              CosNaming::NamingContext::CannotProceed,
              CosNaming::NamingContext::InvalidName,
              CosNaming::NamingContext::AlreadyBound,
              CORBA::SystemException);

    virtual void
    rebind_context(const CosNaming::Name&, CosNaming::NamingContext_ptr)
        throw(CosNaming::NamingContext::NotFound,
              CosNaming::NamingContext::CannotProceed,
              CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    virtual CORBA::Object_ptr
    resolve(const CosNaming::Name&)
        throw(CosNaming::NamingContext::NotFound,
              CosNaming::NamingContext::CannotProceed,
              CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    virtual void
    unbind(const CosNaming::Name&)
        throw(CosNaming::NamingContext::NotFound,
              CosNaming::NamingContext::CannotProceed,
              CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    virtual CosNaming::NamingContext_ptr
    new_context()
        throw(CORBA::SystemException);

    virtual CosNaming::NamingContext_ptr
    bind_new_context(const CosNaming::Name&)
        throw(CosNaming::NamingContext::NotFound,
              CosNaming::NamingContext::AlreadyBound,
              CosNaming::NamingContext::CannotProceed,
              CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    virtual void
    destroy()
        throw(CosNaming::NamingContext::NotEmpty,
              CORBA::SystemException);

    virtual void
    list(CORBA::ULong, CosNaming::BindingList_out,
         CosNaming::BindingIterator_out)
        throw(CORBA::SystemException);

    //
    // Operations from CosNaming::NamingContextExt
    //
    virtual CosNaming::NamingContextExt::StringName
    to_string(const CosNaming::Name&)
        throw(CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    virtual CosNaming::Name*
    to_name(const char*)
        throw(CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    virtual CosNaming::NamingContextExt::URLString
    to_url(const char*, const char*)
        throw(CosNaming::NamingContextExt::InvalidAddress,
              CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    virtual CORBA::Object_ptr
    resolve_str(const char*)
        throw(CosNaming::NamingContext::NotFound,
              CosNaming::NamingContext::CannotProceed,
              CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    //
    // Operations from CosNaming::OBNamingContext
    //
    virtual CosNaming::OBNamingContext::ExtendedBindingList*
    list_extended()
        throw(CORBA::SystemException);

    virtual void
    rename(const CosNaming::Name&, const CosNaming::Name&)
        throw(CosNaming::NamingContext::NotFound,
              CosNaming::NamingContext::AlreadyBound,
              CosNaming::NamingContext::CannotProceed,
              CosNaming::NamingContext::InvalidName,
              CORBA::SystemException);

    virtual CORBA::Boolean
    add_listener(CosNaming::BindingListener_ptr)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean
    remove_listener(CosNaming::BindingListener_ptr)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean
    callbacks_active()
        throw(CORBA::SystemException);

    //
    // ORBacus internal methods
    //
    const PortableServer::ObjectId& _OB_getKey() const;

    void _OB_bindWithTimestamp(const CosNaming::Name&,
			       const CosNaming::OBNamingContext::BindingUnion&,
			       long);
};

class BindingIterator_impl
    : virtual public POA_CosNaming::BindingIterator,
      virtual public PortableServer::RefCountServantBase,
      public JTCMonitor
{
    PortableServer::POA_var poa_;
    CORBA::ULong idx_;
    BindingList bl_;

public:

    BindingIterator_impl(PortableServer::POA_ptr,
                         const CosNaming::BindingList&);

    //
    // Standard IDL to C++ Mapping
    //
    virtual CORBA::Boolean
    next_one(CosNaming::Binding_out)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean
    next_n(CORBA::ULong, CosNaming::BindingList_out)
        throw(CORBA::SystemException);

    virtual void
    destroy()
        throw(CORBA::SystemException);
};

} // End of namespace CosNaming

#endif
