// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/BootManager.h>

#include <NamingService.h>

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

using namespace CORBA;
using namespace PortableServer;
using namespace CosNaming;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << 
	" [options]\n"
	"\n"
	"Options:\n"
	"-h, --help                   "
	"Show this message.\n"
	"-v, --version                "
	"Show Naming Service version.\n"
	"-i, --ior                    "
	"Print IOR on standard output.\n"
	"-n, --no-updates             "
	"No automatic updates with ORBacus Names.\n"
        "-s, --start                  "
	"Run for the first time.\n"
        "-d, --database FILE          "
	"Use FILE as the database.\n"
	"-l, --limit COUNT            "
	"Limit bindings returned by list operation.\n"
        "-t, --timeout MINS           "
	"Purge the database every MINS minutes.\n"
	"                             "
	"(The default is 5 minutes.)\n"
	"-c, --callback-timeout SECS  "
	"Timeout for callbacks in seconds.\n"
	"                             "
	"(The default is 5 seconds.)\n"
	;
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Get options
    //
    bool ior = false;
    OB::NamingServiceOptions options = OB::NamingService::defaultOptions();

    //
    // Check properties first
    //
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    OB::Properties_var properties = oborb -> properties();
    OB::Properties::KeySeq keys = properties -> getKeys("ooc.naming.");
    {
        for(ULong i = 0; i < keys.length(); ++i)
        {
	    const char* key = keys[i];
	    const char* value = properties -> getProperty(key);
	    assert(value != 0);
	    
	    if(strcmp(key, "ooc.naming.database") == 0)
	    {
	        options.databaseName = string_dup(value);
	    }
	    else if(strcmp(key, "ooc.naming.limit") == 0)
	    {
	        options.maxBindings = atoi(value);
	    }
	    else if(strcmp(key, "ooc.naming.timeout") == 0)
	    {
	        options.databaseTimeout = atoi(value);
	    }
	    else if(strcmp(key, "ooc.naming.callback_timeout") == 0)
	    {
	        options.callbackTimeout = atoi(value);
	    }
	    else if(strcmp(key, "ooc.naming.no_updates") == 0)
	    {
	        options.noUpdates = (strcmp(value, "true") == 0) ||
				    (strcmp(value, "TRUE") == 0) ||
				    (strcmp(value, "1") == 0);
	    }
	    else if(strcmp(key, "ooc.naming.trace_level") == 0)
	    {
	        options.traceLevel = atol(value);
	    }
            else if(strcmp(key, "ooc.naming.endpoint") == 0)
            {
                // ignore here
            }
	    else
	    {
	        OB_CERR << argv[0] << "NameService: `" << key 
	             << "': unknown property" << OB_ENDL;
	    }
        }
    }

    //
    // Resolve Root POA
    //
    Object_var poaObj = orb -> resolve_initial_references("RootPOA");
    POA_var rootPOA = POA::_narrow(poaObj.in());

    //
    // Get a reference to the POA manager
    //
    POAManager_var manager = rootPOA -> the_POAManager();

    //
    // Create POA 
    //
    POA_var rootContextPOA;
    POA_var contextPOA;
    try
    {
        PolicyList policies;
        policies.length(2);
        policies[0] =
            rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
        policies[1] =
            rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
        rootContextPOA = 
	    rootPOA -> create_POA("RootContextPOA", manager, policies);

        policies[1] =
            rootPOA -> create_id_assignment_policy(PortableServer::SYSTEM_ID);
        contextPOA = 
	    rootPOA -> create_POA("ContextPOA", manager, policies);
    }
    catch(const POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(rootContextPOA));
    assert(!is_nil(contextPOA));

    for(int i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
        if(strcmp(argv[i], "--help") == 0 || 
	   strcmp(argv[i], "-h") == 0)
        {
	    usage(argv[0]);
	    return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "--version") == 0 || 
		strcmp(argv[i], "-v") == 0)
        {
	    OB_CERR << "ORBacus Naming Service " << OB::Version << OB_ENDL;
	    return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "--ior") == 0 || 
		strcmp(argv[i], "-i") == 0)
        {
	    ior = true;
        }
        else if(strcmp(argv[i], "--no-updates") == 0 ||
		strcmp(argv[i], "-n") == 0)
        {
	    options.noUpdates = true;
        }
        else if(strcmp(argv[i], "--limit") == 0 || 
		strcmp(argv[i], "-l") == 0)
        {
	    ++i;
	    if(i >= argc)
	    {
	        OB_CERR << argv[0] << ": " << argv[i-1]
			<< " expects argument" << OB_ENDL;
	        usage(argv[0]);
	        return EXIT_FAILURE;
	    }
	    options.maxBindings = atoi(argv[i]);
        }
        else if(strcmp(argv[i], "--timeout") == 0 || 
		strcmp(argv[i], "-t") == 0)
        {
	    ++i;
	    if(i >= argc)
	    {
	        OB_CERR << argv[0] << ": " << argv[i-1]
			<< " expects argument" << OB_ENDL;
	        usage(argv[0]);
	        return EXIT_FAILURE;
	    }
	    options.databaseTimeout = atoi(argv[i]);
        }
        else if(strcmp(argv[i], "--callback-timeout") == 0 ||
		strcmp(argv[i], "-c") == 0)
        {
	    ++i;
	    if(i >= argc)
	    {
	        OB_CERR << argv[0] << ": " << argv[i-1]
			<< " expects argument" << OB_ENDL;
	        usage(argv[0]);
	        return EXIT_FAILURE;
	    }
	    options.callbackTimeout = atoi(argv[i]);
        }
        else if(strcmp(argv[i], "--database") == 0 ||
		strcmp(argv[i], "-d") == 0)
        {
	    ++i;
	    if(i >= argc)
	    {
	        OB_CERR << argv[0] << ": " << argv[i-1]
			<< " expects argument" << OB_ENDL;
	        usage(argv[0]);
	        return EXIT_FAILURE;
	    }
	    options.databaseName = string_dup(argv[i]);
        }
        else if(strcmp(argv[i], "--start") == 0 || 
		strcmp(argv[i], "-s") == 0)
        {
	    options.databaseMode = OB::NamingServiceOptions::CreateNew;
        }
        else
        {
	    OB_CERR << argv[0] 
		    << ": unknown option `" 
		    << argv[i]
	    	    << "'\n"
		    << OB_ENDL;
	    usage(argv[0]);
	    return EXIT_FAILURE;
        }
    }
	
    if(options.databaseTimeout <= 0)
    {
        OB_CERR << argv[0] << ": timeout value must be > 0" << OB_ENDL;
        return EXIT_FAILURE;
    }
	
    if(options.maxBindings < 0)
    {
        OB_CERR << argv[0] << ": binding limit must be >= 0" << OB_ENDL;
        return EXIT_FAILURE;
    }
	
    if((options.databaseMode == OB::NamingServiceOptions::CreateNew) &&
       !options.databaseName.in())
    {
        OB_CERR << argv[0]
		<< ": --start (-s) requires the --database (-d) option"
		<< OB_ENDL;
        return EXIT_FAILURE;
    }

    //
    // Initialize the naming service.
    //
    OB::NamingService namingService(orb, rootContextPOA, contextPOA, options);
    namingService.initialize();

    //
    // Get the root naming context.
    //
    NamingContext_var root = namingService.rootNamingContext();

    //
    // Add reference to boot manager
    //
    ObjectId_var nsObjId = PortableServer::string_to_ObjectId("NameService");
    Object_var bmgrObj = orb -> resolve_initial_references("BootManager");
    OB::BootManager_var bootManager = OB::BootManager::_narrow(bmgrObj);
    bootManager -> add_binding(nsObjId.in(), root);

    if(ior)
    {
	//
	// Print IOR on standard output.
	//
	String_var s = orb -> object_to_string(root);
	OB_COUT << s << OB_ENDL;
    }

    //
    // Run implementation
    //
    manager -> activate();
    orb -> run();

    return EXIT_SUCCESS;
}

//
// ORB_ptr for signal handler
//
static CORBA::ORB_ptr signal_orb = CORBA::ORB::_nil();

//
// Signal handler for ORB shutdown
//
#ifdef WIN32
BOOL
handler(DWORD)
{
    //
    // Inform JTC of presence of new thread
    //
    JTCAdoptCurrentThread adopt;

    //
    // Terminate ORB event loop
    //
    try 
    {
        if (!CORBA::is_nil(signal_orb))
	{
            signal_orb->shutdown(false);
	}
    } 
    catch (...) 
    {
        // Can't throw here...
    }
    return TRUE;
}
#else
extern "C"
void
handler(int)
{
    // 
    // Ignore further signals
    // 
    struct sigaction ignore;
    ignore.sa_handler = SIG_IGN;
    sigemptyset(&ignore.sa_mask);
    ignore.sa_flags = 0;

    if (sigaction(SIGINT, &ignore, (struct sigaction *)0) == -1)
        abort();
    if (sigaction(SIGTERM, &ignore, (struct sigaction *)0) == -1)
        abort();
    if (sigaction(SIGHUP, &ignore, (struct sigaction *)0) == -1)
        abort();

    //
    // Terminate ORB event loop
    //
    try 
    {
        if (!CORBA::is_nil(signal_orb))
	{
            signal_orb->shutdown(false);
	}
    } 
    catch (...) 
    {
        // Can't throw here...
    }
}
#endif

//
// Install signal handler for ORB shutdown
//
void install_signal_handler(CORBA::ORB_ptr orb)
{
    // 
    // Set ORB pointer for handling function
    //
    signal_orb = orb;

    //
    // Install signal handler for cleanup
    //
#ifdef WIN32
    BOOL rc = SetConsoleCtrlHandler((PHANDLER_ROUTINE)handler, TRUE);
    if (!rc)
    {
        abort();
    }
#else
    struct sigaction sa;          // New signal state
    sa.sa_handler = handler;      // Set handler function
    sigfillset(&sa.sa_mask);      // Mask all other signals while handler runs
    sa.sa_flags = 0 | SA_RESTART; // Restart interrupted system calls

    if (sigaction(SIGINT, &sa, (struct sigaction *)0) == -1)
        abort();
    if (sigaction(SIGHUP, &sa, (struct sigaction *)0) == -1)
        abort();
    if (sigaction(SIGTERM, &sa, (struct sigaction *)0) == -1)
        abort();
#endif
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
    	//
	// Setup properties
	//
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);

        //
        // Check to see if the concurrency model is usable for nested
        // method invocations. Either reactive/reactive or threaded/
        // thread_per_request/thread_pool is needed.
        //
        const char* oaModel =
	    properties -> getProperty("ooc.orb.oa.conc_model");
        const char* orbModel =
	    properties -> getProperty("ooc.orb.conc_model");

	bool valid;
	if(oaModel != 0 && orbModel != 0)
	{
	    valid = ((strcmp(orbModel, "reactive") == 0 &&
		      strcmp(oaModel, "reactive") == 0) ||
	             (strcmp(orbModel, "threaded") == 0 &&
		      (strcmp(oaModel, "thread_per_request") == 0 ||
		       strcmp(oaModel, "thread_pool") == 0)));
	}
	else
	{
	    valid = false;
	}

	//
	// Set the concurrency model to threaded/thread_per_request if 
	// a vaild concurrency model wasn't selected and compiled with
	// JThreads/C++ or reactive/reactive otherwise
	//
        if(!valid)
        {
	    properties -> setProperty("ooc.orb.conc_model", 
				      "threaded");
	    properties -> setProperty("ooc.orb.oa.conc_model", 
	    			      "thread_per_request");
        }

	const char* port = properties -> getProperty("ooc.naming.endpoint");
	if(port != 0)
	{
	    //
            // Set the ooc.orb.oa.endpoint property from this value
	    // if not already set.
            //
	    const char* oaPort = 
	        properties -> getProperty("ooc.orb.oa.endpoint");
	    if(oaPort == 0)
	    {
                properties -> setProperty("ooc.orb.oa.endpoint", port);
	    }
	}

	orb = CORBA::ORB_init(argc, argv);

	//
	// Install signal handler for ORB shutdown
	//
	install_signal_handler(orb);

	//
	// Run the service
	//
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
