// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <OB/CORBA.h>

#include <OB/CosNaming.h>

#include "TestObject_impl.h"
#include "TestMacro.h"


//
// Test CosNaming::NamingContext::rebind()
//
void
TestRebind(CORBA::ORB_ptr /* orb */, CosNaming::NamingContext_ptr root)
{
    TestObject_impl impl;
    TestObject_var test = impl._this();
    CosNaming::Name name1, name2;

    try
    {
        //
        // Test exceptions
        //

        try
        {
            //
            // Try to rebind an object when an intermediate context
            // doesn't exist
            //
            name1.length(2);
            name1[0].id = CORBA::string_dup("bad");
            name1[0].kind = CORBA::string_dup("node");
            name1[1].id = CORBA::string_dup("my");
            name1[1].kind = CORBA::string_dup("node");
            root -> rebind(name1, test.in());
            assert("rebind should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::missing_node);
        }

        name1.length(1);
        name1[0].id = CORBA::string_dup("object");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(root -> bind(name1, test.in()));

        try
        {
            //
            // Try to rebind an object when an object is in the path
            //
            name2.length(2);
            name2[0].id = CORBA::string_dup("object");
            name2[0].kind = CORBA::string_dup("");
            name2[1].id = CORBA::string_dup("my");
            name2[1].kind = CORBA::string_dup("node");
            root -> rebind(name2, test.in());
            assert("rebind should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::not_context);
        }

        //
        // Rebind nil object to name
        //
        SUCCEED(root -> rebind(name1, CORBA::Object::_nil()));
        SUCCEED(root -> unbind(name1));

        try
        {
            //
            // Try to rebind an object with an invalid name
            //
            name1.length(0);
            root -> rebind(name1, test.in());
            assert("rebind should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::InvalidName&)
        {
            // Expected
        }

        try
        {
            //
            // Try to rebind an object with an invalid name
            //
            name1.length(1);
            name1[0].id = CORBA::string_dup("");
            name1[0].kind = CORBA::string_dup("");
            root -> rebind(name1, test.in());
            assert("rebind should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::InvalidName&)
        {
            // Expected
        }

        //
        // Bind an intermediate context for use in more tests
        //
        CosNaming::NamingContext_var context = root -> new_context();
        name1.length(1);
        name1[0].id = CORBA::string_dup("ctx");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(root -> bind_context(name1, context.in()));

        try
        {
            //
            // Try to rebind an object when an intermediate context
            // doesn't exist
            //
            name1.length(3);
            name1[0].id = CORBA::string_dup("ctx");
            name1[0].kind = CORBA::string_dup("");
            name1[1].id = CORBA::string_dup("bad");
            name1[1].kind = CORBA::string_dup("node");
            name1[2].id = CORBA::string_dup("my");
            name1[2].kind = CORBA::string_dup("node");
            root -> rebind(name1, test.in());
            assert("rebind should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::missing_node);
        }

        name1.length(1);
        name1[0].id = CORBA::string_dup("object");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(context -> bind(name1, test.in()));

        try
        {
            //
            // Try to rebind an object when an object is in the path
            //
            name2.length(3);
            name2[0].id = CORBA::string_dup("ctx");
            name2[0].kind = CORBA::string_dup("");
            name2[1].id = CORBA::string_dup("object");
            name2[1].kind = CORBA::string_dup("");
            name2[2].id = CORBA::string_dup("my");
            name2[2].kind = CORBA::string_dup("node");
            root -> rebind(name2, test.in());
            assert("rebind should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::not_context);
        }

        //
        // Rebind nil object to an existing name
        //
        name2.length(2);
        name2[0].id = CORBA::string_dup("ctx");
        name2[0].kind = CORBA::string_dup("");
        name2[1].id = CORBA::string_dup("object");
        name2[1].kind = CORBA::string_dup("");
        SUCCEED(root -> rebind(name2, CORBA::Object::_nil()));

        SUCCEED(context -> unbind(name1));

        try
        {
            //
            // Try to rebind an object using the name of an existing context
            //
            name1.length(1);
            name1[0].id = CORBA::string_dup("ctx");
            name1[0].kind = CORBA::string_dup("");
            root -> rebind(name1, test.in());
            assert("rebind should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::not_object);
        }

        //
        // Clean up context
        //
        name1.length(1);
        name1[0].id = CORBA::string_dup("ctx");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(root -> unbind(name1));
        SUCCEED(context -> destroy());


        //
        // Test name equality to make sure kind is included in comparisons
        //
        name1.length(1);
        name1[0].id = CORBA::string_dup("object");
        name1[0].kind = CORBA::string_dup("kind1");
        name2.length(1);
        name2[0].id = CORBA::string_dup("object");
        name2[0].kind = CORBA::string_dup("kind2");
        SUCCEED(root -> rebind(name1, test.in()));
        SUCCEED(root -> rebind(name2, test.in()));
        SUCCEED(root -> unbind(name1));
        SUCCEED(root -> unbind(name2));


        //
        // Since impl is a stack-based servant, we need to deactivate
        // it before it goes out of scope
        //
        PortableServer::POA_var poa = impl._default_POA();
        PortableServer::ObjectId_var id = poa -> servant_to_id(&impl);
        poa -> deactivate_object(id.in());
    }
    catch(...)
    {
        //
        // Deactivate stack-based servant
        //
        PortableServer::POA_var poa = impl._default_POA();
        PortableServer::ObjectId_var id = poa -> servant_to_id(&impl);
        poa -> deactivate_object(id.in());
        throw;
    }
}
