#include <OB/CORBA.h>

#include <CCS.hh>               // ORB-specific
#include <CosTradingRepos.hh>   // ORB-specific
#include <CosTrading.hh>        // ORB-specific
#include <assert.h>

#if defined(HAVE_STD_IOSTREAM)
using namespace std;
#endif

//----------------------------------------------------------------

// Generic ostream inserter for exceptions. Inserts the exception
// name, if available, and the repository ID otherwise.

#if 0   // This inserter may or may not be needed for your ORB.

static ostream &
operator<<(ostream & os, const CORBA::Exception & e)
{
    CORBA::Any tmp;
    tmp <<= e;

    CORBA::TypeCode_var tc = tmp.type();
    const char * p = tc->name();
    if (*p != '\0')
        os << p;
    else
        os << tc->id();
    return os;
}

#endif

//----------------------------------------------------------------

// Function templates don't work properly with SGI CC 7.2.1, Sun CC
// 5.0, and VC++ 6.0
#if !defined(__GNUC__) && \
    (defined(__sgi) || defined(__sun) || defined(_MSC_VER))
CosTrading::Lookup_ptr
resolve_init(CORBA::ORB_ptr orb, const char * id)
{
    CORBA::Object_var obj;
    try {
        obj = orb->resolve_initial_references(id);
    }
    catch (const CORBA::ORB::InvalidName &) {
        throw;
    }
    catch (const CORBA::Exception & e) {
        cerr << "Cannot get initial reference for "
             << id << ": " << e << endl;
        throw 0;
    }
    assert(!CORBA::is_nil(obj));

    CosTrading::Lookup_var ref;
    try {
        ref = CosTrading::Lookup::_narrow(obj);
    }
    catch (const CORBA::Exception & e) {
        cerr << "Cannot narrow reference for "
             << id << ": " << e << endl;
        throw 0;
    }
    if (CORBA::is_nil(ref)) {
        cerr << "Incorrect type of reference for "
             << id << endl;
        throw 0;
    }
    return ref._retn();
}
#else
template<class T>
typename T::_ptr_type
resolve_init(CORBA::ORB_ptr orb, const char * id)
{
    CORBA::Object_var obj;
    try {
        obj = orb->resolve_initial_references(id);
    }
    catch (const CORBA::ORB::InvalidName & e) {
        throw;
    }
    catch (const CORBA::Exception & e) {
        cerr << "Cannot get initial reference for "
             << id << ": " << e << endl;
        throw 0;
    }
    assert(!CORBA::is_nil(obj));

    typename T::_var_type ref;
    try {
        ref = T::_narrow(obj);
    }
    catch (const CORBA::Exception & e) {
        cerr << "Cannot narrow reference for "
             << id << ": " << e << endl;
        throw 0;
    }
    if (CORBA::is_nil(ref)) {
        cerr << "Incorrect type of reference for "
             << id << endl;
        throw 0;
    }
    return ref._retn();
}
#endif

//----------------------------------------------------------------

int
run(CORBA::ORB_ptr orb, int /* argc */, char*[])
{
    using namespace CosTradingRepos;
    using namespace CosTrading;

    // Get reference to Lookup interface.
    Lookup_var lookup;
// Function templates don't work properly with SGI CC 7.2.1, Sun CC
// 5.0, and VC++ 6.0
#if !defined(__GNUC__) && \
    (defined(__sgi) || defined(__sun) || defined(_MSC_VER))
    lookup = resolve_init(orb, "TradingService");
#else
    lookup = resolve_init<Lookup>(orb, "TradingService");
#endif

    // Read type_repos attribute to get IOR to type repository.
    CORBA::Object_var obj = lookup->type_repos();

    // Narrow.
    ServiceTypeRepository_var repos;
    repos = ServiceTypeRepository::_narrow(obj);
    if (CORBA::is_nil(repos)) {
        cerr << "Not a type repository reference" << endl;
        throw 0;
    }

    // Fill in property definitions for controllers.
    ServiceTypeRepository::PropStructSeq props;
    props.length(5);
    props[0].name = CORBA::string_dup("Model");
    props[0].value_type = CORBA::TypeCode::_duplicate(CORBA::_tc_string);
    props[0].mode = ServiceTypeRepository::PROP_MANDATORY_READONLY;

    props[1].name = CORBA::string_dup("Manufacturer");
    props[1].value_type = CORBA::TypeCode::_duplicate(
                            Manufacturing::_tc_AddressType
                          );
    props[1].mode = ServiceTypeRepository::PROP_NORMAL;

    props[2].name = CORBA::string_dup("Phone");
    props[2].value_type = CORBA::TypeCode::_duplicate(CORBA::_tc_string);
    props[2].mode = ServiceTypeRepository::PROP_MANDATORY;

    props[3].name = CORBA::string_dup("Supports");
    props[3].value_type = CORBA::TypeCode::_duplicate(
                            Airconditioning::_tc_ModelType
                          );
    props[3].mode = ServiceTypeRepository::PROP_READONLY;

    props[4].name = CORBA::string_dup("MaxDevices");
    props[4].value_type = CORBA::TypeCode::_duplicate(CORBA::_tc_ulong);
    props[4].mode = ServiceTypeRepository::PROP_NORMAL;

    // Create Controllers service type.
    ServiceTypeRepository::ServiceTypeNameSeq base_types;
    repos->add_type(
        "CCS::Controllers",
        "IDL:acme.com/CCS/Controller:1.0",
        props,
        base_types
    );

    return 0;
}

int
main(int argc, char* argv[], char*[])
{
    int status = 0;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << "Uncaught CORBA exception: " << ex << endl;
	status = 1;
    }
    catch(...)
    {
        status = 1;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = 1;
	}
    }
    
    return status;
}
