// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <BroadcasterAdmin_impl.h>
#include <Broadcaster_impl.h>

#include <stdlib.h>
#include <time.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace chat;

#ifdef HAVE_NO_GETHOSTNAME_PROTOTYPE
extern "C"
{
    int gethostname(char*,  int);
}
#endif

BroadcasterAdmin_impl::BroadcasterAdmin_impl(
    CORBA::ORB_ptr orb,
    PortableServer::POA_ptr adminPoa,
    PortableServer::POA_ptr broadcasterPoa
)
    : orb_(CORBA::ORB::_duplicate(orb)),
      adminPoa_(PortableServer::POA::_duplicate(adminPoa)),
      broadcasterPoa_(PortableServer::POA::_duplicate(broadcasterPoa)),
      nextBroadcasterId_(0)
{
    // All initialization done
}

PortableServer::POA_ptr
BroadcasterAdmin_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(adminPoa_);
}

//
// Remove Receiver from internal list
//

void
BroadcasterAdmin_impl::remove_receiver(CORBA::ULong id)
{
    //
    // Search Receiver
    //
    ReceiverList::iterator iter;
    for (iter = receiverList_.begin(); iter != receiverList_.end(); iter++)
    {
	if((*iter).broadcasterId == id)
	{
	    break;
	} 
    }

    //
    // Erase corresponding entry from list
    //
    receiverList_.erase(iter);
}

//
// Internal operations - also used by Broadcaster_impl servants
//

void
BroadcasterAdmin_impl::setNickName(CORBA::ULong id, const char* nick)
    throw(NickExists)
{
    ReceiverList::iterator iter;

    //
    // Check if nick already exists
    //
    for (iter = receiverList_.begin(); iter != receiverList_.end(); iter++)
    {
	if(strcmp(((*iter).receiverDesc.nick), nick) == 0)
	{
	    throw NickExists();  
	} 
    }
	     
    //
    // Search receiver
    //
    for (iter = receiverList_.begin(); iter != receiverList_.end(); iter++)
    {
	if ((*iter).broadcasterId == id)
        {
            //
            // Receiver found, set nick name
            //
	    CORBA::String_var oldNick = (*iter).receiverDesc.nick;
	    (*iter).receiverDesc.nick = nick;

	    //
	    // Send `is known as' message to all
	    //
	    const char* msg = " is now known as ";
	    
	    CORBA::String_var message = CORBA::string_alloc(
		4 + strlen(oldNick) + strlen(nick) + strlen(msg));
	    strcpy(message.inout(), "*** ");
	    strcat(message.inout(), oldNick);
	    strcat(message.inout(), msg);
	    strcat(message.inout(), nick);
	    
	    say(message);

	    return;
	}
    }
}

void
BroadcasterAdmin_impl::say(const char* text)
    throw(CORBA::SystemException)
{
    ReceiverList::iterator iter;

    //
    // Remember `bad' object references in a `bad list'
    //
    ReceiverList badList;

    for (iter = receiverList_.begin(); iter != receiverList_.end(); iter++)
    {
        try
        {
            (*iter).receiverDesc.recv -> print(text);
        }
        catch(const CORBA::Exception& ex)
        {
            cerr << ex << endl;

	    badList.push_back(*iter);
        }
    }
 
    //
    // Unsubsribe `bad' object references
    //
    for (iter = badList.begin(); iter != badList.end(); iter++)
    {
	remove_receiver((*iter).broadcasterId);
    }

    //
    // Send `vanish' messages
    //
    for (iter = badList.begin(); iter != badList.end(); iter++)
    {
	const char* msg = " vanished";
	
        CORBA::String_var message = CORBA::string_alloc(
	    4 +	strlen((*iter).receiverDesc.nick) + strlen(msg));
	strcpy(message.inout(), "*** ");
	strcat(message.inout(), (*iter).receiverDesc.nick);
	strcat(message.inout(), msg);
	
	say(message);
    }

    //
    // Print text as log message, too
    //
    cout << text << endl;
}

void
BroadcasterAdmin_impl::
getReceiverByNick(const char* nick, ReceiverDesc_out desc)
    throw(UnknownNick)
{
    //
    // Search for nick name
    //
    ReceiverList::iterator iter;
    for (iter = receiverList_.begin(); iter != receiverList_.end(); iter++)
    {
	if (!strcmp(((*iter).receiverDesc.nick), nick))
	{
	    desc = new ReceiverDesc((*iter).receiverDesc);
	    return;
    	}
    }

    //
    // Nick name was not found
    //
    throw UnknownNick();
}

StringSeq*
BroadcasterAdmin_impl::getReceiverNames()
{
    StringSeq* names = new StringSeq;
    names -> length(receiverList_.size());

    ReceiverList::iterator iter;
    unsigned long index = 0;
    for (iter = receiverList_.begin(); iter != receiverList_.end(); iter++)
    {
	char* s = CORBA::string_alloc(strlen((*iter).receiverDesc.nick) + 2 
				      + strlen((*iter).receiverDesc.id) + 1
				      + strlen((*iter).receiverDesc.host) +1);
	strcpy(s, (*iter).receiverDesc.nick);
	strcat(s, " (");
	strcat(s, (*iter).receiverDesc.id);
	strcat(s, "@");
	strcat(s, (*iter).receiverDesc.host);
	strcat(s, ")");
	
	(*names)[index++] = s;
    }

    return names;
}


//
// Operations from the .idl file
//

Broadcaster_ptr
BroadcasterAdmin_impl::subscribe(
    const ReceiverDesc& receiver)
    throw(NickExists, CORBA::SystemException)
{
    //
    // Do nothing if this is a nil object reference
    //
    if(CORBA::is_nil(receiver.recv))
    {
        return Broadcaster::_nil();
    }

    //
    // Check if nick already exists
    //
    ReceiverList::iterator iter;
    for (iter = receiverList_.begin(); iter != receiverList_.end(); iter++)
    {
	if(strcmp(((*iter).receiverDesc.nick), receiver.nick) == 0)
	{
	    throw NickExists();  
	} 
    }
	     
    //
    // Add receiver description to list
    //
    ReceiverListItem newReceiver;

    newReceiver.receiverDesc = receiver;
    newReceiver.broadcasterId = nextBroadcasterId_;
    receiverList_.push_back(newReceiver);

    //
    // Create servant
    //
    Broadcaster_impl * broadcasterImpl = 
        new Broadcaster_impl(this, nextBroadcasterId_, broadcasterPoa_);
    
    //
    // Switch to next id value
    //
    nextBroadcasterId_++;

    //
    // Create CORBA object - implicit activation policy assumed
    //
    Broadcaster_var newBroadcaster = broadcasterImpl->_this();

    //
    // Send `join' message to others
    //
    static const char* msg = ") has joined the chat";

    CORBA::String_var message =
	CORBA::string_alloc(4 + strlen(receiver.nick) + 
			    2 + strlen(receiver.id) + 
			    1 + strlen(receiver.host) + 
			    strlen(msg));

    strcpy(message.inout(), "*** ");
    strcat(message.inout(), receiver.nick);
    strcat(message.inout(), " (");
    strcat(message.inout(), receiver.id);
    strcat(message.inout(), "@");
    strcat(message.inout(), receiver.host);
    strcat(message.inout(), msg);

    say(message);

    return  newBroadcaster._retn();
}

Broadcaster_ptr
BroadcasterAdmin_impl::subscribeWithCheck(const ReceiverDesc& receiver)
    throw(ReceiverUnreachable,
	  NickExists,
	  CORBA::SystemException)
{
    //
    // Do nothing if this is a nil object reference
    //
    if(CORBA::is_nil(receiver.recv))
    {
        return Broadcaster::_nil();
    }

    //
    // Try to ping the receiver
    //
    CORBA::Boolean nonExist = true;
    try
    {
	nonExist = receiver.recv -> _non_existent();
    }
    catch(const CORBA::Exception& ex)
    {
        cerr << ex << endl;
    }

    if(nonExist)
    {
	throw ReceiverUnreachable();
    }

    //
    // Register receiver
    //
    return subscribe(receiver);
}

void
BroadcasterAdmin_impl::unregister(CORBA::ULong id)
{
    //
    // Remember nick name
    //
    CORBA::String_var nick;
    ReceiverList::iterator iter;
    for (iter = receiverList_.begin(); iter != receiverList_.end(); iter++)
    {
	if((*iter).broadcasterId == id)
	{
	    nick = (*iter).receiverDesc.nick;
	} 
    }
	     
    //
    // Remove receiver from list
    //
    remove_receiver(id);

    //
    // Send `leave' message to other
    //
    if((const char*)nick != 0)
    {
	const char* msg = " has left the chat";
	
	CORBA::String_var message =
	    CORBA::string_alloc(4 + strlen(nick) + strlen(msg));
	
	strcpy(message.inout(), "*** ");
	strcat(message.inout(), nick);
	strcat(message.inout(), msg);
	
	say(message);
    }
}






