// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Hashtable.h>
#include <OB/HashtableI.h>
#include <OB/Timer.h>

#include <LRUReaper.h>

#include <stdlib.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

// ----------------------------------------------------------------------
// LRUReaper::ObjectIdHasher public member implementations
// ----------------------------------------------------------------------

CORBA::ULong
LRUReaper::ObjectIdHasher::hash(const PortableServer::ObjectId& k)
{
    CORBA::ULong h = 0;
    CORBA::ULong offset = 0;
    CORBA::Long len = (CORBA::Long)k.length();

    if(len < 16)
    {
        for(CORBA::Long i = len ; i > 0 ; i--)
            h = (h * 37) + (int)k[offset++];
    }
    else
    {
        CORBA::ULong skip = len / 8;
        for(CORBA::Long i = len ; i > 0 ; i -= skip, offset += skip)
            h = (h * 39) + (int)k[offset];
    }

    return h;
}

bool
LRUReaper::ObjectIdHasher::comp(const PortableServer::ObjectId& k1,
				const PortableServer::ObjectId& k2)
{
    return ((k1.length() == k2.length()) &&
            (memcmp(k1.get_buffer(), k2.get_buffer(), k1.length()) == 0));
}

// ----------------------------------------------------------------------
// LRUReaper constructor and destructor
// ----------------------------------------------------------------------

LRUReaper::ServantInfo::ServantInfo(PortableServer::Servant s)
    : servant(s), count(0)
{
    last = OB::TimeNow();
}

// ----------------------------------------------------------------------
// LRUReaper constructor and destructor
// ----------------------------------------------------------------------

LRUReaper::LRUReaper(bool debug, CORBA::ULong maxIdle, ReapListener* listener)
    : maxIdle_(maxIdle), listener_(listener), debug_(debug)
{
    servantMap_ = new ServantMap(1023);
}

LRUReaper::~LRUReaper()
{
    ServantMap::Enumerator keys = servantMap_ -> keys();
    for(CORBA::ULong i = 0 ; i < keys.length() ; i++)
    {
        ServantInfo* info = 0;
        bool result = servantMap_ -> get(keys[i], info);
	assert(result);
	assert(info);
        servantMap_ -> remove(keys[i]);
        info -> servant -> _remove_ref();
        delete info;
    }
    delete servantMap_;
}

// ----------------------------------------------------------------------
// LRUReaper public member implementation
// ----------------------------------------------------------------------

void
LRUReaper::addServant(const PortableServer::ObjectId& oid,
                      PortableServer::Servant servant)
{
    JTCSynchronized synchronized(*this);

    if(! servantMap_ -> containsKey(oid))
    {
        ServantInfo* info = new ServantInfo(servant);
        servantMap_ -> put(oid, info);
        servant -> _add_ref();
    }
}

void
LRUReaper::removeServant(const PortableServer::ObjectId& oid)
{
    JTCSynchronized synchronized(*this);

    ServantInfo* info;
    if(servantMap_ -> get(oid, info))
    {
        servantMap_ -> remove(oid);
        info -> servant -> _remove_ref();
        delete info;
    }
}

PortableServer::Servant
LRUReaper::acquireServant(const PortableServer::ObjectId& oid)
{
    JTCSynchronized synchronized(*this);

    ServantInfo* info;

    if(! servantMap_ -> get(oid, info))
        return 0;

    info -> last = OB::TimeNow();
    info -> count++;

    if(debug_)
    {
        CORBA::String_var id = PortableServer::ObjectId_to_string(oid);
	cout << "Touching servant " << id << ", count = "
	     << info -> count << endl;
    }

    return info -> servant;
}

void
LRUReaper::releaseServant(const PortableServer::ObjectId& oid)
{
    JTCSynchronized synchronized(*this);

    ServantInfo* info;

    if(! servantMap_ -> get(oid, info))
        return;

    assert(info -> count > 0);
    info -> count--;

    if(debug_)
    {
        CORBA::String_var id = PortableServer::ObjectId_to_string(oid);
	cout << "Releasing servant " << id << ", count = "
	     << info -> count << endl;
    }
}

void
LRUReaper::reap()
{
    JTCSynchronized synchronized(*this);

    if(debug_)
        cout << "Starting reap" << endl;

    struct timeval now = OB::TimeNow();

    ServantMap::Enumerator keys = servantMap_ -> keys();
    for(CORBA::ULong i = 0 ; i < keys.length() ; i++)
    {
        ServantInfo* info = 0;
        bool result = servantMap_ -> get(keys[i], info);
	assert(result);
	assert(info);

        //
        // Determine the servant's age since last use
        //
        CORBA::ULong age = now.tv_sec - info -> last.tv_sec;

        //
        // Only reap the servant if the age exceeds maxIdle and the
        // usage count == 0 (the count will be 0 when the servant is
        // not currently in use by the application).
        //
        if(age > maxIdle_ && info -> count == 0)
        {
            if(debug_)
            {
                CORBA::String_var id =
                    PortableServer::ObjectId_to_string(keys[i]);
                cout << "servant " << id << ", age = " << age << endl;
            }

            servantMap_ -> remove(keys[i]);
            if(listener_)
                listener_ -> servantReaped(keys[i], info -> servant);
            info -> servant -> _remove_ref();
            delete info;
        }
    }
}
