// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <LocatorFactory_impl.h>
#include <Database.h>
#include <LRUReaper.h>
#include <ReapTimer.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

// ----------------------------------------------------------------------
// DBRecord servant class
// ----------------------------------------------------------------------

class LocRecord_impl : virtual public POA_DBRecord,
                       virtual public PortableServer::RefCountServantBase
{
    DBRecordData* data_;
    bool debug_;

public:

    LocRecord_impl(DBRecordData* data, bool debug)
        : data_(data),
	  debug_(debug)
    {
        if(debug_)
	{
	    cout << "LocRecord_impl (" << data_ -> key << ", `"
	         << data_ -> data << "') instantiated." << endl;
        }
    }

    virtual ~LocRecord_impl()
    {
        if(debug_)
	{
	    cout << "LocRecord_impl (" << data_ -> key << ", `"
	         << data_ -> data << "') deleted." << endl;
        }
    }

    //
    // Standard IDL to C++ Mapping
    //
    virtual CORBA::Long key()
        throw(CORBA::SystemException)
    {
        return data_ -> key;
    }

    virtual char* data()
        throw(CORBA::SystemException)
    {
        return CORBA::string_dup(data_ -> data);
    }

    virtual void data(const char* str)
        throw(CORBA::SystemException)
    {
        data_ -> data = CORBA::string_dup(str);
    }
};

// ----------------------------------------------------------------------
// Servant locator class
// ----------------------------------------------------------------------

class Locator : public POA_PortableServer::ServantLocator
{
    Database* database_;
    Reaper* reaper_;
    bool debug_;

public:

    Locator(Database* database, Reaper* reaper, bool debug)
        : database_(database), reaper_(reaper), debug_(debug)
    {
    }

    //
    // Standard IDL to C++ Mapping
    //
    PortableServer::Servant preinvoke(const PortableServer::ObjectId& oid,
        			      PortableServer::POA_ptr /* poa */,
        			      const char* /* operation */,
        			      PortableServer::ServantLocator::Cookie&
				      /* the_cookie */)
        throw(CORBA::SystemException)
    {
        //
        // First attempt to find the servant in the reaper
        //
        PortableServer::Servant result = reaper_ -> acquireServant(oid);

        //
        // If that fails, then make sure the database contains a
        // record for the key, then create a new servant
        //
        if(result == 0)
        {
            CORBA::String_var id = PortableServer::ObjectId_to_string(oid);
            long key = atol(id);
            DBRecordData* data = database_ -> find(key);
            if(data != 0)
            {
                result = new LocRecord_impl(data, debug_);
                reaper_ -> addServant(oid, result);
                reaper_ -> acquireServant(oid);
                result -> _remove_ref(); // give ownership to reaper
            }
        }

        if(result == 0)
            throw CORBA::OBJECT_NOT_EXIST();

        return result;
    }

    void postinvoke(const PortableServer::ObjectId& oid,
                    PortableServer::POA_ptr /* poa */,
                    const char* /* operation */,
                    PortableServer::ServantLocator::Cookie /* the_cookie */,
                    PortableServer::Servant /* the_servant */)
        throw(CORBA::SystemException)
    {
        reaper_ -> releaseServant(oid);
    }
};

// ----------------------------------------------------------------------
// LocatorFactory_impl private member implementation
// ----------------------------------------------------------------------

PortableServer::ObjectId*
LocatorFactory_impl::makeKey(CORBA::Long key)
{
    //
    // Create the object ID for the record
    //
    char keyBuff[20];
    sprintf(keyBuff, "%ld", key);
    return PortableServer::string_to_ObjectId(keyBuff);
}

DBRecord_ptr
LocatorFactory_impl::makeRef(CORBA::Long key)
{
    //
    // Manufacture an object reference for the record
    //
    PortableServer::ObjectId_var oid = makeKey(key);
    CORBA::Object_var obj =
        poa_ -> create_reference_with_id(oid, "IDL:DBRecord:1.0");
    DBRecord_ptr result = DBRecord::_narrow(obj);
    return result;
}

// ----------------------------------------------------------------------
// LocatorFactory_impl constructor and destructor
// ----------------------------------------------------------------------

LocatorFactory_impl::LocatorFactory_impl(CORBA::ORB_ptr orb,
                                         PortableServer::POA_ptr parent,
					 bool debug)
    : nextKey_(0),
      database_(new Database),
      orb_(CORBA::ORB::_duplicate(orb)),
      mypoa_(PortableServer::POA::_duplicate(parent))
{
    //
    // Create POA for use by factory
    //
    PortableServer::POAManager_var poaManager = parent -> the_POAManager();
    CORBA::PolicyList policies;
    policies.length(6);
    policies[0] =
        parent -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] =
        parent -> create_lifespan_policy(PortableServer::PERSISTENT);
    policies[2] =
        parent -> create_servant_retention_policy(PortableServer::NON_RETAIN);
    policies[3] =
        parent -> create_id_uniqueness_policy(PortableServer::UNIQUE_ID);
    policies[4] = parent ->
        create_request_processing_policy(PortableServer::USE_SERVANT_MANAGER);
    policies[5] = parent ->
     create_implicit_activation_policy(PortableServer::NO_IMPLICIT_ACTIVATION);
    poa_ = mypoa_ -> create_POA("locator_factory", poaManager, policies);

    //
    // Create reaper and timer. The maxIdle value for the reaper is 2,
    // meaning objects that have been idle for 2+ seconds are eligible
    // for reaping. The timer will invoke the reaper once every second.
    //
    // Note: These timeout values are much too small for most applications.
    //
    reaper_ = new LRUReaper(debug, 2);
    timer_ = new ReapTimer(reaper_, 1);
    timer_ -> start();

    //
    // Create and install servant locator
    //
    locator_ = new Locator(database_, reaper_, debug);
    PortableServer::ServantLocator_var locator = locator_ -> _this();
    poa_ -> set_servant_manager(locator);
}

LocatorFactory_impl::~LocatorFactory_impl()
{
    poa_ -> destroy(true, true);
    delete locator_;

    timer_ -> destroy();
    while(timer_ -> isAlive())
    {
	try
	{
	    timer_ -> join();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    delete reaper_;
    delete database_;
}

// ----------------------------------------------------------------------
// LocatorFactory public member implementation
// ----------------------------------------------------------------------

//
// Create a new record
//
// To create a new key we have an constantly incrementing nextKey
// member variable. For a real application this isn't good enough,
// but for this demonstration it will suffice.
//
DBRecord_ptr
LocatorFactory_impl::createRecord()
    throw(CORBA::SystemException)
{
    //
    // Add record in database - the servant locator will create
    // the servant on demand
    //
    CORBA::Long key = nextKey_++;
    database_ -> add(key);
    return makeRef(key);
}

//
// Destroy the record with `key'. If the record isn't found
// then raise a NoRecord exception.
//
void
LocatorFactory_impl::destroyByKey(CORBA::Long key)
    throw(DBFactory::NoRecord,
          CORBA::SystemException)
{
    //
    // Ask the database to destroy the record
    //
    if(database_ -> find(key) == 0)
        throw DBFactory::NoRecord();

    PortableServer::ObjectId_var oid = makeKey(key);
    reaper_ -> removeServant(oid);
    database_ -> destroy(key);
}

//
// Retrieve the record identified by `key'. If the record isn't
// found then raise a NoRecord exception.
//
DBRecord_ptr
LocatorFactory_impl::getByKey(CORBA::Long key)
    throw(DBFactory::NoRecord,
          CORBA::SystemException)
{
    //
    // Look in the database
    //
    if(database_ -> find(key) != 0)
        return makeRef(key);

    throw DBFactory::NoRecord();
    return DBRecord::_nil(); // Some compilers need this
}

//
// Return the keys of all records in the database
//
DBFactory::DBRecordKeySeq*
LocatorFactory_impl::getAllKeys()
    throw(CORBA::SystemException)
{
    DBFactory::DBRecordKeySeq_var r =
        new DBFactory::DBRecordKeySeq(database_ -> size());
    r -> length(database_ -> size());
    for(CORBA::ULong i = 0 ; i < database_ -> size() ; i++)
        r[i] = database_ -> recordAt(i) -> key;

    return r._retn();
}

PortableServer::POA_ptr
LocatorFactory_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(mypoa_);
}
