// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/Timer.h>

#include <Peer_impl.h>

#include <time.h>
#include <stdlib.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

int
run(CORBA::ORB_ptr orb, int /* argc */, char* argv[])
{
    //
    // Get peer
    //
    CORBA::Object_var obj = orb -> string_to_object("relfile:/Peer.ref");
    if(CORBA::is_nil(obj))
    {
        cerr << argv[0] << ": cannot read IOR from Peer.ref"
             << endl;
        return EXIT_FAILURE;
    }

    Peer_var p1 = Peer::_narrow(obj);
    assert(!CORBA::is_nil(p1));

    //
    // Resolve the Root POA
    //
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(
        CORBA::Object_var(orb -> resolve_initial_references("RootPOA")));

    //
    // Create implementation object
    //
    Peer_impl* peerImpl = new Peer_impl(rootPOA);
    PortableServer::ServantBase_var servant = peerImpl;
    Peer_var p2 = peerImpl -> _this();

    //
    // Initialize the servers so that incoming requests are
    // handled from this point on
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();
    manager -> activate();

    int status = EXIT_SUCCESS;
    try
    {
	while(true)
	{
	    //
	    // Start recursive decrement
	    //
	    cout << "Enter a number (0 to exit): ";
	    CORBA::Long value;
	    cin >> value;
	    
	    if(value <= 0)
		break;
	    
	    //
	    // Run demo
	    //
	    timeval tv = OB::TimeNow();
	    p1 -> decrement(value, p2);
	    tv = OB::TimeSub(OB::TimeNow(), tv);
	    cout << "Time needed: "
		 << ((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0)
		 << endl;
	}
    }
    catch(const CORBA::SystemException& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    //
    // Shutdown the ORB. Since a stack based servant is used the ORB
    // must be shutdown before returning from run() otherwise a
    // servant that has gone out of scope can still be attached to the
    // ORB. An alternative to shutting down the ORB is to deactivate
    // the servant.
    //
    orb -> shutdown(true);

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);

        //
        // Check to see if the concurrency model is usable for nested
        // method invocations. Either reactive/reactive or threaded/
        // thread_per_request/thread_pool is needed.
        //
        const char* oaModel =
	    properties -> getProperty("ooc.orb.oa.conc_model");
        const char* orbModel =
	    properties -> getProperty("ooc.orb.conc_model");
        bool valid;
        if(oaModel != 0 && orbModel != 0)
        {
            valid = ((strcmp(orbModel, "reactive") == 0 &&
                     strcmp(oaModel, "reactive") == 0) ||
                     (strcmp(orbModel, "threaded") == 0 &&
                     (strcmp(oaModel, "thread_per_request") == 0 ||
                      strcmp(oaModel, "thread_pool") == 0)));
        }
        else
            valid = false;

	//
	// Set the concurrency model to reactive/reactive if a valid
        // concurrency model wasn't selected
	//
        if(!valid)
        {
	    properties -> setProperty("ooc.orb.conc_model", "reactive");
	    properties -> setProperty("ooc.orb.oa.conc_model", "reactive");
        }

	properties -> setProperty("ooc.orb.id", "Nested-Peer2");

	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
