// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/Net.h>
#include <OB/IMR.h>

#include <IMRUtil.h>

#include <time.h>

const char* emptyBranch = "    ";
const char* verticalBranch = "|   ";
const char* nodeBranch = "|-- ";
const char* bottomNodeBranch = "`-- ";

void
usage(const char* progName)
{
    OB_CERR << "Usage:\n"
        << progName << " options\n"
        "\n"
        "Options:\n"
        "-h, --help                              "
	"Display this information.\n"
        "--add-oad [host]                        "
	"Register an OAD for the specified\n"
        "                                        host.\n"
        "--add-server server-name [exec [host]]  "
	"Register a server under the OAD\n"
        "                                        "
        "specified by host with the given\n"
        "                                        exec attribute.\n"
        "--add-poa server-name poa-name          "
	"Register a POA for the specified\n"
        "                                        server.\n"
        "--remove-oad [host]                     Unregister an OAD.\n"
        "--remove-server server-name             Unregister a server.\n"
        "--remove-poa server-name poa-name       Unregister a POA.\n"
        "--get-oad-status [host]                 "
	"Get the status of an OAD.\n"
        "--get-server-info server-name           "
	"Get the attributes and state\n"
        "                                        "
        "information for a server.\n"
        "--get-poa-status server-name poa-name   Get the status of a POA.\n"
        "--list-oads                             List all OADs.\n"
        "--list-servers                          List all servers.\n"
        "--list-poas server-name                 List all POAs.\n"
        "--tree                                  "
	"Display all OADs, servers and\n"
        "                                        POAs in a tree like format.\n"
        "--tree-oad [host]                       "
	"Display an OAD and its associated\n"
        "                                        "
	"servers and POAs in a tree like\n"
        "                                        format.\n"
        "--tree-server server-name               Display a server and its\n"
        "                                        "
	"associated POAs in a tree like\n"
        "                                        format.\n"
        "--set-server server-name {exec|host|    Set an attribute of a server\n"
        "  args|rundir|mode|activate-poas|       "
	"(e.g., --set-server srv max-spawns 2\n"
        "  update-timeout|failure-timeout|       "
	"sets the max-spawns attribute for\n"
        "  max-spawns} value                     the server srv to 2).\n"
        "--start-server server-name              Start a server.\n"
        "--stop-server server-name               Stop a server.\n"
        "--reset-server server-name              Reset a server.\n";
}

char*
getHost(int argc, char** argv, int pos)
{
    CORBA::String_var host;
    if(pos == argc - 1)
        host = CORBA::string_dup(argv[pos]);
    else
    {
        //
        // MAXHOSTNAMELEN + 1 is used since we're not sure whether the
        // len parameter to gethostname() includes '\0' or not.
        //
        char hostname[MAXHOSTNAMELEN + 1];
        if(gethostname(hostname, MAXHOSTNAMELEN) == -1)
        {
            OB_CERR << "gethostname failed" << OB_ENDL;
            exit(1);
        }
        
        host = CORBA::string_dup(hostname);
    }
    
    return host._retn();
}

IMR::POAName*
parsePOAName(const char* serverName, const char* sPOAName)
{
    IMR::POAName_var poaName = new IMR::POAName;

    CORBA::ULong poaNameLength  = poaName -> length();
    poaName -> length(poaNameLength + 1);
    poaName[poaNameLength] = serverName;

    const char* p = strtok((char *)sPOAName, "/");
    while(p)
    {
        int len = strcspn(p, "/");
        char* s = CORBA::string_alloc(len);
        strncpy(s, p, len);
        s[len] = '\0';

        CORBA::ULong poaNameLength  = poaName -> length();
        poaName -> length(poaNameLength + 1);
        poaName[poaNameLength] = s;
                          
        p = strtok(0, "/");
    }

    return poaName._retn();
}

char*
poaNameToString(const IMR::POAName& poaName)
{
    CORBA::String_var s = CORBA::string_dup("");

    for(unsigned int i = 1; i < poaName.length(); i++)
    {
        if(i != 1)
            s += "/";
        s += poaName[i];
    }

    return s._retn();
}

void
displayTree(const IMR::POAInfo& poa,
            const char* prefix = 0,
            bool lastNode = false)
{
    if(prefix != 0)
    {
        OB_COUT << prefix;

        if(lastNode)
            OB_COUT << bottomNodeBranch;
        else
            OB_COUT << nodeBranch;
    }

    CORBA::String_var s = poaNameToString(poa.poa);
    OB_COUT << s << "  (" << IMRUtil::poaStatusToStr(poa.state)
	    << ")" << OB_ENDL;
}

void
displayTree(IMR::Server_ptr sa,
            const char* prefix = 0,
            bool lastNode = false)
{
    CORBA::String_var newPrefix;
    if(prefix != 0)
    {
        newPrefix = CORBA::string_dup(prefix);
        OB_COUT << prefix;

        if(lastNode)
        {
            newPrefix += emptyBranch;
            OB_COUT << bottomNodeBranch;
        }
        else
        {
            newPrefix += verticalBranch;
            OB_COUT << nodeBranch;
        }
    }
    else
        newPrefix = CORBA::string_dup("");
    
    CORBA::String_var s = sa -> name();
    OB_COUT << s;

    IMR::ServerStatus ss;
    ss = sa -> status();
    OB_COUT << " (" << IMRUtil::serverStatusToStr(ss) << ")"
		 << OB_ENDL;

    IMR::POAInfoSeq_var poas = sa -> list_poas();
    for(CORBA::ULong i = 0; i < poas -> length(); i++)
        displayTree(poas[i], newPrefix, poas -> length() == i + 1);
}

void
displayTree(IMR::ServerDomain_ptr domain,
            const IMR::OADInfo& oi,
            const char* prefix = 0,
            bool lastNode = false)
{
    IMR::ServerFactory_var factory = domain -> get_server_factory();

    CORBA::String_var newPrefix;
    if(prefix != 0)
    {
        newPrefix = CORBA::string_dup(prefix);
        OB_COUT << prefix;

        if(lastNode)
        {
            newPrefix += emptyBranch;
            OB_COUT << bottomNodeBranch;
        }
        else
        {
            newPrefix += verticalBranch;
            OB_COUT << nodeBranch;
        }
    }
    else
        newPrefix = CORBA::string_dup("");
    
    CORBA::String_var oadHost;
    oadHost = oi.host;
    OB_COUT << oadHost << "  ("
		 << (oi.up ? "up" : "down")
		 << ")" << OB_ENDL;

    IMR::ServerSeq_var ss;
    ss = factory -> list_servers_by_host(oadHost);

    for(CORBA::ULong i = 0; i < ss -> length(); i++)
        displayTree(ss[i], newPrefix, ss -> length() == i + 1);
}

void
displayTree(IMR::ServerDomain_ptr domain)
{
    OB_COUT << "domain" << OB_ENDL;

    IMR::OADInfoSeq_var ois;
    ois = domain -> list_oads();
            
    IMR::ServerFactory_var factory = domain -> get_server_factory();

    IMR::ServerSeq_var ss;
    ss = factory -> list_orphaned_servers();
    bool orphaned = ss -> length() != 0;

    for(CORBA::ULong i = 0; i < ois -> length(); i++)
        displayTree(domain, ois[i], "", 
		    orphaned ? false : ois -> length() == i + 1);

    if(ss -> length() != 0)
    {
        OB_COUT << bottomNodeBranch << "Orphaned Servers" << OB_ENDL;

        for(CORBA::ULong i = 0; i < ss -> length(); i++)
            displayTree(ss[i], emptyBranch, ss -> length() == i + 1);
    }
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    const char* progName = argv[0];

    //
    // Parse command-line arguments for operation type and object type
    //
    if(argc < 2)
    {
        usage(progName);
        return EXIT_FAILURE;
    }

    enum {ADD, REMOVE, GET, LIST, TREE, SET, RESET, START, STOP} opType;
    enum {ROOT, OAD, SERVER, POA } objType;

    if(strcmp(argv[1], "--add-oad") == 0)
    {
        opType = ADD;
        objType = OAD;
    }
    else if(strcmp(argv[1], "--add-server") == 0)
    {
        opType = ADD;
        objType = SERVER;
    }
    else if(strcmp(argv[1], "--add-poa") == 0)
    {
        opType = ADD;
        objType = POA;
    }
    else if(strcmp(argv[1], "--remove-oad") == 0)
    {
        opType = REMOVE;
        objType = OAD;
    }
    else if(strcmp(argv[1], "--remove-server") == 0)
    {
        opType = REMOVE;
        objType = SERVER;
    }
    else if(strcmp(argv[1], "--remove-poa") == 0)
    {
        opType = REMOVE;
        objType = POA;
    }
    else if(strcmp(argv[1], "--get-oad-status") == 0)
    {
        opType = GET;
        objType = OAD;
    }
    else if(strcmp(argv[1], "--get-server-info") == 0)
    {
        opType = GET;
        objType = SERVER;
    }
    else if(strcmp(argv[1], "--get-poa-status") == 0)
    {
        opType = GET;
        objType = POA;
    }
    else if(strcmp(argv[1], "--list-oads") == 0)
    {
        opType = LIST;
        objType = OAD;
    }
    else if(strcmp(argv[1], "--list-servers") == 0)
    {
        opType = LIST;
        objType = SERVER;
    }
    else if(strcmp(argv[1], "--list-poas") == 0)
    {
        opType = LIST;
        objType = POA;
    }
    else if(strcmp(argv[1], "--tree") == 0)
    {
        opType = TREE;
        objType = ROOT;
    }
    else if(strcmp(argv[1], "--tree-oad") == 0)
    {
        opType = TREE;
        objType = OAD;
    }
    else if(strcmp(argv[1], "--tree-server") == 0)
    {
        opType = TREE;
        objType = SERVER;
    }
    else if(strcmp(argv[1], "--set-server") == 0)
    {
        opType = SET;
        objType = SERVER;
    }
    else if(strcmp(argv[1], "--reset-server") == 0)
    {
        opType = RESET;
        objType = SERVER;
    }
    else if(strcmp(argv[1], "--start-server") == 0)
    {
        opType = START;
        objType = SERVER;
    }
    else if(strcmp(argv[1], "--stop-server") == 0)
    {
        opType = STOP;
        objType = SERVER;
    }
    else
    {
        usage(progName);
        return EXIT_FAILURE;
    }
    
    //
    // Get ServerFactory
    //
    CORBA::Object_var imrObj;
    try
    {
        imrObj = orb -> resolve_initial_references("IMR");
    }
    catch(const CORBA::ORB::InvalidName&)
    {
        OB_CERR << "No `IMR' initial reference." << OB_ENDL;
        return EXIT_FAILURE;
    }
    if(CORBA::is_nil(imrObj))
    {
        OB_CERR << "No `IMR' initial reference." << OB_ENDL;
        return EXIT_FAILURE;
    }
    IMR::ServerDomain_var domain = IMR::ServerDomain::_narrow(imrObj);
    if(CORBA::is_nil(domain))
    {
        OB_CERR << "`IMR' initial reference is not an IMR::ServerDomain."
		     << OB_ENDL;
        return EXIT_FAILURE;
    }

    IMR::ServerFactory_var factory = domain -> get_server_factory();
        
    //
    // Perform operation
    //
    if(opType == ADD)
    {
        if(objType == OAD)
        {
            if(argc > 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            CORBA::String_var host = getHost(argc, argv, 2);
            
            try
            {
                domain -> create_oad_record(host);
            }
            catch(const IMR::OADAlreadyExists&)
            {
                OB_COUT << "Failed: oad record already exists"
			     << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else if(objType == SERVER)
        {
            if(argc < 3 || argc > 5)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            CORBA::String_var serverName = CORBA::string_dup(argv[2]);
            CORBA::String_var exec = (argc == 3) ? CORBA::string_dup("")
	    				         : CORBA::string_dup(argv[3]);
            CORBA::String_var host = getHost(argc, argv, 4);

            IMR::Server_var s;
            try
            {
                s = factory -> create_server_record(serverName);
                assert(!CORBA::is_nil(s));

                s -> host(host);
                s -> exec(exec);
            }
            catch(const IMR::ServerAlreadyRegistered&)
            {
                OB_COUT << "Failed: server record already exists"
			     << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else if(objType == POA)
        {
            if(argc < 4 || argc > 5)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            CORBA::String_var serverName = CORBA::string_dup(argv[2]);
            IMR::POAName_var poaName = parsePOAName(serverName, argv[3]);

            try
            {
                IMR::Server_var server = factory -> get_server(serverName);
                assert(!CORBA::is_nil(server));

                server -> create_poa_record(poaName);
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: record doesn't exist"
			     << OB_ENDL;
                return EXIT_FAILURE;
            }
            catch(const IMR::POAAlreadyRegistered&)
            {
                OB_COUT << "Failed: POA Already Registered"
			     << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else
        {
            usage(progName);
            return EXIT_FAILURE;
        }
    }
    else if(opType == REMOVE)
    {
        if(objType == OAD)
        {
            if(argc > 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            CORBA::String_var host = getHost(argc, argv, 2);

            try
            {
                domain -> remove_oad_record(host);
            }
            catch(const IMR::NoSuchOAD&)
            {
                OB_COUT << "Failed: OAD record doesn't exist"
			     << OB_ENDL;
                return EXIT_FAILURE;
            }
            catch(const IMR::OADRunning&)
            {
                OB_COUT << "Failed: OAD is running" << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else if(objType == SERVER)
        {
            if(argc != 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }
            CORBA::String_var serverName = CORBA::string_dup(argv[2]);

            IMR::Server_var server;
            try
            {
                server = factory -> get_server(serverName);
                assert(!CORBA::is_nil(server));
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: No such server" << OB_ENDL;
                return EXIT_FAILURE;
            }

	    try
	    {
                server -> destroy();
	    }
	    catch(const IMR::ServerRunning&)
	    {
	        OB_COUT << "Failed: Server still active" << OB_ENDL;
		return EXIT_FAILURE;
	    }
        }
        else if(objType == POA)
        {
            if(argc != 4)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            CORBA::String_var serverName = CORBA::string_dup(argv[2]);
            IMR::POAName_var poaName = parsePOAName(serverName, argv[3]);

            IMR::Server_var server;
            try
            {
                server = factory -> get_server(serverName);
                assert(!CORBA::is_nil(server));
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: No such server" << OB_ENDL;
                return EXIT_FAILURE;
            }
            
            try
            {
                server -> remove_poa_record(poaName);
            }
            catch(const IMR::NoSuchPOA&)
            {
                OB_COUT << "Failed: No such POA" << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else
        {
            usage(progName);
            return EXIT_FAILURE;
        }
    }
    else if(opType == GET)
    {
        if(objType == OAD)
        {
            if(argc > 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            CORBA::String_var host = getHost(argc, argv, 2);

            IMR::OADInfo_var oadInfo;
            try
            {
                oadInfo = domain -> get_oad_record(host);
            }
            catch(const IMR::NoSuchOAD&)
            {
                OB_COUT << "Failed: no such OAD record exists"
			     << OB_ENDL;
                return EXIT_FAILURE;
            }

            OB_COUT << "OAD " << oadInfo -> host << ":" << OB_ENDL;
            OB_COUT << "    Status: ";
            if(oadInfo -> up)
                OB_COUT << "up";
            else
                OB_COUT << "down";
            OB_COUT << OB_ENDL;
        }
        else if(objType == SERVER)
        {
            if(argc != 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }
            CORBA::String_var serverName = CORBA::string_dup(argv[2]);

            IMR::Server_var server;
            try
            {
                server = factory -> get_server(serverName);
                assert(!CORBA::is_nil(server));
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: No such server" << OB_ENDL;
                return EXIT_FAILURE;
            }

            CORBA::Long l;
            CORBA::String_var s;

            s = server -> name();
            OB_COUT << "Server " << s << ":" << OB_ENDL;

            l = server -> id();
            OB_COUT << "    ID:                       " << l
			 << OB_ENDL;

            IMR::ServerStatus ss;
            ss = server -> status();
            OB_COUT << "    Status:                   ";
            OB_COUT << IMRUtil::serverStatusToStr(ss) << OB_ENDL;
        
            s = server -> name();
            OB_COUT << "    Name:                     " << s << OB_ENDL;
            s = server -> host();
            OB_COUT << "    Host:                     " << s << OB_ENDL;
            s = server -> exec();
            OB_COUT << "    Path:                     " << s << OB_ENDL;
            s = server -> directory();
            OB_COUT << "    RunDir:                   " << s << OB_ENDL;

            IMR::ArgList_var args = server -> args();
            OB_COUT << "    Arguments:                ";
            for(CORBA::ULong i = 0; i < args -> length(); i++)
            {
                if(i != 0)
                    OB_COUT << " ";
                OB_COUT << args[i];
            }
            OB_COUT << OB_ENDL;

            IMR::ServerActivationMode mode = server -> mode();
            OB_COUT << "    Activation Mode:          ";
	    switch(mode)
	    {
	    case IMR::SHARED:
	        OB_COUT << "shared\n";
	        break;
	    case IMR::PERSISTENT:
	        OB_COUT << "persistent\n";
	        break;
	    }

            OB_COUT << "    POA Activation:           ";
 	    OB_COUT << (server -> activatePOAs() ? "true" : "false")
			 << OB_ENDL;

            l = server -> updateTimeout();
            OB_COUT << "    Update timeout (ms):      " << l << OB_ENDL;
            l = server -> failureTimeout();
            OB_COUT << "    Failure timeout (secs):   " << l << OB_ENDL;
            l = server -> maxForks();
            OB_COUT << "    Maximum spawn count:      " << l << OB_ENDL;

            CORBA::Boolean b;
            b = server -> manual();
            OB_COUT << "    Started manually:         ";
            if(b)
                OB_COUT << "yes" << OB_ENDL;
            else
                OB_COUT << "no" << OB_ENDL;

            l = server -> timesForked();
            OB_COUT << "    Number of times spawned:  " << l
			 << OB_ENDL;
        }
        else if(objType == POA)
        {
            if(argc != 4)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            CORBA::String_var serverName = CORBA::string_dup(argv[2]);
            IMR::POAName_var poaName = parsePOAName(serverName, argv[3]);

            IMR::Server_var server;
            try
            {
                server = factory -> get_server(serverName);
                assert(!CORBA::is_nil(server));
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: No such server" << OB_ENDL;
                return EXIT_FAILURE;
            }

            IMR::POAInfo_var poaInfo;
        
            try
            {
                poaInfo = server -> get_poa_info(poaName);
            }
            catch(const IMR::NoSuchPOA&)
            {
                OB_COUT << "Failed: No such POA" << OB_ENDL;
                return EXIT_FAILURE;
            }
        
            CORBA::String_var sPOAName = poaNameToString(poaName);
            OB_COUT << "POA " << sPOAName << ":" << OB_ENDL;

            OB_COUT << "    Status:    ";
	    OB_COUT << IMRUtil::poaStatusToStr(poaInfo -> state)
			 << OB_ENDL;
        }
        else
        {
            usage(progName);
            return EXIT_FAILURE;
        }
    }
    else if(opType == LIST)
    {
        if(objType == OAD)
        {
            if(argc != 2)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            IMR::OADInfoSeq_var oadInfos;
            oadInfos = domain -> list_oads();

            OB_COUT << "OAD records:" << OB_ENDL;
        
            CORBA::ULong p;
            for(p = 0; p < oadInfos -> length(); p++ )
            {
                CORBA::String_var s;

                s = oadInfos[p].host;
                OB_COUT << "    " << s;
                if(oadInfos[p].up)
                    OB_COUT << " (up)";
                else
                    OB_COUT << " (down)";
                OB_COUT << OB_ENDL;
            }
        }
        else if(objType == SERVER)
        {
            if(argc != 2)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            IMR::ServerSeq_var servers;

            servers = factory -> list_servers();

            OB_COUT << "Server records:" << OB_ENDL;
        
            CORBA::ULong p;
            for(p = 0; p < servers -> length(); p++ )
            {
                CORBA::String_var s;

                s = servers[p] -> name();
                OB_COUT << "    " << s << OB_ENDL;
            }
        }
        else if(objType == POA)
        {
            if(argc != 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }
            CORBA::String_var serverName = CORBA::string_dup(argv[2]);

            IMR::Server_var server;
            try
            {
                server = factory -> get_server(serverName);
                assert(!CORBA::is_nil(server));
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: No such server" << OB_ENDL;
                return EXIT_FAILURE;
            }

            IMR::POAInfoSeq_var seq;
            seq = server -> list_poas();
        
            OB_COUT << "POA records under server " << serverName << ":"
			 << OB_ENDL;
            for(CORBA::ULong i = 0; i < seq -> length(); i++)
            {
                CORBA::String_var s = poaNameToString(seq[i].poa);
                OB_COUT << s << OB_ENDL;
            }
        }
        else
        {
            usage(progName);
            return EXIT_FAILURE;
        }
    }
    else if(opType == TREE)
    {
        if(objType == ROOT)
        {
            displayTree(domain);
        }
        else if(objType == OAD)
        {
            if(argc > 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            CORBA::String_var host = getHost(argc, argv, 2);

            IMR::OADInfo_var oadInfo;
            try
            {
                oadInfo = domain -> get_oad_record(host);
            }
            catch(const IMR::NoSuchOAD&)
            {
                OB_COUT << "Failed: no such OAD record exists"
			     << OB_ENDL;
                return EXIT_FAILURE;
            }

            displayTree(domain, oadInfo);
            
        }
        else if(objType == SERVER)
        {
            if(argc != 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }
            CORBA::String_var serverName = CORBA::string_dup(argv[2]);

            IMR::Server_var server;
            try
            {
                server = factory -> get_server(serverName);
                assert(!CORBA::is_nil(server));
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: No such server" << OB_ENDL;
                return EXIT_FAILURE;
            }

            displayTree(server);
        }
        else
        {
            usage(progName);
            return EXIT_FAILURE;
        }
    }
    else if(opType == SET)
    {
        if(objType == SERVER)
        {
            if(argc != 5)
            {
                usage(progName);
                return EXIT_FAILURE;
            }

            CORBA::String_var serverName = CORBA::string_dup(argv[2]);
            IMR::Server_var server;
            try
            {
                server = factory -> get_server(serverName);
                assert(!CORBA::is_nil(server));
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: No such server" << OB_ENDL;
                return EXIT_FAILURE;
            }

            if(strcmp(argv[3], "exec") == 0)
            {
                CORBA::String_var exec = CORBA::string_dup(argv[4]);

                server -> exec(exec);
            }
            else if(strcmp(argv[3], "args") == 0)
            {
                IMR::ArgList args;

                const char* whitespace = " \t\n";
                const char* p = strtok(argv[4], whitespace);
                while(p)
                {
                    int len = strcspn(p, whitespace);
                    char* arg = CORBA::string_alloc(len);
                    strncpy(arg, p, len);
                    arg[len] = '\0';

                    CORBA::ULong argsLength = args.length();
                    args.length(argsLength + 1);
                    args[argsLength] = arg;

                    p = strtok(0, whitespace);
                }

                server -> args(args);
            }
            else if(strcmp(argv[3], "rundir") == 0)
            {
                CORBA::String_var dir = CORBA::string_dup(argv[4]);

                server -> directory(dir);
            }
            else if(strcmp(argv[3], "host") == 0)
            {
                CORBA::String_var host = CORBA::string_dup(argv[4]);

                server -> host(host);
            }
            else if(strcmp(argv[3], "update-timeout") == 0)
            {
                CORBA::Long updateTimeout = atol(argv[4]);

                server -> updateTimeout(updateTimeout);
            }
            else if(strcmp(argv[3], "failure-timeout") == 0)
            {
                CORBA::Long failureTimeout = atol(argv[4]);

                server -> failureTimeout(failureTimeout);
            }
            else if(strcmp(argv[3], "max-spawns") == 0)
            {
                CORBA::Long maxForks = atol(argv[4]);

                server -> maxForks((CORBA::Short)maxForks);
            }
	    else if(strcmp(argv[3], "mode") == 0)
	    {
	        IMR::ServerActivationMode mode;

	        if(strcmp(argv[4], "shared") == 0)
		    mode = IMR::SHARED;
	        else if(strcmp(argv[4], "persistent") == 0)
		    mode = IMR::PERSISTENT;
	        else
	        {
	            OB_COUT << "Unknown mode: " << argv[4] <<
			OB_ENDL;
	            OB_COUT << "Valid modes: shared persistent per_client"
		         << OB_ENDL;
	    	    return EXIT_FAILURE;
	        }

	        server -> mode(mode);
	    }
	    else if(strcmp(argv[3], "activate-poas") == 0)
	    {
	        CORBA::Boolean activate = 
		    strcmp(argv[4], "false") ? true : false;

	        server -> activatePOAs(activate);
	    }
            else
            {
                usage(progName);
                return EXIT_FAILURE;
            }
        }
        else
        {
            usage(progName);
            return EXIT_FAILURE;
        }
    }
    else if(opType == RESET)
    {
        if(objType == SERVER)
        {
            if(argc != 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }
            CORBA::String_var serverName = CORBA::string_dup(argv[2]);

            try
            {
                IMR::Server_var server = factory -> get_server(serverName);
                assert(!CORBA::is_nil(server));
                server -> clear_error_state();
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: record doesn't exist" << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else
        {
            usage(progName);
            return EXIT_FAILURE;
        }
    }
    else if(opType == START)
    {
        if(objType == SERVER)
        {
            if(argc != 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }
            CORBA::String_var serverName = CORBA::string_dup(argv[2]);

            try
            {
                IMR::Server_var server = factory -> get_server(serverName);
                server -> start();
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: record doesn't exist" << OB_ENDL;
                return EXIT_FAILURE;
            }
            catch(const IMR::ServerRunning&)
            {
                OB_COUT << "Failed: server already active" << OB_ENDL;
                return EXIT_FAILURE;
            }
	}
        else
        {
            usage(progName);
            return EXIT_FAILURE;
        }
    }
    else if(opType == STOP)
    {
        if(objType == SERVER)
        {
            if(argc != 3)
            {
                usage(progName);
                return EXIT_FAILURE;
            }
            CORBA::String_var serverName = CORBA::string_dup(argv[2]);

            try
            {
                IMR::Server_var server = factory -> get_server(serverName);
                server -> stop();
            }
            catch(const IMR::NoSuchServer&)
            {
                OB_COUT << "Failed: record doesn't exist" << OB_ENDL;
                return EXIT_FAILURE;
            }
            catch(const IMR::ServerNotRunning&)
            {
                OB_COUT << "Failed: server not active" << OB_ENDL;
                return EXIT_FAILURE;
            }
            catch(const IMR::OADNotRunning&)
            {
                OB_COUT << "Failed: OAD is down" << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else
        {
            usage(progName);
            return EXIT_FAILURE;
        }
    }
    else
    {
        usage(progName);
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

#define ARRAYSIZE 19

const char* options[ARRAYSIZE] = 
{ 
    "--add-oad",
    "--add-server",
    "--add-poa",
    "--remove-oad",
    "--remove-server",
    "--remove-poa",
    "--get-oad-status",
    "--get-server-info",
    "--get-poa-status",
    "--list-oads",
    "--list-servers",
    "--list-poas",
    "--tree",
    "--tree-oad",
    "--tree-server",
    "--set-server",
    "--reset-server",
    "--start-server",
    "--stop-server"
};

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	if(argc == 2)
	{
            if(strcmp(argv[1], "--help") == 0 || strcmp(argv[1], "-h") == 0)
            {
                usage(argv[0]);
                return EXIT_SUCCESS;
	    }
	}

	//
	// Remove all the admin arguments so they will not get stripped
	// off by the ORB even if in ORB format. 
	//
	char** adminOpts = new char*[argc];
	int optLen = 1;
	adminOpts[0] = argv[0];
	for(int i = 1 ; i < argc ; ++i)
	{
	    unsigned int j;
	    for(j = 0 ; j < ARRAYSIZE ; ++j)
	    {
	        if(strcmp(argv[i], options[j]) == 0)
		{
		    optLen = argc - i + 1;
		    for(int k = 1; k < optLen; ++k)
		    {
		        adminOpts[k] = argv[i + k - 1];
		    }
		    argc = i;
		    break;
		}
	    }
	    if(j < ARRAYSIZE)
	        break;
	}

	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);
	if(argc != 1)
	{
	    OB_CERR << "Unknown option: " << argv[1] << OB_ENDL;
	    usage(argv[0]);
	    return EXIT_FAILURE;
	}

	orb = CORBA::ORB_init(argc, argv);

	status = run(orb, optLen, adminOpts);
    }
    catch(const CORBA::TRANSIENT&)
    {
        OB_CERR << "Could not contact IMR" << OB_ENDL;
	status = EXIT_FAILURE;
    }
    catch(const CORBA::NO_PERMISSION&)
    {
        OB_CERR << "IMR is not running in administrative mode" << OB_ENDL;
	status = EXIT_FAILURE;
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
