// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/Codec.h>
#include <OB/Net.h>
#include <OB/IMR.h>

#include <ServerDatabase.h>
#include <OADDatabase.h>
#include <ProcessDatabase.h>
#include <IMRUtil.h>

#include <OSLayer.h>
#include <DatabaseUpgrade.h>

void
Usage(const char* progName)
{
    OB_CERR << "Usage: " << progName << " <Database Path>\n";
}

// ---------------------------------------------------------------------
// Method to read any from file (for version 4.0, 4.0.1)
// ---------------------------------------------------------------------

bool
ReadAny(CORBA::ORB_ptr orb, const char* db, CORBA::Any_out any)
{
    IOP::Codec_var codec = IMRUtil::codec(orb);
    assert(!CORBA::is_nil(codec));

    struct stat sbuf;
    if(!OSLayer::stat(db, sbuf))
        return false;

    //
    // Read it.
    //
    CORBA::OctetSeq data(sbuf.st_size);
    data.length(sbuf.st_size);

    const char* mode;
#ifdef WIN32
    mode = "rb";
#else
    mode = "r";
#endif

    FILE* fp = fopen(db, mode);
    if(fp == 0)
        throw CORBA::PERSIST_STORE();

    int n = fread(data.get_buffer(), sbuf.st_size, 1, fp);
    fclose(fp);

    if(n != 1)
        throw CORBA::PERSIST_STORE();

    try
    {
        any = codec -> decode(data);
    }
    catch(const CORBA::MARSHAL&)
    {
        return false;
    }

    return true;
}

// ---------------------------------------------------------------------
// Methods to convert old record formats to new
// ---------------------------------------------------------------------

IMRDB_4_1_0::ServerStatusRecord
ConvertIMR_4_0_2(const IMRDB_4_0_2::ServerStatusRecord* rec)
{
    IMRDB_4_1_0::ServerStatusRecord newReq;

    newReq.id             = rec -> id;
    newReq.server         = rec -> server;
    newReq.host           = rec -> host;
    newReq.exec           = rec -> exec;
    newReq.args           = rec -> args;
    newReq.mode           = rec -> mode;
    newReq.updateTimeout  = rec -> updateTimeout;
    newReq.failureTimeout = rec -> failureTimeout;
    newReq.maxForks       = rec -> maxForks;
    newReq.instance       = rec -> instance;
    newReq.started        = rec -> started;
    newReq.stopped	  = rec -> stopped;
    newReq.status         = rec -> status;
    newReq.updateTime     = rec -> updateTime;
    newReq.timesForked    = rec -> timesForked;
    newReq.activatePOAs   = rec -> activatePOAs;
    newReq.dir		  = rec -> dir;
    newReq.rootORT	  = 0;

    newReq.poas.length(rec -> poas.length());
    for(CORBA::ULong j = 0; j < newReq.poas.length(); ++j)
    {
        newReq.poas[j].poa   = rec -> poas[j].poa;
        newReq.poas[j].ort   = 0;
        newReq.poas[j].state = rec -> poas[j].state;
    }

    return newReq;
}

IMRDB_4_1_0::ServerStatusRecord
ConvertIMR_4_0(const IMRDB::ServerStatusRecord& rec)
{
    IMRDB_4_1_0::ServerStatusRecord newReq;

    newReq.id             = rec.id;
    newReq.server         = rec.server;
    newReq.host           = rec.host;
    newReq.exec           = rec.exec;
    newReq.args           = rec.args;
    newReq.mode           = rec.mode;
    newReq.updateTimeout  = rec.updateTimeout;
    newReq.failureTimeout = rec.failureTimeout;
    newReq.maxForks       = rec.maxForks;
    newReq.instance       = rec.instance;
    newReq.started        = rec.manual;
    newReq.status         = rec.status;
    newReq.updateTime     = rec.updateTime;
    newReq.timesForked    = rec.timesForked;
    newReq.activatePOAs   = rec.activatePOAs;
    newReq.rootORT	  = 0;

    newReq.poas.length(rec.poas.length());
    for(CORBA::ULong j = 0; j < newReq.poas.length(); ++j)
    {
        newReq.poas[j].poa   = rec.poas[j].poa;
        newReq.poas[j].ort   = 0;
        newReq.poas[j].state = rec.poas[j].state;
    }

    newReq.dir     = CORBA::string_dup("");
    newReq.stopped = false;

    return newReq;
}

IMRDB_4_0_2::OADRecord
ConvertOAD_4_0(const IMRDB::OADRecord& rec)
{
    IMRDB_4_0_2::OADRecord newReq;

    newReq.host       = rec.host;
    newReq.controller = rec.controller;
    newReq.manager    = rec.manager;
    newReq.up         = rec.up;

    newReq.id = 0;

    return newReq;
}

OADDB_4_0_2::ServerRecord
ConvertOADHot_4_0(const OADDB::ServerRecord& rec)
{
    OADDB_4_0_2::ServerRecord newReq;

    newReq.instance = rec.id;
    newReq.server   = rec.server;
    newReq.pid      = rec.pid;
    newReq.endpoint = rec.endpoint;

    newReq.id = 0;

    return newReq;
}

// ---------------------------------------------------------------------
// Methods to determine database format
// ---------------------------------------------------------------------

bool
ReadIMRDB(CORBA::ORB_ptr orb,
          const char* IMRdb,
	  bool& upToDate,
          IMRDB_4_1_0::ServerStatusSeq& seq)
{
    //
    // Check if database is already up to date
    //
    try
    {
        IMR::ServerDatabase* db = new IMR::ServerDatabase(orb);
	delete db;
	upToDate = true;
	return true;
    }
    catch(const CORBA::Exception&)
    {
    }

    //
    // Is the database 4.0, 4.0.1 (old style database)
    //
    CORBA::Any_var any;
    if(ReadAny(orb, IMRdb, any))
    {
        const IMRDB::ServerStatusSeq* dbrec;
	if(any >>= dbrec)
	{
	    seq.length(dbrec -> length());
	    for(CORBA::ULong i = 0 ; i < dbrec -> length() ; ++i)
	        seq[i] = ConvertIMR_4_0((*dbrec)[i]);
	    return true;
	}
    }

    //
    // Is it a 4.0.2 database
    //
    try
    {
        OB::Database* db =
	    new OB::Database(IMRUtil::codec(orb), IMRdb, true,
			     IMRDB_4_0_2::_tc_ServerStatusRecord, 512);

        OB_STL::list<CORBA::ULong> ids;
        db -> getAll(ids);

        OB_STL::list<CORBA::ULong>::iterator iter;
        for(iter = ids.begin() ; iter != ids.end() ; ++iter)
        {
            CORBA::Any_var any;
            const IMRDB_4_0_2::ServerStatusRecord* dbrec;

            if(!db -> get(*iter, any) || !(any >>= dbrec))
	        return false;

	    seq.length(seq.length() + 1);
	    seq[seq.length() - 1] = ConvertIMR_4_0_2(dbrec);
        }

	delete db;
        return true;
    }
    catch(const CORBA::Exception&)
    {
    }

    return false;
}

bool
ReadOADDB(CORBA::ORB_ptr orb,
          const char* OADdb,
	  bool& upToDate,
          IMRDB_4_0_2::OADRecordSeq& seq)
{
    //
    // Check if database is already up to date
    //
    try
    {
        IMR::OADDatabase* db = new IMR::OADDatabase(orb);
	delete db;
	upToDate = true;
	return true;
    }
    catch(const CORBA::Exception&)
    {
    }

    //
    // Is the database 4.0, 4.0.1 (old style database)
    //
    CORBA::Any_var any;
    if(ReadAny(orb, OADdb, any))
    {
        const IMRDB::OADRecordSeq* dbrec;
	if(any >>= dbrec)
	{
	    seq.length(dbrec -> length());
	    for(CORBA::ULong i = 0 ; i < dbrec -> length() ; ++i)
	        seq[i] = ConvertOAD_4_0((*dbrec)[i]);
	    return true;
	}
    }

    return false;
}

bool
ReadOADHotDB(CORBA::ORB_ptr orb,
             const char* OADHotdb,
	     bool& upToDate,
             OADDB_4_0_2::ServerRecordSeq& seq)
{
    //
    // Check if database is already up to date
    //
    try
    {
        OAD::ProcessDatabase* db = new OAD::ProcessDatabase(orb);
	delete db;
	upToDate = true;
	return true;
    }
    catch(const CORBA::Exception&)
    {
    }

    //
    // Is the database 4.0, 4.0.1 (old style database)
    //
    CORBA::Any_var any;
    if(ReadAny(orb, OADHotdb, any))
    {
        const OADDB::ServerRecordSeq* dbrec;
	if(any >>= dbrec)
	{
	    seq.length(dbrec -> length());
	    for(CORBA::ULong i = 0 ; i < dbrec -> length() ; ++i)
	        seq[i] = ConvertOADHot_4_0((*dbrec)[i]);
	    return true;
	}
    }

    return false;
}

// ---------------------------------------------------------------------

int
run(CORBA::ORB_ptr orb, const char* database)
{
    //
    // Make sure database is a valid directory
    //
    struct stat sbuf;
    if(!OSLayer::stat(database, sbuf) || !OSLayer::isDir(sbuf))
    {
        OB_CERR << "Directory does not exist: " << database << OB_ENDL;
        return EXIT_FAILURE;
    }

    OB::DatabaseUpgrade upgrade;

    // 
    // Upgrade imr.db
    //
    CORBA::String_var IMRdb = CORBA::string_dup(database);
    IMRdb += "/imr.db";
        
    CORBA::String_var IMRdbBak = CORBA::string_dup(IMRdb);
    IMRdbBak += ".bak";

    if(!OSLayer::stat(IMRdb, sbuf))
    {
        OB_CERR << "No IMR database found in " << database << OB_ENDL;
    }
    else
    {
        if(upgrade.upgrade(IMRdb, IMRdbBak))
	    IMRdbBak += "~";

        bool upToDate = false;
        IMRDB_4_1_0::ServerStatusSeq imrSeq;
        if(!ReadIMRDB(orb, IMRdb, upToDate, imrSeq))
	{
            OB_CERR << "Unknown database format in " << IMRdb << OB_ENDL;
            return EXIT_FAILURE;
	}

	if(!upToDate)
        {
	    if(!OSLayer::rename(IMRdb, IMRdbBak))
	    {
	        OB_CERR << "Could not back up database" << OB_ENDL;
		return EXIT_FAILURE;
	    }

	    OB::Database* db = 
	        new OB::Database(IMRUtil::codec(orb), IMRdb, true,
		                 IMRDB_4_1_0::_tc_ServerStatusRecord, 512);

	    for(CORBA::ULong i = 0 ; i < imrSeq.length() ; ++i)
	    {
	        CORBA::Any any;
		any <<= imrSeq[i];
		db -> add(any);
	    }
	    delete db;
        }
    }

    // 
    // Upgrade oad.db
    //
    CORBA::String_var OADdb = CORBA::string_dup(database);
    OADdb += "/oad.db";
        
    CORBA::String_var OADdbBak = CORBA::string_dup(OADdb);
    OADdbBak += ".bak";

    if(!OSLayer::stat(OADdb, sbuf))
    {
        OB_CERR << "No OAD database found in " << database << OB_ENDL;
    }
    else
    {
        if(upgrade.upgrade(OADdb, OADdbBak))
	    OADdbBak += "~";

        bool upToDate = false;
        IMRDB_4_0_2::OADRecordSeq oadSeq;
        if(!ReadOADDB(orb, OADdb, upToDate, oadSeq))
	{
            OB_CERR << "Unknown database format in " << OADdb << OB_ENDL;
            return EXIT_FAILURE;
	}

	if(!upToDate)
        {
	    if(!OSLayer::rename(OADdb, OADdbBak))
	    {
	        OB_CERR << "Could not back up database" << OB_ENDL;
		return EXIT_FAILURE;
	    }

	    OB::Database* db = 
	        new OB::Database(IMRUtil::codec(orb), OADdb, true,
		                 IMRDB_4_0_2::_tc_OADRecord, 512);

	    for(CORBA::ULong i = 0 ; i < oadSeq.length() ; ++i)
	    {
	        CORBA::Any any;
		any <<= oadSeq[i];
		db -> add(any);
	    }
	    delete db;
        }
    }

    // 
    // Upgrade oad-hot.db
    //
    CORBA::String_var OADHotdb = CORBA::string_dup(database);
    OADHotdb += "/oad-hot.db";

    CORBA::String_var OADHotdbBak = CORBA::string_dup(OADHotdb);
    OADHotdbBak += ".bak";

    if(!OSLayer::stat(OADHotdb, sbuf))
    {
        OB_CERR << "No OAD Hot database found in " << database << OB_ENDL;
    }
    else
    {
        if(upgrade.upgrade(OADHotdb, OADHotdbBak))
	    OADHotdbBak += "~";

        bool upToDate = false;
        OADDB_4_0_2::ServerRecordSeq oadSeq;
        if(!ReadOADHotDB(orb, OADHotdb, upToDate, oadSeq))
	{
            OB_CERR << "Unknown database format in " << OADHotdb << OB_ENDL;
            return EXIT_FAILURE;
	}

	if(!upToDate)
        {
	    if(!OSLayer::rename(OADHotdb, OADHotdbBak))
	    {
	        OB_CERR << "Could not back up database" << OB_ENDL;
		return EXIT_FAILURE;
	    }

	    OB::Database* db = 
	        new OB::Database(IMRUtil::codec(orb), OADHotdb, true,
		                 OADDB_4_0_2::_tc_ServerRecord, 512);

	    for(CORBA::ULong i = 0 ; i < oadSeq.length() ; ++i)
	    {
	        CORBA::Any any;
		any <<= oadSeq[i];
		db -> add(any);
	    }
	    delete db;
        }
    }

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[])
{
    const char* database = 0;
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    if(argc != 2 || strcmp(argv[1], "--help") == 0 ||
       strcmp(argv[1], "-h") == 0)
    {
        Usage(argv[0]);
	return EXIT_SUCCESS;
    }
    database = argv[1];

    try
    {
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);
	properties -> setProperty("ooc.imr.dbdir", database);

        orb = CORBA::ORB_init(argc, argv);
	status = run(orb, database);
    }
    catch(const CORBA::Exception& ex)
    {
        OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
        try
        {
            orb -> destroy();
        }
        catch(const CORBA::Exception& ex)
        {
            OB_CERR << ex << OB_ENDL;
            status = EXIT_FAILURE;
        }
    }

    return status;
}
