// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/OBPOAManager.h>
#include <OB/RefCounted.h>
#include <OB/BootManager.h>

#include <ServerObserver_impl.h>
#include <InternalDomain_impl.h>
#include <OADEndpointManager_impl.h>
#include <ServerDatabase.h>
#include <OADDatabase.h>
#include <LocationForwarder.h>
#include <ServerStarter.h>

#include <IMRLogger.h>
#include <IMRServer.h>

#include <stdlib.h>

using namespace IMR;

// ----------------------------------------------------------------------
// IMRServer constructor/destructor
// ----------------------------------------------------------------------

IMRServer::IMRServer(CORBA::ORB_ptr orb,
		     IMROAD::Logger_ptr logger,
		     OB::Properties_ptr properties,
		     const char* fwdPort,
		     bool administrative)
{
    //
    // Resolve Root POA.
    //
    CORBA::Object_var poaObj = orb -> resolve_initial_references("RootPOA");
    rootPOA_ = PortableServer::POA::_narrow(poaObj);
    PortableServer::POAManager_var manager = rootPOA_ -> the_POAManager();

    //
    // Create the forward POA
    //
    CORBA::PolicyList pl(5);
    pl.length(5);
    pl[0] = rootPOA_ -> create_lifespan_policy(PortableServer::PERSISTENT);
    pl[1] = rootPOA_ -> create_id_assignment_policy(PortableServer::USER_ID);
    pl[2] = rootPOA_ -> create_request_processing_policy(
	PortableServer::USE_SERVANT_MANAGER);
    pl[3] = rootPOA_ -> create_servant_retention_policy(
	PortableServer::NON_RETAIN);
    pl[4] = rootPOA_ -> create_implicit_activation_policy(
	PortableServer::NO_IMPLICIT_ACTIVATION);

    properties -> setProperty("ooc.orb.poamanager.forward.endpoint", fwdPort);

    CORBA::Object_var facObj =
        orb -> resolve_initial_references("POAManagerFactory");
    OBPortableServer::POAManagerFactory_var poaMgrFactory =
        OBPortableServer::POAManagerFactory::_narrow(facObj);
    fwdManager_ = poaMgrFactory -> create_poa_manager("forward");

    fwdPOA_ = rootPOA_ -> create_POA("forward", fwdManager_, pl);

    //
    // Create a persistent POA for the administrative interfaces.
    //
    pl.length(2);
    pl[0] = rootPOA_ -> create_id_assignment_policy(PortableServer::USER_ID);
    pl[1] = rootPOA_ -> create_lifespan_policy(PortableServer::PERSISTENT);

    adminPOA_ = rootPOA_ -> create_POA("IMRServerPOA", manager, pl);

    assert(!CORBA::is_nil(adminPOA_));

    //
    // Resolve the Boot Manager
    //
    CORBA::Object_var bmgrObj =
        orb -> resolve_initial_references("BootManager");
    OB::BootManager_var bootManager = OB::BootManager::_narrow(bmgrObj);

    //
    // Create the database objects.
    //
    serverdb_ = new ServerDatabase(orb);
    oaddb_ = new OADDatabase(orb);

    //
    // Create the server starter thread
    //
    ServerStarter* serverStarter = new IMR::ServerStarter();
    serverStarter_ = serverStarter;

    //
    // Create the OADEndpointManager. This is used for OADs to
    // register their endpoint with the IMR at OAD initialization.
    //
    endpointManager_ = 
        new IMR::OADEndpointManager_impl(orb, adminPOA_, logger, oaddb_,
	                                 serverdb_, serverStarter);
    PortableServer::ObjectId_var registrarOID =
	PortableServer::string_to_ObjectId("OADEndpointManager");
    adminPOA_ -> activate_object_with_id(registrarOID, endpointManager_);
    CORBA::Object_var registrarObj = adminPOA_ -> id_to_reference(registrarOID);
    //bootManager -> add_binding(registrarOID, registrarObj);
    
    //
    // Create the IMR::Domain implementation object
    //
    domainImpl_ = new IMR::InternalDomain_impl(orb, adminPOA_, fwdPOA_, oaddb_, 
    				               serverdb_, endpointManager_,
					       logger, administrative, 
					       serverStarter);


    //
    // Add the bootstrap reference as both "Domain" and "IMR"
    //
    PortableServer::ObjectId_var domainOID =
        PortableServer::string_to_ObjectId("IMR");
    adminPOA_ -> activate_object_with_id(domainOID, domainImpl_);
    CORBA::Object_var domainObj = adminPOA_ -> id_to_reference(domainOID);
    bootManager -> add_binding(domainOID, domainObj);

    //
    // Register IMR initial reference for any collocated OADs
    //
    try
    {
        CORBA::Object_var obj = orb -> resolve_initial_references("IMR");
    }
    catch(CORBA::ORB::InvalidName&)
    {
        orb -> register_initial_reference("IMR", domainObj);
    }

    IMR::ActiveStateFactory* factory =
        domainImpl_ -> get_active_state_factory();

    //
    // Create the ServerObserver. This is used to notify the IMR that a
    // server died.
    //
    observerImpl_ = new IMR::ServerObserver_impl(adminPOA_, factory);
    PortableServer::ObjectId_var callbackOID =
	PortableServer::string_to_ObjectId("ServerObserver");
    adminPOA_ -> activate_object_with_id(callbackOID, observerImpl_);

    IMR::ServerObserver_var observer = observerImpl_ -> _this();

    endpointManager_ -> _OB_setNotifier(observer);

    //
    // Create the servant manager for the fwdPOA_.
    //
    forwarder_ = new  LocationForwarder(orb, logger, factory);
    PortableServer::ServantLocator_var forwarder = forwarder_ -> _this();
    fwdPOA_ -> set_servant_manager(forwarder);
    
    forwarderId_ = rootPOA_ -> servant_to_id(forwarder_);

    //
    // Activate the POA manager for the forward POA
    //
    fwdManager_ -> activate();
}

IMRServer::~IMRServer()
{
    serverStarter_ -> terminate();

    delete forwarder_;

    delete endpointManager_;
    delete observerImpl_;
    delete domainImpl_;

    delete serverdb_;
    delete oaddb_;
}

void
IMRServer::connectOADs(bool collocated)
{
    endpointManager_ -> _OB_connectOADs(collocated);
}

void
IMRServer::startServerStarter()
{
    serverStarter_ -> start();
}
