// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/RefCounted.h>
#include <OB/seq_converter.h>
#include <OB/IMRORT_impl.h>

#include <InternalServer_impl.h>
#include <ServerStarter.h>
#include <ServerDatabase.h>
#include <OADDatabase.h>

#include <stdlib.h>
#include <ctype.h>

#include <IMRLogger.h>
#include <OSLayer.h>
#include <IMRUtil.h>

using namespace IMR;

//
// TODO: have something other than a global database monitor for the
// servers.
//

// ----------------------------------------------------------------------
// InternalServer_impl private member implementation
// ----------------------------------------------------------------------

ServerRecord_ptr
InternalServer_impl::getRecord()
{
    PortableServer::ObjectId_var oid = current_ -> get_object_id();
    CORBA::String_var sid = PortableServer::ObjectId_to_string(oid);
    if(oid -> length() == 0 || !isdigit(sid[0]))
	throw CORBA::INV_IDENT(); // TODO: correct?

    CORBA::Long id = ::atol(sid.in());
    ServerRecord_ptr rec = db_ -> queryRecordById(id);
    if(CORBA::is_nil(rec))
	throw CORBA::OBJECT_NOT_EXIST();
     else
        return rec;
}

void
InternalServer_impl::forkFailed()
{
    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().status = STOPPED;
    rec -> record().updateTime = OSLayer::time();

    db_ -> updateRecord(rec);
}

// ----------------------------------------------------------------------
// InternalServer_impl constructor and destructor
// ----------------------------------------------------------------------

InternalServer_impl::InternalServer_impl(ServerDatabase* db,
					 OADDatabase* oaddb,
					 CORBA::ORB_ptr orb,
					 PortableServer::POA_ptr poa,
					 IMROAD::Logger_ptr logger,
					 ServerStarter* serverStarter)
    : db_(db),
      oaddb_(oaddb),
      logger_(IMROAD::Logger::_duplicate(logger)),
      count_(0),
      serverStarter_(serverStarter)
{
    try
    {
	CORBA::Object_var obj =
	    orb -> resolve_initial_references("POACurrent");
	current_ = PortableServer::Current::_narrow(obj);
    }
    catch(const CORBA::ORB::InvalidName&)
    {
    }
    assert(!CORBA::is_nil(current_));

    //
    // This is an interesting hack. We need an IOR that `points' to
    // the ServantLocator attached to the LocationForward POA. We
    // create a IOR that is then completed by the calling process.
    //
    PortableServer::ObjectId_var oid =
	PortableServer::string_to_ObjectId("dummy");
    CORBA::Object_var obj =
	poa -> create_reference_with_id(oid, "IDL:dummy:1.0");
    
    //
    // Grab the IOR from the object reference.
    //
    OBPortableServer::POA_var obpoa = OBPortableServer::POA::_narrow(poa);
    fwdORT_ = obpoa -> adapter_template();
}

InternalServer_impl::~InternalServer_impl()
{
}

// ----------------------------------------------------------------------
// InternalServer_impl public member implementation
// ----------------------------------------------------------------------

//
// IDL:ooc.com/IMR/ActiveState/startup:1.0
//
void
InternalServer_impl::startup(
    const char* instance,
    PortableInterceptor::ObjectReferenceTemplate* rootORT,
    OAD::ProcessEndpointManager_out manager)
    throw(IMR::NoSuchOAD,
          IMR::OADNotRunning,
	  CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    {
        JTCSynchronized sync(*rec.in());

        IMRDB_4_0_2::OADRecord oadrec;
        if(!oaddb_ -> queryRecordByHost(rec -> record().host, oadrec, true))
	    throw NoSuchOAD(rec -> record().host);

        //
        // Is this OAD currently active?
        //
        // TODO: doesn't this mean that servers cannot be started by hand?
        //
        if(CORBA::is_nil(oadrec.manager))
	    throw OADNotRunning();
    
        manager = OAD::ProcessEndpointManager::_duplicate(oadrec.manager);

        //
        // If the OAD started the server then our state would be
        // FORKED. If the state is STOPPED then this was a manually
        // started server and we need to copy the process-id.
        //
        if(rec -> record().status == STOPPED)
        {
	    rec -> record().started = true;
	    rec -> record().instance = CORBA::string_dup(instance);
        }
        else if(rec -> record().status != FORKED)
	    throw CORBA::BAD_PARAM();
    
        //
        // If this is not the expected process ID then we have a problem.
        //
        if(strcmp(rec -> record().instance, instance) != 0)
	    throw CORBA::BAD_PARAM();

        //
        //
        // Update status, time and rootIOR field
        //
        rec -> record().status = STARTING;
        rec -> record().updateTime = OSLayer::time();
	CORBA::add_ref(rootORT);
        rec -> record().rootORT = rootORT;

        //
        // Run through the POAs and invalidate all IORs
        //
        for(CORBA::ULong i = 0; i < rec -> record().poas.length(); i++)
	    rec -> record().poas[i].state = IMR::NON_EXISTENT;
    
        if(logger_ -> traceServerStatus() >= 2)
        {
	    CORBA::String_var msg = CORBA::string_dup("server_starting: ");
	    msg += rec -> record().server;
	    logger_ -> trace("IMR", msg);
        }

        db_ -> updateRecord(rec);
    }

    JTCSynchronized sync2(*db_);
    db_ -> notifyAll();
}

//
// IDL:ooc.com/IMR/ActiveState/set_status:1.0
//
void
InternalServer_impl::set_status(const char* instance, IMR::ServerStatus status)
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    {
        JTCSynchronized sync(*rec.in());

        //
        // If the instance IDs do not match we are not interested
        //
        if(strcmp(instance, rec -> record().instance) != 0)
	    throw CORBA::BAD_PARAM();

        //
        // We need to validate the server status field. We only allow
        // "valid" state transitions.
        //
        switch(status)
        {
        case STOPPED:
	    //
	    // Reset the Instance Id.
	    //
	    rec -> record().instance = CORBA::string_dup("");
	    break;

        case FORKED:
        case STARTING:
	    //
	    // We shouldn't get this state transition from the server
	    // itself.
	    //
	    throw CORBA::BAD_PARAM();

        case RUNNING:
	    if(rec -> record().status != STARTING)
	        throw CORBA::BAD_PARAM();
	    break;

        case STOPPING:
	    if(rec -> record().status != RUNNING)
	        throw CORBA::BAD_PARAM();
	    break;

        default:
	    throw CORBA::BAD_PARAM();
        }

        //
        // Update status and updateTime field.
        //
        rec -> record().status = status;
        rec -> record().updateTime = OSLayer::time();

        //
        // If we're RUNNING reset the times forked.
        //
        if(status == RUNNING)
	    rec -> record().timesForked = 0;
        else if(status == STOPPED)
        {
	    //
	    // Depending on the status update run through the POAs and
	    // invalidate all IORs.
	    //
	    for(CORBA::ULong poa = 0 ; poa < rec -> record().poas.length() ;
	        poa++)
	        rec -> record().poas[poa].state = IMR::NON_EXISTENT;
        }

        if(logger_ -> traceServerStatus() >= 2)
        {
	    CORBA::String_var msg = CORBA::string_dup("updating: ");
	    msg += rec -> record().server;
	    msg += " state: ";
	    msg += IMRUtil::serverStatusToStr(status);
	    logger_ -> trace("IMR", msg);
        }

        //
        // If this is a persistent server restart it unless it was
        // manually stoped via the console or imradmin
        //
        if(status == STOPPED && rec -> record().mode == IMR::PERSISTENT &&
           !rec -> record().stopped)
            serverStarter_ -> add(rec -> record().server);

        db_ -> updateRecord(rec);
    }

    JTCSynchronized sync2(*db_);
    db_ -> notifyAll();

}

//
// IDL:ooc.com/IMR/ActiveState/poa_create:1.0
//
PortableInterceptor::ObjectReferenceTemplate*
InternalServer_impl::poa_create(
    IMR::POAStatus state,
    PortableInterceptor::ObjectReferenceTemplate* ort)
    throw(IMR::NoSuchPOA,
	  CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    PortableInterceptor::AdapterName_var poaname = ort -> adapter_name();
    CORBA::String_var serverId = ort -> server_id();

    if(logger_ -> traceServerStatus() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("poa_create: ");
	msg += rec -> record().server;
	msg += " ";
	for(CORBA::ULong poa = 0; poa < poaname -> length(); poa++)
	{
	    msg += poaname[poa];
	    if(poa != poaname -> length() -1)
		msg += "/";
	}
	msg += ": ";
	msg += IMRUtil::poaStatusToStr(state);
	logger_ -> trace("IMR", msg);
    }

    //
    // Find the POA record.
    //
    CORBA::ULong poa;
    for(poa = 0; poa < rec -> record().poas.length(); poa++)
	if(IMRUtil::poacmp(poaname, rec -> record().poas[poa].poa, true))
	    break;

    if(poa >= rec -> record().poas.length())
    {
    	//
	// No POA record exists. Implicitly add the POA if the
	// server activatePOAs attribute is set.
	//
	if(rec -> record().activatePOAs)
	{
	    IMRDB_4_1_0::POAStatusRecord poaRec;

	    poaRec.poa.length(poaname -> length() + 1);

	    poaRec.poa[0] = serverId;
	    for(CORBA::ULong i = 0; i < poaname -> length(); i++)
	        poaRec.poa[i + 1] = poaname[i];

	    poa = rec -> record().poas.length();
	    rec -> record().poas.length(poa + 1);
	    rec -> record().poas[poa] = poaRec;
	}
	else
	    throw NoSuchPOA(poa);
    }

    //
    // Update the record content.
    //
    CORBA::add_ref(ort);
    rec -> record().poas[poa].ort = ort;
    rec -> record().poas[poa].state = state;

    //
    // Update and updateRecord the record.
    //
    db_ -> updateRecord(rec);

    //
    // Create IMR Object Reference Template and return it.
    //
    OBPortableInterceptor::IMRORT_impl* imrORT =
        new OBPortableInterceptor::IMRORT_impl(serverId, poaname, fwdORT_);

    CORBA::add_ref(imrORT);
    return imrORT;
}

//
// IDL:ooc.com/IMR/ActiveState/poa_status_update:1.0
//
void
InternalServer_impl::poa_status_update(const IMR::POANameSeq& poas, 
	                               IMR::POAStatus state)
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    {
        JTCSynchronized sync(*rec.in());

        for(CORBA::ULong i = 0; i < poas.length(); ++i)
        {
            for(CORBA::ULong j = 0; j < rec -> record().poas.length(); ++j)
	        if(IMRUtil::poacmp(poas[i], rec -> record().poas[j].poa, true))
	        {
	            rec -> record().poas[j].state = state;
    		    if(logger_ -> traceServerStatus() >= 2)
    		    {
		        CORBA::String_var msg;
		        if(state == IMR::NON_EXISTENT)
                            msg = CORBA::string_dup("remove_poa: ");
			else
                            msg = CORBA::string_dup("poa_status: ");

                        msg += rec -> record().server;
                        msg += " ";

		        for(CORBA::ULong k = 0; k < poas[i].length(); ++k)
		        {
		            msg += poas[i][k];
			    if(k != poas[i].length() -1)
			        msg += '/';
		        }

			if(state != IMR::NON_EXISTENT)
			{
		            msg += ": ";
		            msg += IMRUtil::poaStatusToStr(state);
			}
                        logger_ -> trace("IMR", msg);
                    }
	        }
        }

        db_ -> updateRecord(rec);
    }

    //
    // Ensure that the state transition is propogated.
    //
    JTCSynchronized sync2(*db_);
    db_ -> notifyAll();
}

//
// IDL:ooc.com/IMR/InternalServer/kill:1.0
//
void
InternalServer_impl::kill()
    throw(NoSuchOAD, CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();

    if(logger_ -> traceProcessControl() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("kill: ");
	msg += rec -> record().server;
	logger_ -> trace("IMR", msg);
    }

    IMRDB_4_0_2::OADRecord oadrec;
    if(!oaddb_ -> queryRecordByHost(rec -> record().host, oadrec, true))
	throw NoSuchOAD(rec -> record().host);

    if(CORBA::is_nil(oadrec.manager) || !oadrec.up)
	throw CORBA::TRANSIENT();

    bool failed = false;
    CORBA::String_var reason;

    try
    {
	oadrec.controller -> kill_process(rec -> record().server);
    }
    catch(const OAD::KillFailed& ex)
    {
	failed = true;
	reason = CORBA::string_dup("KillFailed: ");
	reason += ex.reason;
    }
    catch(const CORBA::SystemException& ex)
    {
	//
	// ProcessController isn't reachable. The OAD state
	// record should be updated by our death detector.
	//
	failed = true;
	reason = ex._name();
    }

    if(logger_ -> traceProcessControl() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("kill: ");
	msg += rec -> record().server;
	msg += ": ";
	if(failed)
	    msg += reason;
	else
	    msg += "success";
	logger_ -> trace("IMR", msg);
    }
    if(failed)
	throw CORBA::TRANSIENT();
}

//
// IDL:ooc.com/IMR/InternalServer/fork:1.0
//
void
InternalServer_impl::fork()
    throw(NoSuchOAD, CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    IMRDB_4_0_2::OADRecord oadrec;
    CORBA::String_var instance;
    IMR::ArgList args;

    {
        JTCSynchronized sync(*rec.in());

	//
	// If the server state is not STOPPED do not try
	// to fork
	//
        if(rec -> record().status != STOPPED)
            return;

	//
	// Check if executable is empty. If so do not try to fork.
	//
	if(strlen(rec -> record().exec) == 0)
	{
	    if(logger_ -> traceProcessControl() >= 2)
	    {
	        CORBA::String_var msg = CORBA::string_dup("cannot fork ");
		msg += rec -> record().server;
		msg += ": no executable defined";
	        logger_ -> trace("IMR", msg);
	    }
	    throw CORBA::TRANSIENT();
	}

        //
        // Have we attempted to fork the server too many times?
        //
        if(rec -> record().timesForked >= rec -> record().maxForks)
        {
    	    if(logger_ -> traceProcessControl() >= 2)
    	    {
	        CORBA::String_var msg =
		    CORBA::string_dup("max forks exceeded: ");
	        msg += rec -> record().maxForks;
	        msg += " for server ";
	        msg += rec -> record().server;
	        logger_ -> trace("IMR", msg);
	    }
	    throw CORBA::TRANSIENT();
        }

        if(!oaddb_ -> queryRecordByHost(rec -> record().host, oadrec, true))
	    throw NoSuchOAD(rec -> record().host);

        if(CORBA::is_nil(oadrec.manager) || !oadrec.up)
	    throw CORBA::TRANSIENT();

        if(logger_ -> traceProcessControl() >= 2)
        {
	    CORBA::String_var msg = CORBA::string_dup("exec: ");
	    msg += rec -> record().server;
	    msg += " forked ";
	    msg += rec -> record().timesForked;
	    msg += " times in a row";
	    logger_ -> trace("IMR", msg);
        }

	//
	// Construct an argument list with the -ORBServerId command
	// line param
	//
	args.length(rec -> record().args.length() + 4);
	CORBA::ULong i = 0;

	CORBA::ULong j;
	for(j = 0; j < rec -> record().args.length() ; j++, i++)
	{
	    args[i] = rec -> record().args[j];
	}
	
	args[i++] = CORBA::string_dup("-ORBServerId");
	args[i++] = rec -> record().server;

	instance = CORBA::string_dup("");
	instance += OSLayer::time();
	instance += "-";
	instance += count_++;

	args[i++] = CORBA::string_dup("-ORBserver_instance");
	args[i++] = instance;

	rec -> record().status = FORKED;
	rec -> record().instance = instance;
	rec -> record().started = false;
	rec -> record().stopped = false;
	++rec -> record().timesForked;
	rec -> record().updateTime = OSLayer::time();
	db_ -> updateRecord(rec);
    }

    //
    // Did the exec fail? and if so the reason
    //
    bool failed = false;
    CORBA::String_var reason;

    try
    {
	oadrec.controller -> exec_process(rec -> record().server,
					  rec -> record().exec, args,
					  rec -> record().dir, instance);
    }
    catch(const OAD::ExecFailed& ex)
    {
        forkFailed();
	failed = true;
	reason = CORBA::string_dup("ExecFailed: ");
	reason += ex.reason;
    }
    catch(const CORBA::SystemException& ex)
    {
        forkFailed();
	failed = true;
	reason = ex._name();
    }

    if(logger_ -> traceProcessControl() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("exec: ");
	msg += rec -> record().server;
	msg += ": ";
	if(failed)
	    msg += reason;
	else
	    msg += "success";
	logger_ -> trace("IMR", msg);
    }

    if(failed)
	throw CORBA::TRANSIENT();
}

//
// IDL:ooc.com/IMR/InternalServer/ior_for:1.0
//
CORBA::Object*
InternalServer_impl::make_object(const POAName& poaname,
			         const PortableServer::ObjectId& oid)
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();

    //
    // Locate the ServerStatusRecord and POAStatusRecord.
    //

    CORBA::ULong poa;
    for(poa = 0; poa < rec -> record().poas.length(); poa++)
	if(IMRUtil::poacmp(poaname, rec -> record().poas[poa].poa, true))
	    break;
    if(poa >= rec -> record().poas.length())
	throw CORBA::TRANSIENT();

    //
    // Validate that the OAD for this record is actually running and
    // operational
    //
    // TODO: there is a race condition here (why?) I'm not sure what
    // to do about it
    //
    IMRDB_4_0_2::OADRecord oadrec;
    if(!oaddb_ -> queryRecordByHost(rec -> record().host, oadrec, true))
	throw CORBA::TRANSIENT();
    if(!oadrec.up)
	throw CORBA::TRANSIENT();

    //
    // Convert PS:ObjectID to PI::ObjectId. Note that data is not
    // being copied.
    //
    PortableInterceptor::ObjectId pioid(oid.length(), oid.length(),
    					(unsigned char *) oid.get_buffer(),
					false);

    //
    // If the POA is available return the IOR for the POA itself
    //
    if(rec -> record().poas[poa].state != IMR::NON_EXISTENT)
	return rec -> record().poas[poa].ort -> 
	    make_object("", pioid);

    //
    // Otherwise return the IOR for the root POA
    //
    OBPortableInterceptor::ObjectReferenceTemplate* ort = 
        OBPortableInterceptor::ObjectReferenceTemplate::_downcast(
	    rec -> record().rootORT);
    return ort -> make_object_for("", pioid, poaname);
}

//
// IDL:ooc.com/IMR/InternalServer/wait:1.0
//
void
InternalServer_impl::wait(IMR::ServerStatus status,
			  const IMR::POAName& poaname,
		          IMR::POAStatus poaState)
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    {
    JTCSynchronized sync(*rec.in());

    //
    // Return immediately if server status or poa state has changed
    //
    if(rec -> record().status != status)
        return;

    for(CORBA::ULong poa = 0 ; poa < rec -> record().poas.length() ; poa++)
    {
	if(IMRUtil::poacmp(poaname, rec -> record().poas[poa].poa))
	{
	    if(rec -> record().poas[poa].state != poaState)
	        return;
	    break;
	}
    }
    }

    try
    {
        JTCSynchronized sync2(*db_);
	db_ -> wait(rec -> record().updateTimeout);
    }
    catch(const JTCInterruptedException&)
    {
    }
}

//
// IDL:ooc.com/IMR/InternalServer/timeout:1.0
//
CORBA::Boolean
InternalServer_impl::timeout()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return OSLayer::time() - rec -> record().updateTime >
        rec -> record().failureTimeout;
}

//
// IDL:ooc.com/IMR/InternalServer/poa_status:1.0
//
void
InternalServer_impl::status(const IMR::POAName& poaname,
			    IMR::ServerStatus& status,
			    IMR::POAStatus& poaStatus)
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();

    status = rec -> record().status;

    //
    // Find the status record for this POA
    //
    CORBA::ULong poa;
    for(poa = 0; poa < rec -> record().poas.length(); poa++)
	if(IMRUtil::poacmp(poaname, rec -> record().poas[poa].poa, true))
	{
	    poaStatus =  rec -> record().poas[poa].state;
	    break;
	}

    if(poa == rec -> record().poas.length())
        throw CORBA::TRANSIENT();
}
