// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/Net.h>
#include <OB/seq_converter.h>

#include <OADDatabase.h>
#include <IMRUtil.h>
#include <OSLayer.h>

#ifndef HAVE_STRCASECMP
#   define strcasecmp stricmp
#endif

#ifndef HAVE_STRNCASECMP
#   define strncasecmp strnicmp
#endif


using namespace IMR;

// ----------------------------------------------------------------------
// OADDatabase private member implementation
// ----------------------------------------------------------------------

void
OADDatabase::load()
{
    OB_STL::list<CORBA::ULong> ids;
    db_ -> getAll(ids);

    OB_STL::list<CORBA::ULong>::iterator iter;
    for(iter = ids.begin() ; iter != ids.end() ; ++iter)
    {
        CORBA::Any_var any;
        if(!db_ -> get(*iter, any))
        {
            logger_ -> error("Error reading database");
            throw CORBA::PERSIST_STORE();
        }

        const IMRDB_4_0_2::OADRecord* dbrec;
        if(any >>= dbrec)
        {
            IMRDB_4_0_2::OADRecord rec = *dbrec;
            rec.id = *iter;
            oads_ -> put(rec.host, rec);
        }
        else
        {
            logger_ -> error("Bad record in database");
            throw CORBA::PERSIST_STORE();
        }
    }
}

// ----------------------------------------------------------------------
// OADDatabase constructor/destructor
// ----------------------------------------------------------------------

OADDatabase::OADDatabase(CORBA::ORB_ptr orb)
    : oads_(new HostTable(7))
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    logger_ = oborb -> logger();

    CORBA::String_var OADdb = CORBA::string_dup("");

    OB::Properties_var properties = oborb -> properties();
    const char* dbdir = properties -> getProperty("ooc.imr.dbdir");
    if(dbdir != 0)
    {
        struct stat sbuf;
        if(!OSLayer::stat(dbdir, sbuf))
        {
            if(!OSLayer::mkdir(dbdir, 0700))
            {
                CORBA::String_var msg = CORBA::string_dup("mkdir: ");
                msg += dbdir;
                msg += " failed";
                logger_ -> error(msg);
                throw CORBA::PERSIST_STORE();
            }
            OSLayer::stat(dbdir, sbuf);
        }
        if(!OSLayer::isDir(sbuf))
        {
            CORBA::String_var msg = CORBA::string_dup(dbdir);
            msg += ": not directory.";
            logger_ -> error(msg);
            throw CORBA::PERSIST_STORE();
        }

        OADdb += dbdir;
        OADdb += "/";
    }
        
    OADdb += "oad.db";

    db_ = new OB::Database(IMRUtil::codec(orb), OADdb, true,
    			   IMRDB_4_0_2::_tc_OADRecord, 512);

    try
    {
        load();
    }
    catch(const CORBA::Exception&)
    {
        delete oads_;
	delete db_;
    }
}

OADDatabase::~OADDatabase()
{
    delete oads_;
    delete db_;
}

// ----------------------------------------------------------------------
// OADDatabase public member implementation
// ----------------------------------------------------------------------

bool
OADDatabase::queryRecordByHost(const char* host,
			       IMRDB_4_0_2::OADRecord& rec,
			       bool matchPartial)
{
    JTCSynchronized sync(*this);

    if(oads_ -> get(host, rec))
        return true;

    if(matchPartial)
    {
        static char shortname[MAXHOSTNAMELEN + 1];
        int shortnamelen;
        bool partial = false;
        bool ambiguous = false;

        //
        // Get short name for passed host (host.domain -> host)
        //
        if(matchPartial)
        {
            const char* p = strchr(host,  '.');
            if(p)
            {
                shortnamelen = p - host;
                strncpy(shortname, host, shortnamelen);
	        shortname[shortnamelen] = '\0';
            }
            else
            {
                strcpy(shortname, host);
	        shortnamelen = strlen(shortname);
            }
        }

        IMRDB_4_0_2::OADRecord match;
	HostTable::Enumerator keys = oads_ -> keys();
        for(CORBA::ULong i = 0 ; i < keys.length() ; i++)
	    if(strncasecmp(keys[i], shortname, shortnamelen) == 0 &&
	       (keys[i][shortnamelen] == '.' || keys[i][shortnamelen] == '\0'))
	    {
	        // 
	        // Found partial match
	        //
	        if(partial)
		{
	            //
		    // Had previous partial match, thus ambiguous.
		    //
	            ambiguous = true;
		}
	        else
	        {
	            oads_ -> get(keys[i], match);
	            partial = true;
	        }
	    }

        if(partial && !ambiguous)
	{
            rec = match;
            return true;
        }
    }
    return false;
}

void
OADDatabase::addRecord(IMRDB_4_0_2::OADRecord& rec)
{
    JTCSynchronized sync(*this);

    if(oads_ -> containsKey(rec.host))
        assert(false);

    CORBA::Any any;
    any <<= rec;

    rec.id = db_ -> add(any);
    oads_ -> put(rec.host, rec);
}

void
OADDatabase::updateRecord(const IMRDB_4_0_2::OADRecord& rec)
{
    JTCSynchronized sync(*this);

    if(oads_ -> containsKey(rec.host))
    {
        CORBA::Any any;
	any <<= rec;
	db_ -> update(rec.id, any);

	oads_ -> remove(rec.host);
	oads_ -> put(rec.host, rec);
	return;
    }
    assert(false);
}

void
OADDatabase::removeRecord(const char* host)
{
    JTCSynchronized sync(*this);

    IMRDB_4_0_2::OADRecord rec;
    if(oads_ -> get(host, rec))
    {
        db_ -> remove(rec.id);
	oads_ -> remove(rec.host);
	return;
    }
    assert(false);
}

IMRDB_4_0_2::OADRecordSeq*
OADDatabase::getAllRecords()
{
    JTCSynchronized sync(*this);

    IMRDB_4_0_2::OADRecordSeq_var rc =
        new IMRDB_4_0_2::OADRecordSeq(oads_ -> size());
    rc -> length(oads_ -> size());

    IMRDB_4_0_2::OADRecord rec;
    HostTable::Enumerator keys = oads_ -> keys();
    for(CORBA::ULong i = 0; i < oads_ -> size(); i++)
    {
        oads_ -> get(keys[i], rec);
	rc[i] = rec;
    }
    return rc._retn();
}
