// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/Net.h>
#include <OB/Logger.h>
#include <OB/BootManager.h>

#include <IMRInternal.h>
#include <IMRLogger.h>
#include <OADServer.h>
#include <ProcessDatabase.h>
#include <ProcessController_impl.h>

using namespace OAD;

// ----------------------------------------------------------------------
// OADServer constructor/destructor
// ----------------------------------------------------------------------

OADServer::OADServer(CORBA::ORB_ptr orb,
		     IMROAD::Logger_ptr logger,
		     OB::Properties_ptr properties,
		     const char* slavePort,
		     bool slave)
{
    //
    // Resolve Root POA.
    //
    CORBA::Object_var obj = orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(obj);
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

    //
    // We use MAXHOSTNAMELEN + 1 since we're not sure whether the
    // len parameter to gethostname() includes '\0' or not.
    //
    char hostname[MAXHOSTNAMELEN + 1];
    if(gethostname(hostname, MAXHOSTNAMELEN) == -1)
    {
	CORBA::String_var msg = CORBA::string_dup("OAD: gethostname failed");
	logger -> error(msg);
	//TODO: exit -> exception
	exit(1);
    }

    try
    {
        obj = orb -> resolve_initial_references("IMR");
    }
    catch(...)
    {
	CORBA::String_var msg =
	    CORBA::string_dup("OAD: Could not get initial reference for IMR");
	logger -> error(msg);
	exit(1);
    }

    IMR::OADEndpointManager_var endpointMgr;
    try
    {
        IMR::InternalDomain_var domain = IMR::InternalDomain::_narrow(obj);
	endpointMgr = domain -> get_endpoint_manager();
    }
    catch(const CORBA::SystemException&)
    {
	CORBA::String_var msg =
	    CORBA::string_dup("OAD: Could not contact IMR");
	logger -> warning(msg);
    }

    db_ = new ProcessDatabase(orb);

    //
    // Create Persistent POA
    //
    CORBA::PolicyList pl;
    pl.length(2);
    pl[0] = rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
    pl[1] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);


    //
    // If we are running in slave mode the Root POA manager is 
    // already listening on the proper port. Otherwise we need to
    // create a new POA Manager on the slave port.
    //
    if(slave)
    {
        poa_ = rootPOA -> create_POA("OADServerPOA", manager, pl);
    }
    else
    {
        properties -> 
	    setProperty("ooc.orb.poamanager.slave.endpoint", slavePort);

        CORBA::Object_var facObj =
            orb -> resolve_initial_references("POAManagerFactory");
        OBPortableServer::POAManagerFactory_var poaMgrFactory =
            OBPortableServer::POAManagerFactory::_narrow(facObj);
        slaveManager_ = poaMgrFactory -> create_poa_manager("slave");

        poa_ = rootPOA -> create_POA("OADServerPOA", slaveManager_, pl);
    }

    assert(!CORBA::is_nil(poa_));

    //
    // Resolve the Boot Manager
    //
    CORBA::Object_var bmgrObj =
        orb -> resolve_initial_references("BootManager");
    OB::BootManager_var bootManager = OB::BootManager::_narrow(bmgrObj);

    //
    // Create the OAD implementation.
    //
    controller_ = new OAD::ProcessController_impl(db_, orb, poa_, endpointMgr,
    						  hostname, logger);
    oid_ = PortableServer::string_to_ObjectId("OAD");

    //
    // Register the OAD object with the Boot Manager
    //
    poa_ -> activate_object_with_id(oid_, controller_);
    OAD::ProcessController_var daemonRef = controller_ -> _this();
    bootManager -> add_binding(oid_, daemonRef);

    try
    {
	//
	// Contact the OAD registrar in the IMR and register
	// ourselves.
	//
	IMR::ServerObserver_var observer;
	if(!CORBA::is_nil(endpointMgr))
	{
	    OAD::ProcessEndpointManager_var endpointManager =
		controller_ -> _OB_getEndpointManager();
            OAD::ProcessList_var processes;
            controller_ -> list_processes(processes);

	    endpointMgr -> register_oad(
                hostname, processes, daemonRef, endpointManager, observer);

  	    controller_ -> _OB_setObserver(observer);
	}
    }
    catch(const IMR::NoSuchOAD& ex)
    {
	CORBA::String_var msg =
	    CORBA::string_dup("IMR: No OAD configured for: ");
	msg += ex.name;
	logger -> warning(msg);
    }
    catch(const CORBA::SystemException&)
    {
	CORBA::String_var msg =
	    CORBA::string_dup("OAD: Could not contact IMR");
	logger -> warning(msg);
    }
    
    if(logger -> tracePeerStatus() >= 1)
    {
        CORBA::String_var msg = CORBA::string_dup("ready for ");
        msg += hostname;
        logger -> trace("OAD", msg);
    }

    //
    // Activate the POA manager
    //
    if(!CORBA::is_nil(slaveManager_))
        slaveManager_ -> activate();
}

OADServer::~OADServer()
{
    delete controller_;

    delete db_;
}

void
OADServer::shutdown()
{
    controller_ -> shutdown();

    poa_ -> destroy(true, true);
}
