// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>

#include <IMRLogger.h>
#include <ProcController.h>
#include <ProcObserver.h>

#ifdef HAVE_STRINGS_H
#    include <strings.h>
#endif

#ifndef WIN32
#    include <sys/wait.h>
#    include <sys/stat.h>
#    include <signal.h>
#    include <fcntl.h>
#    include <signal.h>
#    include <unistd.h>
#    include <stdlib.h>
#    include <errno.h>
#endif

using namespace OAD;


// ----------------------------------------------------------------------
// WIN32ProcController constructor and destructor
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// ProcessReaper for WIN32
// ----------------------------------------------------------------------

namespace OAD
{

#ifdef WIN32
static char*
GetErrorText()
{
     DWORD rc;
     char* str = 0;
     char* cstr = 0;
     rc = ::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|
			  FORMAT_MESSAGE_FROM_SYSTEM|
			  FORMAT_MESSAGE_ARGUMENT_ARRAY,
			  0, GetLastError(), LANG_NEUTRAL, (char*)&str, 0, 0);
     if(str)
     {
	 cstr = CORBA::string_dup(str);
	 ::LocalFree((HLOCAL)str);
     }
     else
     {
	 cstr = CORBA::string_dup("FormatMessage failed");
     }
     return cstr;
}
#endif

class ProcessReaper : public JTCThread, public JTCMonitor
{
#ifdef WIN32
    typedef HANDLE Handle;
    typedef DWORD ProcessId;
#else
    typedef int Handle;
    typedef pid_t ProcessId;
#endif

    //
    // Stopped?
    //
    bool stop_;

    //
    // The maximum number of handles
    //
    int max_;

    //
    // Set of process handles
    //
    Handle* handles_;

    //
    // Associated process-ids. The process id index mirror the index
    // of the handle in the handles_ array
    //
    ProcessId* pids_;

    //
    // Number of handles and pids
    //
    int count_;

    //
    // Handle used to interrupt the WFMO
    //
    Handle interrupt_;

    //
    // The process observer
    //
    ProcObserver* observer_;

    //
    // The trace levels
    //
    IMROAD::Logger_var logger_;

    //
    // Wait for an event to occur. Return the index of the active
    // handle.
    //
    int
    waitEvent();

    //
    // Close a file handle
    //
    void
    doclose(Handle);

    //
    // Reset the interrupt flag
    //
    void
    resetInterrupt();

    //
    // Set the interrupt flag
    //
    void
    setInterrupt();

public:

    ProcessReaper(ProcObserver*, IMROAD::Logger_ptr);
    ~ProcessReaper();

    virtual void
    run();

    void
    terminate();

    void
    add(Handle, ProcessId);
};

} // End namespace OAD

//
// A process watcher handle
//
typedef JTCHandleT<ProcessReaper> ProcessReaperHandle;

// ----------------------------------------------------------------------
// ProcessReaper private member implementation
// ----------------------------------------------------------------------

#ifdef WIN32

int
ProcessReaper::waitEvent()
{
    //
    // Copy the handles.
    //
    HANDLE h[MAXIMUM_WAIT_OBJECTS];
    DWORD c;
    {
	JTCSynchronized synchronized(*this);
	c = count_;
	
	for(int i = 0; i < count_; ++i)
	    h[i] = handles_[i];
	
	::ResetEvent(interrupt_);
    }

    //
    // Wait for the event
    //
    DWORD result = ::WaitForMultipleObjects(c, h, false, INFINITE);
    
    //
    // Are we terminating?
    //
    if(stop_)
        return -1;
    
    //
    // No failures, or timeouts
    //
    assert(result != WAIT_FAILED && result != WAIT_TIMEOUT);

    //
    // Determine which process terminated
    //
    int rc = result - WAIT_OBJECT_0;
    assert(rc >= 0 && rc < count_);
    return rc;
}

void
ProcessReaper::doclose(Handle fd)
{
    ::CloseHandle(fd);
}

void
ProcessReaper::resetInterrupt()
{
    //
    // This method does nothing since the interrupt_ event is reset
    // when the handles are copied in waitEvent
    //
}

void
ProcessReaper::setInterrupt()
{
    ::SetEvent(interrupt_);
}

#else

int
ProcessReaper::waitEvent()
{
    int result;
    fd_set rfds;

    while(true)
    {
	//
	// Copy the handles
	//
	FD_ZERO(&rfds);
	int max;
	{
	    JTCSynchronized synchronized(*this);
	    max = 0;
	
	    for(int i = 0; i < count_; ++i)
	    {
		if(handles_[i] > max)
		    max = handles_[i]+1;
		FD_SET(handles_[i], &rfds);
	    }
	}
    
	//
	// Wait for the event
	//
	result = ::select(max, &rfds, 0, 0, 0);

        //
        // Are we terminating?
        //
        if(stop_)
            return -1;
    
	//
	// If the exit was caused by an interrupt continue
	//
	if(result == -1 && errno == EINTR)
	    continue;
	break;
    }

    //
    // No failures
    //
    assert(result != -1);

    //
    // Determine the handle that caused the interrupt
    //
    for(int i = 0 ; i < count_ ; i++)
	if(FD_ISSET(handles_[i], &rfds))
	    return i;

    assert(false);
    return -1;
}

void
ProcessReaper::doclose(Handle fd)
{
    ::close(fd);
}

void
ProcessReaper::resetInterrupt()
{
    char c;
    ::read(handles_[0], &c, 1);
}

void
ProcessReaper::setInterrupt()
{
    char c = 0;
    ::write(interrupt_, &c, 1);
}

#endif
// ----------------------------------------------------------------------
// ProcessReaper constructor/destructor
// ----------------------------------------------------------------------

ProcessReaper::ProcessReaper(ProcObserver* observer, IMROAD::Logger_ptr logger)
    : stop_(false),
      count_(0),
      observer_(observer),
      logger_(IMROAD::Logger::_duplicate(logger))
{
#ifdef WIN32
    max_ = MAXIMUM_WAIT_OBJECTS;
#else
    //
    // Close all file descriptors except the write end of the
    // pipe
    //
    max_ = sysconf(_SC_OPEN_MAX);
    if(max_ <= 0)
    {
        CORBA::String_var msg = CORBA::string_dup("OAD: sysconf() failed: ");
	msg += strerror(errno);
        logger_ -> error(msg);
	//TODO: exit?
	exit(-1);
    }
#endif

    handles_ = new Handle[max_];
    pids_ = new ProcessId[max_];

#ifdef WIN32
    //
    // The first waitable handle is an event handle that is used to
    // notify the reaper that a new process has been added to the set
    // of waitable handles
    //
    interrupt_ = ::CreateEvent(0, false, false, 0);

    handles_[0] = interrupt_;
    pids_[0] = 0;
    ++count_;
#else
    int fds[2];
    if(pipe(fds) != 0)
    {
        CORBA::String_var msg = CORBA::string_dup("OAD: pipe: ");
	msg += strerror(errno);
        logger_ -> error(msg);
	//TODO: exit?
	exit(-1);
    }

    interrupt_ = fds[1];
    handles_[0] = fds[0];
    pids_[0] = 0;
    ++count_;
#endif
}

ProcessReaper::~ProcessReaper()
{
    for(int i = 0; i < count_; i++)
	doclose(handles_[i]);
#ifndef WIN32
    doclose(interrupt_);
#endif
    delete[] handles_;
    delete[] pids_;
}

// ----------------------------------------------------------------------
// ProcessReaper public member implementation
// ----------------------------------------------------------------------

void
ProcessReaper::run()
{
    for(;;)
    {
	int idx = waitEvent();

	//
	// Lock the monitor again
	//
	JTCSynchronized synchronized(*this);

	//
	// Done?
	//
	if(stop_)
	    return;

	//
	// If the interrupt_ event is set then the handle set
	// changed. Cancel the interrupt and Continue.
	//
	if(idx == 0)
	{
	    resetInterrupt();
	    continue;
	}

	if(logger_ -> traceProcessControl() >= 2)
	{
	    CORBA::String_var msg = CORBA::string_dup("process died: ");
	    msg += pids_[idx];
	    logger_ -> trace("OAD", msg);
	}

	//
	// Save the process id
	//
	ProcessId deadpid = pids_[idx];

	//
	// Close the handle
	//
	doclose(handles_[idx]);
	    
	//
	// Remove the pid and the handle from the waiting block. Copy
	// the remainder down.
	//
	for(--count_; idx < count_; idx++)
	{
	    handles_[idx] = handles_[idx+1];
	    pids_[idx] = pids_[idx+1];
	}

	observer_ -> processDestroy(deadpid);
    }
}

void
ProcessReaper::terminate()
{
    JTCSynchronized synchronized(*this);
    stop_ = true;
    setInterrupt();
}

void
ProcessReaper::add(Handle handle, ProcessId pid)
{
    JTCSynchronized synchronized(*this);

    //
    // TODO: resource limits -- we can only handle
    // MAXIMUM_WAIT_OBJECTS, etc
    //
    handles_[count_] = handle;
    pids_[count_] = pid;
    ++count_;

    //
    // Notify the run() method that the handle set has changed
    //
    setInterrupt();
}

#ifdef WIN32
// ----------------------------------------------------------------------
// WIN32ProcController
// ----------------------------------------------------------------------

namespace OAD
{

class WIN32ProcController : public ProcController
{
    ProcessReaperHandle watcher_;

public:

    WIN32ProcController(ProcObserver*, IMROAD::Logger_ptr);
    ~WIN32ProcController();

    virtual IMR::ProcessID
    spawn(const char*, const char*, const IMR::ArgList&, const char*,
          const char*)
	throw(OAD::ExecFailed);

    virtual void
    kill(IMR::ProcessID)
	throw(OAD::KillFailed);
};

} // End namespace OAD

// ----------------------------------------------------------------------
// WIN32ProcController constructor/destructor
// ----------------------------------------------------------------------

WIN32ProcController::WIN32ProcController(ProcObserver* observer,
					 IMROAD::Logger_ptr logger)
    : ProcController(observer, logger)
{
    watcher_ = new ProcessReaper(observer, logger);
    watcher_ -> start();
}

WIN32ProcController::~WIN32ProcController()
{
    //
    // Ask the updater to terminate, and wait for it to stop
    //
    watcher_ -> terminate();
    while(watcher_ -> isAlive())
    {
	try
	{
	    watcher_ -> join();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }
}

// ----------------------------------------------------------------------
// WIN32ProcController public member implementation
// ----------------------------------------------------------------------

IMR::ProcessID
WIN32ProcController::spawn(const char* server, const char* exec,
			   const IMR::ArgList& args, const char* dir,
			   const char* id)
    throw(OAD::ExecFailed)
{
    //
    // Trace the command line parameters
    //
    if(logger_ -> traceProcessControl() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("exec: ");
	msg += exec;
	for(CORBA::ULong i = 0 ; i < args.length() ; i++)
	{
	    msg += " ";
	    msg += args[i];
	}
	logger_ -> trace("OAD", msg);
    }

    //
    // Under WIN32 we must use the CreateProcess API.
    //
    // TODO: fill in the command line environment?
    //
    PROCESS_INFORMATION procInfo;
    STARTUPINFO sinfo;
    memset(&sinfo, 0, sizeof(sinfo));
    sinfo.cb = sizeof(sinfo);

    CORBA::String_var cmdline = CORBA::string_dup(exec);
    unsigned int i;
    for(i = 0; i < args.length(); i++)
    {
        cmdline += " ";
        cmdline += args[i];
    }

    if(!::CreateProcess(exec,
                        cmdline.inout(),
                        0, 0, false,
                        CREATE_NEW_PROCESS_GROUP|DETACHED_PROCESS,
                        0,
                        strlen(dir) == 0 ? "C:\\" : dir,
                        &sinfo,
                        &procInfo))
    {
	CORBA::String_var err = OAD::GetErrorText();
	CORBA::String_var msg = CORBA::string_dup("CreateProcess: ");
	msg += err;
	throw OAD::ExecFailed(msg);
    }

    watcher_ -> add(procInfo.hThread, procInfo.dwProcessId);

    ::CloseHandle(procInfo.hProcess);

    observer_ ->
        processCreate(server, id, (IMR::ProcessID)procInfo.dwProcessId);

    return (IMR::ProcessID)procInfo.dwProcessId;
}

void
WIN32ProcController::kill(IMR::ProcessID pid)
    throw(OAD::KillFailed)
{
    HANDLE h = ::OpenProcess(PROCESS_ALL_ACCESS, false, pid);
    if(h == NULL)
    {
	DWORD err = ::GetLastError();
	CORBA::String_var msg = CORBA::string_dup("OpenProcess failed: ");
	msg += err;
	throw OAD::KillFailed(msg);
    }

    if(!::TerminateProcess(h, 1))
    {
        ::CloseHandle(h);
	DWORD err = ::GetLastError();
	CORBA::String_var msg =
	    CORBA::string_dup("TerminateProcess failed: ");
	msg += err;
	throw OAD::KillFailed(msg);
    }

    ::CloseHandle(h);
}

#else

//
// Called on a SIGCHLD. This function should reap each dead child and
// update the child list. This function must be extern "C" since its
// being used with a C function (signal).
//
extern "C" void
OADSignalHandler(int)
{
    //
    // Reap all dead children
    //
    pid_t pid;
    do
    {
	int status;
	pid = waitpid(-1, &status, WNOHANG);
    }
    while(pid > 0);
}

// ----------------------------------------------------------------------
// UNIXProcController
// ----------------------------------------------------------------------

class UNIXProcController : public ProcController
{
    ProcessReaperHandle watcher_;

public:

    UNIXProcController(ProcObserver*, IMROAD::Logger_ptr);
    ~UNIXProcController();

    virtual IMR::ProcessID
    spawn(const char*, const char*, const IMR::ArgList&, const char*, 
          const char*)
	throw(OAD::ExecFailed);

    virtual void
    kill(IMR::ProcessID)
	throw(OAD::KillFailed);
};

// ----------------------------------------------------------------------
// UNIXProcController constructor/destructor
// ----------------------------------------------------------------------

UNIXProcController::UNIXProcController(ProcObserver* observer,
				       IMROAD::Logger_ptr logger)
    : ProcController(observer, logger)
{
    //
    // Install a signal handler for SIGCHLD so that zombie processes
    // are not created
    //
    struct sigaction sa;
    sa.sa_handler = OADSignalHandler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;
    sigaction(SIGCHLD, &sa, 0);

    watcher_ = new ProcessReaper(observer, logger);
    watcher_ -> start();
}

UNIXProcController::~UNIXProcController()
{
    struct sigaction sa;
    sa.sa_handler = SIG_DFL;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(SIGCHLD, &sa, 0);

    //
    // Ask the updater to terminate, and wait for it to stop
    //
    watcher_ -> terminate();
    while(watcher_ -> isAlive())
    {
	try
	{
	    watcher_ -> join();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }
}

// ----------------------------------------------------------------------
// UNIXProcController public member implementation
// ----------------------------------------------------------------------

IMR::ProcessID
UNIXProcController::spawn(const char* server, const char* exec,
			  const IMR::ArgList& args, const char* dir, 
			  const char* id)
    throw(OAD::ExecFailed)
{
    //
    // Trace the command line parameters
    //
    if(logger_ -> traceProcessControl() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("exec: ");
	msg += exec;
	for(CORBA::ULong i = 0 ; i < args.length() ; i++)
	{
	    msg += " ";
	    msg += args[i];
	}
	logger_ -> trace("OAD", msg);
    }

    //
    // Create a pipe
    //
    int fds[2];
    if(::pipe(fds) < 0)
    {
	//...
    }

    //
    // Fork a new process
    //
    pid_t pid = fork();
    if(pid < 0) // Failed
    {
	const char* err = strerror(errno);
	CORBA::String_var msg = CORBA::string_dup("fork failed: ");
	msg += err;

	::close(fds[0]);
	::close(fds[1]);

	//
	// Signal handling is restored in a catch block below
	//
	throw OAD::ExecFailed(msg);
    }

    //
    // Child.
    //
    if(pid == 0)
    {
	//
	// Start the daemonize process. This means:
	// - becoming a session leader.
	// - changing the working directory to "/"
	// - setting umask
	// - closing all unneeded file descriptors.
	// - resetting signals to default behavior
	//
	// TODO: environment, uid?
	//

	int i;

	//
	// Become a session leader
	//
	setsid();

	//
	// Change working directory to "/"
	//
	if(chdir(strlen(dir) == 0 ? "/" : dir) != 0)
	{
            CORBA::String_var msg =
	        CORBA::string_dup("OAD: chdir(\"");
	    msg += dir;
	    msg += "\") failed: ";
	    msg += strerror(errno);
            logger_ -> error(msg);
	    exit(-1);
	}

	//
	// Set umask to a sensible default
	//
	umask(027);

	//
	// Close all file descriptors except the write end of the
	// pipe.
	//
	int maxFds = sysconf(_SC_OPEN_MAX);
	if(maxFds <= 0)
	{
            CORBA::String_var msg =
	        CORBA::string_dup("OAD: sysconf() failed: ");
	    msg += strerror(errno);
            logger_ -> error(msg);
	    exit(-1);
	}
	for(i = 0 ; i < maxFds ; i++)
	{
	    if(i != fds[1])
		::close(i);
	}

	//
	// Make sure hat fd[1] is file descriptor 3.
	//
	int fd;
	if(fds[1] != 3)
	{
		fd = dup2(fds[1], 3);
		assert(fd == 3);
		::close(fds[1]);
		fds[1] = 3;
	}

	//
	// Next attach file descriptors 0, 1 and 2 to /dev/null.
	// This assists applications that write to stdout and
	// stderr not to do stupid things. We open all three descriptors
	// for read/write because some kernels open them that way and
	// some silly libraries actually expect that they can read from
	// stdout and stderr or write to stdin...
	//
	fd = open("/dev/null", O_RDWR);
	assert(fd == 0);
	fd = dup2(0, 1);
	assert(fd == 1);
	fd = dup2(1, 2);
	assert(fd == 2);

	//
	// Reset all signals to their default behavior.
	//
	struct sigaction sa;
	sa.sa_handler = SIG_DFL;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	for (int sig = 1; i <= 64; i++)
	    sigaction(sig, &sa, (struct sigaction *)0);

	//
	// TODO: environment?
	//

	//
	// Exec the server
	//
	char** const argv = new char*[args.length() + 2];
	argv[0] = strdup(exec);

	CORBA::ULong arg;
	for(arg = 0 ; arg < args.length() ; arg++)
	    argv[arg + 1] = strdup(args[arg]);
	argv[args.length() + 1] = 0;

	execv(exec, argv);

	for(arg = 0 ; arg < args.length() + 1 ; arg++)
	    free(argv[arg]);
	delete [] argv;
        
	//
	// This should not be reached
	//
        CORBA::String_var msg = CORBA::string_dup("OAD: execv() failed: ");
	msg += strerror(errno);
        logger_ -> error(msg);
	exit(-1);
    }

    close(fds[1]);

    watcher_ -> add(fds[0], pid);

    observer_ -> processCreate(server, id, (IMR::ProcessID)pid);

    return (IMR::ProcessID)pid;
}

void
UNIXProcController::kill(IMR::ProcessID id)
    throw(OAD::KillFailed)
{
    //
    // Kill the process. SIGKILL - overkill?
    //
    if(::kill((pid_t)id, SIGKILL) < 0)
    {
	const char* err = strerror(errno);
	CORBA::String_var msg = CORBA::string_dup("kill: ");
	msg += err;
	throw OAD::KillFailed(msg);
    }
}

#endif /*WIN32*/

// ----------------------------------------------------------------------
// ProcController constructor and destructor
// ----------------------------------------------------------------------

ProcController::ProcController(ProcObserver* observer,
			       IMROAD::Logger_ptr logger)
    : observer_(observer),
      logger_(IMROAD::Logger::_duplicate(logger))
{
}

ProcController::~ProcController()
{
}

// ----------------------------------------------------------------------
// ProcControllerFactory public member implementation
// ----------------------------------------------------------------------

/*static*/ ProcController*
ProcControllerFactory::createProcController(ProcObserver* observer,
					    IMROAD::Logger_ptr logger)
{
#ifdef WIN32
    return new WIN32ProcController(observer, logger);
#else
    return new UNIXProcController(observer, logger);
#endif
}
