// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IMR.h>
#include <OB/Logger.h>

#include <IMRLogger.h>
#include <ProcessDatabase.h>
#include <ProcessController_impl.h>
#include <ProcController.h>
#include <ProcObserver.h>
#include <ProcessEndpointManager_impl.h>

using namespace OAD;

// ----------------------------------------------------------------------
// ProcessController_impl private member implementation
// ----------------------------------------------------------------------

void
ProcessController_impl::reconnect()
{
    ProcessEndpointManager_var registrar = endpointManager_ -> _this();

    //
    // Run through children processes
    //
    OADDB_4_0_2::ServerRecordSeq_var children = db_ -> getAllServers();
    for(CORBA::ULong i = 0; i < children -> length(); i++)
    {
	bool removeProcess = false;
	CORBA::String_var err;
	try
	{
	    //
	    // If the monitor is nil OR the monitor no longer exists
	    // or cannot be contacted then then the process is removed
	    // from the child list
	    //
	    if(!CORBA::is_nil(children[i].endpoint))
	    {
		if(children[i].endpoint -> _non_existent())
		{
		    removeProcess = true;
		    err = CORBA::string_dup("_non_existent");
		}
		else
		    children[i].endpoint -> reestablish_link(registrar);
	    }
	    else
	    {
		removeProcess = true;
		err = CORBA::string_dup("nil monitor");
	    }
	}
	//
	// TODO: Not quite right! We should catch the valid
	// exceptions here... (COMM_FAILURE, TRANSIENT, ...)
	//
	catch(const CORBA::SystemException& e)
	{
	    removeProcess = true;
	    err = e._to_string();
	}

	if(removeProcess)
	{
            if(logger_ -> traceServerStatus() >= 2)
    	    {
		CORBA::String_var msg = CORBA::string_dup("reconnect: ");
		msg += children[i].server;
		msg += '(';
		msg += children[i].pid;
		msg += ") register_again: ";
		msg += err;
		logger_ -> trace("OAD", msg);
	    }
	    db_ -> removeRecord(children[i].server);
	}
    }
}

// ----------------------------------------------------------------------
// ProcessController_impl constructor/destructor
// ----------------------------------------------------------------------

ProcessController_impl::ProcessController_impl(ProcessDatabase* db,
			 CORBA::ORB_ptr orb,
                         PortableServer::POA_ptr poa,
                         IMR::OADEndpointManager_ptr oadEndpointManager,
			 const char* host,
			 IMROAD::Logger_ptr logger)
    : host_(host),
      db_(db),
      poa_(PortableServer::POA::_duplicate(poa)),
      logger_(IMROAD::Logger::_duplicate(logger)),
      oadEndpointManager_(
	  IMR::OADEndpointManager::_duplicate(oadEndpointManager))
{
    //
    // Create the process observer
    //
    observer_ = new ProcObserver(db_, logger_);

    //
    // Create the controller registrar object
    //
    endpointManager_ = new ProcessEndpointManager_impl(orb, poa, db_,
    						       observer_, logger_);
    PortableServer::ObjectId_var oid =
        PortableServer::string_to_ObjectId("PEM");
    poa -> activate_object_with_id(oid, endpointManager_);

    //
    // Create the controller
    //
    controller_ = ProcControllerFactory::createProcController(observer_,
    							      logger_);

    //
    // Ask each child process to reconnect to the OAD
    //
    reconnect();
}

ProcessController_impl::~ProcessController_impl()
{ 
    delete endpointManager_;

    delete controller_;

    delete observer_;
}

void
ProcessController_impl::shutdown()
{
    endpointManager_ -> shutdown();
}

// ----------------------------------------------------------------------
// ProcessController_impl public member implementation
// ----------------------------------------------------------------------

//
// IDL:ooc.com/OAD/ProcessController/host:1.0
//
char*
ProcessController_impl::host()
    throw(CORBA::SystemException)
{
    return CORBA::string_dup(host_);
}

//
// IDL:ooc.com/OAD/ProcessController/exec_process:1.0
//
void
ProcessController_impl::exec_process(const char* server,
			             const char* exec,
			             const IMR::ArgList& args,
				     const char* dir,
			             const char* id)
    throw(ExecFailed,
	  CORBA::SystemException)
{
    //
    // IMR ensures that exec_process() is not called concurrently
    // on the same server, thus synchronization is not required
    // here.
    //
    OADDB_4_0_2::ServerRecord rec;
    if(db_ -> queryServerByName(server, rec))
    {
        if(logger_ -> traceProcessControl() >= 2)
	{
	    CORBA::String_var msg = CORBA::string_dup("exec: ");
	    msg += server;
	    msg += ": rejected, server already running";
	    logger_ -> trace("OAD", msg);
	}
	throw ExecFailed("Server running");
    }

    if(logger_ -> traceProcessControl() >= 1)
    {
	CORBA::String_var msg = CORBA::string_dup("exec: ");
	msg += server;
	logger_ -> trace("OAD", msg);
    }
    
    IMR::ProcessID pid = controller_ -> spawn(server, exec, args, dir, id);

    if(logger_ -> traceProcessControl() >= 1)
    {
	CORBA::String_var msg = CORBA::string_dup("exec: ");
	msg += server;
	msg += " ";
	msg += "PID: ";
	msg += pid;
	logger_ -> trace("OAD", msg);
    }
}

//
// IDL:ooc.com/OAD/ProcessController/stop_process:1.0
//
void
ProcessController_impl::stop_process(const char* server)
    throw(CORBA::SystemException)
{
    OADDB_4_0_2::ServerRecord rec;
    if(db_ -> queryServerByName(server, rec))
    {
        if(logger_ -> traceProcessControl() >= 2)
        {
	    CORBA::String_var msg = CORBA::string_dup("stop_process: ");
	    msg += server;
	    logger_ -> trace("OAD", msg);
        }

        rec.endpoint -> stop();
    }
}

//
// IDL:ooc.com/OAD/ProcessController/kill_process:1.0
//
void
ProcessController_impl::kill_process(const char* server)
    throw(KillFailed,
	  CORBA::SystemException)
{
    CORBA::ULong pid = 0xFFFFFFFF;
    OADDB_4_0_2::ServerRecord rec;
    if(db_ -> queryServerByName(server, rec))
	pid = rec.pid;

    if(logger_ -> traceProcessControl() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("kill_process: ");
	msg += server;
	msg += '(';
	if(pid == 0xffffffff)
	    msg += "UNKNOWN";
	else
	    msg += pid;
	msg += ")";
	logger_ -> trace("OAD", msg);
    }

    //
    // Not found?
    //
    if(pid == 0xffffffff)
	throw KillFailed("No such server");

    controller_ -> kill(pid);
}

//
// IDL:ooc.com/OAD/ProcessController/reset_process:1.0
//
void
ProcessController_impl::reset_process(const char* server)
    throw(CORBA::SystemException)
{
    OADDB_4_0_2::ServerRecord rec;
    if(db_ -> queryServerByName(server, rec))
    {
        if(logger_ -> traceProcessControl() >= 2)
        {
            CORBA::String_var msg = CORBA::string_dup("reset_process: ");
            msg += server;
            msg += '(';
            msg += rec.pid;
            msg += ')';
            logger_ -> trace("OAD", msg);
        }
        db_ -> removeRecord(server);
    }
}

//
// IDL:ooc.com/OAD/ProcessController/list_processes:1.0
//
void
ProcessController_impl::list_processes(OAD::ProcessList_out processes)
    throw(CORBA::SystemException)
{
    OADDB_4_0_2::ServerRecordSeq_var children = db_ -> getAllServers();
    processes = new OAD::ProcessList(children -> length());
    processes -> length(children -> length());
    for(CORBA::ULong i = 0; i < children -> length(); i++) {
	processes[i] = children[i].server;
    }
}

//
// IDL:ooc.com/OAD/ProcessController/connect_imr:1.0
//
void
ProcessController_impl::connect_imr(
    IMR::OADEndpointManager_ptr oadEndpointManager)
    throw(IMR::NoSuchOAD, CORBA::SystemException)
{
    if(CORBA::is_nil(oadEndpointManager))
	throw CORBA::BAD_PARAM();

    oadEndpointManager_ =
	IMR::OADEndpointManager::_duplicate(oadEndpointManager);

    ProcessList_var processes;
    list_processes(processes);

    IMR::ServerObserver_var observer;
    oadEndpointManager_ -> register_oad(
        host_, processes, _this(), endpointManager_ -> _this(), observer);

    _OB_setObserver(observer);
}

PortableServer::POA_ptr
ProcessController_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

//
// Internal methods
//
void
ProcessController_impl::_OB_setObserver(
    IMR::ServerObserver_ptr observer)
{
    observer_ -> setObserver(observer);
}

ProcessEndpointManager_ptr
ProcessController_impl::_OB_getEndpointManager()
{
    return endpointManager_ -> _this();
}
