// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/seq_converter.h>

#include <ProcessDatabase.h>
#include <IMRUtil.h>
#include <OSLayer.h>

using namespace OAD;

// ----------------------------------------------------------------------
// ProcessDatabase private member implementation
// ----------------------------------------------------------------------

void
ProcessDatabase::load()
{
    OB_STL::list<CORBA::ULong> ids;
    db_ -> getAll(ids);

    children_.length(ids.size());

    CORBA::ULong idx = 0;
    OB_STL::list<CORBA::ULong>::iterator iter;
    for(iter = ids.begin() ; iter != ids.end() ; ++iter, ++idx)
    {
        CORBA::Any_var any;
        if(!db_ -> get(*iter, any))
        {
            logger_ -> error("Error reading database");
            throw CORBA::PERSIST_STORE();
        }

        const OADDB_4_0_2::ServerRecord* dbrec;
        if(any >>= dbrec)
        {
            OADDB_4_0_2::ServerRecord rec = *dbrec;
            rec.id = *iter;
	    children_[idx] = rec;
        }
        else
        {
            logger_ -> error("Bad record in database");
            throw CORBA::PERSIST_STORE();
        }
    }
}

// ----------------------------------------------------------------------
// ProcessDatabase constructor/destructor
// ----------------------------------------------------------------------

ProcessDatabase::ProcessDatabase(CORBA::ORB_ptr orb)
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    logger_ = oborb -> logger();

    CORBA::String_var OADdb = CORBA::string_dup("");

    OB::Properties_var properties = oborb -> properties();
    const char* dbdir = properties -> getProperty("ooc.imr.dbdir");
    if(dbdir != 0)
    {
        struct stat sbuf;
        if(!OSLayer::stat(dbdir, sbuf))
        {
            if(!OSLayer::mkdir(dbdir, 0700))
            {
                CORBA::String_var msg = CORBA::string_dup("mkdir: ");
                msg += dbdir;
                msg += " failed";
                logger_ -> error(msg);
                throw CORBA::PERSIST_STORE();
            }
            OSLayer::stat(dbdir, sbuf);
        }
        if(!OSLayer::isDir(sbuf))
        {
            CORBA::String_var msg = CORBA::string_dup(dbdir);
            msg += ": not directory.";
            logger_ -> error(msg);
            throw CORBA::PERSIST_STORE();
        }

        OADdb += dbdir;
        OADdb += "/";
    }
        
    OADdb += "oad-hot.db";

    db_ = new OB::Database(IMRUtil::codec(orb), OADdb, true,
    			   OADDB_4_0_2::_tc_ServerRecord, 512);

    try
    {
        load();
    }
    catch(const CORBA::Exception&)
    {
        delete db_;
    }
}

ProcessDatabase::~ProcessDatabase()
{
    delete db_;
}

// ----------------------------------------------------------------------
// ProcessDatabase public member implementation
// ----------------------------------------------------------------------

bool
ProcessDatabase::createRecord(const char* server, const char* instance,
			      IMR::ProcessID pid,
			      OADDB_4_0_2::ServerRecord& rec)
{
    JTCSynchronized sync(*this);

    CORBA::ULong i;
    for(i = 0; i < children_.length(); i++)
	if(strcmp(server, children_[i].server) == 0)
	    break;
    if(i < children_.length())
	return false;

    rec.server = CORBA::string_dup(server);
    rec.instance = CORBA::string_dup(instance);
    rec.pid = pid;
    rec.endpoint = OAD::ProcessEndpoint::_nil();

    CORBA::Any any;
    any <<= rec;
    rec.id = db_ -> add(any);
    if(rec.id == 0)
        return false;

    CORBA::ULong childrenLength = children_.length();
    children_.length(childrenLength + 1);
    children_[childrenLength] = rec;

    return true;
}

void
ProcessDatabase::removeRecord(const char* server)
{
    JTCSynchronized sync(*this);
    for(CORBA::ULong i = 0; i < children_.length(); i++)
	if(strcmp(server, children_[i].server) == 0)
	{
	    db_ -> remove(children_[i].id);
	    OB::remove_seq_item(children_, i);
	    return;
	}
}

void
ProcessDatabase::updateRecord(const OADDB_4_0_2::ServerRecord& rec)
{
    JTCSynchronized sync(*this);

    CORBA::ULong i;
    for(i = 0; i < children_.length(); i++)
	if(strcmp(rec.server, children_[i].server) == 0)
	    break;
    assert(i < children_.length());
    children_[i] = rec;

    CORBA::Any any;
    any <<= rec;
    db_ -> update(rec.id, any);
}

bool
ProcessDatabase::queryServerByName(const char* server,
				   OADDB_4_0_2::ServerRecord& rec)
{
    JTCSynchronized sync(*this);

    CORBA::ULong i;
    for(i = 0; i < children_.length(); i++)
	if(strcmp(server, children_[i].server) == 0)
	{
	    rec = children_[i];
	    return true;
	}
    return false;
}

bool
ProcessDatabase::queryServerById(const char* server,
                                 const char* instance,
				 OADDB_4_0_2::ServerRecord& rec)
{
    JTCSynchronized sync(*this);

    CORBA::ULong i;
    for(i = 0; i < children_.length(); i++)
	if(strcmp(server, children_[i].server) == 0 &&
	   strcmp(instance, children_[i].instance) == 0)
	{
	    rec = children_[i];
	    return true;
	}
    return false;
}

bool
ProcessDatabase::queryServerByPid(IMR::ProcessID pid, OADDB_4_0_2::ServerRecord& rec)
{
    JTCSynchronized sync(*this);

    CORBA::ULong i;
    for(i = 0; i < children_.length(); i++)
	if(pid == children_[i].pid)
	{
	    rec = children_[i];
	    return true;
	}
    return false;
}

OADDB_4_0_2::ServerRecordSeq*
ProcessDatabase::getAllServers()
{
    JTCSynchronized sync(*this);

    OADDB_4_0_2::ServerRecordSeq_var rc =
	new OADDB_4_0_2::ServerRecordSeq(children_.length());
    rc -> length(children_.length());
    for(CORBA::ULong i = 0; i < children_.length(); i++)
	rc[i] = children_[i];

    return rc._retn();
}
