// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <ServerFactory_impl.h>
#include <Server_impl.h>
#include <ServerDatabase.h>
#include <OADDatabase.h>

using namespace IMR;

// ----------------------------------------------------------------------
// ServerFactory_impl private member functions
// ----------------------------------------------------------------------

Server_ptr
ServerFactory_impl::createServerReference(CORBA::Long id)
{
    //
    // Create the ID for the record.
    //
    CORBA::String_var sid = CORBA::string_dup("");
    sid += id;
    PortableServer::ObjectId_var oid = PortableServer::string_to_ObjectId(sid);

    //
    // Create the object-reference.
    //
    CORBA::Object_var obj = serverPOA_ ->
	create_reference_with_id(oid, "IDL:ooc.com/IMR/Server:1.0");
    return Server::_narrow(obj);
}

// ----------------------------------------------------------------------
// ServerFactory_impl constructor and destructor
// ----------------------------------------------------------------------

ServerFactory_impl::ServerFactory_impl(
    ServerDatabase* db,
    OADDatabase* oaddb,
    CORBA::ORB_ptr orb,
    PortableServer::POA_ptr poa,
    IMROAD::Logger_ptr logger,
    bool administrative, 
    ServerStarter* serverStarter)
    : db_(db),
      oaddb_(oaddb),
      poa_(PortableServer::POA::_duplicate(poa)),
      server_(new Server_impl(db, oaddb, orb, logger, administrative, 
      			      serverStarter)),
      administrative_(administrative)
{
    CORBA::PolicyList pl(6);
    pl.length(6);
    pl[0] = poa_ -> create_lifespan_policy(PortableServer::PERSISTENT);
    pl[1] = poa_ -> create_id_assignment_policy(PortableServer::USER_ID);
    pl[2] = poa_ -> create_request_processing_policy(
	PortableServer::USE_DEFAULT_SERVANT);
    pl[3] = poa_ -> create_id_uniqueness_policy(
	PortableServer::MULTIPLE_ID);
    pl[4] = poa_ -> create_servant_retention_policy(
	PortableServer::NON_RETAIN);
    pl[5] = poa_ -> create_implicit_activation_policy(
	PortableServer::NO_IMPLICIT_ACTIVATION);
    PortableServer::POAManager_var manager = poa_ -> the_POAManager();
    serverPOA_ = poa_ -> create_POA("server", manager, pl);
    serverPOA_ -> set_servant(server_);
}

ServerFactory_impl::~ServerFactory_impl()
{
    delete server_;
}

// ----------------------------------------------------------------------
// ServerFactory_impl public member functions
// ----------------------------------------------------------------------

//
// Implementation for IMR::ServerFactory
//
Server_ptr
ServerFactory_impl::get_server(const char* server)
    throw(NoSuchServer,
	  CORBA::SystemException)
{
    ServerRecord_var rec = db_ -> queryRecordByServer(server);
    if(CORBA::is_nil(rec))
	throw NoSuchServer(server);

    return createServerReference(rec -> record().id);
}

//
// Mapping for IMR::ServerFactory
//
Server_ptr
ServerFactory_impl::create_server_record(const char* server)
    throw(IMR::ServerAlreadyRegistered,
	  CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    JTCSynchronized sync(*db_);

    //
    // Create a new record.
    //
    ServerRecord_var rec = db_ -> createRecord(server);
    if(CORBA::is_nil(rec))
	throw ServerAlreadyRegistered(server);

    return createServerReference(rec -> record().id);
}

ServerSeq*
ServerFactory_impl::list_servers()
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*db_);

    IMRDB_4_1_0::ServerIdSeq_var seq = db_ -> getAllServers();

    ServerSeq_var rc = new ServerSeq(seq -> length());
    rc -> length(seq -> length());
    for(CORBA::ULong i = 0; i < seq -> length(); i++)
        rc[i] = createServerReference(seq[i]);

    return rc._retn();
}

ServerSeq*
ServerFactory_impl::list_servers_by_host(const char* host)
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*db_);

    IMRDB_4_1_0::ServerIdSeq_var seq = db_ -> queryAllByHost(host);

    ServerSeq_var rc = new ServerSeq(seq -> length());
    rc -> length(seq -> length());
    for(CORBA::ULong i = 0; i < seq -> length(); i++)
        rc[i] = createServerReference(seq[i]);

    return rc._retn();
}

ServerSeq*
ServerFactory_impl::list_orphaned_servers()
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*db_);

    IMRDB_4_1_0::ServerIdSeq_var seq = db_ -> getAllServers();
    IMRDB_4_0_2::OADRecord rec;

    ServerSeq_var rc = new ServerSeq;
    CORBA::ULong length = 0;
    rc -> length(length);
    for(CORBA::ULong i = 0; i < seq -> length(); i++)
    {
        ServerRecord_var server = db_ -> queryRecordById(seq[i]);
	if(!oaddb_ -> queryRecordByHost(server -> record().host, rec))
	{
	    rc -> length(++length);
            rc[length - 1] = createServerReference(seq[i]);
	}
    }

    return rc._retn();
}

PortableServer::POA_ptr
ServerFactory_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}
