// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/RefCounted.h>
#include <OB/seq_converter.h>

#include <Server_impl.h>
#include <ServerStarter.h>
#include <ServerDatabase.h>
#include <OADDatabase.h>

#include <stdlib.h>
#include <ctype.h>

#include <IMRLogger.h>
#include <OSLayer.h>
#include <IMRUtil.h>

using namespace IMR;

//
// TODO: have something other than a global database monitor for the
// servers.
//

// ----------------------------------------------------------------------
// Server_impl private member implementation
// ----------------------------------------------------------------------

ServerRecord_ptr
Server_impl::getRecord()
{
    PortableServer::ObjectId_var oid = current_ -> get_object_id();
    CORBA::String_var sid = PortableServer::ObjectId_to_string(oid);
    if(oid -> length() == 0 || !isdigit(sid[0]))
	throw CORBA::INV_IDENT(); // TODO: correct?

    CORBA::Long id = ::atol(sid.in());
    ServerRecord_ptr rec = db_ -> queryRecordById(id);
    if(CORBA::is_nil(rec))
	throw CORBA::OBJECT_NOT_EXIST();
    else
        return rec;
}

// ----------------------------------------------------------------------
// Server_impl constructor and destructor
// ----------------------------------------------------------------------

Server_impl::Server_impl(ServerDatabase* db,
			 OADDatabase* oaddb,
			 CORBA::ORB_ptr orb,
			 IMROAD::Logger_ptr logger,
			 bool administrative, 
			 ServerStarter* serverStarter)
    : db_(db),
      oaddb_(oaddb),
      logger_(IMROAD::Logger::_duplicate(logger)),
      administrative_(administrative), 
      serverStarter_(serverStarter)
{
    try
    {
	CORBA::Object_var obj =
	    orb -> resolve_initial_references("POACurrent");
	current_ = PortableServer::Current::_narrow(obj);
    }
    catch(const CORBA::ORB::InvalidName&)
    {
    }
    assert(!CORBA::is_nil(current_));
}

Server_impl::~Server_impl()
{
}

// ----------------------------------------------------------------------
// Server_impl public member implementation
// ----------------------------------------------------------------------

//
// IDL:ooc.com/IMR/Server/id:1.0
//
IMR::ServerID
Server_impl::id()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return rec -> record().id;
}

//
// IDL:ooc.com/IMR/Server/status:1.0
//
IMR::ServerStatus
Server_impl::status()
    throw(CORBA::SystemException) 
{
    ServerRecord_var rec = getRecord();
    return rec -> record().status;
}

//
// IDL:ooc.com/IMR/Server/manual:1.0
//
CORBA::Boolean
Server_impl::manual()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return rec -> record().started;
}

//
// IDL:ooc.com/IMR/Server/updateTime:1.0
//
CORBA::Long
Server_impl::updateTime()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return rec -> record().updateTime;
}

//
// IDL:ooc.com/IMR/Server/timesForked:1.0
//
CORBA::Short
Server_impl::timesForked()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return rec -> record().timesForked;
}

//
// IDL:ooc.com/IMR/Server/name:1.0
//
char*
Server_impl::name()
    throw(CORBA::SystemException) 
{
    ServerRecord_var rec = getRecord();
    return CORBA::string_dup(rec -> record().server);
}

void
Server_impl::name(const char* s)
    throw(CORBA::SystemException) 
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().server = CORBA::string_dup(s);
    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/host:1.0
//
char*
Server_impl::host()
    throw(CORBA::SystemException) 
{
    ServerRecord_var rec = getRecord();
    return CORBA::string_dup(rec -> record().host);
}

void
Server_impl::host(const char* s)
    throw(CORBA::SystemException) 
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    IMRDB_4_0_2::OADRecord oadrec;
    if(!oaddb_ -> queryRecordByHost(s, oadrec, true))
        throw CORBA::OBJECT_NOT_EXIST();

    rec -> record().host = CORBA::string_dup(oadrec.host);
    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/exec:1.0
//
char*
Server_impl::exec()
    throw(CORBA::SystemException) 
{
    ServerRecord_var rec = getRecord();
    return CORBA::string_dup(rec -> record().exec);
}

void
Server_impl::exec(const char* s)
        throw(CORBA::SystemException) 
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().exec = CORBA::string_dup(s);
    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/directory:1.0
//
char*
Server_impl::directory()
    throw(CORBA::SystemException) 
{
    ServerRecord_var rec = getRecord();
    return CORBA::string_dup(rec -> record().dir);
}

void
Server_impl::directory(const char* s)
        throw(CORBA::SystemException) 
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().dir = CORBA::string_dup(s);
    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/args:1.0
//
IMR::ArgList*
Server_impl::args()
    throw(CORBA::SystemException) 
{
    ServerRecord_var rec = getRecord();
    return new IMR::ArgList(rec -> record().args);
}

void
Server_impl::args(const IMR::ArgList& args)
        throw(CORBA::SystemException) 
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().args = args;
    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/mode:1.0
//
IMR::ServerActivationMode
Server_impl::mode()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return rec -> record().mode;
}

void
Server_impl::mode(IMR::ServerActivationMode mode)
    throw(CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().mode = mode;
    db_ -> updateRecord(rec);

    //
    // If setting the mode to persistent and the server is
    // currently stopped, start it.
    //
    if(mode == IMR::PERSISTENT && rec -> record().status == STOPPED)
        serverStarter_ -> add(rec -> record().server);
}

//
// IDL:ooc.com/IMR/Server/updateTimeout:1.0
//
CORBA::Long
Server_impl::updateTimeout()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return rec -> record().updateTimeout;
}

void
Server_impl::updateTimeout(CORBA::Long t)
    throw(CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().updateTimeout = t;
    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/failureTimeout:1.0
//
CORBA::Long
Server_impl::failureTimeout()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return rec -> record().failureTimeout;
}

void
Server_impl::failureTimeout(CORBA::Long t)
    throw(CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().failureTimeout = t;
    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/maxForks:1.0
//
CORBA::Short
Server_impl::maxForks()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return rec -> record().maxForks;
}

void
Server_impl::maxForks(CORBA::Short f)
    throw(CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().maxForks = f;
    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/activatePOAs:1.0
//
CORBA::Boolean
Server_impl::activatePOAs()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    return rec -> record().activatePOAs;
}

void
Server_impl::activatePOAs(CORBA::Boolean a)
    throw(CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    rec -> record().activatePOAs = a;
    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/create_poa_record:1.0
//
void
Server_impl::create_poa_record(const IMR::POAName& poaname)
    throw(IMR::POAAlreadyRegistered,
	  CORBA::SystemException) 
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    //
    // Create the status record for this new POA
    //
    for(CORBA::ULong poa = 0; poa < rec -> record().poas.length(); poa++)
	if(IMRUtil::poacmp(poaname, rec -> record().poas[poa].poa))
	    throw POAAlreadyRegistered(rec -> record().server, poa);

    IMRDB_4_1_0::POAStatusRecord poaRec;
    poaRec.poa = poaname;
    poaRec.state = IMR::NON_EXISTENT;

    CORBA::ULong poasLength = rec -> record().poas.length();
    rec -> record().poas.length(poasLength + 1);
    rec -> record().poas[poasLength] = poaRec;

    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/remove_poa_record:1.0
//
void
Server_impl::remove_poa_record(const IMR::POAName& poaname)
    throw(IMR::NoSuchPOA, CORBA::SystemException)
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    //
    // Create the status record without this POA
    //
    CORBA::ULong poa;
    for(poa = 0; poa < rec -> record().poas.length(); poa++)
	if(IMRUtil::poacmp(poaname, rec -> record().poas[poa].poa))
            break;

    if(poa == rec -> record().poas.length())
        throw NoSuchPOA(poa);

    OB::remove_seq_item(rec -> record().poas, poa);

    db_ -> updateRecord(rec);
}

//
// IDL:ooc.com/IMR/Server/get_poa_info:1.0
//
IMR::POAInfo*
Server_impl::get_poa_info(const IMR::POAName& poaname)
    throw(IMR::NoSuchPOA, CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    CORBA::ULong poa;
    for(poa = 0; poa < rec -> record().poas.length(); poa++)
	if(IMRUtil::poacmp(poaname, rec -> record().poas[poa].poa))
            break;

    if(poa == rec -> record().poas.length())
        throw NoSuchPOA(poaname);

    POAInfo_var poainfo = new POAInfo;

    poainfo -> poa = rec -> record().poas[poa].poa;
    poainfo -> state = rec -> record().poas[poa].state;

    return poainfo._retn();
}

//
// IDL:ooc.com/IMR/Server/list_poas:1.0
//
IMR::POAInfoSeq*
Server_impl::list_poas()
    throw(CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());
    
    POAInfoSeq_var seq = new POAInfoSeq(rec -> record().poas.length());
    seq -> length(rec -> record().poas.length());
    for(CORBA::ULong poa = 0; poa < rec -> record().poas.length(); poa++)
    {
        seq[poa].poa = rec -> record().poas[poa].poa;
	seq[poa].state = rec -> record().poas[poa].state;
    }

    return seq._retn();
}

//
// IDL:ooc.com/IMR/Server/clear_error_state:1.0
//
void
Server_impl::clear_error_state()
    throw(CORBA::SystemException) 
{
    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    //
    // Update the times forked, status and time
    //
    rec -> record().timesForked = 0;
    rec -> record().status = STOPPED;
    rec -> record().updateTime = OSLayer::time();

    //
    // Run through the POAs and invalidate all IORs
    //
    for(CORBA::ULong i = 0 ; i < rec -> record().poas.length() ; i++)
        rec -> record().poas[i].state = IMR::NON_EXISTENT;

    db_ -> updateRecord(rec);

    //
    // Ensure that the state transition is propogated.
    //
    {
        JTCSynchronized sync(*db_);
        db_ -> notifyAll();
    }

    //
    // Update the state record in the OAD
    //
    IMRDB_4_0_2::OADRecord oadrec;
    if(oaddb_ -> queryRecordByHost(rec -> record().host, oadrec) && oadrec.up)
    {
        try
        {
            oadrec.controller -> reset_process(rec -> record().server);
        }
        catch(const CORBA::SystemException& ex)
        {
            //
            // ProcessController isn't reachable. The OAD state
	    // record should be updated by our death detector.
            //
	    if(logger_ -> traceProcessControl() >= 2)
	    {
                CORBA::String_var msg = CORBA::string_dup("kill: ");
                msg += rec -> record().server;
                msg += ": ";
		msg += ex._name();
                logger_ -> trace("IMR", msg);
            }
        }
    }
}

//
// IDL:ooc.com/IMR/Server/start:1.0
//
void
Server_impl::start()
    throw(IMR::ServerRunning, CORBA::SystemException)
{
    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    //
    // Is the server stopped?
    //
    if(rec -> record().status != STOPPED)
        throw IMR::ServerRunning();

    serverStarter_ -> add(rec -> record().server);
}

//
// IDL:ooc.com/IMR/Server/stop:1.0
//
void
Server_impl::stop()
    throw(IMR::OADNotRunning,
          IMR::ServerNotRunning,
	  CORBA::SystemException) 
{
    ServerRecord_var rec = getRecord();
    IMRDB_4_0_2::OADRecord oadrec;

    {
        JTCSynchronized sync(*rec.in());

        //
        // Is the server running.
        //
        if(rec -> record().status != RUNNING)
            throw IMR::ServerNotRunning();

        //
        // Is the OAD running
        //
        if(!oaddb_ -> queryRecordByHost(rec -> record().host, oadrec) ||
	   !oadrec.up)
            throw IMR::OADNotRunning();

	rec -> record().stopped = true;
        db_ -> updateRecord(rec);
    }

    oadrec.controller -> stop_process(rec -> record().server);
}

//
// IDL:ooc.com/IMR/Server/destroy:1.0
//
void
Server_impl::destroy()
    throw(IMR::ServerRunning, CORBA::SystemException) 
{
    if(!administrative_)
        throw CORBA::NO_PERMISSION();

    ServerRecord_var rec = getRecord();
    JTCSynchronized sync(*rec.in());

    //
    // Not permitted to destroy a running server
    //
    if(rec -> record().status != STOPPED)
        throw IMR::ServerRunning();

    if(!db_ -> destroyRecord(rec -> record().server))
	throw CORBA::OBJECT_NOT_EXIST();
}
