// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OCI_impl.h>
#include <CodeSetDatabase.h>
#include <CodeConverter.h>
#include <CodeSetUtil.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// FixedWidth2Reader public member implementations
// ----------------------------------------------------------------------

WChar
FixedWidth2Reader::read_wchar(OCI::BufferImpl*& in, ULong len)
    throw(DATA_CONVERSION)
{
    if(len == 2)
    {
	WChar c = *(in -> cur_++) << 8;
	c |= *(in -> cur_++);

	return c;
    }
    else
	throw DATA_CONVERSION();

    return 0;
}

// ----------------------------------------------------------------------
// FixedWidth2Writer public member implementations
// ----------------------------------------------------------------------

void
FixedWidth2Writer::write_wchar(OCI::BufferImpl*& out, WChar v)
    throw(DATA_CONVERSION)
{
    *(out -> cur_++) = (Octet)(v >> 8);
    *(out -> cur_++) = (Octet)v;
}

// ----------------------------------------------------------------------
// UTF8Reader private member implementations
// ----------------------------------------------------------------------

WChar
UTF8Reader::utf8ToUnicode(OCI::BufferImpl*& in)
    throw(DATA_CONVERSION)
{
    Octet first = *(in -> cur_++);
 
    //
    // Direct mapping for characters < 0x80
    //
    if((first & 0x80) == 0)
	return first;

    WChar value;

    if((first & 0xf8) == 0xc0)
    {
	// 5 free bits
	value = first & 0x1f;
    }
    else if((first & 0xf8) == 0xe0)
    {
	// 4 free bits
	value = first & 0x0f;

	if((*(in -> cur_) & 0xc0) != 0x80)
	    throw DATA_CONVERSION();

	value <<= 6;
	value |= *(in -> cur_++) & 0x3f;
    }
    //
    // 16 bit overflow
    //
    else
	throw DATA_CONVERSION();

    if((*(in -> cur_) & 0xc0) != 0x80)
	throw DATA_CONVERSION();

    value <<= 6;
    value |= *(in -> cur_++) & 0x3f;

    return value;
}

// ----------------------------------------------------------------------
// UTF8Reader public member implementations
// ----------------------------------------------------------------------

ULong
UTF8Reader::count_wchar(Octet first) const
{
    if((first & 0x80) == 0)
	return 1;
    else if((first & 0xf8) == 0xc0)
	return 2;
    else if((first & 0xf8) == 0xe0)
	return 3;

    throw DATA_CONVERSION();
}

// ----------------------------------------------------------------------
// UTF8Writer private member implementations
// ----------------------------------------------------------------------

void
UTF8Writer::unicodeToUtf8(OCI::BufferImpl*& out, WChar value)
    throw(DATA_CONVERSION)
{
    if(value < 0x80)
    {
	// Direct mapping (7 bits) for characters < 0x80

	*(out -> cur_++) = (Octet)value;
    }
    else if(value < 0x7ff)
    {
        // 5 free bits (%110xxxxx | %vvvvv)

	*(out -> cur_++) = (Octet)((value >> 6) | 0xc0);
	*(out -> cur_++) = (Octet)(value & 0x3f);
    }
    else if(value < 0xffff)
    {
        // 4 free bits (%1110xxxx | %vvvv)

	*(out -> cur_++) = (Octet)((value >> 12) | 0xe0);
	*(out -> cur_++) = (Octet)(((value >> 6) & 0x3f) | 0x80);
	*(out -> cur_++) = (Octet)((value & 0x3f) | 0x80);
    }
    else
	throw DATA_CONVERSION();
}

// ----------------------------------------------------------------------
// UTF8Writer public member implementations
// ----------------------------------------------------------------------

ULong
UTF8Writer::count_wchar(WChar value) const
{
    if(value < 0x80)
	return 1;
    else if(value < 0x7ff)
	return 2;
    else
	return 3;
}

// ----------------------------------------------------------------------
// CodeConverterBase constructor
// ----------------------------------------------------------------------

CodeConverterBase::CodeConverterBase(const CodeSetDatabase::CodeSetInfo* from,
				     const CodeSetDatabase::CodeSetInfo* to)
    : reader_(0), writer_(0)
{
    noneInstance_.description = "none";
    noneInstance_.rgy_value = 0;
    noneInstance_.char_values_size = 0;
    noneInstance_.char_values = 0;
    noneInstance_.max_bytes = 2;

    if(!from)
	from = &noneInstance_;
    if(!to)
	to = &noneInstance_;

    from_ = from;
    to_ = to;

    if(from -> rgy_value == CodeSetDatabase::UTF8)
	reader_ = new UTF8Reader();
    else if(from -> max_bytes <= 2)
	reader_ = new FixedWidth2Reader();
    else
    {
	//
	// ORBacus doesn't support characters larger than 16 bit
	//
	assert(false);
    }

    if(to -> rgy_value == CodeSetDatabase::UTF8)
	writer_ = new UTF8Writer();
    else if(to -> max_bytes <= 2)
	writer_ = new FixedWidth2Writer();
    else
    {
	//
	// ORBacus doesn't support characters larger than 16 bit
	//
	assert(false);
    }
}

CodeConverterBase::~CodeConverterBase()
{
    //
    // These instances have definitely been created
    //
    delete reader_;
    delete writer_;
}

// ----------------------------------------------------------------------
// CodeConverterFrom constructor
// ----------------------------------------------------------------------

CodeConverterFrom::CodeConverterFrom(
    const CodeSetDatabase::CodeSetInfo* from,
    const CodeSetDatabase::CodeSetInfo* to,
    const CodeSetDatabase::CharMapInfo* fromMap)
    : CodeConverterBase(from, to)
{
    fromMap_ = fromMap;
}

// ----------------------------------------------------------------------
// CodeConverterFrom public member implementations
// ----------------------------------------------------------------------

Char
CodeConverterFrom::convert(Char v)
{
    if(v <= (Char)fromMap_ -> upper_bound)
    {
	return (Char)fromMap_ -> map_values[v];
    }

    return v;
}

WChar
CodeConverterFrom::convert(WChar v)
{
    if(v <= (WChar)fromMap_ -> upper_bound)
    {
	return (WChar)fromMap_ -> map_values[v];
    }

    return v;
}

// ----------------------------------------------------------------------
// CodeConverterTo constructor
// ----------------------------------------------------------------------

CodeConverterTo::CodeConverterTo(
    const CodeSetDatabase::CodeSetInfo* from,
    const CodeSetDatabase::CodeSetInfo* to,
    const CodeSetDatabase::CharMapInfo* toMap)
    : CodeConverterBase(from, to)
{
    toMap_ = toMap;
}

// ----------------------------------------------------------------------
// CodeConverterTo public member implementations
// ----------------------------------------------------------------------

Char
CodeConverterTo::convert(Char v)
{
    for(ULong i = 0 ; i <= toMap_ -> upper_bound ; i++)
    {
	if(toMap_ -> map_values[i] == v)
	    return (Char)i;
    }

    return v;
}

WChar
CodeConverterTo::convert(WChar v)
{
    for(ULong i = 0 ; i <= toMap_ -> upper_bound ; i++)
    {
	if(toMap_ -> map_values[i] == v)
	    return (WChar)i;
    }

    return v;
}

// ----------------------------------------------------------------------
// CodeConverterBoth constructor
// ----------------------------------------------------------------------

CodeConverterBoth::CodeConverterBoth(
    const CodeSetDatabase::CodeSetInfo* from,
    const CodeSetDatabase::CodeSetInfo* to,
    const CodeSetDatabase::CharMapInfo* fromMap,
    const CodeSetDatabase::CharMapInfo* toMap)
    : CodeConverterBase(from, to)
{
    fromMap_ = fromMap;
    toMap_ = toMap;
}

// ----------------------------------------------------------------------
// CodeConverterBoth public member implementations
// ----------------------------------------------------------------------

Char
CodeConverterBoth::convert(Char v)
{
    if(v <= (Char)fromMap_ -> upper_bound)
    {
	v = (Char)fromMap_ -> map_values[v];
    }

    for(ULong i = 0 ; i <= toMap_ -> upper_bound ; i++)
    {
	if(toMap_ -> map_values[i] == v)
	    return (Char)i;
    }

    return v;
}

WChar
CodeConverterBoth::convert(WChar v)
{
    if(v <= (WChar)fromMap_ -> upper_bound)
    {
	v = (WChar)fromMap_ -> map_values[v];
    }

    for(ULong i = 0 ; i <= toMap_ -> upper_bound ; i++)
    {
	if(toMap_ -> map_values[i] == v)
	    return (WChar)i;
    }

    return v;
}

// ----------------------------------------------------------------------
// CodeConverterNone constructor
// ----------------------------------------------------------------------

CodeConverterNone::CodeConverterNone(const CodeSetDatabase::CodeSetInfo* from,
				     const CodeSetDatabase::CodeSetInfo* to)
    : CodeConverterBase(from, to)
{
}

// ----------------------------------------------------------------------
// CodeConverterNone public member implementations
// ----------------------------------------------------------------------

WChar
CodeConverterNone::read_wchar(OCI::BufferImpl*&, ULong)
	throw(DATA_CONVERSION)
{
    throw BAD_PARAM(MinorNoWcharCodeSet, COMPLETED_NO);
    return 0;
}

void
CodeConverterNone::write_wchar(OCI::BufferImpl*&, WChar)
	throw(DATA_CONVERSION)
{
    throw INV_OBJREF(MinorWcharCodeSetRequired, COMPLETED_NO);
}

ULong
CodeConverterNone::count_wchar(Octet) const
{
    throw BAD_PARAM(MinorNoWcharCodeSet, COMPLETED_NO);
    return 0;
}

ULong
CodeConverterNone::count_wchar(WChar) const
{
    throw INV_OBJREF(MinorWcharCodeSetRequired, COMPLETED_NO);
    return 0;
}

Char
CodeConverterNone::convert(Char)
{
    throw INV_OBJREF(MinorWcharCodeSetRequired, COMPLETED_NO);
    return 0;
}

WChar
CodeConverterNone::convert(WChar)
{
    throw INV_OBJREF(MinorWcharCodeSetRequired, COMPLETED_NO);
    return L'\0';
}

// ----------------------------------------------------------------------
// CodeConverterSimple constructor
// ----------------------------------------------------------------------

CodeConverterSimple::CodeConverterSimple(
    const CodeSetDatabase::CodeSetInfo* from,
    const CodeSetDatabase::CodeSetInfo* to)
    : CodeConverterBase(from, to)
{
}

// ----------------------------------------------------------------------
// CodeConverterNone public member implementations
// ----------------------------------------------------------------------

WChar
CodeConverterSimple::read_wchar(OCI::BufferImpl*&, ULong)
	throw(DATA_CONVERSION)
{
    assert(false);
    return 0;
}

void
CodeConverterSimple::write_wchar(OCI::BufferImpl*&, WChar)
	throw(DATA_CONVERSION)
{
    assert(false);
}

ULong
CodeConverterSimple::count_wchar(Octet) const
{
    assert(false);
    return 0;
}

ULong
CodeConverterSimple::count_wchar(WChar) const
{
    assert(false);
    return 0;
}

Char
CodeConverterSimple::convert(Char)
{
    assert(false);
    return 0;
}

WChar
CodeConverterSimple::convert(WChar)
{
    assert(false);
    return L'\0';
}
