// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_CODE_CONVERTER_H
#define OB_CODE_CONVERTER_H

#include <OB/Basic.h>

#include <OCI_impl.h>
#include <CodeSetDatabase.h>

namespace OB
{

class CodeSetReader
{

public:

    virtual CORBA::Char read_char(OCI::BufferImpl*&)
	throw(CORBA::DATA_CONVERSION) = 0;
    virtual CORBA::WChar read_wchar(OCI::BufferImpl*&, CORBA::ULong)
	throw(CORBA::DATA_CONVERSION) = 0;
    virtual CORBA::ULong count_wchar(CORBA::Octet) const = 0;
};

class CodeSetWriter
{

public:

    virtual void write_char(OCI::BufferImpl*&, CORBA::Char)
	throw(CORBA::DATA_CONVERSION) = 0;
    virtual void write_wchar(OCI::BufferImpl*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION) = 0;
    virtual CORBA::ULong count_wchar(CORBA::WChar) const = 0;
};

class FixedWidth2Reader : public CodeSetReader
{

public:

    FixedWidth2Reader() { };

    CORBA::Char read_char(OCI::BufferImpl*& in)
	throw(CORBA::DATA_CONVERSION)
	{ return *(in -> cur_++); }

    CORBA::WChar read_wchar(OCI::BufferImpl*&, CORBA::ULong)
	throw(CORBA::DATA_CONVERSION);

    CORBA::ULong count_wchar(CORBA::Octet) const { return 2; }
};

class FixedWidth2Writer : public CodeSetWriter
{

public:

    FixedWidth2Writer() { };

    void write_char(OCI::BufferImpl*& out, CORBA::Char v)
	throw(CORBA::DATA_CONVERSION)
	{ *(out -> cur_++) = (CORBA::Octet)v; }

    void write_wchar(OCI::BufferImpl*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION);

    CORBA::ULong count_wchar(CORBA::WChar) const { return 2; }
};

class UTF8Reader : public CodeSetReader
{
    static CORBA::WChar utf8ToUnicode(OCI::BufferImpl*&)
	throw(CORBA::DATA_CONVERSION);

public:

    UTF8Reader() { };
    virtual ~UTF8Reader() { };

    CORBA::Char read_char(OCI::BufferImpl*& in)
	throw(CORBA::DATA_CONVERSION)
	{ return (CORBA::Char)utf8ToUnicode(in); }

    CORBA::WChar read_wchar(OCI::BufferImpl*& in, CORBA::ULong)
	throw(CORBA::DATA_CONVERSION)
	{ return utf8ToUnicode(in); }

    CORBA::ULong count_wchar(CORBA::Octet) const;
};

class UTF8Writer : public CodeSetWriter
{
    static void unicodeToUtf8(OCI::BufferImpl*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION);

public:

    UTF8Writer() { };
    virtual ~UTF8Writer() { };

    void write_char(OCI::BufferImpl*& out, CORBA::Char v)
	throw(CORBA::DATA_CONVERSION)
	{ unicodeToUtf8(out, v); }

    void write_wchar(OCI::BufferImpl*& out, CORBA::WChar v)
	throw(CORBA::DATA_CONVERSION)
	{ unicodeToUtf8(out, v); }

    CORBA::ULong count_wchar(CORBA::WChar) const;
};

class CodeConverterBase
{

protected:

    //
    // Source and destination code set
    //
    const CodeSetDatabase::CodeSetInfo* from_;
    const CodeSetDatabase::CodeSetInfo* to_;

    //
    // The UTF-8 or fixed width reader/writer
    //
    CodeSetReader* reader_;
    CodeSetWriter* writer_;

    CodeSetDatabase::CodeSetInfo noneInstance_;

public:

    CodeConverterBase(const CodeSetDatabase::CodeSetInfo*,
		      const CodeSetDatabase::CodeSetInfo*);
    virtual ~CodeConverterBase();

    bool readerRequired() const
	{ return from_ -> rgy_value == CodeSetDatabase::UTF8; }
    bool writerRequired() const
	{ return to_ -> rgy_value == CodeSetDatabase::UTF8; }

    //
    // Get source/destination code set
    //
    const CodeSetDatabase::CodeSetInfo* getFrom() const { return from_; }
    const CodeSetDatabase::CodeSetInfo* getTo() const { return to_; }

    //
    // Conversion required?
    //
    virtual bool conversionRequired() const = 0;

    //
    // Convert characters
    //
    virtual CORBA::Char convert(CORBA::Char) = 0;
    virtual CORBA::WChar convert(CORBA::WChar) = 0;

    //
    // Read/write/count characters
    //
    virtual CORBA::Char read_char(OCI::BufferImpl*& in)
	throw(CORBA::DATA_CONVERSION)
	{ return reader_ -> read_char(in); }

    virtual CORBA::WChar read_wchar(OCI::BufferImpl*& in, CORBA::ULong len)
	throw(CORBA::DATA_CONVERSION)
	{ return reader_ -> read_wchar(in, len); }

    virtual void write_char(OCI::BufferImpl*& out, CORBA::Char v)
	throw(CORBA::DATA_CONVERSION)
	{ writer_ -> write_char(out, v); }

    virtual void write_wchar(OCI::BufferImpl*& out, CORBA::WChar v)
	throw(CORBA::DATA_CONVERSION)
	{ writer_ -> write_wchar(out, v); }

    virtual CORBA::ULong count_wchar(CORBA::Octet v) const
	{ return reader_ -> count_wchar(v); }

    virtual CORBA::ULong count_wchar(CORBA::WChar v) const
	{ return writer_ -> count_wchar(v); }
};

class CodeConverterFrom : public CodeConverterBase
{
    //
    // Unicode character mapping table
    //
    const CodeSetDatabase::CharMapInfo* fromMap_;

public:

    CodeConverterFrom(const CodeSetDatabase::CodeSetInfo*,
		      const CodeSetDatabase::CodeSetInfo*,
		      const CodeSetDatabase::CharMapInfo*);
    virtual ~CodeConverterFrom() { };

    bool conversionRequired() const { return true; };

    CORBA::Char convert(CORBA::Char);
    CORBA::WChar convert(CORBA::WChar);
};

class CodeConverterTo : public CodeConverterBase
{
    //
    // Unicode character mapping table
    //
    const CodeSetDatabase::CharMapInfo* toMap_;

public:

    CodeConverterTo(const CodeSetDatabase::CodeSetInfo*,
		    const CodeSetDatabase::CodeSetInfo*,
		    const CodeSetDatabase::CharMapInfo*);
    virtual ~CodeConverterTo() { };

    bool conversionRequired() const { return true; };

    CORBA::Char convert(CORBA::Char);
    CORBA::WChar convert(CORBA::WChar);
};

class CodeConverterBoth : public CodeConverterBase
{
    //
    // Unicode character mapping tables
    //
    const CodeSetDatabase::CharMapInfo* fromMap_;
    const CodeSetDatabase::CharMapInfo* toMap_;

public:

    CodeConverterBoth(const CodeSetDatabase::CodeSetInfo*,
		      const CodeSetDatabase::CodeSetInfo*,
		      const CodeSetDatabase::CharMapInfo*,
		      const CodeSetDatabase::CharMapInfo*);
    virtual ~CodeConverterBoth() { };

    bool conversionRequired() const { return true; };

    CORBA::Char convert(CORBA::Char);
    CORBA::WChar convert(CORBA::WChar);
};

class CodeConverterNone : public CodeConverterBase
{

public:

    CodeConverterNone(const CodeSetDatabase::CodeSetInfo*,
		      const CodeSetDatabase::CodeSetInfo*);
    virtual ~CodeConverterNone() { };

    bool conversionRequired() const { return true; };

    CORBA::WChar read_wchar(OCI::BufferImpl*&, CORBA::ULong)
	throw(CORBA::DATA_CONVERSION);
    void write_wchar(OCI::BufferImpl*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION);
    CORBA::ULong count_wchar(CORBA::Octet) const;
    CORBA::ULong count_wchar(CORBA::WChar) const;

    CORBA::Char convert(CORBA::Char v);
    CORBA::WChar convert(CORBA::WChar v);
};

class CodeConverterSimple : public CodeConverterBase
{

public:

    CodeConverterSimple(const CodeSetDatabase::CodeSetInfo*,
			const CodeSetDatabase::CodeSetInfo*);
    virtual ~CodeConverterSimple() { };

    bool conversionRequired() const { return false; };

    CORBA::WChar read_wchar(OCI::BufferImpl*&, CORBA::ULong)
	throw(CORBA::DATA_CONVERSION);
    void write_wchar(OCI::BufferImpl*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION);
    CORBA::ULong count_wchar(CORBA::Octet) const;
    CORBA::ULong count_wchar(CORBA::WChar) const;

    CORBA::Char convert(CORBA::Char v);
    CORBA::WChar convert(CORBA::WChar v);
};

} // End of namespace OB

#endif
