// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/Except.h>
#include <OB/ORBInstance.h>
#include <OB/ValueBase.h>
#include <OB/Any.h> // Needed for gcc
#include <OB/PIIOR.h>
#include <OB/PIManager.h>
#include <OB/Stream.h>
#include <OB/CONV_FRAME.h>

#include <CodeSetDatabase.h>
#include <CodeConverter.h>
#include <CodeSetUtil.h>

#include <stdlib.h>

using namespace CORBA;
using namespace OB;
using namespace PortableInterceptor;

FixSeq< ULong, int > CodeSetUtil::supportedCharCodeSets_;
FixSeq< ULong, int > CodeSetUtil::supportedWcharCodeSets_;

// ----------------------------------------------------------------------
// Add code set
// ----------------------------------------------------------------------

void
CodeSetUtil::addCharCodeSet(ULong id)
{
    CORBA::ULong supportedCharCodeSetsLength =
	supportedCharCodeSets_.length();
    supportedCharCodeSets_.length(supportedCharCodeSetsLength + 1);
    supportedCharCodeSets_[supportedCharCodeSetsLength] = id;
}

void
CodeSetUtil::addWcharCodeSet(ULong id)
{
    CORBA::ULong supportedWcharCodeSetsLength =
	supportedWcharCodeSets_.length();
    supportedWcharCodeSets_.length(supportedWcharCodeSetsLength + 1);
    supportedWcharCodeSets_[supportedWcharCodeSetsLength] = id;
}

CONV_FRAME::CodeSetComponent*
CodeSetUtil::createCodeSetComponent(ULong id, bool wChar)
{
    CodeSetDatabase::instance();

    CONV_FRAME::CodeSetComponent* codeSetComponent =
	new CONV_FRAME::CodeSetComponent();

    codeSetComponent -> native_code_set = id;

    //
    // Add conversion codesets, filter native codeset
    //
    if(wChar)
    {
	for(ULong i = 0 ; i < supportedWcharCodeSets_.length() ; i++)
	{
	    if(supportedWcharCodeSets_[i] != id) {
		OB::FixSeq<CONV_FRAME::CodeSetId,
		    CONV_FRAME::CodeSetComponent::
		    OBUnique_conversion_code_sets>&
		    code_sets = codeSetComponent -> conversion_code_sets;

		CORBA::ULong conversion_code_sets_length = code_sets.length();
		code_sets.length(conversion_code_sets_length + 1);
		code_sets[conversion_code_sets_length] =
		    supportedWcharCodeSets_[i];
	    }
	}
    }
    else
    {
	for(ULong i = 0 ; i < supportedCharCodeSets_.length() ; i++)
	{
	    if(supportedCharCodeSets_[i] != id) {
		OB::FixSeq<CONV_FRAME::CodeSetId,
		    CONV_FRAME::CodeSetComponent::
		    OBUnique_conversion_code_sets>&
		    code_sets = codeSetComponent -> conversion_code_sets;

		CORBA::ULong conversion_code_sets_length = code_sets.length();
		code_sets.length(conversion_code_sets_length + 1);
		code_sets[conversion_code_sets_length] =
		    supportedCharCodeSets_[i];
	    }
	}

    }

    return codeSetComponent;
}

// ----------------------------------------------------------------------
// Extract codeset information
// ----------------------------------------------------------------------

bool
CodeSetUtil::getCodeSetInfoFromComponents(
    ORBInstance_ptr orbInstance,
    const OCI::ProfileInfo& profileInfo,
    CONV_FRAME::CodeSetComponentInfo& info)
{
    //
    // Only IIOP 1.1 or newer has codeset information
    //
    if(profileInfo.major == 1 && profileInfo.minor > 0)
    {
	for(ULong i = 0 ; i < profileInfo.components.length() ; i++)
	{
	    if(checkForCodeSetInfo(profileInfo.components[i], info))
		return true;
	}
    }
    //
    // For IIOP 1.0 use proprietary mechanism (ISOLATIN1 and UCS-2),
    // if configured.
    //
    else
    {
	if(orbInstance -> extendedWchar())
	{
	    info.ForCharData.native_code_set = CodeSetDatabase::ISOLATIN1;
	    info.ForWcharData.native_code_set = CodeSetDatabase::UCS2;

	    return true;
	}
    }

    return false;
}

// ----------------------------------------------------------------------
// Get code converter from ProfileInfo and/or IOR
// ----------------------------------------------------------------------

void
CodeSetUtil::getCodeConverters(ORBInstance_ptr orbInstance,
			       const OCI::ProfileInfo& profileInfo,
			       OB::CodeConverters& conv)
{
    //
    // Set codeset defaults: ISO 8859-1 for char, no default for wchar
    // (13.7.2.4) but ORBacus uses default_wcs, which is initially 0
    //
    CONV_FRAME::CodeSetComponentInfo serverInfo;
    serverInfo.ForCharData.native_code_set = CodeSetDatabase::ISOLATIN1;
    serverInfo.ForWcharData.native_code_set =
	orbInstance -> getDefaultWcs();

    //
    // Set up code converters
    //
    ULong nativeCs = orbInstance -> getNativeCs();
    CONV_FRAME::CodeSetComponent_var client_cs =
	createCodeSetComponent(nativeCs, false);
    ULong tcs_c = CodeSetDatabase::ISOLATIN1;
    ULong nativeWcs = orbInstance -> getNativeWcs();
    CONV_FRAME::CodeSetComponent_var client_wcs =
	createCodeSetComponent(nativeWcs, true);
    ULong tcs_wc = orbInstance -> getDefaultWcs();

    CodeSetDatabase* db = CodeSetDatabase::instance();

    //
    // Other transmission codesets than the defaults can only be
    // determined if a codeset profile was present in the IOR.
    // The fallbacks in this case according to the specification
    // are UTF-8 (not ISOLATIN1!) and UTF-16.
    //

    if(getCodeSetInfoFromComponents(orbInstance, profileInfo, serverInfo))
    {
	tcs_c = db -> determineTCS(client_cs, serverInfo.ForCharData,
				   CodeSetDatabase::UTF8);
	tcs_wc = db -> determineTCS(client_wcs, serverInfo.ForWcharData,
				    CodeSetDatabase::UCS2);
    }

    conv.inputCharConverter = db -> getConverter(nativeCs, tcs_c);
    conv.outputCharConverter = db -> getConverter(tcs_c, nativeCs);
    conv.inputWcharConverter = db -> getConverter(nativeWcs, tcs_wc);
    conv.outputWcharConverter = db -> getConverter(tcs_wc, nativeWcs);
}

// ----------------------------------------------------------------------
// Check for code set information in a tagged component
// ----------------------------------------------------------------------

bool
CodeSetUtil::checkForCodeSetInfo(const IOP::TaggedComponent& comp,
                                 CONV_FRAME::CodeSetComponentInfo& info)
{
    if(comp.tag == IOP::TAG_CODE_SETS)
    {
	const Octet* coct = comp.component_data.get_buffer();
	ULong len = comp.component_data.length();
        OCI::BufferImpl buf((Octet*)coct, len);
	InputStreamImpl in(&buf, 0, false);
	in._OB_readEndian();
	CONV_FRAME::CodeSetComponentInfo::_OB_unmarshal(info, &in);
	return true;
    }

    return false;
}

// ----------------------------------------------------------------------
// Extract code set context from service context
// ----------------------------------------------------------------------

void
CodeSetUtil::extractCodeSetContext(const IOP::ServiceContext& context,
				   CONV_FRAME::CodeSetContext& ctx)
{
    const Octet* coct = context.context_data.get_buffer();
    ULong len = context.context_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    CONV_FRAME::CodeSetContext::_OB_unmarshal(ctx, &in);
}

// ----------------------------------------------------------------------
//  CodeSetIORInterceptor_impl constructor
// ----------------------------------------------------------------------

PortableInterceptor::
CodeSetIORInterceptor_impl::CodeSetIORInterceptor_impl(ULong nativeCs,
						       ULong nativeWcs)
    : nativeCs_(nativeCs), nativeWcs_(nativeWcs)
{
}

// ----------------------------------------------------------------------
//  CodeSetIORInterceptor_impl public member implementation
// ----------------------------------------------------------------------

char*
PortableInterceptor::CodeSetIORInterceptor_impl::name()
    throw(SystemException)
{
    return string_dup("");
}

void
PortableInterceptor::CodeSetIORInterceptor_impl::destroy()
    throw(SystemException)
{
}

void
PortableInterceptor::CodeSetIORInterceptor_impl::establish_components(
    IORInfo_ptr info)
    throw(SystemException)
{
    try
    {
	CONV_FRAME::CodeSetComponent_var c =
	    CodeSetUtil::createCodeSetComponent(nativeCs_, false);
	CONV_FRAME::CodeSetComponent_var wc =
	    CodeSetUtil::createCodeSetComponent(nativeWcs_, true);

	CONV_FRAME::CodeSetComponentInfo_var codeSetInfo =
	    new CONV_FRAME::CodeSetComponentInfo();
	codeSetInfo -> ForCharData = c;
	codeSetInfo -> ForWcharData = wc;

	IOP::TaggedComponent component;
	component.tag = IOP::TAG_CODE_SETS;

	//
	// The Codec could be used here -- but that means that the
	// Any insertion/extraction operators would have to be
	// generated unnecessarily
	//
	OCI::BufferImpl buf;
	OutputStreamImpl out(&buf);
	out._OB_writeEndian();
	codeSetInfo -> _OB_marshal(&out);

	component.component_data.length(out._OB_pos());
	memcpy(component.component_data.get_buffer(),
	       buf.data(), buf.length());
	info -> add_ior_component(component);
    }
    catch(const INV_POLICY&)
    {
	return;
    }
}

void
PortableInterceptor::CodeSetIORInterceptor_impl::components_established(
    IORInfo_ptr)
    throw(SystemException)
{
}

void
PortableInterceptor::CodeSetIORInterceptor_impl::
adapter_manager_state_change(PortableInterceptor::AdapterManagerId,
			     PortableInterceptor::AdapterState)
    throw(SystemException)
{
}
void
PortableInterceptor::CodeSetIORInterceptor_impl::adapter_state_change(
    const ObjectReferenceTemplateSeq&,
    AdapterState)
    throw(SystemException)
{
}
