// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/ORBInstance.h>
#include <OB/Properties.h>
#include <OB/Object.h>
#include <OB/Any.h>
#include <OB/Logger.h>

#include <Dispatch_impl.h>

// #include <pthread.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(ThreadPool_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(ThreadPool_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Definitions for the various built-in dispatch strategies
// ----------------------------------------------------------------------

namespace OB
{

//
// Invoke the request in the same thread
//
class DispatchSameThread_impl : public DispatchStrategy,
				public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy constructor and assignment operator
    //
    DispatchSameThread_impl(const DispatchSameThread_impl&);
    void operator=(const DispatchSameThread_impl&);

public:

    DispatchSameThread_impl() { }
    virtual ~DispatchSameThread_impl() { }

    //
    // Standard IDL to C++ Mapping
    //
    virtual OB::DispatchStrategyId
    id()
    {
	return SAME_THREAD;
    }
    
    virtual Any*
    info()
    {
	return new Any();
    }

    virtual void
    dispatch(DispatchRequest_ptr request)
    {
        //
        // Invoke the request
        //
        request -> invoke();
    }
};

//
// Invoke the request in a new thread
//
class DispatchThreadPerRequest_impl : public DispatchStrategy,
				      public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy constructor and assignment operator
    //
    DispatchThreadPerRequest_impl(const DispatchThreadPerRequest_impl&);
    void operator=(const DispatchThreadPerRequest_impl&);

    //
    // Helper class for dispatching a request
    //
    class Dispatcher : public JTCThread
    {
        DispatchRequest_var request_;

    public:
	
        Dispatcher(DispatchRequest_ptr request)
            : request_(DispatchRequest::_duplicate(request))
        {
        }

        virtual ~Dispatcher() { }

        virtual void run()
        {
            //
            // Invoke the request
            //
            request_ -> invoke();
        }
    };

public:

    DispatchThreadPerRequest_impl() { }
    virtual ~DispatchThreadPerRequest_impl() { }

    //
    // Standard IDL to C++ Mapping
    //
    virtual DispatchStrategyId
    id()
    {
	return THREAD_PER_REQUEST;
    }
    
    virtual Any*
    info()
    {
	return new Any();
    }

    virtual void
    dispatch(DispatchRequest_ptr request)
    {
	try
	{
	    JTCThreadHandle t = new Dispatcher(request);
	    t -> start();
	}
	catch(const JTCException&)
	{
	    throw CORBA::TRANSIENT();
	}
    }
};

//
// Invoke the request using a thread from a thread pool
//
class DispatchThreadPool_impl : public DispatchStrategy,
				public OBCORBA::RefCountLocalObject
{
    DispatchThreadPool_impl(const DispatchThreadPool_impl&);
    void operator=(const DispatchThreadPool_impl&);

    ThreadPoolId id_;
    ThreadPool_var pool_;

public:

    DispatchThreadPool_impl(ThreadPoolId id, ThreadPool_ptr pool)
	: id_(id), pool_(ThreadPool::_duplicate(pool))
    {
    }

    virtual ~DispatchThreadPool_impl() { }

    //
    // IDL:omg.org/PortableServer/DispatchStrategy/dispatch:1.0
    //
    virtual DispatchStrategyId
    id()
    {
	return THREAD_POOL;
    }
    
    virtual Any*
    info()
    {
	Any_var any = new Any();
	any <<= id_;
	return any._retn();
    }

    virtual void
    dispatch(DispatchRequest_ptr request)
    {
	pool_ -> add(request);
    }
};

//
// ThreadPoolDispatcher
//
class ThreadPoolDispatcher : public JTCThread
{
    ThreadPool_var threadPool_;
    
public:
    
    ThreadPoolDispatcher(JTCThreadGroupHandle& group,
                         ThreadPool_ptr threadPool)
	: JTCThread(group),
	  threadPool_(ThreadPool::_duplicate(threadPool))
    {
    }
    
    virtual void
    run()
    {
        while(true)
        {
            DispatchRequest_var req = threadPool_ -> get();
            
            //
            // ThreadPool has terminated
            //
            if(CORBA::is_nil(req))
                break;

            req -> invoke();
        }

	//
	// Release the reference
	//
	threadPool_ = ThreadPool::_nil();
    }
};

} // End of namespace OB

// ----------------------------------------------------------------------
// ThreadPool constructor/destructor
// ----------------------------------------------------------------------

ThreadPool::ThreadPool(ULong id, Long n)
    : destroy_(false), head_(0)
{
    //
    // Create a new thread group. Place each of the threads in the
    // pool in this new group
    //
    String_var name = string_dup("ThreadPool-");
    name += id;
    group_ = new JTCThreadGroup(name.in());

    //
    // Start each of the threads in the pool
    //
    for(Long i = 0 ; i < n ; i++)
    {
	JTCThreadHandle t = new ThreadPoolDispatcher(group_, this);
	t -> start();
    }
}

ThreadPool::~ThreadPool()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// ThreadPool public member implementation
// ----------------------------------------------------------------------

void
ThreadPool::destroy()
{
    {
        JTCSynchronized synchronized(*this);
	if(destroy_)
	    return;
        destroy_ = true;
        notifyAll();
    }

    //
    // Wait for all the threads in the pool to end
    //
    JTCSynchronized groupSync(*group_);
    while(group_ -> activeCount() > 0)
    {
	try
	{
	    group_ -> wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    //
    // Destroy the group
    //
    try
    {
	group_ -> destroy();
    }
    catch(const JTCIllegalThreadStateException&)
    {
	assert(false);
    }

    //
    // Clean out all pending requests
    //
    while(head_)
    {
        RequestRecord* next = head_ -> next;
        delete head_;
        head_ = next;
    }
}

void
ThreadPool::add(DispatchRequest_ptr request)
{
    JTCSynchronized synchronized(*this);

    //
    // If the thread pool has been destroyed then this is an
    // OBJ_ADAPTER error
    //
    if(destroy_)
	throw OBJ_ADAPTER();

    RequestRecord** p = &head_;
    while(*p)
        p = &((*p) -> next);

    *p = new RequestRecord;
    (*p) -> request = DispatchRequest::_duplicate(request);
    (*p) -> next = 0;

    notify();
}

DispatchRequest_ptr
ThreadPool::get()
{
    JTCSynchronized synchronized(*this);

    while(!destroy_ && !head_)
    {
        try
        {
            wait();
        }
        catch(const JTCInterruptedException&)
        {
        }
    }

    DispatchRequest_ptr result = DispatchRequest::_nil();
    if(!destroy_)
    {
	RequestRecord* rec = head_;
	head_ = head_ -> next;
	result = rec -> request._retn();
	delete rec;
    }
    return result;
}

// ----------------------------------------------------------------------
// DispatchStrategyFactory constructor/destructor
// ----------------------------------------------------------------------

DispatchStrategyFactory_impl::DispatchStrategyFactory_impl()
    : haveDefaultThreadPool_(false), destroy_(false)
{
}

DispatchStrategyFactory_impl::~DispatchStrategyFactory_impl()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// DispatchStrategyFactory public member implementation
// ----------------------------------------------------------------------

ThreadPoolId
DispatchStrategyFactory_impl::create_thread_pool(CORBA::Long nthreads)
    throw(CORBA::SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    //
    // Find the first empty thread pool
    //
    ULong i;
    for(i = 0 ; i < pools_.length() ; i++)
	if(is_nil(pools_[i]))
	   break;

    //
    // If there is no empty slot then append an empty slot
    //
    if(i >= pools_.length())
    {
	CORBA::ULong poolsLength = pools_.length();
	pools_.length(poolsLength + 1);
	pools_[poolsLength] = ThreadPool::_nil();
    }

    //
    // Allocate a new ThreadPool
    //
    pools_[i] = new ThreadPool(i, nthreads);

    return i;
}

void
DispatchStrategyFactory_impl::destroy_thread_pool(ThreadPoolId id)
    throw(InvalidThreadPool, CORBA::SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    if(id > pools_.length() || is_nil(pools_[id]))
	throw InvalidThreadPool();

    //
    // Destroy the ThreadPool
    //
    pools_[id] -> destroy();

    //
    // Empty the slot associated with this thread pool
    //
    pools_[id] = ThreadPool::_nil();
}

DispatchStrategy_ptr
DispatchStrategyFactory_impl::create_thread_pool_strategy(ThreadPoolId id)
    throw(InvalidThreadPool, CORBA::SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    if(id > pools_.length() || is_nil(pools_[id]))
	throw InvalidThreadPool();

    return new DispatchThreadPool_impl(id, pools_[id]);
}

DispatchStrategy_ptr
DispatchStrategyFactory_impl::create_same_thread_strategy()
    throw(CORBA::SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    return new DispatchSameThread_impl();
}

DispatchStrategy_ptr
DispatchStrategyFactory_impl::create_thread_per_request_strategy()
	throw(CORBA::SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    return new DispatchThreadPerRequest_impl();
}

DispatchStrategy_ptr
DispatchStrategyFactory_impl::create_default_dispatch_strategy()
	throw(CORBA::SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    //
    // Get the ORB properties
    //
    Properties_var properties = orbInstance_ -> getProperties();

    //
    // Set the dispatch strategy policy as specified by the
    // conc_model property
    //
    const char* value = properties -> getProperty("ooc.orb.oa.conc_model");
    if(value != 0)
    {
	if(strcmp(value, "reactive") == 0)
	{
	    return create_same_thread_strategy();
	}
	else if(strcmp(value, "threaded") == 0 ||
		strcmp(value, "thread_per_client") == 0)
	{
	    return create_same_thread_strategy();
	}
	else if(strcmp(value, "thread_per_request") == 0)
	{
	    return create_thread_per_request_strategy();
	}
	else if(strcmp(value, "thread_pool") == 0)
	{
	    //
	    // If there is no default thread pool yet then create one,
	    // with a default of 10 threads.
	    //
	    if(!haveDefaultThreadPool_)
	    {
		haveDefaultThreadPool_ = true;
		value = properties -> getProperty("ooc.orb.oa.thread_pool");
		ULong nthreads = (value != 0) ? atol(value) : 0;
		if(nthreads == 0)
		    nthreads = 10;
		defaultThreadPool_ = create_thread_pool(nthreads);
	    }
	    return create_thread_pool_strategy(defaultThreadPool_);
	}
	else
	{
	    String_var err =
		string_dup("ooc.orb.oa.conc_model: Unknown value `");
            err += value;
            err += "'";
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> warning(err);
	}
    }

    //
    // The default
    //
    return create_same_thread_strategy();
}

void
DispatchStrategyFactory_impl::_OB_setORBInstance(ORBInstance_ptr orbInstance)
{
    JTCSynchronized synchronized(*this);;
    orbInstance_ = ORBInstance::_duplicate(orbInstance);
}

void
DispatchStrategyFactory_impl::_OB_destroy()
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    destroy_ = true;

    orbInstance_ = ORBInstance::_nil();

    //
    // Destroy each of the thread pools
    //
    for(ULong i = 0 ; i < pools_.length() ; i++)
    {
	if(!is_nil(pools_[i]))
	{
	    pools_[i] -> destroy();
	    pools_[i] = ThreadPool::_nil();
	}
    }
}
