// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/StubImpl.h>
#include <OB/Downcall.h>
#include <OB/Client.h>
#include <OB/NamedValue.h>
#include <OB/DII.h>
#include <OB/PIClient.h>
#include <OB/PIORBInit.h>
#include <OB/PIManager.h>
#include <OB/ORBInstance.h>
#include <OB/Stream.h>
#include <OB/RefCounted.h>

#include <CodeConverter.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(Downcall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(Downcall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

/*
void
OBDuplicate(PIDowncall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(PIDowncall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(PIArgsDowncall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(PIArgsDowncall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(PIDIIDowncall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(PIDIIDowncall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}
*/

void
OBDuplicate(DowncallEmitter_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(DowncallEmitter_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Downcall private and protected members
// ----------------------------------------------------------------------

void
Downcall::checkForException()
    throw(ExceptionBase)
{
    switch(state_)
    {
    case DowncallStateUserException:
	if(ex_) // Only raise if a user exception has been set
	    ex_ -> _raise();
	break;

    case DowncallStateSystemException:
        assert(ex_);
	ex_ -> _raise();
	break;

    case DowncallStateFailureException:
	assert(ex_);
	//
	// COMPILERFIX HPUX 64
	//
	// was throw FailureException(*SystemException::_downcast(ex_));
	//
        throw FailureException(*(SystemException::_downcast(ex_)));

    case DowncallStateForward:
	assert(!is_nil(forwardIOR_));
	throw LocationForward(forwardIOR_, false);

    case DowncallStateForwardPerm:
	assert(!is_nil(forwardIOR_));
	throw LocationForward(forwardIOR_, true);

    default:
	break;
    }
}

// ----------------------------------------------------------------------
// Downcall constructor and destructor
// ----------------------------------------------------------------------

Downcall::Downcall(ORBInstance_ptr orbInstance,
		   Client_ptr client,
		   RefCountProfileInfo_ptr profileInfo,
		   RefCountPolicyList_ptr policies,
		   const char* op,
		   bool resp)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      client_(Client::_duplicate(client)),
      profileInfo_(RefCountProfileInfo::_duplicate(profileInfo)),
      policies_(RefCountPolicyList::_duplicate(policies)),
      op_(op),
      responseExpected_(resp),
      state_(DowncallStateUnsent),
      stateMonitor_(0),
      ex_(0)
{
    //
    // Get the next request ID
    //
    reqId_ = client_ -> requestId();
}

Downcall::~Downcall()
{
    delete ex_;
    delete stateMonitor_;
}

// ----------------------------------------------------------------------
// Downcall public member implementation
// ----------------------------------------------------------------------

OutputStream_ptr
Downcall::preMarshal()
    throw(ExceptionBase)
{
    emitter_ = client_ -> startDowncall(this, out_);
    checkForException();
    return out_;
}

void
Downcall::marshalEx(const SystemException& ex)
    throw(ExceptionBase)
{
    setFailureException(ex, COMPLETED_NO);
    checkForException();
    assert(false);
}

void
Downcall::postMarshal()
    throw(ExceptionBase)
{
}

void
Downcall::locate()
    throw(ExceptionBase)
{
    assert(responseExpected_);
    assert(strcmp(op_, "_locate") == 0);

    //
    // We could also use send() and receive() separately. But using
    // sendReceive() is more efficient.
    //
/*
    bool finished = emitter_ -> send(this, true);
    if(finished)
	checkForException();

    finished = emitter_ -> receive(this, true);
    assert(finished);
    checkForException();
*/

    bool finished = emitter_ -> sendReceive(this);
    assert(finished);
    checkForException();
}

void
Downcall::request()
    throw(ExceptionBase)
{
    assert(responseExpected_);

    //
    // We could also use send() and receive() separately. But using
    // sendReceive() is more efficient.
    //
/*
    bool finished = emitter_ -> send(this, true);
    if(finished)
	checkForException();

    finished = emitter_ -> receive(this, true);
    assert(finished);
    checkForException();
*/

    bool finished = emitter_ -> sendReceive(this);
    assert(finished);
    checkForException();
}

void
Downcall::oneway()
    throw(ExceptionBase)
{
    assert(!responseExpected_);

    if(policies_ -> syncScope == Messaging::SYNC_WITH_TRANSPORT)
    {
	bool finished = emitter_ -> send(this, true);
	assert(finished);
	checkForException();
    }
    else
    {
	bool finished = emitter_ -> send(this, false);
	if(finished)
	    checkForException();
    }
}

void
Downcall::deferred()
    throw(ExceptionBase)
{
    assert(responseExpected_);

    bool finished = emitter_ -> send(this, true);
    if(finished)
	checkForException();
}

void
Downcall::response()
    throw(ExceptionBase)
{
    assert(responseExpected_);

    bool finished = emitter_ -> receive(this, true);
    assert(finished);
    checkForException();
}

bool
Downcall::poll()
    throw(ExceptionBase)
{
    assert(responseExpected_);

    bool finished = emitter_ -> receive(this, false);
    if(finished)
    {
	checkForException();
	return state_ != DowncallStatePending;
    }
    else
	return false;
}

InputStream_ptr
Downcall::preUnmarshal()
    throw(ExceptionBase)
{
    return in_;
}

void
Downcall::unmarshalEx(const SystemException& ex)
    throw(ExceptionBase)
{
    setFailureException(ex, COMPLETED_MAYBE);
    checkForException();
    assert(false);
}

void
Downcall::postUnmarshal()
    throw(ExceptionBase)
{
    //
    // If the result of this downcall is a user exception, but no user
    // exception could be unmarshalled, then use the system exception
    // UNKNOWN
    //
    if(state_ == DowncallStateUserException && !ex_)
	setSystemException(UNKNOWN(MinorUnknownUserException, COMPLETED_YES));

    checkForException();
}

char*
Downcall::unmarshalExceptionId()
{
    assert(state_ == DowncallStateUserException);
    ULong pos = in_ -> _OB_pos();
    String_var id = in_ -> read_string();
    in_ -> _OB_pos(pos);
    return id._retn();
}
void
Downcall::setPending()
    throw()
{
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);

    assert(responseExpected_);
    state_ = DowncallStatePending;

    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
}

void
Downcall::setNoException(InputStream_ptr in)
    throw()
{
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);

    state_ = DowncallStateNoException;
    if(is_nil(in))
    {
	assert(!responseExpected_);
    }
    else
    {
	assert(responseExpected_);
	in_ = InputStream::_duplicate(in);
	in_ -> _OB_ORBInstance(orbInstance_);
	const CodeConverters* codeConverters = client_ -> codeConverters();
	in_ -> _OB_codeConverters(codeConverters,
				  (profileInfo_ -> value.major << 8) |
				  profileInfo_ -> value.minor);
    }

    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
}

void
Downcall::setUserException(InputStream_ptr in)
    throw()
{
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);

    assert(!is_nil(in));
    assert(responseExpected_);
    state_ = DowncallStateUserException;
    in_ = InputStream::_duplicate(in);
    in_ -> _OB_ORBInstance(orbInstance_);
    const CodeConverters* codeConverters = client_ -> codeConverters();
    in_ -> _OB_codeConverters(codeConverters,
                              (profileInfo_ -> value.major << 8) |
			      profileInfo_ -> value.minor);

    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
}

void
Downcall::setUserException(UserException* ex, const char*)
    throw()
{
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);

    assert(responseExpected_);
    assert(!ex_);
    state_ = DowncallStateUserException;
    ex_ = ex;

    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
}

void
Downcall::setUserException(const UserException& ex)
    throw()
{
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);

    assert(responseExpected_);
    assert(!ex_);
    state_ = DowncallStateUserException;
    ex_ = ex._OB_clone();

    if(stateMonitor_)
    {
	stateMonitor_ -> notify(); // TODO: Is notify() necessary here?
	delete sync;
    }
}

void
Downcall::setSystemException(const SystemException& ex)
    throw()
{
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);

    assert(responseExpected_);
    assert(!ex_);
    state_ = DowncallStateSystemException;
    ex_ = ex._OB_clone();

    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
}

void
Downcall::setFailureException(const SystemException& ex,
			      CompletionStatus completed)
    throw()
{
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);

    assert(!ex_);
    state_ = DowncallStateFailureException;
    ex_ = ex._OB_clone();

    if(strcmp(op_, "_locate") == 0)
    {
	//
	// It's safe to assume that locate requests never
	// complete. Also, since locate requests are used to validate
	// the connection with the server, it's necessary to always
	// specify COMPLETED_NO here so that the retry code tries
	// again to establish the connection in case we get a
	// COMM_FAILURE exception with a COMPLETED_MAYBE minor code
	// (something which is likely to happen when the ORB shutdowns
	// because of a race condition when the acceptor is closed).
	//
	(dynamic_cast<SystemException*>(ex_)) ->completed(COMPLETED_NO);
    }
    else
    {
	(dynamic_cast<SystemException*>(ex_)) ->completed(completed);
    }

    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
}

void
Downcall::setLocationForward(RefCountIOR_ptr ior, bool perm)
    throw()
{
    JTCSynchronized* sync = 0;
    if(stateMonitor_)
	sync = new JTCSynchronized(*stateMonitor_);

    assert(responseExpected_);
    assert(is_nil(forwardIOR_));
    if(perm)
	state_ = DowncallStateForwardPerm;
    else
	state_ = DowncallStateForward;
    forwardIOR_ = RefCountIOR::_duplicate(ior);

    if(stateMonitor_)
    {
	stateMonitor_ -> notify();
	delete sync;
    }
}

void
Downcall::initStateMonitor()
{
    assert(!stateMonitor_);
    stateMonitor_ = new JTCMonitor;
}

bool
Downcall::waitUntilCompleted(bool block)
{
    //
    // Get the timeout
    //
    Long t = policies_ -> requestTimeout;

    //
    // Yield if non-blocking or blocking with zero timeout
    //
    if(!block || (block && t == 0))
	JTCThread::yield();

    //
    // Wait for the desired state, taking the timeout and blocking
    // flag into account
    //
    assert(stateMonitor_);
    JTCSynchronized synchronized(*stateMonitor_);

    while(state_ == DowncallStateUnsent || state_ == DowncallStatePending)
    {
	if(!block)
	    return false;

	try
	{
	    if(t < 0)
	    {
		stateMonitor_ -> wait();
	    }
	    else
	    {
		DowncallState oldState = state_;

		stateMonitor_ -> wait(t);

		if(state_ == oldState)
		    throw NO_RESPONSE(0, COMPLETED_MAYBE);
	    }
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    //
    // The downcall has completed
    //
    return true;
}

// ----------------------------------------------------------------------
// PIDowncall private and protected members
// ----------------------------------------------------------------------

void
PIDowncall::checkForException()
    throw(ExceptionBase)
{
    //
    // If ex_ is set, but exId_ is not, then set it now
    //
    if(ex_ && !exId_.in())
	exId_ = ex_ -> _rep_id();

    switch(state_)
    {
    case DowncallStateUserException:
	//
	// Only invoke interceptor if a user exception has been
	// set
	//
	if(ex_)
	    piManager_ -> clientReceiveException(requestInfo_, false,
						 *ex_, exId_);
	break;
	    
    case DowncallStateSystemException:
	assert(ex_);
	piManager_ -> clientReceiveException(requestInfo_, true,
					     *ex_, exId_);
	break;
	    
    case DowncallStateFailureException:
	try
	{
	    assert(ex_);
	    piManager_ -> clientReceiveException(requestInfo_, true,
						 *ex_, exId_);
	}
	catch(const SystemException&)
	{
	    //
	    // Ignore any exception translations for failure
	    // exceptions
	    //
	}
	break;
	    
    case DowncallStateForward:
    case DowncallStateForwardPerm:
	assert(!is_nil(forwardIOR_));
	piManager_ -> clientReceiveLocationForward(requestInfo_,
						   forwardIOR_);
	break;

    default:
	break;
    }

    Downcall::checkForException();
}

// ----------------------------------------------------------------------
// PIDowncall constructor and destructor
// ----------------------------------------------------------------------

PIDowncall::PIDowncall(ORBInstance_ptr orbInstance,
		       Client_ptr client,
		       RefCountProfileInfo_ptr profileInfo,
		       RefCountPolicyList_ptr policies,
		       const char* op,
		       bool resp,
		       RefCountIOR_ptr IOR,
		       RefCountIOR_ptr origIOR,
		       /**/ PIManager_ptr piManager)
    : Downcall(orbInstance, client, profileInfo, policies, op, resp),
      IOR_(RefCountIOR::_duplicate(IOR)),
      origIOR_(RefCountIOR::_duplicate(origIOR)),
      piManager_(piManager)
{
}

PIDowncall::~PIDowncall()
{
}

// ----------------------------------------------------------------------
// PIDowncall public member implementation
// ----------------------------------------------------------------------

OutputStream_ptr
PIDowncall::preMarshal()
    throw(ExceptionBase)
{
    requestInfo_ = piManager_ ->
	clientSendRequest(op_,
			  responseExpected_,
			  IOR_,
			  origIOR_,
			  profileInfo_,
			  policies_,
			  requestSCL_,
			  replySCL_);

    return Downcall::preMarshal();
}

void
PIDowncall::postUnmarshal()
    throw(ExceptionBase)
{
    //
    // If the result of this downcall is a user exception, but no user
    // exception could be unmarshalled, then use the system exception
    // UNKNOWN, but keep the original exception ID
    //
    if(state_ == DowncallStateUserException && !ex_)
    {
	String_var id = unmarshalExceptionId();
	setSystemException(UNKNOWN(MinorUnknownUserException, COMPLETED_YES));
	exId_ = id._retn();
    }
    
    Downcall::postUnmarshal();
    
    if(responseExpected_)
	assert(state_ == DowncallStateNoException);
    else
	assert(state_ == DowncallStateUnsent ||
	       state_ == DowncallStateNoException);
    piManager_ -> clientReceiveReply(requestInfo_);
}

void
PIDowncall::setUserException(UserException* ex, const char* exId)
    throw()
{
    Downcall::setUserException(ex, exId);
    exId_ = exId;
}

// ----------------------------------------------------------------------
// PIArgsDowncall constructor and destructor
// ----------------------------------------------------------------------

PIArgsDowncall::PIArgsDowncall(ORBInstance_ptr orbInstance,
			       Client_ptr client,
			       RefCountProfileInfo_ptr profileInfo,
			       RefCountPolicyList_ptr policies,
			       const char* op,
			       bool resp,
			       RefCountIOR_ptr IOR,
			       RefCountIOR_ptr origIOR,
			       /**/ PIManager_ptr piManager,
			       ParameterDesc argDesc[],
			       ULong nargDesc,
			       ParameterDesc* retDesc,
			       TypeCode_ptr* exceptionTC,
			       ULong nexceptionTC)
    : PIDowncall(orbInstance, client, profileInfo, policies, op, resp,
		 IOR, origIOR, piManager),
      argDesc_(argDesc),
      nargDesc_(nargDesc),
      retDesc_(retDesc),
      exceptionTC_(exceptionTC),
      nexceptionTC_(nexceptionTC)
{
}

PIArgsDowncall::~PIArgsDowncall()
{
}

// ----------------------------------------------------------------------
// PIArgsDowncall public member implementation
// ----------------------------------------------------------------------

OutputStream_ptr
PIArgsDowncall::preMarshal()
    throw(ExceptionBase)
{
    requestInfo_ = piManager_ ->
	clientSendRequest(op_,
			  responseExpected_,
			  IOR_,
			  origIOR_,
			  profileInfo_,
			  policies_,
			  requestSCL_,
			  replySCL_,
			  argDesc_,
			  nargDesc_,
			  retDesc_,
			  exceptionTC_,
			  nexceptionTC_);

    return Downcall::preMarshal();
}

// ----------------------------------------------------------------------
// PIDIIDowncall constructor and destructor
// ----------------------------------------------------------------------

PIDIIDowncall::PIDIIDowncall(ORBInstance_ptr orbInstance,
			     Client_ptr client,
			     RefCountProfileInfo_ptr profileInfo,
			     RefCountPolicyList_ptr policies,
			     const char* op,
			     bool resp,
			     RefCountIOR_ptr IOR,
			     RefCountIOR_ptr origIOR,
			     /**/ PIManager_ptr piManager,
			     NVList_ptr args,
			     NamedValue_ptr result,
			     ExceptionList_ptr exceptions)
    : PIDowncall(orbInstance, client, profileInfo, policies, op, resp,
		 IOR, origIOR, piManager),
      args_(NVList::_duplicate(args)),
      result_(NamedValue::_duplicate(result)),
      exceptionList_(ExceptionList::_duplicate(exceptions))
{
}

PIDIIDowncall::~PIDIIDowncall()
{
}

// ----------------------------------------------------------------------
// PIDIIDowncall public member implementation
// ----------------------------------------------------------------------

OutputStream_ptr
PIDIIDowncall::preMarshal()
    throw(ExceptionBase)
{
    requestInfo_ = piManager_ ->
	clientSendRequest(op_,
			  responseExpected_,
			  IOR_,
			  origIOR_,
			  profileInfo_,
			  policies_,
			  requestSCL_,
			  replySCL_,
			  args_,
			  result_,
			  exceptionList_);
    
    return Downcall::preMarshal();
}
