// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Fixed.h>
#include <OB/AnyDeclaration.h>
#include <OB/Template.h>
#include <OB/TypeCodeConst.h>
#include <OB/Exceptions.h>
#include <OB/Except.h>
#include <OB/Any.h>

#include <stdio.h>
#include <stdlib.h>

using namespace CORBA;

//
// This class implements a fixed point type with 31 digits precision.
// CORBA doesn't support more, this class might. But since it is meant
// to provide only basic support for CORBA the precision was restricted.
//

void
Fixed::mul10(Fixed& tmp, ULong count)
{
    memmove(&tmp.value_[count], &tmp.value_, sizeof(tmp.value_) - count);
    memset(tmp.value_, 0, count);
}

void
Fixed::div10(Fixed& tmp, ULong count)
{
    memmove(&tmp.value_, &tmp.value_[count], sizeof(tmp.value_) - count);
    memset(tmp.value_ + sizeof(tmp.value_) - count, 0, count);
}

// ----------------------------------------------------------------------
// Fixed constructors and destructor
// ----------------------------------------------------------------------

Fixed::Fixed(Double val)
{
    char str[64];

    sprintf(str, "%15.15f", val);
    fromString(str);
}

Fixed::Fixed(LongDouble val)
{
#if SIZEOF_LONG_DOUBLE >= 12
    char str[64];

    sprintf(str, "%31.15Lf", val);

    fromString(str);
#else
    assert_nca(false, OB::NCANoLongDouble);
#endif
}

Fixed::Fixed(const Fixed& val)
{
    scale_ = val.scale_;
    signum_ = val.signum_;
    memcpy(value_, val.value_, sizeof(value_));
}

Fixed::Fixed(const Fixed& val, UShort scale)
{
    scale_ = scale;
    signum_ = val.signum_;
    memcpy(value_, val.value_, sizeof(value_));    
}

// ----------------------------------------------------------------------
// Fixed public member implementation
// ----------------------------------------------------------------------

Fixed
Fixed::round(UShort n) const
{
    if(n >= scale_)
	return *this;
    else
    {
	Fixed result = *this;
	Fixed f = (Long)5 * signum();;
	result += f.movePointLeft(n + 1);

	return result.setScale(n);
    }
}

Fixed
Fixed::truncate(UShort n) const
{
    if(n >= scale_)
	return *this;
    else
	return setScale(n);
}

UShort
Fixed::fixed_digits() const
{
    UShort digits = sizeof(value_);

    while(digits > 0 && !value_[digits - 1])
	digits--;

    return digits;
}

UShort
Fixed::fixed_scale() const
{
    return scale_;
}

char*
Fixed::to_string() const
{
    bool isZero = true;
    Long i;
    String_var s = string_dup("");

    //
    // Digits left from the decimal point.
    //
    i = fixed_digits() - 1;
    while(i >= scale_)
    {
	if(value_[i])
	    isZero = false;

	s += (char)(value_[i] + '0');
	i--;
    }

    if(scale_ > 0)
	s += '.';

    //
    // Digits right from the decimal point.
    //
    i = fixed_scale() - 1;
    while(i >= 0)
    {
	if(value_[i])
	    isZero = false;

	s += (char)(value_[i] + '0');
	i--;
    }

    String_var v;
    if(strlen(s))
    {
	if(signum_ == -1 && !isZero)
	    v = string_dup("-");
    }
    else
	s = string_dup("0");

    v += s;

    return v._retn();
}

// ----------------------------------------------------------------------
// Conversions from other types
// ----------------------------------------------------------------------

void
Fixed::fromLongLong(LongLong val)
{
    if(val >= 0)
	fromULongLong(val);
    else
	fromULongLong(-val, -1);
}

void
Fixed::fromULongLong(ULongLong val, Short signum)
{
    memset(value_, 0, sizeof(value_));
    scale_ = 0;
    signum_ = signum;

    Octet* v = value_;
    while(val)
    {
	*v = (Octet)(val % 10);
	v++;
	val /= 10;
    }
}

void
Fixed::fromString(const char* val)
{
    if(!strlen(val))
	throw DATA_CONVERSION();

    memset(value_, 0, sizeof(value_));
    scale_ = 0;
    signum_ = 1;

    //
    // This array takes the digits only temporarily
    //
    Octet dummy[sizeof(value_)];
    Octet* tmp = dummy + sizeof(value_);

    while(*val == ' ')
	val++;

    if(*val == '-')
    {
	signum_ = -1;
	val++;
    }
    else if(*val == '+')
    {
	val++;
    }

    bool fraction = false;

    while(*val)
    {
	if(*val == '.')
	{
	    fraction = true;
	}
	else if(*val >= '0' && *val <= '9')
	{
	    tmp--;
	    *tmp = *val - '0';

	    //
	    // This digit might be placed right from the decimal point
	    //
	    if(fraction)
		scale_++;
	}
	else if(*val == 'd' || *val == 'D')
	    break;
	else
	    throw DATA_CONVERSION();

	val++;
    }

    //
    // Check for 'd' or 'D' as the last character
    //
    if(strlen(val))
    {
	if(strcmp(val, "D") != 0 && strcmp(val, "d") != 0)
	    throw DATA_CONVERSION();
    }

    memcpy(value_, tmp, dummy + sizeof(value_) - tmp);
}

// ----------------------------------------------------------------------
// Helpers
// ----------------------------------------------------------------------

Fixed
Fixed::abs() const
{
    if(signum_ >= 0)
	return *this;
    else
	return -(*this);
}

Short
Fixed::signum() const
{
    ULong i;
    for(i = 0 ; i < sizeof(value_) ; i++)
    {
	if(value_[i])
	    return signum_;
    }

    return 0;
}

Fixed
Fixed::setScale(UShort scale) const
{
    Fixed result = *this;

    Short diff = scale - scale_;
    if(diff > 0)
    {
	memmove(result.value_ + diff, result.value_,
		sizeof(result.value_) - diff);
	memset(result.value_, 0, diff);
    }
    else if(diff < 0)
    {
	memmove(result.value_, result.value_ - diff,
		sizeof(result.value_) + diff);
	memset(result.value_ + sizeof(result.value_) + diff, 0, -diff);
    }

    result.scale_ = scale;

    return result;
}

Fixed
Fixed::movePointLeft(Short n) const
{
    Fixed result;

    if(n >= 0)
    {
	result = *this;
	result.scale_ += n;
    }
    else
	result = movePointRight(-n);

    return result;
}

Fixed
Fixed::movePointRight(Short n) const
{
    Fixed result;

    if(scale_ >= n)
    {
	result = *this;
	result.scale_ -= n;
    }
    else
    {
	result = setScale(n - scale_);
	result.scale_ = 0;
    }

    return result;
}

// ----------------------------------------------------------------------
// Operators
// ----------------------------------------------------------------------

Fixed::operator LongLong() const
{
    if(fixed_digits() - scale_ > 31)
	throw DATA_CONVERSION();

    Long i;
    LongLong result = 0;

    //
    // Digits left from the decimal point.
    //
    i = sizeof(value_) - 1;
    while(i >= scale_)
    {
	result *= 10;
	result += value_[i];
	i--;
    }

    if(signum_ >= 0)
	return result;
    else
	return -result;
}

Fixed::operator LongDouble() const
{
#if SIZEOF_LONG_DOUBLE >= 12

    Long i;
    LongDouble result = 0;

    //
    // Digits left from the decimal point.
    //
    i = sizeof(value_) - 1;
    while(i >= scale_)
    {
	result *= 10;
	result += value_[i];
	i--;
    }

    //
    // Digits right from the decimal point.
    //
    while(i >= 0)
    {
	result += (LongDouble)value_[i] / (10 * (i + 1));
	i--;
    }

    if(signum_ >= 0)
	return result;
    else
	return -result;

#else
    assert_nca(false, OB::NCANoLongDouble);
    return LongDouble(); // Some compilers need this
#endif
}

Fixed&
Fixed::operator=(const Fixed& val)
{
    scale_ = val.scale_;
    signum_ = val.signum_;
    memcpy(value_, val.value_, sizeof(value_));

    return *this;
}

Fixed&
Fixed::operator+=(const Fixed& val)
{
    *this = *this + val;

    return *this;
}

Fixed&
Fixed::operator-=(const Fixed& val)
{
    *this = *this - val;

    return *this;
}

Fixed&
Fixed::operator*=(const Fixed& val)
{
    *this = *this * val;

    return *this;
}

Fixed&
Fixed::operator/=(const Fixed& val)
{
    *this = *this / val;

    return *this;
}

Fixed&
Fixed::operator++()
{
    *this = *this + Fixed(Short(1));

    return *this;
}

Fixed
Fixed::operator++(int)
{
    *this = *this + Fixed(Short(1));

    return *this;
}

Fixed&
Fixed::operator--()
{
    *this = *this - Fixed(Short(1));

    return *this;
}

Fixed
Fixed::operator--(int)
{
    *this = *this - Fixed(Short(1));

    return *this;
}

Fixed
Fixed::operator+() const
{
    return *this;
}

Fixed
Fixed::operator-() const
{
    Fixed result = *this;
   
    result.signum_ *= -1;

    return result;
}

Boolean
Fixed::operator!() const
{
    ULong i;
    for(i = 0 ; i < sizeof(value_) - 1 ; i++)
	if(value_[i]) return false;

    return true;
}

Fixed
Fixed::opAdd(const Fixed& val1, const Fixed& val2)
{
    if(val1.signum_ == -1 && val2.signum_ == 1)
	return Fixed(val2) - (-val1);
    else if(val2.signum_ == -1 && val1.signum_ == 1)
	return val1 - (-val2);

    Fixed v1 = val1;
    Fixed v2 = val2;

    //
    // Adjust scales, i. e. decimal point
    //
    if(v1.scale_ < v2.scale_)
	v1 = v1.setScale(v2.scale_);
    else if(v2.scale_ < v1.scale_)
	v2 = v2.setScale(v1.scale_);

    //
    // Add digits
    //
    Octet carry = 0;
    ULong i;
    for(i = 0 ; i < sizeof(v2.value_) - 1 ; i++)
    {
	v1.value_[i] += v2.value_[i] + carry;
	if(v1.value_[i] >= 10)
	{
	    v1.value_[i] -= 10;
	    carry = 1;
	}
	else
	    carry = 0;
    }

    return v1;
}

Fixed
Fixed::opSub(const Fixed& val1, const Fixed& val2)
{
    if(val1.signum_ == -1 && val2.signum_ == -1)
	return -Fixed(val2) - (- val1);
    else if(val1.signum_ == -1 && val2.signum_ == 1)
	return -((-val1) + val2);
    else if(val1.signum_ == 1 && val2.signum_ == -1)
	return val1 + (-val2);

    Fixed v1;
    Fixed v2;
    bool neg;
    if(val1 > val2)
    {
	v1 = val1;
	v2 = val2;
	neg = false;
    }
    else
    {
	v2 = val1;
	v1 = val2;
	neg = true;
    }

    //
    // Adjust scales, i. e. decimal point
    //
    if(v1.scale_ < v2.scale_)
	v1 = v1.setScale(v2.scale_);
    else if(v2.scale_ < v1.scale_)
	v2 = v2.setScale(v1.scale_);

    //
    // Subtract digits
    //
    Octet carry = 0;
    ULong i;
    for(i = 0 ; i < sizeof(v1.value_) - 1 ; i++)
    {
	v1.value_[i] -= v2.value_[i] + carry;
	if(v1.value_[i] >= 10)
	{
	    v1.value_[i] += 10;
	    carry = 1;
	}
	else
	    carry = 0;
    }

    if(carry || neg)
	v1.signum_ = -1;

    return v1;
}

Fixed
Fixed::opMul(const Fixed& val1, const Fixed& val2)
{
    Fixed result;

    ULong i;
    for(i = 0 ; i < sizeof(val1.value_) / 2 - 1 ; i++)
    {
	//
	// Multiply absolute value by 10^i and add to result
	//
	if(val2.value_[i])
	{
	    Fixed tmp = val1.abs();
	    mul10(tmp, i);

	    Octet j;
	    for(j = 0 ; j < val2.value_[i] ; j++)
		result += tmp;
	}
    }

    result.signum_ = val1.signum_ * val2.signum_;
    result.scale_ = val1.scale_ + val2.scale_;

    return result;
}

Fixed
Fixed::opDiv(const Fixed& val1, const Fixed& val2)
{
    //
    // Division by 0?
    //
    assert(Fixed((Long)0) != val2);

    //
    // Rescale to largest scale
    //
    Fixed dividend, divisor;
    if(val2.scale_ > val1.scale_)
    {
        dividend = (val1.setScale(val2.scale_)).abs();
        divisor = val2.abs();
    }
    else
    {
        dividend = val1.abs();
        divisor = (val2.setScale(val1.scale_)).abs();
    }

    //
    // Multiply dividend by 10^32 to prevent overflows.
    //
    mul10(dividend, sizeof(val1.value_) / 2);

    Fixed v = divisor;
    ULong i = 0;
    while(v <= dividend)
    {
	mul10(v, 1);
	i++;
    }

    Fixed result = (Long)0;

    while(i)
    {
	Fixed tmp = divisor;
	mul10(tmp, i);

	Fixed v = (Long)1;
	mul10(v, i);

	while(tmp <= dividend)
	{
	    dividend -= tmp;
	    result += v;
	}

	i--;
    }

    //
    // Revert overflow adjustment
    //
    result.scale_ = 32;
    while(!result.value_[0] && result.scale_)
    {
	div10(result, 1);
	result.scale_--;
    }

    result.signum_ = val1.signum_ * val2.signum_;

    return result;
}

Boolean
Fixed::opGt(const Fixed& val1, const Fixed& val2)
{
    if(val1.signum_ != val2.signum_)
	return val1.signum_ > val2.signum_;

    Fixed v1 = val1;
    Fixed v2 = val2;

    //
    // Adjust scales, i. e. decimal point
    //
    if(v1.scale_ < v2.scale_)
	v1 = v1.setScale(v2.scale_);
    else if(v2.scale_ < v1.scale_)
	v2 = v2.setScale(v1.scale_);

    for(Long i = sizeof(v1.value_) - 1 ; i >= 0 ; i--)
    {
	if(v1.value_[i] > v2.value_[i])
	    return true;
	else if(v1.value_[i] < v2.value_[i])
	    return false;
    }

    return false;
}

// ----------------------------------------------------------------------
// Global operators
// ----------------------------------------------------------------------

OB_STD(istream&) operator>>(OB_STD(istream&) in, Fixed &val)
{
    char str[1024];
    in.get(str, 1023);
    assert(in.good());
    val = str;

    return in;
}

OB_STD(ostream&) operator<<(OB_STD(ostream&) out, const Fixed &val)
{
    String_var v = val.to_string();

    return out << v;
}

Fixed operator+(const Fixed& val1, const Fixed& val2)
{
    return Fixed::opAdd(val1, val2);
}

Fixed operator-(const Fixed& val1, const Fixed& val2)
{
    return Fixed::opSub(val1, val2);
}

Fixed operator*(const Fixed& val1, const Fixed& val2)
{
    return Fixed::opMul(val1, val2);
}

Fixed operator/(const Fixed& val1, const Fixed& val2)
{
    return Fixed::opDiv(val1, val2);
}

Boolean operator>(const Fixed& val1, const Fixed& val2)
{
    return Fixed::opGt(val1, val2);
}

Boolean operator<(const Fixed& val1, const Fixed& val2)
{
    return val2 > val1;
}

Boolean operator>=(const Fixed& val1, const Fixed& val2)
{
    return !(val1 < val2);
}

Boolean operator<=(const Fixed& val1, const Fixed& val2)
{
    return !(val1 > val2);
}

Boolean operator==(const Fixed& val1, const Fixed& val2)
{
    return !(val1 - val2);
}

Boolean operator!=(const Fixed& val1, const Fixed& val2)
{
    return !(val1 == val2);
}
