// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/StubImpl.h>
#include <OB/Stream.h>
#include <OB/OCI.h>
#include <OB/RefCounted.h>
#include <OB/ORBInstance.h>
#include <OB/Logger.h>
#include <OB/CoreTraceLevels.h>
#include <OB/Properties.h>
#include <OB/NativeTypes.h>
#include <OB/Reactor.h>

#include <GIOPClient.h>
#include <GIOPClientWorker.h>

#ifdef HAVE_SYS_TIMEB_H
#   include <sys/timeb.h>
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

/*
void
OBDuplicate(OB::GIOPClientWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}
*/

void
OBDuplicate(OB::GIOPClientWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// GIOPClientWorker::UnsentMessage public member implementations
// ----------------------------------------------------------------------

GIOPClientWorker::UnsentMessage::UnsentMessage(OCI::Buffer_ptr b)
{
    buf = OCI::Buffer::_duplicate(b);
    buf -> pos(0);
}

GIOPClientWorker::UnsentMessage::UnsentMessage(Downcall_ptr d)
{
    down = Downcall::_duplicate(d);
    buf = OCI::Buffer::_duplicate(d -> output_nodup() -> _OB_buffer_nodup());
    buf -> pos(0);
}

// ----------------------------------------------------------------------
// GIOPClientWorker private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorker::logCloseConnection()
{
    if(!closeLogged_)
    {
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceConnections())
	{
	    OCI::TransportInfo_var info = transport_ -> get_info();
	    String_var msg = string_dup("closing connection\n");
	    String_var desc = info -> describe();
	    msg += desc;
//          msg += '\n';
//  	    String_var exMsg = exception_ -> _to_string();
//  	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("outgoing", msg);
	}
	
	closeLogged_ = true;
    }
}

void
GIOPClientWorker::add(OCI::Buffer_ptr buf)
{
    //
    // If we are in StateError or StateClosed, don't add.
    //
    if(state_ == StateError || state_ == StateClosed)
	return;

    //
    // Add new message to the message buffers
    //
    unsent_.push_back(UnsentMessage(buf));
}

void
GIOPClientWorker::add(Downcall_ptr down)
{
    //
    // If a failure exception is set, don't add new downcall
    // objects. Just set the exception in the downcall.
    //
    if(exception_)
    {
	down -> setFailureException(*exception_, COMPLETED_NO);
	return;
    }

    //
    // If we are in StateClosing, StateError, or StateClosed, then why
    // is no exception set?
    //
    assert(state_ != StateClosing && state_ != StateError &&
	   state_ != StateClosed);

    //
    // Add the message header
    //
    try
    {
	//
	// Save stream position and rewind
	//
	OutputStream_ptr out = down -> output_nodup();
	ULong pos = out -> _OB_pos();
	out -> _OB_pos(0);
	
	//
	// Write message header
	//
	const OCI::ProfileInfo& profileInfo =
	    down -> profileInfo_nodup() -> value;
	GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);

	//
	// Write protocol header
	//
	const char* op = down -> operation();
	if(*op == '_' && strcmp(op, "_locate") == 0)
	    outgoing.writeMessageHeader(GIOP::LocateRequest, false, pos - 12);
	else
	    outgoing.writeMessageHeader(GIOP::Request, false, pos - 12);
	
	//
	// Restore stream position
	//
	out -> _OB_pos(pos);
    }
    catch(const SystemException& ex)
    {
	down -> setFailureException(ex, COMPLETED_NO);
	return;
    }

    //
    // Update the GIOP minor version
    //
    Octet minor = down -> profileInfo_nodup() -> value.minor;
    if(minor > minor_)
	minor_ = minor;

    //
    // Add new message to the message buffers
    //
    unsent_.push_back(UnsentMessage(down));
}

Downcall_ptr
GIOPClientWorker::moveFirstUnsentToPending()
{
    //
    // If there are no unsent messages, then there must be a failure
    // exception, otherwise it's a bug.
    //
    if(unsent_.empty())
    {
	assert(exception_);
	return Downcall::_nil();
    }

    //
    // Remove first downcall
    //
    Downcall_var down = unsent_.front().down._retn();
    unsent_.pop_front();
    
    //
    // Only add to pending if a response is expected
    //
    if(!is_nil(down))
    {
	if(down -> responseExpected())
	{
	    down -> setPending();
	    pending_.push_front(down);
	}
	else
	{
	    down -> setNoException(InputStream::_nil());
	}
    }

    //
    // Do not set messageSent_ if operation was a LocateRequest since
    // LocateRequest does not contain ServiceContext information
    //
    if(!messageSent_)
    {
	if(strcmp(down -> operation(), "_locate") != 0)
	{
	    messageSent_ = true;
	}
    }  

    return down._retn();
}

//
// Remove pending message and put it back in the unsent list
//
void
GIOPClientWorker::movePendingToUnsent(Downcall_ptr down)
{
    if(down -> responseExpected())
    {
	Downcall_var d = findAndRemovePending(down -> requestId());
	assert(is_nil(d) || d.in() == down);
	unsent_.push_back(UnsentMessage(down));
    }
}

Downcall_ptr
GIOPClientWorker::findAndRemovePending(ULong reqId)
{
    OB_STL::list<Downcall_var>::iterator iter;
    for(iter = pending_.begin() ; iter != pending_.end() ; ++iter)
    {
        if((*iter) -> requestId() == reqId)
	{
	    Downcall_ptr downcall = (*iter)._retn();
	    pending_.erase(iter);
	    return downcall;
	}
    }

    return Downcall::_nil();
}

bool
GIOPClientWorker::exception(State state,
			    const CORBA::SystemException& ex,
			    bool notCompleted)
{
    //
    // Don't do anything if there is no state change
    //
    if(state_ == state)
	return false;

    //
    // It is not possible to transition backwards
    //
    if(state < state_)
	return false;

    //
    // Set the failure exception
    //
    if(!exception_)
    {
	//
	// Note: The completion status of ex is not taken into account
	// here, since the completion status of an exception raised by
	// low-level operations (like OCI send and receive methods) is
	// not applicable as the completion status of the downcall as
	// a whole. Therefore the completion status is passed
	// explicitly with the notCompleted parameter.
	//
	
	//
	// Save the exception, with a completion status of COMPLETED_NO
	//
	exception_ = SystemException::_downcast(ex._OB_clone());
	exception_ -> completed(COMPLETED_NO);
	
	//
	// Always use a completion status of NO for unsent requests
	//
	while(!unsent_.empty())
	{
	    if(!is_nil(unsent_.front().down))
		unsent_.front().down -> setFailureException(*exception_,
							    COMPLETED_NO);
	    
	    unsent_.pop_front();
	}
	
	//
	// If notCompleted is true, use a completion status of
	// COMPLETED_NO for pending requests, otherwise use
	// COMPLETED_MAYBE
	//
	if(notCompleted)
	{
	    while(!pending_.empty())
	    {
		pending_.front() -> setFailureException(*exception_,
							COMPLETED_NO);
		pending_.pop_front();
	    }
	}
	else
	{
	    while(!pending_.empty())
	    {
		pending_.front() -> setFailureException(*exception_,
							COMPLETED_MAYBE);
		pending_.pop_front();
	    }
	}
    }

    switch(state)
    {
    case StateActive:

	//
	// There is nothing special to do
	//
	break;

    case StateClosing:

	//
	// Detach the worker from the client
	//
	client_ -> removeWorker(this);

        //
	// Do a graceful shutdown after state change below...
	//

	break;

    case StateError:

	//
	// Detach the worker from the client
	//
	client_ -> removeWorker(this);

        //
	// Do an abortive shutdown
	//
	abortiveShutdown();

	break;

    case StateClosed:

	//
	// Detach the worker from the client
	//
	client_ -> removeWorker(this);

	//
	// Close the transport
	//
	logCloseConnection();
	transport_ -> close();

	break;
    }

    //
    // Update the state
    //
    state_ = state;

    //
    // If StateClosing, shutdown gracefully. This must be called after
    // the state has been changed since calling gracefulShutdown() may 
    // cause a state change to StateClosed and we do not want to reset it
    // to StateClosing if that occurs.
    //
    if(state_ == StateClosing)
        gracefulShutdown();

    return true;
}

void
GIOPClientWorker::execute()
    throw()
{
    assert(state_ == StateActive || state_ == StateClosing);

    bool complete;
    try
    {
	//
	// Provide the buffer to the incoming message and reset buf_
	// to allow for reading the next message. A return value of
	// false indicates that a message is not available (i.e., a
	// fragment was received).
	//
        complete = incoming_.consumeBuffer(buf_._retn());
    }
    catch(const SystemException& ex)
    {
	exception(StateError, ex);
	return;
    }

    //
    // Keep track of the highest GIOP minor version seen
    //
    Octet minor = incoming_.version().minor;
    if(minor > minor_)
	minor_ = minor;

    //
    // If the message is not complete yet, there is nothing to do
    //
    if(!complete)
	return;

    //
    // OK, message is complete - let's execute it
    //
    switch(incoming_.type())
    {
    case GIOP::Reply:
    {
	ULong reqId;
	GIOP::ReplyStatusType_1_2 status;
	IOP::ServiceContextList scl;
	
	try
	{
	    incoming_.readReplyHeader(reqId, status, scl);
	}
	catch(const SystemException& ex)
	{
	    exception(StateError, ex);
	    return;
	}
	
	Downcall_var down = findAndRemovePending(reqId);
	if(is_nil(down))
	{ 
	    //
	    // Request id is unknown
	    //
	    exception(StateError,
		      COMM_FAILURE(MinorUnknownReqId, COMPLETED_MAYBE));
	    return;
	}
	
	down -> setReplySCL(scl);
	InputStream_var in = incoming_.input();
	
	switch(status)
	{
	case GIOP::NO_EXCEPTION:
	    down -> setNoException(in);
	    break;
	    
	case GIOP::USER_EXCEPTION:
	    down -> setUserException(in);
	    break;
	    
	case GIOP::SYSTEM_EXCEPTION:
	{
	    try
	    {
		SystemException* ex = SystemException::_OB_unmarshal(in);
		ex -> remote(true);
		down -> setSystemException(*ex);
		delete ex;
	    }
	    catch(const SystemException& ex)
	    {
		exception(StateError, ex);
	    }

	    break;
	}
	
	case GIOP::LOCATION_FORWARD:
	{
	    try
	    {
		IOP::IOR ior;
		IOP::IOR::_OB_unmarshal(ior, in);
		RefCountIOR_var refIOR = new RefCountIOR(ior);
		down -> setLocationForward(refIOR, false);
	    }
	    catch(const SystemException& ex)
	    {
		exception(StateError, ex);
	    }

	    break;
	}
	
	case GIOP::LOCATION_FORWARD_PERM:
	{
	    try
	    {
		IOP::IOR ior;
		IOP::IOR::_OB_unmarshal(ior, in);
		RefCountIOR_var refIOR = new RefCountIOR(ior);
		down -> setLocationForward(refIOR, true);
	    }
	    catch(const SystemException& ex)
	    {
		exception(StateError, ex);
	    }

	    break;
	}
	
	case GIOP::NEEDS_ADDRESSING_MODE:
	    // TODO: implement
	    exception(StateError, NO_IMPLEMENT());
	    break;
	    
	default:
	    //
	    // Unknown message
	    //
	    exception(StateError,
		      COMM_FAILURE(MinorUnknownReplyMessage, COMPLETED_MAYBE));
	    break;
	}

	break;
    }
	
    case GIOP::Request:
    case GIOP::LocateRequest:
    case GIOP::CancelRequest:
	
	//
	// These messages may not be sent to a client
	//
	exception(StateError,
		  COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
	break;
	
    case GIOP::LocateReply:
    {
	ULong reqId;
	GIOP::LocateStatusType_1_2 status;
	try
	{
	    incoming_.readLocateReplyHeader(reqId, status);
	}
	catch(const SystemException& ex)
	{
	    exception(StateError, ex);
	    return;
	}

	Downcall_var down = findAndRemovePending(reqId);
	if(is_nil(down))
	{
	    //
	    // Request id is unknown
	    //
	    exception(StateError,
		      COMM_FAILURE(MinorUnknownReqId, COMPLETED_MAYBE));
	    return;
	}

	//
	// Was this a LocateRequest?
	//
	const char* op = down -> operation();
	if(strcmp(op, "_locate") != 0)
	{
	    exception(StateError,
		      COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
	    return;
	}
	InputStream_var in = incoming_.input();

	switch(status)
	{
	case GIOP::UNKNOWN_OBJECT:
	    down -> setSystemException(OBJECT_NOT_EXIST());
	    break;

	case GIOP::OBJECT_HERE:
	    down -> setNoException(in);
	    break;

	case GIOP::OBJECT_FORWARD:
            try
            {
                IOP::IOR ior;
                IOP::IOR::_OB_unmarshal(ior, in);
                RefCountIOR_var refIOR = new RefCountIOR(ior);
                down -> setLocationForward(refIOR, false);
            }
            catch(const SystemException& ex)
            {
                Logger_var logger = orbInstance_ -> getLogger();
                logger -> warning("An error occurred while reading a "
                                   "locate reply, possibly indicating\n"
                                   "an interoperability problem. You may "
                                   "need to set the LocateRequestPolicy\n"
                                   "to false.");
                down -> setSystemException(ex);
                exception(StateError, ex);
            }

	    break;

	case GIOP::OBJECT_FORWARD_PERM:
            try
            {
                IOP::IOR ior;
                IOP::IOR::_OB_unmarshal(ior, in);
                RefCountIOR_var refIOR = new RefCountIOR(ior);
                down -> setLocationForward(refIOR, true);
            }
            catch(const SystemException& ex)
            {
                Logger_var logger = orbInstance_ -> getLogger();
                logger -> warning("An error occurred while reading a "
                                   "locate reply, possibly indicating\n"
                                   "an interoperability problem. You may "
                                   "need to set the LocateRequestPolicy\n"
                                   "to false.");
                down -> setSystemException(ex);
                exception(StateError, ex);
            }

	    break;

	case GIOP::LOC_SYSTEM_EXCEPTION:
            try
            {
                SystemException* ex = SystemException::_OB_unmarshal(in);
                ex -> remote(true);
                down -> setSystemException(*ex);
                delete ex;
            }
            catch(const SystemException& ex)
            {
                down -> setSystemException(ex);
                exception(StateError, ex);
            }

	    break;

	case GIOP::LOC_NEEDS_ADDRESSING_MODE:
	    // TODO: implement
	    exception(StateError, NO_IMPLEMENT());
	    break;
	}

	break;
    }
	    
    case GIOP::CloseConnection:

	//
	// If the peer closes the connection, all outstanding requests
	// can safely be reissued. Thus we send all of them a
	// TRANSIENT exception with a completion status of
	// COMPLETED_NO. This is done by calling exception() with the
	// notCompleted parameter set to true.
	//
	exception(StateClosed,
		  TRANSIENT(MinorCloseConnection, COMPLETED_MAYBE), true);
	break;
	
    case GIOP::MessageError:
	
	//
	// How should we react to MessageError messages? We just send
	// a MessageError back, and close the connection.
	//
	exception(StateError,
		  COMM_FAILURE(MinorMessageError, COMPLETED_MAYBE));
	break;
	
    case GIOP::Fragment:
	
	//
        // At this point there should be no fragments, only complete
        // messages
	//
	assert(false);
	break;

    default:

	//
	// Unknown message
	//
	exception(StateError,
		  COMM_FAILURE(MinorUnknownMessage, COMPLETED_MAYBE));
	break;
    }
}

// ----------------------------------------------------------------------
// GIOPClientWorker constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorker::GIOPClientWorker(ORBInstance_ptr orbInstance,
				   GIOPClient_ptr client,
				   OCI::Transport_ptr transport)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      client_(GIOPClient::_duplicate(client)),
      transport_(OCI::Transport::_duplicate(transport)),
      incoming_(orbInstance),
      messageSent_(false),
      minor_(0),
      closeLogged_(false),
      state_(StateActive),
      exception_(0)
{
    //
    // Cache properties
    //
    Properties_var properties = orbInstance_ -> getProperties();
    const char* value;

    value = properties -> getProperty("ooc.orb.client_shutdown_timeout");
    if(value)
	shutdownTimeout_ = atoi(value);
    else
	shutdownTimeout_ = 2;

    //
    // Trace new outgoing connection
    //
    CoreTraceLevels_var coreTraceLevels = orbInstance_ -> getCoreTraceLevels();
    if(coreTraceLevels -> traceConnections())
    {
	OCI::TransportInfo_var info = transport_ -> get_info();
	String_var msg = string_dup("new connection\n");
	String_var desc = info -> describe();
	msg += desc;
	Logger_var logger = orbInstance_ -> getLogger();
	logger -> trace("outgoing", msg);
    }
}

GIOPClientWorker::~GIOPClientWorker()
{
    assert(state_ == StateClosed);
    delete exception_;
}

// ----------------------------------------------------------------------
// GIOPClientWorker public member implementations
// ----------------------------------------------------------------------

OCI::Transport_ptr
GIOPClientWorker::transport()
{
    return OCI::Transport::_duplicate(transport_);
}

// ----------------------------------------------------------------------
// GIOPClientWorkerReactive private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerReactive::updateReactor()
{
    Mask mask = 0;

    if((state_ == StateActive || state_ == StateClosing) &&
        transport_ -> mode() != OCI::SendOnly)
        mask |= EventRead;

    if(state_ != StateClosed && !unsent_.empty())
	mask |= EventWrite;

    if(mask_ != mask)
    {
	mask_ = mask;

	if(mask_)
	{
	    OCI::Handle handle = transport_ -> handle();
	    assert(handle >= 0);
	    reactor_ -> register_handler(this, mask_, TypeClient, handle);
	}
	else
	{
	    reactor_ -> unregister_handler(this);
	}
    }
}

void
GIOPClientWorkerReactive::add(OCI::Buffer_ptr buf)
{
    GIOPClientWorker::add(buf);
    updateReactor();
    assert(unsent_.empty() || mask_ & EventWrite);
}

void
GIOPClientWorkerReactive::add(Downcall_ptr down)
{
    GIOPClientWorker::add(down);
    updateReactor();
    assert(unsent_.empty() || mask_ & EventWrite);
}

bool
GIOPClientWorkerReactive::exception(State state,
				    const CORBA::SystemException& ex,
				    bool notCompleted)
{
    //
    // Let the base class change the state
    //
    if(!GIOPClientWorker::exception(state, ex, notCompleted))
	return false;

    //
    // Update reactor registration
    //
    updateReactor();
    return true;
}

void
GIOPClientWorkerReactive::gracefulShutdown()
{
    if(minor_ >= 2) // GIOP 1.0 and 1.1 clients don't send CloseConnection
    {
	OCI::Buffer_var buf = new OCI::Buffer(12);
	OutputStream_var out = new OutputStream(buf);
	
	OCI::ProfileInfo profileInfo;
	profileInfo.major = 1;
	profileInfo.minor = minor_;
	
	GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
	outgoing.writeMessageHeader(GIOP::CloseConnection, false, 0);
	
	add(buf);
    }

    //
    // If we are in StateClosing and all messages have been sent, we
    // gracefully shutdown the connection.
    //
    if(unsent_.empty())
    {
	logCloseConnection();
	transport_ -> shutdown();

        if(transport_ -> mode() == OCI::SendReceive)
        {
            //
            // TODO: The following code is temporary. This must be
            // changed when we have reactor timers back.
            //
            /* Begin temporary code */
            //
            // If we have a shutdown timeout, we receive dummy data, until
            // we get an exception.
            //
            if(shutdownTimeout_ > 0)
            {
                try
                {
                    while(true)
                    {
                        if(is_nil(buf_))
                            buf_ = new OCI::Buffer(12);
                        
                        if(buf_ -> pos() < 12)
                        {
                            transport_ -> receive_timeout(buf_,
                                                          shutdownTimeout_ *
                                                          1000);
                            if(!buf_ -> is_full())
                                throw TRANSIENT(MinorForcedShutdown,
                                                COMPLETED_MAYBE);
                        
                            incoming_.extractHeader(buf_);
                            buf_ -> realloc(12 + incoming_.size());
                        }
                        
                        if(!buf_ -> is_full())
                        {
                            transport_ -> receive_timeout(buf_,
                                                          shutdownTimeout_ *
                                                          1000);
                            if(!buf_ -> is_full())
                                throw TRANSIENT(MinorForcedShutdown,
                                                COMPLETED_MAYBE);
                        }
                        
                        execute();
                    }
                }
                catch(const SystemException& ex)
                {
                    exception(StateClosed, ex);
                }
            }
            /* End temporary code */
        }
        else
            exception(StateClosed,
                      TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
    }
}

void
GIOPClientWorkerReactive::abortiveShutdown()
{
    OCI::Buffer_var buf = new OCI::Buffer(12);
    OutputStream_var out = new OutputStream(buf);

    OCI::ProfileInfo profileInfo;
    profileInfo.major = 1;
    profileInfo.minor = minor_;

    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
    outgoing.writeMessageHeader(GIOP::MessageError, false, 0);

    add(buf);
}

bool
GIOPClientWorkerReactive::receiveNoSync(Downcall_ptr down, bool block)
    throw()
{
    //
    // Get the timeout
    //
    Long t = down -> policies_nodup() -> requestTimeout;

    //
    // Try to receive the reply
    //
    try
    {
	while(down -> unsent() || down -> pending())
	{
	    if(block)
	    {
		if(t < 0)
		{
		    //
		    // Dispatch one event, blocking
		    //
		    reactor_ -> dispatch_one_event(-1);
		}
		else
		{
		    //
		    // Dispatch one event, with timeout
		    //
		    if(!reactor_ -> dispatch_one_event(t))
			throw NO_RESPONSE(); // Timeout
		}
	    }
	    else
	    {
		//
		// Dispatch one event, with zero timeout
		//
		if(!reactor_ -> dispatch_one_event(0))
		    return false;
	    }
	}
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return true;
    }

    return true;
}

// ----------------------------------------------------------------------
// GIOPClientWorkerReactive constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorkerReactive::GIOPClientWorkerReactive(
    ORBInstance_ptr orbInstance,
    GIOPClient_ptr client,
    OCI::Transport_ptr transport)
    : GIOPClientWorker(orbInstance, client, transport),
      reactor_(orbInstance -> getReactor()),
      mask_(0)
{
    updateReactor();
}

GIOPClientWorkerReactive::~GIOPClientWorkerReactive()
{
    assert(mask_ == 0);
}

// ----------------------------------------------------------------------
// GIOPClientWorkerReactive public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerReactive::destroy()
{
    JTCSynchronized synchronized(*this);

    exception(StateClosing, INITIALIZE(MinorORBDestroyed, COMPLETED_NO));
}

bool
GIOPClientWorkerReactive::send(Downcall_ptr down, bool block)
    throw()
{
    JTCSynchronized synchronized(*this);

    //
    // Add the downcall
    //
    add(down);
    
    //
    // Get the timeout
    //
    Long t = down -> policies_nodup() -> requestTimeout;

    //
    // Try to send the request
    //
    try
    {
	while(down -> unsent())
	{
	    if(block)
	    {
		if(t < 0)
		{
		    //
		    // Dispatch one event, blocking
		    //
		    reactor_ -> dispatch_one_event(-1);
		}
		else
		{
		    //
		    // Dispatch one event, with timeout
		    //
		    if(!reactor_ -> dispatch_one_event(t))
			throw NO_RESPONSE(); // Timeout
		}
	    }
	    else
	    {
		//
		// Dispatch one event, with zero timeout
		//
		if(!reactor_ -> dispatch_one_event(0))
		    return false;
	    }
	}
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return true;
    }

    if(down -> responseExpected())
	return false;
    else
	return true;
}

bool
GIOPClientWorkerReactive::receive(Downcall_ptr down, bool block)
    throw()
{
    JTCSynchronized synchronized(*this);

    return receiveNoSync(down, block);
}

bool
GIOPClientWorkerReactive::sendReceive(Downcall_ptr down)
    throw()
{
    JTCSynchronized synchronized(*this);

    //
    // We do not need to call send(). Instead, we just add the
    // downcall here. This is faster, and the result of the downcall
    // will be picked up by receiveNoSync() anyway (including any
    // failures).
    //
    add(down);

    return receiveNoSync(down, true);
}

Mask
GIOPClientWorkerReactive::handleEvent(Mask mask)
{
    JTCSynchronized synchronized(*this);
    
    Mask retMask = 0;

    //
    // Handle write events only if we are not in StateClosed
    //
    if((mask & EventWrite) && state_ != StateClosed && !unsent_.empty())
    {
	OCI::Buffer_ptr buf = unsent_.front().buf;

	//
	// Send buffer, non-blocking
	//
	try
	{
	    retMask |= EventWrite;
	    transport_ -> send(buf, false);
	}
	catch(const SystemException& ex)
	{
	    exception(StateClosed, ex);
	    return retMask;
	}
	
	if(buf -> is_full())
	{
	    //
	    // Message was sent completely
	    //
	    Downcall_var down = moveFirstUnsentToPending();
	}
	
	if(unsent_.empty() && state_ == StateClosing)
	{
	    //
	    // If all messages have been sent and we are in
	    // StateClosing, we gracefully shutdown the connection.
	    //
	    logCloseConnection();
	    transport_ -> shutdown();

            if(transport_ -> mode() == OCI::SendReceive)
            {
                //
                // TODO: The following code is temporary. This must be
                // changed when we have reactor timers back.
                //
                /* Begin temporary code */
                //
                // If we have a shutdown timeout, we receive dummy data,
                // until we get an exception.
                //
                if(shutdownTimeout_ > 0)
                {
                    try
                    {
                        while(true)
                        {
                            if(is_nil(buf_))
                                buf_ = new OCI::Buffer(12);
                            
                            if(buf_ -> pos() < 12)
                            {
                                retMask |= EventRead;
                                transport_ -> receive_timeout(
                                    buf_, shutdownTimeout_ * 1000);
                                if(!buf_ -> is_full())
                                {
                                    exception(StateClosed,
                                              TRANSIENT(MinorForcedShutdown,
                                                        COMPLETED_MAYBE));
                                    return retMask;
                                }
                                
                                incoming_.extractHeader(buf_);
                                buf_ -> realloc(12 + incoming_.size());
                            }
                            
                            if(!buf_ -> is_full())
                            {
                                retMask |= EventRead;
                                transport_ -> receive_timeout(
                                    buf_, shutdownTimeout_ * 1000);
                                if(!buf_ -> is_full())
                                {
                                    exception(StateClosed,
                                              TRANSIENT(MinorForcedShutdown,
                                                        COMPLETED_MAYBE));
                                    return retMask;
                                }
                            }
                            
                            execute();
                        }
                    }
                    catch(const SystemException& ex)
                    {
                        exception(StateClosed, ex);
                        return retMask;
                    }
                }
                /* End temporary code */
            }
            else
                exception(StateClosed,
                          TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
	}

	if(unsent_.empty() && state_ == StateError)
	{
	    //
	    // If all messages have been sent and we are in
	    // StateError, we don't go through all the hula hoop with
	    // continuing to receive messages until the peer
	    // closes. Instead, we just close the connection, meaning
	    // that we can't be 100% sure that the peer gets the last
	    // message.
	    //
	    exception(StateClosed,
		      TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
	}
    }

    //
    // Handle read events only if we are in StateActive or
    // StateClosing. We must continue to read if we are in
    // StateClosing to make sure that no messages can get lost upon
    // close, and to make sure that CloseConnection messages from the
    // peer are processed.
    //
    if((mask & EventRead) &&
       (state_ == StateActive || state_ == StateClosing))
    {
	//
	// Setup the incoming message buffer if not already done
	//
	if(is_nil(buf_))
	    buf_ = new OCI::Buffer(12);
	
	if(buf_ -> pos() < 12)
	{
	    //
	    // Receive header, non-blocking
	    //
	    try
	    {
		retMask |= EventRead;
		transport_ -> receive(buf_, false);
	    }
	    catch(const SystemException& ex)
	    {
		exception(StateClosed, ex);
		return retMask;
	    }
	    
	    //
	    // Header complete?
	    //
	    if(buf_ -> is_full())
	    {
		try
		{
		    incoming_.extractHeader(buf_);
		    buf_ -> realloc(12 + incoming_.size());
		}
		catch(const SystemException& ex)
		{
		    exception(StateError, ex);
		    return retMask;
		}
	    }
	}
	
	if(buf_ -> pos() >= 12)
	{
	    if(!buf_ -> is_full())
	    {
		//
		// Receive body, non-blocking
		//
		try
		{
		    retMask |= EventRead;
		    transport_ -> receive(buf_, false);
		}
		catch(const SystemException& ex)
		{
		    exception(StateClosed, ex);
		    return retMask;
		}
	    }
	    
	    //
	    // Execute message if buffer is full
	    //
	    if(buf_ -> is_full())
		execute();
	}
    }

    //
    // Update reactor registration
    //
    updateReactor();

    return retMask;
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::updateTimestamp()
{
    //
    // Only update if a connection timeout is set
    //
    if(connectionTimeout_ > 0)
    {
#ifdef WIN32
	struct _timeb tb;
	_ftime(&tb);
	timestamp_ = tb.time;
#else
	timeval tv;
	gettimeofday(&tv, 0);
	timestamp_ = tv.tv_sec;
#endif
    }
}

void
GIOPClientWorkerThreaded::add(OCI::Buffer_ptr buf)
{
    updateTimestamp();
    GIOPClientWorker::add(buf);
}

void
GIOPClientWorkerThreaded::add(Downcall_ptr down)
{
    if(down -> responseExpected())
	down -> initStateMonitor();
    updateTimestamp();
    GIOPClientWorker::add(down);
}

void
GIOPClientWorkerThreaded::gracefulShutdown()
{
    try
    {
	//
	// Retrieve the thread group for the clients
	//
	JTCThreadGroupHandle group = orbInstance_ -> getClientWorkerGroup();
	
	//
	// Start the shutdown thread, which will take care of proper
	// connection termination
	//
	shutdownThread_ = new ShutdownThread(group, this);
	shutdownThread_ -> start();
    }
    catch(const JTCException&)
    {	
	exception(StateClosed, IMP_LIMIT(MinorThreadLimit, COMPLETED_NO));
	return;
    }
}

void
GIOPClientWorkerThreaded::abortiveShutdown()
{    
    //
    // Send a MessageError message
    //
    try
    {
	OCI::Buffer_var buf = new OCI::Buffer(12);
	OutputStream_var out = new OutputStream(buf);
	
	OCI::ProfileInfo profileInfo;
	profileInfo.major = 1;
	profileInfo.minor = minor_;
	
	GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
	outgoing.writeMessageHeader(GIOP::MessageError, false, 0);
	
	out -> _OB_buffer_nodup() -> pos(0);

	transport_ -> send(out -> _OB_buffer_nodup(), true);
	assert(out -> _OB_buffer_nodup() -> is_full());
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return;
    }
    
    //
    // If we are in StateError, we don't go through all the hula hoop
    // with continuing to receive messages until the peer
    // closes. Instead, we just close the connection, meaning that we
    // can't be 100% sure that the peer gets the last message.
    //
    exception(StateClosed,
	      TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorkerThreaded::GIOPClientWorkerThreaded(
    ORBInstance_ptr orbInstance, 
    GIOPClient_ptr client,
    OCI::Transport_ptr transport)
    : GIOPClientWorker(orbInstance, client, transport)
{
    //
    // Cache properties
    //
    Properties_var properties = orbInstance_ -> getProperties();
    const char* value;

    value = properties -> getProperty("ooc.orb.client_timeout");
    if(value)
    {
	connectionTimeout_ = atoi(value);
	updateTimestamp();
    }
    else
    {
	connectionTimeout_ = 0;
	timestamp_ = 0;
    }

    //
    // Only start a receiver thread if the transport supports
    // twoway invocations
    //
    if(transport_ -> mode() == OCI::SendReceive)
    {
        try
        {
            //
            // Retrieve the thread group for the clients
            //
            JTCThreadGroupHandle group =
                orbInstance_ -> getClientWorkerGroup();

            //
            // Start receiver thread
            //
            receiverThread_ = new ReceiverThread(group, this);
            receiverThread_ -> start();
        }
        catch(const JTCException&)
        {
            transport_ -> close();
            state_ = StateClosed;
            _OB_setRef(0);
            throw IMP_LIMIT(MinorThreadLimit, COMPLETED_NO);
        }
    }
}

GIOPClientWorkerThreaded::~GIOPClientWorkerThreaded()
{
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::destroy()
{
    JTCSynchronized synchronized(*this);

    exception(StateClosing, INITIALIZE(MinorORBDestroyed, COMPLETED_NO));
}

bool
GIOPClientWorkerThreaded::send(Downcall_ptr down, bool block)
    throw()
{
    if(block)
    {
	//
	// The following code is carrefully written in order to not
	// lock the worker state when calling the send() method in
	// blocking mode. We first get a message to send locking the
	// worker state (we assume that the message will successfully
	// be sent) and then we send the message without locking the
	// worker state.
	//

	bool downcallAdded = false;
	Downcall_var downcallToSend;

	//
	// Get the timeout
	//
	Long t = down -> policies_nodup() -> requestTimeout;

	try
	{
	    do
	    {
		//
		// Get a message to send from the unsent queue (and
		// eventually add the downcall if not already done. We
		// add the downcall here to avoid locking the worker
		// state twice).
		//
		OCI::Buffer_ptr buf;
		{
		    JTCSynchronized synchronized(*this);
		    
		    if(!downcallAdded)
		    {
			add(down);
			if(!down -> unsent())
			    return (!down -> responseExpected()); 
			downcallAdded = true;
		    }
		    
		    buf = unsent_.front().buf.ptr();
		    downcallToSend = moveFirstUnsentToPending();
		}
		
		//
		// Send the message
		//
		{
		    JTCSynchronized synchronized(sendMutex_);
		    if(t < 0)
		    {
			//
			// Send buffer, blocking
			//
			transport_ -> send(buf, true);
			assert(buf -> is_full());
		    }
		    else
		    {
			//
			// Send buffer, with timeout
			//
			transport_ -> send_timeout(buf, (ULong)t);
			if(!buf -> is_full())
			    throw NO_RESPONSE(); // Timeout
		    }
		}
	    }
	    while(down -> unsent());
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    if(!is_nil(downcallToSend))
	    {
		//
		// Move back the message to the unsent queue since we
		// failed to send it (this is mainly to have an
		// accurate minor code for set failure exception set
		// on the downcall).
		//
		movePendingToUnsent(downcallToSend);
	    }
	    exception(StateClosed, ex);
	    return true;
	}
    }
    else // Non blocking
    {
	JTCSynchronized synchronized(*this);

	add(down);

	try
	{
	    while(down -> unsent())
	    {
		OCI::Buffer_ptr buf = unsent_.front().buf;	    
		
		//
		// Send buffer, non-blocking
		//
		{
		    JTCSynchronized synchronized(sendMutex_);
		    
		    transport_ -> send(buf, false);
		    if(!buf -> is_full())
			return false;
		}
		
		//
		// Message was sent completely
		//
		moveFirstUnsentToPending();
	    }
	}
	catch(const SystemException& ex)
	{
	    exception(StateClosed, ex);
	    return true;
	}
    }

    if(down -> responseExpected())
	return false;
    else
	return true;
}

bool
GIOPClientWorkerThreaded::receive(Downcall_ptr down, bool block)
    throw()
{
    //
    // Try to receive the reply
    //
    try
    {
	return down -> waitUntilCompleted(block);
    }
    catch(const SystemException& ex)
    {
	JTCSynchronized synchronized(*this);
	exception(StateClosed, ex);
	return true;
    }
}

bool
GIOPClientWorkerThreaded::sendReceive(Downcall_ptr down)
    throw()
{
    if(send(down, true))
	return true;

    return receive(down, true);
}

void
GIOPClientWorkerThreaded::receiverRun()
{
    while(true)
    {
	//
	// Setup the incoming message buffer
	//
	assert(is_nil(buf_));
	buf_ = new OCI::Buffer(12);
	
	//
	// Receive header, blocking
	//
	try
	{
	    transport_ -> receive(buf_, true);
	    assert(buf_ -> is_full());
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateClosed, ex);
	    return;
	}
	    
	//
	// Header is complete
	//
	try
	{
	    incoming_.extractHeader(buf_);
	    buf_ -> realloc(12 + incoming_.size());
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateError, ex);
	    return;
	}
	
	if(!buf_ -> is_full())
	{
	    //
	    // Receive body, blocking
	    //
	    try
	    {
		transport_ -> receive(buf_, true);
		assert(buf_ -> is_full());
	    }
	    catch(const SystemException& ex)
	    {
		JTCSynchronized synchronized(*this);
		exception(StateClosed, ex);
		return;
	    }
	}
	    
	{
	    JTCSynchronized synchronized(*this);

	    //
	    // If we are not in StateActive or StateClosing, stop this
	    // thread. We do *not* stop this thread if we are in
	    // StateClosing, since we must continue to read data from
	    // the Transport to make sure that no messages can get
	    // lost upon close, and to make sure that CloseConnection
	    // messages from the peer are processed.
	    //
	    if(state_ != StateActive && state_ != StateClosing)
		return;

	    //
	    // Execute the message
	    //
	    updateTimestamp();
	    execute();
	}
    }
}

void
GIOPClientWorkerThreaded::shutdownRun()
{
    if(minor_ >= 2) // GIOP 1.0 and 1.1 clients don't send CloseConnection
    {
	//
	// Send a CloseConnection message
	//
	try
	{
	    OCI::Buffer_var buf = new OCI::Buffer(12);
	    OutputStream_var out = new OutputStream(buf);
	    
	    OCI::ProfileInfo profileInfo;
	    profileInfo.major = 1;
	    profileInfo.minor = minor_;
	    
	    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
	    outgoing.writeMessageHeader(GIOP::CloseConnection, false, 0);
	    
	    out -> _OB_buffer_nodup() -> pos(0);
	    
	    transport_ -> send(out -> _OB_buffer_nodup(), true);
	    assert(out -> _OB_buffer_nodup() -> is_full());
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateClosed, ex);
	    return;
	}
    }

    //
    // Now we gracefully shutdown the connection. Synchronization
    // is necessary to avoid a race with the receiver thread.
    //
    {
        JTCSynchronized synchronized(*this);
        if(state_ == StateClosing)
        {
            logCloseConnection();
            transport_ -> shutdown();
        }
    }

    if(transport_ -> mode() == OCI::SendReceive)
    {
        while(true)
        {
            try
            {
                if(shutdownTimeout_ > 0)
                {
                    //
                    // Join with the receiver thread, with timeout
                    //
                    receiverThread_ -> join(shutdownTimeout_ * 1000);
                }
                else
                {
                    //
                    // Join with the receiver thread, no timeout
                    //
                    receiverThread_ -> join();
                }
                
                break;
            }
            catch(const JTCInterruptedException&)
            {
                continue;
            }
        }
    }
    
    //
    // We now close the connection actively, since it may still be
    // open under certain circumstances. (For example, if the receiver
    // thread has not terminated yet, or if the receiver thread sets
    // the state to StateError before termination.)
    //
    {
	JTCSynchronized synchronized(*this);
	exception(StateClosed,
		  TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
    }
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded::ReceiverThread public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::ReceiverThread::run()
{
    try
    {
	worker_ -> receiverRun();
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	assert(false);
    }
    catch(const JTCException& ex)
    {
	String_var err = string_dup(ex.getType());
	err += ": ";
	err += ex.getMessage();
	OB_CERR << err << OB_ENDL;
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    //
    // Break cyclic object dependency
    //
    worker_ = GIOPClientWorkerThreaded::_nil();
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded::ShutdownThread public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::ShutdownThread::run()
{
    try
    {
	worker_ -> shutdownRun();
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	assert(false);
    }
    catch(const JTCException& ex)
    {
	String_var err = string_dup(ex.getType());
	err += ": ";
	err += ex.getMessage();
	OB_CERR << err << OB_ENDL;
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    //
    // Break cyclic object dependency
    //
    worker_ = GIOPClientWorkerThreaded::_nil();
}
