// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_CLIENT_WORKER_H
#define OB_GIOP_CLIENT_WORKER_H

#include <OB/OCIClient_fwd.h>

#include <GIOPClient_fwd.h>
#include <GIOPClientWorker_fwd.h>

#include <OB/Downcall.h>
#include <OB/EventHandlerBase.h>

#include <GIOPMessage.h>

namespace OB
{

//
// The GIOPClientWorker class
//
class GIOPClientWorker : public DowncallEmitter
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPClientWorker(const GIOPClientWorker&);
    void operator=(const GIOPClientWorker&);

protected:

    ORBInstance_var orbInstance_; // The ORBInstance
    GIOPClient_var client_; // The client
    OCI::Transport_var transport_; // The transport
    OCI::Buffer_var buf_; // The incoming message buffer
    GIOPIncomingMessage incoming_; // The incoming message
    bool messageSent_; // True, if at least one message was sent, other
    		       // than a LocateRequest
    CORBA::Octet minor_; // The highest minor version seen in all messages
    bool closeLogged_; // True, if connection closure was already logged
    CORBA::ULong shutdownTimeout_; // The shutdown timeout

    //
    // Unsent messages
    //
    struct UnsentMessage
    {
	OCI::Buffer_var buf;
	Downcall_var down;

	UnsentMessage() { }
	UnsentMessage(const UnsentMessage& msg)
	    : buf(msg.buf), down(msg.down) { }
	UnsentMessage(OCI::Buffer_ptr);
	UnsentMessage(Downcall_ptr);
    };
    OB_STL::list<UnsentMessage> unsent_;

    //
    // Downcalls awaiting replies
    //
    OB_STL::list<Downcall_var> pending_;

    //
    // The worker's state
    //
public:
    enum State
    {
	StateActive,
	StateClosing,
	StateError,
	StateClosed
    };
protected:
    State state_;
    
    //
    // The failure exception. This must be set if state_ >=
    // StateCloseError.
    //
    CORBA::SystemException* exception_;

    //
    // Emit a trace message when closing the connection
    //
    void logCloseConnection();

    //
    // Add new unsent buffer/downcall
    //
    virtual void add(OCI::Buffer_ptr);
    virtual void add(Downcall_ptr);

    //
    // Move the first unsent downcall to the list of pending downcalls
    //
    Downcall_ptr moveFirstUnsentToPending();

    //
    // Move back the given downcall from the pending list of downcalls
    // to the unsent list of downcalls.
    //
    void movePendingToUnsent(Downcall_ptr);

    //
    // Find and remove a pending downcall
    //
    Downcall_ptr findAndRemovePending(CORBA::ULong);

    //
    // Change the state of the worker due to an exception
    //
    virtual bool exception(State, const CORBA::SystemException&,
			   bool = false);

    //
    // Shut down the connection, graceful and abortive
    //
    virtual void gracefulShutdown() = 0;
    virtual void abortiveShutdown() = 0;

    //
    // Execute incoming message
    //
    void execute()
	throw();

public:

    GIOPClientWorker(ORBInstance_ptr, GIOPClient_ptr, OCI::Transport_ptr);
    virtual ~GIOPClientWorker();

    static inline GIOPClientWorker_ptr _duplicate(GIOPClientWorker_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPClientWorker_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy() = 0;

    //
    // Get the OCI transport
    //
    OCI::Transport_ptr transport();

    //
    // Return true if at least one message has been sent successfully
    //
    bool messageSent() { return messageSent_; }
};

//
// The GIOPClientWorkerReactive class
//
class GIOPClientWorkerReactive : public GIOPClientWorker,
				 public EventHandlerBase,
				 public JTCRecursiveMutex
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPClientWorkerReactive(const GIOPClientWorkerReactive&);
    void operator=(const GIOPClientWorkerReactive&);

protected:

    Reactor_var reactor_; // The reactor
    Mask mask_; // The registered mask for the event handler

    //
    // Update the reactor registration
    //
    void updateReactor();

    //
    // Add new unsent buffer/downcall
    //
    virtual void add(OCI::Buffer_ptr);
    virtual void add(Downcall_ptr);

    //
    // Change the state of the worker due to an exception
    //
    virtual bool exception(State, const CORBA::SystemException&, bool = false);

    //
    // Shut down the connection, graceful and abortive
    //
    virtual void gracefulShutdown();
    virtual void abortiveShutdown();

    //
    // Receive downcalls without synchronization
    //
    bool receiveNoSync(Downcall_ptr, bool)
        throw();

public:

    GIOPClientWorkerReactive(ORBInstance_ptr, GIOPClient_ptr,
			     OCI::Transport_ptr);
    ~GIOPClientWorkerReactive();

    static inline GIOPClientWorkerReactive_ptr
    _duplicate(GIOPClientWorkerReactive_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPClientWorkerReactive_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy();

    //
    // Called to emit downcalls
    //
    virtual bool send(Downcall_ptr, bool)
        throw();
    virtual bool receive(Downcall_ptr, bool)
        throw();
    virtual bool sendReceive(Downcall_ptr)
        throw();

    //
    // Handle event
    //
    virtual Mask handleEvent(Mask);
};

//
// The GIOPClientWorkerThreaded class
//
class GIOPClientWorkerThreaded : public GIOPClientWorker, public JTCMutex
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPClientWorkerThreaded(const GIOPClientWorkerThreaded&);
    void operator=(const GIOPClientWorkerThreaded&);

    //
    // Mutex for OCI send_* calls
    //
    JTCMutex sendMutex_;

protected:

    //
    // The receiver thread
    //
    class ReceiverThread : public JTCThread
    {
	GIOPClientWorkerThreaded_var worker_;

    public:
	
	ReceiverThread(JTCThreadGroupHandle& group,
		       GIOPClientWorkerThreaded_ptr worker)
	    : JTCThread(group),
	      worker_(GIOPClientWorkerThreaded::_duplicate(worker)) { }
	virtual ~ReceiverThread() { }
	virtual void run();
    };
    friend class ReceiverThread;
    JTCThreadHandle receiverThread_;

    //
    // The shutdown thread
    //
    class ShutdownThread : public JTCThread
    {
	GIOPClientWorkerThreaded_var worker_;

    public:
	
	ShutdownThread(JTCThreadGroupHandle& group,
		       GIOPClientWorkerThreaded_ptr worker)
	    : JTCThread(group),
	      worker_(GIOPClientWorkerThreaded::_duplicate(worker)) { }
	virtual ~ShutdownThread() { }
	virtual void run();
    };
    friend class ShutdownThread;
    JTCThreadHandle shutdownThread_;

    CORBA::ULong connectionTimeout_; // The connection timeout
    CORBA::ULong timestamp_; // Timestamp for the last connection activity

    //
    // Update the timestamp for the last connection activity
    //
    void updateTimestamp();

    //
    // Add new unsent buffer/downcall
    //
    virtual void add(OCI::Buffer_ptr);
    virtual void add(Downcall_ptr);

    //
    // Shut down the connection, graceful and abortive
    //
    virtual void gracefulShutdown();
    virtual void abortiveShutdown();

public:

    GIOPClientWorkerThreaded(ORBInstance_ptr, GIOPClient_ptr,
			     OCI::Transport_ptr);
    ~GIOPClientWorkerThreaded();

    static inline GIOPClientWorkerThreaded_ptr
    _duplicate(GIOPClientWorkerThreaded_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPClientWorkerThreaded_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy();

    //
    // Called to emit downcalls
    //
    virtual bool send(Downcall_ptr, bool)
        throw();
    virtual bool receive(Downcall_ptr, bool)
        throw();
    virtual bool sendReceive(Downcall_ptr)
        throw();

    //
    // Run method for receiver and shutdown thread
    //
    void receiverRun();
    void shutdownRun();
};

} // End of namespace OB


namespace CORBA
{

inline void
release(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPClientWorker_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPClientWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPClientWorkerReactive_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPClientWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPClientWorkerThreaded_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
