// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/OAInterface.h>
#include <OB/Except.h>
#include <OB/OCI.h>
#include <OB/ORBInstance.h>
#include <OB/Logger.h>
#include <OB/CoreTraceLevels.h>
#include <OB/NativeTypes.h>
#include <OB/Reactor.h>
#include <OB/algorithms.h>
#include <OB/function.h>

#include <GIOPServerStarter.h>
#include <GIOPServerWorker.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPServerStarter_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerStarter_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

/*
void
OBDuplicate(OB::GIOPServerStarterReactive_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerStarterReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}
*/

void
OBDuplicate(OB::GIOPServerStarterThreaded_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerStarterThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// GIOPServerStarter private and protected member implementation
// ----------------------------------------------------------------------

void
GIOPServerStarter::logCloseAcceptor()
{
    CoreTraceLevels_var coreTraceLevels =
	orbInstance_ -> getCoreTraceLevels();
    if(coreTraceLevels -> traceConnections())
    {
	OCI::AcceptorInfo_var info = acceptor_ -> get_info();
        String_var msg =
	    string_dup("stopped accepting connections\n");
	String_var desc = info -> describe();
	msg += desc;
	Logger_var logger = orbInstance_ -> getLogger();
	logger -> trace("incoming", msg);
    }
}

void
GIOPServerStarter::reapWorkers()
{
    //
    // reapWorkers should be called with the mutex locked
    //
    //JTCSynchronized synchronized(*this);

    OB_STL::list<GIOPServerWorker_var>::iterator iter = workers_.begin();
    while(iter != workers_.end())
    {
        if((*iter).in() -> destroyed())
	{
	    OB_STL::list<GIOPServerWorker_var>::iterator tmp = iter;
	    ++iter;
	    workers_.erase(tmp);
	}
	else
	    ++iter;
    }
}

// ----------------------------------------------------------------------
// GIOPServerStarter constructor and destructor
// ----------------------------------------------------------------------

GIOPServerStarter::GIOPServerStarter(
    ORBInstance_ptr orbInstance, OCI::Acceptor_ptr acceptor,
    OAInterface_ptr oaInterface)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      acceptor_(OCI::Acceptor::_duplicate(acceptor)),
      oaInterface_(OAInterface::_duplicate(oaInterface)),
      state_(StateHolding) // Must be holding initially
{
    try
    {
        //
        // Trace acceptor creation
        //
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceConnections())
        {
            OCI::AcceptorInfo_var info = acceptor_ -> get_info();
            String_var msg = string_dup("accepting connections\n");
            String_var desc = info -> describe();
            msg += desc;
	    Logger_var logger = orbInstance_ -> getLogger();
            logger -> trace("incoming", msg);
        }

	//
	// Start listening
	//
	acceptor_ -> listen();
    }
    catch(const SystemException&)
    {
	acceptor_ -> close();
	state_ = StateClosed;
        _OB_setRef(0);
        throw;
    }
}

GIOPServerStarter::~GIOPServerStarter()
{
    assert(state_ == StateClosed);
}

// ----------------------------------------------------------------------
// GIOPServerStarterReactive private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPServerStarterReactive::updateReactor()
{
    Mask mask = 0;

    if(state_ == StateActive)
	mask |= EventRead;

    if(mask_ != mask)
    {
	mask_ = mask;

	if(mask_)
	{
	    OCI::Handle handle = acceptor_ -> handle();
	    assert(handle >= 0);
	    reactor_ -> register_handler(this, mask_, TypeServer, handle);
	}
	else
	{
	    reactor_ -> unregister_handler(this);
	}
    }
}

// ----------------------------------------------------------------------
// GIOPServerStarterReactive constructor and destructor
// ----------------------------------------------------------------------

GIOPServerStarterReactive::GIOPServerStarterReactive(
    ORBInstance_ptr orbInstance, OCI::Acceptor_ptr acceptor,
    OAInterface_ptr oaInterface)
    : GIOPServerStarter(orbInstance, acceptor, oaInterface),
      reactor_(orbInstance -> getReactor()),
      mask_(0)
{
    updateReactor();
}

GIOPServerStarterReactive::~GIOPServerStarterReactive()
{
    assert(mask_ == 0);
}

// ----------------------------------------------------------------------
// GIOPServerStarterReactive public member implementation
// ----------------------------------------------------------------------

bool
GIOPServerStarterReactive::setState(State state)
{
    JTCSynchronized synchronized(*this);

    //
    // Don't do anything if there is no state change
    //
    if(state_ == state)
        return false;

    //
    // It is not possible to transition backwards, except if we are
    // in holding state
    //
    if(state_ != StateHolding && state < state_)
        return false;

    switch(state)
    {
    case StateActive:
    {
        OB_STL::list<GIOPServerWorker_var>::iterator iter;
	for(iter = workers_.begin() ; iter != workers_.end() ; iter++)
	    (*iter) -> setState(GIOPServerWorker::StateActive);
	break;
    }

    case StateHolding:
    {
        OB_STL::list<GIOPServerWorker_var>::iterator iter;
	for(iter = workers_.begin() ; iter != workers_.end() ; iter++)
	    (*iter) -> setState(GIOPServerWorker::StateHolding);
        break;
    }
    
    case StateClosed:
    {
        OB_STL::list<GIOPServerWorker_var>::iterator iter;
	for(iter = workers_.begin() ; iter != workers_.end() ; iter++)
	    (*iter) -> setState(GIOPServerWorker::StateClosing);
	workers_.clear();
    
        //
        // Shutdown the acceptor so that no further connections
        // are accepted
        //
        logCloseAcceptor();
        acceptor_ -> shutdown();
    
        //
        // Accept all connections which might have queued up in
        // the listen() backlog
        //
        while(true)
        {
    	    OCI::Transport_var transport;
    	
    	    try
    	    {
    	        transport = acceptor_ -> accept(false);
    	    }
    	    catch(const SystemException&)
    	    {
    	    }
    	
    	    if(is_nil(transport))
    	        break;
    	
    	    try
    	    {
    	        GIOPServerWorker_var worker =
    		    new GIOPServerWorkerReactive(orbInstance_,
    					         transport,
    					         oaInterface_);
    	    
    	        worker -> setState(GIOPServerWorker::StateClosing);
    	    }
    	    catch(const SystemException&)
    	    {
    	        // Ignore SystemExceptions
    	    }
        }
        
        //
        // Close the acceptor
        //
        acceptor_ -> close();
    
        break;
    }
    }

    //
    // Update the state and the reactor
    //
    state_ = state;
    updateReactor();

    return true;
}

Mask
GIOPServerStarterReactive::handleEvent(Mask mask)
{
    JTCSynchronized synchronized(*this);

    //
    // Don't do anything if we're closed
    //
    if(state_ == StateClosed)
	return 0;

    //
    // Handle read events
    //
    assert(mask == EventRead);
    if(state_ == StateActive)
    {
	//
	// Reap the existing set of workers
	//
	reapWorkers();

	//
	// Get new transport, non-blocking
	//
	OCI::Transport_var transport;
	try
	{
	    transport = acceptor_ -> accept(false);
	}
	catch(const NO_PERMISSION&)
	{
	    //
	    // Ignore NO_PERMISSION exceptions
	    //
	}
	catch(const SystemException&)
	{
	    //
	    // Ignore exception. This probably means that the server
	    // exceeded the number available file descriptors.
	    //
	}

	//
	// Create and add new worker
	//
	if(!is_nil(transport))
	{
	    try
	    {
		GIOPServerWorker_var worker =
		    new GIOPServerWorkerReactive(orbInstance_,
						 transport,
						 oaInterface_);
		
		workers_.push_back(worker);
		worker -> setState(GIOPServerWorker::StateActive);
	    }
	    catch(const SystemException& ex)
	    {
		String_var msg = string_dup("can't accept connection\n");
		String_var exMsg = ex._to_string();
		msg += exMsg;
		Logger_var logger = orbInstance_ -> getLogger();
		logger -> warning(msg);
	    }
	}

	return EventRead;
    }
    else
	return 0;
}

// ----------------------------------------------------------------------
// GIOPServerStarterThreaded constructor and destructor
// ----------------------------------------------------------------------

GIOPServerStarterThreaded::GIOPServerStarterThreaded(
    ORBInstance_ptr orbInstance, OCI::Acceptor_ptr acceptor,
    OAInterface_ptr oaInterface)
    : GIOPServerStarter(orbInstance, acceptor, oaInterface)
{
    try
    {
	//
	// Retrieve the thread group for the servers
	//
	JTCThreadGroupHandle group = orbInstance_ -> getServerWorkerGroup();

	//
	// Start starter thread
	//
	starterThread_ = new StarterThread(group, this);
	starterThread_ -> start();
    }
    catch(const JTCException&)
    {
	acceptor_ -> close();
	state_ = StateClosed;
	_OB_setRef(0);
	throw IMP_LIMIT(MinorThreadLimit, COMPLETED_NO);
    }
}

GIOPServerStarterThreaded::~GIOPServerStarterThreaded()
{
}

// ----------------------------------------------------------------------
// GIOPServerStarterThreaded public member implementation
// ----------------------------------------------------------------------

bool
GIOPServerStarterThreaded::setState(State state)
{
    JTCSynchronized synchronized(*this);

    //
    // Don't do anything if there is no state change
    //
    if(state_ == state)
        return false;

    //
    // It is not possible to transition backwards, except if we are
    // in holding state
    //
    if(state_ != StateHolding && state < state_)
        return false;

    switch(state)
    {
    case StateActive:
    {
        OB_STL::list<GIOPServerWorker_var>::iterator iter;
	for(iter = workers_.begin() ; iter != workers_.end() ; iter++)
	    (*iter) -> setState(GIOPServerWorker::StateActive);
	break;
    }

    case StateHolding:
    {
        OB_STL::list<GIOPServerWorker_var>::iterator iter;
	for(iter = workers_.begin() ; iter != workers_.end() ; iter++)
	    (*iter) -> setState(GIOPServerWorker::StateHolding);
        break;
    }
    
    case StateClosed:
    {
        OB_STL::list<GIOPServerWorker_var>::iterator iter;
	for(iter = workers_.begin() ; iter != workers_.end() ; iter++)
	    (*iter) -> setState(GIOPServerWorker::StateClosing);
	workers_.clear();
    
        //
        // Connect to this starter's acceptor, to unblock the call
        // to accept() in the starter thread
        //
        try
        {
		OCI::Transport_var tr = acceptor_ -> connect_self();
		tr -> close();
        }
        catch(const SystemException&)
        {
		//
		// Ignore all system exceptions
		//
        }

        break;
    }
    }

    //
    // Update the state and notify about the state change
    //
    state_ = state;
    notifyAll();

    return true;
}

void
GIOPServerStarterThreaded::starterRun()
{
    while(true)
    {
	//
	// Get new transport, blocking
	//
	OCI::Transport_var transport;
	try
	{
	    transport = acceptor_ -> accept(true);
	    assert(!is_nil(transport));
	}
	catch(const NO_PERMISSION&)
	{
	    //
	    // Ignore NO_PERMISSION exceptions
	    //
	}
	catch(const SystemException&)
	{
	    //
	    // Ignore exception. This probably means that the server
	    // exceeded the number available file descriptors.
	    //
	}

	{
	    JTCSynchronized synchronized(*this);

	    //
	    // Reap the existing set of workers
	    //
	    reapWorkers();

	    //
	    // Check whether we are on hold
	    //
	    while(state_ == StateHolding)
	    {
		try
		{
		    wait();
		}
		catch(const JTCInterruptedException&)
		{
		}
	    }
	    
	    if(!is_nil(transport))
	    {
		try
		{
		    if(state_ == StateActive)
		    {
			//
			// If we're active, we create and add a new
			// worker to the worker list
			//
			GIOPServerWorker_var worker =
			    new GIOPServerWorkerThreaded(orbInstance_,
							 transport,
							 oaInterface_);
			
			workers_.push_back(worker);
			worker -> setState(GIOPServerWorker::StateActive);
		    }
		    else
		    {
			//
			// If we're closed, we create a new dummy
			// worker, only in order to set it to
			// StateClosing for proper connection shutdown
			//
			assert(state_ == StateClosed);
			
			GIOPServerWorker_var worker =
			    new GIOPServerWorkerThreaded(orbInstance_,
							 transport,
							 oaInterface_);
			
			worker -> setState(GIOPServerWorker::StateClosing);
		    }
		}
		catch(const SystemException& ex)
		{
		    String_var msg = string_dup("can't accept connection\n");
		    String_var exMsg = ex._to_string();
		    msg += exMsg;
		    Logger_var logger = orbInstance_ -> getLogger();
		    logger -> warning(msg);
		}
	    }

	    if(state_ == StateClosed)
		break;
	}
    }
}

// ----------------------------------------------------------------------
// GIOPServerStarterThreaded::SenderThread public member implementations
// ----------------------------------------------------------------------

void
GIOPServerStarterThreaded::StarterThread::run()
{
    try
    {
	starter_ -> starterRun();
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    //
    // Shutdown the acceptor so that no further connections are
    // accepted
    //
    starter_ -> logCloseAcceptor();
    starter_ -> acceptor_ -> shutdown();
    
    //
    // Accept all connections which might have queued up in the
    // listen() backlog
    //
    while(true)
    {
	OCI::Transport_var transport;

	try
	{
	    transport = starter_ -> acceptor_ -> accept(false);
	}
	catch(const SystemException&)
	{
	}

	if(is_nil(transport))
	    break;

	try
	{
	    GIOPServerWorker_ptr worker =
		new GIOPServerWorkerThreaded(starter_ -> orbInstance_,
					     transport,
					     starter_ -> oaInterface_);

	    worker -> setState(GIOPServerWorker::StateClosing);
	}
	catch(const SystemException&)
	{
	    // Ignore SystemExceptions
	}
    }

    //
    // Close the acceptor
    //
    starter_ -> acceptor_ -> close();

    //
    // Break cyclic object dependency
    //
    starter_ = GIOPServerStarterThreaded::_nil();
}
