// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_SERVER_STARTER_H
#define OB_GIOP_SERVER_STARTER_H

#include <OB/OAInterface_fwd.h>
#include <OB/OCI_fwd.h>
#include <OB/ORBInstance_fwd.h>
#include <OB/list.h>

#include <GIOPServerStarter_fwd.h>

#include <OB/EventHandlerBase.h>

#include <GIOPServerWorker.h>

namespace OB
{

//
// The GIOPServerStarter class
//
// Must inherit virtual from RefCount, because
// GIOPServerStarterReactive inherits from EventHandlerBase, which
// also inherits virtual from RefCount
//
class GIOPServerStarter : virtual public RefCount, public JTCMonitor
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerStarter(const GIOPServerStarter&);
    void operator=(const GIOPServerStarter&);

protected:

    ORBInstance_var orbInstance_; // The ORBInstance
    OCI::Acceptor_var acceptor_; // The acceptor
    OAInterface_var oaInterface_; // The object adapter interface
    OB_STL::list<GIOPServerWorker_var> workers_; // The workers

    //
    // The starter's state
    //
public:
    enum State
    {
	StateActive,
	StateHolding,
	StateClosed
    };
protected:
    State state_;

    //
    // Emit a trace message when closing the acceptor
    //
    void logCloseAcceptor();

    //
    // Reap the set of workers
    //
    void reapWorkers();

public:
    
    GIOPServerStarter(ORBInstance_ptr, OCI::Acceptor_ptr, OAInterface_ptr);
    virtual ~GIOPServerStarter();

    static inline GIOPServerStarter_ptr _duplicate(GIOPServerStarter_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerStarter_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    virtual bool setState(State) = 0;
};

//
// The GIOPServerStarterReactive class
//
class GIOPServerStarterReactive : public GIOPServerStarter,
				  public EventHandlerBase
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerStarterReactive(const GIOPServerStarterReactive&);
    void operator=(const GIOPServerStarterReactive&);

protected:

    Reactor_var reactor_; // The reactor
    Mask mask_; // The registered mask for the event handler

    //
    // Update the reactor registration
    //
    void updateReactor();

public:
    
    GIOPServerStarterReactive(ORBInstance_ptr, OCI::Acceptor_ptr,
			      OAInterface_ptr);
    virtual ~GIOPServerStarterReactive();

    static inline GIOPServerStarterReactive_ptr
    _duplicate(GIOPServerStarterReactive_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerStarterReactive_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    virtual bool setState(State);

    //
    // Handle event
    //
    virtual Mask handleEvent(Mask);
};

//
// The GIOPServerStarterThreaded class
//
class GIOPServerStarterThreaded : public GIOPServerStarter
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerStarterThreaded(const GIOPServerStarterThreaded&);
    void operator=(const GIOPServerStarterThreaded&);

protected:

    //
    // The starter thread
    //
    class StarterThread : public JTCThread
    {
	GIOPServerStarterThreaded_var starter_;

    public:
	
	StarterThread(JTCThreadGroupHandle& group,
		      GIOPServerStarterThreaded_ptr starter)
	    : JTCThread(group),
	      starter_(GIOPServerStarterThreaded::_duplicate(starter)) { }
	virtual ~StarterThread() { }
	virtual void run();
    };
    friend class StarterThread;
    JTCThreadHandle starterThread_;

public:
    
    GIOPServerStarterThreaded(ORBInstance_ptr, OCI::Acceptor_ptr,
			      OAInterface_ptr);
    virtual ~GIOPServerStarterThreaded();

    static inline GIOPServerStarterThreaded_ptr
    _duplicate(GIOPServerStarterThreaded_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerStarterThreaded_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    virtual bool setState(State);

    //
    // Run method for starter thread
    //
    void starterRun();
};

} // End of namespace OB

namespace CORBA
{

inline void
release(OB::GIOPServerStarter_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerStarter_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPServerStarterReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerStarterReactive_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPServerStarterThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerStarterThreaded_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
