// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/TypeCode.h>

#include <Logger_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Logger_impl private method implementation
// ----------------------------------------------------------------------

void
Logger_impl::message(const char* msg)
{
    JTCSynchronized synchronized(*this);

    //
    // The stdio library is used instead of the iostream library as a
    // workaround for a bug found on the Windows platform.  In
    // particular, if the /MT or /MTd options are used, then the cout
    // and cerr objects may be destroyed before local variables declared
    // in main() are destroyed.
    //
    fputs(msg, stderr);
    fputc('\n', stderr);
    fflush(stderr);
}

#if !defined(HAVE_NO_WCHAR_T_SUPPORT_IN_STREAMS) && defined(HAVE_IOSTREAM)
void
Logger_impl::wmessage(const wchar_t* msg)
{
    JTCSynchronized synchronized(*this);

    std::wcerr << msg << std::endl;
}
#else
void
Logger_impl::wmessage(const wchar_t* msg)
{
    bool hasWideChar = false;
    String_var s = string_dup("");

    for(size_t i = 0 ; i < OB::wcslen(msg) ; i++)
    {
	if(msg[i] > (wchar_t)0xff)
	    hasWideChar = true;

	s += (const char)msg[i];
    }

    if(hasWideChar)
    {
	s += " - narrowed";
    }

    message(s);
}
#endif


// ----------------------------------------------------------------------
// Logger_impl constructor/destructor
// ----------------------------------------------------------------------

Logger_impl::Logger_impl()
{
}

Logger_impl::~Logger_impl()
{
}

// ----------------------------------------------------------------------
// Logger_impl public member implementation
// ----------------------------------------------------------------------

void
Logger_impl::info(const char* msg)
{
    message(msg);
}

void
Logger_impl::winfo(const wchar_t* msg)
{
    wmessage(msg);
}

void
Logger_impl::error(const char* msg)
{
    const char* head = "error: ";
    String_var s = string_alloc(strlen(head) + strlen(msg));
    strcpy(s.inout(), head);
    strcat(s.inout(), msg);
    message(s);
}

void
Logger_impl::werror(const wchar_t* msg)
{
    const wchar_t* head = L"error: ";
    WString_var s = wstring_alloc(OB::wcslen(head) + OB::wcslen(msg));
    OB::wcscpy(s.inout(), head);
    OB::wcscat(s.inout(), msg);
    wmessage(s);
}

void
Logger_impl::warning(const char* msg)
{
    const char* head = "warning: ";
    String_var s = string_alloc(strlen(head) + strlen(msg));
    strcpy(s.inout(), head);
    strcat(s.inout(), msg);
    message(s);
}

void
Logger_impl::wwarning(const wchar_t* msg)
{
    const wchar_t* head = L"warning: ";
    WString_var s = wstring_alloc(OB::wcslen(head) + OB::wcslen(msg));
    OB::wcscpy(s.inout(), head);
    OB::wcscat(s.inout(), msg);
    wmessage(s);
}

void
Logger_impl::trace(const char* category, const char* msg)
{
    String_var s = (const char*)"[ ";
    s += category;
    s += ": ";
    const char* p = msg;
    const char* next;
    while((next = strchr((char*)p, '\n')))
    {
	next++;
	String_var line = string_alloc(next - p);
	strncpy(line.inout(), p, next - p);
	line[next - p] = '\0';
	s += line;
	s += "  ";
	p = next;
    }
    s += p;
    s += " ]";
    message(s);
}

void
Logger_impl::wtrace(const wchar_t* category, const wchar_t* msg)
{
    WString_var s = (const wchar_t*)L"[ ";
    s += category;
    s += L": ";
    const wchar_t* p = msg;
    const wchar_t* next;
    while((next = OB::wcschr((wchar_t*)p, L'\n')))
    {
	next++;
	WString_var line = wstring_alloc(next - p);
	OB::wcsncpy(line.inout(), p, next - p);
	line[next - p] = L'\0';
	s += line;
	s += L"  ";
	p = next;
    }
    s += p;
    s += L" ]";
    wmessage(s);
}
