// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/Stream.h>
#include <OB/IIOP.h>
#include <OB/OBPolicies.h>
#include <OB/NativeTypes.h>
#include <OB/MarshalStubImpl.h>
#include <OB/OBORB.h>
#include <OB/IORUtil.h>
#include <OB/function.h>
#include <OB/algorithms.h>
#include <OB/seq_converter.h>

#include <OCI_IIOP_impl.h>
#include <IIOPUtil.h>

#include <stdlib.h>
#include <ctype.h>

#ifndef HAVE_STRCASECMP
#   define strcasecmp stricmp
#endif

//
// Some operating systems don't know socklen_t
//
#ifdef HAVE_SOCKLEN_T
    typedef socklen_t socklen_type;
#else
#   ifdef HAVE_SOCKETS_WITH_SIZE_T
        typedef size_t socklen_type;
#   else
        typedef int socklen_type;
#   endif
#endif

//
// Undefine minor (this is needed for SunOS 4.1 and some other Unix
// flavors)
//
#undef minor

using namespace CORBA;
using namespace OB;

//#define OB_OCI_DEBUG

// ----------------------------------------------------------------------
// Some utility functions
// ----------------------------------------------------------------------

#ifdef never
//
// TODO: Potentially hacks.  IIOPCreateIOR and IIOPExtractProfileInfo
// are used to support the IMR.
//

IOP::IOR*
OB::IIOPCreateIOR(const char* host, UShort port, const char* id,
		  const OCI::ProfileInfo& profileInfo)
{
    assert(port > 0);

    IOP::IOR_var ior = new IOP::IOR;
    ior -> type_id = id;
    ior -> profiles.length(1);
    ior -> profiles[0].tag = IOP::TAG_INTERNET_IOP;
    
    if(profileInfo.major == 1 && profileInfo.minor == 0)
    {
	::IIOP::ProfileBody_1_0 body;
	body.iiop_version.major = profileInfo.major;
	body.iiop_version.minor = profileInfo.minor;
	body.host = host;
	body.port = port;
	body.object_key.length(profileInfo.key.length());
	memcpy(body.object_key.get_buffer(), profileInfo.key.get_buffer(),
	       profileInfo.key.length());
	
	OCI::BufferImpl buf;
        OutputStreamImpl out(&buf);
	out._OB_writeEndian();
	body._OB_marshal(&out);
	ULong count = out._OB_pos();
	ior -> profiles[0].profile_data.length(count);
	Octet* oct = ior -> profiles[0].profile_data.get_buffer();
	memcpy(oct, buf.data(), count);
    }
    else
    {
	::IIOP::ProfileBody_1_1 body;
	body.iiop_version.major = profileInfo.major;
	body.iiop_version.minor = profileInfo.minor;
	body.host = host;
	body.port = port;
	body.object_key.length(profileInfo.key.length());
	memcpy(body.object_key.get_buffer(), profileInfo.key.get_buffer(),
	       profileInfo.key.length());
	body.components.length(profileInfo.components.length());
	for(ULong i = 0 ; i < profileInfo.components.length() ; i++)
	    body.components[i] = profileInfo.components[i];
	
	OCI::BufferImpl buf;
        OutputStreamImpl out(&buf);
	out._OB_writeEndian();
	body._OB_marshal(&out);
	ULong count = out._OB_pos();
	ior -> profiles[0].profile_data.length(count);
	Octet* oct = ior -> profiles[0].profile_data.get_buffer();
	memcpy(oct, buf.data(), count);
    }

    return ior._retn();
}

IOP::IOR*
OB::IIOPCreateIOR(const IOP::IOR& ior, const char* id,
		  const OCI::ObjectKey& key)
{
    //
    // Extract the IIOP profile information from the provided IOR
    //
    ULong profile;
    for(profile = 0 ; profile < ior.profiles.length() ; profile++)
	if(ior.profiles[profile].tag == IOP::TAG_INTERNET_IOP)
	    break;
    
    assert(profile < ior.profiles.length()); // TODO: Internal error?
    
    const Octet* coct = ior.profiles[profile].profile_data.get_buffer();
    ULong len = ior.profiles[profile].profile_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    ::IIOP::ProfileBody_1_0 body;
    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);
    
    //
    // Create a new profile based on the information from the provided IOR
    //
    OCI::ProfileInfo profileInfo;
    profileInfo.key = key;
    profileInfo.major = body.iiop_version.major;
    profileInfo.minor = body.iiop_version.minor;
    return IIOPCreateIOR(body.host, body.port, id, profileInfo);
}

bool
OB::IIOPExtractProfileInfo(const OCI::IOR& ior,
			   OCI::ProfileInfo_out profileInfo)
{
    OCI::ProfileInfoSeq profileInfoSeq;
    IIOPExtractAllProfileInfos(ior, profileInfoSeq, false, 0, 0, false);
    if(profileInfoSeq.length() > 0)
    {
        assert(profileInfoSeq.length() == 1);
        profileInfo = new OCI::ProfileInfo(profileInfoSeq[0]);
        return true;
    }

    return false;
}
#endif

bool
OB::IIOPHostMatch(const char* host1, const char* host2, bool loopbackMatches)
{
    //
    // Direct host name comparison
    //
    if(strcmp(host1, host2) != 0)
    {
        //
        //
        // Direct host name comparision failed - must look up
        // addresses to be really sure if the hosts differ
        //
        try
        {
            struct sockaddr_in addr1;
            GetInAddr(host1, addr1);
            
            struct sockaddr_in addr2;
            GetInAddr(host2, addr2);
            
            if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
            {
                //
                // Address comparison failed - shall I extract
                // the key if the profile body contains the
                // loopback address?
                //
                if(loopbackMatches)
                {
                    if(addr2.sin_addr.s_addr != htonl(INADDR_LOOPBACK))
                        return false;
                }
                else
                    return false;
            }
        }
        catch(const COMM_FAILURE&)
        {
            //
            // Continue on hostname lookup failure
            //
            return false;
        }
    }

    return true;
}

void
OB::IIOPExtractAllProfileInfos(const OCI::IOR& ior,
                               OCI::ProfileInfoSeq& profileInfoSeq,
                               bool performMatch,
                               const char* host, UShort port,
                               bool loopbackMatches)
{
    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == IOP::TAG_INTERNET_IOP)
	{
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
	    OCI::BufferImpl buf((Octet*)coct, len);
	    InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);

            //
            // Read components if the IIOP version is > 1.0
            //
            OCI::TaggedComponentSeq components;
            if(body.iiop_version.major > 1 || body.iiop_version.minor > 0)
            {
                len = in.read_ulong();
                components.length(len);
                IOP::TaggedComponent* buf = components.get_buffer();
                for(ULong i = 0 ; i < len ; i++)
                    IOP::TaggedComponent::_OB_unmarshal(buf[i], &in);
            }

            if(performMatch)
            {
                //
                // Check primary host/port
                //
                bool match = false;
                if(port == body.port &&
                   IIOPHostMatch(host, body.host, loopbackMatches))
                {
                    match = true;
                }

                //
                // Check alternate host/port
                //
                if(!match)
                {
                    for(ULong j = 0 ; j < components.length() && !match ; j++)
                    {
                        if(components[j].tag ==
                           IOP::TAG_ALTERNATE_IIOP_ADDRESS)
                        {
                            const Octet* coct =
                                components[j].component_data.get_buffer();
                            len = components[j].component_data.length();
                            OCI::BufferImpl b((Octet*)coct, len);
                            InputStreamImpl s(&b, 0, false);
                            s._OB_readEndian();
                            String_var altHost = s.read_string();
                            UShort altPort = s.read_ushort();
                            if(port == altPort &&
                               IIOPHostMatch(host, altHost, loopbackMatches))
                            {
                                match = true;
                            }
                        }
                    }
                }

                if(!match)
                    continue;
            }

            //
            // OK, found a match
            //
            len = profileInfoSeq.length() + 1;
            profileInfoSeq.length(len);
            profileInfoSeq[len - 1].key.length(body.object_key.length());
            memcpy(profileInfoSeq[len - 1].key.get_buffer(),
                   body.object_key.get_buffer(), body.object_key.length());
            profileInfoSeq[len - 1].minor = body.iiop_version.minor;
            profileInfoSeq[len - 1].major = body.iiop_version.major;
            profileInfoSeq[len - 1].id = ior.profiles[i].tag;
            profileInfoSeq[len - 1].index = i;
            profileInfoSeq[len - 1].components.length(components.length());
            for(ULong j = 0 ; j < components.length() ; j++)
                profileInfoSeq[len - 1].components[j] = components[j];
	}
    }
}

static Boolean
IIOPCompareBodies(const ::IIOP::ProfileBody_1_0& body1,
		  const ::IIOP::ProfileBody_1_0& body2)
{
    //
    // Compare versions
    //
    if(body1.iiop_version.major != body2.iiop_version.major ||
       body1.iiop_version.minor != body2.iiop_version.minor)
	return false;
    
    //
    // Compare ports
    //
    if(body1.port != body2.port)
	return false;
    
    //
    // Compare object keys
    //
    if(body1.object_key.length() != body2.object_key.length())
	return false;
    
    if(memcmp(body1.object_key.get_buffer(), body2.object_key.get_buffer(),
	      body1.object_key.length()) != 0)
	return false;
    
    //
    // Direct host name comparison
    //
    if(strcmp(body1.host, body2.host) != 0)
    {
	//
	// Direct host name comparision failed - must look up
	// addresses to be really sure if the hosts differ
	//
	try
	{
	    struct sockaddr_in addr1;
	    GetInAddr(body1.host, addr1);
	    
	    struct sockaddr_in addr2;
	    GetInAddr(body2.host, addr2);
	    
	    if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
		return false;
	}
	catch(const COMM_FAILURE&)
	{
	    //
	    // Return false on hostname lookup failure
	    //
	    return false;
	}
    }
    
    //
    // OK, found a match
    //
    return true;
}

static Boolean
IIOPEquivalent(const OCI::IOR& ior1, const OCI::IOR& ior2)
{
    ULong p1, p2, b1, b2;
    ULong cnt1 = 0, cnt2 = 0;
    ::IIOP::ProfileBody_1_0** bodies1 = 0;
    ::IIOP::ProfileBody_1_0** bodies2 = 0;

    //
    // Calculate number of IIOP profiles in ior1
    //
    for(p1 = 0 ; p1 < ior1.profiles.length() ; p1++)
	if(ior1.profiles[p1].tag == IOP::TAG_INTERNET_IOP)
	    cnt1++;
    
    //
    // Calculate number of IIOP profiles in ior2
    //
    for(p2 = 0 ; p2 < ior2.profiles.length() ; p2++)
	if(ior2.profiles[p2].tag == IOP::TAG_INTERNET_IOP)
	    cnt2++;

    //
    // Return false now if the number of IIOP profile bodies do not
    // match
    //
    if(cnt1 != cnt2)
	return false;

    //
    // Create an array with all IIOP profile bodies of ior1
    //
    bodies1 = new ::IIOP::ProfileBody_1_0*[cnt1];
    for(p1 = 0, b1 = 0 ; p1 < ior1.profiles.length() ; p1++)
	if(ior1.profiles[p1].tag == IOP::TAG_INTERNET_IOP)
	{
	    const Octet* coct =	ior1.profiles[p1].profile_data.get_buffer();
            ULong len = ior1.profiles[p1].profile_data.length();
	    OCI::BufferImpl buf((Octet*)coct, len);
	    InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0* p = new ::IIOP::ProfileBody_1_0;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(*p, &in);
	    bodies1[b1++] = p;
	}
    
    assert(b1 == cnt1);
    
    //
    // Create an array with all IIOP profile bodies of ior2
    //
    bodies2 = new ::IIOP::ProfileBody_1_0*[cnt2];
    for(p2 = 0, b2 = 0 ; p2 < ior2.profiles.length() ; p2++)
	if(ior2.profiles[p2].tag == IOP::TAG_INTERNET_IOP)
	{
	    const Octet* coct =	ior2.profiles[p2].profile_data.get_buffer();
	    ULong len = ior2.profiles[p2].profile_data.length();
	    OCI::BufferImpl buf((Octet*)coct, len);
	    InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0* p = new ::IIOP::ProfileBody_1_0;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(*p, &in);
	    bodies2[b2++] = p;
	}
    
    assert(b2 == cnt2);

    //
    // Check for profile body matches
    //
    for(b1 = 0 ; b1 < cnt1 ; b1++)
    {
	for(b2 = 0 ; b2 < cnt2 ; b2++)
	{
	    if(!bodies2[b2])
		continue;
	    
	    //
	    // Compare profile bodies
	    //
	    if(IIOPCompareBodies(*bodies1[b1], *bodies2[b2]))
	    {
		//
		// OK, found a match
		//
		delete bodies1[b1];
		bodies1[b1] = 0;
		delete bodies2[b2];
		bodies2[b2] = 0;
		break;
	    }
	}
    }

    //
    // Check whether there are any unmatched IIOP profile bodies
    //
    ULong unmatched1 = 0;
    for(b1 = 0 ; b1 < cnt1 ; b1++)
	if(bodies1[b1])
	{
	    unmatched1++;
	    delete bodies1[b1];
	}

    delete [] bodies1;

    ULong unmatched2 = 0;
    for(b2 = 0 ; b2 < cnt2 ; b2++)
	if(bodies2[b2])
	{
	    unmatched2++;
	    delete bodies2[b2];
	}

    delete [] bodies2;

    return unmatched1 == 0 && unmatched2 == 0;
}

static ULong
IIOPHash(const OCI::IOR& ior, ULong maximum)
{
    ULong hash = 0;

    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == IOP::TAG_INTERNET_IOP)
	{
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
	    OCI::BufferImpl buf((Octet*)coct, len);
	    InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);

	    //
	    // Add port to hash
	    //
	    hash ^= body.port;

	    //
	    // Add object key to hash
	    //
	    for(ULong j = 0 ; j + 1 < body.object_key.length() ; j += 2)
		hash ^= body.object_key[j + 1] * 256 + body.object_key[j];
	}
    }

    return hash % (maximum + 1);
}

// ----------------------------------------------------------------------
// CorbalocProtocol_impl constructor and destructor
// ----------------------------------------------------------------------
        
OCI::IIOP::CorbalocProtocol_impl::CorbalocProtocol_impl()
{
}

OCI::IIOP::CorbalocProtocol_impl::~CorbalocProtocol_impl()
{
}

char*
OCI::IIOP::CorbalocProtocol_impl::name()
    throw(SystemException)
{
    return string_dup("iiop");
}

IOP::TaggedProfile*
OCI::IIOP::CorbalocProtocol_impl::parse_address(
    const char* addr, const CorbalocProtocol::ObjectKey& key)
    throw(SystemException)
{
    Octet major, minor;

    //
    // Do we have an iiop version 'X.Y@' portion? (default is 1.0)
    //
    const char* at = strchr(addr, '@');
    if(at == 0)
    {
        major = 1;
        minor = 0;
    }
    else
    {
        const char* cp = addr;
        const char* dot;
        bool seenDot = false;
        bool ok = true;
        while(cp < at && ok)
        {
            if(*cp == '.')
            {
                if(seenDot || cp == addr || cp == at - 1)
                    ok = false;
                seenDot = true;
                dot = cp;
            }
            else if(!isdigit(*cp))
                ok = false;
            cp++;
        }
        if(!ok || !seenDot)
            throw BAD_PARAM("URL: iiop version must be of the form `X.Y'",
                            MinorBadAddress, COMPLETED_NO);

        int nMajor, nMinor;
        nMajor = atoi(addr);
        nMinor = atoi(dot + 1);
        if(nMajor != 1 || nMinor > 255)
            throw BAD_PARAM("URL: iiop version is invalid or unsupported", 
                            MinorBadAddress, COMPLETED_NO);
        major = (Octet)nMajor;
        minor = (Octet)nMinor;

        // 
        // Can only write up to version 1.2 iiop profile. Server
        // should still talk down minor iiop revisions.
        // 
        if(minor > 2)
            minor = 2;

        //
        // Skip the version
        //
        addr = at + 1;
    }

    // 
    // Empty hostname is illegal (as is port ':YYYY' by itself)
    // 
    if(*addr == '\0' || *addr == ':')
        throw BAD_PARAM("URL: iiop hostname must be specified", 
                        MinorBadAddress, COMPLETED_NO);

    //
    // Hostname is terminated by ':port', or by end of string
    //
    String_var host;
    const char* colon = strchr(addr, ':');
    if(colon == 0)
        host = addr;
    else
    {
        ULong len = colon - addr;
        host = string_alloc(len);
        strncpy(host.inout(), addr, len);
        host[len] = '\0';
    }

    // 
    // Valid range for port is 1 - 65535
    // 
    ULong port = 2809; // default port
    if(colon != 0 && *(colon + 1) != '\0')
    {
        const char* cp = colon + 1;
        while(*cp)
        {
            if(!isdigit(*cp))
                throw BAD_PARAM("URL: iiop port is invalid", MinorBadAddress,
                                COMPLETED_NO);
            cp++;
        }
        port = (ULong)atoi(colon + 1);
        if(port < 1 || port > 65535)
            throw BAD_PARAM("URL: iiop port must be between 1 and 65535", 
                            MinorBadAddress, COMPLETED_NO);
    }

    //
    // Create profile
    //
    IOP::TaggedProfile_var profile = new IOP::TaggedProfile;
    profile -> tag = IOP::TAG_INTERNET_IOP;

    if(major == 1 && minor == 0)
    {
        ::IIOP::ProfileBody_1_0 body;
        body.iiop_version.major = 1;
        body.iiop_version.minor = 0;
        body.host = host;
        body.port = (UShort)port;
        body.object_key.length(key.length());
        memcpy(body.object_key.get_buffer(), key.get_buffer(), key.length());

        OCI::BufferImpl buf;
        OutputStreamImpl out(&buf);
        out._OB_writeEndian();
        body._OB_marshal(&out);
        ULong count = out._OB_pos();
        profile -> profile_data.length(count);
        Octet* oct = profile -> profile_data.get_buffer();
        memcpy(oct, buf.data(), count);
    }
    else
    {
        // 
        // If the IIOP version is 1.1 or *greater*, then the best
        // we can do is a 1.1 profile body with no components.
        // The server or forwarding agent may or may not accept it.
        // 

        ::IIOP::ProfileBody_1_1 body;
        body.iiop_version.major = major;
        body.iiop_version.minor = minor;
        body.host = host;
        body.port = (UShort)port;
        body.object_key.length(key.length());
        memcpy(body.object_key.get_buffer(), key.get_buffer(), key.length());

        OCI::BufferImpl buf;
        OutputStreamImpl out(&buf);
        out._OB_writeEndian();
        body._OB_marshal(&out);
        ULong count = out._OB_pos();
        profile -> profile_data.length(count);
        Octet* oct = profile -> profile_data.get_buffer();
        memcpy(oct, buf.data(), count);
    }

    return profile._retn();
}

void
OCI::IIOP::CorbalocProtocol_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// Plugin_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::Plugin_impl::Plugin_impl(ORB_ptr orb)
    : orb_(ORB::_duplicate(orb))
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "Plugin_impl" << OB_ENDL;
#endif
}

OCI::IIOP::Plugin_impl::~Plugin_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~Plugin_impl" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// Plugin_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::Plugin_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

void
OCI::IIOP::Plugin_impl::init_client(const OCI::ParamSeq& params)
    throw(SystemException)
{
    bool keepAlive = true;
    ULong i = 0;
    while(i < params.length())
    {
        if(strcmp(params[i], "--no-keepalive") == 0)
        {
            keepAlive = false;
            i++;
        }
        else
        {
            String_var err = string_dup("iiop: unknown client parameter `");
            err += params[i];
            err += "'";
            throw INITIALIZE(err);
        }
    }

    try
    {
        Object_var obj =
            orb_ -> resolve_initial_references("OCIConFactoryRegistry");
        OCI::ConFactoryRegistry_var registry =
            OCI::ConFactoryRegistry::_narrow(obj);
        assert(!is_nil(registry));
        OCI::ConFactory_var factory = new ConFactory_impl(keepAlive);
        registry -> add_factory(factory);
    }
    catch(const ORB::InvalidName&)
    {
        assert(false);
    }
    catch(const OCI::FactoryAlreadyExists&)
    {
        throw INITIALIZE("OCI IIOP plug-in already installed");
    }

    try
    {
        Object_var obj =
            orb_ -> resolve_initial_references("URLRegistry");
        URLRegistry_var registry = URLRegistry::_narrow(obj);
        assert(!is_nil(registry));
        URLScheme_var scheme = registry -> find_scheme("corbaloc");
        assert(!is_nil(scheme));
        CorbalocURLScheme_var corbaloc = CorbalocURLScheme::_narrow(scheme);
        assert(!is_nil(corbaloc));
        CorbalocProtocol_var prot = new CorbalocProtocol_impl();
        corbaloc -> add_protocol(prot);
    }
    catch(const ORB::InvalidName&)
    {
        assert(false);
    }
    catch(const CorbalocURLScheme::ProtocolAlreadyExists&)
    {
        assert(false);
    }
}

void
OCI::IIOP::Plugin_impl::init_server(const OCI::ParamSeq&)
    throw(SystemException)
{
    try
    {
        Object_var obj =
            orb_ -> resolve_initial_references("OCIAccFactoryRegistry");
        OCI::AccFactoryRegistry_var registry =
            OCI::AccFactoryRegistry::_narrow(obj);
        assert(!is_nil(registry));
        OCI::AccFactory_var factory = new AccFactory_impl();
        registry -> add_factory(factory);
    }
    catch(const ORB::InvalidName&)
    {
        assert(false);
    }
    catch(const OCI::FactoryAlreadyExists&)
    {
        throw INITIALIZE("OCI IIOP plug-in already installed");
    }
}

// ----------------------------------------------------------------------
// Transport_impl private and protected member implementations
// ----------------------------------------------------------------------

void
OCI::IIOP::Transport_impl::setBlock(Boolean block)
{
    if(block != block_)
    {
	block_ = block;

	if(block_)
	    SetBlock(fd_);
	else
	    SetNoBlock(fd_);
    }
}

// ----------------------------------------------------------------------
// Transport_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::Transport_impl::Transport_impl(Connector_impl* con,
					  OCI::Handle fd)
    : fd_(fd), block_(true)
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "Transport_impl" << OB_ENDL;
#endif

    //
    // Set the socket to blocking initially
    //
    SetBlock(fd_);

    //
    // Since the Constructor of TransportInfo uses this object create
    // it after all members are initialized
    //
    try
    {
	info_ = new TransportInfo_impl(this, con);
    }
    catch(...)
    {
        CloseSocket(fd_);
        _OB_setRef(0);
	throw;
    }
}

OCI::IIOP::Transport_impl::Transport_impl(Acceptor_impl* acc,
					  OCI::Handle fd)
    : fd_(fd), block_(true)
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "Transport_impl" << OB_ENDL;
#endif

    //
    // Set the socket to blocking initially
    //
    SetBlock(fd_);

    //
    // Since the Constructor of TransportInfo uses this object create
    // it after all members are initialized
    //
    try
    {
	info_ = new TransportInfo_impl(this, acc);
    }
    catch(...)
    {
        CloseSocket(fd_);
        _OB_setRef(0);
	throw;
    }

#ifdef WIN32
    //
    // Set a larger send buffer size to avoid a performance problem
    // on Win32. See send_detect().
    //
    // TODO: This is a total hack!
    //
    int bufsz = 64 * 1024; // 64K
    if(setsockopt(fd_, SOL_SOCKET, SO_SNDBUF, (char*)&bufsz,
       sizeof(int)) == -1)
    {
        int err = GetError();
        CloseSocket(fd_);
        _OB_setRef(0);
        SetError(err);
        throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
#endif
}

OCI::IIOP::Transport_impl::~Transport_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~Transport_impl" << OB_ENDL;
#endif

    if(fd_ >= 0)
	close();

    release(info_);
}

// ----------------------------------------------------------------------
// Transport_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::Transport_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

void
OCI::IIOP::Transport_impl::close()
    throw(SystemException)
{
    //
    // Call callbacks
    //
    info_ -> _OB_callCloseCB(info_);

    //
    // Destroy the info object
    //
    info_ -> _OB_destroy();

    //
    // Close the socket
    //
    int saveFd = fd_;
    fd_ = -1; // Must be set to -1 before the shutdown/close
    ::shutdown(saveFd, SHUT_RDWR); // This helps to unblock threads blocking
                                   // in recv()
    CloseSocket(saveFd);
}

void
OCI::IIOP::Transport_impl::shutdown()
    throw(SystemException)
{
    ::shutdown(fd_, SHUT_WR); // Shutdown send side only
}

void
OCI::IIOP::Transport_impl::receive(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
#ifdef __hpux
	if(block)
	{
	    //
	    // On HP-UX 11.x we must always block on select()
	    // first. Otherwise, if we block on recv() and another
	    // thread tries to close the socket, the close() system
	    // call will hang until the other peer close the socket.
	    //
	    fd_set fdSet;
	    FD_ZERO(&fdSet);
	    
	    //
	    // It's possible for the fd_ member to become -1 if this
	    // transport is closed by another thread
	    //
	    int fd = fd_;
	    if(fd != -1)
	    {
		//
		// Use fd here, otherwise FD_SET will crash if given a
		// file descriptor of -1
		//
		FD_SET(fd, &fdSet);

		//
		// Use fd here
		//
		int resultS = ::select(fd + 1, &fdSet, 0, 0, 0);

		if(resultS == 0)
		{
		    //
		    // Shouldn't happen since we didn't set any
		    // timeout on for the select.
		    //
		    assert(false);
		}
		else if(resultS == -1)
		{
		    if(IsInterrupted())
			continue;
                
		    throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
		}
	    }
        }
#endif

	int result = ::recv(fd_, (char*)buf -> rest(), sizeOfPacket, 0);

	if(result == 0)
	{
	    throw COMM_FAILURE("", MinorRecvZero, COMPLETED_NO);
	}
	else if(result == -1)
	{
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
	        sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(WouldBlock())
	    {
		if(!block)
		    return;
		
		setBlock(true);
		continue;
	    }

	    throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
	}

	buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }
}

Boolean
OCI::IIOP::Transport_impl::receive_detect(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
#ifdef __hpux
	if(block)
	{
	    //
	    // On HP-UX 11.x we must always block on select()
	    // first. Otherwise, if we block on recv() and another
	    // thread tries to close the socket, the close() system
	    // call will hang until the other peer close the socket.
	    //
	    fd_set fdSet;
	    FD_ZERO(&fdSet);
	    
	    //
	    // It's possible for the fd_ member to become -1 if this
	    // transport is closed by another thread
	    //
	    int fd = fd_;
	    if(fd != -1)
	    {
		//
		// Use fd here, otherwise FD_SET will crash if given a
		// file descriptor of -1
		//
		FD_SET(fd, &fdSet);

		//
		// Use fd here
		//
		int resultS = ::select(fd + 1, &fdSet, 0, 0, 0);
		
		if(resultS == 0)
		{
		    return false;
		}
		else if(resultS == -1)
		{
		    if(IsInterrupted())
			continue;
		    
		    if(fd_ == -1)
			return false;
		    
		    throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
		}
	    }
        }
#endif

        int result = ::recv(fd_, (char*)buf -> rest(), sizeOfPacket, 0);

	if(result == 0)
	{
	    return false;
	}
	else if(result == -1)
	{
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
	        sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(fd_ == -1 || ConnectionDown())
		return false;

	    if(WouldBlock())
	    {
		if(!block)
		    return true;
		
		setBlock(true);
		continue;
	    }

	    throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
	}

	buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }

    return true;
}

void
OCI::IIOP::Transport_impl::receive_timeout(OCI::Buffer_ptr buf, ULong t)
    throw(SystemException)
{
    if(t == 0 || fd_ == -1 /*TODO*/)
    {
	receive(buf, false);
	return;
    }

    setBlock(false);

    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
	fd_set fdSet;
	FD_ZERO(&fdSet);

        //
        // It's possible for the fd_ member to become -1 if this
        // transport is closed by another thread
        //
        int fd = fd_;
        if(fd != -1)
        {
            //
            // Use fd here, otherwise FD_SET will crash if given a
            // file descriptor of -1
            //
            FD_SET(fd, &fdSet);

            struct timeval tv;
            tv.tv_sec = t / 1000;
            tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

            //
            // Use fd here, since if fd_ == -1, select will simply
            // block for the specified timeout period
            //
            int resultS = ::select(fd + 1, &fdSet, 0, 0, &tv);
            
            if(resultS == 0)
            {
                return;
            }
            else if(resultS == -1)
            {
                if(IsInterrupted())
                    continue;
                
                throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
            }
        }

        //
        // The fd_ member may be -1 at this point. If so, recv will fail.
        //
        int result = ::recv(fd_, (char*)buf -> rest(), sizeOfPacket, 0);

        if(result == 0)
        {
            throw COMM_FAILURE("", MinorRecvZero, COMPLETED_NO);
        }
        else if(result == -1)
        {
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
	        sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(WouldBlock())
		return;

            throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
        }

        buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }
}

Boolean
OCI::IIOP::Transport_impl::receive_timeout_detect(OCI::Buffer_ptr buf, ULong t)
    throw(SystemException)
{
    if(t == 0 || fd_ == -1 /*TODO*/)
	return receive_detect(buf, false);

    setBlock(false);

    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
	fd_set fdSet;
	FD_ZERO(&fdSet);

        //
        // It's possible for the fd_ member to become -1 if this
        // transport is closed by another thread
        //
        int fd = fd_;
        if(fd != -1)
        {
            //
            // Use fd here, otherwise FD_SET will crash if given a
            // file descriptor of -1
            //
            FD_SET(fd, &fdSet);

            struct timeval tv;
            tv.tv_sec = t / 1000;
            tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

            //
            // Use fd here, since if fd_ == -1, select will simply
            // block for the specified timeout period
            //
            int resultS = ::select(fd + 1, &fdSet, 0, 0, &tv);
            
            if(resultS == 0)
            {
                return true;
            }
            else if(resultS == -1)
            {
                if(IsInterrupted())
                    continue;
                
                throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
            }
        }

        //
        // The fd_ member may be -1 at this point. If so, recv will fail.
        //
        int result = ::recv(fd_, (char*)buf -> rest(), sizeOfPacket, 0);

        if(result == 0)
        {
            return false;
        }
        else if(result == -1)
        {
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
	        sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(fd_ == -1 || ConnectionDown())
		return false;

	    if(WouldBlock())
		continue;

            throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
        }

        buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }

    return true;
}

void
OCI::IIOP::Transport_impl::send(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
	int result = ::send(fd_, (const char*)buf -> rest(), sizeOfPacket, 0);

	if(result == 0)
	{
	    throw COMM_FAILURE("", MinorSendZero, COMPLETED_NO);
	}
	else if(result == -1)
	{
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
		sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(WouldBlock())
	    {
		if(!block)
		    return;
		
		setBlock(true);
		continue;
	    }

	    throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
	}

	buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }
}

Boolean
OCI::IIOP::Transport_impl::send_detect(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
#ifdef WIN32
        //          
        // Limit the maximum size of a message, in order to avoid
        // a performance problem on Win32.
        //      
        // Note: The maximum size must be less than the socket's
        // send buffer size set by the constructor.
        //  
        // TODO: This is a total hack!
        //
        ULong toSend = sizeOfPacket;
        if(sizeOfPacket > 65535)
            toSend = 65535;

        int result = ::send(fd_, (const char*)buf -> rest(), toSend, 0);
#else
        int result = ::send(fd_, (const char*)buf -> rest(), sizeOfPacket, 0);
#endif

	if(result == 0)
	{
	    return false;
	}
	else if(result == -1)
	{
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
		sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(fd_ == -1 || ConnectionDown())
		return false;

	    if(WouldBlock())
	    {
		if(!block)
		    return true;
		
		setBlock(true);
		continue;
	    }

	    throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
	}

	buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }

    return true;
}

void
OCI::IIOP::Transport_impl::send_timeout(OCI::Buffer_ptr buf, ULong t)
    throw(SystemException)
{
    if(t == 0 || fd_ == -1 /*TODO*/)
    {
	send(buf, false);
	return;
    }

    setBlock(false);
    
    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
	fd_set fdSet;
	FD_ZERO(&fdSet);

        //
        // It's possible for the fd_ member to become -1 if this
        // transport is closed by another thread
        //
        int fd = fd_;
        if(fd != -1)
        {
            //
            // Use fd here, otherwise FD_SET will crash if given a
            // file descriptor of -1
            //
            FD_SET(fd, &fdSet);

            struct timeval tv;
            tv.tv_sec = t / 1000;
            tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

            //
            // Use fd here, since if fd_ == -1, select will simply
            // block for the specified timeout period
            //
            int resultS = ::select(fd + 1, 0, &fdSet, 0, &tv);
            
            if(resultS == 0)
            {
                return;
            }
            else if(resultS == -1)
            {
                if(IsInterrupted())
                    continue;
                
                throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
            }
        }

        //
        // The fd_ member may be -1 at this point. If so, send will fail.
        //
        int result = ::send(fd_, (const char*)buf -> rest(), sizeOfPacket, 0);

        if(result == 0)
        {
            throw COMM_FAILURE("", MinorSendZero, COMPLETED_NO);
        }
        else if(result == -1)
        {
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
		sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(WouldBlock())
		return;

            throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
        }

        buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }
}

Boolean
OCI::IIOP::Transport_impl::send_timeout_detect(OCI::Buffer_ptr buf, ULong t)
    throw(SystemException)
{
    if(t == 0 || fd_ == -1 /*TODO*/)
	return send_detect(buf, false);

    setBlock(false);
    
    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
	fd_set fdSet;
	FD_ZERO(&fdSet);

        //
        // It's possible for the fd_ member to become -1 if this
        // transport is closed by another thread
        //
        int fd = fd_;
        if(fd != -1)
        {
            //
            // Use fd here, otherwise FD_SET will crash if given a
            // file descriptor of -1
            //
            FD_SET(fd, &fdSet);

            struct timeval tv;
            tv.tv_sec = t / 1000;
            tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

            //
            // Use fd here, since if fd_ == -1, select will simply
            // block for the specified timeout period
            //
            int resultS = ::select(fd + 1, 0, &fdSet, 0, &tv);
            
            if(resultS == 0)
            {
                return true;
            }
            else if(resultS == -1)
            {
                if(IsInterrupted())
                    continue;
                
                throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
            }
        }

        //
        // The fd_ member may be -1 at this point. If so, send will fail.
        //
        int result = ::send(fd_, (const char*)buf -> rest(), sizeOfPacket, 0);

        if(result == 0)
        {
            return false;
        }
        else if(result == -1)
        {
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
		sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(fd_ == -1 || ConnectionDown())
		return false;

	    if(WouldBlock())
		continue;

            throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
        }

        buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }

    return true;
}

OCI::TransportInfo_ptr
OCI::IIOP::Transport_impl::get_info()
    throw(SystemException)
{
    return OCI::TransportInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// Connector_impl private and protected member implementations
// ----------------------------------------------------------------------

void
OCI::IIOP::Connector_impl::close()
{
    //
    // Destroy the info object
    //
    info_ -> _OB_destroy();

    //
    // Close the socket
    //
    int saveFd = fd_;
    fd_ = -1; // Must be set to -1 before the close
    CloseSocket(saveFd);
}

// ----------------------------------------------------------------------
// OCI::IIOP::Connector_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::Connector_impl::Connector_impl(const char* host,
					  UShort port,
                                          bool keepAlive,
					  const OCI::ConnectCBSeq& cb)
    : host_(host), port_(port), keepAlive_(keepAlive), fd_(-1),
      info_(new ConnectorInfo_impl(this, cb))
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "OCI::IIOP::Connector_impl" << OB_ENDL;
#endif
}

OCI::IIOP::Connector_impl::~Connector_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~OCI::IIOP::Connector_impl" << OB_ENDL;
#endif

    if(fd_ >= 0)
	close();
    
    release(info_);
}

// ----------------------------------------------------------------------
// Connector_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::Connector_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

OCI::Transport_ptr
OCI::IIOP::Connector_impl::connect()
    throw(SystemException)
{
    if(fd_ >= 0)
	close();

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	fd_ = -1;
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Set SO_KEEPALIVE option
    //
    if(keepAlive_)
    {
        int flag = 1;
        if(setsockopt(fd_, SOL_SOCKET, SO_KEEPALIVE, (char*)&flag, sizeof(int))
           == -1)
        {
            int err = GetError();
            CloseSocket(fd_);
            fd_ = -1;
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
        }
    }

    //
    // Get the address
    //
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port_);
    GetInAddr(host_, addr);

    //
    // Connect
    //
    while(true)
    {
        if(::connect(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &addr),
		     sizeof(addr)) == -1)
        {
            if(IsInterrupted())
                continue;
	    
	    int err = GetError();
	    CloseSocket(fd_);
	    fd_ = -1;
	    SetError(err);
	    if(ConnectFailed())
		throw TRANSIENT(LastError(), MinorConnectFailed, COMPLETED_NO);
	    else
		throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);
        }

        break;
    }
    
    //
    // Create new transport
    //
    OCI::Transport_var tr = new Transport_impl(this, fd_);
    fd_ = -1;

    //
    // Call callbacks
    //
    OCI::TransportInfo_var trInfo = tr -> get_info();
    try
    {
	info_ -> _OB_callConnectCB(trInfo);
    }
    catch(const SystemException&)
    {
	tr -> close();
	throw;
    }

    //
    // Return new transport
    //
    return tr._retn();
}

OCI::Transport_ptr
OCI::IIOP::Connector_impl::connect_timeout(ULong t)
    throw(SystemException)
{
    if(fd_ >= 0)
	close();

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	fd_ = -1;
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Set socket to non-blocking mode
    //
    SetNoBlock(fd_);
   
    //
    // Get the address
    //
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port_);
    GetInAddr(host_, addr);

    //
    // Connect
    //
    while(true)
    {
        if(::connect(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &addr),
		     sizeof(addr)) == -1)
        {
            if(IsInterrupted())
                continue;

	    if(InProgress())
	    {
		//
		// Call select() until timeout expired or connection
		// fd_ becomes writeable.
		//
		while(true)
		{
		    
		    struct timeval tv;
		    tv.tv_sec = t / 1000;
		    tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

		    fd_set writeFdSet;
		    FD_ZERO(&writeFdSet);
		    FD_SET(fd_, &writeFdSet);

#ifndef WIN32
		    int resultS = ::select(fd_ + 1, 0, &writeFdSet, 0, &tv);
#else
		    //
		    // Under WIN32 ::connect failure notification is
		    // done through the exceptfds.
		    //
		    fd_set exceptFdSet;
		    FD_ZERO(&exceptFdSet);
		    FD_SET(fd_, &exceptFdSet);

		    int resultS = ::select(fd_ + 1, 0, &writeFdSet,
					   &exceptFdSet, &tv);
#endif
		    
		    if(resultS == 0)
		    {
			CloseSocket(fd_);
			fd_ = -1;
			return Transport::_nil();
		    }
		    else if(resultS == -1)
		    {
			if(IsInterrupted())
			    continue;
			
			throw COMM_FAILURE(LastError(), MinorSelect,
					   COMPLETED_NO);
		    }
		    
#ifdef WIN32
		    //
		    // Windows seems to need this call to Sleep(),
		    // otherwise no error is reported through
		    // getsockopt. I have no idea why.
		    //
		    Sleep(0);
#endif

		    //
		    // Check for connect error
		    //
		    socklen_type optlen = sizeof(socklen_type);
		    int optval;
		    if(getsockopt(fd_, SOL_SOCKET, SO_ERROR, (char*)&optval,
				  &optlen) == -1)
		    {
			int err = GetError();
			CloseSocket(fd_);
			fd_ = -1;
			SetError(err);
			throw COMM_FAILURE(LastError(), MinorGetsockopt,
					   COMPLETED_NO);
		    }
		    
		    if(optval > 0)
		    {
			CloseSocket(fd_);
			fd_ = -1;
			SetError(optval);
			if(ConnectFailed())
			    throw TRANSIENT(LastError(), MinorConnectFailed,
					    COMPLETED_NO);
			else
			    throw COMM_FAILURE(LastError(), MinorConnect,
					       COMPLETED_NO);
		    }
		    
		    break;
		}
	    }
	    else
	    {
		int err = GetError();
		CloseSocket(fd_);
		fd_ = -1;
		SetError(err);

		//
		// On Solaris (at least) if our connection is refused (that is
		// the server is down) InProgress() returns false so we need to
		// check for connection failure here as well.
		//
		if(ConnectFailed())
		    throw TRANSIENT(LastError(), MinorConnectFailed,
				    COMPLETED_NO);
		else
		    throw COMM_FAILURE(LastError(), MinorConnect,
		    		       COMPLETED_NO);
	    }
        }
	
        break;
    }
    
    //
    // Create new transport
    //
    OCI::Transport_var tr = new Transport_impl(this, fd_);
    fd_ = -1;

    //
    // Call callbacks
    //
    OCI::TransportInfo_var trInfo = tr -> get_info();
    try
    {
	info_ -> _OB_callConnectCB(trInfo);
    }
    catch(const SystemException&)
    {
	tr -> close();
	throw;
    }

    //
    // Return new transport
    //
    return tr._retn();
}

OCI::ProfileInfoSeq*
OCI::IIOP::Connector_impl::get_usable_profiles(
    const IOP::IOR& ior,
    const CORBA::PolicyList& policies)
    throw(SystemException)
{
    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;
    
    //
    // Make sure that the set of policies is met
    //
    for(ULong i = 0; i < policies.length(); i++)
    {
        if(policies[i] -> policy_type() == PROTOCOL_POLICY_ID)
        {
            ProtocolPolicy_var protocolPolicy =
                ProtocolPolicy::_narrow(policies[i]);
            assert(!is_nil(protocolPolicy));
            if(!protocolPolicy -> contains(PLUGIN_ID))
                return seq._retn();
        }
    }

    IIOPExtractAllProfileInfos(ior, seq.inout(), true, host_, port_, false);

    return seq._retn();
}

CORBA::Boolean
OCI::IIOP::Connector_impl::equal(Connector_ptr con)
    throw(CORBA::SystemException)
{
    Connector_impl* impl = dynamic_cast<OCI::IIOP::Connector_impl*>(con);
    if(!impl)
	return false;
    
    //
    // Compare ports
    //
    if(port_ != impl -> port_)
	return false;
    
    //
    // Direct host name comparison
    //
    if(strcmp(host_, impl -> host_) != 0)
    {
	//
	//
	// Direct host name comparision failed - must look up
	// addresses to be really sure if the hosts differ
	//
	try
	{
	    struct sockaddr_in addr1;
	    GetInAddr(host_, addr1);
	    
	    struct sockaddr_in addr2;
	    GetInAddr(impl -> host_, addr2);
	    
	    if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
		return false;
	}
	catch(const COMM_FAILURE&)
	{
	    //
	    // Return false on hostname lookup failure
	    //
	    return false;
	}
    }

    //
    // OK, connectors are the same
    //
    return true;
}    

OCI::ConnectorInfo_ptr
OCI::IIOP::Connector_impl::get_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// Acceptor_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::Acceptor_impl::Acceptor_impl(const StringSeq& hosts,
                                        bool multiProfile,
                                        UShort port, UShort backlog,
                                        bool keepAlive)
    : hosts_(hosts), multiProfile_(multiProfile), port_(port),
      keepAlive_(keepAlive), fd_(-1), info_(new AcceptorInfo_impl(this))
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "OCI::IIOP::Acceptor_impl" << OB_ENDL;
#endif

    if(backlog == 0)
        backlog_ = 5;
    else
        backlog_ = backlog;

    //
    // Get the local address for use by connect_self. Since we are
    // binding to all network interfaces, we'll use the loopback
    // address.
    //
    memset(&localAddress_, 0, sizeof(localAddress_));
    localAddress_.sin_family = AF_INET;
    localAddress_.sin_port = htons(port_);
    localAddress_.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

    int no;

    //
    // Get the bind address
    //
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port_);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	_OB_setRef(0);
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Bind to all network interfaces
    //
#ifndef WIN32
    no = 1;
    if(setsockopt(fd_, SOL_SOCKET, SO_REUSEADDR, (char*)&no,
		  sizeof(int)) == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
#endif
	
    if(bind(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &addr),
	    sizeof(addr)) == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorBind, COMPLETED_NO);
    }

    socklen_type len = sizeof(addr);
    if(getsockname(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &addr),
		   &len) == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorGetsockname, COMPLETED_NO);
    }

    //
    // Read back the port. This is needed if the port was selected by
    // the operating system.
    //
    port_ = ntohs(addr.sin_port);
    localAddress_.sin_port = addr.sin_port;
}

OCI::IIOP::Acceptor_impl::Acceptor_impl(const char* address,
                                        const StringSeq& hosts,
                                        bool multiProfile,
                                        UShort port, UShort backlog,
                                        bool keepAlive)
    : hosts_(hosts), multiProfile_(multiProfile), port_(port),
      backlog_(backlog), keepAlive_(keepAlive), fd_(-1),
      info_(new AcceptorInfo_impl(this))
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "OCI::IIOP::Acceptor_impl" << OB_ENDL;
#endif

    if(backlog == 0)
        backlog_ = 5;
    else
        backlog_ = backlog;

    //
    // Get the local address for use by connect_self
    //
    memset(&localAddress_, 0, sizeof(localAddress_));
    localAddress_.sin_family = AF_INET;
    localAddress_.sin_port = htons(port_);
    GetInAddr(address, localAddress_);

    int no;

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	_OB_setRef(0);
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Bind to requested network interface
    //
#ifndef WIN32
    no = 1;
    if(setsockopt(fd_, SOL_SOCKET, SO_REUSEADDR, (char*)&no,
		  sizeof(int)) == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
#endif
	
    if(bind(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &localAddress_),
	    sizeof(localAddress_)) == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorBind, COMPLETED_NO);
    }

    socklen_type len = sizeof(localAddress_);
    if(getsockname(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &localAddress_),
		   &len) == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorGetsockname, COMPLETED_NO);
    }

    //
    // Read back the port. This is needed if the port was selected by
    // the operating system.
    //
    port_ = ntohs(localAddress_.sin_port);
}

OCI::IIOP::Acceptor_impl::~Acceptor_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~Acceptor_impl" << OB_ENDL;
#endif

    if(fd_ >= 0)
	close();

    release(info_);
}

// ----------------------------------------------------------------------
// Acceptor_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::Acceptor_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

void
OCI::IIOP::Acceptor_impl::close()
    throw(SystemException)
{
    //
    // Destroy the info object
    //
    info_ -> _OB_destroy();

    //
    // Close the socket
    //
    int saveFd = fd_;
    fd_ = -1;
    CloseSocket(saveFd);
}

void
OCI::IIOP::Acceptor_impl::shutdown()
    throw(SystemException)
{
    //
    // shutdown doesn't work on passive socket. It doesn't seem
    // possible to do an "unlisten" to get time to accept the backlog
    // and shutdown gracefully connections, so there's some small
    // chances that a client may get a COMM_FAILURE exception if it
    // tries to connect to the server when it's shutting
    // down. Fortunately, because of locate requests and retries it's
    // *very unlikely* that it happens.
    //
    // ::shutdown(fd_, SHUT_RD); // Shutdown receive side only
}

void
OCI::IIOP::Acceptor_impl::listen()
    throw(SystemException)
{
    //
    // Listen
    //
    while(true)
    {
	if(::listen(fd_, backlog_) == -1)
	{
            if(IsInterrupted())
		continue;
	    int err = GetError();
	    CloseSocket(fd_);
	    fd_ = -1;
	    SetError(err);
	    throw COMM_FAILURE(LastError(), MinorListen, COMPLETED_NO);
	}
	break;
    }
}

OCI::Transport_ptr
OCI::IIOP::Acceptor_impl::accept(Boolean block)
    throw(SystemException)
{
    if(block)
	SetBlock(fd_);
    else
	SetNoBlock(fd_);

    //
    // Accept
    //
    struct sockaddr_in newAddr;
    socklen_type len = sizeof(newAddr);
    int newFd;
    while(true)
    {
	if((newFd = ::accept(fd_,
			     OB_REINTERPRET_CAST(struct sockaddr*, &newAddr),
			     &len)) == -1)
        {
	    //
	    // Here we have special handling for this case, since
	    // accept can generate other loop-and-retry errors on some
	    // platforms.
	    //
            if(AcceptRetry())
                continue;
	    
	    if(WouldBlock())
	    {
		if(!block)
		    return Transport::_nil();
		
		SetBlock(fd_);
		continue;
	    }

	    //
	    // Accept failed, throw an exception
	    //
            throw COMM_FAILURE(LastError(), MinorAccept, COMPLETED_NO);
        }
	
        break;
    }

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(newFd, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(newFd);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Set SO_KEEPALIVE option
    //
    if(keepAlive_)
    {
        int flag = 1;
        if(setsockopt(newFd, SOL_SOCKET, SO_KEEPALIVE, (char*)&flag,
                      sizeof(int)) == -1)
        {
            int err = GetError();
            CloseSocket(newFd);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
        }
    }
    
    //
    // Create new transport
    //
    OCI::Transport_var tr = new Transport_impl(this, newFd);

    //
    // Call callbacks
    //
    OCI::TransportInfo_var trInfo = tr -> get_info();
    try
    {
	info_ -> _OB_callAcceptCB(trInfo);
    }
    catch(const SystemException&)
    {
	tr -> close();
	throw;
    }

    //
    // Return new transport
    //
    return tr._retn();
}

OCI::Transport_ptr
OCI::IIOP::Acceptor_impl::connect_self()
    throw(SystemException)
{
    //
    // Create socket
    //
    int fd;
    if((fd = CreateSocket()) == -1)
    {
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
   
    //
    // Connect to local address
    //
    while(true)
    {
        if(::connect(fd, OB_REINTERPRET_CAST(struct sockaddr*, &localAddress_),
		     sizeof(localAddress_)) == -1)
        {
            if(IsInterrupted())
                continue;
	    
	    int err = GetError();
	    CloseSocket(fd);
	    SetError(err);
	    if(ConnectFailed())
		throw TRANSIENT(LastError(), MinorConnectFailed, COMPLETED_NO);
	    else
		throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);
        }

        break;
    }
    
    //
    // Create and return new transport
    //
    return new Transport_impl(this, fd);
}

void
OCI::IIOP::Acceptor_impl::add_profiles(const OCI::ProfileInfo& profileInfo,
                                       OCI::IOR& ior)
    throw(SystemException)
{
    assert(port_ > 0);
    
    //
    // Filter components according to IIOP version
    //
    OCI::TaggedComponentSeq components;
    if(profileInfo.major == 1 && profileInfo.minor == 0)
    {
        //
        // No components for IIOP 1.0
        //
    }
    else
    {
        components.length(profileInfo.components.length());
        for(ULong i = 0 ; i < profileInfo.components.length() ; i++)
            components[i] = profileInfo.components[i];
    }

    if(profileInfo.major == 1 && profileInfo.minor == 0)
    {
        //
        // For IIOP 1.0, we always add one profile for each host,
        // since IIOP 1.0 doesn't support tagged components in a
        // profile
        //
        for(ULong i = 0 ; i < hosts_.length() ; i++)
        {
            ::IIOP::ProfileBody_1_0 body;
            body.iiop_version.major = profileInfo.major;
            body.iiop_version.minor = profileInfo.minor;
            body.host = hosts_[i];
            body.port = port_;
            body.object_key.length(profileInfo.key.length());
            memcpy(body.object_key.get_buffer(), profileInfo.key.get_buffer(),
                   profileInfo.key.length());

            ULong len = ior.profiles.length() + 1;
            ior.profiles.length(len);
            ior.profiles[len - 1].tag = tag();

            OCI::BufferImpl buf;
            OutputStreamImpl out(&buf);
            out._OB_writeEndian();
            body._OB_marshal(&out);
            ULong count = out._OB_pos();
            ior.profiles[len - 1].profile_data.length(count);
            Octet* oct = ior.profiles[len - 1].profile_data.get_buffer();
            memcpy(oct, buf.data(), count);
        }
    }
    else
    {
        if(multiProfile_)
        {
            //
            // Add one profile for each host
            //

            for(ULong i = 0 ; i < hosts_.length() ; i++)
            {
                ::IIOP::ProfileBody_1_1 body;
                body.iiop_version.major = profileInfo.major;
                body.iiop_version.minor = profileInfo.minor;
                body.host = hosts_[i];
                body.port = port_;
                body.object_key.length(profileInfo.key.length());
                memcpy(body.object_key.get_buffer(),
                       profileInfo.key.get_buffer(),
                       profileInfo.key.length());
                body.components.length(components.length());
                for(ULong j = 0 ; j < components.length() ; j++)
                    body.components[j] = components[j];
                
                ULong len = ior.profiles.length() + 1;
                ior.profiles.length(len);
                ior.profiles[len - 1].tag = tag();

                OCI::BufferImpl buf;
                OutputStreamImpl out(&buf);
                out._OB_writeEndian();
                body._OB_marshal(&out);
                ULong count = out._OB_pos();
                ior.profiles[len - 1].profile_data.length(count);
                Octet* oct = ior.profiles[len - 1].profile_data.get_buffer();
                memcpy(oct, buf.data(), count);
            }
        }
        else
        {
            //
            // Add a single tagged profile. If there are additional
            // hosts, add a tagged component for each host.
            //

            ::IIOP::ProfileBody_1_1 body;
            body.iiop_version.major = profileInfo.major;
            body.iiop_version.minor = profileInfo.minor;
            body.host = hosts_[0];
            body.port = port_;
            body.object_key.length(profileInfo.key.length());
            memcpy(body.object_key.get_buffer(), profileInfo.key.get_buffer(),
                   profileInfo.key.length());

            ULong i;

            body.components.length(components.length());
            for(i = 0 ; i < components.length() ; i++)
                body.components[i] = components[i];
            
            for(i = 1 ; i < hosts_.length() ; i++)
            {
                ULong len = body.components.length() + 1;
                body.components.length(len);
                body.components[len - 1].tag = IOP::TAG_ALTERNATE_IIOP_ADDRESS;
                OCI::BufferImpl buf;
                OutputStreamImpl out(&buf);
                out._OB_writeEndian();
                out.write_string(hosts_[i]);
                out.write_ushort(body.port);
                ULong count = out._OB_pos();
                body.components[len - 1].component_data.length(count);
                Octet* oct =
                    body.components[len - 1].component_data.get_buffer();
                memcpy(oct, buf.data(), count);
            }

            ULong len = ior.profiles.length() + 1;
            ior.profiles.length(len);
            ior.profiles[len - 1].tag = tag();

            OCI::BufferImpl buf;
            OutputStreamImpl out(&buf);
            out._OB_writeEndian();
            body._OB_marshal(&out);
            ULong count = out._OB_pos();
            ior.profiles[len - 1].profile_data.length(count);
            Octet* oct = ior.profiles[len - 1].profile_data.get_buffer();
            memcpy(oct, buf.data(), count);
        }
    }
}

OCI::ProfileInfoSeq*
OCI::IIOP::Acceptor_impl::get_local_profiles(const OCI::IOR& ior)
    throw(SystemException)
{
    //
    // Get local profiles for all hosts
    //
    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;

    for(ULong i = 0 ; i < hosts_.length() ; i++)
        IIOPExtractAllProfileInfos(ior, seq.inout(), true, hosts_[i], port_,
                                   true);

    return seq._retn();
}

OCI::AcceptorInfo_ptr
OCI::IIOP::Acceptor_impl::get_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// AccFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::AccFactory_impl::AccFactory_impl()
    : info_(new AccFactoryInfo_impl)
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "AccFactory_impl" << OB_ENDL;
#endif
}

OCI::IIOP::AccFactory_impl::~AccFactory_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~AccFactory_impl" << OB_ENDL;
#endif
    release(info_);
}

// ----------------------------------------------------------------------
// AccFactory_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::AccFactory_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

OCI::AccFactoryInfo_ptr
OCI::IIOP::AccFactory_impl::get_info()
    throw(SystemException)
{
    return OCI::AccFactoryInfo::_duplicate(info_);
}

OCI::Acceptor_ptr
OCI::IIOP::AccFactory_impl::create_acceptor(const ParamSeq& params)
    throw(OCI::InvalidParam,
          SystemException)
{
    String_var bind;
    StringSeq hosts;
    bool keepAlive = true;
    bool multiProfile = false;
    UShort port = 0;
    UShort backlog = 0;
    bool numeric = false;

    ULong i = 0;
    while(i < params.length())
    {
        if(strcmp(params[i], "--backlog") == 0)
        {
            if(i + 1 >= params.length())
                throw InvalidParam("argument expected for --backlog");
            const char* arg = params[i + 1];
            for(const char* cp = arg ; *cp != '\0' ; cp++)
                if(!isdigit(*cp))
                    throw InvalidParam("invalid argument for --backlog");
            ULong n = atoi(arg);
            if(n < 1 || n > 65535)
                throw InvalidParam("invalid backlog");
            backlog = (UShort)n;
            i += 2;
        }
        else if(strcmp(params[i], "--bind") == 0)
        {
            if(i + 1 >= params.length())
                throw InvalidParam("argument expected for --bind");
            bind = params[i + 1];
            i += 2;
        }
        else if(strcmp(params[i], "--host") == 0)
        {
            if(i + 1 >= params.length())
                throw InvalidParam("argument expected for --host");
            const char* start = params[i + 1];
            while(true)
            {
                while(*start != '\0' && isspace(*start))
                    start++;
                if(*start == '\0')
                    break;
                const char* cp = strchr(start, ',');
                if(cp == start)
                    start++;
                else
                {
                    if(cp == 0)
                        cp = start + strlen(start);
                    cp--;
                    while(isspace(*cp))
                        cp--;
                    ULong strLen = cp - start + 1;
                    ULong idx = hosts.length();
                    hosts.length(idx + 1);
                    char* str = string_alloc(strLen);
                    strncpy(str, start, strLen);
                    str[strLen] = '\0';
                    hosts[idx] = str;
                    start = cp + 1;
                }
            }
            if(hosts.length() == 0)
                throw InvalidParam("invalid argument for --host");
            i += 2;
        }
        else if(strcmp(params[i], "--multi-profile") == 0)
        {
            multiProfile = true;
            i++;
        }
        else if(strcmp(params[i], "--no-keepalive") == 0)
        {
            keepAlive = false;
            i++;
        }
        else if(strcmp(params[i], "--numeric") == 0)
        {
            numeric = true;
            i++;
        }
        else if(strcmp(params[i], "--port") == 0)
        {
            if(i + 1 >= params.length())
                throw InvalidParam("argument expected for --port");
            const char* arg = params[i + 1];
            for(const char* cp = arg ; *cp != '\0' ; cp++)
                if(!isdigit(*cp))
                    throw InvalidParam("invalid argument for --port");
            ULong n = atoi(arg);
            if(n < 1 || n > 65535)
                throw InvalidParam("invalid port");
            port = (UShort)n;
            i += 2;
        }
        else
        {
            String_var err = string_dup("unknown parameter: ");
            err += params[i];
            throw InvalidParam(err);
        }
    }

    if(hosts.length() == 0)
    {
        hosts.length(1);
        hosts[0] = GetCanonicalHostname(numeric);
    }

    if(bind.in() == 0)
        return new Acceptor_impl(hosts, multiProfile, port, backlog,
                                 keepAlive);
    else
        return new Acceptor_impl(bind, hosts, multiProfile, port, backlog,
                                 keepAlive);
}

void
OCI::IIOP::AccFactory_impl::change_key(IOP::IOR& ior, const ObjectKey& key)
    throw(SystemException)
{
    //
    // Extract the IIOP profile information from the provided IOR
    //
    for(ULong profile = 0 ; profile < ior.profiles.length() ; profile++)
    {
	if(ior.profiles[profile].tag == IOP::TAG_INTERNET_IOP)
	{
	    //
	    // Extract the 1_0 profile body
	    //
	    const Octet* coct =
		ior.profiles[profile].profile_data.get_buffer();
	    ULong len = ior.profiles[profile].profile_data.length();
	    OCI::BufferImpl buf((Octet*)coct, len);
	    InputStreamImpl in(&buf, 0, false);
	    in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);

            //
            // Read components if the IIOP version is > 1.0
            //
            OCI::TaggedComponentSeq components;
            if(body.iiop_version.major > 1 || body.iiop_version.minor > 0)
            {
                len = in.read_ulong();
                components.length(len);
                IOP::TaggedComponent* buf = components.get_buffer();
                for(ULong i = 0 ; i < len ; i++)
                    IOP::TaggedComponent::_OB_unmarshal(buf[i], &in);
            }

	    //
	    // Fill in the new object-key
	    //
	    body.object_key.length(key.length());
	    memcpy(body.object_key.get_buffer(), key.get_buffer(),
		   key.length());

	    //
	    // Remarshal the new body
	    //
	    OCI::BufferImpl buf2;
	    OutputStreamImpl out(&buf2);
	    out._OB_writeEndian();
	    body._OB_marshal(&out);

	    //
	    // Remarshal the components if the IIOP version is > 1.0
	    //
            if(body.iiop_version.major > 1 || body.iiop_version.minor > 0)
            {
		out.write_ulong(components.length());
		for(ULong i = 0 ; i < components.length() ; i++)
		    components[i]._OB_marshal(&out);
	    }
	    ULong count = out._OB_pos();
	    ior.profiles[profile].profile_data.length(count);
	    Octet* oct = ior.profiles[profile].profile_data.get_buffer();
	    memcpy(oct, buf2.data(), count);
	}
    }
}

// ----------------------------------------------------------------------
// ConFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::ConFactory_impl::ConFactory_impl(bool keepAlive)
    : keepAlive_(keepAlive), info_(new ConFactoryInfo_impl)
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "ConFactory_impl" << OB_ENDL;
#endif
}

OCI::IIOP::ConFactory_impl::~ConFactory_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~ConFactory_impl" << OB_ENDL;
#endif
    release(info_);
}

// ----------------------------------------------------------------------
// ConFactory_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::ConFactory_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::IIOP::ConFactory_impl::describe_profile(
    const IOP::TaggedProfile& profile)
    throw(CORBA::SystemException)
{
    assert(profile.tag == IOP::TAG_INTERNET_IOP);

    //
    // Get the IIOP profile body
    //
    const Octet* coct = profile.profile_data.get_buffer();
    ULong len = profile.profile_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    ::IIOP::ProfileBody_1_0 body;
    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);

    String_var result = string_dup("");

    //
    // Show general info
    //
    result += "iiop_version: ";
    result += int(body.iiop_version.major);
    result += '.';
    result += int(body.iiop_version.minor);
    result += '\n';
    result += "host: ";
    result += body.host;
    result += '\n';
    result += "port: ";
    result += body.port;
    result += '\n';
    result += "object_key: (";
    result += body.object_key.length();
    result += ")\n";
    String_var tmp = dump_octets(body.object_key.get_buffer(),
                                 body.object_key.length());
    result += tmp;

    //
    // Print IIOP 1.1 information (components)
    //
    if(body.iiop_version.major > 1 || body.iiop_version.minor >= 1)
    {
        ULong l = in.read_ulong();

        for(ULong i = 0 ; i < l ; i++)
        {
            IOP::TaggedComponent component;
            IOP::TaggedComponent::_OB_unmarshal(component, &in);

            String_var desc = describe_component(component);
            result += desc;
	}
    }

    return result._retn();
}

OCI::ConnectorSeq*
OCI::IIOP::ConFactory_impl::create_connectors(const OCI::IOR& ior,
					      const PolicyList& policies)
    throw(SystemException)
{
    ULong i;

    //
    // Check whether policies are satisfied
    //
    for(i = 0; i < policies.length(); i++)
    {
	if(policies[i] -> policy_type() == PROTOCOL_POLICY_ID)
	{
	    ProtocolPolicy_var protocolPolicy =
		ProtocolPolicy::_narrow(policies[i]);
	    assert(!is_nil(protocolPolicy));
            if(!protocolPolicy -> contains(PLUGIN_ID))
		return new OCI::ConnectorSeq;
	}
    }

    //
    // Create Connectors from profiles
    //
    OCI::ConnectorSeq_var seq = new OCI::ConnectorSeq;
    for(i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == tag())
	{
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
	    ULong len = ior.profiles[i].profile_data.length();
	    OCI::BufferImpl buf((Octet*)coct, len);
	    InputStreamImpl in(&buf, 0, false);
	    in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);
	    
	    //
	    // Create new connector for this profile
	    //
	    OCI::ConnectCBSeq_var cbs = info_ -> _OB_getConnectCBSeq();

	    CORBA::ULong seqLength  = seq -> length();
	    seq -> length(seqLength + 1);
	    seq[seqLength] = new Connector_impl(body.host, body.port,
                                                keepAlive_, cbs);

            //
            // If this is a 1.1 profile, check for
            // TAG_ALTERNATE_IIOP_ADDRESS in the components
            //
            if(body.iiop_version.major > 1 || body.iiop_version.minor > 0)
            {
		ULong i;

		//
		// Unmarshal the tagged components
		//
		TaggedComponentSeq components;
		ULong len = in.read_ulong();
		components.length(len);
		IOP::TaggedComponent* buf = components.get_buffer();
		for(i = 0 ; i < len ; i++)
		    IOP::TaggedComponent::_OB_unmarshal(buf[i], &in);

		//
		// Check for TAG_ALTERNATE_IIOP_ADDRESS
		//
		for(i = 0 ; i < components.length() ; i++)
		    if(components[i].tag == IOP::TAG_ALTERNATE_IIOP_ADDRESS)
		    {
			const Octet* coct =
			    components[i].component_data.get_buffer();
			ULong len = components[i].component_data.length();
			OCI::BufferImpl buf((Octet*)coct, len);
			InputStreamImpl in(&buf, 0, false);
			in._OB_readEndian();
			String_var host = in.read_string();
			UShort port = in.read_ushort();

			//
			// Create new connector for this component
			//
			OCI::ConnectCBSeq_var cbs =
			    info_ -> _OB_getConnectCBSeq();

			CORBA::ULong seqLength  = seq -> length();
			seq -> length(seqLength + 1);
			seq[seqLength] = new Connector_impl(host, port,
                                                            keepAlive_, cbs);
		    }
	    }
	}
    }

    return seq._retn();
}

Boolean
OCI::IIOP::ConFactory_impl::equivalent(const OCI::IOR& ior1,
				       const OCI::IOR& ior2)
    throw(SystemException)
{
    return IIOPEquivalent(ior1, ior2);
}

ULong
OCI::IIOP::ConFactory_impl::hash(const OCI::IOR& ior, ULong maximum)
    throw(SystemException)
{
    return IIOPHash(ior, maximum);
}

OCI::ConFactoryInfo_ptr
OCI::IIOP::ConFactory_impl::get_info()
    throw(SystemException)
{
    return ConFactoryInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// TransportInfo_impl private and protected member implementation
// ----------------------------------------------------------------------

void
OCI::IIOP::TransportInfo_impl::init()
{
    //
    // I have to save the local and remote address and port, since I
    // get garbage if I calculate these values "on the fly" during the
    // close callback. getsockname or getpeername may fail here if the
    // socket has been shutdown by the peer.
    //

    socklen_type len = sizeof(struct sockaddr_in);

    //
    // Get the local address and port
    //
    if(getsockname(transport_ -> fd_,
		   OB_REINTERPRET_CAST(struct sockaddr*, &addr_),
		   &len) == -1)
    {
	throw COMM_FAILURE(LastError(), MinorGetsockname, COMPLETED_NO);
    }
    

    //
    // Get the remote address and port
    //
    if(getpeername(transport_ -> fd_,
		   OB_REINTERPRET_CAST(struct sockaddr*, &remoteAddr_),
		   &len) == -1)
    {
	throw COMM_FAILURE(LastError(), MinorGetpeername, COMPLETED_NO);
    }
}

// ----------------------------------------------------------------------
// TransportInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::TransportInfo_impl::TransportInfo_impl(
    Transport_impl* transport,
    Connector_impl* connector)
    : transport_(transport)
{
    try
    {
	connectorInfo_ = connector -> get_info();
	
	//
	// Initialize some data members
	//
	init();
    }
    catch(...)
    {
	_OB_setRef(0);
	throw;
    }
}

OCI::IIOP::TransportInfo_impl::TransportInfo_impl(Transport_impl* transport,
                                       Acceptor_impl* acceptor)
    : transport_(transport)
{
    try
    {
	acceptorInfo_ = acceptor -> get_info();
	
	//
	// Initialize some data members
	//
	init();
    }
    catch(...)
    {
	_OB_setRef(0);
	throw;
    }
}

OCI::IIOP::TransportInfo_impl::~TransportInfo_impl()
{
}

// ----------------------------------------------------------------------
// TransportInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::TransportInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::IIOP::TransportInfo_impl::describe()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    InetAddr_var localAddr = addr();
    UShort localPort = port();
    desc += "\nlocal address: ";
    desc += (int)localAddr[0];
    desc += ".";
    desc += (int)localAddr[1];
    desc += ".";
    desc += (int)localAddr[2];
    desc += ".";
    desc += (int)localAddr[3];
    desc += ":";
    desc += localPort;

    InetAddr_var remoteAddr = remote_addr();
    UShort remotePort = remote_port();
    desc += "\nremote address: ";
    desc += (int)remoteAddr[0];
    desc += ".";
    desc += (int)remoteAddr[1];
    desc += ".";
    desc += (int)remoteAddr[2];
    desc += ".";
    desc += (int)remoteAddr[3];
    desc += ":";
    desc += remotePort;

    return desc._retn();
}

OCI::ConnectorInfo_ptr
OCI::IIOP::TransportInfo_impl::connector_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_duplicate(connectorInfo_);
}

OCI::AcceptorInfo_ptr
OCI::IIOP::TransportInfo_impl::acceptor_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(acceptorInfo_);
}

void
OCI::IIOP::TransportInfo_impl::add_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin();
	iter != closeCBs_.end();
	iter++)
    {
	if((*iter).in() == cb)
	    return;
    }

    closeCBs_.push_back(OCI::CloseCB::_duplicate(cb));
}

void
OCI::IIOP::TransportInfo_impl::remove_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);
	
    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin();
	iter != closeCBs_.end();
	iter++)
    {
	if((*iter).in() == cb)
	{
	    closeCBs_.erase(iter);
	    return;
	}
    }
}

OCI::IIOP::InetAddr_slice*
OCI::IIOP::TransportInfo_impl::addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
	throw NO_RESOURCES();

    InetAddr_slice* addr = new InetAddr;
    memcpy(addr, (const char*)&addr_.sin_addr.s_addr, 4);
    return addr;
}

UShort
OCI::IIOP::TransportInfo_impl::port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
	throw NO_RESOURCES();

    return ntohs(addr_.sin_port);
}

OCI::IIOP::InetAddr_slice*
OCI::IIOP::TransportInfo_impl::remote_addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
	throw NO_RESOURCES();

    InetAddr_slice* addr = new InetAddr;
    memcpy(addr, (const char*)&remoteAddr_.sin_addr.s_addr, 4);
    return addr;
}

UShort
OCI::IIOP::TransportInfo_impl::remote_port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
	throw NO_RESOURCES();

    return ntohs(remoteAddr_.sin_port);
}

void
OCI::IIOP::TransportInfo_impl::_OB_callCloseCB(OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for (OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin();
         iter != closeCBs_.end();
	 iter++)
    {
        (*iter) -> close_cb(info);
    }
}

void
OCI::IIOP::TransportInfo_impl::_OB_destroy()
{
    JTCSynchronized synchronized(*this);

    transport_ = 0;
}

// ----------------------------------------------------------------------
// ConnectorInfo_impl constructor/destructor
// ----------------------------------------------------------------------

OCI::IIOP::ConnectorInfo_impl::ConnectorInfo_impl(Connector_impl* connector,
						  const OCI::ConnectCBSeq& cb)
    : connector_(connector)
{
    copy_from_seq(cb, OB_STL::back_inserter(connectCBs_));
}

OCI::IIOP::ConnectorInfo_impl::~ConnectorInfo_impl()
{
}

// ----------------------------------------------------------------------
// ConnectorInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::ConnectorInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::IIOP::ConnectorInfo_impl::describe()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    InetAddr_var remoteAddr = remote_addr();
    UShort remotePort = remote_port();
    desc += "\nremote address: ";
    desc += (int)remoteAddr[0];
    desc += ".";
    desc += (int)remoteAddr[1];
    desc += ".";
    desc += (int)remoteAddr[2];
    desc += ".";
    desc += (int)remoteAddr[3];
    desc += ":";
    desc += remotePort;

    return desc._retn();
}

void
OCI::IIOP::ConnectorInfo_impl::add_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin();
	iter != connectCBs_.end();
	iter++)
    {
	if((*iter).in() == cb)
	    return;
    }

    connectCBs_.push_back(OCI::ConnectCB::_duplicate(cb));
}

void
OCI::IIOP::ConnectorInfo_impl::remove_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);
	
    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin();
	iter != connectCBs_.end();
	iter++)
    {
	if((*iter).in() == cb)
	{
	    connectCBs_.erase(iter);
	    return;
	}
    }
}

OCI::IIOP::InetAddr_slice*
OCI::IIOP::ConnectorInfo_impl::remote_addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(connector_ == 0)
	throw NO_RESOURCES();

    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(connector_ -> port_);
    GetInAddr(connector_ -> host_, addr);
    InetAddr_var rc = new InetAddr;
    memcpy(rc.inout(), (const char*)&addr.sin_addr.s_addr, 4);

    return rc._retn();
}

UShort
OCI::IIOP::ConnectorInfo_impl::remote_port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(connector_ == 0)
	throw NO_RESOURCES();

    return connector_ -> port_;
}

void
OCI::IIOP::ConnectorInfo_impl::_OB_callConnectCB(OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for (OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin();
         iter != connectCBs_.end();
	 iter++)
    {
        (*iter) -> connect_cb(info);
    }
}

void
OCI::IIOP::ConnectorInfo_impl::_OB_destroy()
{
    JTCSynchronized synchronized(*this);

    connector_ = 0;
}

// ----------------------------------------------------------------------
// AcceptorInfo_impl constructor/destructor
// ----------------------------------------------------------------------

OCI::IIOP::AcceptorInfo_impl::AcceptorInfo_impl(Acceptor_impl* acceptor)
    : acceptor_(acceptor)
{
}

OCI::IIOP::AcceptorInfo_impl::~AcceptorInfo_impl()
{
}

// ----------------------------------------------------------------------
// AcceptorInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::AcceptorInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::IIOP::AcceptorInfo_impl::describe()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    InetAddr_var localAddr = addr();
    UShort localPort = port();
    desc += "\nlocal address: ";
    desc += (int)localAddr[0];
    desc += ".";
    desc += (int)localAddr[1];
    desc += ".";
    desc += (int)localAddr[2];
    desc += ".";
    desc += (int)localAddr[3];
    desc += ":";
    desc += localPort;

    StringSeq_var hostNames = hosts();
    desc += "\nhosts: ";
    for(ULong i = 0 ; i < hostNames -> length() ; i++)
    {
        if(i > 0)
            desc += ", ";
        desc += hostNames[i];
    }

    return desc._retn();
}

void
OCI::IIOP::AcceptorInfo_impl::add_accept_cb(OCI::AcceptCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<AcceptCB_var>::iterator iter = acceptCBs_.begin();
	iter != acceptCBs_.end();
	iter++)
    {
	if((*iter).in() == cb)
	    return;
    }

    acceptCBs_.push_back(OCI::AcceptCB::_duplicate(cb));
}

void
OCI::IIOP::AcceptorInfo_impl::remove_accept_cb(OCI::AcceptCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);
	
    for(OB_STL::list<AcceptCB_var>::iterator iter = acceptCBs_.begin();
	iter != acceptCBs_.end();
	iter++)
    {
	if((*iter).in() == cb)
	{
	    acceptCBs_.erase(iter);
	    return;
	}
    }
}

StringSeq*
OCI::IIOP::AcceptorInfo_impl::hosts()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(acceptor_ == 0)
	throw NO_RESOURCES();

    return new StringSeq(acceptor_ -> hosts_);
}

OCI::IIOP::InetAddr_slice*
OCI::IIOP::AcceptorInfo_impl::addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(acceptor_ == 0)
	throw NO_RESOURCES();

    struct sockaddr_in addr;
    socklen_type len = sizeof(addr);
    int result = getsockname(acceptor_ -> fd_,
			     OB_REINTERPRET_CAST(struct sockaddr*, &addr),
			     &len);
    InetAddr_var rc = new InetAddr;
    if(result == 0)
        memcpy(rc.inout(), (const char*)&addr.sin_addr.s_addr, 4);

    return rc._retn();
}

UShort
OCI::IIOP::AcceptorInfo_impl::port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(acceptor_ == 0)
	throw NO_RESOURCES();

    struct sockaddr_in addr;
    socklen_type len = sizeof(addr);
    int result = getsockname(acceptor_ -> fd_,
			     OB_REINTERPRET_CAST(struct sockaddr*, &addr),
			     &len);

    if(result == 0)
        return ntohs(addr.sin_port);
    return 0;
}

void
OCI::IIOP::AcceptorInfo_impl::_OB_callAcceptCB(OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for (OB_STL::list<AcceptCB_var>::iterator iter = acceptCBs_.begin();
         iter != acceptCBs_.end();
	 iter++)
    {
        (*iter) -> accept_cb(info);
    }
}

void
OCI::IIOP::AcceptorInfo_impl::_OB_destroy()
{
    JTCSynchronized synchronized(*this);

    acceptor_ = 0;
}

// ----------------------------------------------------------------------
// AccFactoryInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::AccFactoryInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::IIOP::AccFactoryInfo_impl::describe()
    throw(SystemException)
{
    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;
    return desc._retn();
}

// ----------------------------------------------------------------------
// ConFactoryInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::IIOP::ConFactoryInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::IIOP::ConFactoryInfo_impl::describe()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    return desc._retn();
}

void
OCI::IIOP::ConFactoryInfo_impl::add_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin();
	iter != connectCBs_.end();
	iter++)
    {
	if((*iter).in() == cb)
	    return;
    }

    connectCBs_.push_back(OCI::ConnectCB::_duplicate(cb));
}

void
OCI::IIOP::ConFactoryInfo_impl::remove_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin();
	iter != connectCBs_.end();
	iter++)
    {
	if((*iter).in() == cb)
	{
	    connectCBs_.erase(iter);
	    return;
	}
    }
}

OCI::ConnectCBSeq*
OCI::IIOP::ConFactoryInfo_impl::_OB_getConnectCBSeq()
{
    JTCSynchronized synchronized(*this);

    OCI::ConnectCBSeq* seq = new OCI::ConnectCBSeq;
    seq -> length(connectCBs_.size());
    copy_to_seq(connectCBs_.begin(), connectCBs_.end(), *seq);
    return seq;
}
