// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/algorithms.h>
#include <OB/function.h>
#include <OB/seq_converter.h>

#include <OCI_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// AccFactoryRegistry_impl public member implementation
// ----------------------------------------------------------------------

void
OCI::AccFactoryRegistry_impl::add_factory(AccFactory_ptr factory)
    throw(OCI::FactoryAlreadyExists,
          SystemException)
{
    JTCSynchronized synchronized(*this);

    PluginId_var id = factory -> id();

    for(OB_STL::list<AccFactory_var>::iterator iter = factories_.begin() ;
        iter != factories_.end() ;
        iter++)
    {
        PluginId_var s = (*iter) -> id();
        if(strcmp(s, id) == 0)
            throw FactoryAlreadyExists(id);
    }
    
    factories_.push_back(AccFactory::_duplicate(factory));
}

OCI::AccFactory_ptr
OCI::AccFactoryRegistry_impl::get_factory(const char* id)
    throw(OCI::NoSuchFactory,
          SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<AccFactory_var>::iterator iter = factories_.begin() ;
        iter != factories_.end() ;
        iter++)
    {
        PluginId_var s = (*iter) -> id();
        if(strcmp(s, id) == 0)
            return AccFactory::_duplicate((*iter));
    }
    
    throw NoSuchFactory(id);
}

OCI::AccFactorySeq*
OCI::AccFactoryRegistry_impl::get_factories()
    throw(SystemException)
{
    OCI::AccFactorySeq* seq = new AccFactorySeq();
    seq -> length(factories_.size());
    copy_to_seq(factories_.begin(), factories_.end(), *seq);
    return seq;
}

// ----------------------------------------------------------------------
// ConFactoryRegistry_impl public member implementation
// ----------------------------------------------------------------------

void
OCI::ConFactoryRegistry_impl::add_factory(ConFactory_ptr factory)
    throw(OCI::FactoryAlreadyExists,
          SystemException)
{
    JTCSynchronized synchronized(*this);

    PluginId_var id = factory -> id();

    for(OB_STL::list<ConFactory_var>::iterator iter = factories_.begin() ;
        iter != factories_.end() ;
        iter++)
    {
        PluginId_var s = (*iter) -> id();
        if(strcmp(s, id) == 0)
            throw FactoryAlreadyExists(id);
    }
    
    factories_.push_back(ConFactory::_duplicate(factory));
}

OCI::ConFactory_ptr
OCI::ConFactoryRegistry_impl::get_factory(const char* id)
    throw(OCI::NoSuchFactory,
          SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConFactory_var>::iterator iter = factories_.begin() ;
        iter != factories_.end() ;
        iter++)
    {
        PluginId_var s = (*iter) -> id();
        if(strcmp(s, id) == 0)
            return ConFactory::_duplicate((*iter));
    }
    
    throw NoSuchFactory(id);
}

OCI::ConFactorySeq*
OCI::ConFactoryRegistry_impl::get_factories()
    throw(SystemException)
{
    OCI::ConFactorySeq* seq = new ConFactorySeq();
    seq -> length(factories_.size());
    copy_to_seq(factories_.begin(), factories_.end(), *seq);
    return seq;
}

// ----------------------------------------------------------------------
// CurrentState class. This holds the internal state of the OCI
// Current implementation object.
// ----------------------------------------------------------------------

namespace OCI
{

//
// A holder type is necessary to easily deal with the thread specific
// data
//
struct CurrentStateHolder
{
    //
    // The stack of OCI transports
    //
    OB_STL::list<OCI::TransportInfo_var> transportStack;
};

}; // End namespace OCI

//
// TSS destructor for the CurrentStateHolder
//
static void
DestroyCurrentTSD(void* arg)
{
    OCI::CurrentStateHolder* holder = (OCI::CurrentStateHolder*)arg;
    delete holder;
}

// ----------------------------------------------------------------------
// Current_impl private member implementation
// ----------------------------------------------------------------------

OCI::CurrentStateHolder*
OCI::Current_impl::getStateHolder()
{
    CurrentStateHolder* holder_ = (CurrentStateHolder*)JTCTSS::get(stateKey_);

    //
    // If the data isn't already allocated then allocate a new one
    //
    if(holder_ == 0)
    {
	holder_ = new CurrentStateHolder();
	JTCTSS::set(stateKey_, holder_);
    }

    return holder_;
}

// ----------------------------------------------------------------------
// Current_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::Current_impl::Current_impl()
{
    //
    // Initialize the TSS key
    //
    stateKey_ = JTCTSS::allocate(DestroyCurrentTSD);
}

OCI::Current_impl::~Current_impl()
{
    CurrentStateHolder* holder = getStateHolder();
    delete holder;

    JTCTSS::set(stateKey_, 0);
    JTCTSS::release(stateKey_);
}

// ----------------------------------------------------------------------
// Current_impl public member implementation
// ----------------------------------------------------------------------

OCI::TransportInfo_ptr
OCI::Current_impl::get_oci_transport_info()
    throw(SystemException)
{
    CurrentStateHolder* holder = getStateHolder();

    //
    // TODO: Exception?
    //
    if(holder -> transportStack.empty())
	return TransportInfo::_nil();

    return TransportInfo::_duplicate(holder -> transportStack.front());
}

OCI::AcceptorInfo_ptr
OCI::Current_impl::get_oci_acceptor_info()
    throw(SystemException)
{
    TransportInfo_var info = get_oci_transport_info();
    if(!is_nil(info))
        return info -> acceptor_info();
    else
        return AcceptorInfo::_nil();
}

void
OCI::Current_impl::_OB_preinvoke(TransportInfo_ptr info)
{
    CurrentStateHolder* holder = getStateHolder();
    holder -> transportStack.push_front(TransportInfo::_duplicate(info));
}

void
OCI::Current_impl::_OB_postinvoke()
{
    CurrentStateHolder* holder = getStateHolder();
    holder -> transportStack.pop_front();
}
