// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/Any.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/CoreTraceLevels.h>
#include <OB/ORB_init.h>
#include <OB/ORBInstance.h>

#include <CodeSetDatabase.h>
#include <CodeSetUtil.h>
#include <GIOPMessage.h>
#include <ORB_impl.h>
#include <SelectReactor.h>
#include <Logger_impl.h>

#include <stdlib.h> // for strtoul
#include <ctype.h>
#include <errno.h>

#ifdef WIN32
# include <direct.h>
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// ORB Initialization
// ----------------------------------------------------------------------

void
OB::ParseArgs(int& ac, char** av, Properties_ptr properties,
	      Logger_ptr logger)
{
    assert(!is_nil(properties));

    OB::Logger_var dummy;
    if(is_nil(logger))
    {
	logger = new Logger_impl();
	dummy = logger;
    }

    String_var configFile;

    //
    // Scan through the argument list looking for the config file
    // variable. We need to do this since command line arguments have
    // precendence over the environment variables.
    //
    int i = 0;
    while(i < ac)
    {
        if(strcmp(av[i], "-ORBconfig") == 0)
	{
	    if(i + 1 >= ac)
	    {
		logger -> error("ORB_init: argument expected for -ORBconfig");
		throw INITIALIZE();
	    }
            configFile = string_dup(av[i + 1]);
	    for(int j = i ; j + 2 < ac ; j++)
		av[j] = av[j + 2];

	    ac -= 2;
	}
        else
            i++;
    }

    //
    // Load the contents of the configuration file if present
    //
    if(configFile && configFile[0] != '\0')
    {
	if(!properties -> load(configFile))
	{
	    String_var msg = string_dup(
		"ORB_init: could not load configuration file ");
	    msg += configFile;
	    logger -> warning(msg);
	}
    }

    //
    // Set the default value of the ORB and OA concurrency models if
    // they are not already set
    //
    if(properties -> getProperty("ooc.orb.conc_model") == 0)
        properties -> setProperty("ooc.orb.conc_model",
				  "threaded");

    if(properties -> getProperty("ooc.orb.oa.conc_model") == 0)
        properties -> setProperty("ooc.orb.oa.conc_model",
				  "thread_per_client");

    //
    // Process the remaining -ORB arguments. Turn each argument into
    // an appropriate property.
    //
    i = 0;
    while(i < ac)
    {
        const char* option = av[i];
	if(strcmp(option, "-ORBServerId") == 0)
	{
	    if(i + 1 >= ac)
	    {
		logger -> error("ORB_init: argument expected for "
				"-ORBServerId");
		throw INITIALIZE();
	    }

            properties -> setProperty("ooc.orb.server_name", av[i + 1]);
            
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
	else if(strcmp(option, "-ORBregister") == 0)
	{
	    if(i + 1 >= ac)
	    {
		logger -> error("ORB_init: argument expected for "
				"-ORBregister");
		throw INITIALIZE();
	    }

            properties -> setProperty("ooc.orb.server_name", av[i + 1]);

	    String_var path;
	    char buf[1024];
#ifdef WIN32
            if(strncmp(av[0], "\\\\", 2) == 0 || strchr(av[0], ':') != 0)
            {
                path = string_dup(av[0]);
            }
            else
            {
                _getcwd(buf, 1024);
                path = string_dup(buf);
                path += "\\";
                path += av[0];
            }
#else
            if(av[0][0] == '/')
            {
                path = string_dup(av[0]);
            }
            else
            {
                getcwd(buf, 1024);
                path = string_dup(buf);
                path += "/";
                path += av[0];
            }
#endif
            properties -> setProperty("ooc.orb.imr.register", path);

            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
	else if(strcmp(option, "-ORBserver_instance") == 0)
	{
	    if(i + 1 >= ac)
	    {
		logger -> error("ORB_init: argument expected for "
				"-ORBserver_instance");
		throw INITIALIZE();
	    }

            properties -> setProperty("ooc.orb.server_instance", av[i + 1]);
            
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
	else if(strcmp(option, "-ORBListenEndpoints") == 0)
	{
	    if(i + 1 >= ac)
	    {
		logger -> error("ORB_init: argument expected for "
				"-ORBListenEndpoints");
		throw INITIALIZE();
	    }

            properties -> setProperty("ooc.orb.oa.endpoint", av[i + 1]);

	    ac -= 2;
	}
	else if(strcmp(option, "-ORBNoProprietaryAction") == 0)
	{
	    properties -> setProperty("ooc.orb.noIMR", "true");

	    ac -= 1;
	}
        else if(strcmp(option, "-ORBservice") == 0)
        {
            if(i + 2 >= ac)
            {
                logger -> error("ORB_init: arguments expected for "
				"-ORBservice");
                throw INITIALIZE();
            }

            String_var prop = string_dup("ooc.orb.service.");
            prop += av[i + 1];
            properties -> setProperty(prop, av[i + 2]);
                
            for(int j = i ; j + 3 < ac ; j++)
                av[j] = av[j + 3];
            
            ac -= 3;
        }
        else if(strcmp(option, "-ORBInitRef") == 0)
        {
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
				"-ORBInitRef");
                throw INITIALIZE();
            }

            const char* cp = strchr(av[i + 1], '=');
            if(cp != 0 && av[i + 1][0] != '=' && strlen(cp + 1) > 0)
            {
                String_var id = string_alloc(cp - av[i + 1]);
                strncpy(id.inout(), av[i + 1], cp - av[i + 1]);
                id[cp - av[i + 1]] = '\0';
                String_var prop = string_dup("ooc.orb.service.");
                prop += id;
                properties -> setProperty(prop, cp + 1);
            }
            else
            {
                logger -> error("ORB_init: invalid argument for "
                                "-ORBInitRef");
                throw INITIALIZE();
            }

            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
        else if(strcmp(option, "-ORBDefaultInitRef") == 0)
        {
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
                                "-ORBDefaultInitRef");
                throw INITIALIZE();
            }

            if(strlen(av[i + 1]) == 0)
            {
                logger -> error("ORB_init: invalid argument for "
                                "-ORBDefaultInitRef");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.default_init_ref",
				      av[i + 1]);

            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];

            ac -= 2;
        }
        else if(strcmp(option, "-ORBproperty") == 0)
        {
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
				"-ORBproperty");
                throw INITIALIZE();
            }

            const char* cp = strchr(av[i + 1], '=');
            if(cp != 0 && av[i + 1][0] != '=' && strlen(cp + 1) > 0)
            {
                String_var name = string_alloc(cp - av[i + 1]);
                strncpy(name.inout(), av[i + 1], cp - av[i + 1]);
                name[cp - av[i + 1]] = '\0';
                properties -> setProperty(name, cp + 1);
            }
            else
            {
                logger -> error("ORB_init: invalid argument for -ORBproperty");
                throw INITIALIZE();
            }

            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
        else if(strcmp(option, "-ORBrepository") == 0)
        {
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
                                "-ORBrepository");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.service.InterfaceRepository",
				      av[i + 1]);

            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
        else if(strcmp(option, "-ORBnaming") == 0)
        {
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for -ORBnaming");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.service.NameService",
				      av[i + 1]);
                
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
        else if(strcmp(option, "-ORBtrace_connections") == 0 ||
		strcmp(option, "-ORBtrace_retry") == 0 )
        {
            if(i + 1 >= ac)
            {
		String_var msg =
		(const char*)"ORB_init: argument expected for ";
		msg += option;
                logger -> error(msg);
                throw INITIALIZE();
            }

	    String_var prop = (const char*)"ooc.orb.trace.";
	    prop += option + 10;
            properties -> setProperty(prop, av[i + 1]);
            
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }        
        else if(strcmp(option, "-ORBblocking") == 0)
        {
            properties -> setProperty("ooc.orb.conc_model",
				      "blocking");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-ORBreactive") == 0)
        {
            properties -> setProperty("ooc.orb.conc_model",
				      "reactive");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-ORBthreaded") == 0)
        {
            properties -> setProperty("ooc.orb.conc_model",
				      "threaded");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-ORBversion") == 0)
        {
            logger -> info(Version);
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
	else if(strcmp(option, "-ORBnative_cs") == 0)
	{
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
                                "-ORBnative_cs");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.native_cs", av[i + 1]);
                
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
	else if(strcmp(option, "-ORBnative_wcs") == 0)
	{
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
                                "-ORBnative_wcs");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.native_wcs", av[i + 1]);
                
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
	else if(strcmp(option, "-ORBdefault_wcs") == 0)
	{
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
                                "-ORBdefault_wcs");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.default_wcs", av[i + 1]);
                
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
	else if(strcmp(option, "-ORBmodules") == 0)
	{
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
                                "-ORBmodules");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.modules", av[i + 1]);
                
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
	else if(strcmp(option, "-OAhost") == 0 ||
                strcmp(option, "-OAport") == 0)
	{
            //
            // These options are deprecated, but for
            // backward compatibility they are still
            // supported by the IIOP plug-in
            //
            i += 2;
	}
	else if(strcmp(option, "-OAnumeric") == 0)
	{
            //
            // This option is deprecated, but for
            // backward compatibility it is still
            // supported by the IIOP plug-in
            //
            i++;
	}
	else if(strcmp(option, "-OAversion") == 0)
	{
	    if(i + 1 >= ac)
	    {
                logger -> error("ORB_init: argument expected for -OAversion");
		throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.oa.version", av[i + 1]);

	    for(int j = i ; j + 2 < ac ; j++)
		av[j] = av[j + 2];

	    ac -= 2;
	}
        else if(strcmp(option, "-OAreactive") == 0)
        {
            properties -> setProperty("ooc.orb.oa.conc_model", "reactive");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAthreaded") == 0)
        {
            properties -> setProperty("ooc.orb.oa.conc_model", "threaded");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAthread_per_client") == 0)
        {
            properties -> setProperty("ooc.orb.oa.conc_model",
                                      "thread_per_client");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAthread_per_request") == 0)
        {
            properties -> setProperty("ooc.orb.oa.conc_model",
                                      "thread_per_request");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAthread_pool") == 0)
        {
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
                                "-OAthread_pool");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.oa.conc_model",
				      "thread_pool");
            properties -> setProperty("ooc.orb.oa.thread_pool", av[i + 1]);
            
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
	else if(strcmp(option, "-ORBlicense") == 0)
	{
            if(i + 1 >= ac)
            {
                logger -> error("ORB_init: argument expected for "
                                "-ORBlicense");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.license", av[i + 1]);
            
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
        else if(strncmp(option, "-OA", 3) == 0)
        {
            String_var err =
	    string_dup("ORB_init: unknown option: ");
            err += option;
            logger -> warning(err);
	    
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
	    
            ac -= 1;
        }
	else if(strncmp(option, "-ORB", 4) == 0)
	{
	    String_var err =
	    string_dup("ORB_init: unknown option: ");
	    err += option;
	    logger -> warning(err);
	    
	    for(int j = i ; j + 1 < ac ; j++)
		av[j] = av[j + 1];
	    
	    ac -= 1;
	}
	else
	    i++;
    }
}

OBCORBA::ORB_ptr
OBCORBA::ORB_init(int& ac, char** av, Properties_ptr properties,
		  Logger_ptr logger, Reactor_ptr reactor,
		  const char* orb_identifier, const char* version)
{
    assert_nca(orb_identifier != 0, NCANullString);

    //
    // Scan through the argument list looking for -ORBid.  This is
    // needed to initialize the ORB itself. This variable can take the
    // form "-ORBidfoo" or "-ORBid foo".
    //
    int i = 0;
    while(i < ac)
    {
        if(strncmp(av[i], "-ORBid", 6) == 0)
        {
	    int n;

	    //
	    // Is the entire argument -ORBid?
	    //
	    if(strlen(av[i]) == 6)
	    {
		if(i + 1 >= ac)
		    throw INITIALIZE();
		orb_identifier = av[i+1];
		n = 2;
	    }
	    else
	    {
		orb_identifier = av[i] + 6;
		n = 1;
	    }
	    for(int j = i ; j + n < ac ; j++)
		av[j] = av[j + n];
	    ac -= n;
	    break;
        }
	else
	    i++;
    }

    //
    // Initialize the Logger
    //
    OB::Logger_var loggerDummy;
    if(is_nil(logger))
    {
	loggerDummy = new Logger_impl();
	logger = loggerDummy.in();
    }

    //
    // Check the version
    //
    if(strcmp(version, OB::Version) != 0)
    {
        String_var err = string_dup("ORB_init: version mismatch:");
        err += "\nlibrary version is ";
        err += Version;
        err += "\nheader file version is ";
        err += version;
        logger -> error(err);
        throw INITIALIZE();
    }

    //
    // Need to sync on the orbImplMap_ here to ensure that the
    // properties and what-not are initialized atomically
    //
    JTCSynchronized orbImplMapSync(ORB_impl::orbImplMap_);

    //
    // Initialize the properties
    //
    Properties_var propertyDummy;
    if(is_nil(properties))
    {
	Properties_var defaultProps = Properties::getDefaultProperties();
	propertyDummy = new Properties(defaultProps);
	properties = propertyDummy.in();
    }

    ParseArgs(ac, av, properties, logger);

    //
    // Process each property
    //
    String_var orbId = orb_identifier;
    String_var serverId = string_dup("");
    String_var serverInstance = string_dup("");
    Client::ConcModel concModel = Client::Threaded;
    int nativeCs = CodeSetDatabase::ISOLATIN1;
    int nativeWcs = CodeSetDatabase::UCS2;
    int defaultWcs = 0;

    Properties::KeySeq keys = properties -> getKeys("ooc.orb.");
    ULong j;
    for(j = 0 ; j < keys.length() ; j++)
    {
        const char* key = keys[j];
        const char* value = properties -> getProperty(key);
        assert(value != 0);

        if(strcmp(key, "ooc.orb.conc_model") == 0)
        {
            if(strcmp(value, "reactive") == 0)
                concModel = Client::Reactive;
            else if(strcmp(value, "threaded") == 0)
                concModel = Client::Threaded;
            else
            {
                String_var err =
		    string_dup("ORB_init: unknown value for "
			       "ooc.orb.conc_model: ");
                err += value;
                logger -> warning(err);
            }
        }
	else if(strncmp(key, "ooc.orb.trace.", 14) == 0)
        {
	    // Ignore -- handled in CoreTraceLevels
        }
	else if(strncmp(key, "ooc.orb.policy.", 15) == 0)
        {
	    // Ignore -- handled in initializeDefaultPolicies()
        }
        else if(strcmp(key, "ooc.orb.id") == 0)
        {
            orbId = string_dup(value);
        }
	else if(strcmp(key, "ooc.orb.server_name") == 0)
	{
	    //
	    // The server name must begin with an alpha-numeric
	    // character
	    //
	    if(!isalnum(*value))
	    {
		String_var err =
		    string_dup("ORB_init: illegal value for "
			       "ooc.orb.server_name: ");
                err += value;
                logger -> error(err);
		throw INITIALIZE();
	    }
	    serverId = string_dup(value);
	}
	else if(strcmp(key, "ooc.orb.server_instance") == 0)
	{
	    serverInstance = string_dup(value);
	}
	else if(strcmp(key, "ooc.orb.native_cs") == 0)
	{
	    ULong csid = CodeSetDatabase::instance() -> nameToId(value);
	    if(CodeSetDatabase::instance() -> getCharMapInfo(csid) != 0)
		nativeCs = csid;
	    else
	    {
		String_var err =
		    string_dup("ORB_init: unknown value for "
			       "ooc.orb.native_cs: ");
		err += value;
		logger -> error(err);
		throw INITIALIZE();
	    }
	}
	else if(strcmp(key, "ooc.orb.native_wcs") == 0)
	{
	    ULong csid = CodeSetDatabase::instance() -> nameToId(value);
	    if(CodeSetDatabase::instance() -> getCharMapInfo(csid) != 0)
		nativeWcs = csid;
	    else
	    {
		String_var err =
		    string_dup("ORB_init: unknown value for "
			       "ooc.orb.native_wcs: ");
		err += value;
		logger -> error(err);
		throw INITIALIZE();
	    }
	}
	else if(strcmp(key, "ooc.orb.default_wcs") == 0)
	{
	    ULong csid = CodeSetDatabase::instance() -> nameToId(value);
	    if(csid && csid != CodeSetDatabase::UTF8)
		defaultWcs = csid;
	    else
	    {
		String_var err =
		    string_dup("ORB_init: unknown value for "
			       "ooc.orb.default_wcs: ");
		err += value;
		logger -> error(err);
		throw INITIALIZE();
	    }
	}
	else if(strcmp(key, "ooc.orb.extended_wchar") == 0)
	{
	    if(strcmp(value, "true") != 0 && strcmp(value, "false") != 0) 
	    {
		String_var err =
		    string_dup("ORB_init: unknown value for "
			       "ooc.orb.extended_wchar: ");
		err += value;
		logger -> error(err);
		throw INITIALIZE();
	    }
	}
        else if(strcmp(key, "ooc.orb.default_init_ref") == 0)
        {
            if(strlen(value) == 0)
                logger -> warning("ORB_init: invalid value for "
                                  "ooc.orb.default_init_ref");
        }
	else if(strcmp(key, "ooc.orb.server_timeout") == 0 ||
	        strcmp(key, "ooc.orb.server_shutdown_timeout") == 0 )
	{
	    // Used by GIOPServerWorker
	}
	else if(strcmp(key, "ooc.orb.client_timeout") == 0 ||
	        strcmp(key, "ooc.orb.client_shutdown_timeout") == 0)
	{
	    // Used by GIOPClientWorker
	}
	else if(strncmp(key, "ooc.orb.service.", 16) == 0)
	{
	    // Ignore
	}
	else if(strncmp(key, "ooc.orb.oa.", 11) == 0)
	{
	    // Ignore
	}
	else if(strncmp(key, "ooc.orb.poamanager.", 19) == 0)
	{
	    // Ignore
	}
	else if(strcmp(key, "ooc.orb.imr.register") == 0)
	{
	    // Used by POAManagerFactory
	}
	else if(strcmp(key, "ooc.orb.use_type_code_cache") == 0)
	{
	    if(strcmp(value, "true") != 0 && strcmp(value, "false") != 0) 
	    {
		String_var err =
		    string_dup("ORB_init: unknown value for "
			       "ooc.orb.use_type_code_cache: ");
		err += value;
		logger -> error(err);
		throw INITIALIZE();
	    }
	}
	else if(strcmp(key, "ooc.orb.modules") == 0)
        {
	    // Ignore -- handled in ORB_impl() and ModuleManager
        }
	else if(strncmp(key, "ooc.orb.module.", 15) == 0)
        {
	    // Ignore -- handled in ModuleManager
        }
        else if(strcmp(key, "ooc.orb.giop.max_message_size") == 0)
        {
            char* end;
            ULong max = strtoul(value, &end, 0);
            if(*end != '\0' || errno == ERANGE)
            {
                String_var err =
		    string_dup("ORB_init: invalid value for "
			       "ooc.orb.giop.max_message_size: ");
                err += value;
                logger -> warning(err);
            }
            else
            {
		//
		// TODO: singleton?
		//
                GIOPIncomingMessage::setMaxMessageSize(max);
                GIOPOutgoingMessage::setMaxMessageSize(max);
            }
        }
	else if(strcmp(key, "ooc.orb.license") == 0)
        {
	    // Ignore -- handled by translators
        }
        else
        {
            String_var err = string_dup("ORB_init: unknown property: ");
            err += key;
            logger -> warning(err);
        }
    }
   
    //
    // Parse the tracing levels from the properties
    //
    CoreTraceLevels_var coreTraceLevels =
	new CoreTraceLevels(logger, properties);

    //
    // Is there an existing ORB implementation for this orb id?
    //
    // TODO: Is this correct? Why did we bother to parse all of those
    // command line options and stuff if we're only going to return
    // the existing ORB?
    //
    ORB_var orb;
    if(ORB_impl::orbImplMap_.get(orbId, orb))
	return orb._retn();

    //
    // Create a SelectReactor if none was provided
    //
    OB::Reactor_var reactorDummy;
    if(is_nil(reactor))
    {
	reactorDummy = new SelectReactor();
	reactor = reactorDummy.in();
    }

    //
    // Create new ORB
    //
    orb = new ORB_impl(ac, av, orbId, serverId, serverInstance, concModel,
    		       coreTraceLevels, properties, logger, reactor,
		       nativeCs, nativeWcs, defaultWcs);

    return orb._retn();
}

OBCORBA::ORB_ptr
OBCORBA::ORB_init(int& ac, char** av, Properties_ptr properties,
		  const char* orb_identifier, const char* version)
{
    return OBCORBA::ORB_init(ac, av, properties, Logger::_nil(),
			     Reactor::_nil(), orb_identifier, version);
			     
}

CORBA::ORB_ptr
CORBA::ORB_init(int& ac, char** av, const char* orb_identifier,
		const char* version)
{
    return OBCORBA::ORB_init(ac, av, Properties::_nil(),
			     Logger::_nil(), Reactor::_nil(),
			     orb_identifier, version);
}
