// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/ObjectKey.h>

#include <stdlib.h>

using namespace OB;
using namespace CORBA;

// ----------------------------------------------------------------------
// ObjectKey OB namespace functions
// ----------------------------------------------------------------------

//
// OB 3.x object keys:
// Transient: 0 + data
// Persistent: non-zero + data
//
// OB 4.x object key:
//
// The object key is: 0xabacab
// ['1'=transient|'0'=persistent]|[time\0]`poa \0 poa \0 poa \0\0 id'.
// NOTE: we then know that we've reached the id by the presence of the
// two \0\0.  The time field we add for the POA create time (if
// transient) is current seconds. This is a tradeoff - it's most
// likely to be good enough.
//

OCI::ObjectKey*
OB::CreateObjectKey(const ObjectKeyData& id)
{
    OCI::ObjectKey_var key = new OCI::ObjectKey;
    
    ULong i;



    //
    // Count the number of bytes for the poas.
    //
    ULong len = strlen(id.serverId) + 1;
    for(i = 0; i < id.poaId.length(); ++i)
        len += strlen(id.poaId[i]) + 1;
    
    //
    // Add one for the '\0'
    //
    len += 1;

    //
    // Add 4 for the magic number
    //
    len += 3;

    //
    // Add one for persistent bit
    //
    len += 1;
    String_var time;

    //
    // If the POA is transient then add the time the poa was created
    //
    if(!id.persistent)
    {
        time = string_dup("");

        time += id.createTime;
        len += strlen(time) + 1;
    }
    
    //
    // Set the length
    //
    key -> length(id.oid.length() + len);

    char* data = (char*)key -> get_buffer();

    //
    // Add the magic number
    //
    *data++ = (Char)0xab;
    *data++ = (Char)0xac;
    *data++ = (Char)0xab;

    //
    // Add the persistent/transient information
    //
    if(id.persistent)
    {
        *data++ = '0';
    }
    else
    {
        *data++ = '1';
        strcpy(data, time);
        data += strlen(time)+1;
    }
    
    //
    // Copy in the data for the poas
    //
    strcpy(data, id.serverId);
    data += strlen(id.serverId)+1;
    for(i = 0; i < id.poaId.length(); ++i)
    {
        strcpy(data, id.poaId[i]);
        data += strlen(id.poaId[i])+1;
    }

    //
    // Add additional '\0'
    //
    *data++ = '\0';

    memcpy((Octet*)data, id.oid.get_buffer(), id.oid.length());
    
    return key._retn();
}

//
// Parse an object key, filling in the fields of the ObjectKeyData.
// Return false if the object-key cannot be parsed, true otherwise.
//
bool
OB::ParseObjectKey(const OCI::ObjectKey& key,
                   ObjectKeyData& keyData)
{
    //
    // Parse out the POA's. This should be a sequence of strings
    // terminated by two null values.
    //
    const Octet* data = key.get_buffer();
    const Octet* end = data + key.length();
    
    //
    // First try to figure out whether the object-key is OB 4.x
    // format.  Must be at least 4 bytes - magic + 1|0
    //
    //     OB 4.x magic number: abacab
    //
    if(key.length() > 4 &&
       data[0] == 0xab && data[1] == 0xac && data[2] == 0xab)
    {
        data += 3;
	if(*data == '0') // persistent
	{
	    keyData.persistent = true;
	    keyData.createTime = 0;
	    ++data;
	}
	else
	{
	    if(*data != '1')
		return false;
	    keyData.persistent = false;
	    ++data;
	    //
	    // Remember the start of the time stamp
	    //
	    const char* start = (const char*)data;
	    while(data < end && *data != '\0')
		data++;
	    if(data >= end)
		return false;
	    
	    keyData.createTime = atol(start);
	    //
	    // skip the null byte
	    //
	    ++data;
	}

	bool first = true;
	keyData.poaId.length(0);
	while(data < end)
	{
	    //
	    // Remember the start of the POA name
	    //
	    const char* start = (const char*)data;
	    while(data < end && *data != '\0')
		data++;
	    
	    //
	    // Ensure that we haven't gone too far...
	    //
	    if(data >= end)
		return false;
	    
	    if(first)
	    {
	        keyData.serverId = start;
	        first = false;
	    }
	    else
	    {
	        //
	        // Append this to the sequence of POA's.
	        //
	        CORBA::ULong poaIdLength = keyData.poaId.length();
	        keyData.poaId.length(poaIdLength + 1);
	        keyData.poaId[poaIdLength] = start;
	    }
	    
	    //
	    // Skip this byte, check to see if we have another '\0'
	    //
	    if(*++data == '\0')
	    {
		//
		// Skip this byte
		//
		++data;
		break;
	    }
	}

        //
        // Verify that we haven't gone too far.
        //
        if(data >= end)
            return false;
    
        //
        // Remaining portion is the ObjectId.
        //
        ULong len = end - data;
        keyData.oid.length(len);
        memcpy(keyData.oid.get_buffer(), data, len);

        return true;
    }
    else
    {
        return false;
    }

}

// ----------------------------------------------------------------------
// ObjectKeyData constructor
// ----------------------------------------------------------------------

ObjectKeyData::ObjectKeyData()
{
}

ObjectKeyData::ObjectKeyData(const char* _serverId,
			     const StringSeq& _poaId,
			     const PortableServer::ObjectId& _oid,
			     bool _persistent,
			     time_t _createTime)
    : serverId(_serverId),
      poaId(_poaId),
      oid(_oid),
      persistent(_persistent),
      createTime(_createTime)
{
}
