// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <PICurrent_impl.h>

//
// Implementation Notes:
//
// The Current implementation needs to have a stack of slot
// information.  This is to deal with collocated calls that occur in
// the same thread as the client. In this case the server current
// should not overwrite the content of the client side current.
//

using namespace CORBA;
using namespace OB;

namespace PortableInterceptor
{

//
// This is the contents of the slot data
//
struct SlotData
{
    SlotData(CORBA::AnySeq* s, bool r)
        : slots(s), release(r), next(0) { }

    ~SlotData()
    {
	if(release)
	{
	    delete slots;
	}
    }
    
    CORBA::AnySeq* slots;
    bool release;
    SlotData* next;
};

//
// A holder type for the SlotData
//
struct SlotDataHolder
{
    SlotDataHolder()
        : head(0) { }
    ~SlotDataHolder()
    {
        while(head != 0)
        {
            SlotData* tmp = head;
            head = head -> next;
            delete tmp;
        }
    }
    SlotData* head;
};

} // End namespace PortableInterceptor

// ----------------------------------------------------------------------
// JTCTSS destructor function
// ----------------------------------------------------------------------

static void
DestroyCurrentTSD(void* arg)
{
    PortableInterceptor::SlotDataHolder* holder =
	(PortableInterceptor::SlotDataHolder*)arg;
    delete holder;
}

// ----------------------------------------------------------------------
// Current_impl private member implementation
// ----------------------------------------------------------------------

//
// Establish the current thread-specific data
//
PortableInterceptor::SlotDataHolder*
PortableInterceptor::Current_impl::establishTSD(bool partial)
{
    SlotDataHolder* holder_ = (SlotDataHolder*)JTCTSS::get(stateKey_);

    //
    // If the data isn't already allocated then allocate a new
    // SlotDataHolder and a new set of slots
    //
    if(holder_ == 0)
    {
	holder_ = new SlotDataHolder();
	JTCTSS::set(stateKey_, holder_);

	AnySeq* slots = 0;
	//
	// This is an optimization. If this is a partial allocation
	// then it's not necessary to allocate data for the actual
	// slots.
	//
	if(!partial)
	{
	    slots = new AnySeq(maxSlots_);
	    slots -> length(maxSlots_);
	}
	holder_ -> head = new SlotData(slots, true);
	holder_ -> head -> next = 0;
    }

    return holder_;
}

// ----------------------------------------------------------------------
// Current_impl constructor and destructor
// ----------------------------------------------------------------------

PortableInterceptor::Current_impl::Current_impl()
    : maxSlots_(0)
{
    //
    // Allocate a new TSD key
    //
    stateKey_ = JTCTSS::allocate(DestroyCurrentTSD);
}

PortableInterceptor::Current_impl::~Current_impl()
{
    SlotDataHolder* holder = establishTSD(false);
    delete holder;

    JTCTSS::set(stateKey_, 0);
    JTCTSS::release(stateKey_);
}

// ----------------------------------------------------------------------
// Current_impl public member implementation
// ----------------------------------------------------------------------

Any*
PortableInterceptor::Current_impl::get_slot(SlotId id)
    throw(SystemException, PortableInterceptor::InvalidSlot)
{
    if(id >= maxSlots_)
    {
        throw InvalidSlot();
    }

    SlotDataHolder* holder = establishTSD(false);

    return new Any((*holder -> head -> slots)[id]);
}

void
PortableInterceptor::Current_impl::set_slot(SlotId id, const Any& any)
    throw(SystemException, PortableInterceptor::InvalidSlot)
{
    if(id >= maxSlots_)
    {
        throw InvalidSlot();
    }

    SlotDataHolder* holder = establishTSD(false);

    (*holder -> head -> slots)[id] = any;
}

AnySeq*
PortableInterceptor::Current_impl::_OB_currentSlotData()
{
    SlotDataHolder* holder = establishTSD(false);

    return new AnySeq(*holder -> head -> slots);
}

//
// On the client side a completely new set of slots are needed during
// the actual interceptor call, the client interceptor manage the slot
// data.
//
// On the server side the set of slots are shared between the
// interceptor and the server side PICurrent
//
void
PortableInterceptor::Current_impl::_OB_pushSlotData(AnySeq& slots)
{
    SlotDataHolder* holder = establishTSD(false);

    SlotData* newSlots = new SlotData(&slots, false);
    newSlots -> next = holder -> head;
    holder -> head = newSlots;
}

void
PortableInterceptor::Current_impl::_OB_popSlotData()
{
    SlotDataHolder* holder = establishTSD(false);

    SlotData* tmp = holder -> head;
    holder -> head = holder -> head -> next;
    assert(holder -> head != 0);
    delete tmp;
}

AnySeq*
PortableInterceptor::Current_impl::_OB_newSlotTable() const
{
    AnySeq_var t = new AnySeq(maxSlots_);
    t -> length(maxSlots_);
    return t._retn();
}

void
PortableInterceptor::Current_impl::_OB_setMaxSlots(ULong max)
{
    maxSlots_ = max;
}
