// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/MarshalStubImpl.h>
#include <OB/DirectStubImpl.h>
#include <OB/OAInterface.h>
#include <OB/Upcall.h>
#include <OB/ValueBase.h>
#include <OB/IMRState.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/Timer.h>
#include <OB/RefCounted.h>
#include <OB/OCI_IIOP.h>
#include <OB/Net.h> // For MAXHOSTNAMELEN
#include <OB/seq_converter.h> // For MAXHOSTNAMELEN

#include <OB/ValueBase.h>
#include <OB/PIIOR.h>
#include <OB/PIManager.h>
#include <OB/OBPIORT.h>

#include <Server.h>
#include <POAManagerFactory_impl.h>
#include <POAManager_impl.h>
#include <POA_impl.h>
#include <OADProcessEndpoint_impl.h>
#include <POALocator.h>
#include <ParseParams.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// POAManagerFactory_impl private and protected member implementation
// ----------------------------------------------------------------------

char*
OBPortableServer::POAManagerFactory_impl::getUniqueName()
{
    timeval tv = TimeNow();

    String_var name = string_dup("POAManager-");
    name += tv.tv_sec;
    name += count_++;

    assert(!managers_.containsKey(name));

    return name._retn();
}

void
OBPortableServer::POAManagerFactory_impl::validateName(const char* name)
    throw(OBPortableServer::POAManagerFactory::POAManagerAlreadyExists)
{
    //
    // Does the POAManager exist?
    //
    POAManager_var manager;
    if(managers_.get(name, manager))
    {
        //
        // If the POAManager is INACTIVE then remove it from the list,
        // and allow the user to re-add the POAManager
        //
        if(manager -> get_state() == PortableServer::POAManager::INACTIVE)
            managers_.remove(name);
        else
            throw POAManagerAlreadyExists();
    }
}

// ----------------------------------------------------------------------
// POAManagerFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OBPortableServer::POAManagerFactory_impl::POAManagerFactory_impl()
    : count_(0),
      poaLocator_(new POALocator)
{
}

OBPortableServer::POAManagerFactory_impl::~POAManagerFactory_impl()
{
    assert(managers_.size() == 0);

    delete poaLocator_;
}

// ----------------------------------------------------------------------
// POAManagerFactory_impl public member implementation
// ----------------------------------------------------------------------

OBPortableServer::POAManager_ptr
OBPortableServer::POAManagerFactory_impl::create_poa_manager(const char* name)
    throw(OBPortableServer::POAManagerFactory::POAManagerAlreadyExists,
          OCI::InvalidParam,
          SystemException)
{
    JTCSynchronized sync(managers_);

    String_var nameDummy;
    if(strlen(name) == 0)
    {
        nameDummy = getUniqueName();
        name = nameDummy.in();
    }
    else
    {
        validateName(name);
	++count_;
    }

    Properties_var props = orbInstance_ -> getProperties();
    Logger_var logger = orbInstance_ -> getLogger();

    //
    // Get the endpoint configuration
    //
    const char* rootStr = 0;
    const char* paramStr = 0;
    if(strcmp(name, "RootPOAManager") == 0)
        rootStr = props -> getProperty("ooc.orb.oa.endpoint");
    String_var propName = string_dup("ooc.orb.poamanager.");
    propName += name;
    propName += ".endpoint";
    paramStr = props -> getProperty(propName);
    if(paramStr == 0 && rootStr == 0)
        paramStr = "iiop";
    else if(paramStr == 0)
        paramStr = rootStr;

    OCI::AccFactoryRegistry_var registry =
        orbInstance_ -> getAccFactoryRegistry();
    OCI::AccFactorySeq_var factories = registry -> get_factories();

    OCI::AcceptorSeq acceptors;
    const char* next = paramStr;
    while(next != 0)
    {
        OB_STL::list<String_var> params;
        next = parse_params(next, params);
        if(!params.empty())
        {
            String_var prot = params.front();
            params.pop_front();
            ULong i;
            for(i = 0 ; i < factories -> length() ; i++)
            {
                String_var s = factories[i] -> id();
                if(strcmp(s, prot) == 0)
                {
                    try
                    {
                        OCI::ParamSeq paramSeq;
                        paramSeq.length(params.size());
                        copy_to_seq(params.begin(), params.end(), paramSeq);

                        OCI::Acceptor_var acceptor =
                             factories[i] -> create_acceptor(paramSeq);
                        acceptors.length(acceptors.length() + 1);
                        acceptors[acceptors.length() - 1] = acceptor;
                        break;
                    }
                    catch(const OCI::InvalidParam& ex)
                    {
                        String_var err = string_dup("unable to create "
                                                    "acceptor: ");
                        err += ex.reason;
                        logger -> error(err);
                        throw INITIALIZE();
                    }
                }
            }
            if(i >= factories -> length())
            {
                String_var err = string_dup("unknown endpoint protocol `");
                err += prot;
                err += "'";
                logger -> error(err);
                throw INITIALIZE();
            }
        }
    }

    if(acceptors.length() == 0)
    {
        String_var err = string_dup("no endpoints defined");
        logger -> error(err);
        throw INITIALIZE(err);
    }

    //
    // Create the new POAManager_impl and add to the table
    //
    POAManager_var manager = new POAManager_impl(orbInstance_,
                                                 poaLocator_, name,
                                                 count_, acceptors);

    managers_.put(name, manager);

    return manager._retn();
}

OBPortableServer::POAManager_ptr
OBPortableServer::POAManagerFactory_impl::create_poa_manager_with_config(
    const char* name,
    const AcceptorConfigSeq& config)
    throw(OBPortableServer::POAManagerFactory::POAManagerAlreadyExists,
          OCI::NoSuchFactory,
          OCI::InvalidParam,
	  SystemException)
{
    JTCSynchronized sync(managers_);

    String_var nameDummy;
    if(strlen(name) == 0)
    {
        nameDummy = getUniqueName();
        name = nameDummy.in();
    }
    else
    {
        validateName(name);
	++count_;
    }	 

    //
    // Create acceptors
    //
    OCI::AccFactoryRegistry_var registry =
        orbInstance_ -> getAccFactoryRegistry();
    OCI::AcceptorSeq acceptors;
    acceptors.length(config.length());
    for(ULong i = 0 ; i < config.length() ; i++)
    {
        OCI::AccFactory_var factory = registry -> get_factory(config[i].id);
        acceptors[i] = factory -> create_acceptor(config[i].params);
    }

    //
    // Create the new POAManager_impl and add to the table
    //
    POAManager_var manager = new POAManager_impl(orbInstance_,
                                                 poaLocator_, name,
                                                 count_, acceptors);
    managers_.put(name, manager);

    return manager._retn();
}

OBPortableServer::POAManagerSeq*
OBPortableServer::POAManagerFactory_impl::get_poa_managers()
    throw(SystemException)
{
    POAManagerTable::Enumerator keys = managers_.keys();
    POAManagerSeq_var result = new POAManagerSeq(keys.length());

    POAManager_var manager;
    for(ULong i = 0 ; i < keys.length() ; i++)
    	if(managers_.get(keys[i], manager))
	{
	    ULong resultLength  = result -> length();
	    result -> length(resultLength + 1);
	    result[resultLength] = manager;
        }

    return result._retn();
}

void
OBPortableServer::POAManagerFactory_impl::destroy()
    throw(SystemException)
{
    //
    // Shutdown, if necessary (this should be called when the ORB is
    // destroyed).
    //
    if(managers_.size() != 0)
	_OB_deactivate();

    //
    // Remove the references to the orbInstance_
    //
    orbInstance_ = ORBInstance::_nil();
}

void
OBPortableServer::POAManagerFactory_impl::_OB_deactivate()
    throw(SystemException)
{
    //
    // Deactivate each of the POAManagers
    //
    POAManagerTable::Enumerator keys = managers_.keys();
    for(ULong i = 0 ; i < keys.length() ; i++)
    {
	POAManager_var manager;
	if(managers_.get(keys[i], manager))
	{
	    try
	    {
		manager -> deactivate(true, true);
	    }
	    catch(const PortableServer::POAManager::AdapterInactive&)
	    {
		// Ignore
	    }
	}
    }
    managers_.clear();

    //
    // Tell the IMR that the ORB is STOPPING
    //
    if(!is_nil(activeState_))
    {
	Logger_var logger = orbInstance_ -> getLogger();
	CORBA::String_var serverInstance = orbInstance_ -> getServerInstance();
	try
	{
	    activeState_ -> set_status(serverInstance, IMR::STOPPING);
	}
	catch(...)
	{
	    String_var msg = orbInstance_ -> getServerId();
	    msg += ": Cannot contact IMR on shutdown";
	    logger -> warning(msg);
	}

	//
	// Clear the IMR::Server record
	//
	activeState_ = IMR::ActiveState::_nil();
    }
}

DirectStubImpl_ptr
OBPortableServer::POAManagerFactory_impl::_OB_createDirectStubImpl(
    RefCountIOR_ptr ior, 
    RefCountPolicyList_ptr policies)
    throw(SystemException, LocationForward)
{
    POAManagerTable::Enumerator keys = managers_.keys();
    for(ULong i = 0 ; i < keys.length() ; i++)
    {
	try
	{
	    POAManager_var manager;
	    if(managers_.get(keys[i], manager))
	    {
		OCI::AcceptorSeq_var acceptors = manager -> get_acceptors();
                for(ULong j = 0 ; j < acceptors -> length() ; j++)
                {
                    OCI::ProfileInfoSeq_var profileInfos =
                        acceptors[j] -> get_local_profiles(ior -> value);

                    //
                    // If the IOR is local then at least one ProfileInfo
                    // will be returned
                    //
                    if(profileInfos -> length() > 0)
                    {
                        POAManager_impl* managerImpl =
                            dynamic_cast<POAManager_impl*>(manager.in());
                        
                        //
                        // In the case that the servant cannot support a
                        // DirectStubImpl a null-pointer will be returned
                        //
                        return managerImpl ->
                            _OB_createDirectStubImpl(profileInfos[0].key,
			    			     policies);
                    }
                }
	    }
	}
	catch(const PortableServer::POAManager::AdapterInactive&)
	{
	    //
	    // Ignore -- the POAManager isn't valid anymore
	    //
	}
    }

    return DirectStubImpl::_nil();
}

void
OBPortableServer::POAManagerFactory_impl::_OB_setORBInstance(
    ORBInstance_ptr instance)
{
    orbInstance_ = ORBInstance::_duplicate(instance);
}

namespace PortableInterceptor
{

    class IMRIORInterceptor_impl : public PortableInterceptor::IORInterceptor,
				   public OBCORBA::RefCountLocalObject
    {
	//
	// TODO: XXX: Document
	//
	typedef OB::Hashtable< AdapterManagerId,
			       IMR::POANameSeq_var,
			       ULongHasher > POAHashTable;

	POAHashTable poas_;
	OB::Logger_var logger_;
	IMR::ActiveState_var as_;
	CORBA::String_var serverInstance_;
	bool running_;

	IMR::POAStatus convertState(AdapterState state)
	{
	    IMR::POAStatus status;
	    switch(state)
	    {
	    case PortableInterceptor::INACTIVE:
	        status = IMR::INACTIVE;
		break;
	    case PortableInterceptor::ACTIVE:
	        status = IMR::ACTIVE;
		break;
	    case PortableInterceptor::HOLDING:
	        status = IMR::HOLDING;
		break;
	    case PortableInterceptor::DISCARDING:
	        status = IMR::DISCARDING;
		break;
	    }
	    return status;
	}
    
    public:

	IMRIORInterceptor_impl(OB::Logger_ptr logger,
			       IMR::ActiveState_ptr as,
			       const char* serverInstance)
	    : logger_(OB::Logger::_duplicate(logger)),
	      as_(IMR::ActiveState::_duplicate(as)),
	      serverInstance_(serverInstance),
	      running_(false)
	{
	}

	virtual char*
	name()
	    throw(SystemException)
	{
	    return string_dup("IMRInterceptor");
	}

        virtual void
        destroy()
	    throw(SystemException)
	{
	}

	virtual void
	establish_components(IORInfo_ptr)
	    throw(SystemException)
	{
	}

	virtual void
	components_established(IORInfo_ptr info)
	    throw(SystemException)
	{
	    //
	    // This method does nothing if this is not a persistent POA
	    //
            try
            {
                Policy_var p = info -> get_effective_policy(
                    PortableServer::LIFESPAN_POLICY_ID);
                PortableServer::LifespanPolicy_var policy =
                    PortableServer::LifespanPolicy::_narrow(p);
                if(policy -> value() != PortableServer::PERSISTENT)
                    return;
            }
            catch(const INV_POLICY &)
            {
                // 
                // Default Lifespan policy is TRANSIENT
                // 
                return;
            }

	    //
	    // Get the primary object-reference template
	    //
	    PortableInterceptor::ObjectReferenceTemplate_var primary =  
		info -> adapter_template();
	
	    try
	    {
	        AdapterState state = info -> state();
		IMR::POAStatus status = convertState(state);

		PortableInterceptor::ObjectReferenceTemplate_var secondary =
		    as_ -> poa_create(status, primary);

		info -> current_factory(secondary);
	    }
	    catch(const IMR::NoSuchPOA& e)
	    {
		String_var msg = string_dup("IMR: POA not registered: ");
		for(ULong i = 0; i < e.poa.length(); i++)
		{
		    msg += e.poa[i];
		    if(i != e.poa.length()-1)
			msg += "/";
		}
		logger_ -> error(msg);
		throw INITIALIZE();
	    }
	    catch(SystemException& ex)
	    {
		String_var msg = string_dup("IMR: Cannot contact: ");
		//XXX - memory leak
		msg += ex._to_string();
		logger_ -> error(msg);
		//
		// TODO: XXX: Are these exceptions swallowed? If so,
		// that's a bug. Note that this probably breaks the POA
		// code since this happens in init() and that's declared
		// throw().
		//
		throw INITIALIZE(); // TODO: - some exception
	    }

	    //
	    // Update the poa hash table
	    //
	    AdapterManagerId id = info -> manager_id();
	    AdapterName_var name = primary -> adapter_name();

	    IMR::POANameSeq_var poas;
	    if(poas_.get(id, poas))
	    {
	        //
		// Add poa to exiting entry
		//
	        ULong length = poas -> length();
		poas -> length(length + 1);
		poas[length] = name;
		poas_.put(id, poas);
	    }
	    else
	    {
	        //
		// Add a new entry for this adapter manager
		//
	        poas = new IMR::POANameSeq(1);
		poas -> length(1);
		poas[0] = name;
		poas_.put(id, poas);
	    }
	}

	virtual void
	adapter_state_change(const ObjectReferenceTemplateSeq& templates,
			     AdapterState state)
	    throw(SystemException)
	{
	    //
	    // Only update the IMR from this point if the POAs have
	    // been destroyed.
	    //
	    if(state != PortableInterceptor::NON_EXISTENT)
	        return;

	    IMR::POANameSeq poanames;
	    for(CORBA::ULong i = 0 ; i < templates.length() ; ++i)
	    {
	        OBPortableInterceptor::PersistentORT* ort = 
		    OBPortableInterceptor::PersistentORT::_downcast(
		        templates[i]);

		//
		// If not a Persistent ORT continue
		//
		if(ort == 0)
		    continue;
	        
	        AdapterName_var adpaterName = templates[i] -> adapter_name();

		//
		// Add the POA to the list of POAs to send to the
		// IMR for the status update
		//
		CORBA::ULong length = poanames.length();
		poanames.length(length + 1);
		poanames[length] = adpaterName;

		//
		// Find the POA in the POAManager -> POAs map and
		// remove it.
		//
		POAHashTable::Enumerator keys = poas_.keys();
		for(CORBA::ULong j = 0 ; j < keys.length() ; ++j)
		{
		    IMR::POANameSeq_var poas;
	            if(poas_.get(keys[j], poas))
	            {
	                //
		        // Find the poa being deleted
		        //
	                ULong k;
		        for(k = 0 ; k < poas -> length() ; ++k)
		        {
		            if(poas[k].length() != adpaterName -> length())
		                continue;

		            bool found = true;
		            for(ULong l = 0 ; l < adpaterName -> length() ; ++l)
		            {
		                if(strcmp(poas[k][l], adpaterName[l]) != 0)
			        {
			            found = false;
			            break;
			        }
		            }
		            if(found)
		                break;
		        }

		        //
		        // Shift back the remaining poas if match found
		        //
		        if(k != poas -> length())
		        {
		            for(; k < poas -> length() - 1 ; ++k)
		                poas[k] = poas[k + 1];
		            poas -> length(poas -> length() - 1);
			    break;
		        }
	            }
		}
	    }

	    if(poanames.length() != 0)
	    {
	        try
	        {
		    as_ -> poa_status_update(poanames, IMR::NON_EXISTENT);
	        }
	        catch(const SystemException& ex)
	        {
		    String_var msg = string_dup("IMR: poa_destroy: ");
		    //XXX - memory leak

		    msg += ex._to_string();
		    logger_ -> warning(msg);
	        }
	    }
	}

	//
	// Update POA states for this adapter
	//
	void
	adapter_manager_state_change(AdapterManagerId id,
				     AdapterState state)
	    throw(SystemException)
	{
	    if(!running_)
	    {
	    	//
		// Inform the IMR the server is now running if this
		// is the first call.
		//
		try
		{
		    as_ -> set_status(serverInstance_, IMR::RUNNING);
		}
		catch(const CORBA::OBJECT_NOT_EXIST&)
		{
		    String_var msg = string_dup("IMR: Not registered");
		    logger_ -> error(msg);
		    throw INITIALIZE();
		}
		catch(const CORBA::SystemException&)
		{
		    String_var msg = string_dup("IMR: Cannot contact");
		    logger_ -> error(msg);
		    throw INITIALIZE();
		}

	        running_ = true;
	    }


	    //
	    // Inform the IMR of the POA status update
	    //
	    IMR::POANameSeq_var poas;
	    if(poas_.get(id, poas))
	    {
	        if(poas -> length() != 0)
		{
	            try
		    {
		        IMR::POAStatus status = convertState(state);
	                as_ -> poa_status_update(poas, status);
	    	    }
		    catch(const CORBA::SystemException&)
		    {
		        //
			// XXX ????
			//
		    }
		}
	    }
	}
    };

} // End namespace PortableInterceptor

void
OBPortableServer::POAManagerFactory_impl::_OB_initializeIMR(
    OBPortableServer::POA_impl* root,
    OB::ORBControl_ptr orbControl)
{
    CORBA::String_var serverId = orbInstance_ -> getServerId();
    CORBA::String_var serverInstance = orbInstance_ -> getServerInstance();

    Properties_var properties = orbInstance_ -> getProperties();
    const char* noIMR = properties -> getProperty("ooc.orb.noIMR");
    if(strlen(serverId) == 0 || noIMR != 0)
	return;

    //
    // Create the OAD::ProcessMonitor servant
    //
    OAD::ProcessEndpoint_impl* processEndpointImpl =
	new OAD::ProcessEndpoint_impl(serverId, serverInstance, root,
				      orbControl);

    //
    // We assign this to a _var type to ensure that it's destroyed
    // appropriately
    //
    processEndpoint_ = processEndpointImpl;

    IMR::Domain_var imrDomain;
    try
    {
        InitialServiceManager_var initServiceManager =
            orbInstance_ -> getInitialServiceManager();
        Object_var imrObj = initServiceManager ->
            resolveInitialReferences("IMR");
        imrDomain = IMR::Domain::_narrow(imrObj);
    }
    catch(const ORB::InvalidName&)
    {
	// Ignore -- this will be handled later
    }

    Logger_var logger = orbInstance_ -> getLogger();

    //
    // IMR::IMRDomain not reachable?
    //
    if(is_nil(imrDomain))
    {
	String_var msg = string_dup(serverId);
	msg += ": IMRDomain not reachable";
	logger -> error(msg);
	throw INITIALIZE();
    }

    //
    // Check if we should try to register with the IMR.
    //
    const char* exec = properties -> getProperty("ooc.orb.imr.register");
    if(exec != 0)
    {
        char host[MAXHOSTNAMELEN + 1];
	gethostname(host, MAXHOSTNAMELEN);

	try
	{
	    imrDomain -> registerServer(serverId, exec, host);
	}
	catch(const IMR::ServerAlreadyRegistered&)
	{
	    String_var msg = string_dup(serverId);
	    msg += ": (IMR) Server already registered";
	    logger -> error(msg);
	    throw INITIALIZE();
	}
	catch(const NO_PERMISSION&)
	{
	    String_var msg = string_dup(serverId);
	    msg += ": (IMR) Not running in administrative mode";
	    logger -> error(msg);
	    throw INITIALIZE();
	}
    }

    //
    // Tell the IMR that we're starting up
    //
    try
    {
	//
	// This is the ProcessEndpointManager
	//
        OAD::ProcessEndpointManager_var endpoint;

	PortableInterceptor::ObjectReferenceTemplate_var primary =
	    root -> adapter_template();

	//
	// Tell the IMR that we are STARTING.
	//
	activeState_  = imrDomain -> startup(serverId, serverInstance,
				 	     primary, endpoint);

	IMR::ProcessID pid;
#ifdef WIN32
	pid = GetCurrentProcessId();
#else
        pid = getpid();
#endif

	//
	// Link with the OAD ProcessEndpoint
	//
	OAD::ProcessEndpoint_var ref = processEndpointImpl -> _this();
	endpoint -> establish_link(serverId, serverInstance, pid, ref);
        endpointManager_ = Object::_duplicate(endpoint);

	//
	// Create an register the IORInterceptor for the IMR
	//
	PortableInterceptor::IORInterceptor_var i = new
	    PortableInterceptor::IMRIORInterceptor_impl(logger, activeState_,
	    						serverInstance);
	PIManager_var manager = orbInstance_ -> getPIManager();

	try
	{
	    manager -> addIORInterceptor(i, true);
	}
	catch(const PortableInterceptor::ORBInitInfo::DuplicateName&)
	{
	    assert(false);
	}
    }
    catch(const BAD_PARAM&)
    {
	String_var msg = string_dup(serverId);
	msg += ": (IMR) Server already running";
	logger -> error(msg);
	throw INITIALIZE();
    }
    catch(const IMR::NoSuchServer&)
    {
	String_var msg = string_dup(serverId);
	msg += ": (IMR) Not registered with IMR";
	logger -> error(msg);
	throw INITIALIZE();
    }
    catch(const IMR::NoSuchOAD&)
    {
	String_var msg = string_dup(serverId);
	msg += ": (IMR) No OAD for host";
	logger -> error(msg);
	throw INITIALIZE();
    }
    catch(const IMR::OADNotRunning&)
    {
	String_var msg = string_dup(serverId);
	msg += ": (IMR) OAD not registered";
	logger -> error(msg);
	throw INITIALIZE();
    }
    catch(const OAD::AlreadyLinked&)
    {
	String_var msg = string_dup(serverId);
	msg += ": (IMR) Process registered with OAD";
	logger -> error(msg);
	throw INITIALIZE();
    }
}
