// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/MarshalStubImpl.h>
#include <OB/DirectStubImpl.h>
#include <OB/Upcall.h>
#include <OB/ValueBase.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/RefCounted.h>
#include <OB/OAInterface.h>
#include <OB/PIManager.h>

#include <Server.h>
#include <POAManager_impl.h>
#include <POA_impl.h>
#include <POAOAInterface_impl.h>
#include <POACurrent_impl.h>
#include <BootManager_impl.h>
#include <POALocator.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// POAManager_impl private member implementation
// ----------------------------------------------------------------------

//
// If we're in the context of a method invocation returns true, false
// otherwise.
//
bool
OBPortableServer::POAManager_impl::isInORBUpcall()
{
    //
    // Find out whether we're inside a method invocation
    //
    bool inInvocation = false;
    try
    {
	InitialServiceManager_var initialServiceManager =
	    orbInstance_ -> getInitialServiceManager();
	CORBA::Object_var o = initialServiceManager ->
	    resolveInitialReferences("POACurrent");
	PortableServer::Current_impl* current =
	    dynamic_cast<PortableServer::Current_impl*>(o.in());

	inInvocation = (current != 0 && current -> _OB_inUpcall());
	if(inInvocation)
	{
	    //
 	    // Check whether or not the request is dispatched in this
 	    // POAManager's ORB or another ORB.
	    //
	    PortableServer::POA_var poa = current -> get_POA();
	    POA_impl* p = dynamic_cast<POA_impl*>(poa.in());
	    inInvocation = (p -> _OB_ORBInstance() == orbInstance_.in());
	}
    }
    catch(const ORB::InvalidName&)
    {
    }

    return inInvocation;
}

//
// Wait for all pending requests to complete
//
void
OBPortableServer::POAManager_impl::waitPendingRequests()
    throw(SystemException)
{
    //
    // Wait for all pending requests from all POAs to complete
    //
    POASyncHashTable::Enumerator keys = poas_.keys();
    for(ULong i = 0; i < keys.length(); ++i)
    {
	PortableServer::POA_var poa;
	if(poas_.get(keys[i], poa))
	{
	    POA_impl* poaImpl = dynamic_cast<POA_impl*>(poa.in());
	    poaImpl -> _OB_waitPendingRequests();
	}
    }
}

//
// Etherealize each of the servants associated with each POA
//
void
OBPortableServer::POAManager_impl::etherealizePOAs()
{
    try
    {
	InitialServiceManager_var initialServiceManager =
	    orbInstance_ -> getInitialServiceManager();
	CORBA::Object_var o = initialServiceManager ->
	    resolveInitialReferences("RootPOA");
	OBPortableServer::POA_impl* rootPOA =
	    dynamic_cast<OBPortableServer::POA_impl*>(o.in());
	assert(rootPOA);

	//
	// Etherealize recursively from the RootPOA and only POAs
	// associated to this POAManager.
	//
	rootPOA -> _OB_etherealize(this);
    }
    catch(const ORB::InvalidName&)
    {
	assert(false);
    }
}

// ----------------------------------------------------------------------
// POAManager_impl constructor and destructor
// ----------------------------------------------------------------------

OBPortableServer::POAManager_impl::POAManager_impl(
    ORBInstance_ptr orbInstance,
    POALocator* poaLocator,
    const char* name,
    CORBA::Long id,
    const OCI::AcceptorSeq& acceptors)
    throw(SystemException)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      state_(PortableServer::POAManager::HOLDING),
      acceptors_(acceptors),
      name_(name),
      id_(id),
      bootManager_(orbInstance -> getBootManager()),
      poaLocator_(poaLocator)
{
    try
    {
        //
	// Set the server id
	//
	serverId_ = orbInstance_ -> getServerId();
	if(strlen(serverId_) == 0)
	    serverId_ = string_dup("_RootPOA");

        //
        // Create the OAInterface
        //
        oaInterface_ = new POAOAInterface_impl(this, orbInstance_);

        //
        // Construct the root name of the property key for this POAManager
        // instance
        //
        String_var rootKey = string_dup("ooc.orb.poamanager.");
        rootKey += name_;
        rootKey += ".";
        ULong rootKeyLen = strlen(rootKey);

        //
        // Get the Logger
        //
        Logger_var logger = orbInstance_ -> getLogger();

        //
        // Get the ORB properties
        //
        Properties_var properties = orbInstance -> getProperties();

        //
        // The set of properties that the POAManager supports
        //
        const char* props[] =
        {
            "conc_model",
            "endpoint",
            "version"
        };
        int numProps = sizeof(props)/sizeof(props[0]);

        //
        // Check over the POAManager properties and find out whether an
        // unknown property is present
        //
        Properties::KeySeq keys = properties -> getKeys();
        for(ULong j = 0 ; j < keys.length() ; j++)
        {
            const char* key = keys[j];

            //
            // Ignore ooc.orb.oa.thread_pool
            //
            if(strcmp(key, "ooc.orb.oa.thread_pool") == 0)
                continue;

	    //
    	    // Remove the property prefix
	    //
            const char* prop = key;
            if(strncmp(key, rootKey, rootKeyLen) == 0)
                prop += rootKeyLen;
            else if(strncmp(key, "ooc.orb.oa.", 11) == 0)
                prop += 11;
            else
                continue;

            //
            // Check for a match among the supported properties
            //
            int i;
            for(i = 0 ; i < numProps ; i++)
                if(strcmp(prop, props[i]) == 0)
                    break;

	    //
	    // Was the property present?
	    //
	    if(i >= numProps)
	    {
                String_var err = key;
	        err += ": unknown property";
                logger -> warning(err);
	    }
        }

        //
        // Configuration parameters
        //
        version_.major = 1;
        version_.minor = 2;
        Server::ConcModel concModel = Server::Reactive;

        //
        // Parse the individual properties
        //
        for(int i = 0 ; i < numProps ; i++)
        {
	    const char* key = props[i];

	    //
	    // First check the specific POAManager key
	    //
	    String_var fullkey = rootKey;
	    fullkey += key;
	    const char* value = properties -> getProperty(fullkey);

	    //
	    // If the specific POAManager key doesn't have a value then
	    // check the default value "ooc.orb.oa.*"
	    //
	    if(value == 0)
	    {
	        fullkey = string_dup("ooc.orb.oa.");
	        fullkey += key;
	        value = properties -> getProperty(fullkey);
	    }

	    //
	    // No value
	    //
	    if(value == 0)
	        continue;

	    if(strcmp(key, "conc_model") == 0)
	    {
	        if(strcmp(value, "reactive") == 0)
		    concModel = Server::Reactive;
	        else if(strcmp(value, "threaded") == 0)
		    concModel = Server::Threaded;
	        //
	        // Technically the only valid values for
	        // ooc.orb.poamanager.*.conc_model are "reactive" and
	        // "threaded" since this is the communications conc model.
	        // However, we'll also accept the following values since
	        // we might be parsing "ooc.orb.oa.conc_model" (which
	        // represents the default value for both the comm conc
	        // model *and* the method dispatch model).
	        //
	        else if(strcmp(value, "thread_per_client") == 0)
		    concModel = Server::Threaded;
	        else if(strcmp(value, "thread_per_request") == 0)
		    concModel = Server::Threaded;
	        else if(strcmp(value, "thread_pool") == 0)
		    concModel = Server::Threaded;
	        else
	        {
		    String_var err = fullkey;
		    err += ": unknown value";
		    logger -> warning(err);
	        }
    	    }
            else if(strcmp(key, "version") == 0)
            {
                if(strcmp(value, "1.0") == 0)
                {
                    version_.major = 1;
                    version_.minor = 0;
                }
                else if(strcmp(value, "1.1") == 0)
                {
                    version_.major = 1;
                    version_.minor = 1;
                }
                else if(strcmp(value, "1.2") == 0)
                {
                    version_.major = 1;
                    version_.minor = 2;
                }
                else
                {
                    String_var err = fullkey;
                    err += ": expected `1.0', `1.1', or `1.2'";
                    logger -> error(err);
                    throw INITIALIZE();
                }
            }
        }

        //
        // Create the server manager
        //
        serverManager_ = new ServerManager(orbInstance_,
					   acceptors_, 
					   oaInterface_,
					   concModel);

	//
	// Get the boot manager implementation
	//
	bootManagerImpl_ = dynamic_cast<BootManager_impl*>(bootManager_.in());
    }
    catch(...)
    {
        _OB_setRef(0);
        throw;
    }
}

OBPortableServer::POAManager_impl::~POAManager_impl()
{
}

// ----------------------------------------------------------------------
// POAManager_impl public member implementation
// ----------------------------------------------------------------------

//
// Mapping for PortableServer::POAManager
//
void
OBPortableServer::POAManager_impl::activate()
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // If the POA manager is in inactive state then raise the
    // AdapterInactive exception, if it's already in the active state
    // do nothing.
    //
    if(state_ == PortableServer::POAManager::INACTIVE)
        throw PortableServer::POAManager::AdapterInactive();

    //
    // COMPILERFIX: Needed for the VA C++ 5.0.0.2 compiler.
    //
#if defined(_AIX)
  if(state_ != PortableServer::POAManager::ACTIVE)
  {
#else
    if(state_ == PortableServer::POAManager::ACTIVE)
	return;
#endif

    //
    // Switch to the active state.
    //
    state_ = PortableServer::POAManager::ACTIVE;
    
    //
    // Notify all of a state transition
    //
    notifyAll();

    //
    // Activate the server manager
    //
    serverManager_ -> activate();
    
    //
    // Tell the OAInterface to accept requests
    //
    oaInterface_ -> activate();
    
    //
    // Call the state change interceptor
    //
    PIManager_var piManager = orbInstance_ -> getPIManager();
    piManager -> adapterManagerStateChange(id_, _OB_getAdapterState());
#if defined(_AIX)
  }
#endif
}

void
OBPortableServer::POAManager_impl::hold_requests(Boolean waitCompletion)
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException)
{
    {
	JTCSynchronized synchronized(*this);
	
	//
	// If the POA manager is in inactive state then raise the
	// AdapterInactive exception, if it's already in the holding
	// state, do nothing.
	//
	if(state_ == PortableServer::POAManager::INACTIVE)
	    throw PortableServer::POAManager::AdapterInactive();
	
	if(state_ == PortableServer::POAManager::HOLDING)
	    return;

	if(waitCompletion && isInORBUpcall())
	    throw BAD_INV_ORDER(0, COMPLETED_NO);   

	//
	// Switch to the holding state.
	//
	state_ = PortableServer::POAManager::HOLDING;
	
	//
	// Notify all of a state transition
	//
	notifyAll();
        
	//
	// Ask the server manager to hold
	//
	serverManager_ -> hold();
	
	//
	// Tell the OAInterface to accept requests
	//
	oaInterface_ -> activate();
	
	//
	// Call the state change interceptor
	//
	PIManager_var piManager = orbInstance_ -> getPIManager();
	piManager -> adapterManagerStateChange(id_, _OB_getAdapterState());
    }

    //
    // Wait for all pending requests to complete, if asked
    //
    if(waitCompletion)
	waitPendingRequests();
}

void
OBPortableServer::POAManager_impl::discard_requests(Boolean waitCompletion)
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException)
{
    {
	JTCSynchronized synchronized(*this);
	
	//
	// If the POA manager is in inactive state then raise the
	// AdapterInactive exception, if it's already in the
	// discarding state, do nothing.
	//
	if(state_ == PortableServer::POAManager::INACTIVE)
	    throw PortableServer::POAManager::AdapterInactive();
	
	if(state_ == PortableServer::POAManager::DISCARDING)
	    return;

	if(waitCompletion && isInORBUpcall())
	    throw BAD_INV_ORDER(0, COMPLETED_NO);   
	
	//
	// Switch to the discarding state.
	//
	state_ = PortableServer::POAManager::DISCARDING;
	
	//
	// Notify all of a state transition
	//
	notifyAll();
	    
	//
	// Tell the OAInterface to discard requests
	//
	oaInterface_ -> discard();
	
	//
	// Call the state change interceptor
	//
	PIManager_var piManager = orbInstance_ -> getPIManager();
	piManager -> adapterManagerStateChange(id_, _OB_getAdapterState());
    }

    //
    // Wait for all pending requests to complete, if asked
    //
    if(waitCompletion)
	waitPendingRequests();
}

void
OBPortableServer::POAManager_impl::deactivate(Boolean etherealize,
					      Boolean waitCompletion)
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException)
{
    {
	JTCSynchronized synchronized(*this);
    
	if(state_ == PortableServer::POAManager::INACTIVE)
	{
	    //
	    // COMPILERFIX: Needed for the VA C++ 5.0.0.2 compiler.
	    //
#if defined(_AIX)
	    synchronized.~JTCSynchronized();
#endif
	    return;
	}

	if(waitCompletion && isInORBUpcall())
	    throw BAD_INV_ORDER(0, COMPLETED_NO);

	//
	// Destroy the server manager
	//
	serverManager_ -> destroy();
	
	//
	// Clear the acceptor sequence - we must not hold references
	// to the acceptors, because they could be implemented in
	// dynamically-loaded libraries which are unloaded before the
	// POAManager destructor is invoked
	//
	acceptors_.length(0);
	
	//
	// Set the state to INACTIVE *after* the serverManager_ has
	// been destroyed, to avoid a race condition.
	//
	state_ = PortableServer::POAManager::INACTIVE;
	
	//
	// Notify all of a state transition
	//
	notifyAll();
	
	//
	// Call the state change interceptor
	//
	PIManager_var piManager = orbInstance_ -> getPIManager();
	piManager -> adapterManagerStateChange(id_, _OB_getAdapterState());
    }
    
    //
    // Wait for all pending requests to complete, if asked
    //
    if(waitCompletion)
	waitPendingRequests();
	
    //
    // Etherealize each of the servants associated with each POA
    //
    if(etherealize)
	etherealizePOAs();
}

PortableServer::POAManager::State
OBPortableServer::POAManager_impl::get_state()
    throw(SystemException)
{
    //
    // We'll remove the synchronization at present since this is a
    // simple state variable
    //
    //JTCSynchronized synchronized(*this);

    return state_;
}

//
// Mapping for OBPortableServer::POAManager
//

char*
OBPortableServer::POAManager_impl::name()
    throw(SystemException)
{
    return string_dup(name_.in());
}

OCI::AcceptorSeq*
OBPortableServer::POAManager_impl::get_acceptors()
     throw(PortableServer::POAManager::AdapterInactive,
	   SystemException)
{
    JTCSynchronized synchronized(*this);

    if(state_ == PortableServer::POAManager::INACTIVE)
	throw PortableServer::POAManager::AdapterInactive();

    return new OCI::AcceptorSeq(acceptors_);
}

//
// Internal methods.
//

//
// Register a POA with this POAManager
//
void
OBPortableServer::POAManager_impl::_OB_addPOA(PortableServer::POA_ptr poa,
					      const StringSeq& id)
    throw()
{
    JTCSynchronized synchronized(*this);

    assert(!poas_.containsKey(id));
    poas_.put(id, PortableServer::POA::_duplicate(poa));

    poaLocator_ -> add(poa, id);
}

//
// Un-register a POA with this POAManager
//
void
OBPortableServer::POAManager_impl::_OB_removePOA(const StringSeq& id)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    assert(poas_.containsKey(id));
    poas_.remove(id);

    poaLocator_ -> remove(id);
}

DirectStubImpl_ptr
OBPortableServer::POAManager_impl::_OB_createDirectStubImpl(
    const OCI::ObjectKey& key,
    RefCountPolicyList_ptr policies)
    throw(PortableServer::POAManager::AdapterInactive,
	  SystemException, LocationForward)
{
    JTCSynchronized synchronized(*this);

    if(state_ == PortableServer::POAManager::INACTIVE)
	throw PortableServer::POAManager::AdapterInactive();
    
    ObjectKeyData data;
    if(ParseObjectKey(key, data))
    {
        PortableServer::POA_var poa = _OB_locatePOA(data);
        if(!is_nil(poa))
        {
            POA_impl* poaImpl = dynamic_cast<POA_impl*>(poa.in());
            return poaImpl -> _OB_createDirectStubImpl(data.oid, policies);
        }
    }

    //
    // Check to see if the BootManager knows of a reference for the
    // ObjectKey. If so, forward the request.
    //
    RefCountIOR_var ior = bootManagerImpl_ -> _OB_locate(key);
    if(!is_nil(ior))
	throw LocationForward(ior, false);

    //
    // In this case there is no POA for a local servant. This is an
    // OBJECT_NOT_EXIST exception.
    //
    throw OBJECT_NOT_EXIST();
    return DirectStubImpl::_nil(); // Some compilers need this
}

PortableServer::POA_ptr
OBPortableServer::POAManager_impl::_OB_locatePOA(const ObjectKeyData& data)
    throw(LocationForward,
	  SystemException)
{
    //
    // If the GIOP engine sends a request while the POAManager is in
    // INACTIVE state, then something is wrong.
    //
    assert(get_state() != PortableServer::POAManager::INACTIVE);

    PortableServer::POA_var poa;
        

    //
    // Now check the POA name
    //
    if(strcmp(data.serverId, serverId_) == 0 && !poas_.get(data.poaId, poa))
    {
	//
	// The POA isn't contained in our local POA table. Ask the
	// POALocator to locate the POA.
	//
	poa = poaLocator_ -> locate(data);

        //
        // If the POA is connected to some other POAManager (and hence
        // some other end-point) then location forward
        //
        if(!is_nil(poa))
        {
            PortableServer::POAManager_var manager = poa -> the_POAManager();
            if(manager.in() != this)
            {
		Object_var obj = poa -> create_reference_with_id(data.oid, "");
                RefCountIOR_var ior = obj -> _OB_IOR();
                throw LocationForward(ior, false);
            }
        }
    }

    //
    // If the POA doesn't match the ObjectKeyData then this POA isn't
    // present.
    //
    if(!is_nil(poa))
    {
        POA_impl* poaImpl = dynamic_cast<POA_impl*>(poa.in());
        if(!poaImpl -> _OB_poaMatches(data, false))
            poa = POA::_nil();
    }

    return poa._retn();
}

CollocatedServer_ptr
OBPortableServer::POAManager_impl::_OB_getCollocatedServer()
    throw()
{
    return serverManager_ -> getCollocatedServer();
}

void
OBPortableServer::POAManager_impl::_OB_validateState()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    while(true)
    {
	//
	// If POAManager::activate() has been called then we're done
	//
	if(state_ == PortableServer::POAManager::ACTIVE)
	    break;

	//
	// If the POAManager is INACTIVE or DISCARDING then throw a
	// TRANSIENT exception
	//
	if(state_ == PortableServer::POAManager::INACTIVE ||
	  state_ == PortableServer::POAManager::DISCARDING)
	    throw TRANSIENT(0, COMPLETED_NO);
	
	//
	// Wait for a state transition
	//
	try
	{
	    wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }
}

const GIOP::Version&
OBPortableServer::POAManager_impl::_OB_getGIOPVersion() const
    throw()
{
    return version_;
}

PortableInterceptor::AdapterManagerId
OBPortableServer::POAManager_impl::_OB_getAdapterManagerId()
    throw(CORBA::SystemException)
{
    return id_;
}

PortableInterceptor::AdapterState
OBPortableServer::POAManager_impl::_OB_getAdapterState()
    throw()
{
    //
    // We'll remove the synchronization at present since this is a
    // simple state variable
    //
    //JTCSynchronized synchronized(*this);

    switch(state_)
    {
    case PortableServer::POAManager::INACTIVE:
	return PortableInterceptor::INACTIVE;

    case PortableServer::POAManager::ACTIVE:
	return PortableInterceptor::ACTIVE;

    case PortableServer::POAManager::HOLDING:
	return PortableInterceptor::HOLDING;

    case PortableServer::POAManager::DISCARDING:
	return PortableInterceptor::DISCARDING;
    }

    assert(false);
    return PortableInterceptor::NON_EXISTENT;
}

OCI::AcceptorSeq* 
OBPortableServer::POAManager_impl::_OB_getAcceptors() const
    throw()
{
    return new OCI::AcceptorSeq(acceptors_);
}
