// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/ORBInstance.h>
#include <OB/RefCounted.h>
#include <OB/NativeTypes.h>
#include <OB/Except.h>
#include <OB/Object.h>
#include <OB/BootManager.h>
#include <OB/OCI.h>
#include <OB/Upcall.h>
#include <OB/Stream.h>

#include <POAOAInterface_impl.h>
#include <POAManager_impl.h>
#include <POA_impl.h>
#include <BootManager_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OAInterface_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OAInterface_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// POAOAInterface_impl constructor
// ----------------------------------------------------------------------

POAOAInterface_impl::POAOAInterface_impl(
    OBPortableServer::POAManager_impl* poaManager, ORBInstance_ptr orbInstance)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      poaManager_(poaManager),
      bootManager_(orbInstance -> getBootManager()),
      discard_(false)
{
    bootManagerImpl_ = dynamic_cast<BootManager_impl*>(bootManager_.in());
}

POAOAInterface_impl::~POAOAInterface_impl()
{
}

//
// Note: For efficiency reasons, this operation assumes ownership
// for all parameter marked with "/**/"
//
Upcall_ptr
POAOAInterface_impl::createUpcall(UpcallReturn_ptr upcallReturn,
				  /**/ OCI::ProfileInfo* profileInfo,
				  /**/ OCI::TransportInfo_ptr transportInfo,
				  ULong requestId,
				  /**/ char* op,
				  /**/ InputStream_ptr in,
				  /**/ IOP::ServiceContextList* requestSCL)
    throw()
{
    Upcall_var upcall;
    try
    {
	//
	// If discarding then throw a TRANSIENT exception
	//
	if(discard_)
	    throw TRANSIENT(0, COMPLETED_NO);

	ObjectKeyData data;
	if(ParseObjectKey(profileInfo -> key, data))
	{
	    while(true)
	    {
		//
		// Locate the POA. This may also throw a TRANSIENT
		// exception if the POA manager is discarding.
		//
		PortableServer::POA_var poa =
		    poaManager_ -> _OB_locatePOA(data);
		if(!is_nil(poa))
		{
		    OBPortableServer::POA_impl* poaImpl =
			dynamic_cast<OBPortableServer::POA_impl*>(poa.in());
		    upcall =
			poaImpl -> _OB_createUpcall(data.oid, upcallReturn,
						    profileInfo,
						    transportInfo, requestId,
						    op, in, requestSCL);
		    //
		    // If _OB_createUpcall returns a nil Upcall object
		    // then we should retry since that means that the
		    // POA is being destroyed
		    //
		    if(is_nil(upcall))
			continue;
		}
		break;
	    }
	}
	else if(!is_nil(upcallReturn))
	{
	    //
	    // Check to see if the BootManager knows of a reference
	    // for the ObjectKey. If so, forward the request.
	    //
	    RefCountIOR_var ior =
		bootManagerImpl_ -> _OB_locate(profileInfo -> key);
	    if(!is_nil(ior))
		throw LocationForward(ior, false);
	}
	//
	// If no upcall has been created then the object simply
	// doesn't exist
	//
	if(is_nil(upcall))
        {
            if(strcmp(op, "_non_existent") == 0 ||
               strcmp(op, "_not_existent") == 0)
            {
                upcall = new Upcall(orbInstance_, upcallReturn, profileInfo,
                                    transportInfo, requestId, op, in,
                                    requestSCL);
                upcall -> preUnmarshal();
                upcall -> postUnmarshal();
                OutputStream_ptr out = upcall -> preMarshal();
                out -> write_boolean(true);
                upcall -> postMarshal();
            }
            else
                throw OBJECT_NOT_EXIST(MinorCannotDispatch, COMPLETED_NO);
        }
    }
    catch(const SystemException& ex)
    {
	upcall = new Upcall(orbInstance_, upcallReturn, profileInfo,
			    transportInfo, requestId, op, in, requestSCL);
	upcall -> setSystemException(ex);
    }
    catch(const LocationForward& ex)
    {
	upcall = new Upcall(orbInstance_, upcallReturn, profileInfo,
			    transportInfo, requestId, op, in, requestSCL);
	upcall -> setLocationForward(ex.ior, ex.perm);
    }

    assert(!is_nil(upcall));
    return upcall._retn();
}

OAInterface::LocateStatus
POAOAInterface_impl::findByKey(const OCI::ObjectKey& key, RefCountIOR_out ior)
    throw()
{
    ObjectKeyData data;
    if(ParseObjectKey(key, data))
    {
	try
	{
	    PortableServer::POA_var poa =
		poaManager_ -> _OB_locatePOA(data);
	    if(!is_nil(poa))
	    {
		OBPortableServer::POA_impl* poaImpl =
		    dynamic_cast<OBPortableServer::POA_impl*>(poa.in());
		poaImpl -> _OB_locateServant(data.oid);
		return OAInterface::OBJECT_HERE;
	    }
	}
	catch(const SystemException&)
	{
	    //
	    // Fall through to default
	    //
	}
	catch(const LocationForward& fwd)
	{
	    ior = RefCountIOR::_duplicate(fwd.ior);
	    return fwd.perm
		? OAInterface::OBJECT_FORWARD_PERM
		: OAInterface::OBJECT_FORWARD;
	}
    }
    else
    {
        //
        // Check to see if the BootManager knows of a reference
        // for the ObjectKey. 
        //
        ior = bootManagerImpl_ -> _OB_locate(key);
        if(!is_nil(ior))
	    return OAInterface::OBJECT_FORWARD;
    }
    return OAInterface::UNKNOWN_OBJECT;
}

OCI::ProfileInfoSeq*
POAOAInterface_impl::getUsableProfiles(const OCI::IOR& ior, const PolicyList&)
    throw()
{
    OCI::AcceptorSeq_var acceptors = poaManager_ -> get_acceptors();

    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;
    for(ULong i = 0 ; i < acceptors -> length() ; i++)
    {
        OCI::ProfileInfoSeq_var seq2 = acceptors[i] -> get_local_profiles(ior);
        for(ULong j = 0 ; j < seq2 -> length() ; j++)
        {
            seq -> length(seq -> length() + 1);
            seq[seq -> length() - 1] = seq2[j];
        }
    }

    return seq._retn();
}

//
// Discard all incoming requests with a TRANSIENT exception
//
// ASYNC SAFE
//
void
POAOAInterface_impl::discard()
    throw()
{
    discard_ = true;
}

//
// Allow associated POAs to receive requests
//
void
POAOAInterface_impl::activate()
    throw()
{
    discard_ = false;
}
