// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/TypeCode.h>
#include <OB/Stream.h>
#include <OB/TypeCodeCache.h>

#include <TypeCodeFactory.h>

using namespace CORBA;
using namespace OB;

//#define OB_TYPECODE_DEBUG

#ifdef OB_TYPECODE_DEBUG
static ULong TCCount = 0;
#endif

static JTCRecursiveMutex* GCMutex = new JTCRecursiveMutex;

class GCMutexDestroyer
{
public:

    ~GCMutexDestroyer()
    {
	delete GCMutex;
	GCMutex = 0;
    }
};

static GCMutexDestroyer destroyer;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(TypeCode_ptr p)
{
    if(p)
	p -> _OB_duplicateWithGC();
}

void
OBRelease(TypeCode_ptr p)
{
    if(p)
	p -> _OB_releaseWithGC();
}

// ----------------------------------------------------------------------
// TypeCode::Label private member implementation
// ----------------------------------------------------------------------
void*
TypeCode::Label::copyValue(TCKind kind, const void* value)
{
    void* val = 0;

    switch(kind)
    {
    case tk_short:
	val = new Short(*(const Short*)value);
	break;	
	
    case tk_ushort:
	val = new UShort(*(const UShort*)value);
	break;
	
    case tk_long:
	val = new Long(*(const Long*)value);
	break;
	
    case tk_ulong:
    case tk_enum:
	val = new ULong(*(const ULong*)value);
	break;
	
    case tk_longlong:
	val = new LongLong(*(const LongLong*)value);
	break;
	
    case tk_ulonglong:
	val = new ULongLong(*(const ULongLong*)value);
	break;
		
    case tk_boolean:
	val = new Boolean(*(const Boolean*)value);
	break;
	
    case tk_char:
	val = new Char(*(const Char*)value);
	break;
    default:
	assert(false);
    }

    return val;
}

void
TypeCode::Label::replace(TCKind kind, void *value)
{
    if(value_)
    {
	switch(kind_)
	{
	case tk_short:
	    delete (Short*)value_;
	    break;
	    
	case tk_ushort:
	    delete (UShort*)value_;
	    break;
	    
	case tk_long:
	    delete (Long*)value_;
	    break;

	case tk_ulong:
	case tk_enum:
	    delete (ULong*)value_;
	    break;

	case tk_longlong:
	    delete (LongLong*)value_;
	    break;

	case tk_ulonglong:
	    delete (ULongLong*)value_;
	    break;

	case tk_boolean:
	    delete (Boolean*)value_;
	    break;
	    
	case tk_char:
	    delete (Char*)value_;
	    break;

	default:
	    assert(false);
	}
    }

    kind_ = kind;
    value_ = value;
    isDefault_ = (value == 0);
}

// ----------------------------------------------------------------------
// TypeCode::Label constructor/desctructor
// ----------------------------------------------------------------------

TypeCode::Label::Label()
    : isDefault_(true), value_(0), kind_(tk_null)
{
}

TypeCode::Label::~Label()
{
    replace(tk_null, 0);
}

// ----------------------------------------------------------------------
// TypeCode::Label public member implementation
// ----------------------------------------------------------------------

TypeCode::Label&
TypeCode::Label::operator=(const TypeCode::Label& label)
{
    if(this != &label)
    {
	if(label.value_)
	{
	    replace(label.kind_, copyValue(label.kind_, label.value_));
	}
	else
	{
	    replace(tk_null, 0);
	}
    }

    return *this;
}

TypeCode::Label&
TypeCode::Label::operator=(const Any& any)
{
    TypeCode_var tc = any.type();
    TypeCode_ptr origTC = GetOrigType_nodup(tc);
    
    if(origTC -> kind() != tk_octet)
    {
	replace(origTC -> kind(), copyValue(origTC -> kind(), any.value()));
    }
    else
    {
	replace(tk_null, 0);
    }

    return *this;
}

void 
TypeCode::Label::setDefault()
{
    replace(tk_null, 0);
}

void
TypeCode::Label::value(TCKind kind, void *value)
{
    replace(kind, value);
}

// ----------------------------------------------------------------------
// TypeCode private and protected member implementation
// ----------------------------------------------------------------------

bool
TypeCode::equivalentRecHelper(TypeCode_ptr t,
                              OB::FixSeq< TypeCode_ptr, int >& history,
                              OB::FixSeq< TypeCode_ptr, int >& otherHistory)
                              const
{
    ULong i;

    if(is_nil(t))
        return false;

    if(t == this)
        return true;

    //
    // Avoid infinite loops
    //
    bool foundLoop = false, foundOtherLoop = false;
    for(i = 0 ; i < history.length() && !foundLoop ; i++)
        if(this == history[i])
            foundLoop = true;
    for(i = 0 ; i < otherHistory.length() && !foundOtherLoop ; i++)
        if(t == otherHistory[i])
            foundOtherLoop = true;
    if(foundLoop && foundOtherLoop)
        return true;

    history.length(history.length() + 1);
    history[history.length() - 1] = (TypeCode_ptr)this;
    otherHistory.length(otherHistory.length() + 1);
    otherHistory[otherHistory.length() - 1] = (TypeCode_ptr)t;

    bool result = equivalentRec(t, history, otherHistory);

    history.length(history.length() - 1);
    otherHistory.length(otherHistory.length() - 1);

    return result;
}

bool
TypeCode::equivalentRec(TypeCode_ptr t,
                        OB::FixSeq< TypeCode_ptr, int >& history,
                        OB::FixSeq< TypeCode_ptr, int >& otherHistory) const
{
    ULong i;

    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
        return recType_ -> equivalentRecHelper(t, history, otherHistory);
    }

    if(t -> recId_.in())
    {
        if(is_nil(t -> recType_))
            throw BAD_PARAM(MinorIncompleteTypeCodeParameter, COMPLETED_NO);
        return equivalentRecHelper(t -> recType_, history, otherHistory);
    }

    TypeCode_ptr tc1 = GetOrigType_nodup((TypeCode_ptr)this);
    TypeCode_ptr tc2 = GetOrigType_nodup(t);

    if(tc1 -> kind_ != tc2 -> kind_)
        return false;

    if(tc1 -> kind_ == tk_objref ||
       tc1 -> kind_ == tk_struct ||
       tc1 -> kind_ == tk_union ||
       tc1 -> kind_ == tk_enum ||
       tc1 -> kind_ == tk_alias ||
       tc1 -> kind_ == tk_value ||
       tc1 -> kind_ == tk_value_box ||
       tc1 -> kind_ == tk_native ||
       tc1 -> kind_ == tk_abstract_interface ||
       tc1 -> kind_ == tk_local_interface ||
       tc1 -> kind_ == tk_except)
    {
        if(strcmp(tc1 -> id_, "") != 0 && strcmp(tc2 -> id_, "") != 0)
        {
            if(strcmp(tc1 -> id_, tc2 -> id_) == 0)
                return true;
            else
                return false;
        }
    }

    // names_ and memberNames_ must be ignored

    if(tc1 -> kind_ == tk_struct ||
       tc1 -> kind_ == tk_union ||
       tc1 -> kind_ == tk_value ||
       tc1 -> kind_ == tk_except)
    {
        if(tc1 -> memberTypes_.length() != tc2 -> memberTypes_.length())
            return false;

        for(i = 0 ; i < tc1 -> memberTypes_.length() ; i++)
            if(!(tc1 -> memberTypes_[i] -> equivalentRecHelper(
                 tc2 -> memberTypes_[i], history, otherHistory)))
                return false;
    }

    if(tc1 -> kind_ == tk_union)
    {
        if(tc1 -> labels_.length() != tc2 -> labels_.length())
            return false;

        //
        // Don't use equivalentRecHelper here
        //
        if(!(tc1 -> discriminatorType_ -> equivalent(tc2 -> discriminatorType_)))
            return false;

        for(i = 0 ; i < tc1 -> labels_.length() ; i++)
        {
	    if(tc1 -> labels_[i].isDefault() != 
	       tc2 -> labels_[i].isDefault()
            )
	    {
		return false;
	    }

	    if(tc1 -> labels_[i].isDefault())
	    {
		continue;
	    }

            const void* v1 = tc1 -> labels_[i].value();
            const void* v2 = tc2 -> labels_[i].value();

	    tc1 = GetOrigType_nodup(tc1);
            switch(tc1 -> kind())
            {
            case tk_short:
                if(*(Short*)v1 != *(Short*)v2)
                    return false;
                break;

            case tk_ushort:
                if(*(UShort*)v1 != *(UShort*)v2)
                    return false;
                break;

            case tk_long:
                if(*(Long*)v1 != *(Long*)v2)
                    return false;
                break;

            case tk_ulong:
            case tk_enum:
                if(*(ULong*)v1 != *(ULong*)v2)
                    return false;
                break;

            case tk_longlong:
                if(*(LongLong*)v1 != *(LongLong*)v2)
                    return false;
                break;

            case tk_ulonglong:
                if(*(ULongLong*)v1 != *(ULongLong*)v2)
                    return false;
                break;

            case tk_char:
                if(*(Char*)v1 != *(Char*)v2)
                    return false;
                break;

            case tk_boolean:
                if(*(Boolean*)v1 != *(Boolean*)v2)
                    return false;
                break;

            case tk_octet:
            default:
                assert(false);
                break;
            }
        }
    }

    if(tc1 -> kind_ == tk_string ||
       tc1 -> kind_ == tk_wstring ||
       tc1 -> kind_ == tk_sequence ||
       tc1 -> kind_ == tk_array)
    {
        if(tc1 -> length_ != tc2 -> length_)
            return false;
    }

    if(tc1 -> kind_ == tk_sequence ||
       tc1 -> kind_ == tk_array ||
       tc1 -> kind_ == tk_value_box ||
       tc1 -> kind_ == tk_alias)
    {
        if(!(tc1 -> contentType_ -> equivalentRecHelper(
             tc2 -> contentType_, history, otherHistory)))
            return false;
    }

    if(tc1 -> kind_ == tk_fixed)
    {
        if(tc1 -> fixedDigits_ != tc2 -> fixedDigits_ ||
           tc1 -> fixedScale_ != tc2 -> fixedScale_)
            return false;
    }

    if(tc1 -> kind_ == tk_value)
    {
         if(tc1 -> memberVisibility_.length() !=
            tc2 -> memberVisibility_.length())
             return false;

         for(i = 0 ; i < tc1 -> memberVisibility_.length() ; i++)
             if(tc1 -> memberVisibility_[i] != tc2 -> memberVisibility_[i])
                 return false;

        if(tc1 -> typeModifier_ != tc2 -> typeModifier_)
            return false;

        if(tc1 -> concreteBaseType_.in() || tc2 -> concreteBaseType_.in())
        {
            if(!(tc1 -> concreteBaseType_.in() &&
                 tc2 -> concreteBaseType_.in()))
                return false;

            //
            // Don't use equivalentRecHelper here
            //
            if(!(tc1 -> concreteBaseType_ ->
                 equivalent(tc2 -> concreteBaseType_)))
                return false;
        }
    }

    return true;
}

TypeCode_ptr
TypeCode::getCompactTypeCodeRec(
    OB::FixSeq< TypeCode_ptr, int >& history,
    OB::FixSeq< TypeCode_ptr, int >& compacted) const
{
    ULong i;

    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> getCompactTypeCodeRec(history, compacted);
    }

    //
    // Avoid infinite loops
    //
    for(i = 0 ; i < history.length(); i++)
        if(this == history[i])
	    return TypeCode::_duplicate(compacted[i]);

    history.length(history.length() + 1);
    history[history.length() - 1] = (TypeCode_ptr)this;

    //
    // Create the new compacted type code (needed for recursive type
    // codes).
    //
    TypeCode_var result = new TypeCode;
    compacted.length(compacted.length() + 1);
    compacted[compacted.length() - 1] = result.in();

    //
    // Compact member names and types.
    //
    OB::StrSeq< int > names;
    names.length(memberNames_.length());
    for(i = 0 ; i < memberNames_.length() ; i++)
	names[i] = string_dup("");

    OB::ObjSeq< TypeCode, int > types;
    types.length(memberTypes_.length());
    for(i = 0 ; i < memberTypes_.length() ; i++)
	types[i] = memberTypes_[i] -> getCompactTypeCodeRec(history, 
							    compacted);
    //
    // Compact content type.
    //
    TypeCode_var content;
    if(contentType_.in())
        content = contentType_ -> getCompactTypeCodeRec(history, 
							compacted);

    //
    // Compact discriminator type.
    //
    TypeCode_var discriminator;
    if(discriminatorType_.in())
	discriminator = discriminatorType_ -> getCompactTypeCodeRec(history, 
								    compacted);

    //
    // Compact concrete base type.
    //
    TypeCode_var concrete;
    if(concreteBaseType_.in())
	concrete = concreteBaseType_ -> getCompactTypeCodeRec(history,
							      compacted);

    switch(kind_)
    {
    case tk_null:
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
        result -> kind_ = kind_;
        break;

    case tk_fixed:
        result -> kind_ = kind_;
        result -> fixedDigits_ = fixedDigits_;
        result -> fixedScale_ = fixedScale_;
        break;

    case tk_objref:
    case tk_abstract_interface:
    case tk_local_interface:
    case tk_native:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        break;

    case tk_struct:
    case tk_except:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        result -> memberNames_ = names;
        result -> memberTypes_ = types;
        break;

    case tk_union:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        result -> memberNames_ = names;
        result -> memberTypes_ = types;
        result -> labels_ = labels_;
        result -> discriminatorType_ = discriminator._retn();
        break;

    case tk_enum:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        result -> memberNames_ = names;
        break;

    case tk_string:
    case tk_wstring:
        result -> kind_ = kind_;
        result -> length_ = length_;
        break;

    case tk_sequence:
    case tk_array:
        result -> kind_ = kind_;
        result -> length_ = length_;
        result -> contentType_ = content._retn();
        break;

    case tk_alias:
    case tk_value_box:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        result -> contentType_ = content._retn();
        break;

    case tk_value:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = "";
        result -> memberNames_ = names;
        result -> memberTypes_ = types;
        result -> memberVisibility_ = memberVisibility_;
        result -> typeModifier_ = typeModifier_;
        if(concrete.in())
            result -> concreteBaseType_ = concrete._retn();
        break;

    default:
        assert(false);
    }

    if(is_nil(result))
        assert(false);

    return result._retn();
}

void
TypeCode::resetTryRelease()
{
    if(releaseVisit_)
	return;
    releaseVisit_ = true;

    tryReleaseVisit_ = false;
    tryRefCount_ = _OB_getRef();
    
    if(recId_.in())
    {
	if(!is_nil(recType_))
	    recType_ -> resetTryRelease();
    }
    else
    {
	if(kind_ == tk_sequence ||
	   kind_ == tk_value_box ||
	   kind_ == tk_array ||
	   kind_ == tk_alias)
	{
	    if(!is_nil(contentType_))
		contentType_ -> resetTryRelease();
	}
	else if(kind_ == tk_struct ||
		kind_ == tk_union ||
		kind_ == tk_value ||
		kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < memberTypes_.length() ; i++)
		if(!is_nil(memberTypes_[i]))
		    memberTypes_[i] -> resetTryRelease();
	}

        if(kind_ == tk_value && !is_nil(concreteBaseType_))
            concreteBaseType_ -> resetTryRelease();
    }

    releaseVisit_ = false;
}

void
TypeCode::tryRelease()
{
    if(tryReleaseVisit_)
	return;
    tryReleaseVisit_ = true;

    if(recId_.in())
    {
	if(!is_nil(recType_))
	{
	    recType_ -> tryRefCount_--;
	    recType_ -> tryRelease();
	}
    }
    else
    {
	if(kind_ == tk_sequence ||
	   kind_ == tk_value_box ||
	   kind_ == tk_array ||
	   kind_ == tk_alias)
	{
	    if(!is_nil(contentType_))
	    {
		contentType_ -> tryRefCount_--;
		contentType_ -> tryRelease();
	    }
	}
	else if(kind_ == tk_struct ||
		kind_ == tk_union ||
		kind_ == tk_value ||
		kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < memberTypes_.length() ; i++)
		if(!is_nil(memberTypes_[i]))
		{
		    memberTypes_[i] -> tryRefCount_--;
		    memberTypes_[i] -> tryRelease();
		}
	}

        if(kind_ == tk_value && !is_nil(concreteBaseType_))
        {
            concreteBaseType_ -> tryRefCount_--;
            concreteBaseType_ -> tryRelease();
        }
    }
}

bool
TypeCode::checkTryRelease(TypeCode_ptr tc)
{
    if(releaseVisit_)
	return true;
    releaseVisit_ = true;

    if(this == tc)
    {
	if(tryRefCount_ > 1)
	{
	    releaseVisit_ = false;
	    return false;
	}
    }
    else
    {
	if(tryRefCount_ > 0)
	{
	    if(isReachable(tc))
	    {
		releaseVisit_ = false;
		return false;
	    }
	}
    }

    if(recId_.in())
    {
	if(!is_nil(recType_))
	{
	    bool res = recType_ -> checkTryRelease(tc);
	    releaseVisit_ = false;
	    return res;
	}
    }
    else
    {
	if(kind_ == tk_sequence ||
	   kind_ == tk_value_box ||
	   kind_ == tk_array ||
	   kind_ == tk_alias)
	{
	    if(!is_nil(contentType_))
	    {
		bool res = contentType_ -> checkTryRelease(tc);
		releaseVisit_ = false;
		return res;
	    }
	}
	else if(kind_ == tk_struct ||
		kind_ == tk_union ||
		kind_ == tk_value ||
		kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < memberTypes_.length() ; i++)
		if(!is_nil(memberTypes_[i]))
		    if(!memberTypes_[i] -> checkTryRelease(tc))
		    {
			releaseVisit_ = false;
			return false;
		    }
	}

        if(kind_ == tk_value && !is_nil(concreteBaseType_))
            if(!concreteBaseType_ -> checkTryRelease(tc))
	    {
		releaseVisit_ = false;
                return false;
	    }
    }

    releaseVisit_ = false;
    return true;
}

bool
TypeCode::isReachable(TypeCode_ptr tc)
{
    if(reachableVisit_)
	return false;
    reachableVisit_ = true;

    if(recId_.in())
    {
	if(!is_nil(recType_))
	{
	    if(recType_.in() == tc)
	    {
		reachableVisit_ = false;
		return true;
	    }
	    else
	    {
		bool res = recType_ -> isReachable(tc);
		reachableVisit_ = false;
		return res;
	    }
	}
    }
    else
    {
	if(kind_ == tk_sequence ||
	   kind_ == tk_value_box ||
	   kind_ == tk_array ||
	   kind_ == tk_alias)
	{
	    if(!is_nil(contentType_))
	    {
		if(contentType_.in() == tc)
		{
		    reachableVisit_ = false;
		    return true;
		}
		else
		{
		    bool res = contentType_ -> isReachable(tc);
		    reachableVisit_ = false;
		    return res;
		}
	    }
	}
	else if(kind_ == tk_struct ||
		kind_ == tk_union ||
		kind_ == tk_value ||
		kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < memberTypes_.length() ; i++)
		if(!is_nil(memberTypes_[i]))
		{
		    if(memberTypes_[i].in() == tc)
		    {
			reachableVisit_ = false;
			return true;
		    }
		    else
		    {
			if(memberTypes_[i] -> isReachable(tc))
			{
			    reachableVisit_ = false;
			    return true;
			}
		    }
		}
	}

        if(kind_ == tk_value && !is_nil(concreteBaseType_))
        {
            if(concreteBaseType_.in() == tc)
	    {
		reachableVisit_ = false;
                return true;
	    }
            else
            {
                if(concreteBaseType_ -> isReachable(tc))
		{
		    reachableVisit_ = false;
                    return true;
		}
            }
        }
    }

    reachableVisit_ = false;
    return false;
}

// ----------------------------------------------------------------------
// TypeCode constructor and destructor
// ----------------------------------------------------------------------

#ifdef OB_TYPECODE_DEBUG
TypeCode::TypeCode(bool st)
#else
TypeCode::TypeCode(bool)
#endif
    : kind_(tk_null), releaseVisit_(false), reachableVisit_(false), gc_(true)
{ 
#ifdef OB_TYPECODE_DEBUG
    if(st && GCMutex != 0)
	GCMutex -> lock();

    TCCount++;

    if(st && GCMutex != 0)
	GCMutex -> unlock();

    OB_COUT << " + " << TCCount << OB_ENDL;
#endif
}

TypeCode::~TypeCode()
{
#ifdef OB_TYPECODE_DEBUG
    if(GCMutex != 0)
	GCMutex -> lock();

    assert(TCCount > 0);
    TCCount--;

    if(GCMutex != 0)
	GCMutex -> unlock();

    OB_COUT << " - " << TCCount << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// TypeCode public member implementation
// ----------------------------------------------------------------------

Boolean
TypeCode::equal(TypeCode_ptr p) const
{
    if(is_nil(p))
        return false;

    if(p == this)
	return true;

    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> equal(p);
    }
    
    if(p -> recId_.in())
    {
        if(is_nil(p -> recType_))
            throw BAD_PARAM(MinorIncompleteTypeCodeParameter, COMPLETED_NO);
	return equal(p -> recType_);
    }
    
    if(kind_ != p -> kind_)
	return false;

    if(kind_ == tk_objref ||
       kind_ == tk_struct ||
       kind_ == tk_union ||
       kind_ == tk_enum ||
       kind_ == tk_alias ||
       kind_ == tk_value ||
       kind_ == tk_value_box ||
       kind_ == tk_native ||
       kind_ == tk_abstract_interface ||
       kind_ == tk_local_interface ||
       kind_ == tk_except)
    {
        if(strcmp(id_, "") != 0 || strcmp(p -> id_, "") != 0)
        {
            if(strcmp(id_, p -> id_) == 0)
                return true;
            else
                return false;
        }

        if(strcmp(name_, "") != 0 || strcmp(p -> name_, "") != 0)
        {
            if(strcmp(name_, p -> name_) != 0)
                return false;
        }
    }
    
    if(kind_ == tk_struct ||
       kind_ == tk_union ||
       kind_ == tk_enum ||
       kind_ == tk_value ||
       kind_ == tk_except)
    {
 	if(memberNames_.length() != p -> memberNames_.length())
 	    return false;
	
 	for(ULong i = 0 ; i < memberNames_.length() ; i++)
        {
            if(strcmp(memberNames_[i], "") != 0 ||
               strcmp(p -> memberNames_[i], "") != 0)
            {
                if(strcmp(memberNames_[i], p -> memberNames_[i]) != 0)
                    return false;
            }
        }
    }
    
    if(kind_ == tk_struct ||
       kind_ == tk_union ||
       kind_ == tk_value ||
       kind_ == tk_except)
    {
	if(memberTypes_.length() != p -> memberTypes_.length())
	    return false;
	
	for(ULong i = 0 ; i < memberTypes_.length() ; i++)
	    if(!memberTypes_[i] -> equal(p -> memberTypes_[i]))
		return false;
    }
    
    if(kind_ == tk_union)
    {
	if(labels_.length() != p -> labels_.length())
	    return false;
	
	if(!discriminatorType_ -> equal(p -> discriminatorType_))
	    return false;

	for(ULong i = 0 ; i < labels_.length() ; i++)
	{
	    if(labels_[i].isDefault() != p -> labels_[i].isDefault())
	    {
		return false;
	    }

	    if(labels_[i].isDefault())
	    {
		continue;
	    }

	    const void* v1 = labels_[i].value();
	    const void* v2 = p -> labels_[i].value();
	    
	    TypeCode_ptr origDisc = GetOrigType_nodup(discriminatorType_);
	    switch(origDisc -> kind())
	    {
	    case tk_short:
		if(*(Short*)v1 != *(Short*)v2)
		    return false;
		break;

	    case tk_ushort:
		if(*(UShort*)v1 != *(UShort*)v2)
		    return false;
		break;

	    case tk_long:
		if(*(Long*)v1 != *(Long*)v2)
		    return false;
		break;

	    case tk_ulong:
	    case tk_enum:
		if(*(ULong*)v1 != *(ULong*)v2)
		    return false;
		break;

	    case tk_longlong:
		if(*(LongLong*)v1 != *(LongLong*)v2)
		    return false;
		break;

	    case tk_ulonglong:
		if(*(ULongLong*)v1 != *(ULongLong*)v2)
		    return false;
		break;

	    case tk_char:
		if(*(Char*)v1 != *(Char*)v2)
		    return false;
		break;

	    case tk_boolean:
		if(*(Boolean*)v1 != *(Boolean*)v2)
		    return false;
		break;

	    case tk_octet:
	    default:
		assert(false);
		break;
	    }
	}
    }
    
    if(kind_ == tk_string ||
       kind_ == tk_wstring ||
       kind_ == tk_sequence ||
       kind_ == tk_array)
    {
	if(length_ != p -> length_)
	    return false;
    }
    
    if(kind_ == tk_sequence ||
       kind_ == tk_array ||
       kind_ == tk_value_box ||
       kind_ == tk_alias)
    {
	if(!contentType_ -> equal(p -> contentType_))
	    return false;
    }

    if(kind_ == tk_fixed)
    {
	if(fixedDigits_ != p -> fixedDigits_ ||
	   fixedScale_ != p -> fixedScale_)
	    return false;
    }

    if(kind_ == tk_value)
    {
 	if(memberVisibility_.length() != p -> memberVisibility_.length())
 	    return false;
	
 	for(ULong i = 0 ; i < memberVisibility_.length() ; i++)
 	    if(memberVisibility_[i] != p -> memberVisibility_[i])
 		return false;

        if(typeModifier_ != p -> typeModifier_)
            return false;

        if(concreteBaseType_.in() || p -> concreteBaseType_.in())
        {
            if(!(concreteBaseType_.in() && p -> concreteBaseType_.in()))
                return false;

            if(!(concreteBaseType_ -> equal(p -> concreteBaseType_)))
                return false;
        }
    }

    return true;
}

Boolean
TypeCode::equivalent(TypeCode_ptr p) const
{
    OB::FixSeq< TypeCode_ptr, int > history, otherHistory;

    bool result = equivalentRecHelper(p, history, otherHistory);

    assert(history.length() == 0);
    assert(otherHistory.length() == 0);

    return result;
}

TypeCode_ptr
TypeCode::get_compact_typecode() const
{
    OB::FixSeq< TypeCode_ptr, int > history, compacted;
    return  getCompactTypeCodeRec(history, compacted);
}

TCKind
TypeCode::kind() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> kind();
    }

    return kind_;
}

const char*
TypeCode::id() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> id();
    }

    if(!(kind_ == tk_objref ||
	 kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_alias ||
	 kind_ == tk_value ||
	 kind_ == tk_value_box ||
	 kind_ == tk_native ||
	 kind_ == tk_abstract_interface ||
	 kind_ == tk_local_interface ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }

    return id_;
}

const char*
TypeCode::name() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> name();
    }

    if(!(kind_ == tk_objref ||
	 kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_alias ||
	 kind_ == tk_value ||
	 kind_ == tk_value_box ||
	 kind_ == tk_native ||
	 kind_ == tk_abstract_interface ||
	 kind_ == tk_local_interface ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }
    
    return name_;
}

ULong
TypeCode::member_count() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> member_count();
    }

    if(!(kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_value ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }
    
    return memberNames_.length();
}

const char*
TypeCode::member_name(ULong index) const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> member_name(index);
    }

    if(!(kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_value ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }

    if(index >= memberNames_.length())
    {
	throw Bounds();
    }

    return memberNames_[index];
}

TypeCode_ptr
TypeCode::member_type(ULong index) const
{
    return TypeCode::_duplicate(_OB_memberType_nodup(index));
}

Any*
TypeCode::member_label(ULong index) const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> member_label(index);
    }

    if(!(kind_ == tk_union))
    {
	throw BadKind();
    }

    if(index >= memberNames_.length())
    {
	throw Bounds();
    }

    if(labels_[index].isDefault())
    {
	//
	// Default label value is the zero octet
	//
	Any_var any = new Any();
	any <<= Any::from_octet(0);
	return any._retn();
    }
    else
    {
	//
	// We create first an any to wrap the label and then, return a
	// copy of it (to return an Any with its own copy of the
	// value).
	//
	Any any(discriminatorType_, (void*)labels_[index].value(), false);
	return new Any(any);
    }
}

TypeCode_ptr
TypeCode::discriminator_type() const
{
    return TypeCode::_duplicate(_OB_discriminatorType_nodup());
}

Long
TypeCode::default_index() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> default_index();
    }

    if(!(kind_ == tk_union))
    {
	throw BadKind();
    }

    for(Long i = 0 ; i < (Long)labels_.length() ; i++)
    {
	if(labels_[i].isDefault())
	{
	    return i;
	}
    }

    return -1;
}

ULong
TypeCode::length() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> length();
    }

    if(!(kind_ == tk_string ||
	 kind_ == tk_wstring ||
	 kind_ == tk_sequence ||
	 kind_ == tk_array))
    {
	throw BadKind();
    }

    return length_;
}

TypeCode_ptr
TypeCode::content_type() const
{
    return TypeCode::_duplicate(_OB_contentType_nodup());
}

UShort
TypeCode::fixed_digits() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> fixed_digits();
    }

    if(!(kind_ == tk_fixed))
    {
	throw BadKind();
    }

    return fixedDigits_;
}

Short
TypeCode::fixed_scale() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> fixed_scale();
    }

    if(!(kind_ == tk_fixed))
    {
	throw BadKind();
    }

    return fixedScale_;
}

Visibility
TypeCode::member_visibility(ULong index) const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> member_visibility(index);
    }

    if(!(kind_ == tk_value))
    {
	throw BadKind();
    }

    if(index >= memberVisibility_.length())
    {
	throw Bounds();
    }

    return memberVisibility_[index];
}

ValueModifier
TypeCode::type_modifier() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> type_modifier();
    }

    if(!(kind_ == tk_value))
    {
	throw BadKind();
    }

    return typeModifier_;
}

TypeCode_ptr
TypeCode::concrete_base_type() const
{
    return TypeCode::_duplicate(_OB_concreteBaseType_nodup());
}

bool
TypeCode::_OB_isSystemException() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> _OB_isSystemException();
    }

    if(kind_ != tk_except)
	return false;

    if(strcmp(id_, "IDL:omg.org/CORBA/UNKNOWN:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_PARAM:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_MEMORY:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/IMP_LIMIT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/COMM_FAILURE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_OBJREF:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_PERMISSION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INTERNAL:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/MARSHAL:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INITIALIZE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_IMPLEMENT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_TYPECODE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_OPERATION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_RESOURCES:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_RESPONSE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/PERSIST_STORE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_INV_ORDER:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TRANSIENT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/FREE_MEM:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_IDENT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_FLAG:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INTF_REPOS:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_CONTEXT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/OBJ_ADAPTER:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/DATA_CONVERSION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/OBJECT_NOT_EXIST:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TRANSACTION_REQUIRED:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TRANSACTION_ROLLEDBACK:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INVALID_TRANSACTION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_POLICY:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/CODESET_INCOMPATIBLE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/REBIND:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TIMEOUT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TRANSACTION_UNAVAILABLE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TRANSACTION_MODE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_QOS:1.0") == 0)
    {
	return true;
    }
    else
    {
	return false;
    }
}

void
TypeCode::_OB_duplicateWithGC()
{
    JTCSyncT< JTCMutex > sync(*this);

    _OB_incRef();
}

void
TypeCode::_OB_releaseWithGC()
{
    //
    // We must manually decrement the reference count since we can't
    // destroy this object while the mutex is locked.
    //
    bool deleteThis = false;
    {
	JTCSyncT< JTCMutex > sync(*this);

	if(!gc_)
	{
	    if(_OB_getRef() > 1)
	    {
		_OB_decRef();
		return;
	    }
	    else
	    {
		_OB_setRef(0);
		deleteThis = true;
	    }
	}
    }

    if(deleteThis)
    {
	delete this;
	return;
    }

    //
    // Only run the garbage collection algorithm if the gc_ flag
    // is true. gc_ is set to false when another object holds a
    // reference on the TypeCode (e.g., the TypeCodeCache) and
    // thus we have the guarantee that the TypeCode won't be
    // released. It's also false for leaf TypeCode (primitive
    // kind, strings,...)
    //
    if(GCMutex != 0)
	GCMutex -> lock();
    
    resetTryRelease();
    tryRelease();
    
    if(checkTryRelease(this))
    {
	//
	// Disable garbage collection of the TypeCode since we are
	// about to release it. This also prevent to run again the
	// algorithm for this TypeCode.
	//
	_OB_disableGC();

	recType_ = _nil();
	contentType_ = _nil();
	memberTypes_.length(0);
	concreteBaseType_ = _nil();

	assert(_OB_getRef() == 1);
    }
    
    if(GCMutex != 0)
	GCMutex -> unlock();
    
    //
    // We must manually decrement the reference count since we can't
    // destroy this object while the mutex is locked.
    //
    {
	JTCSyncT< JTCMutex > sync(*this);
	if(_OB_getRef() > 1)
	{
	    _OB_decRef();
	}
	else
	{
	    _OB_setRef(0);
	    deleteThis = true;
	}
    }

    if(deleteThis)
	delete this;
}

void 
TypeCode::_OB_disableGC()
{ 
    JTCSyncT< JTCMutex > sync(*this); 
    gc_ = false; 
}

void 
TypeCode::_OB_enableGC()
{ 
    JTCSyncT< JTCMutex > sync(*this); 
    gc_ = true; 
}

TypeCode_ptr
TypeCode::_OB_contentType_nodup() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> _OB_contentType_nodup();
    }

    if(!(kind_ == tk_sequence ||
	 kind_ == tk_array ||
	 kind_ == tk_value_box ||
	 kind_ == tk_alias))
    {
	throw BadKind();
    }

    return contentType_;
}

TypeCode_ptr 
TypeCode::_OB_memberType_nodup(ULong index) const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> _OB_memberType_nodup(index);
    }

    if(!(kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_value ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }

    if(index >= memberNames_.length())
    {
	throw Bounds();
    }

    return memberTypes_[index];
}

TypeCode_ptr 
TypeCode::_OB_discriminatorType_nodup() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> _OB_discriminatorType_nodup();
    }

    if(!(kind_ == tk_union))
    {
	throw BadKind();
    }

    return discriminatorType_;
}

TypeCode_ptr 
TypeCode::_OB_concreteBaseType_nodup() const
{
    if(recId_.in())
    {
        if(is_nil(recType_))
            throw BAD_TYPECODE(MinorIncompleteTypeCode, COMPLETED_NO);
	return recType_ -> _OB_concreteBaseType_nodup();
    }

    if(!(kind_ == tk_value))
    {
	throw BadKind();
    }

    return concreteBaseType_;
}

void
TypeCode::_OB_embedRecTC(TypeCode_ptr outer)
{
    //
    // Recursive placeholder TypeCodes are illegal as "outer" argument
    //
    assert(!outer -> recId_.in());

    //
    // Check for illegal recursion
    //
    if(!(outer -> kind_ == tk_struct ||
	 outer -> kind_ == tk_except ||
	 outer -> kind_ == tk_union ||
	 outer -> kind_ == tk_value))
    {
	throw BAD_TYPECODE();
    }

    _OB_embedRecTC(outer, outer);
}

void
TypeCode::_OB_embedRecTC(TypeCode_ptr outer, TypeCode_ptr inner)
{
    //
    // Embed recursive placeholder TypeCodes
    //
    if(inner -> recId_.in())
    {
	if(strcmp(inner -> recId_, outer -> id_) == 0)
	{
	    if(inner -> recType_.in())
	    {
		assert_nca(inner -> recType_.in() == outer, NCAEmbeddedRecTC);
	    }
	    else
	    {
		//
		// Embed the recursive placeholder TypeCode
		//
		inner -> recType_ = TypeCode::_duplicate(outer);
	    }
	}
    }
    else
    {
	//
	// Embed content type
	//
	if(inner -> kind_ == tk_sequence ||
	   inner -> kind_ == tk_value_box ||
	   inner -> kind_ == tk_array ||
	   inner -> kind_ == tk_alias)
	{
	    assert(inner -> contentType_.in() != outer);
	    _OB_embedRecTC(outer, inner -> contentType_);
	}
    
	//
	// Embed member types
	//
	if(inner -> kind_ == tk_struct ||
	   inner -> kind_ == tk_union ||
	   inner -> kind_ == tk_value ||
	   inner -> kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < inner -> memberTypes_.length() ; i++)
	    {
		assert(inner -> memberTypes_[i].in() != outer);
		_OB_embedRecTC(outer, inner -> memberTypes_[i]);
	    }
	}

        //
        // Embed base valuetype
        //
        if(inner -> kind_ == tk_value && inner -> concreteBaseType_.in())
            _OB_embedRecTC(outer, inner -> concreteBaseType_);
    }
}

// ----------------------------------------------------------------------
// TypeCode::Bounds public member implementation
// ----------------------------------------------------------------------

void
TypeCode::Bounds::_OB_insertInAny(Any& any)
{
    any <<= this;
}

void
TypeCode::Bounds::_OB_unmarshal(TypeCode::Bounds&, InputStreamImpl* in)
{
    String_var id = in -> read_string();
    assert(strcmp(id, "IDL:omg.org/CORBA/TypeCode/Bounds:1.0") == 0);
}

// ----------------------------------------------------------------------
// Insertion and extraction for TypeCode::Bounds
// ----------------------------------------------------------------------

void
operator<<=(Any& any, TypeCode::Bounds* v)
{
    static const Info< TypeCode::Bounds > info;
    any.replace(TypeCode::_tc_Bounds, v, true, &info);
}

void
operator<<=(Any& any, const TypeCode::Bounds& v)
{
    any <<= new TypeCode::Bounds(v);
}

Boolean
operator>>=(const Any& any, TypeCode::Bounds*& v)
{
    if(any.check_type(TypeCode::_tc_Bounds))
    {
        if(!any.info())
        {
            InputStream_var in = any.create_input_stream();
            TypeCode::Bounds* val = new TypeCode::Bounds;
            TypeCode::Bounds::_OB_unmarshal(*val, in);
            (Any&)any <<= val;
        }

        v = (TypeCode::Bounds*)any.value();
        return true;
    }
    else
        return false;
}

// ----------------------------------------------------------------------
// TypeCode::BadKind public member implementation
// ----------------------------------------------------------------------

void
TypeCode::BadKind::_OB_insertInAny(Any& any)
{
    any <<= this;
}

void
TypeCode::BadKind::_OB_unmarshal(TypeCode::BadKind&, InputStreamImpl* in)
{
    String_var id = in -> read_string();
    assert(strcmp(id, "IDL:omg.org/CORBA/TypeCode/BadKind:1.0") == 0);
}

// ----------------------------------------------------------------------
// Insertion and extraction for TypeCode::BadKind
// ----------------------------------------------------------------------

void
operator<<=(Any& any, TypeCode::BadKind* v)
{
    static const Info< TypeCode::BadKind > info;
    any.replace(TypeCode::_tc_BadKind, v, true, &info);
}

void
operator<<=(Any& any, const TypeCode::BadKind& v)
{
    any <<= new TypeCode::BadKind(v);
}

Boolean
operator>>=(const Any& any, TypeCode::BadKind*& v)
{
    if(any.check_type(TypeCode::_tc_BadKind))
    {
        if(!any.info())
        {
            InputStream_var in = any.create_input_stream();
            TypeCode::BadKind* val = new TypeCode::BadKind;
            TypeCode::BadKind::_OB_unmarshal(*val, in);
            (Any&)any <<= val;
        }

        v = (TypeCode::BadKind*)any.value();
        return true;
    }
    else
        return false;
}

// ----------------------------------------------------------------------
// _tc_<type> constants and class members for _tc_<type> constants
// ----------------------------------------------------------------------

TypeCodePrimitiveConst::TypeCodePrimitiveConst(CORBA::TCKind kind)
    : kind_(kind)
{
    typeCode_ = TypeCodeFactory::createPrimitiveTC(kind_, true);
}

namespace CORBA
{

TypeCodePrimitiveConst _tc_null(tk_null);
TypeCodePrimitiveConst _tc_void(tk_void);
TypeCodePrimitiveConst _tc_short(tk_short);
TypeCodePrimitiveConst _tc_long(tk_long);
TypeCodePrimitiveConst _tc_longlong(tk_longlong);
TypeCodePrimitiveConst _tc_ushort(tk_ushort);
TypeCodePrimitiveConst _tc_ulong(tk_ulong);
TypeCodePrimitiveConst _tc_ulonglong(tk_ulonglong);
TypeCodePrimitiveConst _tc_float(tk_float);
TypeCodePrimitiveConst _tc_double(tk_double);
TypeCodePrimitiveConst _tc_longdouble(tk_longdouble);
TypeCodePrimitiveConst _tc_boolean(tk_boolean);
TypeCodePrimitiveConst _tc_char(tk_char);
TypeCodePrimitiveConst _tc_wchar(tk_wchar);
TypeCodePrimitiveConst _tc_octet(tk_octet);
TypeCodePrimitiveConst _tc_any(tk_any);
TypeCodePrimitiveConst _tc_TypeCode(tk_TypeCode);
TypeCodePrimitiveConst _tc_Principal(tk_Principal);

} // End namespace CORBA

TypeCodeStringConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
    {
	if(wide_)
	    typeCode_ = TypeCodeFactory::createWstringTC(0);
	else
	    typeCode_ = TypeCodeFactory::createStringTC(0);
    }

    return typeCode_; // No _duplicate() here !
}

TypeCode_ptr
TypeCodeStringConst::_OB_bounded(ULong bound)
{
    if(wide_)
	return TypeCodeFactory::createWstringTC(bound);
    else
	return TypeCodeFactory::createStringTC(bound);
}

namespace CORBA
{

TypeCodeStringConst _tc_string(false);
TypeCodeStringConst _tc_wstring(true);

}

TypeCodeFixedConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
	typeCode_ = TypeCodeFactory::createFixedTC(0, 0);

    return typeCode_; // No _duplicate() here !
}

TypeCode_ptr
TypeCodeFixedConst::_OB_fixed(UShort digits, UShort scale)
{
    return TypeCodeFactory::createFixedTC(digits, scale);
}

namespace CORBA
{

TypeCodeFixedConst _tc_fixed;

}

TypeCodeObjectConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
	typeCode_ =
	    TypeCodeFactory::createInterfaceTC("IDL:omg.org/CORBA/Object:1.0",
					       "Object");

    return typeCode_; // No _duplicate() here !
}

namespace CORBA
{

TypeCodeObjectConst _tc_Object;

}

TypeCodeAbstractBaseConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
	typeCode_ = TypeCodeFactory::createNativeTC(
            "IDL:omg.org/CORBA/AbstractBase:1.0", "AbstractBase");

    return typeCode_; // No _duplicate() here !
}

namespace CORBA
{

TypeCodeAbstractBaseConst _tc_AbstractBase;

}

TypeCodeValueBaseConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
    {
        ValueMemberSeq members;
	typeCode_ =
	    TypeCodeFactory::createValueTC("IDL:omg.org/CORBA/ValueBase:1.0",
					   "ValueBase", VM_ABSTRACT,
                                           TypeCode::_nil(), members);

	//
	// Disable garbage collection, this TypeCode is not recursive!
	//
	typeCode_ -> _OB_disableGC();
    }

    return typeCode_; // No _duplicate() here !
}

namespace CORBA
{

TypeCodeValueBaseConst _tc_ValueBase;

}

TypeCodeSysExConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
    {
	String_var id = string_dup("IDL:omg.org/CORBA/");
	id += name_;
	id += ":1.0";
	String_var name = string_dup(name_);
	StructMemberSeq members;
	members.length(2);
	members[0].name = string_dup("minor");
	members[0].type = TypeCode::_duplicate(_tc_ulong);
	members[1].name = string_dup("completed");
	members[1].type = TypeCode::_duplicate(_tc_completion_status);
	typeCode_ = TypeCodeFactory::createExceptionTC(id, name, members);

	//
	// Disable garbage collection. System exception TypeCode are
	// not recursive!
	//
	typeCode_ -> _OB_disableGC();
    }

    return typeCode_; // No _duplicate() here !
}

namespace CORBA
{

TypeCodeSysExConst _tc_UNKNOWN("UNKNOWN");
TypeCodeSysExConst _tc_BAD_PARAM("BAD_PARAM");
TypeCodeSysExConst _tc_NO_MEMORY("NO_MEMORY");
TypeCodeSysExConst _tc_IMP_LIMIT("IMP_LIMIT");
TypeCodeSysExConst _tc_COMM_FAILURE("COMM_FAILURE");
TypeCodeSysExConst _tc_INV_OBJREF("INV_OBJREF");
TypeCodeSysExConst _tc_NO_PERMISSION("NO_PERMISSION");
TypeCodeSysExConst _tc_INTERNAL("INTERNAL");
TypeCodeSysExConst _tc_MARSHAL("MARSHAL");
TypeCodeSysExConst _tc_INITIALIZE("INITIALIZE");
TypeCodeSysExConst _tc_NO_IMPLEMENT("NO_IMPLEMENT");
TypeCodeSysExConst _tc_BAD_TYPECODE("BAD_TYPECODE");
TypeCodeSysExConst _tc_BAD_OPERATION("BAD_OPERATION");
TypeCodeSysExConst _tc_NO_RESOURCES("NO_RESOURCES");
TypeCodeSysExConst _tc_NO_RESPONSE("NO_RESPONSE");
TypeCodeSysExConst _tc_PERSIST_STORE("PERSIST_STORE");
TypeCodeSysExConst _tc_BAD_INV_ORDER("BAD_INV_ORDER");
TypeCodeSysExConst _tc_TRANSIENT("TRANSIENT");
TypeCodeSysExConst _tc_FREE_MEM("FREE_MEM");
TypeCodeSysExConst _tc_INV_IDENT("INV_IDENT");
TypeCodeSysExConst _tc_INV_FLAG("INV_FLAG");
TypeCodeSysExConst _tc_INTF_REPOS("INTF_REPOS");
TypeCodeSysExConst _tc_BAD_CONTEXT("BAD_CONTEXT");
TypeCodeSysExConst _tc_OBJ_ADAPTER("OBJ_ADAPTER");
TypeCodeSysExConst _tc_DATA_CONVERSION("DATA_CONVERSION");
TypeCodeSysExConst _tc_OBJECT_NOT_EXIST("OBJECT_NOT_EXIST");
TypeCodeSysExConst _tc_TRANSACTION_REQUIRED("TRANSACTION_REQUIRED");
TypeCodeSysExConst _tc_TRANSACTION_ROLLEDBACK("TRANSACTION_ROLLEDBACK");
TypeCodeSysExConst _tc_INVALID_TRANSACTION("INVALID_TRANSACTION");
TypeCodeSysExConst _tc_INV_POLICY("INV_POLICY");
TypeCodeSysExConst _tc_CODESET_INCOMPATIBLE("CODESET_INCOMPATIBLE");
TypeCodeSysExConst _tc_REBIND("REBIND");
TypeCodeSysExConst _tc_TIMEOUT("TIMEOUT");
TypeCodeSysExConst _tc_TRANSACTION_UNAVAILABLE("TRANSACTION_UNAVAILABLE");
TypeCodeSysExConst _tc_TRANSACTION_MODE("TRANSACTION_MODE");
TypeCodeSysExConst _tc_BAD_QOS("BAD_QOS");

} // End namespace CORBA

TypeCodeConst::~TypeCodeConst()
{
    if(!is_nil(typeCode_))
    {
	//
	// Re-enable garbage collection of this TypeCode.
	//
	typeCode_ -> _OB_enableGC();
    }    
}

TypeCodeConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
    {
	assert(str_ != 0 || arr_ != 0);
	Octet* oct;
        ULong len;
        if(str_ != 0)
        {
            oct = AsciiToOctets(str_);
            len = strlen(str_);
        }
        else
        {
            len = 0;
            const char** cp;
            for(cp = arr_ ; *cp != 0 ; cp++)
                len += strlen(*cp);
            String_var s = string_alloc(len);
            s[0] = '\0';
            for(cp = arr_ ; *cp != 0 ; cp++)
                strcat(s.inout(), *cp);
            oct = AsciiToOctets(s);
        }
        OCI::BufferImpl buf(len / 2);
        memcpy(buf.data(), oct, len / 2);
        delete [] oct;
	InputStreamImpl in(&buf, 0, false);
	in._OB_readEndian();
	typeCode_ = in.read_TypeCode();

	//
	// Disable garbage collection. As long as the TypeCodeConst
	// object is not destroyed we hold a reference on it and we're
	// sure that the TypeCode will never be destroyed (garbage
	// collection is re-enabled when the TypeCodeConst object is
	// destroyed).
	//
	typeCode_ -> _OB_disableGC();

	//
	// If the repository ID of this TypeCode is not empty, add the
	// TypeCode to the cache.
	//
	if(*(typeCode_ -> id_.in()) != '\0')
	{
	    TypeCodeCache::instance() -> put(typeCode_ -> id_, typeCode_);
	}
    }
    
    return typeCode_; // No _duplicate() here !
}

TypeCode::TypeCodeConstForBounds::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
    {
	const char* id = "IDL:omg.org/CORBA/TypeCode/Bounds:1.0";
	const char* name = "Bounds";
	StructMemberSeq members;
	typeCode_ = TypeCodeFactory::createExceptionTC(id, name, members);

	//
	// Disable garbage collection, this TypeCode is not recursive!
	//
	typeCode_ -> _OB_disableGC();
    }
    
    return typeCode_; // No _duplicate() here !
}

TypeCode::TypeCodeConstForBounds CORBA::TypeCode::_tc_Bounds;

TypeCode::TypeCodeConstForBadKind::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
    {
	const char* id = "IDL:omg.org/CORBA/TypeCode/BadKind:1.0";
	const char* name = "BadKind";
	StructMemberSeq members;
	typeCode_ = TypeCodeFactory::createExceptionTC(id, name, members);

	//
	// Disable garbage collection, this TypeCode is not recursive!
	//
	typeCode_ -> _OB_disableGC();
    }
    
    return typeCode_; // No _duplicate() here !
}

TypeCode::TypeCodeConstForBadKind CORBA::TypeCode::_tc_BadKind;

// ----------------------------------------------------------------------
// Get original type, i.e. remove aliases
// ----------------------------------------------------------------------

TypeCode_ptr
OB::GetOrigType(TypeCode_ptr p)
{
    TypeCode_var tc = TypeCode::_duplicate(p);

    while(tc -> kind() == tk_alias)
	tc = tc -> content_type();

    return tc._retn();
}

TypeCode_ptr
OB::GetOrigType_nodup(TypeCode_ptr p)
{
    TypeCode_ptr tc = p;

    while(tc -> kind() == tk_alias)
	tc = tc -> _OB_contentType_nodup();

    return tc;
}

