// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/ValueBase.h>
#include <OB/ValueFactoryBase.h>
#include <OB/ValueFactoryManager.h>
#include <OB/Any.h>
#include <OB/Except.h>
#include <OB/TypeCodeConst.h>
#include <OB/StringValue.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::ValueFactoryManager_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::ValueFactoryManager_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// ValueFactoryManager constructor and destructor
// ----------------------------------------------------------------------

ValueFactoryManager::ValueFactoryManager()
    : destroy_(false)
{
    //
    // Create the hashtable
    //
    factories_ = new FactoryTable(1023);

    //
    // Install factories for standard value box types
    //
    ValueFactory factory;
    factory = new StringValue_init;
    registerValueFactory(StringValue::_OB_id(), factory);
    factory -> _remove_ref(); // give ownership to hashtable
    factory = new WStringValue_init;
    registerValueFactory(WStringValue::_OB_id(), factory);
    factory -> _remove_ref(); // give ownership to hashtable
}

ValueFactoryManager::~ValueFactoryManager()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// ValueFactoryManager public member functions
// ----------------------------------------------------------------------

void
ValueFactoryManager::destroy()
{
    JTCSynchronized synchronized(*this);

    assert(!destroy_); // May only be destroyed once
    destroy_ = true;

    //
    // Destroy the hashtable
    //
    delete factories_;
}

ValueFactory
ValueFactoryManager::registerValueFactory(const char* id, ValueFactory factory)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    assert_nca(id != 0, NCANullString);
    assert_nca(factory != 0, NCANullValue);

    ValueFactoryBase_var old;
    factories_ -> get(id, old);

    factory -> _add_ref();
    factories_ -> put(id, factory);

    //
    // Caller assumes ownership of previously registered factory (if any)
    //
    return old._retn();
}

void
ValueFactoryManager::unregisterValueFactory(const char* id)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    assert_nca(id != 0, NCANullString);

    if(!factories_ -> containsKey(id))
        throw BAD_PARAM(MinorValueFactoryError, COMPLETED_NO);

    factories_ -> remove(id);
}

ValueFactory
ValueFactoryManager::lookupValueFactory(const char* id)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    assert_nca(id != 0, NCANullString);
    
    //
    // Caller must invoke _remove_ref
    //
    ValueFactoryBase_var factory;
    if(factories_ -> get(id, factory))
        return factory._retn();

    return 0;
}
