// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>

#include <Util.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include <CPPUtil.h>

using namespace CORBA;
using namespace OB;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] idl-files ...\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show ORBacus version.\n"
	"-d, --debug            Print diagnostic messages.\n"
	"-DNAME                 Define NAME as 1.\n"
	"-DNAME=DEF             Define NAME as DEF.\n"
	"-UNAME                 Remove any definition for NAME.\n"
	"-IDIR                  Put DIR in the include file search path.\n"
        "--case-sensitive       Allow case-sensitive identifiers.\n"
        "                       (Not CORBA compliant.)\n"
	;
}

int
run(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Get options
    //
    String_var cppArgs;

    bool caseSensitive = false;

    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
	if(strcmp(argv[i], "--help") == 0 ||
	   strcmp(argv[i], "-h") == 0)
	{
	    usage(progName);
	    return 0;
	}
	else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
	{
	    OB_CERR << "ORBacus " << Version << OB_ENDL;
	    return 0;
	}
	else if(strcmp(argv[i], "--debug") == 0 ||
		strcmp(argv[i], "-d") == 0)
	{
	    extern int yydebug;
	    yydebug = 1;
	}
	else if(strncmp(argv[i], "-D", 2) == 0 ||
		strncmp(argv[i], "-U", 2) == 0 ||
		strncmp(argv[i], "-I", 2) == 0)
	{
	    if(strlen(argv[i]) == 2)
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i] << "'"
			<< OB_ENDL; // not argv[i - 1] !
		return 1;
	    }

	    cppArgs += ' ';
	    cppArgs += argv[i];
	}
	else if(strcmp(argv[i], "--case-sensitive") == 0)
	{
            OB_CERR << progName << ": Warning: use of the --case-sensitive "
                    << "option is not CORBA compliant!" << OB_ENDL;
	    caseSensitive = true;
	}
	else
	{
	    OB_CERR << progName << ": unknown option `"
		    << argv[i] << "'\n" << OB_ENDL;
	    usage(progName);
	    return 1;
	}
    }

    if(i == argc)
    {
	OB_CERR << progName << ": no input file" << OB_ENDL;
	usage(progName);
	return 1;
    }

    Object_var obj;
    try
    {
        obj = orb -> resolve_initial_references("InterfaceRepository");
    }
    catch(ORB::InvalidName&)
    {
    }
    if(is_nil(obj))
    {
	OB_CERR << "No `InterfaceRepository' initial reference." << OB_ENDL;
	return 1;
    }

    Repository_var repository = Repository::_narrow(obj);
    if(is_nil(repository))
    {
	OB_CERR
	    << "`InterfaceRepository' initial reference is not a repository."
	    << OB_ENDL;
	return 1;
    }

    char* idlcpp = IdlGetCPreprocessor(progName);
    if(!idlcpp)
	return 1;
    String_var cppCmd = idlcpp;

    int status = 0;
    while(status == 0 && i < argc)
    {
	String_var file = string_dup(argv[i++]);
	
	//
	// Check for input file
	//
	struct stat statBuf;
	if(stat(file, &statBuf) == -1)
	{
	    OB_CERR << progName << ": can't open \"" << file
		    << "\": " << strerror(errno) << OB_ENDL;
	    return 1;
	}

	//
	// Run C preprocessor with the input file as argument
	//
	OB_TRANSLATOR_IMPORT extern FILE* yyin;
	String_var cmd = cppCmd;
	cmd += cppArgs;
	cmd += ' ';
	cmd += file;

	OBProcessIO pio;
	if((yyin = pio.open(cmd, "r")) == NULL)
	    return 1;

	//
	// Do parsing
	//
	int IdlParse(ORB_ptr, Repository_ptr, StringSeq&,
		     bool, bool, bool, bool, bool, bool, bool,
		     IdlStringHash&, IdlStringHash&);
	StringSeq includeSeq;
	IdlStringHash idHash;
	IdlStringHash fwdHash;
	status = IdlParse(orb, repository, includeSeq, true, false,
                          caseSensitive, true, true, true, true,
                          idHash, fwdHash);
 	
	//
	// Close input file
	//
	pio.close();
    }

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;

    try
    {
	//
	// Initialize the ORB
	//
	orb = ORB_init(argc, argv);

	//
	// Generate code
	//
	status = run(orb, argc, argv);
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
