// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the forward declaration header file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusFwdH(Container_ptr container,
				  IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

            Identifier_var name = fixKwd(desc -> name);

            out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusFwdH(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_LocalInterface:
	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

	    Identifier_var name = fixKwd(desc -> name);

	    out.sep();
	    out << "\nclass " << name << ';';
	    out << "\ntypedef " << name << "* " << name << "_ptr;";
	    out << "\ntypedef " << name << "* " << name << "Ref;";

	    //
	    // Emit the TypeCode constant declaration if necessary
	    //
	    if(!noTC_)
	    {
		out.sep();
		
		out << '\n';
		
		if(strlen(dllImport_) > 0)
		    out << dllImport_ << ' ';

		out << "extern OB::TypeCodeConst " << "_tc_"
		    << name << ';';
	    }
            break;
	}

	case dk_Value:
	{
	    //
	    // Valuetype
	    //
	    const ValueDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

            if(strcmp(desc -> id, "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
                continue;

	    Identifier_var name = fixKwd(desc -> name);

	    out.sep();
	    out << "\nclass " << name << ';';

	    //
	    // Emit the TypeCode constant declaration if necessary
	    //
            if(!noTC_)
            {
                out.sep();

                out << '\n';

                if(strlen(dllImport_) > 0)
                    out << dllImport_ << ' ';

		out << "extern OB::TypeCodeConst " << "_tc_"
		    << name << ';';
            }
            break;
	}

	case dk_ValueBox:
	{
	    //
	    // Value box
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

	    Identifier_var name = fixKwd(desc -> name);

	    out.sep();
	    out << "\nclass " << name << ';';

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the forward declaration header file, global
// namespace part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusFwdHGlobalNamespace(Container_ptr container,
						 IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    genCPlusPlusFwdHGlobalNamespace(cont, out);

	    break;
	}

	case dk_LocalInterface:
	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

            ScopedName_var absolute = getAbsolute(desc -> id);

	    //
	    // Emit OBUpcast declaration if necessary
	    //
	    if(withInterceptorArgs_)
	    {
		if(seqElem.kind == dk_Interface)
		{
		    out.sep();
		    out << "\nCORBA::Object_ptr OBUpcast("
			<< absolute << "_ptr);";
		}
		else if(seqElem.kind == dk_AbstractInterface)
		{
		    out.sep();
		    out << "\nCORBA::AbstractBase_ptr OBUpcast("
			<< absolute << "_ptr);";
		}
	    }

	    out.sep();
	    out << "\nvoid OBDuplicate(" << absolute << "_ptr);";
	    out << "\nvoid OBRelease(" << absolute << "_ptr);";

	    out.sep();
	    out << "\nvoid OBMarshal(" << absolute
		<< "_ptr, OB::OutputStreamImpl*);";
	    out << "\nvoid OBUnmarshal(" << absolute
		<< "_ptr&, OB::InputStreamImpl*);";

            break;
	}

	case dk_Value:
	{
	    //
	    // Valuetype
	    //
	    const ValueDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

            if(strcmp(desc -> id, "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
                continue;

            ScopedName_var absolute = getAbsolute(desc -> id);

	    //
	    // Emit OBUpcast declaration if necessary
	    //
	    if(withInterceptorArgs_)
	    {
		out.sep();
		out << "\nCORBA::ValueBase* OBUpcast(" << absolute << "*);";
	    }

	    out.sep();
	    out << "\nvoid OBDuplicate(" << absolute << "*);";
	    out << "\nvoid OBRelease(" << absolute << "*);";

	    out.sep();
	    out << "\nvoid OBMarshal(" << absolute
		<< "*, OB::OutputStreamImpl*);";
	    out << "\nvoid OBUnmarshal(" << absolute
		<< "*&, OB::InputStreamImpl*);";

            break;
	}

	case dk_ValueBox:
	{
	    //
	    // Value box
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

            ScopedName_var absolute = getAbsolute(desc -> id);

	    out.sep();
	    out << "\nvoid OBDuplicate(" << absolute << "*);";
	    out << "\nvoid OBRelease(" << absolute << "*);";

	    out.sep();
	    out << "\nvoid OBMarshal(" << absolute
		<< "*, OB::OutputStreamImpl*);";
	    out << "\nvoid OBUnmarshal(" << absolute
		<< "*&, OB::InputStreamImpl*);";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the forward declaration header file, _var and
// _out definitions
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusFwdHVar(Container_ptr container,
				     IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

            Identifier_var name = fixKwd(desc -> name);

            out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusFwdHVar(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_LocalInterface:
	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

	    Identifier_var name = fixKwd(desc -> name);

	    out.sep();
	    out << "\ntypedef OB::ObjVar< " << name << " > "
                << name << "_var;";
            out << "\ntypedef OB::ObjOut< " << name << " > "
                << name << "_out;";

            break;
	}

	case dk_Value:
	{
	    //
	    // Valuetype
	    //
	    const ValueDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

            if(strcmp(desc -> id, "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
                continue;

	    Identifier_var name = fixKwd(desc -> name);

	    out.sep();
	    out << "\ntypedef OB::ObjVar< " << name << " > "
                << name << "_var;";
            out << "\ntypedef OB::ObjOut< " << name << " > "
                << name << "_out;";

            break;
	}

	case dk_ValueBox:
	{
	    //
	    // Value box
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

	    Identifier_var name = fixKwd(desc -> name);

	    out.sep();
	    out << "\ntypedef OB::ObjVar< " << name << " > "
                << name << "_var;";
            out << "\ntypedef OB::ObjOut< " << name << " > "
                << name << "_out;";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the forward declaration header file, StubImpl
// part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusFwdHStubImpl(Container_ptr container,
                                          IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("OBStubImpl", seqElem.contained_object);

	    out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusFwdHStubImpl(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_LocalInterface:
	    break;

	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

            String_var name =
                getPrefixedName("OBStubImpl", seqElem.contained_object);

	    out.sep();
	    out << "\nclass " << name << ';';
	    out << "\ntypedef " << name << "* " << name << "_ptr;";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the forward declaration header file, StubImpl
// global namespace part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusFwdHStubImplGlobalNamespace(Container_ptr
							 container,
							 IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    assert(!is_nil(cont));
	    genCPlusPlusFwdHStubImplGlobalNamespace(cont, out);

	    break;
	}

	case dk_LocalInterface:
	    break;

	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

            ScopedName_var absolute =
                getPrefixedAbsolute("OBStubImpl", desc -> id, "");

	    out.sep();
	    out << "\nvoid OBDuplicate(" << absolute << "_ptr);";
	    out << "\nvoid OBRelease(" << absolute << "_ptr);";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the forward declaration header file, StubImpl
// _var definitions
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusFwdHStubImplVar(Container_ptr container,
					     IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("OBStubImpl", seqElem.contained_object);

	    out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusFwdHStubImplVar(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_LocalInterface:
	    break;

	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

            String_var name =
                getPrefixedName("OBStubImpl", seqElem.contained_object);

	    out.sep();
	    out << "\ntypedef OB::ObjVar< " << name << " > "
                << name << "_var;";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the forward declaration header file, MarshalStubImpl
// part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusFwdHMarshalStubImpl(Container_ptr container,
						 IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("OBMarshalStubImpl", seqElem.contained_object);

	    out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusFwdHMarshalStubImpl(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_LocalInterface:
	    break;

	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!checkFwd(desc -> id))
                continue;

            String_var name =
                getPrefixedName("OBMarshalStubImpl", seqElem.contained_object);

	    out.sep();
	    out << "\nclass " << name << ';';
	    out << "\ntypedef " << name << "* " << name << "_ptr;";

            break;
	}

        default:
            break;
        }
    }
}
