// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenECPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate initialization code for the base classes
// ----------------------------------------------------------------------

void
IdlECPPGenerator::writeInitBaseClasses(
    Container_ptr container,
    InterfaceDefSeq& baseIntfs,
    IdlPrettyPrint& out
)
{
    InterfaceDef_var intf = InterfaceDef::_narrow(container);
    IdlInterfaceInfoSeq baseInfo;
    IdlGetBaseInterfaces(intf, baseInfo, false);

    for(ULong j = 0 ; j < baseInfo.length() ; j++)
    {
	//
	// Impl classes are not generated for abstract interfaces
	//
	if(!baseInfo[j].is_abstract)
	{
	    ULong i;
	    InterfaceDef:: FullInterfaceDescription_var baseDesc =
		baseInfo[j].intf->describe_interface();
	    for(i = 0; i < baseIntfs.length(); i++)
	    {
		InterfaceDef:: FullInterfaceDescription_var addedDesc =
		    baseIntfs[i]->describe_interface();

		if (strcmp(baseDesc -> name, addedDesc -> name) == 0)
		    break;
	    }
	    if(i == baseIntfs.length())
	    {
		InterfaceDef:: FullInterfaceDescription_var desc =
		    baseInfo[j].intf->describe_interface();
		ScopedName_var absolute = getAbsolute(desc -> id);
		String_var name = getFlattenedName(absolute);

		out << ",\n" << name << "_impl(poa)";

		baseIntfs.length(baseIntfs.length() + 1);
		baseIntfs[baseIntfs.length() - 1] = baseInfo[j].intf;
	    }
	}

	if(!noVirtual_)
	    writeInitBaseClasses(baseInfo[j].intf, baseIntfs, out);
    }
}

// ----------------------------------------------------------------------
// Generate code for the example source file
// ----------------------------------------------------------------------

void
IdlECPPGenerator::genCPlusPlusImplC(Container_ptr container, bool all,
                                   IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j, k;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

        switch(seqElem.kind)
        {
        case dk_Module:
        {
            //
            // Module
            //
            const ModuleDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
		continue;

            comment(desc -> id, out);

            out.sep();

            assert(!is_nil(cont));
            genCPlusPlusImplC(cont, all, out);

            break;
        }

        case dk_AbstractInterface:
            //
            // No sample implementation for abstract interface
            //
            break;

	case dk_LocalInterface:
        case dk_Interface:
        {
            //
            // Interface
            //
            const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);
	    String_var flatAbsolute = getFlattenedName(absolute);

            comment(desc -> id, out);

	    if(seqElem.kind == dk_LocalInterface)
            {
                out.sep();
                out << '\n' << flatAbsolute << "_impl::"
		    << flatAbsolute << "_impl()";
                out.start();
                out.end();
            }
            else
            {
                out.sep();
                out << '\n' << flatAbsolute << "_impl::" << flatAbsolute
		    << "_impl(PortableServer::POA_ptr poa)";
                out.inc();
                out << "\n: poa_(PortableServer::POA::_duplicate(poa))";
		InterfaceDefSeq baseIntfs;
		writeInitBaseClasses(cont, baseIntfs, out);
                out.dec();
                out.start();
                out.end();
            }
            out.sep();
            out << '\n' << flatAbsolute << "_impl::~"
		<< flatAbsolute << "_impl()";
            out.start();
            out.end();

	    if(seqElem.kind != dk_LocalInterface)
            {
                //
                // _default_POA
                //
                out.sep();
                out << "\nPortableServer::POA_ptr"
                    << '\n' << flatAbsolute << "_impl::_default_POA("
		    << "CORBA::Environment& ev)";
                out.start();
                out << "\nreturn PortableServer::POA::_duplicate(poa_);";
                out.end();
            }

            //
            // Generate attributes and operations for any base abstract
            // interfaces
            //
            if(!all)
            {
                InterfaceDef_var intf = InterfaceDef::_narrow(cont);
                assert(!is_nil(intf));
                IdlInterfaceInfoSeq baseInfo;
                IdlGetBaseInterfaces(intf, baseInfo, false);

                StringSeq ids;

                for(j = 0 ; j < baseInfo.length() ; j++)
                {
                    if(!baseInfo[j].is_abstract)
                        continue;

		    ScopedName_var dummy;
		    // No _var for seq! See comments for the
		    // implementation of "describe"
		    Container::DescriptionSeq* contents =
			describe(baseInfo[j].intf, dummy.out());

                    for(k = 0 ; k < contents -> length() ; k++)
                    {
                        if((*contents)[k].kind == dk_Operation)
                        {
                            const OperationDescription* opDesc;
                            Boolean b = ((*contents)[k].value) >>= opDesc;
                            assert(b);
			    ULong l;
                            for(l = 0 ; l < ids.length() ; l++)
                                if(strcmp(ids[l], opDesc -> id) == 0)
                                    break;
                            if(l >= ids.length())
                            {
                                genCPlusPlusImplCOperation(flatAbsolute,
							   opDesc, true, out);
                                ids.length(ids.length()+1);
                                ids[ids.length()-1] = opDesc -> id;
                            }
                        }
                        else if((*contents)[k].kind == dk_Attribute)
                        {
                            const AttributeDescription* atDesc;
                            Boolean b = ((*contents)[k].value) >>= atDesc;
                            assert(b);
			    ULong l;
                            for(l = 0 ; l < ids.length() ; l++)
                                if(strcmp(ids[l], atDesc -> id) == 0)
                                    break;
                            if(l >= ids.length())
                            {
                                genCPlusPlusImplCAttribute(flatAbsolute,
							   atDesc, true, out);
                                ids.length(ids.length()+1);
                                ids[ids.length()-1] = atDesc -> id;
                            }
                        }
                    }
                }
            }

            //
            // Attributes and operations
            //

            Container::DescriptionSeq_var contents =
                cont -> describe_contents(dk_all, !all, -1);

            for(j = 0 ; j < contents -> length() ; j++)
            {
                if(contents[j].kind == dk_Attribute)
                {
                    const AttributeDescription* atDesc;
                    Boolean b = (contents[j].value >>= atDesc);
                    assert(b);
                    genCPlusPlusImplCAttribute(flatAbsolute, atDesc,
					       true, out);
                }
                else if(contents[j].kind == dk_Operation)
                {
                    const OperationDescription* opDesc;
                    Boolean b = (contents[j].value >>= opDesc);
                    assert(b);
                    genCPlusPlusImplCOperation(flatAbsolute, opDesc,
					       true, out);
                }
            }

            break;
        }

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for an example attribute implementation
// ----------------------------------------------------------------------

void
IdlECPPGenerator::genCPlusPlusImplCAttribute(
    const char* absolute,
    const AttributeDescription* desc,
    bool /* ex */,
    IdlPrettyPrint& out
)
{
    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);
    out.sep();

    TypeCode_var origType = GetOrigType(desc -> type);
    String_var ret = getTypeString("", desc -> type, GetTypeReturn);
    out << '\n' << ret << '\n' << absolute << "_impl::" << name
	<< "(CORBA::Environment& ev)";
    out.start();
    out << "\n// TODO: Implementation";
    genCPlusPlusImplCReturn(desc -> type, "_r", out);
    out << "\nreturn _r;";
    out.end();

    if(desc -> mode == ATTR_NORMAL)
    {
        out.sep();

        String_var in = getTypeString("", desc -> type, GetTypeIn);

        out << "\nvoid\n" << absolute << "_impl::" << name << '(' << in
            << " a, CORBA::Environment& ev)";
        out.start();
        out << "\n// TODO: Implementation";
        out.end();
    }
}

// ----------------------------------------------------------------------
// Generate code for an example operation implementation
// ----------------------------------------------------------------------

void
IdlECPPGenerator::genCPlusPlusImplCOperation(const char* absolute,
                                             const OperationDescription* desc,
                                             bool /* ex */,
                                             IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);
    out.sep();

    String_var ret = getTypeString("", desc -> result, GetTypeReturn);

    out << '\n' << ret << '\n' << absolute << "_impl::" << name << '(';

    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    ULong i;

    for(i = 0 ; i < desc -> parameters.length() ; i++)
    {
        GetType tt;
        switch(desc -> parameters[i].mode)
        {
        case PARAM_IN:
            tt = GetTypeIn;
            break;

        case PARAM_INOUT:
            tt = GetTypeInOut;
            break;

        case PARAM_OUT:
            tt = GetTypeOut;
            break;
        }

        String_var argName = fixKwd(desc -> parameters[i].name);

        String_var s = getTypeString("", desc -> parameters[i].type,
                                           tt, false, argName);

        if(i > 0)
            out << ",\n";

        out << s;
    }

    if(desc -> contexts.length() > 0)
    {
        if(desc -> parameters.length() > 0)
            out << ",\n";
    
        out << "CORBA::Context_ptr context";
    }

    if(desc -> parameters.length() > 0)
	out << ",\n";
    out << "CORBA::Environment& ev";

    out << ')';

    out.setIndent(indent);

    out.start();
    out << "\n// TODO: Implementation";

    if(desc -> result -> kind() != tk_void)
        genCPlusPlusImplCReturn(desc -> result, "_r", out);

    //
    // Show example code for those out and inout parameters requiring
    // allocation
    //
    for(i = 0 ; i < desc -> parameters.length() ; i++)
    {
        if(desc -> parameters[i].mode == PARAM_IN)
            continue;

        String_var argName = fixKwd(desc -> parameters[i].name);

        //
        // Get original type
        //
        TypeCode_var origType =
            GetOrigType(desc -> parameters[i].type);

        //
        // Check if original type is fixed-sized or variable-sized
        //
        bool variable = IdlIsVariable(origType);

        String_var norm =
            getTypeString("", desc -> parameters[i].type, GetTypeNormal);
        String_var ret =
            getTypeString("", desc -> parameters[i].type, GetTypeReturn);

        TCKind kind = origType -> kind();
        switch(kind)
        {
        case tk_void:
        case tk_null:
        case tk_alias:
        case tk_except:
            assert(false);
            break;

        case tk_boolean:
        case tk_short:
        case tk_long:
        case tk_longlong:
        case tk_ushort:
        case tk_ulong:
        case tk_ulonglong:
        case tk_octet:
        case tk_float:
        case tk_double:
        case tk_longdouble:
        case tk_char:
        case tk_wchar:
        case tk_enum:
        case tk_fixed:
            // nothing to do
            break;

        case tk_string:
            if(desc -> parameters[i].mode == PARAM_INOUT)
                out << "\nCORBA::string_free(" << argName << ");";
            out << '\n' << argName << " = CORBA::string_dup(\"\");";
            break;

        case tk_wstring:
            if(desc -> parameters[i].mode == PARAM_INOUT)
                out << "\nCORBA::wstring_free(" << argName << ");";
            out << '\n' << argName << " = CORBA::wstring_dup(L\"\");";
            break;

        case tk_any:
            if(desc -> parameters[i].mode == PARAM_OUT)
                out << '\n' << argName << " = new " << norm << ';';
            break;

        case tk_TypeCode:
        case tk_Principal:
        case tk_objref:
        case tk_abstract_interface:
	case tk_local_interface:
            if(desc -> parameters[i].mode == PARAM_INOUT)
                out << "\nCORBA::release(" << argName << ");";
            out << '\n' << argName << " = " << norm << "::_nil();";
            break;

        case tk_struct:
        case tk_union:
            if(desc -> parameters[i].mode == PARAM_OUT && variable)
                out << '\n' << argName << " = new " << norm << ';';
            break;

        case tk_sequence:
            if(desc -> parameters[i].mode == PARAM_OUT)
                out << '\n' << argName << " = new " << norm << ';';
            break;

        case tk_array:
            if(desc -> parameters[i].mode == PARAM_OUT && variable)
            {
                String_var arr =
                    getTypeString("", desc -> parameters[i].type,
                                  GetTypeArray);
                out << '\n' << argName << " = " << arr << "_alloc();";
            }
            break;

        case tk_native:
            out << '\n' << argName << " = (" << norm << ")0;";
            break;

        case tk_value:
        case tk_value_box:
            out << '\n' << argName << " = 0;";
            break;
        }
    }

    if(desc -> result -> kind() != tk_void)
        out << "\nreturn _r;";

    out.end();
}

// ----------------------------------------------------------------------
// Generate example return statement
// ----------------------------------------------------------------------

void
IdlECPPGenerator::genCPlusPlusImplCReturn(TypeCode_ptr type,
                                         const char* name,
                                         IdlPrettyPrint& out)
{
    //
    // Get original type
    //
    TypeCode_var origType = GetOrigType(type);

    //
    // Check if original type is fixed-sized or variable-sized
    //
    bool variable = IdlIsVariable(origType);

    String_var norm = getTypeString("", type, GetTypeNormal);
    String_var ret = getTypeString("", type, GetTypeReturn);

    //
    // Get result
    //
    TCKind kind = origType -> kind();
    switch(kind)
    {
    case tk_void:
    case tk_null:
    case tk_alias:
    case tk_except:
        assert(false);
        break;

    case tk_boolean:
        out << '\n' << ret << ' ' << name << " = false;";
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_octet:
        out << '\n' << ret << ' ' << name << " = 0;";
        break;

    case tk_float:
    case tk_double:
    case tk_longdouble:
        out << '\n' << ret << ' ' << name << " = 0.0;";
        break;

    case tk_char:
        out << '\n' << ret << ' ' << name << " = '\\0';";
        break;

    case tk_wchar:
        out << '\n' << ret << ' ' << name << " = L'\\0';";
        break;

    case tk_enum:
        out << '\n' << ret << ' ' << name << " = (" << ret << ")0;";
        break;

    case tk_fixed:
        out << '\n' << ret << ' ' << name << "(\"0\");";
        break;

    case tk_string:
        out << '\n' << ret << ' ' << name << " = CORBA::string_dup(\"\");";
        break;

    case tk_wstring:
        out << '\n' << ret << ' ' << name << " = CORBA::wstring_dup(L\"\");";
        break;

    case tk_any:
        out << '\n' << ret << ' ' << name << " = new " << norm << ';';
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
    case tk_local_interface:
        out << '\n' << ret << ' ' << name << " = " << norm << "::_nil();";
        break;

    case tk_struct:
    case tk_union:
        if(variable)
            out << '\n' << ret << ' ' << name << " = new " << norm << ';';
        else
            out << '\n' << ret << ' ' << name << ';';
        break;

    case tk_sequence:
        out << '\n' << ret << ' ' << name << " = new " << norm << ';';
        break;

    case tk_array:
    {
        String_var arr = getTypeString("", type, GetTypeArray);
        out << '\n' << ret << ' ' << name << " = " << arr << "_alloc();";
        break;
    }

    case tk_native:
        out << '\n' << ret << ' ' << name << " = (" << norm << ")0;";
        break;

    case tk_value:
    case tk_value_box:
        out << '\n' << ret << ' ' << name << " = 0;";
        break;
    }
}

// ----------------------------------------------------------------------
// Generate code for the sample client
// ----------------------------------------------------------------------

void
IdlECPPGenerator::genECPPImplClient(Container_ptr container,
				    IdlPrettyPrint& out)
{
    //
    // Generate the caller functions
    //
    genECPPImplClientCallers(container, out);

    //
    // run
    //
    out.sep();
    out << "\nstatic int"
        << "\nrun(";
    out.inc();
    out << "\nCORBA::ORB_ptr orb,";
    out << "\nint argc,";
    out << "\nchar** argv,";
    out << "\nCORBA::Environment& ev";
    out.dec();
    out << "\n)";
    out.start();
    genECPPImplClientCallCallers(container, out);
    out << "\nreturn EXIT_SUCCESS;";
    out.end();

    //
    // main
    //
    out.sep();
    out << "\nint"
        << "\nmain(";
    out.inc();
    out << "\nint argc,";
    out << "\nchar** argv";
    out.dec();
    out << "\n)";
    out.start();
    out << "\nint status = EXIT_SUCCESS;";
    out << "\nCORBA::Environment ev;";
    out << "\nCORBA::ORB_var orb;";

    out << "\norb = CORBA::ORB_init(argc, argv, \"\", ev);";
    out << "\nif (ev.exception() == 0)";
    out.start();
    out << "\nstatus = run(orb, argc, argv, ev);";
    out.end();
    out << "\nif (ev.exception() != 0)";
    out.start();
    out << "\nCORBA::Exception* ex = ev.exception();";
    out << "\nfprintf(stderr, \"%s\\n\", ex->_rep_id());";
    out << "\nstatus = EXIT_FAILURE;";
    out.end();

    out.sep();
    out << "\nif (!CORBA::is_nil(orb))";
    out.start();
    out << "\nev.clear();";
    out << "\norb->destroy(ev);";
    out << "\nif (ev.exception() != 0)";
    out.start();
    out << "\nCORBA::Exception* ex = ev.exception();";
    out << "\nfprintf(stderr, \"%s\\n\", ex->_rep_id());";
    out << "\nstatus = EXIT_FAILURE;";
    out.end();
    out.end();

    out.sep();
    out << "\nreturn status;";

    out.end();
}

// ----------------------------------------------------------------------
// Generate caller functions for the sample client
// ----------------------------------------------------------------------

void
IdlECPPGenerator::genECPPImplClientCallers(Container_ptr container,
					   IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
        //
        // Get element description
        //
        Container::Description& seqElem = (*seq)[i];

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seqElem.contained_object);

        switch(seqElem.kind)
        {
        case dk_Module:
        {
            //
            // Module
            //
            const ModuleDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            assert(!is_nil(cont));
            genECPPImplClientCallers(cont, out);

            break;
        }

        case dk_Interface:
        {
            //
            // Interface
            //
            const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            //
            // Generate functions to call attributes and operations
            //
            genECPPImplClientCallers(cont, out);

            break;
        }

        case dk_Attribute:
        {
            //
            // Attribute
            //
            const AttributeDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            String_var definedIn = getAbsolute(desc -> defined_in);
	    String_var flatIntfName = getFlattenedName(definedIn);

            //
            // Getter
            //
            {
                out.sep();
                out << "\nstatic void"
                    << "\ncall_" << flatIntfName << "__get_" << desc -> name
                    << '(';
                out.inc();
                out << "\nCORBA::Object_ptr _obj,";
                out << "\nCORBA::ORB_ptr _orb,";
		out << "\nCORBA::Environment& ev";
                out.dec();
                out << "\n)";
                out.start();
                String_var str = getTypeString("", desc -> type, GetTypeReturn);
		out << '\n' << definedIn << "_var " << flatIntfName
		    << " = " << definedIn << "::_narrow(_obj, ev);";
		out << "\nassert(!CORBA::is_nil(" << flatIntfName << "));";
                out << '\n' << str << " _val = " << flatIntfName << "->"
                    << desc -> name << "(ev);";
		out << "\nif (ev.exception() == 0)";
		out.start();
                out << "\n// Success";
		out.end();
                out.end();
            }

            //
            // Setter
            //
            if(desc -> mode == ATTR_NORMAL)
            {
                out.sep();
                out << "\nstatic void"
                    << "\ncall_" << flatIntfName << "__set_"
		    << desc -> name << '(';
                out.inc();
                out << "\nCORBA::Object_ptr _obj,";
                out << "\nCORBA::ORB_ptr _orb,";
		out << "\nCORBA::Environment& ev";
                out.dec();
                out << "\n)";
                out.start();
                String_var str = getTypeString("", desc -> type, GetTypeNormal);
                TypeCode_ptr origType = GetOrigType_nodup(desc -> type);
                if(origType -> kind() == tk_objref ||
                   origType -> kind() == tk_TypeCode)
		{
		    str += "_var";
		}
		out << '\n' << definedIn << "_var " << flatIntfName
		    << " = " << definedIn << "::_narrow(_obj, ev);";
		out << "\nassert(!CORBA::is_nil(" << flatIntfName << "));";
                out << '\n' << str << " _val;";
                out << '\n' << flatIntfName << "->" << desc -> name
                    << "(" << "_val, ev);";
		out << "\nif (ev.exception() == 0)";
		out.start();
                out << "\n// Success";
		out.end();
                out.end();
            }

            break;
        }

        case dk_Operation:
        {
            //
            // Operation
            //
            const OperationDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            String_var definedIn = getAbsolute(desc -> defined_in);
	    String_var flatIntfName = getFlattenedName(definedIn);

            out.sep();
            out << "\nstatic void"
                << "\ncall_" << flatIntfName << '_' << desc -> name << '(';
            out.inc();
            out << "\nCORBA::Object_ptr _obj,";
            out << "\nCORBA::ORB_ptr _orb,";
	    out << "\nCORBA::Environment& ev";
            out.dec();
            out << "\n)";
            out.start();
	    out << '\n' << definedIn << "_var " << flatIntfName
		<< " = " << definedIn << "::_narrow(_obj, ev);";
	    out << "\nassert(!CORBA::is_nil(" << flatIntfName << "));";
            if(desc -> result -> kind() != tk_void)
            {
                String_var str = getTypeString("",
					       desc -> result, GetTypeReturn,
                                               false, "_result");
                out << '\n' << str << ';';
            }
            for(j = 0 ; j < desc -> parameters.length() ; j++)
            {
                String_var paramName = fixKwd(desc -> parameters[j].name);
                TypeCode_ptr origType =
                    GetOrigType_nodup(desc -> parameters[j].type);
		String_var str = getTypeString("",
					       desc -> parameters[j].type,
					       GetTypeNormal);
                switch(desc -> parameters[j].mode)
                {
                case PARAM_IN:
		{
		    if (origType -> kind() == tk_array)
		    {
			str += "_var";
		    }

		    // Fall through
		}
                case PARAM_INOUT:
                {
		    if(origType -> kind() == tk_objref ||
		       origType -> kind() == tk_TypeCode)
		    {
			str += "_var";
		    }
                    out << '\n' << str << ' ' << paramName << ';';
                    break;
                }
                case PARAM_OUT:
                {
                    if((origType -> kind() == tk_struct ||
                        origType -> kind() == tk_union ||
                        origType -> kind() == tk_array) &&
                        IdlIsVariable(origType))
                    {
                        str += "_var";
                    }
                    else if(origType -> kind() == tk_any ||
                            origType -> kind() == tk_sequence ||
			    origType -> kind() == tk_objref ||
			    origType -> kind() == tk_TypeCode)
                    {
                        str += "_var";
                    }
                    out << '\n' << str << ' ' << paramName << ';';
                    break;
                }
                }
            }

            out << '\n';
            if(desc -> result -> kind() != tk_void)
                out << "_result = ";

            out << flatIntfName << "->" << desc -> name << "(";
            for(j = 0 ; j < desc -> parameters.length() ; j++)
            {
                String_var paramName = fixKwd(desc -> parameters[j].name);
                TypeCode_ptr origType =
                    GetOrigType_nodup(desc -> parameters[j].type);

		if (j != 0)
		    out << ", ";
                out << paramName;
                switch(desc -> parameters[j].mode)
                {
                case PARAM_IN:
                case PARAM_OUT:
                    break;

                case PARAM_INOUT:
                {
                    switch(origType -> kind())
                    {
                    case tk_null:
                    case tk_void:
                        assert(false);
                        break;

                    case tk_short:
                    case tk_long:
                    case tk_ushort:
                    case tk_ulong:
                    case tk_float:
                    case tk_double:
                    case tk_boolean:
                    case tk_char:
                    case tk_octet:
                    case tk_longlong:
                    case tk_ulonglong:
                    case tk_longdouble:
                    case tk_wchar:
                    case tk_enum:
                    case tk_string:
                    case tk_wstring:
                    case tk_any:
                    case tk_sequence:
                    case tk_array:
                    case tk_struct:
                    case tk_union:
                        break;

                    case tk_TypeCode:
                    case tk_objref:
                    case tk_local_interface:
			out << ".inout()";
                        break;

                    case tk_except:
                    case tk_Principal:
                    case tk_fixed:
                    case tk_native:
                    case tk_value:
                    case tk_value_box:
                    case tk_abstract_interface:
                    case tk_alias:
                        assert(false);
                        break;
                    }
                    break;
                }
                }
            }
	    if(desc -> parameters.length() > 0)
		out << ", ";
            out << "ev);";
	    out << "\nif (ev.exception() == 0)";
	    out.start();
            out << "\n// Success";
	    out.end();

            if(desc -> exceptions.length() > 0)
            {
                for(j = 0 ; j < desc -> exceptions.length() ; j++)
                {
                    String_var exId = getAbsolute(desc -> exceptions[j].id);
		    String_var flatExId = getFlattenedName(exId);
		    out << '\n' << exId << "* " << flatExId << " = "
		        << exId << "::_downcast(ev.exception());";
		    out << "\nif (" << flatExId << " != 0)";
		    out.start();
		    out << "\n// Handle " << exId;
		    out.end();
		}
            }

            out.end();

            break;
        }

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code to call the caller functions for the sample client
// ----------------------------------------------------------------------

void
IdlECPPGenerator::genECPPImplClientCallCallers(Container_ptr container,
					       IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
        //
        // Get element description
        //
        Container::Description& seqElem = (*seq)[i];

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seqElem.contained_object);

        switch(seqElem.kind)
        {
        case dk_Module:
        {
            //
            // Module
            //
            const ModuleDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            assert(!is_nil(cont));
            genECPPImplClientCallCallers(cont, out);

            break;
        }

        case dk_Interface:
        {
            //
            // Interface
            //
            const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            String_var absolute = getAbsolute(desc -> id);
	    String_var flatIntfName = getFlattenedName(absolute);

            out << "\n// Calling " << absolute;
            out.start();
            out << "\nCORBA::Object_var obj = orb->string_to_object("
	        << "\"relfile:/" << flatIntfName << ".ref\", ev);";
	    out << "\nif (ev.exception() != 0)";
	    out.start();
	    out << "\nreturn EXIT_FAILURE;";
	    out.end();
            out << "\nif (CORBA::is_nil(obj))";
            out.start();
            out << "\nfprintf(stderr, \"%s: cannot read IOR from "
                << flatIntfName << ".ref\\n\", argv[0]);";
            out << "\nreturn EXIT_FAILURE;";
            out.end();
            out.sep();
	    out << '\n' << absolute << "_var " << flatIntfName << " = "
	        << absolute << "::_narrow(obj, ev);";
	    out << "\nassert(!CORBA::is_nil(" << flatIntfName << "));";
	    out.sep();

            //
            // Operations & attributes
            //
	    ScopedName_var scope;
            // No _var for seq! See comments for the implementation of
            // "describe"
            Container::DescriptionSeq* intfSeq = describe(cont, scope.out());
            for(j = 0 ; j < intfSeq -> length() ; j++)
            {
                if((*intfSeq)[j].kind == dk_Attribute)
                {
                    const AttributeDescription* desc;
                    Boolean b = (*intfSeq)[j].value >>= desc;
                    assert(b);
                    out << "\ncall_" << flatIntfName << "__get_"
			<< desc -> name << "(obj, orb, ev);";
		    out << "\nif (ev.exception() != 0)";
		    out.start();
                    out << "\nreturn EXIT_FAILURE;";
		    out.end();
                    if(desc -> mode == ATTR_NORMAL)
                    {
                        out.sep();
                        out << "\ncall_" << flatIntfName << "__set_"
                            << desc -> name << "(obj, orb, ev);";
			out << "\nif (ev.exception() != 0)";
			out.start();
                        out << "\nreturn EXIT_FAILURE;";
			out.end();
                    }
                    out.sep();
                }
                else if((*intfSeq)[j].kind == dk_Operation)
                {
                    const OperationDescription* desc;
                    Boolean b = (*intfSeq)[j].value >>= desc;
                    assert(b);
                    out << "\ncall_" << flatIntfName << '_' << desc -> name
                        << "(obj, orb, ev);";
		    out << "\nif (ev.exception() != 0)";
		    out.start();
                    out << "\nreturn EXIT_FAILURE;";
		    out.end();
                    out.sep();
                }
            }

            out.end();
            out.sep();

            break;
        }

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the sample server
// ----------------------------------------------------------------------

void
IdlECPPGenerator::genECPPImplServer(
    Container_ptr container,
    IdlPrettyPrint& out
)
{
    StringSeq scopedInterfaces;
    StringSeq interfaces;
    ULong i;

    collectInterfaces(container, scopedInterfaces, interfaces);

    //
    // run
    //
    out.sep();
    out << "\nstatic int"
        << "\nrun(";
    out.inc();
    out << "\nCORBA::ORB_ptr orb,";
    out << "\nint argc,";
    out << "\nchar** argv,";
    out << "\nCORBA::Environment& ev";
    out.dec();
    out << "\n)";
    out.start();
    out << "\nCORBA::Object_var obj = orb->resolve_initial_references("
	<< "\"RootPOA\", ev);";
    out << "\nif (ev.exception() != 0)";
    out.start();
    out << "\nreturn EXIT_FAILURE;";
    out.end();
    out << "\nPortableServer::POA_var rootPOA = PortableServer::POA::_narrow("
        << "obj, ev);";
    out << "\nassert(ev.exception() == 0);";
    out << "\nPortableServer::POAManager_var manager = "
	<< "rootPOA->the_POAManager(ev);\n";
    out << "\nassert(ev.exception() == 0);";
    out << "\nFILE* f = 0;";
    out << "\nCORBA::String_var s;";
    out.sep();
    for(i = 0 ; i < scopedInterfaces.length() ; i++)
    {
	CORBA::String_var flatIntfName = getFlattenedName(scopedInterfaces[i]);
        out.sep();
        out << '\n' << flatIntfName << "_impl* " << flatIntfName
            << "_servant = new " << flatIntfName << "_impl(rootPOA);";
	out << "\nPortableServer::ServantBase_var " << flatIntfName
	    << "servant = ";
	if(strcmp(interfaces[i], "") != 0)
	    out << "(POA_" << scopedInterfaces[i] << "*)";
	out << flatIntfName << "_servant;";
	out << '\n' << scopedInterfaces[i] << "_var " << flatIntfName
	    << " = " << flatIntfName << "_servant->";
	if (strcmp(interfaces[i], "") != 0)
	{
	    if (strcmp(interfaces[i], scopedInterfaces[i]) == 0)
		out << "POA_";
	    out << interfaces[i] << "::";
	}
	out << "_this(ev);";
	out << "\nassert(ev.exception() == 0);";
	out.sep();
	out << "\ns = orb->object_to_string("
	    << flatIntfName << ", ev);";
	out << "\nassert(ev.exception() == 0);";
	out.sep();
	out << "\nf = fopen(\"" << flatIntfName << ".ref\", \"w\");";
	out << "\nif (f == 0)";
	out.start();
        out << "\nfprintf(stderr, \"%s: can't open " << flatIntfName
            << ".ref: %s\\n\", argv[0], strerror(errno));";
        out.end();
	out << "\nfputs(s.in(), f);";
        out << "\nfclose(f);";
    }
    out.sep();
    out << "\nmanager->activate(ev);";
    out << "\nif (ev.exception() == 0)";
    out.start();
    out << "\norb->run(ev);";
    out.end();
    out.sep();
    out << "\nif (ev.exception() == 0)";
    out.start();
    out << "\nreturn EXIT_SUCCESS;";
    out.end();
    out << "\nelse";
    out.start();
    out << "\nreturn EXIT_FAILURE;";
    out.end();
    out.end();

    //
    // main
    //
    out.sep();
    out << "\nint"
        << "\nmain(";
    out.inc();
    out << "\nint argc,";
    out << "\nchar** argv";
    out.dec();
    out << "\n)";
    out.start();
    out << "\nint status = EXIT_SUCCESS;";
    out << "\nCORBA::Environment ev;";
    out << "\nCORBA::ORB_var orb;";

    out.sep();
    out << "\norb = CORBA::ORB_init(argc, argv, \"\", ev);";
    out << "\nif (ev.exception() == 0)";
    out.start();
    out << "\nstatus = run(orb, argc, argv, ev);";
    out.end();

    out<< "\nif (ev.exception() != 0)";
    out.start();
    out << "CORBA::Exception* ex = ev.exception();";
    out << "\nfprintf(stderr, \"%s\\n\", ex->_rep_id());";
    out << "\nstatus = EXIT_FAILURE;";
    out.end();

    out.sep();
    out << "\nif (!CORBA::is_nil(orb))";
    out.start();
    out << "\nev.clear();";
    out << "\norb->destroy(ev);";
    out << "\nif (ev.exception() != 0)";
    out.start();
    out << "CORBA::Exception* ex = ev.exception();";
    out << "\nfprintf(stderr, \"%s\\n\", ex->_rep_id());";
    out << "\nstatus = EXIT_FAILURE;";
    out.end();
    out.end();

    out.sep();
    out << "\nreturn status;";

    out.end();
}

// ----------------------------------------------------------------------
// Collect the absolute names of all interfaces
// ----------------------------------------------------------------------

void
IdlECPPGenerator::collectInterfaces(
    Container_ptr container,
    StringSeq& scopedNameSeq,
    StringSeq& nameSeq
)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
        //
        // Get element description
        //
        Container::Description& seqElem = (*seq)[i];

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seqElem.contained_object);

        switch(seqElem.kind)
        {
        case dk_Module:
        {
            //
            // Module
            //
            const ModuleDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            assert(!is_nil(cont));
            collectInterfaces(cont, scopedNameSeq, nameSeq);

            break;
        }

        case dk_Interface:
        {
            //
            // Interface
            //
            const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            scopedNameSeq.length(scopedNameSeq.length() + 1);
            scopedNameSeq[scopedNameSeq.length() - 1] =
		getAbsolute(desc -> id);

	    InterfaceDef_var intf = InterfaceDef::_narrow(cont);
	    IdlInterfaceInfoSeq baseInfo;
	    IdlGetBaseInterfaces(intf, baseInfo, false);

	    nameSeq.length(nameSeq.length() + 1);
	    if (baseInfo.length() > 0)
	    {
		nameSeq[nameSeq.length() - 1] =
		    string_dup(scopedNameSeq[scopedNameSeq.length() - 1]);
		CORBA::String_var scope = getScope(cont);
		IdlRemoveScope(scope, nameSeq[nameSeq.length() - 1].inout());
	    }
	    else
	    {
		//
		// An empty non-scoped name signifies there is only a
		// single base class (the generated POA class) and there
		// is no need for casting to avoid ambiguities.
		//
		nameSeq[nameSeq.length() - 1] = string_dup("");
	    }
            break;
        }

        default:
            break;
        }
    }
}
