// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/list.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STRINGS_H
#   include <strings.h>
#endif

#include <ctype.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenTextBase.h>
#include <GenHTML.h>

using namespace CORBA;
using namespace OB;

//
// The top-level file name (without ".html")
//
const char* IdlHTMLGenerator::top_ = "_top_";

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the start of a code block.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::startBlock(IdlPrettyPrint& out)
{
    out.direct("\n<pre>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the end of a code block.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::endBlock(IdlPrettyPrint& out)
{
    out.direct("\n</pre>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the start of the @ (extra) clauses of a comment.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::startExtra(IdlPrettyPrint& out)
{
    out.lineBreak();
    out.direct("\n<p />\n<dl>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the end of the @ (extra) clauses of a comment.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::endExtra(IdlPrettyPrint& out)
{
    out.direct("</dl>\n");
    out.lineBreak();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the start of an index.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::startIndex(IdlPrettyPrint& out)
{
    out.lineBreak();
    if(useTables_)
        out.direct("\n<table noborders>");
    else
        out.direct("\n<dl>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the end of an index.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::endIndex(IdlPrettyPrint& out)
{
    if(useTables_)
        out.direct("</table>\n");
    else
        out.direct("</dl>\n");
    out.lineBreak();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs the header of an object declaration.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::startObject(const char* name, IdlPrettyPrint& out,
                              bool first)
{
    if(!first)
        out.direct("\n<p />");
    out.lineBreak();
    out.direct("\n<dl>\n<dt>");
    textBold(true, out);
    putLinkTag(name, name, out);
    textBold(false, out);
    out.direct("</dt><dd>");
    out.lineBreak();
    startBlock(out);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method completes an object declaration.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::endObject(const char* id, IdlPrettyPrint& out,
                            Container_ptr container, const char* name)
{
    endBlock(out);
    out.direct("\n<p />");
    putComment(id, out, container, name);
    out.direct("</dd></dl>");
    out.lineBreak();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove bold format from text.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::textBold(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        out.direct("<b>");
    else
        out.direct("</b>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove italic format from text.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::textItalic(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        out.direct("<i>");
    else
        out.direct("</i>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove underline format from text.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::textUnderline(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        out.direct("<u>");
    else
        out.direct("</u>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove literal format from text.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::textLiteral(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        out.direct("<code>");
    else
        out.direct("</code>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove large format from text.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::textLarge(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        out.direct("<big>");
    else
        out.direct("</big>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove small format from text.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::textSmall(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        out.direct("<small>");
    else
        out.direct("</small>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Generate a hyperlink reference string for an object.
// ----------------------------------------------------------------------

char*
IdlHTMLGenerator::getLinkRef(const char* id)
{
    for(ULong i = 0 ; i < linkSeq_.length() ; i += 2)
    {
        if(strcmp(linkSeq_[i], id) == 0)
            return string_dup(linkSeq_[i + 1]);
    }

    String_var absolute = getAbsolute(id);
    absolute = fixAbsolute(absolute);

    String_var link;

    Contained_var contained = repository_ -> lookup_id(id);
    assert(!is_nil(contained));

    DefinitionKind defKind = contained -> def_kind();
    if(defKind == dk_Module || defKind == dk_Interface ||
       defKind == dk_AbstractInterface || defKind == dk_LocalInterface ||
       defKind == dk_Value)
    {
        link = absolute._retn();
        link += ".html";
    }
    else
    {
        char* name = strrchr(absolute.inout(), '.');
        if(name)
        {
            name++;
            assert(*name != '\0');
            *(name - 1) = '\0';
            link = string_dup(absolute);
            link += ".html#";
            link += name;
        }
        else
        {
            link = string_dup(top_);
            link += ".html#";
            link += absolute;
        }
    }

    ULong len = linkSeq_.length();
    linkSeq_.length(len + 2);
    linkSeq_[len] = id;
    linkSeq_[len + 1] = link;
    return link._retn();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output a hyperlink reference string.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putLinkRef(const char* link, const char* text,
                             IdlPrettyPrint& out)
{
    out.lineStart();
    if(link && *link)
    {
        out.direct("<a href=\"");
        out.direct(link);
        out.direct("\">");
        out << text;
        out.direct("</a>");
    }
    else
        out << text;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output a hyperlink reference string.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putLinkTag(const char* link, const char* text,
                             IdlPrettyPrint& out)
{
    out.lineStart();
    if(link && *link)
    {
        out.direct("<a name=\"");
        out.direct(link);
        out.direct("\">");
        out << text;
        out.direct("</a>");
    }
    else
        out << text;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output a horizontal rule (line).
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putRule(IdlPrettyPrint& out)
{
    out.direct("\n<hr />");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs the title of the file.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putTitle(const char* label, const char* name,
                           IdlPrettyPrint& out)
{
    out.setIndent(0);
    out.direct("\n<title>");
    if(label)
        out << label;
    if(name)
        out << name;
    out.direct("</title>");
    out.lineBreak();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a module declaration (level 1 heading).
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putLevel1(const char* label, const char* name,
                            IdlPrettyPrint& out)
{
    out.setIndent(0);
    out.direct("\n<h1>");
    if(label)
        out << label;
    if(name)
        out << name;
    out.direct("</h1>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a group association (level 2 heading).
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putLevel2(const char* label, const char* name,
                            IdlPrettyPrint& out)
{
    out.setIndent(0);
    out.direct("\n<h2>");
    if(label)
        out << label;
    if(name)
        out << name;
    out.direct("</h2>");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs an index entry.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putEntry(const char* label, const char* text,
                           IdlPrettyPrint& out, bool /* first */,
                           const char* link)
{
    if(useTables_)
    {
        out.direct("<tr><td>");
        textBold(true, out);
        putLinkRef(link, label, out);
        textBold(false, out);
        out.direct("</td><td>");
        out.lineBreak();
        putParagraph(text, out);
        out.direct("</td></tr>");
    }
    else
    {
        out.direct("<dt>");
        textBold(true, out);
        putLinkRef(link, label, out);
        textBold(false, out);
        out.direct("</dt><dd>");
        putParagraph(text, out);
        out.direct("</dd>");
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a single @ clause from a comment.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putClause(const char* label, const char* link,
                            const char* name, const char* text,
                            IdlPrettyPrint& out, bool /* first */, bool embed)
{
    if(label && *label)
    {
        out.lineBreak();
        out.direct("<dt>");
        textBold(true, out);
        out << label;
        textBold(false, out);
        out.direct("</dt>");
    }

    if(name && *name || text && *text)
    {
        out.lineBreak();
        if(embed)
            out.direct("<dd>");
        if(name && *name)
        {
            textLiteral(true, out);
            putLinkRef(link, name, out);
            textLiteral(false, out);
            if(text && *text)
                out << " - ";
        }
        if(text && *text)
            putParagraph(text, out);
        if(embed)
            out.direct("</dd>");
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a paragraph of (formatted) text.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putParagraph(const char* text, IdlPrettyPrint& out)
{
    if(text)
    {
        String_var str = string_dup(text);
        char* hdr = str.inout();
        char* ptr;
        char val;

        //
        // Strip leading spaces
        //
        while(*hdr && isspace(*hdr))
            hdr++;
        ptr = hdr + strlen(hdr) - 1;

        //
        // Strip trailing spaces
        //
        while(ptr > hdr && isspace(*ptr))
            ptr--;
        ptr[1] = '\0';

        //
        // Print preserving editing tags. This approach
        // preserves column info maintained by pos_. For
        // some reason, strtok() doesn't work right with
        // adjacent tags (second '<' gets lost).
        //
        while(*hdr)
        {
            ptr = hdr;
            while((val = *ptr) && val != '<')
                ptr++;
            *ptr = '\0';
            out << hdr;
            if(val)
            {
                hdr = ++ptr;
                while((val = *ptr) && val != '>')
                    ptr++;
                *ptr = '\0';
                if(val)
                {
                    out.direct('<');
                    out.direct(hdr);
                    out.direct('>');
                    ptr++;
                }
                else
                    out << '<' << hdr;
            }
            hdr = ptr;
        }
        out << '\n';
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Write the introduction (first sentence) of the comment.
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::putIntro(const char* id, const char* name,
                           IdlPrettyPrint& out, bool first, bool factory)
{
    const char* ptr = 0;
    String_var link = getLinkRef(id);
    String_var comment;

    bool getResult = idHash_.get(id, comment);
    if(getResult)
    {
        if(factory)
        {
            StringSeq factorySeq;

            const char* tok = tokenize(comment.inout());

            while(tok)
            {
                if(strncmp(tok, keyFactory_, strlen(keyFactory_)) == 0)
                {
                    factorySeq.length(factorySeq.length() + 1);
                    factorySeq[factorySeq.length() - 1] =
                        tok + strlen(keyFactory_);
                }
                tok = tokenize(0);
            }

            ULong i;
            for(i = 0 ; i < factorySeq.length() ; i++)
            {
                String_var s = factorySeq[i];
                const char* limit = s.inout() + strlen(s.inout());
                const char* token = strtok(s.inout(), " \t\r\n");
                const char* dtext;

                if(token && !strcmp(token, name))
                {
                    dtext = token + strlen(token) + 1;
                    while(dtext < limit && isspace(*dtext))
                        dtext++;
                    comment = string_dup(dtext);
                    link = string_dup(link);
                    link += "#";
                    link += name;
                    break;
                }
            }

            if(i == factorySeq.length())
                comment = string_dup("");
        }

        if(strlen(comment))
        {
            const char* end1;
            ptr = comment.inout();
            const char* start = ptr;

            while(true)
            {
                end1 = strchr(start, '.');

                if(end1 == 0)
                    break;

                end1++;
                if(*end1 == ' ' || *end1 == '\t' || *end1 == '\n' ||
                   *end1 == '\r')
                    break;

                start = end1;
            }

            const char* end2 = strchr(comment.inout(), '@');

            const char* end;
            if(end1 && end2)
                end = end1 < end2 ? end1 : end2;
            else if(end1)
                end = end1;
            else
                end = end2;

            if(end)
                *((char*)end) = '\0';
        }
    }

    putEntry(name, ptr, out, first, link);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of modules.
// ----------------------------------------------------------------------

int
IdlHTMLGenerator::formatModule(Container_ptr container,
                               const char* /* scope */,
                               const Container::DescriptionSeq& seq,
                               IdlPrettyPrint& /* out */)
{
    int status = 0;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const ModuleDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        //
        // HTML file for the module
        //
        ScopedName_var absolute = getAbsolute(desc -> id);
        String_var path = fixAbsolute(absolute);
        PrettyPrint out(prog_, path, outDir_);
        if(!out.good())
            return 1;

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);
        assert(!is_nil(cont));

        //
        // Generate code
        //
        putTitle("Module ", absolute, out);
        putLevel1("Module ", absolute, out);

        putComment(desc -> id, out, container);

        status = genScope(cont, out);
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of interfaces.
// ----------------------------------------------------------------------

int
IdlHTMLGenerator::formatInterface(Container_ptr container,
                                  const char* /* scope */,
                                  const Container::DescriptionSeq& seq,
                                  IdlPrettyPrint& /* out */)
{
    int status = 0;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const InterfaceDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        //
        // New file for the interface
        //
        ScopedName_var absolute = getAbsolute(desc -> id);
        String_var path = fixAbsolute(absolute);
        PrettyPrint out(prog_, path, outDir_);
        if(!out.good())
            return 1;

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);
        assert(!is_nil(cont));

        //
        // Generate code
        //
        if(seq[i].kind == dk_AbstractInterface)
        {
            putTitle("Abstract Interface ", absolute, out);
            putLevel1("Abstract Interface ", absolute, out);
            putInterface(container, desc, out,  seq[i].kind);
        }
        else if(seq[i].kind == dk_LocalInterface)
        {
            putTitle("Local Interface ", absolute, out);
            putLevel1("Local Interface ", absolute, out);
            putInterface(container, desc, out,  seq[i].kind);
        }
        else
        {
            putTitle("Interface ", absolute, out);
            putLevel1("Interface ", absolute, out);
            putInterface(container, desc, out, seq[i].kind);
        }

        status = genScope(cont, out);
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of valuetypes.
// ----------------------------------------------------------------------

int
IdlHTMLGenerator::formatValuetype(Container_ptr container,
                                  const char* /* scope */,
                                  const Container::DescriptionSeq& seq,
                                  IdlPrettyPrint& /* out */)
{
    int status = 0;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const ValueDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        //
        // HTML file for the valuetype
        //
        ScopedName_var absolute = getAbsolute(desc -> id);
        String_var path = fixAbsolute(absolute);
        PrettyPrint out(prog_, path, outDir_);
        if(!out.good())
            return 1;

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);
        assert(!is_nil(cont));

        ValueDef_var value = ValueDef::_narrow(cont);
        assert(!is_nil(value));

        //
        // Generate code
        //
        putTitle("Valuetype ", absolute, out);
        putLevel1("Valuetype ", absolute, out);
        putValuetype(container, desc, out);

        InitializerSeq_var initializers = value -> initializers();
        if(initializers -> length() > 0)
        {
            String_var valueScope;
            getScope(value, valueScope.out());

            genScope(cont, out, GEN_INDICES);
            putInitializers(container, initializers, out, desc -> id,
                            valueScope, GEN_INDICES);

            genScope(cont, out, GEN_CONTENT);
            putInitializers(container, initializers, out, desc -> id,
                            valueScope, GEN_CONTENT);
        }
        else
            status = genScope(cont, out);
    }

    return status;
}

// ----------------------------------------------------------------------
// IDL code generator constructor and destructor
// ----------------------------------------------------------------------

IdlHTMLGenerator::IdlHTMLGenerator(const char* prog,
                                   Repository_ptr rep,
                                   bool all,
                                   const IdlStringHash& idHash,
                                   const char* outDir,
                                   bool noSort,
                                   bool ignoreCase,
                                   bool useTables,
				   bool altIndent)
    : GenTextBase(rep, all, idHash, noSort, ignoreCase, useTables, altIndent),
      prog_(prog),
      outDir_(outDir)
{
}

// ----------------------------------------------------------------------
// Generate HTML code
// ----------------------------------------------------------------------

int
IdlHTMLGenerator::gen()
{
    PrettyPrint out(prog_, top_, outDir_);

    putTitle("Top-Level Module", 0, out);
    putLevel1("Top-Level Module", 0, out);
    return genScope(repository_, out);
}
