// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <Util.h>
#include <GenUtil.h>
#include <GenTextBase.h>
#include <GenRTFUtil.h>

#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

using namespace CORBA;
using namespace OB;

//
// HTML character replacements, where the "&" is used to
// delimit a special character otherwise reserved to HTML
// or denoting a special insertion.
//
struct HTMLCharacters
{
    const char* name;
    const char* repl;
};

static HTMLCharacters replace[] =
{
    {"&gt;",  ">"},
    {"&lt;",  "<"},
    {"&amp;", "&"},
    {NULL,    NULL}
};

// ----------------------------------------------------------------------
// RTFFontDescriptor public member implementation
// ----------------------------------------------------------------------

RTFFontDescriptor::RTFFontDescriptor(unsigned int idno, const char* tag)
    : tag_(tag), font_(""), idno_(idno), size_(0)
{
}

RTFFontDescriptor::RTFFontDescriptor(RTFFontDescriptor* p)
    : tag_(p -> tag_), font_(p -> font_), idno_(p -> idno_), size_(p -> size_)
{
}

void
RTFFontDescriptor::emit(IdlPrettyPrint& out, bool style) const
{
    String_var text = string_dup("");
    if(style)
    {
        text += "\\s";
        text += idno_;
    }
    text += "\\f";
    text += idno_;
    text += "\\fs";
    text += size_ * FONTPPNT;
    text += ' ';
    out.direct(text);
}

void
RTFFontDescriptor::define(IdlPrettyPrint& out) const
{
    String_var text = string_dup("\n{\\f");
    text += idno_;
    text += "\\fcharset0\\f";
    text += font_;
    text += ";}";
    out.direct(text);
}

// ----------------------------------------------------------------------
// RTFDisplayList public member implementation
// ----------------------------------------------------------------------

RTFDisplayList::RTFDisplayList(ListType type)
    : lstype_(type), indent_(0), number_(0), output_(false)
{
}

//
// Output list characteristics. It would be nice to use the real list
// properties, but Word 97 & Word 2000 really changed the way this worked.
// The usual result can't be interpreted, and it would require another pass
// per file to build the new list table structures.
//
// Note: can't use local indent, which is the indent which must be restored
// after popping the list.
//
void
RTFDisplayList::emit(unsigned int indent, unsigned int adjust,
                     IdlPrettyPrint& out)
{
    if(!output_)
    {
        switch(lstype_)
        {
        case Bullet:
        {
            String_var move = string_dup("\\fi-360\\tx");
            move += WDINDENT * indent + adjust;
            move += ("\\f6\\'b7");
            out.direct(move);
            out.direct("\\f0\\tab  ");
            break;
        }

        case Number:
        {
            String_var move = string_dup("\\fi-360\\tx");
            move += WDINDENT * indent + adjust;
            move += ' ';
            move += number_;
            move += ".\\tab  ";
            out.direct(move);
            break;
        }

        default:
            break;
        }
        output_ = true;
    }
}

// ----------------------------------------------------------------------
// RTFParagraph public member implementation
// ----------------------------------------------------------------------

RTFParagraph::RTFParagraph(unsigned int sn, RTFFontDescriptor* fd)
    : margin_(0), stylen_(sn), fntdsc_(0), border_(false)
{
    if(fd)
        fntdsc_ = new RTFFontDescriptor(fd);
}

RTFParagraph::RTFParagraph(const RTFParagraph* p)
    : margin_(p -> margin_), stylen_(p -> stylen_), fntdsc_(0),
      border_(p -> border_)
{
    if(p -> fntdsc_)
        fntdsc_ = new RTFFontDescriptor(p -> fntdsc_);
}

RTFParagraph::~RTFParagraph()
{
    if(fntdsc_)
        delete fntdsc_;
}

//
// Output paragraph and cell descriptions
//
void
RTFParagraph::emit(unsigned int blanks, unsigned int indent,
                   unsigned int margin, RTFDisplayList* dlistp,
                   IdlPrettyPrint& out)
{
    String_var text = string_dup("");
    unsigned int adjust = 0;

    out.direct("\n\\pard\\plain\\widctlpar");

    //
    // Output whether a cell in a table
    //
    if(inCell())
    {
        out.direct("\\intbl");
        adjust = 108;
    }

    //
    // Output blank space before line
    //
    if(blanks)
    {
        text += "\\sb";
        text += TWIPPPNT * FULLLINE * blanks;
    }

    //
    // Output special paragraph formatting
    //
    switch(stylen_)
    {
    case FN_HEADING:
        out.direct("\\b\\keep\\keepn\\brdrb\\brdrdb\\brdrw10\\brsp20\\pagebb");
        break;
    case FN_INDEX:
        out.direct("\\b\\keep\\keepn\\brdrb\\brdrs\\brdrw10\\brsp20\\brdrcf2");
        break;
    case FN_ENTRY:
    case FN_EXTRA:
        out.direct("\\b");
        if(!inCell())
            out.direct("\\keep\\keepn");
    default:
        if(border_)
            out.direct("\\brdrb\\brdrs\\brdrw30\\brsp20");
    }

    //
    // Output character formatting
    //
    fntdsc_ -> emit(out);

    //
    // Output indentation and margin, with cell bias adjustment
    //
    if(indent)
    {
        text += "\\li";
        if(margin && margin < indent)
        {
            text += WDINDENT * margin + adjust;
            text += ' ';
            for(unsigned int i = margin ; i < indent ; i++ )
                text += ' ';
        }
        else
        {
            text += WDINDENT * indent + adjust;
            text += ' ';
        }
    }
    else
        text += ' ';
    out.direct(text);

    //
    // If in a list, output list preface text
    //
    if(dlistp)
        dlistp -> emit(indent, adjust, out);
}

// ----------------------------------------------------------------------
// RTFTable public member implementation
// ----------------------------------------------------------------------

RTFTable::RTFTable(unsigned int ncells, bool border)
    : pstack_(0), lstndx_(0), margin_(0), ncells_(ncells), rcells_(0),
      cindex_(0), config_(false), border_(border), output_(false),
      incell_(false)
{
    if(ncells_ == 0)
        ncells_ = 1;
    for(unsigned int i = 0 ; i < MAXCELLS ; i++)
        redges_[i] = widths_[i] = 0;
}

bool
RTFTable::calcEdges()
{
    if(!config_)
    {
        unsigned int i;
        if(widths_[0] == 0)
            widths_[0] = CLRIGHT1;
        widths_[0] += margin_;
        for(i = 1 ; i < ncells_ - 1 ; i++)
        {
            if(widths_[i] == 0)
                widths_[i] = CLRIGHT2;
            widths_[i] += widths_[i - 1];
        }
        widths_[ncells_ - 1] = PGRIGHTX;
        config_ = true;

        for(i = 0 ; i < ncells_ ; i++)
            if(redges_[i] != widths_[i])
                return true;
    }
    return false;
}

void
RTFTable::setStack(RTFParagraph* psp, RTFDisplayList** dlp, unsigned int lnx)
{
    pstack_ = psp;
    lstndx_ = lnx;
    for(unsigned int i = 0 ; i < lstndx_ ; i++)
        lstack_[i] = dlp[i];
}

void
RTFTable::getStack(RTFParagraph*& psp, RTFDisplayList** dlp,
                                 unsigned int& lnx)
{
    for(unsigned int i = 0 ; i < lstndx_ ; i++)
        dlp[i] = lstack_[i];
    lnx = lstndx_;
    psp = pstack_;
}

void
RTFTable::newRow(unsigned int rcells, IdlPrettyPrint& out)
{
    //
    // Start a new row, flushing previous row and emitting table definition
    //
    if(!output_)
    {
        unsigned int oldnum = rcells_;
        rcells_ = (rcells < ncells_) ? rcells : ncells_;
        if(calcEdges() || oldnum != rcells_)
        {
            String_var text = string_dup("\n\\trowd\\trgaph108");
            if(margin_)
            {
                text += "\\trleft";
                text += WDINDENT * margin_;
                text += ' ';
            }
            out.direct(text);
            for(unsigned int i = 1 ; i <= ncells_ ; i++)
            {
                text = string_dup("");
                if(i == rcells_  &&  rcells_ < ncells_)
                    text += "\\clmgf";
                else if(i > rcells_)
                    text += "\\clmrg";
                if (border_)
                    text += "\\clbrdrt\\brdrth\\clbrdrl\\brdrth\\clbrdrr"
                            "\\brdrth\\clbrdrb\\brdrth";
                text += "\\cellx";
                text += WDINDENT * (redges_[i - 1] = widths_[i - 1]);
                text += ' ';
                out.direct(text);
            }
        }
        output_ = true;
        cindex_ = 0;
    }
}

void
RTFTable::endRow(IdlPrettyPrint& out)
{
    //
    // Finish the current row, adding unspecified (merged) cells
    //
    if(output_)
    {
        endCell(out);
        for ( ; cindex_ < ncells_ ; cindex_++)
            out.direct("\\cell ");
        out.direct("\\row ");
        output_ = false;
    }
    else
        cindex_ = ncells_;
}

void
RTFTable::newCell(IdlPrettyPrint& out)
{
    //
    // Start a new cell; paragraph start will be forced.
    //
    assert(cindex_ < rcells_);
    RTFPrettyPrint& prt = (RTFPrettyPrint&)out;
    endCell(prt);
    prt.setSpace(0);
    prt.getStyle() -> setMargin(redges_[cindex_++]);
    prt.lineStart();
    incell_ = true;
}

void
RTFTable::endCell(IdlPrettyPrint& out)
{
    //
    // Terminate the current cell; paragraph start has been forced
    //
    if(incell_)
    {
        out.direct("\\cell ");
        out.lineBreak();
        incell_ = false;
    }
}

// ----------------------------------------------------------------------
// PrettyPrint public member implementation
// ----------------------------------------------------------------------

RTFPrettyPrint::RTFPrettyPrint(const char* prog, const char* fullName,
                               const char* dir, RTFFontDescriptor** fonts,
                               unsigned int num)
    : IdlPrettyPrint(out_, num), style_(0), listp_(0), margin_(0), blanks_(0)
{
    //
    // Create complete file name path
    //
    path_ = string_dup(dir);
    if(strlen(path_) > 0)
        path_ += '/';
    path_ += fullName;
    path_ += ".rtf";

    //
    // Create the file
    //
    out_.open(path_);

    //
    // Write header
    //
    if(good())
    {
        Long i;
        out_ << "{";

        //
        // Header
        //
        out_ << "\\rtf1" // RTF version 1
             << "\\ansi" // ANSI character set
             << "\\deff0";

        //
        // Font definitions
        //
        out_ << "\n{\\fonttbl";
        for(i = 0 ; i < MAXFONTS ; i++)
            fonts[i] -> define(*this);
        out_ << "\n}";

        //
        // Color definitions
        //
        out_ << "\n{\\colortbl"
             << "\n\\red0\\green0\\blue0;"
             << "\n\\red128\\green128\\blue128;"
             << "\n}";

        //
        // Style definitions - converting points to RTF half-points
        //
        out_ << "\n{\\stylesheet";
        out_ << "\n{"; fonts[FN_BODY] -> emit(*this);
        out_ << "\\widctlpar\\snext0 Body;}";
        out_ << "\n{"; fonts[FN_ENTRY] -> emit(*this);
        out_ << "\\widctlpar\\snext0\\sbasedon0\\b Entry;}";
        out_ << "\n{"; fonts[FN_EXTRA] -> emit(*this);
        out_ << "\\widctlpar\\snext0\\sbasedon0\\b Extra;}";
        out_ << "\n{"; fonts[FN_INDEX] -> emit(*this);
        out_ << "\\widctlpar\\snext0\\sbasedon0\\keep\\keepn\\b"
             << "\\brdrb\\brdrs\\brdrw10\\brsp20\\brdrcf2 Group;}";
        out_ << "\n{"; fonts[FN_HEADING] -> emit(*this);
        out_ << "\\widctlpar\\snext0\\sbasedon0\\keep\\keepn\\b\\brdrb"
             << "\\brdrdb\\brdrw10\\brsp20\\pagebb Module;}";
        out_ << "\n{"; fonts[FN_LITERAL] -> emit(*this);
        out_ << "\\widctlpar\\snext5\\sbasedon0 Literal;}";
        out_ << "\n}";

        //
        // Document info
        //
        out_ << "\n{\\info"
             << "\n{\\title\nDocumentation for \"" << fullName << "\"\n}"
             << "\n{\\comment"
             << "\nGenerated by the ORBacus IDL-to-RTF translator"
             << "\nVersion: " << Version
             << "\n}";

        out_ << "\n}"
             << "\n"
             << "\\margl1440\\margr1440\\widowctrl\\sectd ";
    }
    else
    {
        OB_CERR << prog << ": can't open \"" << path_
                << "\": " << strerror(errno) << OB_ENDL;
    }
}

RTFPrettyPrint::~RTFPrettyPrint()
{
    out_ << "\n}\n";
}

void
RTFPrettyPrint::nl()
{
    if(emitIndent_)
        blanks_++;
    else
        out_ << "\\par ";
    lineBreak();
}

void
RTFPrettyPrint::indent()
{
    if(style_)
        style_ -> emit(blanks_, indent_, margin_, listp_, *this);
    else
        out_ << "\n\\pard";
    emitIndent_ = false;
    pos_ = indent_;
    blanks_ = 0;
}

void
RTFPrettyPrint::start()
{
    setMargin(getIndent());
    *this << "\n{";
    inc();
}

void
RTFPrettyPrint::end()
{
    dec();
    setMargin(0);
    *this << "\n}";
}

void
RTFPrettyPrint::printChar(char c)
{
    switch(c)
    {
    case '{':
    case '}':
    case '\\':
        lineStart();
        out_ << '\\';
    }
    IdlPrettyPrint::printChar(c);
}

void
RTFPrettyPrint::printWChar(wchar_t c)
{
    switch(c)
    {
    case '{':
    case '}':
    case '\\':
        lineStart();
        out_ << '\\';
    }
    IdlPrettyPrint::printWChar(c);
}

void
RTFPrettyPrint::printString(const char* s)
{
    if(!s)
        return;

    while(*s)
    {
        if(*s == '&')
        {
            bool found = false;
            HTMLCharacters* rp = replace;
            while(!found && rp -> name)
            {
                unsigned int len = strlen(rp -> name);
                if(!strncmp(s, rp -> name, len))
                {
                    const char* q = rp -> repl;
                    while(*q)
                    {
                        printChar(*q);
                        q++;
                    }
                    found = true;
                    s += len;
                }
            }
            if(found)
                continue;
        }
        printChar(*s);
        s++;
    }
}
