// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/list.h>

#include <GenUtil.h>

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Pretty-Print private and protected members
// ----------------------------------------------------------------------

void
IdlPrettyPrint::nl()
{
    out_ << '\n';
    lineBreak();
}

void
IdlPrettyPrint::indent()
{
    emitIndent_ = false;
    for(unsigned int i = 0 ; i < indent_ ; i++)
        out_ << ' ';
}

// ----------------------------------------------------------------------
// Pretty-Print public members
// ----------------------------------------------------------------------

void
IdlPrettyPrint::lineBreak()
{
    pos_ = indent_;
    emitIndent_ = true;
}

void
IdlPrettyPrint::lineStart()
{
    if(emitIndent_)
        indent();
}

void
IdlPrettyPrint::start()
{
    *this << "\n{";
    inc();
    ignoreNextSep_ = true;
}

void
IdlPrettyPrint::end()
{
    dec();
    *this << "\n}";
}

void
IdlPrettyPrint::sep()
{
    if(ignoreNextSep_)
        ignoreNextSep_ = false;
    else
        out_ << '\n';
}

void
IdlPrettyPrint::ignoreNextSep()
{
    ignoreNextSep_ = true;
}

void
IdlPrettyPrint::printChar(char c)
{
    if(c == '\r')
    {
        // Do nothing
    }
    else if(c == '\n')
    {
        nl();
    }
    else
    {
        if(emitIndent_)
            indent();
        if(c == '\t')
        {
            while(pos_ % 8)
            {
                out_ << ' ';
                pos_++;
            }
        }
        else
        {
            if(c != ' ')
                ignoreNextSep_ = false;
            out_ << c;
            pos_++;
        }
    }
}

void
IdlPrettyPrint::printWChar(wchar_t c)
{
    if(c == '\r')
    {
        // Do nothing
    }
    else if(c == '\n')
    {
        nl();
    }
    else
    {
        if(emitIndent_)
            indent();
        if(c == '\t')
        {
            while(pos_ % 8)
            {
                out_ << ' ';
                pos_++;
            }
        }
        else
        {
            out_ << c;
            pos_++;
        }
    }
}

void
IdlPrettyPrint::printString(const char* s)
{
    if(!s)
        return;

    while(*s)
    {
        printChar(*s);
        s++;
    }
}

void
IdlPrettyPrint::printWString(const wchar_t* s)
{
    if(!s)
        return;

    while(*s)
    {
        printWChar(*s);
        s++;
    }
}

// ----------------------------------------------------------------------
// Pretty-Print output operators
// ----------------------------------------------------------------------

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, char c)
{
    pp.printChar(c);
    return pp;
}

#ifndef HAVE_WIN32_WCHAR_TYPE
IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, wchar_t c)
{
    pp.printWChar(c);
    return pp;
}
#endif

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, const char* s)
{
    pp.printString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, const wchar_t* s)
{
    pp.printWString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, short val)
{
    return pp << (LongLong)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, int val)
{
    return pp << (LongLong)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, long val)
{
    return pp << (LongLong)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, LongLong val)
{
    String_var s = OB::LongLongToString(val);
    pp.printString(s.in());
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, unsigned short val)
{
    return pp << (ULongLong)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, unsigned int val)
{
    return pp << (ULongLong)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, unsigned long val)
{
    return pp << (ULongLong)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, ULongLong val)
{
    String_var s = OB::ULongLongToString(val);
    pp.printString(s.in());
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, float val)
{
    return pp << (double)val;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, double val)
{
    char s[1024];
    sprintf(s, "%.100g", val);
    pp.printString(s);
    return pp;
}

IdlPrettyPrint&
operator<<(IdlPrettyPrint& pp, LongDouble val)
{
#if SIZEOF_LONG_DOUBLE < 12
    assert(SIZEOF_LONG_DOUBLE != 0);
#else
    char s[1024];
    sprintf(s, "%.100Lg", val);
    pp.printString(s);
#endif
    return pp;
}

// ----------------------------------------------------------------------
// Remove scope from absolute name
// ----------------------------------------------------------------------

void
IdlRemoveScope(const char* scope, char*& absolute, bool fullyScoped)
{
    assert_nca(scope != 0, NCANullString);
    assert_nca(absolute != 0, NCANullString);

    ULong len = strlen(scope);
    if(len > 0)
    {
        assert(scope[len - 1] == ':');
        assert(scope[len - 2] == ':');

        if(strncmp(scope, absolute, len) == 0)
        {
            if(fullyScoped)
            {
                //
                // If there are no nested scopes, then we can use the
                // unscoped name, otherwise use a fully-scoped name
                //
                if(strchr(absolute + len, ':') == 0)
                {
                    char* a = string_dup(absolute + len);
                    string_free(absolute);
                    absolute = a;
                }
                else
                {
                    String_var tmp = string_dup("::");
                    tmp += absolute;
                    string_free(absolute);
                    absolute = tmp._retn();
                }
            }
            else
            {
                //
                // Remove the scope
                //
                char* a = string_dup(absolute + len);
                string_free(absolute);
                absolute = a;
            }
        }
        else if(strchr(scope, ':') != scope + len - 2)
        {
            //
            // Try again with a shortened scope, if possible
            //
            String_var s = scope;
            s[strlen(s) - 2] = '\0';
            char* s2 = (char*)strrchr(s, ':');
            if(s2)
            {
                assert(*(s2 - 1) == ':');
                *(s2 + 1) = '\0';
                IdlRemoveScope(s, absolute, fullyScoped);
            }
        }
        else if(len > 2 && len == strlen(absolute) + 2 &&
                strncmp(scope, absolute, len - 2) == 0)
        {
            //
            // If scope matches the absolute name, then use the
            // unscoped name. For example, if the scope is A::B::,
            // and the absolute name is A::B, then use B.
            //
            const char* cp = strrchr(absolute, ':');
            String_var tmp;
            if(cp != 0)
            {
                tmp = string_dup(cp + 1);
                string_free(absolute);
                absolute = tmp._retn();
            }
        }
        else if(fullyScoped)
        {
            //
            // Use the fully-scoped name
            //
            String_var tmp = string_dup("::");
            tmp += absolute;
            string_free(absolute);
            absolute = tmp._retn();
        }
    }
}

// ----------------------------------------------------------------------
// Get original type, i.e. translate aliases to the real type
// ----------------------------------------------------------------------

IDLType_ptr
IdlGetOrigIDLType(IDLType_ptr idlType)
{
    IDLType_var origType = IDLType::_duplicate(idlType);

    while(true)
    {
        AliasDef_var alias = AliasDef::_narrow(origType);

        if(is_nil(alias))
            break;
        else
            origType = alias -> original_type_def();
    }

    return IDLType::_duplicate(origType);
}

// ----------------------------------------------------------------------
// Check for variable size type
// ----------------------------------------------------------------------

bool
IdlIsVariable(TypeCode_ptr t)
{
    TypeCode_var type = GetOrigType(t);

    switch(type -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
        return false;

    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_string:
    case tk_wstring:
    case tk_sequence:
    case tk_except:
    case tk_value:
    case tk_value_box:
    case tk_abstract_interface:
    case tk_local_interface:
        return true;

    case tk_struct:
    case tk_union:
    {
        for(ULong i = 0 ; i < type -> member_count() ; i++)
        {
            TypeCode_var memberType = type -> member_type(i);
            if(IdlIsVariable(memberType))
                return true;
        }

        return false;
    }

    case tk_array:
    {
        TypeCode_var contentType = type -> content_type();

        if(IdlIsVariable(contentType))
            return true;
        else
            return false;
    }

    case tk_native:
        return false;

    case tk_alias: // No alias possible here
        break;     // this is to keep the compiler happy
    }

    return false; // This is to keep the compiler happy
}

// ----------------------------------------------------------------------
// Check whether a type contains native
// ----------------------------------------------------------------------

bool
IdlHasNative(TypeCode_ptr type)
{
    if(type -> _OB_isRecursive())
        return false;

    switch(type -> kind())
    {
    case tk_struct:
    case tk_union:
    case tk_value:
    case tk_except:
    {
        for(ULong i = 0 ; i < type -> member_count() ; i++)
        {
            TypeCode_var memberType = type -> member_type(i);
            if(IdlHasNative(memberType))
                return true;
        }

        return false;
    }

    case tk_alias:
    case tk_sequence:
    case tk_array:
    case tk_value_box:
    {
        TypeCode_var contentType = type -> content_type();

        if(IdlHasNative(contentType))
            return true;
        else
            return false;
    }

    case tk_native:
        return true;

    default:
        return false;
    }

    return false; // This is to keep the compiler happy
}

// ----------------------------------------------------------------------
// Check whether an interface contains native
// ----------------------------------------------------------------------

bool
IdlHasNative(const InterfaceDef::FullInterfaceDescription& desc)
{
    ULong i;
    for(i = 0 ; i < desc.operations.length(); i++)
        for(ULong j = 0 ; j < desc.operations[i].parameters.length() ; j++)
            if(IdlHasNative(desc.operations[i].parameters[j].type))
                return true;

    for(i = 0 ; i < desc.attributes.length(); i++)
        if(IdlHasNative(desc.attributes[i].type))
            return true;

    return false;
}

// ----------------------------------------------------------------------
// Determines whether a type contains a valuetype
// ----------------------------------------------------------------------

bool
IdlHasValuetype(TypeCode_ptr tc)
{
    bool result = false;

    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_string:
    case tk_wstring:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_enum:
    case tk_native:
    case tk_longdouble:
    case tk_local_interface:
        break;

    case tk_struct:
    case tk_union:
    case tk_except:
    {
        for(ULong i = 0 ; !result && i < tc -> member_count() ; i++)
        {
            TypeCode_var memberType = tc -> member_type(i);
            if(!memberType -> _OB_isRecursive())
                result = IdlHasValuetype(memberType);
        }
        break;
    }

    case tk_sequence:
    case tk_array:
    case tk_alias:
    {
        TypeCode_var contentType = tc -> content_type();
        if(!contentType -> _OB_isRecursive())
            result = IdlHasValuetype(contentType);
        break;
    }

    case tk_value:
    case tk_value_box:
    case tk_abstract_interface:
        result = true;
        break;
    }

    return result;
}

// ----------------------------------------------------------------------
// Get TypeCode converted to ASCII
// Output TypeCode converted to ASCII
// ----------------------------------------------------------------------

char*
IdlGetAsciiTypeCode(TypeCode_ptr tc)
{
    OCI::BufferImpl buf;
    OutputStreamImpl out(&buf);
    out._OB_writeEndian();
    out.write_TypeCode(tc);

    char* typeCodeStr = OctetsToAscii(buf.data(), buf.pos());

    return typeCodeStr;
}

void
IdlOutputAsciiTypeCode(TypeCode_ptr tc, IdlPrettyPrint& out)
{
    String_var typeCodeStr = IdlGetAsciiTypeCode(tc);

    unsigned int indent = out.getIndent();
    out.setIndent(0);

    ULong j = 0;

    do
    {
        out << "\n\"";
        for(ULong i = 0 ; i < 77 && typeCodeStr[j] != '\0' ; i++)
            out << typeCodeStr[j++];
        out << '"';
    }
    while(typeCodeStr[j] != '\0');

    out << '\n';

    out.setIndent(indent);
}

// ----------------------------------------------------------------------
// Convert path name to identifier
// ----------------------------------------------------------------------

char*
IdlPathToIdent(const char* s)
{
    String_var ident;

    for(ULong i = 0 ; i < strlen(s) ; i++)
    {
        char c = s[i];

        if(c == '.' || c == '/')
            c = '_';

        if(isalnum(c) || c == '_')
            ident += c;
    }

    return ident._retn();
}

// ----------------------------------------------------------------------
// Check if union has an implicit default
// ----------------------------------------------------------------------

bool
IdlImplicitUnionDefault(TypeCode_ptr type)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_union);

    bool result = false;

    if(origType -> default_index() == -1)
    {
        result = true;

        TypeCode_var tc = origType -> discriminator_type();
        tc = GetOrigType(tc);

        if(tc -> kind() == tk_char)
            if(origType -> member_count() == 256)
                result = false;

        if(tc -> kind() == tk_boolean)
            if(origType -> member_count() == 2)
                result = false;

        if(tc -> kind() == tk_enum)
            if(origType -> member_count() == tc -> member_count())
                result = false;
    }

    return result;
}

// ----------------------------------------------------------------------
// Get union info
// ----------------------------------------------------------------------

static void
convertUnionLabel(ULongLong value, bool neg, TypeCode_ptr disc, Any& any)
{
    TypeCode_var origDisc = GetOrigType(disc);

    switch(origDisc -> kind())
    {
    case tk_short:
        if(neg)
            any <<= (Short)(-1 * (Short)value);
        else
            any <<= (Short)value;
        break;

    case tk_ushort:
        assert(!neg);
        any <<= (UShort)value;
        break;

    case tk_long:
        if(neg)
            any <<= (Long)(-1 * (Long)value);
        else
            any <<= (Long)value;
        break;

    case tk_ulong:
        assert(!neg);
        any <<= (ULong)value;
        break;

    case tk_longlong:
        if(neg)
            any <<= (LongLong)(-1 * (LongLong)value);
        else
            any <<= (LongLong)value;
        break;

    case tk_ulonglong:
        assert(!neg);
        any <<= value;
        break;

    case tk_enum:
        assert(!neg);
        any.replace(disc, new ULong((ULong)value), true);
        break;

    case tk_char:
        any <<= Any::from_char((Char)value);
        break;

    case tk_boolean:
        assert(!neg);
        any <<= Any::from_boolean((Boolean)(value ? true : false));
        break;

    case tk_octet:
        assert(!neg);
        any <<= Any::from_octet((Octet)value);
        break;

    default:
        assert(false);
        break;
    }
}

OB_STL::list<IdlUnionMemberInfo>*
IdlGetUnionInfo(TypeCode_ptr type)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_union);

    TypeCode_var discType = origType -> discriminator_type();
    TypeCode_var origDiscType = GetOrigType(discType);

    IdlUnionMemberInfoList_var list = new OB_STL::list<IdlUnionMemberInfo>;

    bool hasDefault = false;
    OB_STL::list<IdlUnionMemberInfo>::iterator defaultPos;

    ULong i, k;

    for(i = 0 ; i < origType -> member_count() ; i++)
    {
        Identifier_var name = origType -> member_name(i);

        if(i > 0)
        {
            Identifier_var s = origType -> member_name(i - 1);

            if(strcmp(s, name) != 0)
            {
                IdlUnionMemberInfo new_member;
                new_member.name = name;
                new_member.type = origType -> member_type(i);
                new_member.isDefault = false;
                list -> push_back(new_member);
            }
            else
            {
                assert(list -> back().nLabels.length() > 0 ||
                       list -> back().pLabels.length() > 0 ||
                       list -> back().isDefault);
            }
        }
        else
        {
            IdlUnionMemberInfo new_member;
            new_member.name = name;
            new_member.type = origType -> member_type(i);
            new_member.isDefault = false;
            list -> push_back(new_member);
        }

        IdlUnionMemberInfo& info = list -> back();

        Any_var label = origType -> member_label(i);
        TypeCode_var labelType = label -> type();
        TypeCode_var origLabelType = GetOrigType(labelType);

        if(labelType -> kind() != tk_octet)
        {
            info.labels.length(info.labels.length() + 1);
            info.labels[info.labels.length() - 1] = label;
        }

        switch(origLabelType -> kind())
        {
        case tk_short:
        {
            Short val;
            Boolean b = label >>= val;
            assert(b);
            if(val >= 0)
            {
                info.pLabels.length(info.pLabels.length() + 1);
                info.pLabels[info.pLabels.length() - 1] = (ULongLong)val;
            }
            else
            {
                info.nLabels.length(info.nLabels.length() + 1);
                info.nLabels[info.nLabels.length() - 1] = (ULongLong)-val;
            }
            break;
        }

        case tk_ushort:
        {
            UShort val;
            Boolean b = label >>= val;
            assert(b);
            info.pLabels.length(info.pLabels.length() + 1);
            info.pLabels[info.pLabels.length() - 1] = (ULongLong)val;
            break;
        }

        case tk_long:
        {
            Long val;
            Boolean b = label >>= val;
            assert(b);
            if(val >= 0)
            {
                info.pLabels.length(info.pLabels.length() + 1);
                info.pLabels[info.pLabels.length() - 1] = (ULongLong)val;
            }
            else
            {
                info.nLabels.length(info.nLabels.length() + 1);
                info.nLabels[info.nLabels.length() - 1] = (ULongLong)-val;
            }
            break;
        }

        case tk_ulong:
        {
            ULong val;
            Boolean b = label >>= val;
            assert(b);
            info.pLabels.length(info.pLabels.length() + 1);
            info.pLabels[info.pLabels.length() - 1] = (ULongLong)val;
            break;
        }

        case tk_longlong:
        {
            LongLong val;
            Boolean b = label >>= val;
            assert(b);
            if(val >= 0)
            {
                info.pLabels.length(info.pLabels.length() + 1);
                info.pLabels[info.pLabels.length() - 1] = (ULongLong)val;
            }
            else
            {
                info.nLabels.length(info.nLabels.length() + 1);
                info.nLabels[info.nLabels.length() - 1] = (ULongLong)-val;
            }
            break;
        }

        case tk_ulonglong:
        {
            ULongLong val;
            Boolean b = label >>= val;
            assert(b);
            info.pLabels.length(info.pLabels.length() + 1);
            info.pLabels[info.pLabels.length() - 1] = (ULongLong)val;
            break;
        }

        case tk_enum:
        {
            ULong val = *(ULong*)(label -> value());
            info.pLabels.length(info.pLabels.length() + 1);
            info.pLabels[info.pLabels.length() - 1] = (ULongLong)val;
            break;
        }

        case tk_char:
        {
            Char val;
            Boolean b = label >>= Any::to_char(val);
            assert(b);
            LongLong lval = (LongLong)val;
            if(lval < 0)
                lval += 256; // convert signed char to unsigned char
            info.pLabels.length(info.pLabels.length() + 1);
            info.pLabels[info.pLabels.length() - 1] = (ULongLong)lval;
            break;
        }

        case tk_boolean:
        {
            Boolean val;
            Boolean b = label >>= Any::to_boolean(val);
            assert(b);
            info.pLabels.length(info.pLabels.length() + 1);
            info.pLabels[info.pLabels.length() - 1] = (ULongLong)val;
            break;
        }

        case tk_octet:
            assert(hasDefault == false);
            hasDefault = true;
            defaultPos = --(list -> end());
            info.isDefault = true;
            break;

        default:
            assert(false);
            break;
        }
    }

    ULongLong min, max;

    switch(origDiscType -> kind())
    {
    case tk_short:
        min = 32768;
        max = 32767;
        break;

    case tk_ushort:
        min = 0;
        max = 65535;
        break;

    case tk_long:
        min = 2147483648U;
        max = 2147483647U;
        break;

    case tk_ulong:
        min = 0;
        max = 4294967295U;
        break;

    case tk_longlong:
        min = 2147483648U;
        max = 2147483647U;
        break;

    case tk_ulonglong:
        min = 0;
        max = 4294967295U;
        break;

    case tk_enum:
        min = 0;
        max = origDiscType -> member_count() - 1;
        break;

    case tk_char:
        min = 128;
        max = 127;
        break;

    case tk_boolean:
        min = 0;
        max = 1;
        break;

    default:
        assert(false);
        break;
    }

    bool hasPDefault = false;
    bool hasNDefault = false;
    ULongLong n;
    ULongLong defPLabel;
    ULongLong defNLabel;

    for(n = 0 ; n <= max ; n++)
    {
        bool ok = true;
        OB_STL::list<IdlUnionMemberInfo>::iterator iter;
        for(iter = list -> begin(); ok && iter != list -> end(); iter++)
            for(k = 0 ; ok && k < (*iter).pLabels.length() ; k++)
                if((*iter).pLabels[k] == n)
                    ok = false;

        if (ok)
        {
            hasPDefault = true;
            defPLabel = n;
            break;
        }
    }

    for(n = 1 ; n <= min ; n++)
    {
        bool ok = true;

        OB_STL::list<IdlUnionMemberInfo>::iterator iter;
        for(iter = list -> begin(); ok && iter != list -> end(); iter++)
            for(k = 0 ; ok && k < (*iter).nLabels.length() ; k++)
                if((*iter).nLabels[k] == n)
                    ok = false;

        if (ok)
        {
            hasNDefault = true;
            defNLabel = n;
            break;
        }
    }

    if(hasPDefault || hasNDefault)
    {
        if(!hasDefault)
        {
            IdlUnionMemberInfo new_member;
            new_member.name = string_dup("");
            new_member.isDefault = true;
            list -> push_back(new_member);

            hasDefault = true;
            defaultPos = --(list -> end());
        }

        // This is legal:
        //
        //     union uni switch(short)
        //     {
        //     case 1 :
        //     default : long value;
        //     };
        //
        // Therefore the following asserts are wrong:
        //
        //assert((*defaultPos).pLabels.length() == 0);
        //assert((*defaultPos).nLabels.length() == 0);
        assert((*defaultPos).isDefault);

        if(hasPDefault)
        {
            (*defaultPos).pLabels.length((*defaultPos).pLabels.length() + 1);
            (*defaultPos).pLabels[(*defaultPos).pLabels.length() - 1] =
                defPLabel;

            //
            // Convert default value to discriminator type and add to
            // info.labels
            //
            ULong len = (*defaultPos).labels.length();
            (*defaultPos).labels.length(len + 1);
            convertUnionLabel(defPLabel, false, discType,
                              (*defaultPos).labels[len]);
        }

        if(hasNDefault)
        {
            (*defaultPos).nLabels.length((*defaultPos).nLabels.length() + 1);
            (*defaultPos).nLabels[(*defaultPos).nLabels.length() - 1] =
                defNLabel;

            //
            // Convert default value to discriminator type and add to
            // info.labels
            //
            ULong len = (*defaultPos).labels.length();
            (*defaultPos).labels.length(len + 1);
            convertUnionLabel(defNLabel, true, discType,
                              (*defaultPos).labels[len]);
        }
    }
    else
    {
        if(hasDefault)
        {
            list -> erase(defaultPos);
            hasDefault = false;
        }
    }

    return list._retn();
}

// ----------------------------------------------------------------------
// Get all base interfaces
// ----------------------------------------------------------------------

void
IdlGetAllBaseInterfaces(InterfaceDef_ptr interf, InterfaceDefSeq& seq)
{
    ULong i, j;

    InterfaceDefSeq_var baseSeq = interf -> base_interfaces();
    for(i = 0 ; i < baseSeq -> length() ; i++)
    {
        RepositoryId_var baseId = baseSeq[i] -> id(); 
        for(j = 0 ; j < seq.length() ; j++)
        {
            RepositoryId_var id = seq[j] -> id(); 
            if(strcmp(baseId, id) == 0)
                break;
        }

        if(j == seq.length())
        {
            seq.length(seq.length() + 1);
            seq[seq.length() - 1] = InterfaceDef::_duplicate(baseSeq[i]);
            IdlGetAllBaseInterfaces(baseSeq[i], seq);
        }
    }
}

// ----------------------------------------------------------------------
// Get information about base interfaces
// ----------------------------------------------------------------------

void
IdlGetBaseInterfaces(InterfaceDef_ptr interf, IdlInterfaceInfoSeq& seq,
                     bool all)
{
    InterfaceDefSeq_var baseSeq = interf -> base_interfaces();
    for(ULong i = 0 ; i < baseSeq -> length() ; i++)
    {
        RepositoryId_var baseId = baseSeq[i] -> id(); 
        ULong j;
        for(j = 0 ; j < seq.length() ; j++)
            if(strcmp(baseId, seq[j].id) == 0)
                break;

        if(j == seq.length())
        {
            if(all)
                IdlGetBaseInterfaces(baseSeq[i], seq, all);

            seq.length(seq.length() + 1);
            seq[seq.length() - 1].intf = InterfaceDef::_duplicate(baseSeq[i]);
            seq[seq.length() - 1].id = baseId;
            DefinitionKind kind = baseSeq[i] -> def_kind();
            seq[seq.length() - 1].is_abstract = (kind == dk_AbstractInterface);
            seq[seq.length() - 1].is_local = (kind == dk_LocalInterface);
        }
    }
}

// ----------------------------------------------------------------------
// Get all base values
// ----------------------------------------------------------------------

void
IdlGetAllBaseValues(ValueDef_ptr value, ValueDefSeq& seq)
{
    ULong i, j;

    ValueDefSeq_var baseSeq = value -> abstract_base_values();
    for(i = 0 ; i < baseSeq -> length() ; i++)
    {
        RepositoryId_var baseId = baseSeq[i] -> id(); 
        for(j = 0 ; j < seq.length() ; j++)
        {
            RepositoryId_var id = seq[j] -> id(); 
            if(strcmp(baseId, id) == 0)
                break;
        }

        if(j == seq.length())
        {
            seq.length(seq.length() + 1);
            seq[seq.length() - 1] = ValueDef::_duplicate(baseSeq[i]);
            IdlGetAllBaseValues(baseSeq[i], seq);
        }
    }
}

// ----------------------------------------------------------------------
// Get all truncatable base values
// ----------------------------------------------------------------------

void
IdlGetAllTruncatableBaseValues(ValueDef_ptr value,
                               ValueDefSeq& seq)
{
    if(value -> is_truncatable())
    {
        ValueDef_var base = value -> base_value();
        assert(!is_nil(base));
        seq.length(seq.length() + 1);
        seq[seq.length() - 1] = ValueDef::_duplicate(base);
        IdlGetAllTruncatableBaseValues(base, seq);
    }
}

// ----------------------------------------------------------------------
// Get all value members
// ----------------------------------------------------------------------

void
IdlGetAllValueMembers(CORBA::ValueDef_ptr value, CORBA::ValueMemberSeq& seq)
{
    ValueDef_var base = value -> base_value();
    if(!is_nil(base))
        IdlGetAllValueMembers(base, seq);

    Container::DescriptionSeq_var contents =
        value -> describe_contents(dk_ValueMember, true, -1);

    for(ULong i = 0 ; i < contents -> length() ; i++)
    {
        const ValueMember* vm;
        Boolean b = contents[i].value >>= vm;
        assert(b);
        seq.length(seq.length() + 1);
        seq[seq.length() - 1] = *vm;
    }
}

// ----------------------------------------------------------------------
// Collects all supported interfaces
// ----------------------------------------------------------------------

static void
IdlGetAllSupportedInterfacesHelper(ValueDef_ptr value,
                                   IdlSupportedInterfaceInfoSeq& seq,
                                   ULong& nreg,
                                   ULong level)
{
    ULong i, j;
    bool abstractValue = value -> is_abstract();
    RepositoryId_var valueId = value -> id();

    InterfaceDefSeq_var intfSeq = value -> supported_interfaces();
    for(i = 0 ; i < intfSeq -> length() ; i++)
    {
        RepositoryId_var id = intfSeq[i] -> id();
        ULong len = seq.length();
        for(j = 0 ; j < len ; j++)
        {
            if(strcmp(seq[j].interface_info.id, id) == 0)
                break;
        }

        if(j >= len)
        {
            seq.length(len + 1);
            seq[len].interface_info.id = id;
            seq[len].interface_info.intf =
                InterfaceDef::_duplicate(intfSeq[i]);
            seq[len].interface_info.is_abstract =
                (intfSeq[i] -> def_kind() == dk_AbstractInterface);
            seq[len].abstract_value = abstractValue;
            seq[len].value_id = valueId;
            seq[len].inheritance_level = level;
            if(!seq[len].interface_info.is_abstract)
                nreg++;
        }
        else
        {
            //
            // If an interface is supported by an abstract valuetype
            // and by a stateful valuetype, the stateful valuetype has
            // precedence. We also ensure that a supported interface
            // is assigned the highest possible inheritance level.
            //
            if(!abstractValue &&
               (seq[j].abstract_value || level > seq[j].inheritance_level))
            {
                seq[j].abstract_value = false;
                seq[j].value_id = valueId;
                seq[j].inheritance_level = level;
            }
        }
    }

    //
    // Interfaces supported by abstract base valuetypes are given the
    // same inheritance level as this valuetype
    //
    ValueDefSeq_var abstractBases = value -> abstract_base_values();
    for(i = 0 ; i < abstractBases -> length() ; i++)
        IdlGetAllSupportedInterfacesHelper(abstractBases[i], seq, nreg, level);

    ValueDef_var base = value -> base_value();
    if(!is_nil(base))
        IdlGetAllSupportedInterfacesHelper(base, seq, nreg, level + 1);
}

void
IdlGetAllSupportedInterfaces(ValueDef_ptr value,
                             IdlSupportedInterfaceInfoSeq& seq,
                             ULong& nreg)
{
    IdlGetAllSupportedInterfacesHelper(value, seq, nreg, 0);
}

// ----------------------------------------------------------------------
// Select the primary interface supported by a valuetype
// ----------------------------------------------------------------------

char*
IdlGetPrimarySupportedInterface(const IdlSupportedInterfaceInfoSeq& info)
{
    RepositoryId_var id;
    ULong i;

    //
    // Our first choice is the directly supported interface,
    // if any. If this valuetype doesn't directly support an
    // interface, then we'll use the one supported by the
    // concrete base type, if any. If there is no concrete
    // base type, or it doesn't support a regular interface,
    // then pick the first regular interface supported by an
    // abstract base type.
    //
    for(i = 0 ; i < info.length() ; i++)
    {
        if(info[i].inheritance_level == 0 &&
           !info[i].abstract_value &&
           !info[i].interface_info.is_abstract)
        {
            id = info[i].interface_info.id;
            break;
        }
    }
    if(id.in() == 0)
    {
        for(i = 0 ; i < info.length() ; i++)
        {
            if(info[i].inheritance_level == 1 &&
               !info[i].abstract_value &&
               !info[i].interface_info.is_abstract)
            {
                id = info[i].interface_info.id;
                break;
            }
        }
    }
    if(id.in() == 0)
    {
        for(i = 0 ; i < info.length() ; i++)
        {
            if(!info[i].interface_info.is_abstract)
            {
                id = info[i].interface_info.id;
                break;
            }
        }
    }
    assert(id.in() != 0);

    return id._retn();
}

// ----------------------------------------------------------------------
// Determine whether value supports an abstract interface
// ----------------------------------------------------------------------

bool
IdlCheckValueSupportsAbstractInterface(ValueDef_ptr value)
{
    ULong i, j, k;

    InterfaceDefSeq_var intfSeq = value -> supported_interfaces();
    for(i = 0 ; i < intfSeq -> length() ; i++)
    {
        if(intfSeq[i] -> def_kind() == dk_AbstractInterface)
            return true;
        else
        {
            IdlInterfaceInfoSeq info;
            IdlGetBaseInterfaces(intfSeq[i], info, true);
            for(j = 0 ; j < info.length() ; j++)
                if(info[j].is_abstract)
                    return true;
        }
    }

    ValueDef_var base = value -> base_value();
    if(!is_nil(base))
    {
        intfSeq = base -> supported_interfaces();
        for(i = 0 ; i < intfSeq -> length() ; i++)
        {
            if(intfSeq[i] -> def_kind() == dk_AbstractInterface)
                return true;
            else
            {
                IdlInterfaceInfoSeq info;
                IdlGetBaseInterfaces(intfSeq[i], info, true);
                for(j = 0 ; j < info.length() ; j++)
                    if(info[j].is_abstract)
                        return true;
            }
        }
    }

    ValueDefSeq_var abstractBases = value -> abstract_base_values();
    for(i = 0 ; i < abstractBases -> length() ; i++)
    {
        intfSeq = abstractBases[i] -> supported_interfaces();
        for(j = 0 ; j < intfSeq -> length() ; j++)
        {
            if(intfSeq[j] -> def_kind() == dk_AbstractInterface)
                return true;
            else
            {
                IdlInterfaceInfoSeq info;
                IdlGetBaseInterfaces(intfSeq[j], info, true);
                for(k = 0 ; k < info.length() ; k++)
                    if(info[k].is_abstract)
                        return true;
            }
        }
    }

    return false;
}

// ----------------------------------------------------------------------
// Get all method names
// ----------------------------------------------------------------------

StringSeq*
IdlGetMethodNames(const ContainedSeq& attributes,
                  const ContainedSeq& operations)
{
    StringSeq_var names = new StringSeq;
    ULong i, j;

    //
    // Add each attribute to the method table
    //
    for(i = 0 ; i < attributes.length() ; i++)
    {
        Contained::Description_var contDesc =
            attributes[i] -> describe();
        const AttributeDescription* desc;
        Boolean b = (contDesc -> value) >>= desc;
        assert(b);

        String_var name = string_dup("_get_");
        name += desc -> name;
        names -> length(names -> length() + 1);
        names[names -> length() - 1] = name._retn();

        if(desc -> mode == ATTR_NORMAL)
        {
            String_var name = string_dup("_set_");
            name += desc -> name;
            names -> length(names -> length() + 1);
            names[names -> length() - 1] = name._retn();
        }
    }

    //
    // Add each operation to the method table
    //
    for(i = 0 ; i < operations.length() ; i++)
    {
        Contained::Description_var contDesc =
            operations[i] -> describe();
        const OperationDescription* desc;
        Boolean b = (contDesc -> value) >>= desc;
        assert(b);

        names -> length(names -> length() + 1);
        names[names -> length() - 1] = desc -> name;
    }

    //
    // Sort names (bubblesort)
    //
    for(i = 0 ; i < names -> length() ; i++)
    {
        for(j = i ; j < names -> length() ; j++)
        {
            if(strcmp(names[i], names[j]) > 0)
            {
                String_var name = names[i]._retn();
                names[i] = names[j]._retn();
                names[j] = name._retn();
            }
        }
    }

    return names._retn();
}

// ----------------------------------------------------------------------
// Get all method names and descriptions
// ----------------------------------------------------------------------

IdlNameValuePairSeq*
IdlGetMethodNamesAndDescriptions(const ContainedSeq& attributes,
                                 const ContainedSeq& operations)
{
    IdlNameValuePairSeq_var result = new IdlNameValuePairSeq;
    ULong i, j;

    //
    // Add each attribute to the method table
    //
    for(i = 0 ; i < attributes.length() ; i++)
    {
        Contained::Description_var contDesc =
            attributes[i] -> describe();
        const AttributeDescription* desc;
        Boolean b = (contDesc -> value) >>= desc;
        assert(b);

        String_var name = string_dup("_get_");
        name += desc -> name;
        result -> length(result -> length() + 1);
        result[result -> length() - 1].name = name._retn();
        result[result -> length() - 1].value = contDesc -> value;

        if(desc -> mode == ATTR_NORMAL)
        {
            String_var name = string_dup("_set_");
            name += desc -> name;
            result -> length(result -> length() + 1);
            result[result -> length() - 1].name = name._retn();
            result[result -> length() - 1].value = contDesc -> value;
        }
    }

    //
    // Add each operation to the method table
    //
    for(i = 0 ; i < operations.length() ; i++)
    {
        Contained::Description_var contDesc =
            operations[i] -> describe();
        const OperationDescription* desc;
        Boolean b = (contDesc -> value) >>= desc;
        assert(b);

        result -> length(result -> length() + 1);
        result[result -> length() - 1].name = desc -> name;
        result[result -> length() - 1].value = contDesc -> value;
    }

    //
    // Sort names (bubblesort)
    //
    for(i = 0 ; i < result -> length() ; i++)
    {
        for(j = i ; j < result -> length() ; j++)
        {
            if(strcmp(result[i].name, result[j].name) > 0)
            {
                IdlNameValuePair tmp = result[i];
                result[i] = result[j];
                result[j] = tmp;
            }
        }
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Get descriptions of all inherited operations and attributes of
// abstract interfaces that are not inherited by concrete base interfaces
// ----------------------------------------------------------------------

IdlNameValuePairSeq*
IdlGetAbstractMethods(InterfaceDef_ptr intf)
{
    ULong i;

    //
    // Get the immediate base interfaces
    //
    IdlInterfaceInfoSeq baseInfo;
    IdlGetBaseInterfaces(intf, baseInfo, false);

    ULong numAbstractBases = 0;
    for(i = 0 ; i < baseInfo.length() ; i++)
        if(baseInfo[i].is_abstract)
            numAbstractBases++;

    //
    // Return now if there are no immediate abstract base interfaces
    //
    if(numAbstractBases == 0)
        return new IdlNameValuePairSeq;

    //
    // For each base interface, collect all operations and
    // attributes into two lists, depending on whether the
    // immediate base interface is abstract or concrete.
    //
    IdlNameValuePairSeq abstract;
    StringSeq concrete;
    for(i = 0 ; i < baseInfo.length() ; i++)
    {
        Container::DescriptionSeq_var seq =
            baseInfo[i].intf -> describe_contents(dk_all, false, -1);

        for(ULong j = 0 ; j < seq -> length() ; j++)
        {
            IdlNameValuePair pair;

            if(seq[j].kind == dk_Attribute)
            {
                const AttributeDescription* desc;
                Boolean b = seq[j].value >>= desc;
                assert(b);
                pair.name = desc -> name;
                pair.value = seq[j].value;
            }
            else if(seq[j].kind == dk_Operation)
            {
                const OperationDescription* desc;
                Boolean b = seq[j].value >>= desc;
                assert(b);
                pair.name = desc -> name;
                pair.value = seq[j].value;
            }
            else
                continue;

            if(baseInfo[i].is_abstract)
            {
                ULong k;
                for(k = 0 ; k < abstract.length() ; k++)
                    if(strcmp(abstract[k].name, pair.name) == 0)
                        break;

                if(k == abstract.length())
                {
                    abstract.length(abstract.length() + 1);
                    abstract[abstract.length() - 1] = pair;
                }
            }
            else
            {
                ULong k;
                for(k = 0 ; k < concrete.length() ; k++)
                    if(strcmp(concrete[k], pair.name) == 0)
                        break;

                if(k == concrete.length())
                {
                    concrete.length(concrete.length() + 1);
                    concrete[concrete.length() - 1] = pair.name;
                }
            }
        }
    }

    //
    // Merge the lists, taking only those entries in the abstract
    // list that do not appear in the concrete list
    //
    IdlNameValuePairSeq_var result = new IdlNameValuePairSeq;
    for(i = 0 ; i < abstract.length() ; i++)
    {
        ULong j;
        for(j = 0 ; j < concrete.length() ; j++)
            if(strcmp(abstract[i].name, concrete[j]) == 0)
                break;

        if(j == concrete.length())
        {
            result -> length(result -> length() + 1);
            result[result -> length() - 1] = abstract[i];
        }
    }

    return result._retn();
}
