// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GEN_UTIL_H
#define GEN_UTIL_H

#ifndef OB_TRANSLATOR_IMPORT
#   ifdef OB_DLL
#       define OB_TRANSLATOR_IMPORT __declspec(dllimport)
#   else
#       define OB_TRANSLATOR_IMPORT /**/
#   endif
#endif

#define OBE_VERSION "2.1.0"
#define OBEC_INTEGER_VERSION "2010000L"

//
// Forward declarations
//

namespace OB
{
    template<class T> class list;
}

//
// The "Pretty-Print" class
//
class IdlPrettyPrint
{
    OB_STD(ostream)& out_;

    unsigned int _ob_num_;

protected:

    unsigned int indent_;
    unsigned int pos_;

    bool ignoreNextSep_;
    bool emitIndent_;

    //
    // Newline with indent
    //
    virtual void nl();
    virtual void indent();

public:

    IdlPrettyPrint(OB_STD(ostream)& out, unsigned int num = 4)
        : out_(out), _ob_num_(num), indent_(0), pos_(0),
          ignoreNextSep_(false), emitIndent_(true) { }
    virtual ~IdlPrettyPrint() { }

    //
    // Set and get indent
    //
    void setIndent(unsigned int in) { indent_ = in; }
    unsigned int getIndent() const { return indent_; }

    //
    // Get and set output position
    //
    unsigned int getPos() const { return pos_; }
    void setPos(unsigned int pos) { pos_ = pos; }

    //
    // Increment and decrement indent
    //
    void inc() { indent_ += _ob_num_; }
    void dec() { assert(indent_ >= _ob_num_); indent_ -= _ob_num_; }

    //
    // Print directly, without character interpretation
    //
    void direct(char c) { out_ << c; }
    void direct(const char* s) { out_ << s; }

    //
    // Line control to mark where indent needed
    //
    virtual void lineBreak();
    virtual void lineStart();

    //
    // Block start and end
    //
    virtual void start();
    virtual void end();

    //
    // Separator
    //
    virtual void sep();
    virtual void ignoreNextSep();

    //
    // Print
    //
    virtual void printChar(char);
    virtual void printWChar(wchar_t);
    virtual void printString(const char*);
    virtual void printWString(const wchar_t*);
};

//
// Output operators
//
IdlPrettyPrint& operator<<(IdlPrettyPrint&, char);
#ifndef HAVE_WIN32_WCHAR_TYPE
IdlPrettyPrint& operator<<(IdlPrettyPrint&, wchar_t);
#endif
IdlPrettyPrint& operator<<(IdlPrettyPrint&, const char*);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, const wchar_t*);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, short);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, int);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, long);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, CORBA::LongLong);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, unsigned short);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, unsigned int);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, unsigned long);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, CORBA::ULongLong);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, float);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, double);
IdlPrettyPrint& operator<<(IdlPrettyPrint&, CORBA::LongDouble);

//
// Remove scope from absolute name
//
void IdlRemoveScope(const char*, char*&, bool = false);

//
// Get original type, i.e. translate aliases to the real type
//
CORBA::IDLType_ptr IdlGetOrigIDLType(CORBA::IDLType_ptr);

//
// Check for variable size type
//
bool IdlIsVariable(CORBA::TypeCode_ptr);

//
// Check whether a type contains native
//
bool IdlHasNative(CORBA::TypeCode_ptr);

//
// Check whether an interface contains native
//
bool IdlHasNative(const CORBA::InterfaceDef::FullInterfaceDescription&);

//
// Check whether a type contains a valuetype
//
bool IdlHasValuetype(CORBA::TypeCode_ptr);

//
// Get TypeCode converted to ASCII
// Output TypeCode converted to ASCII
//
char* IdlGetAsciiTypeCode(CORBA::TypeCode_ptr);
void IdlOutputAsciiTypeCode(CORBA::TypeCode_ptr, IdlPrettyPrint&);

//
// Convert path name to identifier
//
char* IdlPathToIdent(const char*);

//
// Check if union has an implicit default
//
bool IdlImplicitUnionDefault(CORBA::TypeCode_ptr);

//
// Get union info
//
struct IdlUnionMemberInfo
{
    CORBA::String_var name;
    CORBA::TypeCode_var type;
    CORBA::ULongLongSeq pLabels;
    CORBA::ULongLongSeq nLabels;
    CORBA::AnySeq labels;
    bool isDefault;
};
typedef OB::VarSeq< IdlUnionMemberInfo, int > IdlUnionMemberInfoSeq;
typedef OB::SeqVar< OB::VarSeq< IdlUnionMemberInfo, int > >
     IdlUnionMemberInfoSeq_var;
typedef OB::VarVar<OB_STL::list<IdlUnionMemberInfo> >
     IdlUnionMemberInfoList_var;

OB_STL::list<IdlUnionMemberInfo>* IdlGetUnionInfo(CORBA::TypeCode_ptr);

//
// Create sequence with all base interfaces
//
void IdlGetAllBaseInterfaces(CORBA::InterfaceDef_ptr, CORBA::InterfaceDefSeq&);

//
// Create sequence with information about base interfaces
//
struct IdlInterfaceInfo
{
    CORBA::InterfaceDef_var intf;
    CORBA::RepositoryId_var id;
    bool is_abstract;
    bool is_local;
};
typedef OB::VarSeq< IdlInterfaceInfo, int > IdlInterfaceInfoSeq;

void IdlGetBaseInterfaces(CORBA::InterfaceDef_ptr, IdlInterfaceInfoSeq&, bool);

//
// Create sequence with all base values
//
void IdlGetAllBaseValues(CORBA::ValueDef_ptr, CORBA::ValueDefSeq&);

//
// Create sequence with all truncatable base values
//
void IdlGetAllTruncatableBaseValues(CORBA::ValueDef_ptr, CORBA::ValueDefSeq&);

//
// Create sequence with all value members, including all inherited members,
// in the proper marshalling order (i.e., from top to bottom)
//
void IdlGetAllValueMembers(CORBA::ValueDef_ptr, CORBA::ValueMemberSeq&);

//
// Create sequence with information about supported interfaces
//
struct IdlSupportedInterfaceInfo
{
    IdlInterfaceInfo interface_info;
    bool abstract_value;
    CORBA::RepositoryId_var value_id;
    CORBA::ULong inheritance_level;
};
typedef OB::VarSeq< IdlSupportedInterfaceInfo, int >
    IdlSupportedInterfaceInfoSeq;

//
// Creates sequence containing all supported interfaces, including
// supported interfaces of base values
//
void IdlGetAllSupportedInterfaces(CORBA::ValueDef_ptr,
                                  IdlSupportedInterfaceInfoSeq&,
                                  CORBA::ULong&);

//
// Select the primary interface supported by a valuetype
//
char* IdlGetPrimarySupportedInterface(const IdlSupportedInterfaceInfoSeq&);

//
// Determine whether the given valuetype directly or indirectly supports
// an abstract interface
//
bool IdlCheckValueSupportsAbstractInterface(CORBA::ValueDef_ptr);

//
// Get all method names
//
CORBA::StringSeq* IdlGetMethodNames(const CORBA::ContainedSeq&,
                                    const CORBA::ContainedSeq&);

//
// Get all method names and descriptions
//
struct IdlNameValuePair
{
    CORBA::String_var name;
    CORBA::Any value;
};
typedef OB::VarSeq< IdlNameValuePair, int > IdlNameValuePairSeq;
typedef OB::SeqVar< OB::VarSeq< IdlNameValuePair, int > >
    IdlNameValuePairSeq_var;

IdlNameValuePairSeq*
IdlGetMethodNamesAndDescriptions(const CORBA::ContainedSeq&,
                                 const CORBA::ContainedSeq&);

//
// Get descriptions of all inherited operations and attributes of
// abstract interfaces that are not inherited by concrete base interfaces
//
IdlNameValuePairSeq*
IdlGetAbstractMethods(CORBA::InterfaceDef_ptr);

#endif
