// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>

#include <Util.h>
#include <GramUtil.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#ifdef HAVE_STRINGS_H
#   include <strings.h>
#endif

#ifndef HAVE_STRCASECMP
#   define strcasecmp stricmp
#endif

using namespace CORBA;
using namespace OB;

extern int yynerrs;

long IdlLineno; // The current line number
CORBA::String_var IdlDocComment; // The current documentation comment
CORBA::ORB_ptr IdlORB; // The ORB
IdlNameManager* IdlNameMgr; // The name manager
bool IdlTypeCodeEnabled; // If true, TypeCode has been enabled
bool IdlPrincipalEnabled; // If true, Principal has been enabled
bool IdlIgnoreError; // If true, parsing tries to continue as long as possible
bool IdlCaseInsensitive; // If true, case is ignored when comparing identifiers
bool IdlAllowLongDouble; // If true, the IDL long double type is allowed
bool IdlAllowFixed; // If true, the IDL fixed type is allowed
bool IdlAllowValueTypes; // If true, value types are allowed
bool IdlAllowModuleReopening; // If true, it is allowed to reopen modules

// ----------------------------------------------------------------------
// IdlNameManager::StringCaseHasher public member implementations
// ----------------------------------------------------------------------

ULong
IdlNameManager::StringCaseHasher::hash(const char* key)
{
    ULong h = 0;
    ULong offset = 0;
    Long len = (Long)strlen(key);

    if(len < 16)
    {
        for(Long i = len ; i > 0 ; i--)
            h = (h * 37) + tolower(key[offset++]);
    }
    else
    {
        Long skip = len / 8;
        for(Long i = len ; i > 0 ; i -= skip, offset += skip)
            h = (h * 39) + tolower(key[offset]);
    }

    return h;
}

bool
IdlNameManager::StringCaseHasher::comp(const char* k1, const char* k2)
{
    return IdlStringCompare(k1, k2) == 0;
}

// ----------------------------------------------------------------------
// IdlNameManager private and protected member implementations
// ----------------------------------------------------------------------

void
IdlNameManager::processContainer(Container_ptr container)
{
    //
    // Create new scope and add it to the hashtable
    //
    Scope* scope = new Scope;
    scope -> container = Container::_duplicate(container);
    Contained_var contained = Contained::_narrow(container);
    if(!is_nil(contained))
    {
	scope -> id = contained -> id();
	scope -> name = contained -> name();
    }
    else
    {
	scope -> id = string_dup("");
	scope -> name = string_dup("");
    }
    scopeHash_.put(scope -> id, scope);

    //
    // Add super types, if any
    //
    InterfaceDef_var interfaceDef = InterfaceDef::_narrow(container);
    if(!is_nil(interfaceDef))
    {
	InterfaceDefSeq_var base = interfaceDef -> base_interfaces();
	ULong len = scope -> super.length();
	scope -> super.length(len + base -> length());
	for(ULong i = 0 ; i < base -> length() ; i++)
	    scope -> super[len + i] = base[i] -> id();
    }

    ValueDef_var valueDef = ValueDef::_narrow(container);
    if(!is_nil(valueDef))
    {
	ValueDefSeq_var base = valueDef -> abstract_base_values();
	ULong len = scope -> super.length();
	scope -> super.length(len + base -> length());
	for(ULong i = 0 ; i < base -> length() ; i++)
	    scope -> super[len + i] = base[i] -> id();
    }

    //
    // Add contents and recurse
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);
    for(ULong i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Add the new name definition
	//
	RepositoryId_var id = seq[i].contained_object -> id();
	Identifier_var name = seq[i].contained_object -> name();
	scope -> nameHash.put(name, id);

	//
	// Recurse if this is a container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);
	if(!is_nil(cont))
	    processContainer(cont);
    }
}

void
IdlNameManager::alreadyDefined(const char* name) const
{
    String_var err = string_dup("`");
    err += name;
    err += "' is already defined";
    IdlError(err);
}

void
IdlNameManager::repositoryIdChanged(const char* name) const
{
    String_var err = string_dup("can't change repository id of `");
    err += name;
    err += "'";
    IdlError(err);
}

void
IdlNameManager::meaningChanged(const char* name) const
{
    String_var err = string_dup("`");
    err += name;
    err += "' has changed its meaning";
    IdlError(err);
}

char*
IdlNameManager::definedInScope(Scope* scope, const char* name) const
{
    RepositoryId_var id;
    if(scope -> nameHash.get(name, id))
	return id._retn();

    for(ULong i = 0 ; i < scope -> super.length() ; i++)
    {
	Scope* scope2;
	bool getResult = scopeHash_.get(scope -> super[i], scope2);
	assert(getResult);

	id = definedInScope(scope2, name);
	if(strlen(id) > 0)
	    return id._retn();
    }
    
    return string_dup("");
}

// ----------------------------------------------------------------------
// IdlNameManager constructor and destructor
// ----------------------------------------------------------------------

IdlNameManager::IdlNameManager(Repository_ptr repository,
			       StringSeq& includeSeq,
			       bool all,
			       IdlStringHash& content,
			       IdlStringHash& contentFwd)
    : repository_(Repository::_duplicate(repository)),
      includeSeq_(includeSeq),
      all_(all),
      content_(content),
      contentFwd_(contentFwd),
      dummyIdCount_(0)
{
    //
    // Feed the repository
    //
    processContainer(repository);

    //
    // The repository is the first container
    //
    pushContainer("");

    //
    // Add empty file name to file name stack
    //
    fileStack_.length(fileStack_.length() + 1);
    fileStack_[fileStack_.length()-1] = string_dup("");

    //
    // Add a prefix stack with one empty prefix to the prefix stack
    // stack
    //
    prefStackStack_.length(1);
    prefStackStack_[0].length(1);
    prefStackStack_[0][0] = string_dup("");
}

IdlNameManager::~IdlNameManager()
{
    ULong i;

    //
    // Do some sanity checks
    //
    assert(yynerrs > 0 || scopeStack_.length() == 1);
#ifndef NDEBUG
    for(i = 0 ; i < prefStackStack_.length() ; i++)
	assert(yynerrs > 0 || prefStackStack_[i].length() == 1);
#endif

    //
    // Do cleanup
    //
    ScopeHash::Enumerator e = scopeHash_.keys();
    for(i = 0 ; i < e.length() ; i++)
    {
	Scope* p;
	bool getResult = scopeHash_.get(e[i], p);
	assert(getResult);
	delete p;
    }
}

// ----------------------------------------------------------------------
// IdlNameManager public member implementations
// ----------------------------------------------------------------------

Repository_ptr
IdlNameManager::getRepository() const
{
    return Repository::_duplicate(repository_);
}

void
IdlNameManager::setContainer(const char* id, Container_ptr container)
{
    //
    // Find the scope by ID
    //
    Scope* scope;
    bool getResult = scopeHash_.get(id, scope);
    assert(getResult);
	
    //
    // Add the container
    //
    assert(is_nil(scope -> container));
    scope -> container = Container::_duplicate(container);
}

void
IdlNameManager::pushContainer(const char* id)
{
    //
    // Find the scope by ID
    //
    Scope* scope;
    bool getResult = scopeHash_.get(id, scope);
    assert(getResult);
	
    //
    // Add scope to the scope stack
    //
    ULong idx = scopeStack_.length();
    scopeStack_.length(idx + 1);
    scopeStack_[idx] = scope;

    //
    // Push a new prefix on all prefix stacks
    //
    for(ULong i = 0 ; i < prefStackStack_.length() ; i++)
    {
	String_var prefix;
	ULong len = prefStackStack_[i].length();
	if(len)
	    prefix = prefStackStack_[i][len - 1];

	prefix += scope -> name;
	prefix += '/';

        prefStackStack_[i].length(prefStackStack_[i].length() + 1);
        prefStackStack_[i][prefStackStack_[i].length() - 1] = prefix;
    }
}

void
IdlNameManager::popContainer()
{
    scopeStack_.length(scopeStack_.length() - 1);
    
    //
    // Pop a prefix from all prefix stacks
    //
    for(ULong i = 0 ; i < prefStackStack_.length() ; i++)
    {
	assert(prefStackStack_[i].length() > 0);
	prefStackStack_[i].length(prefStackStack_[i].length() - 1);
    }
}

Container_ptr
IdlNameManager::getContainer() const
{
    assert(scopeStack_.length() > 0);
    Container_var cont = scopeStack_[scopeStack_.length() - 1] -> container;
    return cont._retn();
}

char*
IdlNameManager::getContainerID() const
{
    assert(scopeStack_.length() > 0);
    RepositoryId_var id = scopeStack_[scopeStack_.length() - 1] -> id;
    return id._retn();
}

bool
IdlNameManager::defineName(const char* name, const char* id, bool isContainer,
			   StringSeq* super, bool allowFwd, bool allowPriorUse)
{
    ULong idx = scopeStack_.length() - 1;
    
    RepositoryId_var id2;
    bool getResult;

    if(!allowFwd)
    {
	//
	// Since this is a type that does not support forward
	// declarations, we must check for a name clash with an
	// existing forward declaration.
	//
	getResult = scopeStack_[idx] -> nameFwdHash.get(name, id2);
	if(getResult)
	{
	    alreadyDefined(name);
	    return false;
	}
    }

    //
    // Check whether the name is already in use
    //
    getResult = scopeStack_[idx] -> useHash.get(name, id2);
    if(getResult && !allowPriorUse)
    {
	assert(!id || strcmp(id, id2) != 0);
	meaningChanged(name);
	return false;
    }

    //
    // Check whether the name is already defined
    //
    getResult = scopeStack_[idx] -> nameHash.get(name, id2);
    if(getResult)
    {
	alreadyDefined(name);
	return false;
    }
    
    //
    // Add the new name
    //
    scopeStack_[idx] -> nameHash.put(name, id);

    //
    // If this is a type that supports forward declarations, add the
    // new name also to the list of forward declarations. This is OK
    // even if the ID does not match existing forward declarations, in
    // order to allow changing the ID of interfaces or valuetypes
    // after a forward declaration.
    //
    if(allowFwd)
	scopeStack_[idx] -> nameFwdHash.put(name, id);

    //
    // If this name represents a container, create a new scope and add
    // it to the hashtable
    //
    if(isContainer)
    {
	assert(!scopeHash_.containsKey(id));

	Scope* scope = new Scope;
	scopeHash_.put(id, scope);
	
	scope -> name = name;
	scope -> id = id;
	
	//
	// Set super interfaces/valuetypes, if any
	//
	if(super)
	{
	    //
	    // Make sure the super interfaces/valuetypes are defined
	    //
	    for(ULong i = 0 ; i < super -> length() ; i++)
	    {
		if(!scopeHash_.containsKey((*super)[i]))
		{
		    IdlError("can't derive from forward declaration");
		    return false;
		}
	    }
	    
	    scope -> super = *super;
	}
    }

    //
    // Get and reset the comment
    //
    String_var comment = IdlDocComment._retn();
    IdlDocComment = string_dup("");

    //
    // Add the ID
    //
    if(isTopLevelFile())
    {
	content_.put(id, comment);
	contentFwd_.remove(id);
    }

    return true;
}

bool
IdlNameManager::defineDecl(const char* name)
{
    RepositoryId_var dummyId;
    dummyId += dummyIdCount_++;
    return defineName(name, dummyId);
}

void
IdlNameManager::updateID(const char* id)
{
    //
    // Get and reset the comment
    //
    String_var comment = IdlDocComment._retn();
    IdlDocComment = string_dup("");
	
    //
    // Update/add the ID. If the new comment is empty, use the old
    // comment. If there is no old comment either, we use an empty
    // comment.
    //
    if(isTopLevelFile())
    {
	if(strlen(comment) == 0)
	    content_.get(id, comment);

	content_.put(id, comment);
    }
}

bool
IdlNameManager::addForwardDecl(const char* name, const char* id)
{
    ULong idx = scopeStack_.length() - 1;
    
    RepositoryId_var id2;
    bool getResult;

    //
    // Check whether the name is already in use
    //
    getResult = scopeStack_[idx] -> useHash.get(name, id2);
    if(getResult)
    {
	assert(strcmp(id, id2) != 0);
	meaningChanged(name);
	return false;
    }

    //
    // Check whether there is already a forward declaration
    //
    // Note: For each forward declarable definition, there is always a
    // forward declaration, so the code after this if-statement is not
    // reached if an interface or abstract interface is declared after
    // it has been defined.
    //
    getResult = scopeStack_[idx] -> nameFwdHash.get(name, id2);
    if(getResult)
    {
	//
	// Always OK, even if IDs do not match, in order to allow
	// forward declarations after id-changed interface
	// definitions.
	//
	return true;

	/*
	//
	// OK as long as the IDs still match
	//
	if(strcmp(id, id2) != 0)
	{
	    repositoryIdChanged(name);
	    return false;
	}
	else
	    return true;
	*/
    }

    //
    // Check whether the name is already defined
    //
    getResult = scopeStack_[idx] -> nameHash.get(name, id2);
    if(getResult)
    {
	alreadyDefined(name);
	return false;
    }

    //
    // Add the forward declaration
    //
    scopeStack_[idx] -> nameFwdHash.put(name, id);

    //
    // Add the ID
    //
    if(isTopLevelFile())
	contentFwd_.put(id, (const char*)"");

    return true;
}

void
IdlNameManager::checkIncompleteDecl() const
{
    //
    // Get container from container stack
    //
    Container_var container = IdlNameMgr -> getContainer();

    ContainedSeq_var contained;
    ULong i;

    //
    // Check for incomplete struct declarations
    //
    contained = container -> contents(dk_Struct, true);
    for(i = 0 ; i < contained -> length() ; i++)
    {
	StructDef_var st = StructDef::_narrow(contained[i]);
	assert(!is_nil(st));

	//
	// 0 members means that there is a forward declaration only
	//
	StructMemberSeq_var members = st -> members();
	if(members -> length() == 0)
	{
	    Identifier_var name = st -> name();
	    String_var err = 
		string_dup("missing declaration for `struct ");
	    err += name;
	    err += "'";
	    IdlError(err);
	}
    }

    //
    // Check for incomplete union declarations
    //
    contained = container -> contents(dk_Union, true);
    for(i = 0 ; i < contained -> length() ; i++)
    {
	UnionDef_var un = UnionDef::_narrow(contained[i]);
	assert(!is_nil(un));

	//
	// 0 members means that there is a forward declaration only
	//
	UnionMemberSeq_var members = un -> members();
	if(members -> length() == 0)
	{
	    Identifier_var name = un -> name();
	    String_var err = 
		string_dup("missing declaration for `union ");
	    err += name;
	    err += "'";
	    IdlError(err);
	}
    }
}

bool
IdlNameManager::useName(const char* name, const char* id)
{
    for(Long idx = scopeStack_.length() - 1 ; idx >= 0 ; idx--)
    {
	RepositoryId_var id2;
	bool getResult;

	//
	// Check whether the name is already defined
	//
	id2 = definedInScope(scopeStack_[idx], name);
	if(strlen(id2) > 0)
	{
	    //
	    // As long as the meaning is not changed, that's ok
	    //
	    if(strcmp(id, id2) != 0)
	    {
		meaningChanged(name);
		return false;
	    }
	    else
		return true;
	}
	
	//
	// Check whether there is already a forward declaration
	//
	getResult = scopeStack_[idx] -> nameFwdHash.get(name, id2);
	if(getResult)
	{
	    //
	    // As long as the meaning is not changed, that's ok
	    //
	    if(strcmp(id, id2) != 0)
	    {
		meaningChanged(name);
		return false;
	    }
	    else
		return true;
	}
	
	//
	// Check whether the name is already in use
	//
	getResult = scopeStack_[idx] -> useHash.get(name, id2);
	if(getResult)
	{
	    //
	    // As long as the meaning is not changed, that's ok
	    //
	    if(strcmp(id, id2) != 0)
	    {
		meaningChanged(name);
		return false;
	    }
	    else
		return true;
	}
	
	//
	// Add the name usage
	//
	scopeStack_[idx] -> useHash.put(name, id);
    }

    if(IdlTypeCodeEnabled &&
       ((strcmp(name, "CORBA") == 0 &&
	 strcmp(id, "IDL:omg.org/CORBA:1.0") == 0) ||
       ((strcmp(name, "TypeCode") == 0 &&
	 strcmp(id, "IDL:omg.org/CORBA/TypeCode:1.0") == 0))))
    {
	return true;
    }
    else if(IdlPrincipalEnabled &&
	    ((strcmp(name, "CORBA") == 0 &&
	      strcmp(id, "IDL:omg.org/CORBA:1.0") == 0) ||
	     ((strcmp(name, "Principal") == 0 &&
	       strcmp(id, "IDL:omg.org/CORBA/Principal:1.0") == 0))))
    {
	return true;
    }
    else
    {
	//
	// This line should never be reached, since this operation
	// should only be called if the name to be used is defined.
	//
	OB_CERR << "id = " << id << OB_ENDL;
	OB_CERR << "name = " << name << OB_ENDL;
	IdlError("");
  	assert(false);
  	return false;
    }
}

bool
IdlNameManager::changeID(const char* oldID, const char* newID)
{
    if(strcmp(oldID, newID) == 0)
	return true;

    if(changedIDs_.containsKey(oldID))
    {
	String_var err = string_dup("`");
	err += oldID;
	err += "' can't be changed again";
	IdlError(err);
	return false;
    }
    changedIDs_.put(newID, (const char*)"");

    bool getResult;
    ScopeHash::Enumerator e = scopeHash_.keys();
    for(ULong i = 0 ; i < e.length() ; i++)
    {
	Scope* scope;
	getResult = scopeHash_.get(e[i], scope);
	assert(getResult);
	assert(strcmp(e[i], scope -> id) == 0);

	if(strcmp(scope -> id, oldID) == 0)
	{
	    scopeHash_.remove(oldID);
	    scopeHash_.put(newID, scope);
	    scope -> id = newID;
	}

	ULong j;

	for(j = 0 ; j < scope -> super.length() ; j++)
	    if(strcmp(scope -> super[j], oldID) == 0)
		scope -> super[j] = newID;

	NameHash::Enumerator e2;
	
	e2 = scope -> nameHash.keys();
	for(j = 0 ; j < e2.length() ; j++)
	{
	    RepositoryId_var id;
	    getResult = scope -> nameHash.get(e2[j], id);
	    assert(getResult);
	    if(strcmp(id, oldID) == 0)
		scope -> nameHash.put(e2[j], newID);
	}

	e2 = scope -> nameFwdHash.keys();
	for(j = 0 ; j < e2.length() ; j++)
	{
	    RepositoryId_var id;
	    getResult = scope -> nameFwdHash.get(e2[j], id);
	    assert(getResult);
	    if(strcmp(id, oldID) == 0)
		scope -> nameFwdHash.put(e2[j], newID);
	}

	e2 = scope -> useHash.keys();
	for(j = 0 ; j < e2.length() ; j++)
	{
	    RepositoryId_var id;
	    getResult = scope -> useHash.get(e2[j], id);
	    assert(getResult);
	    if(strcmp(id, oldID) == 0)
		scope -> useHash.put(e2[j], newID);
	}
    }
    
    String_var comment;

    getResult = content_.get(oldID, comment);
    if(getResult)
	content_.put(newID, comment);

    getResult = contentFwd_.get(oldID, comment);
    if(getResult)
	contentFwd_.put(newID, comment);

    return true;
}

void
IdlNameManager::setCurrentFile(const char* name, int enterOrLeave)
{
    ULong len = fileStack_.length();
    assert(len > 0);
    assert(prefStackStack_.length() == len);
    
    const char* old = fileStack_[len - 1];
    if(strcmp(old, name) != 0)
    {
	switch(enterOrLeave)
	{
	case 0:
	    assert(IdlLineno == 0);
	    fileStack_.length(len + 1);
	    fileStack_[len] = name;
	    prefStackStack_.length(len + 1);
	    prefStackStack_[len] = prefStackStack_[0];
	    break;
	    
	case 1:
	    fileStack_.length(len + 1);
	    fileStack_[len] = name;
	    prefStackStack_.length(len + 1);
	    prefStackStack_[len] = prefStackStack_[0];
	    if(len == 2)
	    {
		ULong k;
		for(k = 0 ; k < includeSeq_.length() ; k++)
		    if(!strcmp(name, includeSeq_[k]))
			break;
		if(k == includeSeq_.length())
		{
		    includeSeq_.length(includeSeq_.length() + 1);
		    includeSeq_[includeSeq_.length() - 1] = name;
		}
	    }
	    break;
	    
	case 2:
	    assert(strcmp(fileStack_[len - 2], name) == 0);
	    fileStack_.length(len - 1);
	    prefStackStack_.length(len - 1);
	    break;
	    
	default:
	    assert(false);
	    break;
	}
    }
}

char*
IdlNameManager::getCurrentFile() const
{
    return string_dup(fileStack_[fileStack_.length() - 1]);
}

bool
IdlNameManager::isTopLevelFile() const
{
    return all_ || fileStack_.length() <= 2;
}

void
IdlNameManager::setCurrentPrefix(const char* prefix)
{
    ULong stackStackLen = prefStackStack_.length();
    assert(stackStackLen > 0);
    
    ULong stackLen = prefStackStack_[stackStackLen - 1].length();
    assert(stackLen > 0);
    
    prefStackStack_[stackStackLen - 1][stackLen - 1] = prefix;
}

char*
IdlNameManager::getCurrentPrefix() const
{
    ULong stackStackLen = prefStackStack_.length();
    assert(stackStackLen > 0);
    
    ULong stackLen = prefStackStack_[stackStackLen - 1].length();
    assert(stackLen > 0);
    
    return string_dup(prefStackStack_[stackStackLen - 1][stackLen - 1]);
}

// ----------------------------------------------------------------------
// Find scoped name
// ----------------------------------------------------------------------

IRObject_ptr
IdlFindScopedName(Container_ptr con, const char* scopedName)
{
    Repository_var rep = IdlNameMgr -> getRepository();
    IRObject_var irObj;

    if(strcmp(scopedName, "CORBA::TypeCode") == 0 ||
       strcmp(scopedName, "::CORBA::TypeCode") == 0)
    {
	if(IdlTypeCodeEnabled)
	{
	    irObj = rep -> get_primitive(pk_TypeCode);

	    if(scopedName[0] != ':')
		IdlNameMgr -> useName("CORBA", "IDL:omg.org/CORBA:1.0");
	}
	else
	{	
	    IdlError("To enable TypeCode, include `orb.idl' "
		     "or `TypeCode.idl'");
	}
    }
    else if(strcmp(scopedName, "CORBA::Principal") == 0 ||
	    strcmp(scopedName, "::CORBA::Principal") == 0)
    {
	if(IdlPrincipalEnabled)
	{
	    irObj = rep -> get_primitive(pk_Principal);

	    if(scopedName[0] != ':')
		IdlNameMgr -> useName("CORBA", "IDL:omg.org/CORBA:1.0");
	}
	else
	{	
	    IdlError("To enable Principal, include `orb.idl', "
		     "or use `#pragma Principal'");
	}
    }
    else
    {
	if(strcmp(scopedName, "TypeCode") == 0 ||
	   strcmp(scopedName, "Principal") == 0)
	{
	    Contained_var contained = Contained::_narrow(con);
	    if(!is_nil(contained))
	    {
		ScopedName_var contScopedName = contained -> absolute_name();
		if(strcmp(contScopedName, "::CORBA") == 0 ||
		   strncmp(contScopedName, "::CORBA::", 7) == 0)
		{
		    if(strcmp(scopedName, "TypeCode") == 0)
		    {
			if(IdlTypeCodeEnabled)
			{
			    irObj = rep -> get_primitive(pk_TypeCode);

			    IdlNameMgr ->
				useName("TypeCode",
					"IDL:omg.org/CORBA/TypeCode:1.0");
			}
			else
			{	
			    IdlError("To enable TypeCode, include `orb.idl' "
				     "or `TypeCode.idl'");
			}
		    }
		    else if(strcmp(scopedName, "Principal") == 0)
		    {
			if(IdlPrincipalEnabled)
			{
			    irObj = rep -> get_primitive(pk_Principal);

			    IdlNameMgr ->
				useName("Principal",
					"IDL:omg.org/CORBA/Principal:1.0");
			}
			else
			{	
			    IdlError("To enable Principal, include `orb.idl', "
				     "or use `#pragma Principal'");
			}
		    }
		    else
			assert(false);
		}
	    }
	}

	if(is_nil(irObj))
	{
	    Container_var container = Container::_duplicate(con);
	    Contained_var contained;

	    while(is_nil(contained))
	    {
		try
		{
		    contained = container -> lookup(scopedName);
		}
		catch(const INTF_REPOS& ex)
		{
		    IdlPrintException(ex, "", scopedName);
		    return IRObject::_nil();
		}
		
		if(is_nil(contained))
		{
		    Contained_var c = Contained::_narrow(container);

		    if(is_nil(c))
			break;
		    
		    container = c -> defined_in();
		}
	    }

	    if(scopedName[0] != ':' && !is_nil(contained))
	    {
		const char* s = strchr((char*)scopedName, ':');
		if(s)
		{
		    Identifier_var name = string_alloc(s - scopedName);
		    strncpy(name.inout(), scopedName, s - scopedName);
		    name[s - scopedName] = '\0';

		    Contained_var c = container -> lookup(name);
		    assert(!is_nil(c));
		    RepositoryId_var id = c -> id();

		    IdlNameMgr -> useName(name, id);
		}
		else
		{
		    RepositoryId_var id = contained -> id();
		    IdlNameMgr -> useName(scopedName, id);
		}
	    }

	    irObj = IRObject::_duplicate(contained);
	}
    }
    
    if(is_nil(irObj))
    {	
	String_var err = string_dup("`");
	err += scopedName;
	err += "' undeclared";
	IdlError(err);
	return IRObject::_nil();
    }

    return irObj._retn();
}

// ----------------------------------------------------------------------
// Find a primary expression for a scoped name
// ----------------------------------------------------------------------

Any*
IdlFindPrimaryExpr(Container_ptr container, const char* scopedName)
{
    //
    // First try to find matching enum members
    //

    const char* n = strrchr(scopedName, ':');
    ScopedName_var scope;
    
    if(n)
    {
	assert(n > scopedName);
	assert(*(n - 1) == ':');
	n++;
	
	scope = string_alloc(n - scopedName - 2);
	strncpy(scope.inout(), scopedName, n - scopedName - 2);
	scope[n - scopedName - 2] = '\0';
    }
    else
    {
	n = scopedName;
	scope = string_dup("");
    }

    Container_var cont = Container::_duplicate(container);
    while(true)
    {
        Container_var parent;
        if(strlen(scope) > 0)
        {
            try
            {
                Contained_var contained = cont -> lookup(scope);
                parent = Container::_narrow(contained);
            }
            catch(const INTF_REPOS& ex)
            {
                IdlPrintException(ex, "", scope);
                return 0;
            }
        }
        else
            parent = cont;

        if(!is_nil(parent))
        {
            Container::DescriptionSeq_var desc =
                parent -> describe_contents(dk_Enum, false, -1);

	    Any_var any;

            for(ULong i = 0 ; i < desc -> length() ; i++)
            {
                const TypeDescription* td;
                Boolean b = desc[i].value >>= td;
                assert(b);
                for(ULong j = 0 ; j < td -> type -> member_count() ; j++)
                {
                    if(strcmp(n, td -> type -> member_name(j)) == 0)
                    {
			if(any.ptr())
			{
			    INTF_REPOS ex(MinorLookupAmbiguous, COMPLETED_NO);
			    IdlPrintException(ex, "", scopedName);
			    return 0;
			}

                        any = new Any(td -> type, new ULong(j), true);
			break;
                    }
                }
            }

	    if(any.ptr())
		return any._retn();
        }
            
        Contained_var contained = Contained::_narrow(cont);
        
        if(is_nil(contained))
            break;
        
        cont = contained -> defined_in();
    }

    //
    // Now do a regular search
    //
    IRObject_var irObj = IdlFindScopedName(container, scopedName);
    
    if(is_nil(irObj))
	return 0;
    
    ConstantDef_var co = ConstantDef::_narrow(irObj);
    if(!is_nil(co))
    {
	return co -> value();
    }
    else
    {
	String_var err = string_dup("`");
	err += scopedName;
	err += "' is not a constant or an enum";
	IdlError(err);
	return 0;
    }
}

// ----------------------------------------------------------------------
// Output errors and warnings
// ----------------------------------------------------------------------

void
IdlError(const char* s)
{
    String_var n = IdlNameMgr -> getCurrentFile();

    if(strcmp(n, "") == 0)
	n = (const char*)"standard input";

    fflush(stderr);
    fprintf(stderr, "%s(%ld): error: %s\n", n.in(), IdlLineno, s);

    //
    // Increment the error counter
    //
    yynerrs++;
}

void
IdlWarning(const char* s)
{
    String_var n = IdlNameMgr -> getCurrentFile();

    if(strcmp(n, "") == 0)
	n = (const char*)"standard input";

    fflush(stderr);

    fprintf(stderr, "%s(%ld): warning: %s\n", n.in(), IdlLineno, s);
}

// ----------------------------------------------------------------------
// Parse line number and file name
// ----------------------------------------------------------------------

void
IdlParseLineAndFile(const char* yytext)
{
    String_var str = string_dup(yytext + 1);
    char* tok;
    static const char* sep = " \t\r";

    tok = strtok(str.inout(), sep);

    //
    // Filter out "line"
    //
    if(tok && strcmp(tok, "line") == 0)
	tok = strtok(0, sep);

    if(tok)
    {
	//
	// Get line number information
	//
	IdlLineno = atol(tok);
	
	//
        // Subtract 1, because the line number information refers to
        // the next line, not to this line.
	//
	IdlLineno--;

	if((tok = strtok(0, sep)))
	{
	    //
	    // Get file name information
	    //
	    if(tok[0] == '"')
		tok++;
	    
	    if(tok[strlen(tok) - 1] == '"')
		tok[strlen(tok) - 1] = '\0';
	    
	    if(strlen(tok))
	    {
#ifdef WIN32
		String_var newFile = string_alloc(strlen(tok));
		ULong i, j;
		for(i = 0, j = 0 ; i < strlen(tok) ; i++, j++)
		{
		    if(tok[i] == '\\' && tok[i + 1] == '\\')
		    {
			i++;
			newFile[j] = '\\';
		    }
		    else
			newFile[j] = tok[i];
		}

		newFile[j] = '\0';
#else
		String_var newFile = string_dup(tok);
#endif

		int enterOrLeave = 0;
		if((tok = strtok(0, sep)))
		    enterOrLeave = atoi(tok);

		IdlNameMgr -> setCurrentFile(newFile, enterOrLeave);
	    }
	}
    }
}

// ----------------------------------------------------------------------
// Check a repository id
// ----------------------------------------------------------------------

bool
IdlCheckRepositoryId(const char* id)
{
    String_var err = string_dup("`");
    err += id;
    err += "' is not a valid repository id";
    
    if(!strchr((char*)id, ':'))
    {
	IdlError(err);
	return false;
    }

    if(strncmp((char*)id, "IDL:", 4) == 0)
    {
	id += 4;
	char* s = strrchr((char*)id, ':');
	if(!s || s == id)
	{
	    IdlError(err);
	    return false;
	}

	while(id < s)
	{
	    if(!(isalnum(*id) ||
		 *id == '/' || *id == '_' || *id == '-' || *id == '.'))
	    {
		IdlError(err);
		return false;
	    }

	    id++;
	}

	return IdlCheckVersion(s + 1);	    
    }

    return true;
}

// ----------------------------------------------------------------------
// Check a version
// ----------------------------------------------------------------------

bool
IdlCheckVersion(const char* str)
{
    String_var err = string_dup("`");
    err += str;
    err += "' is not a valid version";

    String_var version = string_dup(str);
    
    char* majorStr = strtok(version.inout(), ".");
    if(majorStr)
    {
	char* minorStr = strtok(0, ".");
	if(minorStr)
	{
	    char* dummy = strtok(0, ".");
	    if(!dummy)
	    {
		ULong i;
		
		for(i = 0 ; i < strlen(majorStr) ; i++)
		    if(!isdigit(majorStr[i]))
		    {
			IdlError(err);
			return false;
		    }
		
		for(i = 0 ; i < strlen(minorStr) ; i++)
		    if(!isdigit(minorStr[i]))
		    {
			IdlError(err);
			return false;
		    }

		return true;
	    }
	}
    }

    IdlError(err);
    return false;
}

// ----------------------------------------------------------------------
// Parse pragma
// ----------------------------------------------------------------------

bool
IdlParsePragma(const char* yytext)
{
    static const char* pragmaStr = "pragma";

    while(strncmp(yytext, pragmaStr, strlen(pragmaStr)) != 0)
    {
	yytext++;
	assert(*yytext);
    }

    String_var str = string_dup(yytext + strlen(pragmaStr));
    char* tok;
    static const char* sep = " \t\r";
    bool formatError = true;

    if((tok = strtok(str.inout(), sep)))
    {
	//
	// The prefix pragma
	//
	if(strcmp(tok, "prefix") == 0)
	{
	    if((tok = strtok(0, sep)))
	    {
		int len = strlen(tok);
		if(len >= 2 && tok[0] == '"' && tok[len - 1] == '"')
		{
		    if(len == 2)
			tok[len - 1] = '\0'; // For #pragma prefix ""
		    else
			tok[len - 1] = '/';

		    tok++;

		    IdlNameMgr -> setCurrentPrefix(tok);

		    formatError = false;
		}
	    }
	}
	//
	// The ID pragma
	//
	else if(strcmp(tok, "ID") == 0)
	{
	    if((tok = strtok(0, sep)))
	    {
                String_var name = IdlUnescapeName(tok);
		Container_var container = IdlNameMgr -> getContainer();
		IRObject_var irObj = IdlFindScopedName(container, name);
		if(is_nil(irObj))
		    return false;

		Contained_var contained = Contained::_narrow(irObj);
		
		if(is_nil(contained))
		{
		    String_var err =
			string_dup("can't set repository id for `");
		    err += tok;
		    err += "'";
		    IdlError(err);
		    return false;
		}
		else if((tok = strtok(0, sep)))
		{
		    int len = strlen(tok);
		    if(len >= 2 && tok[0] == '"' && tok[len - 1] == '"')
		    {
			tok[len - 1] = '\0';
			tok++;

			if(!IdlCheckRepositoryId(tok))
			    return false;

			RepositoryId_var old = contained -> id();

			if(!IdlNameMgr -> changeID(old, tok))
			    return false;

			contained -> id(tok);

			formatError = false;
		    }
		}
	    }
	}
	//
	// The version pragma
	//
	else if(strcmp(tok, "version") == 0)
	{
	    if((tok = strtok(0, sep)))
	    {
		Container_var container = IdlNameMgr -> getContainer();
		IRObject_var irObj = IdlFindScopedName(container, tok);
		if(is_nil(irObj))
		    return false;

		Contained_var contained = Contained::_narrow(irObj);
		
		if(is_nil(contained))
		{
		    String_var err = string_dup("can't set version for `");
		    err += tok;
		    err += "'";
		    IdlError(err);
		    return false;
		}
		else if((tok = strtok(0, sep)))
		{
		    if(!IdlCheckVersion(tok))
			return false;

		    RepositoryId_var old = contained -> id();
		    
		    RepositoryId_var id = old;
		    char* p = strrchr(id.inout(), ':');
		    if(p)
			*p = '\0';
		    
		    id += ':';
		    id += tok;
		    
		    if(!IdlNameMgr -> changeID(old, id))
			return false;

		    contained -> version(tok);
		    contained -> id(id);
		    
		    formatError = false;
		}
	    }
	}
	//
	// The ORBacus proprietary TypeCode pragma
	//
	else if(strcmp(tok, "TypeCode") == 0)
	{
	    IdlTypeCodeEnabled = true;
	    formatError = false;
	}
	//
	// The ORBacus proprietary Principal pragma
	//
	else if(strcmp(tok, "Principal") == 0)
	{
	    IdlPrincipalEnabled = true;
	    formatError = false;
	}
    }

    if(formatError)
	IdlWarning("unknown #pragma format");

    return true;
}

// ----------------------------------------------------------------------
// Parse C-style comment (i.e. comment within /**/)
// ----------------------------------------------------------------------

void
IdlParseComment(const char* doc)
{
    assert(doc);

    String_var str = doc;

    if(str[0] == '*')
    {
	IdlDocComment = string_dup("");

	char* s1 = str.inout() + 1;
	char* s2;

	while(true)
	{
	    while(*s1 == '*')
		s1++;
	    
	    s2 = s1;
	    
	    while(*s2 != '\n' && *s2 != '\0')
		s2++;
	    
	    bool end = false;
	    if(*s2 == '\0')
		end = true;
	    else
		*s2 = '\0';
	    
	    IdlDocComment += s1;
	    
	    if(end)
		break;

	    IdlDocComment += '\n';

	    s1 = s2 + 1;

	    while(*s1 == ' ' || *s1 == '\t' || *s1 == '\r')
		s1++;
	}
    }
}

// ----------------------------------------------------------------------
// Get repository id
// ----------------------------------------------------------------------

char*
IdlGetRepositoryId(const char* name, const char* version)
{
    String_var prefix = IdlNameMgr -> getCurrentPrefix();

    String_var str = string_dup("IDL:");
    str += prefix;
    str += name;
    str += ':';
    str += version;

    return str._retn();
}

// ----------------------------------------------------------------------
// Print INTF_REPOS exception as error message
// ----------------------------------------------------------------------

void
IdlPrintException(const SystemException& ex, const char* id,
                  const char* scopedName)
{
    String_var err = string_dup("interface repository error");

    //
    // See IFR chapter for mandatory exceptions and minor codes
    //
    if(BAD_PARAM::_downcast(&ex) != 0)
    {
        switch(ex.minor())
        {
        case MinorRepositoryIdExists:
            err = string_dup("redefinition of repository id `");
            err += id;
            err += "'";
            break;

        case MinorNameExists:
            err = string_dup("`");
            err += scopedName;
            err += "' is already defined";
	    // This exception shouldn't be raised anymore, as such
	    // name clashes are now detected directly in the
	    // translator
	    assert(false);
            break;

        case MinorInvalidContainer:
            err = string_dup("`");
            err += scopedName;
            err += "' is not a valid container";
            break;

        case MinorNameClashInInheritedContext:
            err = string_dup("collision in inherited context of `");
            err += scopedName;
            err += "'";
            break;

	case MinorDuplicateDeclarator:
	    err = string_dup("duplicate declarator in `");
	    err += scopedName;
	    err += "'";
	    // This exception shouldn't be raised anymore, as such
	    // name clashes are now detected directly in the
	    // translator
	    assert(false);
	    break;

	case MinorInvalidValueInheritance:
	    err = string_dup("invalid valuetype inheritance for `");
	    err += scopedName;
	    err += "'";
	    break;

	case MinorInvalidValueModifier:
	    err = string_dup("invalid valuetype modifier for `");
	    err += scopedName;
	    err += "'";
	    break;

	case MinorDuplicateValueInit:
	    err = string_dup("duplicate valuetype initializer in `");
	    err += scopedName;
	    err += "'";
	    break;

	case MinorAbstractValueInit:
	    err = string_dup("abstract valuetype `");
	    err += scopedName;
	    err += "' cannot have initializers";
	    break;

	case MinorDuplicateBaseType:
	    err = string_dup("duplicate base types for `");
	    err += scopedName;
	    err += "'";
	    break;

	case MinorInvalidAbstractInterfaceInheritance:
	    err = string_dup("invalid abstract interface inheritance for `");
	    err += scopedName;
	    err += "'";
	    break;

	case MinorNameRedefinitionInImmediateScope:
	    err = string_dup("invalid redefinition of `");
	    err += scopedName;
	    err += "' in an immediate scope";
	    break;

	case MinorInvalidValueBoxType:
	    err = string_dup("invalid type for valuebox `");
	    err += scopedName;
	    err += "'";
	    break;

	case MinorInvalidLocalInterfaceInheritance:
	    err = string_dup("invalid local interface inheritance for `");
	    err += scopedName;
	    err += "'";
	    break;
	}
    }
    else if(BAD_INV_ORDER::_downcast(&ex) != 0)
    {
        switch(ex.minor())
        {
        case MinorDependencyPreventsDestruction:
            err = string_dup("cannot destroy object due to dependency");
            break;
	    
        case MinorIndestructibleObject:
            err = string_dup("destroy invoked on indestructible object");
            break;
        }
    }
    else if(INTF_REPOS::_downcast(&ex) != 0)
    {
        switch(ex.minor())
        {
        case MinorNoIntfRepos:
            err = string_dup("interface repository is not available");
            break;

        case MinorLookupAmbiguous:
            err = string_dup("`");
            err += scopedName;
            err += "' is ambiguous";
            break;
            
        case MinorIllegalRecursion:
            err = string_dup("illegal recursion in `");
            err += scopedName;
            err += "'";
            break;
        }
    }
    else
        ex._raise();

    IdlError(err);
}

// ----------------------------------------------------------------------
// Determine if an identifier represents an IDL keyword
// ----------------------------------------------------------------------

bool
IdlIsKeyword(const char* ident)
{
    static const char* kwds[] =
    {
        "abstract", "any", "attribute", "boolean", "case", "char", "const",
        "context", "custom", "default", "double", "enum", "exception",
        "factory", "FALSE", "fixed", "float", "in", "inout", "interface",
        "long", "module", "native", "Object", "octet", "oneway", "out",
        "private", "public", "raises", "readonly", "sequence", "short",
        "string", "struct", "supports", "switch", "TRUE", "truncatable",
        "typedef", "unsigned", "union", "ValueBase", "valuetype", "void",
        "wchar", "wstring",
        0
    };

    for(ULong i = 0 ; kwds[i] != 0 ; i++)
    {
        if(strcasecmp(ident, kwds[i]) == 0)
            return true;
    }

    return false;
}

// ----------------------------------------------------------------------
// Remove escapes from a scoped name
// ----------------------------------------------------------------------

char*
IdlUnescapeName(const char* name)
{
    char* result = string_alloc(strlen(name));
    char* p = result;
    bool checkEscape = true;

    while(*name)
    {
        if(*name == '_' && checkEscape)
        {
            checkEscape = false;
        }
        else if(*name == ':')
        {
            *p++ = *name;
            checkEscape = true;
        }
        else
        {
            *p++ = *name;
            checkEscape = false;
        }

        name++;
    }

    *p = '\0';
    return result;
}
