// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/Properties.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <Util.h>
#include <GenUtil.h>
#include <GenTextBase.h>
#include <GenHTML.h>
#include <IFR3_impl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>

#include <CPPUtil.h>

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] idl-files ...\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show ORBacus version.\n"
	"-d, --debug            Print diagnostic messages.\n"
	"-DNAME                 Define NAME as 1.\n"
	"-DNAME=DEF             Define NAME as DEF.\n"
	"-UNAME                 Remove any definition for NAME.\n"
	"-IDIR                  Put DIR in the include file search path.\n"
        "--all                  Generate code for included files.\n"
	"--no-sort              Don't sort symbols alphabetically.\n"
	"--ignore-case          Sort case-insensitive.\n"
	"--use-tables           Output indices using tables.\n"
	"--output-dir DIR       Write .html files to directory DIR.\n"
	"--alt-indent           Use alternative indentation of arguments.\n"
        "--case-sensitive       Allow case-sensitive identifiers.\n"
        "                       (Not CORBA compliant.)\n"
	;
}

int
run(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Get options
    //
    String_var cppArgs;
    bool all = false;
    bool noSort = false;
    bool altIndent = false;
    bool ignoreCase = false;
    bool useTables = false;
    String_var outDir = string_dup("");
    bool caseSensitive = false;

    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
	if(strcmp(argv[i], "--help") == 0 ||
	   strcmp(argv[i], "-h") == 0)
	{
	    usage(progName);
	    return 0;
	}
	else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
	{
	    OB_CERR << "ORBacus " << Version << OB_ENDL;
	    return 0;
	}
	else if(strcmp(argv[i], "--debug") == 0 ||
		strcmp(argv[i], "-d") == 0)
	{
	    extern int yydebug;
	    yydebug = 1;
	}
	else if(strncmp(argv[i], "-D", 2) == 0 ||
		strncmp(argv[i], "-U", 2) == 0 ||
		strncmp(argv[i], "-I", 2) == 0)
	{
	    if(strlen(argv[i]) == 2)
	    {
		OB_CERR << progName << ": argument expected for `"
			<< argv[i] << "'"
			<< OB_ENDL; // not argv[i - 1] !
		return 1;
	    }

	    cppArgs += ' ';
	    cppArgs += argv[i];
	}
        else if(strcmp(argv[i], "--all") == 0)
        {
            all = true;
        }
        else if(strcmp(argv[i], "--no-sort") == 0)
        {
            noSort = true;
        }
	else if(strcmp(argv[i], "--ignore-case") == 0)
        {
            ignoreCase = true;
        }
	else if(strcmp(argv[i], "--use-tables") == 0)
        {
            useTables = true;
        }
	else if(strcmp(argv[i], "--output-dir") == 0)
	{
	    if(++i < argc)
		outDir = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
			<< argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--alt-indent") == 0)
	{
	    altIndent = true;
	}
	else if(strcmp(argv[i], "--case-sensitive") == 0)
        {
            OB_CERR << progName << ": Warning: use of the --case-sensitive "
                    << "option is not CORBA compliant!" << OB_ENDL;
            caseSensitive = true;
        }
	else
	{
	    OB_CERR << progName << ": unknown option `"
                    << argv[i] << "'\n" << OB_ENDL;
	    usage(progName);
	    return 1;
	}
    }

    if(i == argc)
    {
	OB_CERR << progName << ": no input file" << OB_ENDL;
	usage(progName);
	return 1;
    }

    char* idlcpp = IdlGetCPreprocessor(progName);
    if(!idlcpp)
	return 1;
    String_var cppCmd = idlcpp;

    //
    // Get the RootPOA
    //
    PortableServer::POA_var root;
    try
    {
        CORBA::Object_var poaObj =
            orb -> resolve_initial_references("RootPOA");
        root = PortableServer::POA::_narrow(poaObj);
    }
    catch(const CORBA::ORB::InvalidName&)
    {
    }
    assert(!is_nil(root));

    PortableServer::POAManager_var manager = root -> the_POAManager();

    //
    // Create POA for the Interface Repository
    //
    PolicyList policies;
    policies.length(2);
    policies[0] = root -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] = root -> create_lifespan_policy(PortableServer::PERSISTENT);

    PortableServer::POA_var poa;
    try
    {
        poa = root -> create_POA("InterfaceRepositoryPOA", manager, policies);
    }
    catch(const PortableServer::POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(poa));

    //
    // Activate the POA manager
    //
    manager -> activate();

    int status = 0;
    while(status == 0 && i < argc)
    {
	String_var file = string_dup(argv[i++]);
	
	//
	// Remove .idl or .IDL from idl file name
	//
	String_var base = file;
	const char* suffix = strrchr(base, '.');
	if(suffix)
	{
	    if(strcmp(suffix, ".idl") == 0 || strcmp(suffix, ".IDL") == 0)
	    {
		char* newBase = string_alloc(suffix - (const char*)base);
		strncpy(newBase, base, suffix - (const char*)base);
		newBase[suffix - (const char*)base] = '\0';
		base = newBase;
	    }
            else
            {
                OB_CERR << progName << ": Warning: Using an IDL file without "
                        << ".idl extension is not CORBA compliant!" << OB_ENDL;

            }
	}
	
	//
	// Check for input file
	//
	struct stat statBuf;
	if(stat(file, &statBuf) == -1)
	{
	    OB_CERR << progName << ": couldn't open " << file << OB_ENDL;
	    return 1;
	}

	//
	// Open input file
	//
	OB_TRANSLATOR_IMPORT extern FILE* yyin;
	String_var cmd = cppCmd;
	cmd += cppArgs;
	cmd += ' ';
	cmd += file;

	OBProcessIO pio;
	if((yyin = pio.open(cmd, "r")) == NULL)
	    return 1;

	//
	// Create implementation object
	//
	Repository_impl* repositoryImpl = new Repository_impl(orb, poa);
        ObjectId_var repositoryID = string_to_ObjectId("DefaultRepository");
        poa -> activate_object_with_id(repositoryID, repositoryImpl);
	Repository_var repository = repositoryImpl -> _this();
	assert(!is_nil(repository));

	//
	// Do parsing
	//
	int IdlParse(ORB_ptr, Repository_ptr, StringSeq&,
		     bool, bool, bool, bool, bool, bool, bool,
		     IdlStringHash&, IdlStringHash&);
	StringSeq includeSeq;
	IdlStringHash idHash;
	IdlStringHash fwdHash;
	status = IdlParse(orb, repository, includeSeq, all, true,
                          caseSensitive, true, true, true, true,
                          idHash, fwdHash);
	
	//
	// Close input file
	//
	pio.close();
	
	if(status == 0)
	{
	    //
	    // Generate HTML code
	    //
#ifdef WIN32
	    const char* name1 = strrchr(base, '\\');
	    const char* name2 = strrchr(base, '/');
	    const char* name = name1 > name2 ? name1 : name2;
#else
	    const char* name = strrchr(base, '/');
#endif
	    if(name)
		name++;
	    else
		name = base;

	    IdlHTMLGenerator generator(progName, repository, all, idHash,
				       outDir, noSort, ignoreCase, useTables,
				       altIndent);

	    generator.gen();
	}

	repositoryImpl -> destroyRepository();
	repositoryImpl -> _remove_ref();
    }

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;

    try
    {
	//
	// Set ooc.iiop.host to localhost, to avoid DNS lookups
	//
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);
	properties -> setProperty("ooc.iiop.host", "127.0.0.1");

	//
	// Run in reactive/reactive concurrency model.
	//
        properties -> setProperty("ooc.orb.conc_model", "reactive");
	properties -> setProperty("ooc.orb.oa.conc_model", "reactive");

	//
	// Initialize the ORB
	//
	orb = ORB_init(argc, argv);

	//
	// Generate code
	//
	status = run(orb, argc, argv);
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
