// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Hashtable.h>

#include <Util.h>
#include <IFR2_impl.h>
#include <IFR3_impl.h>
#include <IFR4_impl.h>
#ifdef HAVE_STRINGS_H
#   include <strings.h>
#endif

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

// ----------------------------------------------------------------------
// Contained_impl member implementations
// ----------------------------------------------------------------------

Contained_impl::Contained_impl(ORB_ptr orb,
                               POA_ptr poa,
			       Container_impl* c,
			       Repository_impl* r)
    : IRObject_impl(orb, poa, r),
      definedIn_(c)
{
    definedIn_ -> _add_ref();
}

Contained_impl::~Contained_impl()
{
}

void
Contained_impl::destroy()
    throw(SystemException)
{
    if(id_.in())
	repository_ -> remove(this);
    id_ = (const char*)"";

    IRObject_impl::destroy();

    //
    // Calling definedIn_ -> remove(this) may delete this object
    //
    Container_impl* c = definedIn_;
    definedIn_ = 0;
    c -> remove(this);
    c -> _remove_ref();
}

char*
Contained_impl::id()
    throw(SystemException)
{
    if(!id_.in())
	throw BAD_INV_ORDER();

    return string_dup(id_);
}

void
Contained_impl::id(const char* s)
    throw(SystemException)
{
    if(!s || strlen(s) == 0)
	throw BAD_PARAM(); // TODO: Is this exception correct?

    if(id_.in() && strcmp(id_, s) == 0)
	return;

    //
    // Find out whether id already exists.
    //
    Contained_var con = repository_ -> lookup_id(s);
    if(!is_nil(con))
	throw BAD_PARAM(MinorRepositoryIdExists, COMPLETED_NO);

    if(id_.in())
	repository_ -> remove(this);

    id_ = s;

    if(id_.in())
	repository_ -> add(this);
}

char*
Contained_impl::name()
    throw(SystemException)
{
    if(!name_.in())
	throw BAD_INV_ORDER();

    return string_dup(name_);
}

void
Contained_impl::name(const char* s)
    throw(SystemException)
{
    if(!s || strlen(s) == 0)
	throw BAD_PARAM(); // TODO: Is this exception correct?

    if(name_.in() && strcmp(name_, s) == 0)
	return;

    //
    // Check for name redefinition in an immediate scope
    //
    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);
    if(contained)
    {
	Identifier_var containedName = contained -> name();
	if(IdlStringCompare(s, containedName) == 0)
	    throw BAD_PARAM(MinorNameRedefinitionInImmediateScope,
			    COMPLETED_NO);
    }

    //
    // Find out whether name already exists. Exclude inherited names.
    //
    definedIn_ -> validateName(s, true, false);

    name_ = s;
}

char*
Contained_impl::version()
    throw(SystemException)
{
    if(!version_.in())
	throw BAD_INV_ORDER();

    return string_dup(version_);
}

void
Contained_impl::version(const char* s)
    throw(SystemException)
{
    if(!s || strlen(s) == 0)
	throw BAD_PARAM(); // TODO: Is this exception correct?

    version_ = s;
}

Container_ptr
Contained_impl::defined_in()
    throw(SystemException)
{
    return definedIn_ -> _this();
}

char*
Contained_impl::absolute_name()
    throw(SystemException)
{
    if(!name_.in())
	throw BAD_INV_ORDER();

    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);

    if(contained)
    {
	String_var s = contained -> absolute_name();
	char* result = string_alloc(strlen(s) + 2 + strlen(name_));
	strcpy(result, s);
	strcat(result, "::");
	strcat(result, name_);
	return result;
    }
    else
    {
	char* result = string_alloc(2 + strlen(name_));
	strcpy(result, "::");
	strcat(result, name_);
	return result;
    }
}

Repository_ptr
Contained_impl::containing_repository()
    throw(SystemException)
{
    return repository_ -> _this();
}

void
Contained_impl::move(Container_ptr new_container,
		     const char* identifier,
		     const char* new_version)
    throw(SystemException)
{
    if(is_nil(new_container))
	throw BAD_PARAM(); // TODO: Is this exception correct?

    POA_var poa = _default_POA();
    Container_impl* newContainer = 0;
    try
    {
        new_container -> _non_existent();
	ServantBase_var servant = poa -> reference_to_servant(new_container);
	newContainer = dynamic_cast<Container_impl*>(servant.in());
    }
    catch(const POA::WrongAdapter&)
    {
    }
    catch(const POA::WrongPolicy&)
    {
    }
    catch(const POA::ObjectNotActive&)
    {
    }
    if(!newContainer)
    {
	// Must be same repository
	throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
    }

    // TODO: BAD_PARAM(MinorInvalidContainer, COMPLETED_NO) must also
    // be raised in case the container is invalid, for example, if an
    // operation is moved into a container for a module.

    //
    // Remember the old name, the old version, and the old container
    //
    Identifier_var oldName = name();
    VersionSpec_var oldVersion = version();
    Container_impl* oldContainer = definedIn_;
    oldContainer -> _add_ref();
    ServantBase_var dummy = oldContainer;

    //
    // Remove from old container
    //
    definedIn_ -> remove(this);
    definedIn_ -> _remove_ref();

    //
    // Add to new container
    //
    definedIn_ = newContainer;
    definedIn_ -> _add_ref();
    try
    {
	definedIn_ -> add(this);
    }
    catch(...)
    {
	definedIn_ -> _remove_ref();
	definedIn_ = oldContainer;
	definedIn_ -> _add_ref();
	definedIn_ -> add(this);
	throw;
    }

    //
    // Set the new name and version
    //
    try
    {
	name(identifier);
	version(new_version);
    }
    catch(...)
    {
	definedIn_ -> remove(this);
	definedIn_ -> _remove_ref();
	definedIn_ = oldContainer;
	definedIn_ -> _add_ref();
	definedIn_ -> add(this);
	name(oldName);
	version(oldVersion);
	throw;
    }
}

// ----------------------------------------------------------------------
// Container_impl member implementations
// ----------------------------------------------------------------------

void
Container_impl::activateImpl(const char* id, Contained_impl* cont)
    throw(SystemException)
{
    ObjectId_var oid = string_to_ObjectId(id);

    containedImplList_.push_back(cont);

    try
    {
	poa_ -> activate_object_with_id(oid, cont);
    }
    catch(const PortableServer::POA::ObjectAlreadyActive&)
    {
	cont -> destroy();
	throw BAD_PARAM(MinorRepositoryIdExists, COMPLETED_NO);
    }
}

void
Container_impl::add(Contained_impl* cont)
{
#ifndef NDEBUG
    OB_STL::list<Contained_impl*>::iterator containedIter;
    for(containedIter = containedImplList_.begin();
	containedIter != containedImplList_.end();
	containedIter++)
    {
	assert(*containedIter != cont);
    }
#endif

    containedImplList_.push_back(cont);
    cont -> _add_ref();
}

void
Container_impl::remove(Contained_impl* cont)
{
    OB_STL::list<Contained_impl*>::iterator iter;
    for(iter = containedImplList_.begin();
	iter != containedImplList_.end();
	iter++)
    {
	if(*iter == cont)
	{
	    containedImplList_.erase(iter);
	    cont -> _remove_ref();
	    return;
	}	 
    }

    assert(false);
}

void
Container_impl::mergeContainedSeq(ContainedSeq& seq1, const ContainedSeq& seq2)
{
    ULong i;

    RepositoryIdSeq repIdSeq(seq1.length());
    repIdSeq.length(seq1.length());
    
    for(i = 0 ; i < seq1.length() ; i++)
        repIdSeq[i] = seq1[i] -> id();

    for(i = 0 ; i < seq2.length() ; i++)
    {
        RepositoryId_var id = seq2[i] -> id();

	ULong j;
        for(j = 0 ; j < repIdSeq.length() ; j++)
            if(strcmp(id, repIdSeq[j]) == 0)
		break;

        if(j == repIdSeq.length())
	{
            seq1.length(seq1.length() + 1);
            seq1[seq1.length() - 1] = seq2[i];
            
            repIdSeq.length(repIdSeq.length() + 1);
            repIdSeq[repIdSeq.length() - 1] = id;
	}
    }
}

void
Container_impl::checkIllegalRecursion(const char* id, TypeCode_ptr tc)
{
    TypeCode_var origTC = GetOrigType(tc);
    if(origTC -> kind() == tk_struct || origTC -> kind() == tk_union ||
       origTC -> kind() == tk_except)
    {
        String_var tcId = origTC -> id();
        if(strcmp(id, tcId) == 0)
            throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);

        for(ULong i = 0 ; i < origTC -> member_count() ; i++)
        {
            TypeCode_var memberType = origTC -> member_type(i);
            checkIllegalRecursion(id, memberType);
        }
    }
    else if(origTC -> kind() == tk_array)
    {
        TypeCode_var contentType = origTC -> content_type();
        checkIllegalRecursion(id, contentType);
    }
}

Container_impl::Container_impl(ORB_ptr orb,
                               POA_ptr poa,
                               Repository_impl* r)
    : IRObject_impl(orb, poa, r)
{
}

Container_impl::~Container_impl()
{
}

void
Container_impl::destroy()
    throw(SystemException)
{
    destroyContents();
    IRObject_impl::destroy();
}

void
Container_impl::destroyContents()
    throw(SystemException)
{
    OB_STL::list<Contained_impl*> containedImplListCopy = containedImplList_;
    OB_STL::list<Contained_impl*>::iterator iter;

    for(iter = containedImplListCopy.begin();
	iter != containedImplListCopy.end();
	iter++)
    {
	(*iter) -> destroy();
    }

    assert(containedImplList_.size() == 0);
}

Contained_ptr
Container_impl::lookup(const char* search_name)
    throw(SystemException)
{
    if(strncmp(search_name, "::", 2) == 0) // Absolute name, begining with "::"
    {
	Container_var self = _this();
	Contained_var contained = CORBA::Contained::_narrow(self);

	//
        // Must be contained, since it cannot be a repository (the
        // repository overwrites lookup())
	//
	assert(!is_nil(contained));

	Repository_var repository = contained -> containing_repository();
	return repository -> lookup(search_name);
    }
    else // Relative name, not begining with "::"
    {
	const char* s2 = strstr(search_name, "::");
	String_var s1;

	if(s2)
	{
	    ULong len = s2 - search_name;
	    s1 = string_alloc(len);
	    strncpy(s1.inout(), search_name, len);
	    s1[len] = '\0';
	    s2 += 2;
	}
	else
	    s1 = search_name;

        OB_STL::list<Contained_impl*>::iterator iter;
        for(iter = containedImplList_.begin();
	    iter != containedImplList_.end();
	    iter++)
	{
	    Contained_impl* p = *iter;

	    Identifier_var n = p -> name();
	    if(strcmp(n, s1) == 0)
	    {
		if(s2)
		{
		    Container_impl* container =
			dynamic_cast<Container_impl*>(p);
		    
		    if(container)
			return container -> lookup(s2);
		    else
			return Contained::_nil();
		}
		else
		    return p -> _this();
	    }
	}

	return Contained::_nil();
    }
}

ContainedSeq*
Container_impl::contents(DefinitionKind limit_type,
			 Boolean /*exclude_inherited*/)
    throw(SystemException)
{
    ContainedSeq_var s = new ContainedSeq;

    OB_STL::list<Contained_impl*>::iterator iter;
    for(iter = containedImplList_.begin();
	iter != containedImplList_.end();
	iter++)
    {
	if(limit_type == dk_all ||
	   limit_type == (*iter) -> def_kind())
	{
            s -> length(s -> length() + 1);
            s[s -> length() - 1] = (*iter) -> _this();
	}
    }
    
    return s._retn();
}

void
Container_impl::validateName(const char* name,
                             bool /*exclude_inherited*/,
                             bool in_inherited)
{
    bool found = false;

    OB_STL::list<Contained_impl*>::iterator iter;
    for(iter = containedImplList_.begin();
	!found && iter != containedImplList_.end();
	iter++)
    {
        try
        {
            DefinitionKind kind = (*iter) -> def_kind();
            bool isType = (kind == dk_Constant || kind == dk_Exception ||
                           kind == dk_Alias || kind == dk_Struct ||
                           kind == dk_Union || kind == dk_Enum ||
                           kind == dk_Native);

            //
            // Type names can be redefined by subtypes
            //
            if(in_inherited && isType)
                continue;

            Identifier_var name2 = (*iter) -> name();
            if(IdlStringCompare(name, name2) == 0)
                found = true;
            else if(kind == dk_Enum)
            {
                EnumDef_impl* enumDef =
                    dynamic_cast<EnumDef_impl*>(*iter);
                EnumMemberSeq_var members = enumDef -> members();
                for(ULong j = 0 ; j < members -> length() ; j++)
                    if(IdlStringCompare(name, members[j]) == 0)
                        found = true;
            }
        }
        catch(const BAD_INV_ORDER&)
        {
        }
    }

    if(found)
    {
        if(in_inherited)
	    throw BAD_PARAM(MinorNameClashInInheritedContext, COMPLETED_NO);
        else
            throw BAD_PARAM(MinorNameExists, COMPLETED_NO);
    }
}

ContainedSeq*
Container_impl::lookup_name(const char* search_name,
			    Long levels_to_search,
			    DefinitionKind limit_type,
			    Boolean exclude_inherited)
    throw(SystemException)
{
    ContainedSeq_var s = new ContainedSeq;

    if(levels_to_search != 0)
    {
	levels_to_search--;

        OB_STL::list<Contained_impl*>::iterator iter;
        for(iter = containedImplList_.begin();
	    iter != containedImplList_.end();
	    iter++)
	{
	    Contained_impl* p = *iter;

	    if(limit_type == dk_all || limit_type == p -> def_kind())
	    {
		Identifier_var name = p -> name();
		if(strcmp(search_name, name) == 0)
                {
                    s -> length(s -> length() + 1);
                    s[s -> length() - 1] = p -> _this();
                }

		if(levels_to_search != 0)
		{
		    Container_impl* container =
			dynamic_cast<Container_impl*>(p);
		    
		    if(container)
		    {
			ContainedSeq_var s2 =
			    container -> lookup_name(search_name,
						     levels_to_search,
						     limit_type,
						     exclude_inherited);
			mergeContainedSeq(s.inout(), s2);
		    }
		}
	    }
	}
    }
		
    return s._retn();
}

CORBA::Container::DescriptionSeq*
Container_impl::describe_contents(DefinitionKind limit_type,
				  Boolean exclude_inherited,
				  Long max_returned_objs)
    throw(SystemException)
{
    // TODO: This operation should be optimized. It should make use of
    // containedImplList_ directly.

    ContainedSeq_var cs = contents(limit_type, exclude_inherited);

    ULong len = cs -> length();

    if(max_returned_objs >= 0 && len > (ULong)max_returned_objs)
	len = (ULong)max_returned_objs;

    CORBA::Container::DescriptionSeq_var s =
	new CORBA::Container::DescriptionSeq;

    for(ULong i = 0 ; i < len ; i++)
    {
	if(limit_type == dk_all ||
	   limit_type == cs[i] -> def_kind())
	{
	    CORBA::Contained::Description_var description =
                cs[i] -> describe();
	    
	    CORBA::Container::Description desc;
	    desc.contained_object = cs[i];
	    desc.kind = description -> kind;
	    desc.value = description -> value;
	    
            s -> length(s -> length() + 1);
            s[s -> length() - 1] = desc;
	}
    }
    
    return s._retn();
}

ModuleDef_ptr
Container_impl::create_module(const char* id,
			      const char* name,
			      const char* version)
    throw(SystemException)
{
    ModuleDef_impl* p = new ModuleDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    ModuleDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
    }
    catch(...)
    {
	p -> destroy();
        throw;
    }

    return ref._retn();
}

ConstantDef_ptr
Container_impl::create_constant(const char* id,
			        const char* name,
			        const char* version,
			        IDLType_ptr type,
			        const Any& value)
    throw(SystemException)
{
    ConstantDef_impl* p = new ConstantDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    ConstantDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> type_def(type);
        p -> value(value);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

StructDef_ptr
Container_impl::create_struct(const char* id,
			      const char* name,
			      const char* version,
			      const StructMemberSeq& members)
    throw(SystemException)
{
    StructDef_impl* p = new StructDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    StructDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> members(members);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

UnionDef_ptr
Container_impl::create_union(const char* id,
			     const char* name,
			     const char* version,
			     IDLType_ptr discriminator_type,
			     const UnionMemberSeq& members)
    throw(SystemException)
{
    UnionDef_impl* p = new UnionDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    UnionDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> discriminator_type_def(discriminator_type);
        p -> members(members);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }
        
    return ref._retn();
}

EnumDef_ptr
Container_impl::create_enum(const char* id,
			    const char* name,
			    const char* version,
			    const EnumMemberSeq& members)
    throw(SystemException)
{
    EnumDef_impl* p = new EnumDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    EnumDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> members(members);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }
        
    return ref._retn();
}

AliasDef_ptr
Container_impl::create_alias(const char* id,
			     const char* name,
			     const char* version,
			     IDLType_ptr original_type)
    throw(SystemException)
{
    AliasDef_impl* p = new AliasDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    AliasDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> original_type_def(original_type);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }
    
    return ref._retn();
}

InterfaceDef_ptr
Container_impl::create_interface(const char* id,
				 const char* name,
				 const char* version,
				 const InterfaceDefSeq&
				 base_interfaces)
    throw(SystemException)
{
    InterfaceDef_impl* p = new InterfaceDef_impl(orb_, poa_, this,
                                                 repository_);

    activateImpl(id, p);

    InterfaceDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> base_interfaces(base_interfaces);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

AbstractInterfaceDef_ptr
Container_impl::create_abstract_interface(const char* id,
                                          const char* name,
                                          const char* version,
                                          const AbstractInterfaceDefSeq&
                                          base_interfaces)
    throw(SystemException)
{
    AbstractInterfaceDef_impl* p =
        new AbstractInterfaceDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    AbstractInterfaceDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);

        InterfaceDefSeq seq;
        seq.length(base_interfaces.length());
        for(ULong i = 0 ; i < base_interfaces.length() ; i++)
            seq[i] = InterfaceDef::_duplicate(base_interfaces[i]);
        p -> base_interfaces(seq);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

LocalInterfaceDef_ptr
Container_impl::create_local_interface(const char* id,
				       const char* name,
				       const char* version,
				       const InterfaceDefSeq& base_interfaces)
    throw(SystemException)
{
    LocalInterfaceDef_impl* p =
        new LocalInterfaceDef_impl(orb_, poa_, this, repository_);
    containedImplList_.push_back(p);
    ObjectId_var oid = string_to_ObjectId(id);
    poa_ -> activate_object_with_id(oid, p);
    LocalInterfaceDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> base_interfaces(base_interfaces);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

ValueDef_ptr
Container_impl::create_value(const char* id,
			     const char* name,
			     const char* version,
			     Boolean is_custom,
			     Boolean is_abstract,
			     ValueDef_ptr base_value,
			     Boolean is_truncatable,
			     const ValueDefSeq&
			     abstract_base_values,
			     const InterfaceDefSeq& supported_interfaces,
			     const InitializerSeq& initializers)
    throw(SystemException)
{
    ValueDef_impl* p = 
	new ValueDef_impl(
	    orb_, 
	    poa_, 
	    this, 
	    repository_,
	    is_abstract,
	    is_custom,
	    is_truncatable);

    activateImpl(id, p);

    ValueDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> base_value(base_value);
        p -> abstract_base_values(abstract_base_values);
        p -> supported_interfaces(supported_interfaces);
        p -> is_custom(is_custom);
        p -> is_abstract(is_abstract);
        p -> is_truncatable(is_truncatable);
        p -> initializers(initializers);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

ValueBoxDef_ptr
Container_impl::create_value_box(const char* id,
				 const char* name,
				 const char* version,
				 IDLType_ptr original_type_def)
    throw(SystemException)
{
    ValueBoxDef_impl* p = new ValueBoxDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    ValueBoxDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> original_type_def(original_type_def);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

ExceptionDef_ptr
Container_impl::create_exception(const char* id,
				 const char* name,
				 const char* version,
				 const StructMemberSeq& members)
    throw(SystemException)
{
    ExceptionDef_impl* p =
	new ExceptionDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    ExceptionDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> members(members);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

NativeDef_ptr
Container_impl::create_native(const char* id,
			      const char* name,
			      const char* version)
    throw(SystemException)
{
    NativeDef_impl* p = new NativeDef_impl(orb_, poa_, this, repository_);

    activateImpl(id, p);

    NativeDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

// ----------------------------------------------------------------------
// ExceptionDef_impl member implementations
// ----------------------------------------------------------------------

ExceptionDef_impl::ExceptionDef_impl(ORB_ptr orb,
                                     POA_ptr poa,
                                     Container_impl* c,
				     Repository_impl* r)
    : IRObject_impl(orb, poa, r),
      Container_impl(orb, poa, r),
      Contained_impl(orb, poa, c, r)
{
}

ExceptionDef_impl::~ExceptionDef_impl()
{
}

void
ExceptionDef_impl::destroy()
    throw(SystemException)
{
    for(ULong i = 0 ; i < membersIS_.length() ; i++)
	membersIS_[i] -> _remove_ref();

    membersIS_.length(0);
    members_.length(0);

    destroyContents();
    Contained_impl::destroy();
}

void
ExceptionDef_impl::validateName(const char* name,
                                bool exclude_inherited,
                                bool in_inherited)
{
    Container_impl::validateName(name, exclude_inherited, in_inherited);
    
    for(ULong i = 0 ; i < membersIS_.length() ; i++)
	if(IdlStringCompare(members_[i].name, name) == 0)
	    throw BAD_PARAM(MinorNameExists, COMPLETED_NO);
}

CORBA::Contained::Description*
ExceptionDef_impl::describe()
    throw(SystemException)
{
    ExceptionDescription_var desc = new ExceptionDescription;
    
    desc -> name = name();
    desc -> id = id();
    
    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);
    if(contained)
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> type = type();

    CORBA::Contained::Description_var description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc._retn();
    
    return description._retn();
}

TypeCode_ptr
ExceptionDef_impl::type()
    throw(SystemException)
{
    ContainedIS dummy;
    TypeCode_var tc = typeRecursive(dummy);
    assert(dummy.length() == 0);
    return tc._retn();
}

TypeCode_ptr
ExceptionDef_impl::typeRecursive(ContainedIS& seq)
    throw(SystemException)
{
    seq.length(seq.length() + 1);
    seq[seq.length() - 1] = this;
    
    //
    // Recalculate TypeCodes every time, as ids may have changed, and
    // also in order to calculate recursive TypeCodes correctly.
    //
    for(ULong i = 0 ; i < membersIS_.length() ; i++)
	members_[i].type = membersIS_[i] -> typeRecursive(seq);

    seq.length(seq.length() - 1);
    return orb_ -> create_exception_tc(id_, name_, members_);
}

StructMemberSeq*
ExceptionDef_impl::members()
    throw(SystemException)
{
    //
    // Recalculate TypeCodes every time, as ids may have changed, and
    // also in order to calculate recursive TypeCodes correctly.
    //
    for(ULong i = 0 ; i < membersIS_.length() ; i++)
    {
	ContainedIS dummy;
	members_[i].type = membersIS_[i] -> typeRecursive(dummy);
	assert(dummy.length() == 0);
    }

    return new StructMemberSeq(members_);
}

void
ExceptionDef_impl::members(const StructMemberSeq& seq)
    throw(SystemException)
{
    ULong i;

    //
    // Ensure we don't have duplicate declarator names
    //
    for(i = 0 ; i < seq.length() ; i++)
	for(ULong j = i + 1 ; j < seq.length() ; j++)
	    if(IdlStringCompare(seq[i].name, seq[j].name) == 0)
		throw BAD_PARAM(MinorDuplicateDeclarator, COMPLETED_NO);

    IDLTypeIS implSeq;
    implSeq.length(seq.length());
    POA_var poa = _default_POA();
    for(i = 0 ; i < seq.length() ; i++)
    {
        if(strlen(seq[i].name) == 0)
	    throw BAD_PARAM(); // TODO: Is this exception correct?

	//
	// Find out whether name already exists. Note:
	// Container_impl::validateName() must be used, so that the
	// current member names are not take into account.
	//
	Container_impl::validateName(seq[i].name, false, false);

	//
	// Check for name redefinition in an immediate scope
	//
	if(IdlStringCompare(seq[i].name, name_) == 0)
	    throw BAD_PARAM(MinorNameRedefinitionInImmediateScope,
			    COMPLETED_NO);

	if(is_nil(seq[i].type_def))
	    throw BAD_PARAM(); // TODO: Is this exception correct?

	implSeq[i] = 0;
	try
	{
	    ServantBase_var servant =
		poa -> reference_to_servant(seq[i].type_def);
	    implSeq[i] = dynamic_cast<IDLType_impl*>(servant.in());
	}
	catch(const POA::WrongAdapter&)
	{
	}
	catch(const POA::WrongPolicy&)
	{
	}
	catch(const POA::ObjectNotActive&)
	{
	}
	if(!implSeq[i])
	{
	    // Must be same repository
	    throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
	}
    }

    //
    // Check for illegal recursion
    //
    for(i = 0 ; i < seq.length() ; i++)
        checkIllegalRecursion(id_, seq[i].type);

    for(i = 0 ; i < implSeq.length() ; i++)
	implSeq[i] -> _add_ref();
    
    for(i = 0 ; i < membersIS_.length() ; i++)
	membersIS_[i] -> _remove_ref();

    membersIS_ = implSeq;
    members_ = seq;
}

// ----------------------------------------------------------------------
// AttributeDef_impl member implementations
// ----------------------------------------------------------------------

AttributeDef_impl::AttributeDef_impl(ORB_ptr orb,
                                     POA_ptr poa,
                                     Container_impl* c,
				     Repository_impl* r)
    : IRObject_impl(orb, poa, r),
      Contained_impl(orb, poa, c, r),
      typeI_(0),
      attributeMode_(ATTR_NORMAL)
{
}

AttributeDef_impl::~AttributeDef_impl()
{
}

void
AttributeDef_impl::destroy()
    throw(SystemException)
{
    if(typeI_)
    {
	typeI_ -> _remove_ref();
	typeI_ = 0;
    }
    
    Contained_impl::destroy();
}

void
AttributeDef_impl::name(const char* s)
    throw(SystemException)
{
    if(name_.in() && strcmp(name_, s) == 0)
	return;

    //
    // Find out whether attribute name already exists. Do NOT exclude
    // inherited names.
    //
    definedIn_ -> validateName(s, false, false);

    Contained_impl::name(s);
}

CORBA::Contained::Description*
AttributeDef_impl::describe()
    throw(SystemException)
{
    AttributeDescription_var desc = new AttributeDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);
    if(contained)
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> type = type();
    desc -> mode = mode();

    CORBA::Contained::Description_var description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc._retn();

    return description._retn();
}

TypeCode_ptr
AttributeDef_impl::type()
    throw(SystemException)
{
    if(!typeI_)
        throw BAD_INV_ORDER();

    return typeI_ -> type();
}

IDLType_ptr
AttributeDef_impl::type_def()
    throw(SystemException)
{
    if(!typeI_)
        throw BAD_INV_ORDER();

    return typeI_ -> _this();
}

void
AttributeDef_impl::type_def(IDLType_ptr type)
    throw(SystemException)
{
    if(is_nil(type))
	throw BAD_PARAM(); // TODO: Is this exception correct?

    POA_var poa = _default_POA();
    IDLType_impl* impl = 0;
    try
    {
	ServantBase_var servant = poa -> reference_to_servant(type);
	impl = dynamic_cast<IDLType_impl*>(servant.in());
    }
    catch(const POA::WrongAdapter&)
    {
    }
    catch(const POA::WrongPolicy&)
    {
    }
    catch(const POA::ObjectNotActive&)
    {
    }
    if(!impl)
    {
	// Must be same repository
	throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
    }
    
    impl -> _add_ref();
    if(typeI_)
	typeI_ -> _remove_ref();
    typeI_ = impl;
}

AttributeMode
AttributeDef_impl::mode()
    throw(SystemException)
{
    return attributeMode_;
}

void
AttributeDef_impl::mode(AttributeMode mode)
    throw(SystemException)
{
    attributeMode_ = mode;
}

// ----------------------------------------------------------------------
// OperationDef_impl member implementations
// ----------------------------------------------------------------------

OperationDef_impl::OperationDef_impl(ORB_ptr orb,
                                     POA_ptr poa,
                                     Container_impl* c,
				     Repository_impl* r)
    : IRObject_impl(orb, poa, r),
      Contained_impl(orb, poa, c, r),
      resultI_(0),
      mode_(OP_NORMAL)
      
{
}

OperationDef_impl::~OperationDef_impl()
{
}

void
OperationDef_impl::destroy()
    throw(SystemException)
{
    ULong i;

    if(resultI_)
    {
	resultI_ -> _remove_ref();
	resultI_ = 0;
    }

    for(i = 0 ; i < paramsIS_.length() ; i++)
	paramsIS_[i] -> _remove_ref();

    paramsIS_.length(0);
    params_.length(0);
    
    for(i = 0 ; i < exceptionsIS_.length() ; i++)
	exceptionsIS_[i] -> _remove_ref();
    exceptionsIS_.length(0);

    Contained_impl::destroy();
}

void
OperationDef_impl::name(const char* s)
    throw(SystemException)
{
    if(name_.in() && strcmp(name_, s) == 0)
	return;

    //
    // Find out whether operation name already exists. Do NOT exclude
    // inherited names.
    //
    definedIn_ -> validateName(s, false, false);

    Contained_impl::name(s);
}

CORBA::Contained::Description*
OperationDef_impl::describe()
    throw(SystemException)
{
    ULong i;

    OperationDescription_var desc = new OperationDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);
    if(contained)
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> result = result();
    desc -> mode = mode();
    desc -> contexts = contexts_;

    //
    // Recalculate TypeCodes every time, as ids may have changed, and
    // also in order to calculate recursive TypeCodes correctly.
    //
    for(i = 0 ; i < paramsIS_.length() ; i++)
    {
	ContainedIS dummy;
	params_[i].type = paramsIS_[i] -> typeRecursive(dummy);
	assert(dummy.length() == 0);
    }

    desc -> parameters = params_;

    desc -> exceptions.length(exceptionsIS_.length());
    for(i = 0 ; i < exceptionsIS_.length() ; i++)
    {
	CORBA::Contained::Description_var d = exceptionsIS_[i] -> describe();
	const ExceptionDescription* excDesc;
	Boolean b = d -> value >>= excDesc;
	assert(b);
	desc -> exceptions[i] = *excDesc;
    }

    CORBA::Contained::Description_var description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc._retn();
    
    return description._retn();
}

TypeCode_ptr
OperationDef_impl::result()
    throw(SystemException)
{
    if(!resultI_)
        throw BAD_INV_ORDER();

    return resultI_ -> type();
}

IDLType_ptr
OperationDef_impl::result_def()
    throw(SystemException)
{
    if(!resultI_)
        throw BAD_INV_ORDER();

    return resultI_ -> _this();
}

void
OperationDef_impl::result_def(IDLType_ptr result)
    throw(SystemException)
{
    if(is_nil(result))
	throw BAD_PARAM(); // TODO: Is this exception correct?

    POA_var poa = _default_POA();
    IDLType_impl* impl = 0;
    try
    {
	ServantBase_var servant = poa -> reference_to_servant(result);
	impl = dynamic_cast<IDLType_impl*>(servant.in());
    }
    catch(const POA::WrongAdapter&)
    {
    }
    catch(const POA::WrongPolicy&)
    {
    }
    catch(const POA::ObjectNotActive&)
    {
    }
    if(!impl)
    {
	// Must be same repository
	throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
    }
    
    impl -> _add_ref();
    if(resultI_)
	resultI_ -> _remove_ref();
    resultI_ = impl;
}

ParDescriptionSeq*
OperationDef_impl::params()
    throw(SystemException)
{
    //
    // Recalculate TypeCodes every time, as ids may have changed, and
    // also in order to calculate recursive TypeCodes correctly.
    //
    for(ULong i = 0 ; i < paramsIS_.length() ; i++)
    {
	ContainedIS dummy;
	params_[i].type = paramsIS_[i] -> typeRecursive(dummy);
	assert(dummy.length() == 0);
    }

    return new ParDescriptionSeq(params_);
}

void
OperationDef_impl::params(const ParDescriptionSeq& seq)
    throw(SystemException)
{
    ULong i;

    //
    // Ensure we don't have duplicate declarator names
    //
    for(i = 0 ; i < seq.length() ; i++)
	for(ULong j = i + 1 ; j < seq.length() ; j++)
	    if(IdlStringCompare(seq[i].name, seq[j].name) == 0)
		throw BAD_PARAM(MinorDuplicateDeclarator, COMPLETED_NO);

    IDLTypeIS implSeq;
    implSeq.length(seq.length());
    POA_var poa = _default_POA();
    for(i = 0 ; i < seq.length() ; i++)
    {
	if(is_nil(seq[i].type_def))
	    throw BAD_PARAM(); // TODO: Is this exception correct?
	implSeq[i] = 0;
	try
	{
	    ServantBase_var servant =
		poa -> reference_to_servant(seq[i].type_def);
	    implSeq[i] = dynamic_cast<IDLType_impl*>(servant.in());
	}
	catch(const POA::WrongAdapter&)
	{
	}
	catch(const POA::WrongPolicy&)
	{
	}
	catch(const POA::ObjectNotActive&)
	{
	}
	if(!implSeq[i])
	{
	    // Must be same repository
	    throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
	}
    }
    
    for(i = 0 ; i < implSeq.length() ; i++)
	implSeq[i] -> _add_ref();
    
    for(i = 0 ; i < paramsIS_.length() ; i++)
	paramsIS_[i] -> _remove_ref();

    paramsIS_ = implSeq;
    params_ = seq;
}

OperationMode
OperationDef_impl::mode()
    throw(SystemException)
{
    return mode_;
}

void
OperationDef_impl::mode(OperationMode m)
    throw(SystemException)
{
    mode_ = m;
}

ContextIdSeq*
OperationDef_impl::contexts()
    throw(SystemException)
{
    return new ContextIdSeq(contexts_);
}

void
OperationDef_impl::contexts(const ContextIdSeq& c)
    throw(SystemException)
{
    contexts_ = c;
}

ExceptionDefSeq*
OperationDef_impl::exceptions()
    throw(SystemException)
{
    ExceptionDefSeq_var s = new ExceptionDefSeq;
    s -> length(exceptionsIS_.length());
    for(ULong i = 0 ; i < s -> length() ; i++)
	s[i] = exceptionsIS_[i] -> _this();
    return s._retn();
}

void
OperationDef_impl::exceptions(const ExceptionDefSeq& seq)
    throw(SystemException)
{
    ULong i;

    ExceptionDefIS implSeq;
    implSeq.length(seq.length());
    POA_var poa = _default_POA();
    for(i = 0 ; i < seq.length() ; i++)
    {
	if(is_nil(seq[i]))
	    throw BAD_PARAM(); // TODO: Is this exception correct?
	implSeq[i] = 0;
	try
	{
	    ServantBase_var servant = poa -> reference_to_servant(seq[i]);
	    implSeq[i] = dynamic_cast<ExceptionDef_impl*>(servant.in());
	}
	catch(const POA::WrongAdapter&)
	{
	}
	catch(const POA::WrongPolicy&)
	{
	}
	catch(const POA::ObjectNotActive&)
	{
	}
	if(!implSeq[i])
	{
	    // Must be same repository
	    throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
	}
    }
    
    for(i = 0 ; i < implSeq.length() ; i++)
	implSeq[i] -> _add_ref();
    
    for(i = 0 ; i < exceptionsIS_.length() ; i++)
	exceptionsIS_[i] -> _remove_ref();

    exceptionsIS_ = implSeq;
}

// ----------------------------------------------------------------------
// InterfaceDef_impl member implementations
// ----------------------------------------------------------------------

InterfaceDef_impl::InterfaceDef_impl(ORB_ptr orb,
                                     POA_ptr poa,
                                     Container_impl* c,
				     Repository_impl* r)
    : IRObject_impl(orb, poa, r),
      Container_impl(orb, poa, r),
      Contained_impl(orb, poa, c, r),
      IDLType_impl(orb, poa, r)
{
}

InterfaceDef_impl::~InterfaceDef_impl()
{
}

void
InterfaceDef_impl::validateName(const char* name,
                                bool exclude_inherited,
                                bool in_inherited)
{
    Container_impl::validateName(name, exclude_inherited, in_inherited);

    if(!exclude_inherited)
    {
        for(ULong i = 0 ; i < baseInterfacesIS_.length() ; i++)
            baseInterfacesIS_[i] -> validateName(name, false, true);
    }
}

void
InterfaceDef_impl::destroy()
    throw(SystemException)
{
    for(ULong i = 0 ; i < baseInterfacesIS_.length() ; i++)
	baseInterfacesIS_[i] -> _remove_ref();
    baseInterfacesIS_.length(0);

    destroyContents();
    Contained_impl::destroy();
}

TypeCode_ptr
InterfaceDef_impl::type()
    throw(SystemException)
{
    return orb_ -> create_interface_tc(id_, name_);
}

CORBA::Contained::Description*
InterfaceDef_impl::describe()
    throw(SystemException)
{
    InterfaceDescription_var desc = new InterfaceDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);
    if(contained)
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();

    for(ULong i = 0 ; i < baseInterfacesIS_.length() ; i++) 
    {
        desc -> base_interfaces.length(desc -> base_interfaces.length() + 1);
        desc -> base_interfaces[desc -> base_interfaces.length() - 1] =
            baseInterfacesIS_[i] -> id();
    }
    
    CORBA::Contained::Description_var description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc._retn();

    return description._retn();
}

Contained_ptr
InterfaceDef_impl::lookup(const char* search_name)
    throw(SystemException)
{
    Contained_var contained = Container_impl::lookup(search_name);

    if(is_nil(contained))
    {
	for(ULong i = 0 ; i < baseInterfacesIS_.length() ; i++)
	{
	    Contained_var contained2 =
		baseInterfacesIS_[i] -> lookup(search_name);
	    
	    if(!is_nil(contained2))
	    {
		if(!is_nil(contained))
		{
		    RepositoryId_var id = contained -> id();
		    RepositoryId_var id2 = contained2 -> id();
		    if(strcmp(id, id2) != 0)
			throw INTF_REPOS(MinorLookupAmbiguous, COMPLETED_NO);
		}

		contained = contained2;
	    }
	}
    }

    return contained._retn();
}

ContainedSeq*
InterfaceDef_impl::contents(DefinitionKind limit_type,
			    Boolean exclude_inherited)
    throw(SystemException)
{
    ContainedSeq_var s = Container_impl::contents(limit_type,
						  exclude_inherited);
    
    if(!exclude_inherited)
    {
	for(ULong i = 0 ; i < baseInterfacesIS_.length() ; i++)
	{
	    ContainedSeq_var s2 =
		baseInterfacesIS_[i] -> contents(limit_type, false);
	    mergeContainedSeq(s.inout(), s2);
	}
    }
    
    return s._retn();
}

ContainedSeq*
InterfaceDef_impl::lookup_name(const char* search_name,
			       Long levels_to_search,
			       DefinitionKind limit_type,
			       Boolean exclude_inherited)
    throw(SystemException)
{
    ContainedSeq_var s =
	Container_impl::lookup_name(search_name, levels_to_search,
				    limit_type, exclude_inherited);

    if(!exclude_inherited)
    {
	for(ULong i = 0 ; i < baseInterfacesIS_.length() ; i++)
	{
	    ContainedSeq_var s2 =
		baseInterfacesIS_[i] -> lookup_name(search_name,
							 levels_to_search,
							 limit_type,
							 false);
	    mergeContainedSeq(s.inout(), s2);
	}
    }
		
    return s._retn();
}

InterfaceDefSeq*
InterfaceDef_impl::base_interfaces()
    throw(SystemException)
{
    InterfaceDefSeq_var s = new InterfaceDefSeq;
    s -> length(baseInterfacesIS_.length());
    for(ULong i = 0 ; i < s -> length() ; i++)
	s[i] = baseInterfacesIS_[i] -> _this();
    return s._retn();
}

void
InterfaceDef_impl::base_interfaces(const InterfaceDefSeq& seq)
    throw(SystemException)
{
    ULong i;

    //
    // Make sure that the interface doesn't inherit from a local
    // interface if this interface itself isn't a local interface
    //
    if(def_kind() != dk_LocalInterface)
    {
	for(i = 0 ; i < seq.length() ; i++)
	    if(seq[i] -> def_kind() == dk_LocalInterface)
		throw BAD_PARAM(MinorInvalidLocalInterfaceInheritance,
				COMPLETED_NO);
    }

    ContainedSeq conSeq2;

    for(i = 0 ; i < seq.length() ; i++)
    {
	ContainedSeq_var conSeq1 = seq[i] -> contents(dk_Operation, false);

	for(ULong j = 0 ; j < conSeq1 -> length() ; j++)
	{
	    String_var name1 = conSeq1[j] -> name();
	    String_var id1 = conSeq1[j] -> id();

	    for(ULong k = 0 ; k < conSeq2.length() ; k++)
	    {
		String_var name2 = conSeq2[k] -> name();
		String_var id2 = conSeq2[k] -> id();

		if(strcmp(id1, id2) != 0 &&
                   IdlStringCompare(name1, name2) == 0)
		    throw BAD_PARAM(MinorNameClashInInheritedContext,
				    COMPLETED_NO);
	    }

            conSeq2.length(conSeq2.length() + 1);
            conSeq2[conSeq2.length() - 1] = conSeq1[j];
	}
    }

    for(i = 0 ; i < seq.length() ; i++)
    {
	ContainedSeq_var conSeq1 = seq[i] -> contents(dk_Attribute, false);

	for(ULong j = 0 ; j < conSeq1 -> length() ; j++)
	{
	    String_var name1 = conSeq1[j] -> name();
	    String_var id1 = conSeq1[j] -> id();

	    for(ULong k = 0 ; k < conSeq2.length() ; k++)
	    {
		String_var name2 = conSeq2[k] -> name();
		String_var id2 = conSeq2[k] -> id();

		if(strcmp(id1, id2) != 0 &&
                   IdlStringCompare(name1, name2) == 0)
		    throw BAD_PARAM(MinorNameClashInInheritedContext,
				    COMPLETED_NO);
	    }

            conSeq2.length(conSeq2.length() + 1);
            conSeq2[conSeq2.length() - 1] = conSeq1[j];
	}
    }

    InterfaceDefIS implSeq;
    implSeq.length(seq.length());
    POA_var poa = _default_POA();
    for(i = 0 ; i < seq.length() ; i++)
    {
	if(is_nil(seq[i]))
	    throw BAD_PARAM(); // TODO: Is this exception correct?
	implSeq[i] = 0;
	try
	{
	    ServantBase_var servant = poa -> reference_to_servant(seq[i]);
	    implSeq[i] = dynamic_cast<InterfaceDef_impl*>(servant.in());
	}
	catch(const POA::WrongAdapter&)
	{
	}
	catch(const POA::WrongPolicy&)
	{
	}
	catch(const POA::ObjectNotActive&)
	{
	}
	if(!implSeq[i])
	{
	    // Must be same repository
	    throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
	}
    }
    
    for(i = 0 ; i < implSeq.length() ; i++)
	implSeq[i] -> _add_ref();
    
    for(i = 0 ; i < baseInterfacesIS_.length() ; i++)
	baseInterfacesIS_[i] -> _remove_ref();

    baseInterfacesIS_ = implSeq;
}

Boolean
InterfaceDef_impl::is_a(const char* interface_id)
    throw(SystemException)
{
    if(strcmp(id_, interface_id) == 0)
	return true;

    if(strcmp(interface_id, "IDL:omg.org/CORBA/Object:1.0") == 0)
        return true;

    for(ULong i = 0 ; i < baseInterfacesIS_.length() ; i++)
	if(baseInterfacesIS_[i] -> is_a(interface_id))
	    return true;

    return false;
}

CORBA::InterfaceDef::FullInterfaceDescription*
InterfaceDef_impl::describe_interface()
    throw(SystemException)
{
    ULong i;

    CORBA::InterfaceDef::FullInterfaceDescription_var desc =
        new CORBA::InterfaceDef::FullInterfaceDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);
    if(contained)
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();

    ContainedSeq_var s;

    s = contents(dk_Operation, false);
    desc -> operations.length(s -> length());
    for(i = 0 ; i < s->length() ; i++)
    {
	CORBA::Contained::Description_var d = s[i] -> describe();
	assert(d -> kind == dk_Operation);
	const OperationDescription* opDesc;
	bool ok = (d -> value >>= opDesc);
	assert(ok);
	desc -> operations[i] = *opDesc;
    }

    s = contents(dk_Attribute, false);
    desc -> attributes.length(s -> length());
    for(i = 0 ; i < s->length() ; i++)
    {
	CORBA::Contained::Description_var d = s[i] -> describe();
	assert(d -> kind == dk_Attribute);
	const AttributeDescription* atDesc;
	bool ok = (d -> value >>= atDesc);
	assert(ok);
	desc -> attributes[i] = *atDesc;
    }

    for(i = 0 ; i < baseInterfacesIS_.length() ; i++)
    {
        desc -> base_interfaces.length(desc -> base_interfaces.length() + 1);
        desc -> base_interfaces[desc -> base_interfaces.length() - 1] =
            baseInterfacesIS_[i] -> id();
    }

    desc -> type = type();

    return desc._retn();
}

AttributeDef_ptr
InterfaceDef_impl::create_attribute(const char* id,
				    const char* name,
				    const char* version,
				    IDLType_ptr type,
				    AttributeMode mode)
    throw(SystemException)
{
    AttributeDef_impl* p = new AttributeDef_impl(orb_, poa_, this,
                                                 repository_);

    activateImpl(id, p);

    AttributeDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> type_def(type);
        p -> mode(mode);
    }
    catch(...)
    {
	p -> destroy();
        throw;
    }

    return ref._retn();
}

OperationDef_ptr
InterfaceDef_impl::
create_operation(const char* id,
		 const char* name,
		 const char* version,
		 IDLType_ptr result,
		 OperationMode mode,
		 const ParDescriptionSeq& params,
		 const ExceptionDefSeq& exceptions,
		 const ContextIdSeq& contexts)
    throw(SystemException)
{
    OperationDef_impl* p = new OperationDef_impl(orb_, poa_, this,
                                                 repository_);

    activateImpl(id, p);

    OperationDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> result_def(result);
        p -> mode(mode);
        p -> params(params);
        p -> exceptions(exceptions);
        p -> contexts(contexts);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}
