// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/Properties.h>
#include <OB/BootManager.h>

#include <Util.h>
#include <IFR3_impl.h>

#include <stdlib.h>
#include <stdio.h>

#include <CPPUtil.h>

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] idl-files ...\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show ORBacus version.\n"
	"-d, --debug            Print diagnostic messages.\n"
	"-i, --ior              Print repository IOR on standard output.\n"
	"-DNAME                 Define NAME as 1.\n"
	"-DNAME=DEF             Define NAME as DEF.\n"
	"-UNAME                 Remove any definition for NAME.\n"
	"-IDIR                  Put DIR in the include file search path.\n"
        "--case-sensitive       Allow case-sensitive identifiers.\n"
        "                       (Not CORBA compliant.)\n"
	;
}

int
run(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    assert(!is_nil(oborb));

    //
    // Get options
    //
    String_var cppArgs;
    bool ior = false;
    bool caseSensitive = false;

    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
	if(strcmp(argv[i], "--help") == 0 ||
	   strcmp(argv[i], "-h") == 0)
	{
	    usage(progName);
	    return 0;
	}
	else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
	{
	    OB_CERR << "ORBacus " << Version << OB_ENDL;
	    return 0;
	}
	else if(strcmp(argv[i], "--debug") == 0 ||
		strcmp(argv[i], "-d") == 0)
	{
	    extern int yydebug;
	    yydebug = 1;
	}
	else if(strcmp(argv[i], "--ior") == 0 ||
		strcmp(argv[i], "-i") == 0)
	{
	    ior = true;
	}
	else if(strncmp(argv[i], "-D", 2) == 0 ||
		strncmp(argv[i], "-U", 2) == 0 ||
		strncmp(argv[i], "-I", 2) == 0)
	{
	    if(strlen(argv[i]) == 2)
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i] << "'" << OB_ENDL; // not argv[i - 1] !
		return 1;
	    }

	    cppArgs += ' ';
	    cppArgs += argv[i];
	}
	else if(strcmp(argv[i], "--case-sensitive") == 0)
	{
            OB_CERR << progName << ": Warning: use of the --case-sensitive "
                    << "option is not CORBA compliant!" << OB_ENDL;
	    caseSensitive = true;
	}
	else
	{
	    OB_CERR << progName << ": unknown option `"
		    << argv[i] << "'\n" << OB_ENDL;
	    usage(progName);
	    return 1;
	}
    }

    char* idlcpp = IdlGetCPreprocessor(progName);
    if(!idlcpp)
	return 1;
    String_var cppCmd = idlcpp;

    //
    // Grab the ooc.ifr.endpoint property. If this exists then set the
    // ooc.orb.oa.endpoint property from this value.
    //
    OB::Properties_var properties = oborb -> properties();
    const char* value = properties -> getProperty("ooc.ifr.endpoint");
    if(value != 0)
	properties -> setProperty("ooc.orb.oa.endpoint", value);

    //
    // Resolve Root POA
    //
    CORBA::Object_var poaObj = orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);

    //
    // Get a reference to the POA manager
    //
    POAManager_var manager = rootPOA -> the_POAManager();

    //
    // Create POA
    //
    // TODO: This may not be correct. Should the Repository be
    // activated on PERSISTENT POA only while it activates objects on
    // a NON-PERSISTENT POA (rootPOA?).
    //
    PolicyList policies;
    policies.length(3);
    policies[0] =
        rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] =
        rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
    policies[2] =
        rootPOA -> create_thread_policy(PortableServer::SINGLE_THREAD_MODEL);

    PortableServer::POA_var poa;
    try
    {
        poa = rootPOA -> create_POA("InterfaceRepositoryPOA", manager,
				    policies);
    }
    catch(const PortableServer::POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(poa));

    //
    // Resolve the Boot Manager
    //
    OB::BootManager_var bootManager = OB::BootManager::_narrow(
        Object_var(orb -> resolve_initial_references("BootManager")));

    //
    // Create and activate implementation object
    //
    Repository_impl* repositoryImpl = new Repository_impl(orb, poa);

    //
    // Register the repository with the Boot Manager
    //
    ObjectId_var repositoryID = string_to_ObjectId("DefaultRepository");
    poa -> activate_object_with_id(repositoryID, repositoryImpl);
    Repository_var repository = repositoryImpl -> _this();
    bootManager -> add_binding(repositoryID, repository);

    manager -> activate();

    int status = 0;
    while(status == 0 && i < argc)
    {
	String_var file = string_dup(argv[i++]);
	
	//
	// Open input file
	//
	OB_TRANSLATOR_IMPORT extern FILE* yyin;
	String_var cmd = cppCmd;
	cmd += cppArgs;
	cmd += ' ';
	cmd += file;

	OBProcessIO pio;
	if((yyin = pio.open(cmd, "r")) == NULL)
	    return 1;

	//
	// Do parsing
	//
	int IdlParse(ORB_ptr, Repository_ptr, StringSeq&,
		     bool, bool, bool, bool, bool, bool, bool,
		     IdlStringHash&, IdlStringHash&);
	StringSeq includeSeq;
	IdlStringHash idHash;
	IdlStringHash fwdHash;
	status = IdlParse(orb, repository, includeSeq, true, true,
                          caseSensitive, true, true, true, true,
                          idHash, fwdHash);
	
	//
	// Close input file
	//
	pio.close();
    }

    if(status == 0)
    {
	if(ior)
	{
	    String_var s = orb -> object_to_string(repository);
	    OB_COUT << s << OB_ENDL;
	}

        orb -> run();
    }
    
    repositoryImpl -> destroyRepository();
    repositoryImpl -> _remove_ref();

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;
    try
    {
	//
	// Initialize the ORB
	//
	orb = ORB_init(argc, argv);

	//
	// Generate code
	//
	status = run(orb, argc, argv);
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
