// **********************************************************************

// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IIOP.h>

#include <stdlib.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace OB;

static void
MergeIOR(const char* progName, bool /* alternate */,
	 IOP::IOR& to, const IOP::IOR& from)
{
    ULong i;

    if(strlen(to.type_id) == 0)
    {
	to.type_id = from.type_id;
    }
    else
    {
	if(progName && strcmp(to.type_id, from.type_id) != 0)
	{
	    OB_CERR << progName << ": warning: type id `" << from.type_id
			 << "' does not match `" << to.type_id << "'"
			 << OB_ENDL;
	}
    }

    //
    // Merge all profiles
    //
    ULong toLen = to.profiles.length();
    ULong fromLen = from.profiles.length();
    ULong newLen = toLen + fromLen;
    to.profiles.length(newLen);
    for(i = 0 ; i < fromLen ; i++)
	to.profiles[toLen + i] = from.profiles[i];
}

static void
MergeIOR(const char* progName, bool alternate, IOP::IOR& to, const char* s)
{
    assert_nca(s != 0, NCANullString);
    
    if(strncmp(s, "IOR:", 4) != 0)
	throw INV_OBJREF();
    
    s += 4;
    
    Octet* poct = AsciiToOctets(s);
    OCI::BufferImpl buf(poct, strlen(s) / 2);
    InputStreamImpl in(&buf, 0, false);

    Boolean endian = in.read_boolean();
    in._OB_swap(endian != Endian);

    IOP::IOR from;
    IOP::IOR::_OB_unmarshal(from, &in);
    
    MergeIOR(progName, alternate, to, from);
}

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] [-f FILE ... | IOR ...]\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show ORBacus version.\n"
        "-w, --nowarn           Don't print any warning messages.\n"
        "-a, --alternate        Use alternate IIOP address components if\n"
	"                       possible.\n"
        "-f                     Read IORs from files instead from the\n"
	"                       command line.\n"
	;
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];
	
    //
    // Get options
    //
    bool files = false;
    bool nowarn = false;
    bool alternate = false;
    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
        if(strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0)
        {
	    usage(progName);
	    return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
        {
	    OB_CERR << "ORBacus " << OB::Version << OB_ENDL;
	    return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "--nowarn") == 0 ||
		strcmp(argv[i], "-w") == 0)
        {
	    nowarn = true;
        }
        else if(strcmp(argv[i], "--alternate") == 0 ||
		strcmp(argv[i], "-a") == 0)
        {
	    alternate = true;
        }
        else if(strcmp(argv[i], "-f") == 0)
        {
	    files = true;
        }
        else
        {
	    OB_CERR << progName << ": unknown option `" << argv[i]
			 << "'\n" << OB_ENDL;
	    usage(progName);
	    return EXIT_FAILURE;
        }
    }
	
    if(i == argc)
    {
        OB_CERR << progName << ": no IORs\n" << OB_ENDL;
        usage(progName);
        return EXIT_FAILURE;
    }
	
    //
    // The merged IOR
    //
    IOP::IOR ior;

    if(!files)
    {
	const char* isIor = strstr(argv[i], "IOR:");
	const char* isCorbaloc = strstr(argv[i], "corbaloc:");

	if(!isIor && !isCorbaloc)
	{
	    OB_CERR << "[No valid IOR in the command line, assuming -f]"
		    << OB_ENDL;

	    files = true;
	}
    }

    if(!files)
    {
        //
        // Merge IORs given as arguments
        //
        for(; i < argc ; i++)
	{
	    try
	    {
		//
		// This results in support for corbaloc
		//
		Object_var obj = orb -> string_to_object(argv[i]);
		String_var s = orb -> object_to_string(obj);

		MergeIOR(nowarn ? 0 : progName, alternate, ior, s);
	    }
	    catch(const BAD_PARAM&)
	    {
		OB_CERR << "IOR is invalid" << OB_ENDL;
	    }
	}
    }
    else
    {
        //
        // Merge all IORs from the specified files
        //
        for(; i < argc ; i++)
        {
	    OB_IFSTREAM in;
	    // Must use open(name), not ifstream in(name) (VC++ bug)
	    in.open(argv[i]);
	    if(in.fail())
	    {
	        OB_CERR << progName << ": can't open `" << argv[i]
			     << "': " << strerror(errno) << OB_ENDL;
	        return EXIT_FAILURE;
    	    }
		
    	    String_var str = string_dup("");
		
    	    while(!in.eof())
    	    {
	        char c;
	        in.read(&c, 1);
	        if(in.eof() || c == '\n' || c == '\r' || c == '\0')
	        {
		    if(strlen(str) > 0)
		    {
			Object_var obj = orb -> string_to_object(str);
			String_var s = orb -> object_to_string(obj);

		        MergeIOR(nowarn ? 0 : progName, alternate, ior, s);

		        str = string_dup("");
		    }
		    }
		    else
			str += c;
	    }
        }
    }

    //
    // Stringify the merged IOR and print it on standard output
    //
    OCI::BufferImpl buf;
    OutputStreamImpl out(&buf);

    out._OB_writeEndian();
    ior._OB_marshal(&out);
	
    String_var s = OctetsToAscii(buf.data(), buf.length());
	
    static const char* prefix = "IOR:";
    char* result = string_alloc(strlen(prefix) + strlen(s));
    strcpy(result, prefix);
    strcat(result, s);
	
    OB_COUT << result << OB_ENDL;

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
