// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <stdlib.h>

using namespace CORBA;
using namespace OB;

//
// Create a registry key, and all subkeys leading to this key.
//
HKEY
createKey(HKEY base, const char* key)
{
    char path[MAX_PATH];
    strcpy(path, key);
    char* p = path;
    const char* end = p + strlen(path);
    for(;;)
    {
	while(*p && *p != '\\')
	    ++p;

	//
	// Failure - badly formatted key.
	//
	if(*p == '\0')
	{
	    OB_CERR << "createKey: expected \\ \"" << key << "\""
			 << OB_ENDL;
	    return 0;
	}

	*p = '\0';

	//
	// Try to open this key.
	//
	HKEY key;
	DWORD rc = RegOpenKeyEx(base, path, 0, KEY_WRITE|KEY_EXECUTE, &key);
	if(rc != ERROR_SUCCESS)
	{
	    rc = RegCreateKey(base, path, &key);
	    if(rc != ERROR_SUCCESS)
	    {
		OB_CERR << "RegCreateKey \"" << path << "\" failed: "
			     << rc << OB_ENDL;
		return 0;
	    }
	}
	*p = '\\';
	++p;
	//
	// Have we opened the full key?
        //
	if(p == end)
	    return key;
	RegCloseKey(key);
    }
}

//
// Add the key and value to the registry.
//
bool
addRegistryKey(HKEY base, const char* basePath, const char* k, const char* v)
{
    char path[MAX_PATH];
    char valueName[MAX_PATH];

    valueName[0] = '\0';
    strcpy(path, basePath);

    //
    // Ensure that the path ends with '\'
    //
    if(path[strlen(path)-1] != '\\')
	strcat(path, "\\");

    char keyname[MAX_PATH];
    strcpy(keyname, k);

    char* start = keyname;
    char* p = start;

    //
    // Take each dotted section of the name. The last dotted section
    // is the value name, and the other section names are the keys,
    // and sub-keys.
    //
    while(p != 0)
    {
	p = strchr(start, '.');
	if(p != 0)
	{
	    *p = '\0';
	}
	if(*valueName != '\0')
	{
	    strcat(path, valueName);
	    strcat(path, "\\");
	}
	strcpy(valueName, start);
	if(p != 0)
	{
	    start = p + 1;
	    p = start;
	}
    }
    
    //
    // Try to open the key.
    //
    HKEY key;
    DWORD rc = RegOpenKeyEx(base, path, 0, KEY_WRITE|KEY_EXECUTE, &key);
    if(rc != ERROR_SUCCESS)
    {
	//
	// If open fails, then create the key.
	//
	key = createKey(base, path);

	//
	// If that fails then we can't add the value.
	//
	if(key == 0)
	    return false;
    }

    rc = RegSetValueEx(key, valueName, 0, REG_SZ, (BYTE*)v, strlen(v));
    if(rc != ERROR_SUCCESS)
    {
	OB_CERR << "RegSetValueEx failed: " << rc << OB_ENDL;
    }
    RegCloseKey(key);

    return true;
}

//
// Clean each of the values from the registry under the OOC\\Properties
// tree.
//
void
cleanRegistry(HKEY base, const char* path, bool deleteThis)
{
    HKEY key;
    DWORD rc = RegOpenKeyEx(base, path, 0,
			 KEY_ENUMERATE_SUB_KEYS|KEY_EXECUTE|KEY_QUERY_VALUE,
			 &key);
    if(rc != ERROR_SUCCESS)
	return;

    //
    // Get all sub-keys, and add to a string list. Next run through
    // each sub-key, and clean all the contained entries.
    //
    char name[MAX_PATH];
    StrSeq< int > trash;
    int i;
    for(i = 0, rc = ERROR_SUCCESS; rc == ERROR_SUCCESS; i++)
    {
	rc = RegEnumKey(key, i, name, MAX_PATH);
	if(rc == ERROR_SUCCESS)
	{
	    char buf[MAX_PATH];
	    strcpy(buf, path);
	    strcat(buf, "\\");
	    strcat(buf, name);
	    ULong l = trash.length();
	    trash.length(l + 1);
	    trash[l] = string_dup(buf);
	}
    }
    for(ULong j = 0; j < trash.length(); ++j)
	cleanRegistry(base, trash[j], true);

    RegCloseKey(key);

    //
    // Delete all values in this key.
    //
    if(deleteThis)
    {
	rc = RegDeleteKey(base, path);
	if(rc != ERROR_SUCCESS)
	    OB_CERR << "RegDeleteKey failed: " << rc << OB_ENDL;
    }
}

void
usage(const char* progName)
{
    OB_CERR << "Usage: " << progName
		 << " HKEY_CURRENT_USER|HKEY_LOCAL_MACHINE file"
		 << OB_ENDL;
}

int
main(int argc, char** argv)
{
    if(argc != 3)
    {
	usage(argv[0]);
	return EXIT_FAILURE;
    }
    HKEY base;
    if(strcmp(argv[1], "HKEY_CURRENT_USER") == 0)
	base = HKEY_CURRENT_USER;
    else if(strcmp(argv[1], "HKEY_LOCAL_MACHINE") == 0)
	base = HKEY_LOCAL_MACHINE;
    else
    {
	usage(argv[0]);
	return EXIT_FAILURE;
    }

    const char* file = argv[2];
    Properties_var properties = new Properties();
    if(!properties -> load(file))
    {
	OB_CERR << "Cannot open: " << file << OB_ENDL;
	return EXIT_FAILURE;
    }

    const char* basePath = "Software\\OOC\\Properties";

    //
    // Clean all registry keys under
    //
    cleanRegistry(base, basePath, false);

    //
    // Get all registry keys.
    //
    Properties::KeySeq keys = properties -> getKeys("");
    for(ULong i = 0; i < keys.length(); ++i)
    {
	if(!addRegistryKey(base, basePath, keys[i],
			   properties -> getProperty(keys[i])))
	{
	    OB_CERR << "Add of " << keys[i] << " failed. Bailing."
			 << OB_ENDL;
	    return EXIT_FAILURE;
	}
    }
    return EXIT_SUCCESS;
}
