// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <URLTest_impl.h>

// 
// Included for IMR full key construction test
// 
#include <OB/ObjectKey.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;

// 
// URL Test Failure Exception
// 
class TestFailure 
{
};

// 
// Test string_to_object functionality with various URL types 
// 
class ClientTest 
{
public:

    CORBA::ORB_var orb;

    //
    // Number of times to repeat the test
    //
    int count;

    //
    // Object references to use during testing 
    //
    URLTest::IIOPAddress_var ior[2];

    ClientTest()
    {
    }
    
    // 
    // Initialize the ClientTest
    // 
    void
    init(ORB_ptr aOrb, int aCount)
    {
        orb = CORBA::ORB::_duplicate(aOrb);
        count = aCount;
    }

    // 
    // Given a URL string, return a narrow'd URLTest::IIOPAddress_ptr
    // 
    URLTest::IIOPAddress_ptr 
    stringToIIOPAddress(const char*objStr)
    {
        Object_var obj;
        URLTest::IIOPAddress_ptr nObj;
        bool stringToObjOK = false;
        
        try
        {
            obj = orb -> string_to_object(objStr);
            stringToObjOK = true;
            
            nObj = URLTest::IIOPAddress::_narrow(obj);
            
            if(CORBA::is_nil(nObj))
            {
                OB_ERROR(OB_ENDL << "Error stringToIIOPAddress:");
                OB_ERROR("object `"  <<  objStr << "' is nil reference.");
                throw TestFailure();
            }
        }
        catch(const CORBA::SystemException &se)
        {
            OB_ERROR(OB_ENDL << "Error stringToIIOPAddress:");
            OB_ERROR(se);
            OB_ERROR("Original URL: " << objStr << OB_ENDL 
		     << "Resultant URL: ");
            if(stringToObjOK == false)
                OB_ERROR("NONE, string_to_object raised the exception");
            else
            {
                String_var iorStr = orb -> object_to_string(obj);
                OB_ERROR(iorStr.in());
            }
            throw TestFailure();
        }

        return nObj;
    }

    // 
    // Log test start
    // 
    void 
    declare_test(const char*str)
    {
        OB_INFO("Testing " << str << "... ");
    }
    
    // 
    // Log test end, raise exception if test failed
    // 
    void 
    finish_test(bool isOk)
    {
        OB_INFO_LN((isOk ? "Done!" : "Failed!"));

        if(isOk == false)
            throw TestFailure();
    }
        
    // 
    // Logical comparison of two URLTest::IIOPAddress objects
    // considered equal if they return the same host, port, and key
    // 
    bool 
    compareObjs(URLTest::IIOPAddress_ptr iop1, 
                URLTest::IIOPAddress_ptr iop2)
    {
        
        String_var cKey[2];
        String_var cHost[2];
        Long cPort[2];
        URLTest::IIOPAddress_ptr activeIOR;
        
        try
        {
            activeIOR = iop1;
            cKey[0] = iop1 -> getKey();
            cHost[0] = iop1 -> getHost();
            cPort[0] = iop1 -> getPort();

            activeIOR = iop2;
            
            cKey[1] = iop2 -> getKey();
            cHost[1] = iop2 -> getHost();
            cPort[1] = iop2 -> getPort();
        }
        catch(const CORBA::SystemException &se)
        {
            OB_ERROR(OB_ENDL << "Error compareObj, invoking :");
            OB_ERROR(se);
            String_var iorStr = orb -> object_to_string(activeIOR);
            OB_INFO_LN(iorStr.in());
            throw;
        } 

        return strcmp(cKey[0].in(), cKey[1].in())==0 &&
            strcmp(cHost[1].in(), cHost[1].in())==0 &&
            cPort[0] == cPort[1];
    }

    // 
    // Queries a IIOPAddress object and returns the "iiop:" URL
    // component to contact that object. ":555objs.com:2809" for
    // example.
    // 
    char* getURLAddressComponent(URLTest::IIOPAddress_ptr nObj,
				 const char*prefix = 0)
    {
        String_var url(CORBA::string_dup((prefix == 0 ) ? ":" : prefix));
        String_var str = nObj -> getHost();
        url += str.in();
        url += ":";
        url += nObj -> getPort();
        return url._retn();
    }
    
    // 
    // Run a single pass of the URL tests
    // 
    void 
    runPass(int)
    {
        // 
        // Start with very basic tests
        // 
        CORBA::String_var host[2];
        CORBA::String_var key[2];
        CORBA::Long port[2];
        
        // 
        // Test that initial object strings are ok (tests scheme
        // provided on command line)
        // 
        declare_test("string_to_object");
        
	unsigned int i;
        for(i = 0; i < 2; ++i)
        {
            key[i] = ior[i] -> getKey();
            host[i] = ior[i] -> getHost();
            port[i] = ior[i] -> getPort();
        }
        finish_test(true);

        // 
        // Check that object_to_string(string_to_object()) is ok
        // 
        declare_test("object_to_string");

        for(i = 0; i < 2; ++i)
        {
            String_var iorStr = orb -> object_to_string(ior[i]);
            Object_var nObj= orb -> string_to_object(iorStr.in());
            URLTest::IIOPAddress_var nAddr = 
                URLTest::IIOPAddress::_narrow(nObj);

            if(compareObjs(ior[i].in(), nAddr.in()) == 0)
                finish_test(false);
        }

        finish_test(true);

        // 
        // Check that file URL handling is correct with Unix and NT
        // line termination and extra lines
        // 
        declare_test("file URLs");

        for(i = 0; i < 2; ++i)
        {
            int ltor;
            for(ltor = URLTest::NoEOL; ltor <= URLTest::Win32EOLExtraText;
                ++ltor)
            {
                URLTest::TempIORFile tempFile(orb, ior[i], "ftestURL.ref",
					      ltor);
                Object_var nObj= orb ->
		    string_to_object("relfile:ftestURL.ref");
                URLTest::IIOPAddress_var nAddr = 
                    URLTest::IIOPAddress::_narrow(nObj);
                if(compareObjs(ior[i].in(), nAddr.in()) == 0)
                    finish_test(false);
            }
        }
        finish_test(true);

        // 
        // Test that null URL string is handled correctly
        // 
        declare_test("null URL string exception");
        {
            bool correctException = false;
            try
            {
                const char* nullStr = 0;
                Object_var nObj = orb -> string_to_object(nullStr);
            }
            catch(const CORBA::BAD_PARAM &)
            {
                correctException = true;
            }

            finish_test(correctException);
        }

        declare_test("bad corbaloc port and scheme exceptions");
        {
            // 
            // Try URL with legal port first, followed
            // by low and high illegal port numbers
            // 
            const char* urlStr[]=
            {
                "corbaloc:iiop:localhost:2809/a/b/c",
                "corbaloc:iiop:localhost:0/a/b/c",
                "corbaloc:iiop:localhost:65536/a/b/c",
                "corbaloc:iiop:localhost:-90/a/b/c",
                "corbaloc:iiop:localhost:not_a_number/a/b/c",
                "corbaloc:corbaloc-bad-scheme:localhost:not_a_number/a/b/c"
            };
            int exceptionCount =0;
            
            for(i = 0; i < 6; ++i)
            {
                try
                {
                    Object_var nObj = orb -> string_to_object(urlStr[i]);
                }
                catch(const CORBA::BAD_PARAM &)
                {
                    // 
                    // First URL should be valid
                    // 
                    if(i == 0)
                        finish_test(false);
                    
                    ++exceptionCount;
                }            
            }

            // 
            // Last 5 URLs should have failed
            // 
            finish_test(exceptionCount == 5);
        }
        
        // 
        // Basic corbaloc test
        // 
        declare_test("iiop default version 1.0 corbaloc URL");

        for(i = 0; i < 2; ++i)
        {
            String_var url = ior[i] -> getCorbalocURL();
            URLTest::IIOPAddress_var nObj = stringToIIOPAddress(url.in());

            if(compareObjs(ior[i], nObj.in()) == 0)
                finish_test(false);

            String_var str = nObj -> getString();
            
            if(strcmp(str.in(), "corbaloc") != 0)
                finish_test(false);
        }

        finish_test(true);

        // 
        // Versioned corbaloc test. Minor versions beyond what
        // ORBacus supports will be downgraded to the maximum
        // the orb supports.
        // 
        declare_test("iiop version 1.0 - 1.9 corbaloc URL");

        char iiopVersion[6] = {':', '1', '.', '0', '@','\0' };

        for(i = 1; i < 2; ++i)
        {
            unsigned int iv;
            for(iv = 0; iv < 10; ++ iv)
            {
                iiopVersion[3] = '0' + iv;
            
                String_var url((const char* )"corbaloc:");
                String_var addr = getURLAddressComponent(ior[i], 
                                                         iiopVersion);
                url += addr.in();
                url += "/";
                CORBA::String_var keyStr = ior[i] -> getKey();
                url += keyStr;
                
                URLTest::IIOPAddress_var nObj = stringToIIOPAddress(url.in());

                if(compareObjs(ior[i], nObj.in()) == 0)
                    finish_test(false);
                
                String_var str = nObj -> getString();
            
                if(strcmp(str.in(), "corbaloc") != 0)
                    finish_test(false);
            }
        }

        finish_test(true);

        // 
        // Versioned corbaloc rejection test.
        // 
        declare_test("unsupported iiop version corbaloc URL");

        for(i = 0; i < 2; ++i)
        {
            // 
            // Try versions 0.0 - 0.9 and 2.0 - 2.9
            // 
            iiopVersion[1] = (i==0) ? '0' : '2';

            unsigned int iv;
            for(iv = 0; iv < 10; ++iv)
            {
                iiopVersion[3] = '0' + iv;
                String_var url((const char* )"corbaloc:");

                String_var addr = getURLAddressComponent(ior[i], 
                                                         iiopVersion);
                url += addr.in();
                url += "/";
                CORBA::String_var keyStr = ior[i] -> getKey();
                url += keyStr;

                try
                {
                    CORBA::Object_var obj = orb -> string_to_object(url.in());
                    finish_test(false);
                }
                catch(const CORBA::BAD_PARAM &)
                {
                }
            }
        }
        finish_test(true);

        // 
        // Test corbaloc access the same way the
        // IMR mkref utility generates corbaloc's.
        // 
        declare_test("IMR corbaloc");

        for(i = 0; i < 2; ++i)
        {
            // 
            // Create the ORBacus specific object key
            // 
            String_var iorKey = ior[i] -> getKey();

            PortableServer::ObjectId_var oid =
                PortableServer::string_to_ObjectId(iorKey.in());
            
            OB::ObjectKeyData data;
            data.persistent = true;
            data.createTime = 0;
            data.oid = oid;
	    data.serverId = CORBA::string_dup("_RootPOA");
            data.poaId.length(1);
            data.poaId[0] = CORBA::string_dup("testPOA");

            OCI::ObjectKey_var oct = OB::CreateObjectKey(data);
            String_var keyStr = URLTest::keyToStr(oct, true);
            
            // 
            // Create a corbaloc URL for this
            // 
            String_var url((const char* )"corbaloc:");
            String_var addr = getURLAddressComponent(ior[i], 
                                                     (i & 1) ? 
                                                     ":1.1@" : ":1.2@");
            url += addr.in();
            url += '/';
            url += keyStr;
            
            URLTest::IIOPAddress_var nObj = stringToIIOPAddress(url.in());

            if(compareObjs(ior[i], nObj.in()) == 0)
                finish_test(false);

            String_var str = nObj -> getString();
            
            if(strcmp(str.in(), "corbaloc") != 0)
                finish_test(false);
        }

        finish_test(true);

        // 
        // Multi-profile test 1
        // 
        declare_test("active profile 1 in multi-profile IOR");

        for(i = 0; i < 2; ++i)
        {
            String_var str;
            String_var url = CORBA::string_dup("corbaloc:");
            String_var addr = getURLAddressComponent(ior[i]);
            url += addr.in();
            
            // 
            // Add 2 dummy profiles at corbaloc reserved port
            // 
            url += ",:localhost:2809";
            url += ",:localhost:2809";
            
            url += "/";
            str = ior[i] -> getKey();
            url += str.in();
            
            URLTest::IIOPAddress_var nObj = stringToIIOPAddress(url.in());
            
            if(compareObjs(ior[i], nObj.in()) == 0)
                finish_test(false);
            
            str = nObj -> getString();
            
            if(strcmp(str.in(), "corbaloc") != 0)
                finish_test(false);
        }
        
        finish_test(true);

        const bool performMultiProfileTest3 = false;
        
        if(performMultiProfileTest3)
        {
            // 
            // Multi-profile test 3 
            // 
            declare_test("active profile 3 in multi-profile IOR");
            
            for(i = 0; i < 2; ++i)
            {
                String_var str;
                String_var url = CORBA::string_dup("corbaloc:");
                
                // 
                // 2 dummy profiles at corbaloc reserved port
                // 
                url += ":localhost:2809";
                url += ",:localhost:2809,";
                
                // 
                // Usable profile
                // 
                String_var addr = getURLAddressComponent(ior[i]);
                url += addr.in();
                
                url += "/";
                str = ior[i] -> getKey();
                url += str.in();
                
                URLTest::IIOPAddress_var nObj = stringToIIOPAddress(url.in());
                
                if(compareObjs(ior[i], nObj.in()) == 0)
                    finish_test(false);
                
                str = nObj -> getString();
                
                if(strcmp(str.in(), "corbaloc") != 0)
                    finish_test(false);
            }
            finish_test(true);
        }
    }

    // 
    // Run the client URL tests
    // 
    void 
    run()
    {
        for(int c = 0; c < count; ++c)
        {
            if(count != 1)
                OB_INFO_LN("Test Pass: " << c + 1);
            try
            {
                runPass(c);
            }
            catch(TestFailure &)
            {
                // 
                // In multi-pass test, continue for TestFailure exceptions
                // 
                if(count == 1)
                    throw;
            }
        }
    }
};

//
// URL test main, arguments are: test loop count and the URL of the
// server, a second server URL can optionally be specified
// 
int
run(ORB_ptr orb, int argc, char* argv[])
{
    if(argc < 3)
    {
	OB_INFO("usage: " << argv[0] << " loop_count ior_url1 [ior_url2]");
	return EXIT_FAILURE;
    }
    
    int count = atoi(argv[1]);
    
    if(count <= 0)
    {
	OB_ERROR("bad test count ");
	return EXIT_FAILURE;
    }
    
    ClientTest test;
    test.init(orb, count);
    
    int status = EXIT_SUCCESS;
    
    try
    {
	test.ior[0] = test.stringToIIOPAddress(argv[2]);
	test.ior[1] = (argc != 4) ?
	    URLTest::IIOPAddress::_duplicate(test.ior[0]) : 
	    test.stringToIIOPAddress(argv[3]);
	
	test.run();
    }
    catch(const TestFailure&)
    {
	status = EXIT_FAILURE;
    }
	
    if(!is_nil(test.ior[0]))
	test.ior[0] -> deactivate();

    return status;
}

int
main(int argc, char* argv[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;

    try
    {
        orb = ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const Exception& ex)
    {
        OB_ERROR(ex);
        status = EXIT_SUCCESS;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_ERROR(ex);
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
