// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OBPolicies.h>
#include <OB/Timer.h>
#include <TestMacro.h>

#include <Test.h>

#include <stdlib.h>

using namespace CORBA;

static void
testLocationForward(ORB_ptr /* orb */, RetryServer_ptr server)
{
    OB_INFO("Testing forwarding loop detection... ");
    //
    // Verify that we detect excessive forwarding
    //
    Test_var test = server -> get_location_forward_object();
    try
    {
        test -> aMethod();
        TEST(false);
    }
    catch(const TRANSIENT& ex)
    {
        // Expected
        TEST(ex.minor() == OB::MinorLocationForwardHopCountExceeded);
    }
    OB_INFO_LN("Done!");
}

static Retry_ptr
getRetry(ORB_ptr orb, Retry_ptr orig, short mode, ULong interval,
         ULong max, bool remote)
{
    OB::RetryAttributes attrib;
    attrib.mode = mode;
    attrib.interval = interval;
    attrib.max = max;
    attrib.remote = remote;
    Any a;
    a <<= attrib;
    PolicyList policies;
    policies.length(1);
    policies[0] = orb -> create_policy(OB::RETRY_POLICY_ID, a);
    Object_var obj = orig -> _set_policy_overrides(policies, SET_OVERRIDE);
    return Retry::_narrow(obj);
}

static void
testRetry(ORB_ptr orb, RetryServer_ptr server)
{
    Retry_var retry;

    Retry_var orig = server -> get_retry_object();

    //
    // Test: Do not retry on remote exceptions
    //
    OB_INFO("Testing retry for remote exceptions... ");
    retry = getRetry(orb, orig, OB::RETRY_STRICT, 0, 0, false);
    retry -> raise_exception(1, false);
    try
    {
        retry -> aMethod();
        TEST(false);
    }
    catch(const TRANSIENT&)
    {
        // Expected
        TEST(retry -> get_count() == 1);
    }

    //
    // Test: Retry once on remote exceptions
    //
    retry = getRetry(orb, orig, OB::RETRY_STRICT, 0, 0, true);
    retry -> raise_exception(1, false);
    try
    {
        retry -> aMethod();
        TEST(retry -> get_count() == 2);
    }
    catch(const SystemException&)
    {
        TEST(false);
    }
    OB_INFO_LN("Done!");

    //
    // Test: Don't retry if completion status is COMPLETED_MAYBE and
    //       mode is RETRY_STRICT
    //
    OB_INFO("Testing RETRY_STRICT... ");
    retry = getRetry(orb, orig, OB::RETRY_STRICT, 0, 0, true);
    retry -> raise_exception(1, true);
    try
    {
        retry -> aMethod();
        TEST(false);
    }
    catch(const SystemException&)
    {
        // Expected
        TEST(retry -> get_count() == 1);
    }
    OB_INFO_LN("Done!");

    //
    // Test: RETRY_NEVER
    //
    OB_INFO("Testing RETRY_NEVER... ");
    retry = getRetry(orb, orig, OB::RETRY_NEVER, 0, 0, true);
    retry -> raise_exception(1, false);
    try
    {
        retry -> aMethod();
        TEST(false);
    }
    catch(const TRANSIENT&)
    {
        // Expected
        TEST(retry -> get_count() == 1);
    }
    OB_INFO_LN("Done!");

    //
    // Test: RETRY_ALWAYS
    //
    OB_INFO("Testing RETRY_ALWAYS... ");
    retry = getRetry(orb, orig, OB::RETRY_ALWAYS, 0, 0, true);
    retry -> raise_exception(1, true);
    try
    {
        retry -> aMethod();
        TEST(retry -> get_count() == 2);
    }
    catch(const SystemException&)
    {
        TEST(false);
    }
    OB_INFO_LN("Done!");

    //
    // Test: Retry five times
    //
    OB_INFO("Testing multiple retries... ");
    retry = getRetry(orb, orig, OB::RETRY_STRICT, 0, 0, true);
    retry -> raise_exception(5, false);
    try
    {
        retry -> aMethod();
        TEST(retry -> get_count() == 6);
    }
    catch(const SystemException&)
    {
        TEST(false);
    }

    //
    // Test: Retry five times with failure
    //
    retry = getRetry(orb, orig, OB::RETRY_STRICT, 0, 5, true);
    retry -> raise_exception(6, false);
    try
    {
        retry -> aMethod();
        TEST(false);
    }
    catch(const TRANSIENT&)
    {
        // Expected
        TEST(retry -> get_count() == 6);
    }
    OB_INFO_LN("Done!");

    //
    // Test: Retry interval
    //
    OB_INFO("Testing retry interval... ");
    retry = getRetry(orb, orig, OB::RETRY_STRICT, 100, 0, true);
    retry -> raise_exception(12, false);
    try
    {
        timeval start = OB::TimeNow();
        retry -> aMethod();
        timeval stop = OB::TimeNow();
        timeval diff = OB::TimeSub(stop, start);
        TEST(diff.tv_sec > 0);
    }
    catch(const SystemException&)
    {
        TEST(false);
    }
    OB_INFO_LN("Done!");
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;

    try
    {
        orb = ORB_init(argc, argv);

        Object_var obj = orb -> string_to_object("relfile:/Test.ref");
        if(CORBA::is_nil(obj))
        {
            OB_ERROR(argv[0] << ": cannot read IOR from Test.ref");
            return EXIT_FAILURE;
        }
        RetryServer_var server = RetryServer::_narrow(obj);
        TEST(!is_nil(server));

        testLocationForward(orb, server);
        testRetry(orb, server);

        server -> deactivate();
    }
    catch(const Exception& ex)
    {
        OB_ERROR(ex);
        status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
        try
        {
            orb -> destroy();
        }
        catch(const Exception& ex)
        {
            OB_ERROR(ex);
            status = EXIT_FAILURE;
        }
    }

    return status;
}
