// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <TestMacro.h>

#include <Test_impl.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_IO_H
#   include <io.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace PortableServer;

class FwdLocator_impl : public PortableServer::ServantLocator,
                        public OBCORBA::RefCountLocalObject
{
    Object_var obj_;

public:

    FwdLocator_impl(Object_ptr obj)
        : obj_(Object::_duplicate(obj))
    {
    }

    ~FwdLocator_impl()
    {
    }

    Servant
    preinvoke(const ObjectId& /*oid*/,
              POA_ptr /*poa*/,
              const char* /*operation*/,
              PortableServer::ServantLocator::Cookie& /*the_cookie*/)
        throw(ForwardRequest, SystemException)
    {
        throw ForwardRequest(obj_);

        return 0; // Some compilers need this
    }

    void
    postinvoke(const ObjectId& /*oid*/,
               POA_ptr /*poa*/,
               const char* /*operation*/,
               PortableServer::ServantLocator::Cookie /*the_cookie*/,
               Servant /*the_servant*/)
        throw(SystemException)
    {
    }
};

int
run(ORB_ptr orb, int /* argc */, char* argv[])
{
    //
    // Resolve Root POA
    //
    CORBA::Object_var poaObj = orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);
    
    //
    // Get a reference to the POA manager and activate it
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();
    manager -> activate();

    PolicyList policies;
    policies.length(3);
    policies[0] = rootPOA ->
     create_request_processing_policy(PortableServer::USE_SERVANT_MANAGER);
    policies[1] = rootPOA ->
        create_servant_retention_policy(PortableServer::NON_RETAIN);
    policies[2] = rootPOA -> create_implicit_activation_policy(
        PortableServer::NO_IMPLICIT_ACTIVATION);

    //
    // Create two POAs with servant locators, each with its own
    // POAManager (and therefore, each with unique endpoints).
    // The locators will simply raise ForwardRequest for the
    // given object, thus creating an infinite loop.
    //
    POA_var poa1 =
        rootPOA -> create_POA("poa1", POAManager::_nil(), policies);
    POA_var poa2 =
        rootPOA -> create_POA("poa2", POAManager::_nil(), policies);

    Object_var obj1 = poa1 -> create_reference("IDL:Test:1.0");
    Object_var obj2 = poa2 -> create_reference("IDL:Test:1.0");

    ServantLocator_var locator1 = new FwdLocator_impl(obj2);
    ServantLocator_var locator2 = new FwdLocator_impl(obj1);

    poa1 -> set_servant_manager(locator1);
    poa2 -> set_servant_manager(locator2);

    Test_var test = Test::_narrow(obj1);

    Retry_impl* retryImpl = new Retry_impl();
    Retry_var retry = retryImpl -> _this();

    RetryServer_impl* serverImpl = new RetryServer_impl(orb, test, retry);
    RetryServer_var server = serverImpl -> _this();

    //
    // Save reference. This must be done after POA manager
    // activation, otherwise there is a potential for a race
    // condition between the client sending a request and the
    // server not being ready yet.
    //
    const char* refFile = "Test.ref";
    OB_OFSTREAM out(refFile);
    if(out.fail())
    {
        OB_ERROR(argv[0] << ": can't open `" << refFile << "': "
                 << strerror(errno));
        return 1;
    }

    String_var ref = orb -> object_to_string(server);
    out << ref << OB_ENDL;
    out.close();

    //
    // Run implementation
    //
    POAManager_var mgr = poa1 -> the_POAManager();
    mgr -> activate();
    mgr = poa2 -> the_POAManager();
    mgr -> activate();
    orb -> run();

    //
    // Clean up
    //
    retryImpl -> _remove_ref();
    serverImpl -> _remove_ref();

#ifdef HAVE__UNLINK
    _unlink(refFile);
#else
    unlink(refFile);
#endif

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
        orb = CORBA::ORB_init(argc, argv);
        status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
        OB_ERROR(ex);
        status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
        try
        {
            orb -> destroy();
        }
        catch(const CORBA::Exception& ex)
        {
            OB_ERROR(ex);
            status = EXIT_FAILURE;
        }
    }

    return status;
}
