// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/BootManager.h>

#include <CosProperty_impl.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

#include <stdlib.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace OB;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options]\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show version number.\n"
	"-i, --ior              Print IOR on standard output.\n"
	;
}

int
run(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Resolve Root POA
    //
    PortableServer::POA_var rootPOA =
    PortableServer::POA::_narrow(Object_var(
	orb -> resolve_initial_references("RootPOA")));
    
    //
    // Get a reference to the POA manager
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();
    
    //
    // Create persistent POA
    //
    PolicyList policies;
    policies.length(2);
    policies[0] =
    rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] =
    rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);

    PortableServer::POA_var poa;
    try
    {
	poa = rootPOA -> create_POA("PropertyServicePOA", manager,
				    policies);
    }
    catch(const PortableServer::POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(poa));

    //
    // Get options
    //
    bool ior = false;

    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
	if(strcmp(argv[i], "--help") == 0 ||
	   strcmp(argv[i], "-h") == 0)
	{
	    usage(progName);
	    return EXIT_SUCCESS;
	}
	else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
	{
	    OB_CERR << "ORBacus Property Service " << Version << OB_ENDL;
	    return EXIT_SUCCESS;
	}
	else if(strcmp(argv[i], "--ior") == 0 ||
		strcmp(argv[i], "-i") == 0)
	{
	    ior = true;
	}
	else
	{
	    OB_CERR << progName << ": unknown option `"
		    << argv[i] << "'\n" << OB_ENDL;
	    usage(progName);
	    return EXIT_FAILURE;
	}
    }

    //
    // Create factory
    //
    CosPropertyService::PropertySetDefFactory_impl defFactory(rootPOA,
							      poa);
    PortableServer::ObjectId_var oid =
	PortableServer::string_to_ObjectId("PropertyService");
    poa -> activate_object_with_id(oid, &defFactory);
    Object_var facObj = defFactory._this();

    //
    // Register with BootManager
    //
    BootManager_var bootManager = BootManager::_narrow(
	Object_var(orb -> resolve_initial_references("BootManager")));
    bootManager -> add_binding(oid, facObj);

    //
    // Activate the POAManager before displaying the IOR so that
    // connection attempts are not rejected
    //
    manager -> activate();

    //
    // Print IOR on standard output
    //
    if(ior)
    {
	String_var s = orb -> object_to_string(facObj);
	OB_COUT << s << OB_ENDL;
    }

    //
    // Run implementation
    //
    orb -> run();
 
    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	//
	// Process the command line arguments into a property set
	//
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);
 
        //
        // Check properties first
        //
	OB::Properties::KeySeq keys =
	properties -> getKeys("ooc.property.");
	for(ULong i = 0; i < keys.length(); ++i)
	{
	    const char* key = keys[i];
	    const char* value = properties -> getProperty(key);
	    assert(value != 0);

	    if(strcmp(key, "ooc.property.endpoint") == 0)
	    {
		//
		// Set the ooc.orb.oa.endpoint property from this value if
		// not already set
		//
		if(properties -> getProperty("ooc.orb.oa.endpoint") == 0)
		    properties -> setProperty("ooc.orb.oa.endpoint", value);
	    }
	    else
	    {
		OB_CERR << argv[0] << ": warning: `" << key
		        << "': unknown property" << OB_ENDL;
	    }
	}

        //
        // Create ORB
        //
	orb = CORBA::ORB_init(argc, argv);

	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
