// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/CosProperty.h>

//
// Tests for constrained PropertySetDef
//
void
TestConstrained(CosPropertyService::PropertySetDefFactory_ptr factory)
{
    CosPropertyService::PropertySetDef_var set;
    CORBA::Any any;
    bool except;

    //
    // Set up PropertySetDef constraints, only tk_string is allowed
    //
    CosPropertyService::PropertyTypes allowedTypes;
    allowedTypes.length(1);
    allowedTypes[0] = CORBA::TypeCode::_duplicate(CORBA::_tc_string);

    any <<= "property_value";

    CosPropertyService::PropertyDefs allowedProperties;
    allowedProperties.length(3);
    CosPropertyService::PropertyDef p;

    p.property_name = CORBA::string_dup("read_only");
    p.property_value = any;
    p.property_mode = CosPropertyService::read_only;
    allowedProperties[0] = p;

    p.property_name = CORBA::string_dup("fixed_normal");
    p.property_value = any;
    p.property_mode = CosPropertyService::fixed_normal;
    allowedProperties[1] = p;

    p.property_name = CORBA::string_dup("fixed_readonly");
    p.property_value = any;
    p.property_mode = CosPropertyService::fixed_readonly;
    allowedProperties[2] = p;

    set = factory -> create_constrained_propertysetdef(allowedTypes,
						       allowedProperties);

    {
	OB_INFO("  Testing get_allowed_property_types()... ");

	CosPropertyService::PropertyTypes_var types;
	set -> get_allowed_property_types(types.out());

	assert(types -> length() == 1);

	assert(types[0] -> equal(CORBA::_tc_string));

	OB_INFO_LN("Done!");
    }

    {
	OB_INFO("  Testing get_allowed_properties()... ");

	CosPropertyService::PropertyDefs_var properties;
	set -> get_allowed_properties(properties);

	assert(properties -> length() == 3);

	OB_INFO_LN("Done!");
    }

    {
	OB_INFO("  Testing define_property_with_mode()... ");

	//
	// Try to add allowed properties
	//
	any <<= "property_value";
	try
	{
	    set -> define_property_with_mode("fixed_normal", any,
					     CosPropertyService::fixed_normal);
	    set -> define_property_with_mode("read_only", any,
					     CosPropertyService::read_only);
	    set -> define_property_with_mode(
		"fixed_readonly", any, CosPropertyService::fixed_readonly);
	}
	catch(const CosPropertyService::ReadOnlyProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::ConflictingProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedTypeCode&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedMode&)
	{
	    assert(false);
	}

	//
	// Try to change readonly property
	//
	any <<= "read_only";
	except = false;
	try
	{
	    set -> define_property("read_only", any);
	}
	catch(const CosPropertyService::ReadOnlyProperty&)
	{
	    // This exception must be thrown
	    except = true;
	}
	catch(const CosPropertyService::ConflictingProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedTypeCode&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}

	assert(except);

	//
	// Try to add illegal TypeCode
	//
	any <<= (CORBA::ULong)1;
	except = false;
	try
	{
	    set -> define_property("fixed_normal", any);
	}
	catch(const CosPropertyService::ReadOnlyProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::ConflictingProperty&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::UnsupportedProperty&)
	{
	    // This exception must be thrown
	    except = true;
	}
	catch(const CosPropertyService::UnsupportedTypeCode&)
	{
	    // This exception must be thrown
	    except = true;
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}
      
	assert(except);

	OB_INFO_LN("Done!");
    }

    {
	OB_INFO("  Testing get_property_mode()... ");

	CosPropertyService::PropertyModeType fixed_normal;
	CosPropertyService::PropertyModeType read_only;
	CosPropertyService::PropertyModeType fixed_readonly;
	try
	{
	    fixed_normal = set -> get_property_mode("fixed_normal");
	    read_only = set -> get_property_mode("read_only");
	    fixed_readonly = set -> get_property_mode("fixed_readonly");
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::PropertyNotFound&)
	{
	    assert(false);
	}

	assert(fixed_normal == CosPropertyService::fixed_normal ||
               read_only == CosPropertyService::read_only ||
               fixed_readonly == CosPropertyService::fixed_readonly);

	OB_INFO_LN("Done!");
    }

    {
	OB_INFO("  Testing delete_property()... ");

	except = false;
	//
	// Try to delete fixed_normal property;
	//
	try
	{
	    set -> delete_property("fixed_normal");
	}
	catch(const CosPropertyService::FixedProperty&)
	{
	    // This exception must be thrown
	    except = true;
	}
	catch(const CosPropertyService::InvalidPropertyName&)
	{
	    assert(false);
	}
	catch(const CosPropertyService::PropertyNotFound&)
	{
	    assert(false);
	}

	assert(except);

	OB_INFO_LN("Done!");
    }
}
