// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/TimeBase.h>
#include <OB/CosTime.h>
#include <OB/TimeHelper.h> 

#include <CosTime_impl.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

using namespace CORBA;
using namespace TimeBase;
using namespace OB;

//
// The UTO implementation
//
class UTO_impl : public CosTime::UTO,
                 public OBCORBA::RefCountLocalObject
{
    //
    // Time in units of 100 nanoseconds
    //
    TimeT time_;

    //
    // Time inaccuracy in units of 100 nanoseconds
    //
    InaccuracyT inaccuracy_;
 
    //
    // Time displacement from Greenwich meridian in minutes
    //
    TdfT tdf_;

public:

    UTO_impl(TimeT, InaccuracyT, TdfT);
    UTO_impl(TimeT, InaccuracyT);
    UTO_impl(const UtcT&);

    //
    // Standard IDL to C++ Mapping
    //
    virtual TimeT time()
        throw(SystemException);

    virtual InaccuracyT inaccuracy()
        throw(SystemException);

    virtual TdfT tdf()
        throw(SystemException);

    virtual UtcT utc_time()
        throw(SystemException);

    virtual CosTime::UTO_ptr absolute_time()
        throw(SystemException);

    virtual CosTime::TimeComparison compare_time(CosTime::ComparisonType,
						CosTime::UTO_ptr)
        throw(SystemException);

    virtual CosTime::TIO_ptr time_to_interval(CosTime::UTO_ptr)
        throw(SystemException);

    virtual CosTime::TIO_ptr interval()
        throw(SystemException);
};

//
// The TIO implementation
//
class TIO_impl : public CosTime::TIO,
                 public OBCORBA::RefCountLocalObject
{
    //
    // Lower range
    //
    TimeT lower_;

    //
    // Upper range
    //
    TimeT upper_;
    
    CosTime::OverlapType checkOverlap(TimeT, TimeT, CosTime::TIO_out);

public:

    TIO_impl(TimeT, TimeT);

    //
    // Standard IDL to C++ Mapping
    //
    virtual IntervalT time_interval()
        throw(SystemException);

    virtual CosTime::OverlapType spans(CosTime::UTO_ptr, CosTime::TIO_out)
        throw(SystemException);

    virtual CosTime::OverlapType overlaps(CosTime::TIO_ptr, CosTime::TIO_out)
        throw(SystemException);

    virtual CosTime::UTO_ptr time()
        throw(SystemException);
};

// ----------------------------------------------------------------------
// TimeService_impl constructor/destructor
// ----------------------------------------------------------------------

TimeService_impl::TimeService_impl(InaccuracyT inaccuracy)
    throw(SystemException)
    : inaccuracy_(inaccuracy)
{
    if(inaccuracy > OB::TimeHelper::MaxInaccuracyT)
	throw BAD_PARAM();
}

// ----------------------------------------------------------------------
// TimeService_impl public member implementation
// ----------------------------------------------------------------------

CosTime::UTO_ptr
TimeService_impl::universal_time()
    throw(CosTime::TimeUnavailable, SystemException)
{
    return new UTO_impl(OB::TimeHelper::utcNow(inaccuracy_));
}

CosTime::UTO_ptr
TimeService_impl::secure_universal_time()
    throw(CosTime::TimeUnavailable, SystemException)
{
    throw CosTime::TimeUnavailable();
    return 0; // Some compilers need this
}

CosTime::UTO_ptr
TimeService_impl::new_universal_time(TimeT time,
				     InaccuracyT inaccuracy,
				     TdfT tdf)
    throw(SystemException)
{
    if(inaccuracy > OB::TimeHelper::MaxInaccuracyT)
	throw BAD_PARAM();

    return new UTO_impl(time, inaccuracy, tdf);
}

CosTime::UTO_ptr
TimeService_impl::uto_from_utc(const UtcT& utc)
    throw(SystemException)
{
    return new UTO_impl(utc);
}

CosTime::TIO_ptr
TimeService_impl::new_interval(TimeT lower, TimeT upper)
    throw(SystemException)
{
    if(lower > upper)
	throw BAD_PARAM();

    return new TIO_impl(lower, upper);
}

// ----------------------------------------------------------------------
// UTO_impl constructor/destructor
// ----------------------------------------------------------------------

UTO_impl::UTO_impl(TimeT time,
		   InaccuracyT inaccuracy,
		   TdfT tdf)
    : time_(time), inaccuracy_(inaccuracy), tdf_(tdf)
{
}

UTO_impl::UTO_impl(TimeT time,
		   InaccuracyT inaccuracy)
    : time_(time), inaccuracy_(inaccuracy), tdf_(0)
{
}

UTO_impl::UTO_impl(const UtcT& utc)
    : time_(utc.time),
      inaccuracy_(((ULongLong)utc.inacclo) |
		  (((ULongLong)utc.inacchi) << 32)),
      tdf_(utc.tdf)
{
}

// ----------------------------------------------------------------------
// UTO_impl public member implementation
// ----------------------------------------------------------------------

TimeT
UTO_impl::time()
    throw(SystemException)
{
    return time_;
}

InaccuracyT
UTO_impl::inaccuracy()
    throw(SystemException)
{
    return inaccuracy_;
}

TdfT
UTO_impl::tdf()
    throw(SystemException)
{
    return tdf_;
}

UtcT
UTO_impl::utc_time()
    throw(SystemException)
{
    return OB::TimeHelper::toUtcT(time_, inaccuracy_, tdf_);
}

CosTime::UTO_ptr
UTO_impl::absolute_time()
    throw(SystemException)
{
    //
    // time_ is an interval in this case
    //
    UtcT now = OB::TimeHelper::utcNow();

    //
    // Will we overflow?
    //
    if(OB::TimeHelper::MaxTimeT - now.time < time_)
	throw DATA_CONVERSION();

    return new UTO_impl(now + time_);
}

CosTime::TimeComparison
UTO_impl::compare_time(CosTime::ComparisonType comparison_type,
		       CosTime::UTO_ptr uto)
    throw(SystemException)
{
    if(is_nil(uto))
	throw BAD_PARAM();

    TimeT otherTime = uto -> time();
    InaccuracyT otherInacc = uto -> inaccuracy();

    if(inaccuracy_ > OB::TimeHelper::MaxInaccuracyT)
	throw BAD_PARAM();

    IntervalT my = OB::TimeHelper::toIntervalT(time_, inaccuracy_);
    IntervalT other = OB::TimeHelper::toIntervalT(otherTime, otherInacc);

    if(comparison_type == CosTime::IntervalC)
    {
	//
	// For IntervalC comparison, two UTOs are deemed to contain
	// the same time only if the Time attribute of the two objects
	// are equal and the Inaccuracy attributes of both the objects
	// are zero.
	//
	if(time_ == otherTime && inaccuracy_ == 0 && otherInacc == 0)
	    return CosTime::TCEqualTo;

	//
	// In an IntervalC comparison, TCIndeterminate value is
	// returned if the error envelopes around the two times being
	// compared overlap.
	//
	// <------- t1 -------->  <--------- t2 ------------>
	//
	// First test left hand side of t1 against right hand side of
	// t2.  Next test right hand side of t1 against left hand side
	// of t2.  If neither then there are overlapping areas.
	//
	if(my.upper_bound < other.lower_bound)
	    return CosTime::TCLessThan;
	else if(my.lower_bound > other.upper_bound)
	    return CosTime::TCGreaterThan;
	else
	    return CosTime::TCIndeterminate;
    }
    else
    {
	//
	// MidC comparison just compares the base times. A MidC
	// comparison can never return TCIndeterminate.
	//
	if(time_ == otherTime)
	    return CosTime::TCEqualTo;
	else if(time_ < otherTime)
	    return CosTime::TCLessThan;
	else
	    return CosTime::TCGreaterThan;
    }
}

CosTime::TIO_ptr
UTO_impl::time_to_interval(CosTime::UTO_ptr uto)
    throw(SystemException)
{
    if(is_nil(uto))
	throw BAD_PARAM();

    TimeT t = uto -> time();

    TIO_impl* tioImpl;
    if(time_ < t)
	tioImpl = new TIO_impl(time_, t);
    else
	tioImpl = new TIO_impl(t, time_);

    return tioImpl;
}

CosTime::TIO_ptr
UTO_impl::interval()
    throw(SystemException)
{
    IntervalT inter = OB::TimeHelper::toIntervalT(time_, inaccuracy_);

    return new TIO_impl(inter.lower_bound, inter.upper_bound);
}

// ----------------------------------------------------------------------
// TIO_impl private member implementation
// ----------------------------------------------------------------------

CosTime::OverlapType
TIO_impl::checkOverlap(TimeT lower,
		       TimeT upper,
		       CosTime::TIO_out overlap)
{
    //
    // First handle no overlap
    //
    if(upper_ < lower)
    {
	overlap = new TIO_impl(upper_, lower);
	return CosTime::OTNoOverlap;
    }
    else if(lower_ > upper)
    {
	overlap = new TIO_impl(upper, lower_);
	return CosTime::OTNoOverlap;
    }
    //
    // Is this TIO contained in the interval or vice versa? The spec
    // isn't clear what happens if the intervals are exactly the
    // same. This code will return Contained.
    //
    else if(lower_ >= lower && upper_ <= upper)
    {
	overlap = CosTime::TIO::_duplicate(this);
	return CosTime::OTContained;
    }
    else if(lower_ < lower && upper_ > upper)
    {
	overlap = new TIO_impl(lower, upper);
	return CosTime::OTContainer;
    }
    //
    // Overlap
    //
    else if(upper_ > upper)
    {
	overlap = new TIO_impl(lower_, upper);
    }
    else
    {
	overlap = new TIO_impl(lower, upper_);
    }

    return CosTime::OTOverlap;
}

// ----------------------------------------------------------------------
// TIO_impl constructor/destructor
// ----------------------------------------------------------------------

TIO_impl::TIO_impl(TimeT lower, TimeT upper)
    : lower_(lower), upper_(upper)
{
}

// ----------------------------------------------------------------------
// TIO_impl public member implementation
// ----------------------------------------------------------------------

IntervalT
TIO_impl::time_interval()
    throw(SystemException)
{
    IntervalT interval;

    interval.lower_bound = lower_;
    interval.upper_bound = upper_;

    return interval;
}

CosTime::OverlapType
TIO_impl::spans(CosTime::UTO_ptr uto, CosTime::TIO_out overlap)
    throw(SystemException)
{
    if(is_nil(uto))
	throw BAD_PARAM();

    TimeT time = uto -> time();
    InaccuracyT inacc = uto -> inaccuracy();

    if(inacc > OB::TimeHelper::MaxInaccuracyT)
	throw BAD_PARAM();

    IntervalT inter = OB::TimeHelper::toIntervalT(time, inacc);

    return checkOverlap(inter.lower_bound, inter.upper_bound, overlap);
}

CosTime::OverlapType
TIO_impl::overlaps(CosTime::TIO_ptr interval, CosTime::TIO_out overlap)
    throw(SystemException)
{
    if(is_nil(interval))
	throw BAD_PARAM();

    IntervalT inter = interval -> time_interval();

    if(inter.lower_bound > inter.upper_bound)
	throw BAD_PARAM();

    return checkOverlap(inter.lower_bound, inter.upper_bound, overlap);
}

CosTime::UTO_ptr
TIO_impl::time()
    throw(SystemException)
{
    //
    // Note that "time = (upper_ + lower_) / 2" may cause overflow.
    //
    InaccuracyT inaccuracy = (upper_ - lower_) / 2;
    TimeT time = lower_ + inaccuracy;

    //
    // Is inaccuracy too big?
    //
    if(inaccuracy > OB::TimeHelper::MaxInaccuracyT)
	throw DATA_CONVERSION();

    return new UTO_impl(time, inaccuracy);
}

// ----------------------------------------------------------------------
