// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/PortableInterceptor.h>
#include <OB/IORUtil.h>

#include <OB/Callstack.h>
#include <CorbaOut.h>
#include <RILogger.h>
#include <CRILogger_impl.h>

using namespace OB;

// ---------------------------------------------------------------------
// CRILogger private member implementation
// ---------------------------------------------------------------------

void
CRILogger_impl::displayClientRequestInfo(
    PortableInterceptor::ClientRequestInfo_ptr info)
{
    displayRequestInfo(info);

    if(logLevel_ > 3)
    {
	//
	// Display effective profile
	//
	IOP::TaggedProfile_var profile = info -> effective_profile();

	out_ -> skip();
	str_ += "effective profile = ";
	if (profile -> tag == IOP::TAG_INTERNET_IOP)
	    str_ += "TAG_INTERNET_IOP";
	else if (profile -> tag == IOP::TAG_MULTIPLE_COMPONENTS)
	    str_ += "TAG_MULTIPLE_COMPONENTS";
	else
	    str_ += "UNKNOWN_TAG";
	str_ += ": ";
	str_ += '\n';

	CORBA::OctetSeq_var octets =  
	    new CORBA::OctetSeq(profile -> profile_data.length());
	octets -> length(profile -> profile_data.length());
	for (CORBA::ULong i = 0; i < profile -> profile_data.length(); i++)
	{
	    octets[i] = profile -> profile_data[i];
	} 
	CORBA::String_var str = dumpOctetSeq("        ", octets);
	str_ += str;
     }

    //
    // Display received exception
    //
    try
    {
	CORBA::Any_var exception = info -> received_exception();

	out_ -> skip();
    	str_ += "received exception = ";
	str_ += '\n';
	out_ -> down();
	out_ -> display_Any(exception);
	out_ -> up();
    }
    catch(const CORBA::BAD_INV_ORDER&)
    {
	// Operation not supported in this context
    }

    //
    // Display received exception id
    //
    try
    {
	CORBA::RepositoryId exception_id = info -> received_exception_id();
	out_ -> skip();
    	str_ += "received exception id = ";
	str_ += exception_id;
	str_ += '\n';
    }
    catch(const CORBA::BAD_INV_ORDER&)
    {
	// Operation not supported in this context
    }
}

// ---------------------------------------------------------------------
// CRILogger constructor/destructor
// ---------------------------------------------------------------------

CRILogger_impl::CRILogger_impl(PortableInterceptor::ORBInitInfo_ptr info,
			       PortableInterceptor::SlotId slotId,
			       CORBA::ULong logLevel,
			       OB::WLogger_ptr logger)
    : RILogger(info, slotId, logLevel, logger)
{
}

CRILogger_impl::~CRILogger_impl()
{
}

// ----------------------------------------------------------------------
// IDL to C++ Mapping
// ----------------------------------------------------------------------

void
CRILogger_impl::destroy()
{
    // Intentionally empty
}

char*
CRILogger_impl::name()
{
    return CORBA::string_dup("com.ooc.watson.CRILogger_impl");
}

void
CRILogger_impl::send_request(PortableInterceptor::ClientRequestInfo_ptr ri)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
	//
	// Try to get the CallStack from slot
	//
	try
	{
	    CORBA::Any_var any = ri -> get_slot(mySlotId_);
	    CORBA::TypeCode_var tc = any -> type();
	    if(tc -> kind() != CORBA::tk_null)
	    {

		//
		// Encode CallStack
		//
		CORBA::OctetSeq_var data = cdrCodec_ -> encode_value(any);

		//
		// Add encoded CallStack to service context
		//
		IOP::ServiceContext sc;
		sc.context_id = REQUEST_CONTEXT_ID;
		sc.context_data.length(data -> length());
		memcpy(sc.context_data.get_buffer(), 
		       data -> get_buffer(),
		       data -> length());
		ri -> add_request_service_context(sc, false);
	    }
	}
	catch(const PortableInterceptor::InvalidSlot&)
	{
	    // Ignore failure
	}
	catch(const IOP::Codec::InvalidTypeForEncoding&)
	{
	    // Ignore failure
	}

        displayClientRequestInfo(ri);
	logTraceOutput(L"send request");
    }
}

void
CRILogger_impl::send_poll(PortableInterceptor::ClientRequestInfo_ptr ri)
    throw(CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
        displayClientRequestInfo(ri);
	logTraceOutput(L"send poll");
    }
}

void
CRILogger_impl::receive_reply(PortableInterceptor::ClientRequestInfo_ptr ri)
    throw(CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
        displayClientRequestInfo(ri);
	logTraceOutput(L"receive reply");
    }
}

void
CRILogger_impl::receive_other(PortableInterceptor::ClientRequestInfo_ptr ri)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
        displayClientRequestInfo(ri);
	logTraceOutput(L"receive other");
    }
}

void
CRILogger_impl::receive_exception(PortableInterceptor::ClientRequestInfo_ptr ri)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
        displayClientRequestInfo(ri);
	logTraceOutput(L"receive exception");
    }
}

