// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/PortableInterceptor.h>

#include <OB/Callstack.h>

#include <CorbaOut.h>
#include <RILogger.h>
#include <SRILogger_impl.h>

using namespace OB;

// ---------------------------------------------------------------------
// SRILogger private member implementation
// ---------------------------------------------------------------------

void
SRILogger_impl::displayServerRequestInfo(
    PortableInterceptor::ServerRequestInfo_ptr info)
{
    displayRequestInfo(info);

    //
    // Display sending exception
    //
    try
    {
	CORBA::Any_var exception = info -> sending_exception();
	out_ -> skip();
    	str_ += "sending exception =\n";
	out_ -> down();
	out_ -> display_Any(exception);
	out_ -> up();
    }
    catch(const CORBA::BAD_INV_ORDER&)
    {
	// Operation not supported in this context
    }

    if(logLevel_ > 3)
    {
	CORBA::String_var str;
	//
	// Display object id
	//
	try
	{
	    out_ -> skip();
	    str_ += "object id =\n";
	    str = dumpOctetSeq("        ",
			       *(info -> object_id()));
	    str_ += str;
	}
	catch(const CORBA::BAD_INV_ORDER&)
	{
	    // Operation not supported in this context
	}

	//
	// Display adapter id
	//
	try
	{
	    out_ -> skip();
	    str_ += "adapter id =\n";
	    str = dumpOctetSeq("        ", 
			       *(info -> adapter_id()));
	    str_ += str;
	}
	catch(const CORBA::BAD_INV_ORDER&)
	{
	    // Operation not supported in this context
	}
    }

    //
    // Display target most derived interface
    //
    try
    {
	CORBA::RepositoryId tmdi = info -> target_most_derived_interface();

	out_ -> skip();
    	str_ += "target most derived interface = ";
	str_ += tmdi;
	str_ += '\n';
    }
    catch(const CORBA::BAD_INV_ORDER&)
    {
	// Operation not supported in this context
    }
}

// ---------------------------------------------------------------------
// SRILogger constructor/destructor
// ---------------------------------------------------------------------

SRILogger_impl::SRILogger_impl(PortableInterceptor::ORBInitInfo_ptr info,
			       PortableInterceptor::SlotId slotId,
			       CORBA::ULong logLevel,
			       OB::WLogger_ptr logger)
    : RILogger(info, slotId, logLevel, logger)
{
}

SRILogger_impl::~SRILogger_impl()
{
}

// ----------------------------------------------------------------------
// IDL to C++ Mapping
// ----------------------------------------------------------------------

void
SRILogger_impl::destroy()
{
    // Intentionally empty
}

char*
SRILogger_impl::name()
{
    return CORBA::string_dup("com.ooc.watson.SRILogger_impl");
}

void
SRILogger_impl::receive_request_service_contexts(
    PortableInterceptor::ServerRequestInfo_ptr ri)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
        displayServerRequestInfo(ri);

	//
	// Get CallStack from service context
	//
	watson::CallStack callStack(0);
	try
	{
	    //
	    // Get encoded CallStack from service context
	    //
	    IOP::ServiceContext_var sc =
		ri -> get_request_service_context(REQUEST_CONTEXT_ID);

	    //
	    // Decode CallStack
	    //
	    CORBA::Any_var any;
	    CORBA::OctetSeq data(sc -> context_data.length());
	    data.length(sc -> context_data.length());
	    memcpy(data.get_buffer(), sc -> context_data.get_buffer(),
		   sc -> context_data.length());
	    try
	    {
		any = cdrCodec_ -> decode_value(data, watson::_tc_CallStack);
	    }
	    catch(const IOP::Codec::TypeMismatch&)
	    {
		// Ignore failure
	    }
	    catch(const IOP::Codec::FormatMismatch&)
	    {
		// Ignore failure
	    }

	    const watson::CallStack* stack = 0;
	    if(!(any >>= stack))
	    {
		// Ignore failure
	    }
	    assert(stack);

	    //
	    // Set length of new CallStack to the length of
	    // the old CallStack, plus one for the new call
	    //
	    callStack.length(stack -> length() + 1);

	    //
	    // Copy old CallStack to new CallStack
	    //
	    for(CORBA::ULong i = 0 ; i < stack -> length() ; i++)
		callStack[i] = (*stack)[i];
	}
	catch(const CORBA::BAD_PARAM&)
	{
	    //
	    // Service context was not previously set -- set length
	    // of the new CallStack to one for the new call
	    //
	    callStack.length(1);
	}

	//
	// Add new call to the CallStack
	//
	CORBA::String_var op(ri -> operation());
	callStack[callStack.length() - 1] = op;

	//
	// Save the new CallStack in slot
	//
	CORBA::Any any;
	any <<= callStack;
	try
	{
	    ri -> set_slot(mySlotId_, any);
	}
	catch(const PortableInterceptor::InvalidSlot&)
	{
	    // Ignore failure
	}

	//
	// Print the CallStack
	//
	out_ -> skip();
	str_ += "stack =\n";
	out_ -> down();
	for(CORBA::ULong i = 0 ; i < callStack.length() ; i++)
	{
	    out_ -> skip();
	    str_ += (i + 1);
	    str_ += ".\t";
	    str_ += callStack[i];
	    str_ += '\n';
	}
	out_ -> up();

	logTraceOutput(L"receive request service contexts");
    }
}

void
SRILogger_impl::receive_request(
    PortableInterceptor::ServerRequestInfo_ptr ri)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
        displayServerRequestInfo(ri);
	logTraceOutput(L"receive request");
    }
}
    
void
SRILogger_impl::send_reply(
    PortableInterceptor::ServerRequestInfo_ptr ri)
    throw(CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
        displayServerRequestInfo(ri);
	logTraceOutput(L"send reply");
    }
}
    
void
SRILogger_impl::send_other(
    PortableInterceptor::ServerRequestInfo_ptr ri)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
        displayServerRequestInfo(ri);
	logTraceOutput(L"send other");
    }
}

void
SRILogger_impl::send_exception(
    PortableInterceptor::ServerRequestInfo_ptr ri)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
    if(logLevel_ > 0)
    {
        displayServerRequestInfo(ri);
	logTraceOutput(L"send exception");
    }
}









