// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OCI_IIOP.h>
#include <OB/Properties.h>
#include <OB/Logger.h>
#include <OB/OBPolicies.h>
#include <OB/Net.h>
#include <OB/LocalObject.h>

#include <IMRLogger.h>
#include <IMRUtil.h>
#include <OADEndpointManager_impl.h>
#include <OADDatabase.h>
#include <ServerDatabase.h>
#include <ServerStarter.h>
#include <OSLayer.h>

using namespace IMR;

//
// CloseCB implementation class. An instance of this class is
// registered with each connected oad. When the connection from the
// OAD to the IMR is closed (ie: the OAD dies) the close_cb method is
// called by the transport.
//
class OADCloseCB_impl : public OCI::CloseCB,
			public OBCORBA::RefCountLocalObject
{
    //
    // The message logger
    //
    IMROAD::Logger_var logger_;

    //
    // An OAD registrar instance
    //
    OADEndpointManager_impl* oadRegistrar_;
    CORBA::String_var host_; // The host name

public:

    OADCloseCB_impl(OADEndpointManager_impl*, const char*, IMROAD::Logger_ptr);
    ~OADCloseCB_impl();

    //
    // Standard IDL to C++ Mapping
    //

    virtual void close_cb(OCI::TransportInfo_ptr);
};

// ----------------------------------------------------------------------
// OADCloseCB_impl constructor and destructor
// ----------------------------------------------------------------------

OADCloseCB_impl::OADCloseCB_impl(OADEndpointManager_impl* registrar,
                                 const char* host,
				 IMROAD::Logger_ptr logger)
      : logger_(IMROAD::Logger::_duplicate(logger)),
        oadRegistrar_(registrar),
        host_(CORBA::string_dup(host))
{
}

OADCloseCB_impl::~OADCloseCB_impl()
{
}

// ----------------------------------------------------------------------
// OADCloseCB_impl public member functions
// ----------------------------------------------------------------------

void
OADCloseCB_impl::close_cb(OCI::TransportInfo_ptr info)
{
    if(logger_ -> tracePeerStatus() >= 2)
    {

	CORBA::String_var msg = CORBA::string_dup("close_cb: ");
	msg += host_;
	
	OCI::IIOP::TransportInfo_var iiopInfo =
	    OCI::IIOP::TransportInfo::_narrow(info);
	
	if(!CORBA::is_nil(iiopInfo))
	{
	    OCI::IIOP::InetAddr_var remoteAddr = iiopInfo -> remote_addr();
	    CORBA::UShort remotePort = iiopInfo -> remote_port();
	    
	    msg += ": ";
	    msg += (int)remoteAddr[0];
	    msg += '.';
	    msg += (int)remoteAddr[1];
	    msg += '.';
	    msg += (int)remoteAddr[2];
	    msg += '.';
	    msg += (int)remoteAddr[3];
	    msg += ':';
	    msg += remotePort;
	}
	else
	    msg += " (not iiop)";
	logger_ -> trace("IMR", msg);
    }
    oadRegistrar_ -> _OB_disconnect(host_);
}

// ----------------------------------------------------------------------
// OADEndpointManager_impl constructor/destructor implementation
// ----------------------------------------------------------------------

void
OADEndpointManager_impl::updateServersForOAD(
    IMRDB_4_0_2::OADRecord& rec, const OAD::ProcessList& processes)
{
    JTCSynchronized sync(*db_);
	
    if(logger_ -> tracePeerStatus() >= 2)
    {
        CORBA::String_var msg = CORBA::string_dup("OAD for ");
        msg += rec.host;
        msg += " processes: ";
        if(processes.length() == 0)
    	    msg += "EMPTY";
        for(CORBA::ULong i = 0; i < processes.length(); i++)
        {
    	    msg += processes[i];
    	    if(i != processes.length()-1)
    	        msg += ", ";
        }
        logger_ -> trace("IMR", msg);
    }

    IMRDB_4_1_0::ServerIdSeq_var servers = db_ -> queryAllByHost(rec.host);
    CORBA::ULong i;
    for(i = 0; i < servers -> length(); i++)
    {
        ServerRecord_var rec = db_ -> queryRecordById(servers[i]);
        CORBA::ULong p;
        for(p = 0; p < processes.length(); p++)
	    if(strcmp(processes[p], rec -> record().server) == 0)
	        break;
        //
        // If the record isn't in the list of active
        // processes then it's STOPPED.
        //
        if(p >= processes.length() && 
	   (rec -> record().status != STOPPED ||
	    rec -> record().timesForked != 0))
        {
            //
            // TODO: This should be some common code -- it's
            // shared in several places
            //
            rec -> record().timesForked = 0;
            rec -> record().status = STOPPED;
            rec -> record().updateTime = OSLayer::time();
                
            //
            // Run through the POAs and invalidate all IORs
            //
            for(CORBA::ULong p = 0 ; p < rec -> record().poas.length() ; p++)
                rec -> record().poas[p].state = IMR::NON_EXISTENT;

	    if(logger_ -> tracePeerStatus() >= 2)
	    {
		CORBA::String_var msg = CORBA::string_dup("updating: ");
		msg += rec -> record().server;
		msg += " state: ";
		msg += IMRUtil::serverStatusToStr(rec -> record().status);
		logger_ -> trace("IMR", msg);
	    }

	    db_ -> updateRecord(rec);
        }

        //
        // Start any persistent servers that are not running
        //
        if(rec -> record().mode == IMR::PERSISTENT &&
           rec -> record().status == IMR::STOPPED)
            serverStarter_ -> add(rec -> record().server);
    }
}

// ----------------------------------------------------------------------
// OADEndpointManager_impl constructor/destructor implementation
// ----------------------------------------------------------------------

OADEndpointManager_impl::OADEndpointManager_impl(
    CORBA::ORB_ptr orb,
    PortableServer::POA_ptr poa,
    IMROAD::Logger_ptr logger,
    OADDatabase* oaddb,
    ServerDatabase* db, 
    ServerStarter* serverStarter)
    : orb_(OBCORBA::ORB::_narrow(orb)),
      poa_(PortableServer::POA::_duplicate(poa)),
      logger_(IMROAD::Logger::_duplicate(logger)),
      oaddb_(oaddb),
      db_(db),
      serverStarter_(serverStarter)
{
    CORBA::Object_var obj =
        orb -> resolve_initial_references("OCICurrent");
    current_ = OCI::Current::_narrow(obj);
    assert(!CORBA::is_nil(current_));

    CORBA::Any any;
    any <<= OB::RETRY_ALWAYS;
    retryPolicy_ = orb -> create_policy(OB::RETRY_POLICY_ID, any);
}

OADEndpointManager_impl::~OADEndpointManager_impl()
{
}

// ----------------------------------------------------------------------
// OADEndpointManager_impl public member implementation
// ----------------------------------------------------------------------

//
// IDL:ooc.com/IMR/OADEndpointManager/register_oad:1.0
//
void
OADEndpointManager_impl::register_oad(
    const char* host,
    const OAD::ProcessList& processes,
    OAD::ProcessController_ptr controller,
    OAD::ProcessEndpointManager_ptr endpointManager,
    ServerObserver_out notifier)
    throw(IMR::NoSuchOAD, CORBA::SystemException)
{
    JTCSynchronized sync(*oaddb_);

    OCI::TransportInfo_var info = current_ -> get_oci_transport_info();

    if(logger_ -> tracePeerStatus() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("register_oad: ");
	msg += host;
	logger_ -> trace("IMR", msg);
    }

    IMRDB_4_0_2::OADRecord rec;
    if(!oaddb_ -> queryRecordByHost(host, rec, true))
    {
    	//
	// OAD has not been registered yet so add it
	//
        rec.host = CORBA::string_dup(host);
	oaddb_ -> addRecord(rec);
    }

    CORBA::PolicyList pl(1);
    pl.length(1);
    pl[0] = CORBA::Policy::_duplicate(retryPolicy_);
    CORBA::Object_var obj =
        controller -> _set_policy_overrides(pl, CORBA::ADD_OVERRIDE);
    rec.controller = OAD::ProcessController::_narrow(obj);
    rec.manager =
        OAD::ProcessEndpointManager::_duplicate(endpointManager);
    rec.up = true;
    
    oaddb_ -> updateRecord(rec);

    //
    // Update the servers list for this OAD.
    //
    updateServersForOAD(rec, processes);

    notifier = ServerObserver::_duplicate(observer_);

    //
    // Add the close callback to the current transport. If info is
    // nil then it is a collocated oad and do not need the close
    // callback.
    //
    if(!CORBA::is_nil(info))
    {
        OCI::CloseCB_var cb = new OADCloseCB_impl(this, host, logger_);
        info -> add_close_cb(cb);
    }
}

// ----------------------------------------------------------------------
// OADEndpointManager_impl private member implementation
// ----------------------------------------------------------------------

void
OADEndpointManager_impl::_OB_setNotifier(IMR::ServerObserver_ptr observer)
{
    observer_ = IMR::ServerObserver::_duplicate(observer);
}

void
OADEndpointManager_impl::_OB_connectOADs(bool collocated)
{
    //
    // TODO: datebase updates should be smarter.
    //

    char hostname[MAXHOSTNAMELEN + 1];
    if(collocated)
        gethostname(hostname, MAXHOSTNAMELEN) ;

    IMRDB_4_0_2::OADRecordSeq_var seq;
    {
        JTCSynchronized sync(*oaddb_);

        seq = oaddb_ -> getAllRecords();
        for(CORBA::ULong i = 0; i < seq -> length(); i++)
        {
	    if(collocated && strcmp(hostname, seq[i].host) == 0)
	        continue;

            seq[i].up = false;
            seq[i].manager = OAD::ProcessEndpointManager::_nil();

            oaddb_ -> updateRecord(seq[i]);
        }
    }

    OB::Properties_var properties = orb_ -> properties();
    const char* port = properties -> getProperty("ooc.imr.slave_port");
    if(port == 0)
        port = properties -> getProperty("ooc.orb.oa.port");

    for(CORBA::ULong i = 0; i < seq -> length(); i++)
    {
	if(collocated && strcmp(hostname, seq[i].host) == 0)
	    continue;

        try
        {
	    seq[i].controller -> connect_imr(_this());
        }
	catch(const IMR::NoSuchOAD& ex)
	{
	    CORBA::String_var msg =
	        CORBA::string_dup("IMR: Contacted OAD not registered: ");
	    msg += ex.name;
	    logger_ -> warning(msg);
	}
        catch(const CORBA::SystemException&)
        {
            CORBA::String_var msg =
                CORBA::string_dup("IMR: Could not contact OAD on host ");
            msg += seq[i].host;
            logger_ -> warning(msg);
        }
    }
}

//
// Called on a connection close callback
//
void
OADEndpointManager_impl::_OB_disconnect(const char* host)
{
    JTCSynchronized sync(*oaddb_);

    if(logger_ -> tracePeerStatus() >= 2)
    {
        CORBA::String_var msg = CORBA::string_dup("connection_close: ");
        msg += host;
        logger_ -> trace("IMR", msg);
    }

    //
    // Mark the OAD as dead
    //
    IMRDB_4_0_2::OADRecord rec;
    if(!oaddb_ -> queryRecordByHost(host, rec, true))
        throw NoSuchOAD(host);

    rec.up = false;
    rec.manager = OAD::ProcessEndpointManager::_nil();
    oaddb_ -> updateRecord(rec);
}

PortableServer::POA_ptr
OADEndpointManager_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}
