// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/NTService.h>
#include <OB/LocalObject.h>
#include <OB/Logger.h>
#include <OB/Except.h>

using namespace CORBA;
using namespace OB;

//
// Static function to return a formatted error message for the last
// WIN32 error.
//
static char*
GetErrorText()
{
    DWORD rc;
    char* str = 0;
    char* cstr = 0;
    rc = ::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|
		        FORMAT_MESSAGE_FROM_SYSTEM|
		        FORMAT_MESSAGE_ARGUMENT_ARRAY,
		        0, GetLastError(), LANG_NEUTRAL, (char*)&str, 0, 0);
    if(str)
    {
	cstr = string_dup(str);
	::LocalFree((HLOCAL)str);
    }
    else
    {
	cstr = string_dup("FormatMessage failed");
    }
    return cstr;
}

namespace OB
{

//
// This is a custom message logger that is used to dump messages to
// the NT system log -- the user must ensure that this logger is
// used for NT native services
//
class NTEventLogger : public WLogger,
		      public OBCORBA::RefCountLocalObject
{
    HANDLE eventSource_; // The event source
    String_var service_; // Service name
    DWORD eventId_; // The event ID

    //
    // Report the event to the NT event log
    //
    void emitMessage(int, const char*);
    void emitWMessage(int, const wchar_t*);

    bool installValues(HKEY);

public:

    NTEventLogger(const char*, DWORD);
    ~NTEventLogger();

    //
    // Install the registry entries for the event message resources
    //
    bool install();

    //
    // Remove the registry entries.
    //
    bool uninstall();
    
    //
    // Overrides for the Logger methods
    //
    virtual void info(const char*);
    virtual void winfo(const wchar_t*);
    virtual void error(const char*);
    virtual void werror(const wchar_t*);
    virtual void warning(const char*);
    virtual void wwarning(const wchar_t*);
    virtual void trace(const char*, const char*);
    virtual void wtrace(const wchar_t*, const wchar_t*);
};

} // End namespace OB

// ----------------------------------------------------------------------
// NTEventLogger private and protected member implementation
// ----------------------------------------------------------------------

void
NTEventLogger::emitMessage(int eventType, const char* msg)
{
    const char* strings[1];
    
    strings[0] = msg;
    
    ::ReportEvent(eventSource_,         // handle of event source
		  eventType,            // event type
		  0,                    // event category
		  eventId_,             // event ID
		  NULL,                 // current user's SID
		  1,                    // strings in lpszStrings
		  0,                    // no bytes of raw data
		  strings,              // array of error strings
		  NULL);                // no raw data
}

void
NTEventLogger::emitWMessage(int eventType, const wchar_t* msg)
{
    bool hasWideChar = false;
    String_var s = string_dup("");

    for(size_t i = 0 ; i < OB::wcslen(msg) ; i++)
    {
	if(msg[i] > (wchar_t)0xff)
	    hasWideChar = true;

	s += (const char)msg[i];
    }

    if(hasWideChar)
    {
	s += " - narrowed";
    }

    emitMessage(eventType, s);
}

bool
NTEventLogger::installValues(HKEY key)
{
    char path[512];
    
    //
    // Get path to executable.
    //
    if(::GetModuleFileName(NULL, path, sizeof(path)) == 0)
    {
	String_var err = GetErrorText();
	OB_CERR << "GetModuleFileName failed: " << err << OB_ENDL;
	return false;
    }
    
    const char* keyname;

    //
    // Add two keys EventMessageFile which contains the path to the
    // resources file, and TypesSupported which contains a DWORD with
    // the permitted event types.
    //
    keyname = "EventMessageFile";
    if(::RegSetValueEx(key, keyname, 0,
		       REG_EXPAND_SZ, (unsigned char*)path,
		       strlen(path)+1) == ERROR_SUCCESS)
    {
	DWORD supported = EVENTLOG_ERROR_TYPE|EVENTLOG_WARNING_TYPE|
	    EVENTLOG_INFORMATION_TYPE;
	keyname = "TypesSupported";
	if(::RegSetValueEx(key, "TypesSupported", 0, REG_DWORD,
			   (unsigned char*)&supported, sizeof(supported))
	   == ERROR_SUCCESS)
	{
	    return true;
	}
    }

    String_var err = GetErrorText();
    OB_CERR << "RegSetValueEx " << keyname << " failed: " << err << OB_ENDL;
    return false;
}

// ----------------------------------------------------------------------
// NTEventLogger constructor/destructor
// ----------------------------------------------------------------------

NTEventLogger::NTEventLogger(const char* service, DWORD eventId)
    : service_(service),
      eventId_(eventId)
{
    eventSource_ = ::RegisterEventSource(NULL, service);
    if(eventSource_ == 0)
	throw INITIALIZE("RegisterEventSource failed.");
}

NTEventLogger::~NTEventLogger()
{
    ::DeregisterEventSource(eventSource_);
}

// ----------------------------------------------------------------------
// NTEventLogger public member implementation
// ----------------------------------------------------------------------

bool
NTEventLogger::install()
{
    //
    // Constructor the key name
    //
    String_var keyName =
	string_dup("SYSTEM\\CurrentControlSet\\Services\\EventLog"
			  "\\Application\\");
    keyName += service_;

    //
    // The key goes into HKEY_LOCAL_MACHINE
    //
    HKEY regHandle = HKEY_LOCAL_MACHINE;

    //
    // First try to open the key.
    //
    HKEY keyHandle;
    if(::RegOpenKeyEx(regHandle, keyName, 0, KEY_ALL_ACCESS, &keyHandle)
       != ERROR_SUCCESS)
    {
	//
	// Create the key
	//
	DWORD d;
	if(::RegCreateKeyEx(regHandle, keyName, 0, "REG_SZ",
			    REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0,
			    &keyHandle, &d) != ERROR_SUCCESS)
	{
	    String_var err = GetErrorText();
	    OB_CERR << "RegCreateKeyEx " << keyName << " failed: " << err << OB_ENDL;
	    return false;
	}
    }

    bool rc = installValues(keyHandle);

    ::RegCloseKey(keyHandle);

    return rc;
}

bool
NTEventLogger::uninstall()
{
    //
    // Constructor the key name
    //
    String_var keyName =
	string_dup("SYSTEM\\CurrentControlSet\\Services\\EventLog"
			  "\\Application\\");
    keyName += service_;

    return ::RegDeleteKey(HKEY_LOCAL_MACHINE, keyName) == ERROR_SUCCESS;
}

void
NTEventLogger::info(const char* msg)
{
    emitMessage(EVENTLOG_INFORMATION_TYPE, msg);
}

void
NTEventLogger::winfo(const wchar_t* msg)
{
    emitWMessage(EVENTLOG_INFORMATION_TYPE, msg);
}

void
NTEventLogger::error(const char* msg)
{
    emitMessage(EVENTLOG_ERROR_TYPE, msg);
}

void
NTEventLogger::werror(const wchar_t* msg)
{
    emitWMessage(EVENTLOG_ERROR_TYPE, msg);
}

void
NTEventLogger::warning(const char* msg)
{
    emitMessage(EVENTLOG_WARNING_TYPE, msg);
}

void
NTEventLogger::wwarning(const wchar_t* msg)
{
    emitWMessage(EVENTLOG_WARNING_TYPE, msg);
}

void
NTEventLogger::trace(const char* category, const char* msg)
{
    String_var s = category;
    s += ": ";
    s += msg;
    emitMessage(EVENTLOG_INFORMATION_TYPE, s);
}

void
NTEventLogger::wtrace(const wchar_t* category, const wchar_t* msg)
{
    WString_var s = category;
    s += ": ";
    s += msg;
    emitWMessage(EVENTLOG_INFORMATION_TYPE, s);
}

//
// Define callbacks
//
static void WINAPI
OB::_OB_serviceCtrl(DWORD ctrlCode)
{
    NTService::instance() -> control(ctrlCode);
}

static void WINAPI
OB::_OB_serviceMain(DWORD argc, LPTSTR* argv)
{
    NTService::instance() -> main(argc, argv);
}

static BOOL WINAPI
_OB_controlHandler(DWORD dwCtrlType)
{
    switch( dwCtrlType )
    {
        case CTRL_BREAK_EVENT:  // use Ctrl+C or Ctrl+Break to simulate
        case CTRL_C_EVENT:      // SERVICE_CONTROL_STOP in debug mode
            NTService::instance() -> stop();
	    return true;
            break;

    }
    return false;
} 

//
// The NTService - only one instance is allowed
//
NTService* NTService::instance_ = 0;

// ----------------------------------------------------------------------
// NTService private and protected member implementation
// ----------------------------------------------------------------------

void
NTService::statusUpdate(DWORD currentState, DWORD exitCode, DWORD waitHint)
{
    //
    // when debugging we don't report to the SCM
    //
    if(debug_)
        return;
    
    if (currentState == SERVICE_START_PENDING)
        status_.dwControlsAccepted = 0;
    else
        status_.dwControlsAccepted = SERVICE_ACCEPT_STOP;

    status_.dwCurrentState = currentState;
    status_.dwWin32ExitCode = exitCode;
    status_.dwWaitHint = waitHint;

    if ( ( currentState == SERVICE_RUNNING ) ||
         ( currentState == SERVICE_STOPPED ) )
        status_.dwCheckPoint = 0;
    else
        status_.dwCheckPoint = checkPoint_++;

    //
    // Report the status of the service to the service control
    // manager.
    //
    if(!SetServiceStatus( statusHandle_, &status_) && logger_ != 0)
	logger_ -> error("SetServiceStatus");
    SetServiceStatus(statusHandle_, &status_);
}

void
NTService::control(DWORD ctrlCode)
{
    //
    // Handle the requested control code.
    //
    switch(ctrlCode)
    {
    case SERVICE_CONTROL_STOP:
        //
        // Stop the service.
        //
        // SERVICE_STOP_PENDING should be reported before setting the
        // Stop Event - hServerStopEvent - in ServiceStop().  This
        // avoids a race condition which may result in a 1053 - The
        // Service did not respond...  error.
        //
        statusUpdate(SERVICE_STOP_PENDING);
        stop();
        return;
        
    case SERVICE_CONTROL_INTERROGATE:
        //
        // Update the service status.
        //
        break;

    default:
        //
        // invalid control code
        //
        break;
    }

    statusUpdate(status_.dwCurrentState);
}

void
NTService::main(int argc, char** argv)
{
    //
    // Register our service control handler:
    //
    statusHandle_ = ::RegisterServiceCtrlHandler(name_, _OB_serviceCtrl);
    if(statusHandle_)
    {
        //
        // SERVICE_STATUS members that don't change in example
        //
        status_.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
        status_.dwServiceSpecificExitCode = 0;
        
        //
        // report the status to the service control manager.
        //
        statusUpdate(SERVICE_START_PENDING, NO_ERROR, 3000);
        start(argc, argv, logger_);
    }

    //
    // try to report the stopped status to the service control manager.
    //
    if (statusHandle_)
        statusUpdate(SERVICE_STOPPED);
}

// ----------------------------------------------------------------------
// NTService constructor and destructor
// ----------------------------------------------------------------------

NTService::NTService(const char* name, const char* title)
    : name_(name),
      title_(title),
      debug_(false),
      checkPoint_(0),
      logger_(0)
{
    assert(instance_ == 0);
    instance_ = this;

}

NTService::NTService(const char* name, const char* title, DWORD eventId)
    : name_(name),
      title_(title),
      debug_(false),
      checkPoint_(0),
      logger_(new NTEventLogger(name, eventId))
{
    assert(instance_ == 0);
    instance_ = this;
}

NTService::~NTService()
{
    if(logger_ != 0)
    {
	release(logger_);
	logger_ = 0;
    }
    assert(instance_ != 0);
    instance_ = 0;
}

// ----------------------------------------------------------------------
// NTService public member implementation
// ----------------------------------------------------------------------

NTService*
NTService::instance()
{
    return instance_;
}

bool
NTService::install(bool autoStart)
{
    bool rc = false;
    char path[512];

    //
    // Get path to executable.
    //
    if(::GetModuleFileName(NULL, path, sizeof(path)) == 0)
    {
	String_var err = GetErrorText();
	OB_CERR << "GetModuleFileName failed: " << err << OB_ENDL;
        return rc;
    }

    //
    // Open the service manager.
    //
    SC_HANDLE managerHandle =
	::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if(!managerHandle)
    {
	String_var err = GetErrorText();
        OB_CERR << "OpenSCManager failed: " <<  err << OB_ENDL;
        return rc;
    }

    //
    // Create a new service.
    //
    DWORD start = (autoStart) ? SERVICE_AUTO_START : SERVICE_DEMAND_START;
    SC_HANDLE serviceHandle =
	::CreateService(managerHandle,              // SCManager database
			name_,                      // name of service
			title_,                     // name to display
			SERVICE_ALL_ACCESS,         // desired access
			SERVICE_WIN32_OWN_PROCESS,  // service type
			start,                      // start type
			SERVICE_ERROR_NORMAL,       // error control type
			path,                       // service's binary
			NULL,                       // no load ordering group
			NULL,                       // no tag identifier
			"",                         // dependencies
			NULL,                       // LocalSystem account
			NULL);                      // no password
    
    if(serviceHandle)
    {
        OB_CERR << title_ << ": installed." << OB_ENDL;
        ::CloseServiceHandle(serviceHandle);
        rc = true;
    }
    else
    {
	String_var err = GetErrorText();
        OB_CERR << "CreateService Failed: " << err << OB_ENDL;
    }
    
    ::CloseServiceHandle(managerHandle);

    if(rc && logger_ != 0)
    {
	//
	// Initialize the event logger.
	//
	rc = logger_ -> install();
    }

    return rc;
}

bool
NTService::uninstall()
{
    bool rc = false;

    SC_HANDLE managerHandle =
        ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if(!managerHandle)
    {
	String_var err = GetErrorText();
        OB_CERR << "OpenSCManager failed: " << err << OB_ENDL;
        return rc;
    }

    SC_HANDLE serviceHandle = ::OpenService(managerHandle, name_,
					    SERVICE_ALL_ACCESS);
    if(serviceHandle)
    {
        //
        // try to stop the service
        //
        if(::ControlService(serviceHandle, SERVICE_CONTROL_STOP, &status_))
        {
            OB_CERR << "Stopping: " << title_ << OB_ENDL;
            Sleep( 1000 );

            while(::QueryServiceStatus(serviceHandle, &status_))
            {
                if(status_.dwCurrentState == SERVICE_STOP_PENDING)
                {
                    OB_CERR << "." << OB_FLUSH;
                    Sleep( 1000 );
                }
                else
                    break;
            }
            if(status_.dwCurrentState == SERVICE_STOPPED)
                OB_CERR << title_ << ": stopped." << OB_ENDL;
            else
                OB_CERR << title_ << ": failed to stop." << OB_ENDL;
        }
        
        //
        // now remove the service
        //
        if(::DeleteService(serviceHandle))
        {
            rc = true;
            OB_CERR << title_ << ": removed." << OB_ENDL;
        }
        else
        {
	    String_var err = GetErrorText();
            OB_CERR << "DeleteService failed: " << err << OB_ENDL;
        }
        ::CloseServiceHandle(serviceHandle);
    }
    else
    {
	String_var err = GetErrorText();
        OB_CERR << "OpenService failed: " << err << OB_ENDL;
    }
    ::CloseServiceHandle(managerHandle);

    if(rc && logger_ != 0)
    {
	//
	// Initialize the event logger.
	//
	rc = logger_ -> uninstall();
    }

    return rc;
}

void
NTService::run(int argc, char** argv)
{
    if(debug_)
    {
        ::SetConsoleCtrlHandler(_OB_controlHandler, true);
        start(argc, argv, logger_);
    }
    else
    {
        //
        // Define the service dispatch table.
        //
        SERVICE_TABLE_ENTRY dispatchTable[] =
        {
            { name_.inout(), (LPSERVICE_MAIN_FUNCTION)_OB_serviceMain },
            { NULL, NULL }
        };

        //
        // Start the service.
        //
        if(!::StartServiceCtrlDispatcher(dispatchTable))
	{
	    String_var err =
		string_dup("StartServiceCtrlDispatcher: ");
	    err += GetErrorText();
	    if(logger_ != 0)
		logger_ -> error(err);
	}
    }
}
