// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/LocalObject.h>
#include <OB/Except.h>
#include <OB/PIPolicy.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/PolicyFactoryManager.h>
#include <OB/Logger.h>
#include <OB/Properties.h>

#include <ORB_impl.h>
#include <ORBPolicyManager_impl.h>
#include <OBPolicies_impl.h>
#include <Messaging_impl.h>

#include <stdlib.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// The ORBPolicyFactory_impl
// ----------------------------------------------------------------------

namespace PortableInterceptor
{

class ORBPolicyFactory_impl : public PortableInterceptor::PolicyFactory,
			      public OBCORBA::RefCountLocalObject
{
public:

    //
    // IDL to C++ Mapping
    //
    virtual CORBA::Policy_ptr create_policy(CORBA::PolicyType,
                                            const CORBA::Any&);
};

} // End namespace PortableInterceptor

// ----------------------------------------------------------------------
// ORB private member functions
// ----------------------------------------------------------------------

void
OBCORBA::ORB_impl::initializeDefaultPolicies()
{
    Logger_var logger = orbInstance_ -> getLogger();
    Properties_var properties = orbInstance_ -> getProperties();

    Properties::KeySeq keys = properties -> getKeys("ooc.orb.policy.");
    ULong i;
    for(i = 0 ; i < keys.length() ; i++)
    {
        const char* key = keys[i];
        const char* value = properties -> getProperty(key);
        assert(value != 0);

        if(strcmp(key, "ooc.orb.policy.protocol") == 0)
        {
            OCI::PluginIdSeq seq;
            String_var str = value;
            const char* tok = strtok(str.inout(), ", ");
            while(tok != 0)
            {
                seq.length(seq.length() + 1);
                seq[seq.length() - 1] = tok;
                tok = strtok(0, ", ");
            }

            if(seq.length() == 0)
            {
                String_var err = string_dup("ORB_init: invalid value for "
                                            "ooc.orb.policy.protocol: `");
                err += value;
                err += "'";
                logger -> error(err);
                throw INITIALIZE(err);
            }

            ProtocolPolicy_impl* p = new ProtocolPolicy_impl(seq);
            ULong len = policies_.length();
            policies_.length(len + 1);
            policies_[len] = p;
        }
        else if(strcmp(key, "ooc.orb.policy.connection_reuse") == 0)
        {
            bool b;
            if(strcmp(value, "true") == 0)
                b = true;
            else if(strcmp(value, "false") == 0)
                b = false;
            else
            {
                String_var err = string_dup("ORB_init: invalid value for "
                                            "ooc.orb.policy.connection_reuse: "
                                            "`");
                err += value;
                err += "'";
                logger -> error(err);
                throw INITIALIZE(err);
            }

            ConnectionReusePolicy_impl* p = new ConnectionReusePolicy_impl(b);
            ULong len = policies_.length();
            policies_.length(len + 1);
            policies_[len] = p;
        }
        else if(strncmp(key, "ooc.orb.policy.retry", 20) == 0)
        {
	    // Ignore here
        }
        else if(strcmp(key, "ooc.orb.policy.timeout") == 0)
        {
            Long val = atoi(value);
            if(val != -1)
            {
                TimeoutPolicy_impl* p = new TimeoutPolicy_impl((ULong)val);
                ULong len = policies_.length();
                policies_.length(len + 1);
                policies_[len] = p;
            }
        }
        else if(strcmp(key, "ooc.orb.policy.location_transparency") == 0)
        {
            Short val;
            if(strcmp(value, "strict") == 0)
                val = OB::LOCATION_TRANSPARENCY_STRICT;
            else if(strcmp(value, "relaxed") == 0)
                val = OB::LOCATION_TRANSPARENCY_RELAXED;
            else
            {
                String_var err = string_dup("ORB_init: invalid value for "
                                            "ooc.orb.policy."
                                            "location_transparency: `");
                err += value;
                err += "'";
                logger -> error(err);
                throw INITIALIZE(err);
            }

            LocationTransparencyPolicy_impl* p =
                new LocationTransparencyPolicy_impl(val);
            CORBA::ULong len = policies_.length();
            policies_.length(len + 1);
            policies_[len] = p;
        }
        else if(strcmp(key, "ooc.orb.policy.interceptor") == 0)
        {
            bool b;
            if(strcmp(value, "true") == 0)
                b = true;
            else if(strcmp(value, "false") == 0)
                b = false;
            else
            {
                String_var err = string_dup("ORB_init: invalid value for "
                                            "ooc.orb.policy.interceptor: `");
                err += value;
                err += "'";
                logger -> error(err);
                throw INITIALIZE(err);
            }

            InterceptorPolicy_impl* p = new InterceptorPolicy_impl(b);
            CORBA::ULong len = policies_.length();
            policies_.length(len + 1);
            policies_[len] = p;
        }
        else if(strcmp(key, "ooc.orb.policy.connect_timeout") == 0)
        {
            Long val = atoi(value);
            if(val != -1)
            {
                ConnectTimeoutPolicy_impl* p =
                    new ConnectTimeoutPolicy_impl((ULong)val);
                ULong len = policies_.length();
                policies_.length(len + 1);
                policies_[len] = p;
            }
        }
        else if(strcmp(key, "ooc.orb.policy.request_timeout") == 0)
        {
            Long val = atoi(value);
            if(val != -1)
            {
                RequestTimeoutPolicy_impl* p =
                    new RequestTimeoutPolicy_impl((ULong)val);
                ULong len = policies_.length();
                policies_.length(len + 1);
                policies_[len] = p;
            }
        }
        else if(strcmp(key, "ooc.orb.policy.locate_request") == 0)
        {
            bool b;
            if(strcmp(value, "true") == 0)
                b = true;
            else if(strcmp(value, "false") == 0)
                b = false;
            else
            {
                String_var err = string_dup("ORB_init: invalid value for "
                                            "ooc.orb.policy.locate_request:"
                                            " `");
                err += value;
                err += "'";
                logger -> error(err);
                throw INITIALIZE(err);
            }

            LocateRequestPolicy_impl* p = new LocateRequestPolicy_impl(b);
            CORBA::ULong len = policies_.length();
            policies_.length(len + 1);
            policies_[len] = p;
        }
        else if(strcmp(key, "ooc.orb.policy.rebind") == 0)
        {
            Short val;
            if(strcmp(value, "transparent") == 0)
                val = Messaging::TRANSPARENT;
            else if(strcmp(value, "no_rebind") == 0)
                val = Messaging::NO_REBIND;
            else if(strcmp(value, "no_reconnect") == 0)
                val = Messaging::NO_RECONNECT;
            else
            {
                String_var err = string_dup("ORB_init: invalid value for "
                                            "ooc.orb.policy.rebind: `");
                err += value;
                err += "'";
                logger -> error(err);
                throw INITIALIZE(err);
            }

            Messaging::RebindPolicy_impl* p =
                new Messaging::RebindPolicy_impl(val);
            CORBA::ULong len = policies_.length();
            policies_.length(len + 1);
            policies_[len] = p;
        }
        else if(strcmp(key, "ooc.orb.policy.sync_scope") == 0)
        {
            Short val;
            if(strcmp(value, "none") == 0)
                val = Messaging::SYNC_NONE;
            else if(strcmp(value, "transport") == 0)
                val = Messaging::SYNC_WITH_TRANSPORT;
            else if(strcmp(value, "server") == 0)
                val = Messaging::SYNC_WITH_SERVER;
            else if(strcmp(value, "target") == 0)
                val = Messaging::SYNC_WITH_TARGET;
            else
            {
                String_var err = string_dup("ORB_init: invalid value for "
                                            "ooc.orb.policy.sync_scope: `");
                err += value;
                err += "'";
                logger -> error(err);
                throw INITIALIZE(err);
            }

            Messaging::SyncScopePolicy_impl* p =
                new Messaging::SyncScopePolicy_impl(val);
            CORBA::ULong len = policies_.length();
            policies_.length(len + 1);
            policies_[len] = p;
        }
        else
        {
            String_var err = string_dup("ORB_init: unknown property `");
            err += key;
            err += "'";
            logger -> error(err);
            throw INITIALIZE(err);
        }
    }

    //
    // Set the default policies, if not already set
    //
    if(properties -> getProperty("ooc.orb.policy.connection_reuse") == 0)
    {
        ConnectionReusePolicy_impl* p = new ConnectionReusePolicy_impl(true);
        CORBA::ULong len = policies_.length();
        policies_.length(len + 1);
        policies_[len] = p;
    }

    //
    // Set up the retry policy. Default STRICT, immediate retry, 2
    // attempts, not retry on remote exception
    //
    Short retry_mode = OB::RETRY_STRICT;
    ULong retry_interval = 0;
    ULong max_retries = 1;
    bool remote_retry = false;

    const char* value;
    if((value = properties -> getProperty("ooc.orb.policy.retry")) != 0)
    {
        if(strcmp(value, "never") == 0)
            retry_mode = OB::RETRY_NEVER;
        else if(strcmp(value, "strict") == 0)
            retry_mode = OB::RETRY_STRICT;
        else if(strcmp(value, "always") == 0)
            retry_mode = OB::RETRY_ALWAYS;
        else
        {
            String_var err = string_dup("ORB_init: invalid value for "
                                        "ooc.orb.policy.retry: `");
            err += value;
            err += "'";
            logger -> error(err);
            throw INITIALIZE(err);
        }
    }
    if((value = properties -> getProperty("ooc.orb.policy.retry.interval"))
        != 0)
    {
        char* end;
        retry_interval = strtoul(value, &end, 0);
    }
    if((value = properties -> getProperty("ooc.orb.policy.retry.max")) != 0)
    {
        char* end;
        max_retries = strtoul(value, &end, 0);
    }
    if((value = properties -> getProperty("ooc.orb.policy.retry.remote"))
        != 0)
        remote_retry = (strcmp(value, "true") == 0);

    RetryPolicy_impl* p = new RetryPolicy_impl(retry_mode, retry_interval,
    					       max_retries, remote_retry);
    CORBA::ULong len = policies_.length();
    policies_.length(len + 1);
    policies_[len] = p;

    //
    // Create the ORBPolicyManager
    //
    PolicyManager_var pm = new ORBPolicyManager_impl(policies_);
    InitialServiceManager_var initServiceManager =
        orbInstance_ -> getInitialServiceManager();
    initServiceManager -> addInitialReference("ORBPolicyManager", pm);
    ObjectFactory_var objectFactory = orbInstance_ -> getObjectFactory();
    objectFactory -> setPolicyManager(pm);

    //
    // Register the default PolicyFactory policies for the ORB
    //
    PolicyFactoryManager_var pfm = orbInstance_ -> getPolicyFactoryManager();

    PortableInterceptor::PolicyFactory_var factory = new
        PortableInterceptor::ORBPolicyFactory_impl;
    pfm -> registerPolicyFactory(CONNECTION_REUSE_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(PROTOCOL_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(RETRY_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(TIMEOUT_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(LOCATION_TRANSPARENCY_POLICY_ID, factory,
                                 true);
    pfm -> registerPolicyFactory(Messaging::REBIND_POLICY_TYPE, factory, true);
    pfm -> registerPolicyFactory(Messaging::SYNC_SCOPE_POLICY_TYPE, factory,
                                 true);
    pfm -> registerPolicyFactory(INTERCEPTOR_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(CONNECT_TIMEOUT_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(REQUEST_TIMEOUT_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(Messaging::REQUEST_START_TIME_POLICY_TYPE,
    				 factory, true);
    pfm -> registerPolicyFactory(Messaging::REQUEST_END_TIME_POLICY_TYPE,
    				 factory, true);
    pfm -> registerPolicyFactory(Messaging::REPLY_START_TIME_POLICY_TYPE,
    				 factory, true);
    pfm -> registerPolicyFactory(Messaging::REPLY_END_TIME_POLICY_TYPE,
    				 factory, true);
    pfm -> registerPolicyFactory(Messaging::RELATIVE_REQ_TIMEOUT_POLICY_TYPE,
    				 factory, true);
    pfm -> registerPolicyFactory(Messaging::RELATIVE_RT_TIMEOUT_POLICY_TYPE,
    				 factory, true);
}

// ----------------------------------------------------------------------
// ORBPolicyFactory public member functions
// ----------------------------------------------------------------------

Policy_ptr
PortableInterceptor::ORBPolicyFactory_impl::create_policy(
    PolicyType policy_type,
    const Any& val)
{
    //
    // A case stmt would be better here, but egcs complains: case
    // label `CONNECTION_REUSE_POLICY_ID' does not reduce to an
    // integer constant
    //
    if(policy_type == CONNECTION_REUSE_POLICY_ID)
    {
	Boolean b;
	if(val >>= Any::to_boolean(b))
        {
	    return new ConnectionReusePolicy_impl(b);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    if(policy_type == PROTOCOL_POLICY_ID)
    {
        const OCI::PluginIdSeq* seq;
        if(val >>= seq)
        {
            return new ProtocolPolicy_impl(*seq);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    if(policy_type == RETRY_POLICY_ID)
    {
        Short v;
        if(val >>= v)
        {
            if(v > RETRY_ALWAYS)
                throw PolicyError(BAD_POLICY_VALUE);
            return new RetryPolicy_impl(v, 0, 1, false);
        }
	const RetryAttributes* attr;
	if(val >>= attr)
	{
	    return new RetryPolicy_impl(attr -> mode, attr -> interval, 
	    				attr -> max, attr -> remote);
	}
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == TIMEOUT_POLICY_ID)
    {
        ULong t;
        if(val >>= t)
        {
            return new TimeoutPolicy_impl(t);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == LOCATION_TRANSPARENCY_POLICY_ID)
    {
        Short v;
        if(val >>= v)
            return new LocationTransparencyPolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == Messaging::REBIND_POLICY_TYPE)
    {
        Messaging::RebindMode v;
        if(val >>= v)
            return new Messaging::RebindPolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == Messaging::SYNC_SCOPE_POLICY_TYPE)
    {
        Messaging::SyncScope v;
        if(val >>= v)
            return new Messaging::SyncScopePolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == Messaging::REQUEST_START_TIME_POLICY_TYPE)
    {
        const TimeBase::UtcT* v;
        if(val >>= v)
            return new Messaging::RequestStartTimePolicy_impl(*v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == Messaging::REQUEST_END_TIME_POLICY_TYPE)
    {
        const TimeBase::UtcT* v;
        if(val >>= v)
            return new Messaging::RequestEndTimePolicy_impl(*v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == Messaging::REPLY_START_TIME_POLICY_TYPE)
    {
        const TimeBase::UtcT* v;
        if(val >>= v)
            return new Messaging::ReplyStartTimePolicy_impl(*v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == Messaging::REPLY_END_TIME_POLICY_TYPE)
    {
        const TimeBase::UtcT* v;
        if(val >>= v)
            return new Messaging::ReplyEndTimePolicy_impl(*v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == Messaging::RELATIVE_REQ_TIMEOUT_POLICY_TYPE)
    {
        TimeBase::TimeT v;
        if(val >>= v)
            return new Messaging::RelativeRequestTimeoutPolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == Messaging::RELATIVE_RT_TIMEOUT_POLICY_TYPE)
    {
        TimeBase::TimeT v;
        if(val >>= v)
            return new Messaging::RelativeRoundTripTimeoutPolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == INTERCEPTOR_POLICY_ID)
    {
        Boolean v;
        if(val >>= Any::to_boolean(v))
            return new InterceptorPolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == CONNECT_TIMEOUT_POLICY_ID)
    {
        ULong t;
        if(val >>= t)
        {
            return new ConnectTimeoutPolicy_impl(t);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == REQUEST_TIMEOUT_POLICY_ID)
    {
        ULong t;
        if(val >>= t)
        {
            return new RequestTimeoutPolicy_impl(t);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }

    if(policy_type == LOCATE_REQUEST_POLICY_ID)
    {
        Boolean v;
        if(val >>= Any::to_boolean(v))
            return new LocateRequestPolicy_impl(v);
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    throw PolicyError(BAD_POLICY);
    return Policy::_nil(); // Some compilers need this
}
