// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/NativeTypes.h>
#include <OB/Stream.h>
#include <OB/OBObject.h>
#include <OB/LocalObject.h>
#include <OB/RefCounted.h>
#include <OB/ORBInstance.h>
#include <OB/MarshalStubImpl.h>
#include <OB/OBORB.h>
#include <OB/InitialServiceManager.h>
#include <OB/INS.h>
#include <OB/DynamicAny.h>
#include <OB/DII.h>
#include <OB/TypeCode.h>
#include <OB/Environment.h>

#include <URLUtil.h>
#include <URL_impl.h>
#include <TypeCodeFactory.h>

#include <ctype.h>
#include <stdio.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Convert a hex character into an integer
// ----------------------------------------------------------------------

static int
ConvertHex(char ch)
{
    if(ch >= '0' && ch <= '9')
        return ch - '0';
    if(ch >= 'A' && ch <= 'F')
        return (ch - 'A') + 10;
    if(ch >= 'a' && ch <= 'f')
        return (ch - 'a') + 10;

    throw BAD_PARAM("URL: bad escape sequence character",
                    MinorBadAddress, COMPLETED_NO);

    return 0; // some compilers need this
}

// ----------------------------------------------------------------------
// Unescape a URL
// ----------------------------------------------------------------------

char*
OB::UnescapeURL(char* start)
{
    const char* end = start + strlen(start);

    char* cp = start;

    while(start != end)
    {
        char ch = *start;

        // 
        // Escape sequence '%' must be followed by
        // 2 valid characters 0-9, A-F, a-f
        // 
        if(ch == '%')
        {
            if(start + 2 >= end)
                throw BAD_PARAM("URL: bad escape sequence length",
                                MinorBadAddress,COMPLETED_NO);

            char c1 = *++start;
            char c2 = *++start;
            ch = (ConvertHex(c1) << 4) | ConvertHex(c2);
        }
        *cp++ = ch;
        ++start;
    }

    *cp = '\0';

    return cp;
}

// ----------------------------------------------------------------------
// URLRegistry_impl constructor/destructor
// ----------------------------------------------------------------------

URLRegistry_impl::URLRegistry_impl()
    : schemes_(11)
{
}

URLRegistry_impl::~URLRegistry_impl()
{
}

// ----------------------------------------------------------------------
// URLRegistry_impl public member implementation
// ----------------------------------------------------------------------

void
URLRegistry_impl::add_scheme(URLScheme_ptr scheme)
    throw(URLRegistry::SchemeAlreadyExists,
          SystemException)
{
    String_var name = scheme -> name();
    if(schemes_.containsKey(name))
        throw SchemeAlreadyExists();

    schemes_.put(name, URLScheme::_duplicate(scheme));
}

URLScheme_ptr
URLRegistry_impl::find_scheme(const char* name)
    throw(SystemException)
{
    URLScheme_var scheme;
    schemes_.get(name, scheme);
    return scheme._retn();
}

Object_ptr
URLRegistry_impl::parse_url(const char* url)
    throw(SystemException)
{
    if(url == 0)
        throw BAD_PARAM();

    String_var urlCopy = url;
    char* colon = strchr(urlCopy, ':');
    if(colon == 0 || colon == urlCopy.in())
        throw BAD_PARAM(MinorBadSchemeName, COMPLETED_NO);

    for(char* cp = urlCopy.inout() ; cp < colon ; cp++)
        *cp = tolower(*cp);

    String_var schemeName = urlCopy;
    schemeName[colon - urlCopy.in()] = '\0';

    URLScheme_var scheme = find_scheme(schemeName);
    if(is_nil(scheme))
        throw BAD_PARAM(MinorBadSchemeName, COMPLETED_NO);

    return scheme -> parse_url(urlCopy);
}

void
URLRegistry_impl::destroy()
    throw(SystemException)
{
    schemes_.clear();
}

// ----------------------------------------------------------------------
// IORURLScheme_impl constructor/destructor
// ----------------------------------------------------------------------

IORURLScheme_impl::IORURLScheme_impl(ORBInstance_ptr orbInstance)
    : orbInstance_(ORBInstance::_duplicate(orbInstance))
{
}

IORURLScheme_impl::~IORURLScheme_impl()
{
}

// ----------------------------------------------------------------------
// IORURLScheme_impl public member implementation
// ----------------------------------------------------------------------

char*
IORURLScheme_impl::name()
    throw(SystemException)
{
    return string_dup("ior");
}

Object_ptr
IORURLScheme_impl::parse_url(const char* url)
    throw(SystemException)
{
    const char* s = url + 4; // skip "IOR:"
    ULong len = strlen(s);

    if(len % 2)
        throw BAD_PARAM("URL: invalid length",
                        MinorBadSchemeSpecificPart, COMPLETED_NO);

    Octet* octBuf = 0;
    octBuf = AsciiToOctets(s);
    OCI::BufferImpl buf(len / 2);
    memcpy(buf.data(), octBuf, len / 2);
    delete [] octBuf;

    try
    {
        InputStreamImpl in(&buf, 0, false);
        in._OB_readEndian();
        IOP::IOR_var ior = new IOP::IOR;
        IOP::IOR::_OB_unmarshal(ior.inout(), &in);
        ObjectFactory_var objectFactory = orbInstance_ -> getObjectFactory();
        RefCountIOR_var refIOR = new RefCountIOR(ior);
        return objectFactory -> createObject(refIOR);
    }
    catch(const MARSHAL&)
    {
        delete [] octBuf;
        throw BAD_PARAM("URL: invalid", MinorBadSchemeSpecificPart,
                        COMPLETED_NO);
    }

    return 0; // some compilers need this
}

void
IORURLScheme_impl::destroy()
    throw(SystemException)
{
    orbInstance_ = ORBInstance::_nil();
}

// ----------------------------------------------------------------------
// FileURLScheme_impl constructor/destructor
// ----------------------------------------------------------------------

FileURLScheme_impl::FileURLScheme_impl(bool relative, URLRegistry_ptr registry)
    : relative_(relative), registry_(URLRegistry::_duplicate(registry))
{
}

FileURLScheme_impl::~FileURLScheme_impl()
{
}

// ----------------------------------------------------------------------
// FileURLScheme_impl public member implementation
// ----------------------------------------------------------------------

char*
FileURLScheme_impl::name()
    throw(SystemException)
{
    return relative_ ? string_dup("relfile") : string_dup("file");
}

Object_ptr
FileURLScheme_impl::parse_url(const char* url)
    throw(SystemException)
{
    // 
    // WIN32 doesn't have an absolute root for the `file:' scheme.
    // A drive letter or network machine name may proceed the 
    // filename.
    // 
#ifdef WIN32
    char fileAbsolutePrefix = 0;
#else
    char fileAbsolutePrefix = '/';
#endif

    const char* start = url;
    if(relative_)
        start += 8; // skip "relfile:"
    else
        start += 5; // skip "file:"

    //
    // Allow up to three leading '/''s to match commonly used forms
    // of the "file:" URL scheme
    // 
    for(int n = 0 ; *start != '\0' && n < 3 ; n++, start++)
        if(*start != '/')
            break;

    if(*start == '\0')
        throw BAD_PARAM("URL: no file name specified", 
                        MinorBadSchemeSpecificPart, COMPLETED_NO);

    String_var fileName;
    if(!relative_ && fileAbsolutePrefix != 0)
        fileName += fileAbsolutePrefix;
    String_var urlFile = start;
    UnescapeURL(urlFile.inout());
    fileName += urlFile;

    FILE* f = fopen(fileName, "r");
    if(f != 0)
    {
        String_var s;
        char buf[64];

        while(fgets(buf, 64, f))
            s += buf;

        //
        // Explicit close required for NT file locking
        //
        fclose(f);

        //
        // Null terminate the read string
        //
        char *p = strpbrk(s.inout(), "\r\n");
        if(p != 0)
            *p = '\0';

        return registry_ -> parse_url(s);
    }

    throw BAD_PARAM("URL: bad file name",
                    MinorBadSchemeSpecificPart, COMPLETED_NO);

    return 0; // some compilers need this
}

void
FileURLScheme_impl::destroy()
    throw(SystemException)
{
    registry_ = URLRegistry::_nil();
}

// ----------------------------------------------------------------------
// CorbalocURLScheme_impl constructor/destructor
// ----------------------------------------------------------------------

CorbalocURLScheme_impl::CorbalocURLScheme_impl(ORBInstance_ptr orbInstance)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)), protocols_(11)
{
}

CorbalocURLScheme_impl::~CorbalocURLScheme_impl()
{
}

// ----------------------------------------------------------------------
// CorbalocURLScheme_impl private member implementation
// ----------------------------------------------------------------------

Object_ptr
CorbalocURLScheme_impl::parse_addresses(const char* start, const char* end,
                                       char*& keyStr)
{
    //
    // Check for rir:
    //
    if(strncmp(start, "rir:", 4) == 0)
    {
        const char* comma = strchr(start, ',');
        if(comma != 0 && comma <= end)
            throw BAD_PARAM("URL: rir cannot be used with other protocols",
                            MinorBadSchemeSpecificPart, COMPLETED_NO);

        if(start + 3 != end)
            throw BAD_PARAM("URL: rir does not allow an address",
                            MinorBadAddress, COMPLETED_NO);

        try 
        {
            InitialServiceManager_var initialServiceManager =
                orbInstance_ -> getInitialServiceManager();
            return initialServiceManager -> resolveInitialReferences(keyStr);
        }
        catch(const ORB::InvalidName&)
        {
            throw BAD_PARAM("URL: invalid initial reference token",
                            MinorBadSchemeSpecificPart, COMPLETED_NO);
        }
    }

    //
    // Unescape stringified object key and convert to octets
    //
    CorbalocProtocol::ObjectKey key;
    const char* keyEnd = UnescapeURL(keyStr);
    if(keyEnd == 0)
        throw BAD_PARAM("URL: invalid escape sequence",
                        MinorBadAddress, COMPLETED_NO);
    ULong keyLen = keyEnd - keyStr;
    key.length(keyLen);
    memcpy((char*)key.get_buffer(), keyStr, keyLen);

    OCI::IOR ior;
    ior.type_id = string_dup("");

    // 
    // Convert addresses (separated by ',') into IOR profiles
    // 
    const char* cp = start;
    while(cp <= end)
    {
        ULong len;

        //
        // Get the protocol identifier - we'll assume that protocols are
        // terminated by a ':'
        //
        String_var protocol;
        const char* colon = strchr(cp, ':');
        if(colon == 0)
            throw BAD_PARAM("URL: no protocol",
                            MinorBadSchemeSpecificPart, COMPLETED_NO);
        else if(colon > end)
            break;
        if(colon == cp) // ":" is shorthand for "iiop:"
            protocol = string_dup("iiop");
        else
        {
            len = colon - cp;
            protocol = string_alloc(len);
            strncpy(protocol.inout(), cp, len);
            protocol[len] = '\0';
        }
        cp = colon;

        LowerCase(protocol.inout());

        //
        // Check for rir (again)
        //
        if(strcmp(protocol, "rir") == 0)
            throw BAD_PARAM("URL: rir cannot be used with other protocols",
                            MinorBadSchemeSpecificPart, COMPLETED_NO);

        //
        // Get the protocol address
        //
        String_var addr;
        if(cp == end)
        {
            addr = string_dup("");
            cp++;
        }
        else
        {
            const char* addrStart = cp + 1; // skip ':'
            const char* addrEnd;
            const char* comma = strchr(addrStart, ',');
            if(comma == 0 || comma > end)
            {
                addrEnd = end;
                cp = end + 1;
            }
            else
            {
                addrEnd = comma - 1;
                cp = comma + 1;
            }
            len = addrEnd - addrStart + 1;
            addr = string_alloc(len);
            strncpy(addr.inout(), addrStart, len);
            addr[len] = '\0';
        }

        //
        // Find the protocol object
        //
        CorbalocProtocol_var p = find_protocol(protocol);
        if(!is_nil(p))
        {
            IOP::TaggedProfile_var profile = p -> parse_address(addr, key);
            ULong n = ior.profiles.length();
            ior.profiles.length(n + 1);
            ior.profiles[n].tag = profile -> tag;
            ULong dataMax = profile -> profile_data.maximum();
            ULong dataLen = profile -> profile_data.length();
            Octet* buf = profile -> profile_data.get_buffer(true);
            ior.profiles[n].profile_data.replace(dataMax, dataLen, buf, true);
        }
    }

    if(ior.profiles.length() == 0)
        throw BAD_PARAM("URL: no valid protocol addresses",
                        MinorBadSchemeSpecificPart, COMPLETED_NO);

    ObjectFactory_var objectFactory = orbInstance_ -> getObjectFactory();
    RefCountIOR_var refIOR = new RefCountIOR(ior);
    return objectFactory -> createObject(refIOR);
}

// ----------------------------------------------------------------------
// CorbalocURLScheme_impl public member implementation
// ----------------------------------------------------------------------

char*
CorbalocURLScheme_impl::name()
    throw(SystemException)
{
    return string_dup("corbaloc");
}

Object_ptr
CorbalocURLScheme_impl::parse_url(const char* url)
    throw(SystemException)
{
    //
    // Get the object key
    //
    const char* cp = strchr(url, '/');

    //
    // Although an object key is optional according to the specification,
    // we consider this to be an invalid URL
    //
    if(cp == 0)
        throw BAD_PARAM("URL: no key specified", MinorOther, COMPLETED_NO);

    String_var keyStr = string_dup(cp + 1);

    const char* addrStart = url + 9; // skip "corbaloc:"
    const char* addrEnd = cp - 1;
    if(addrStart == cp)
        throw BAD_PARAM("URL: no protocol address",
                        MinorBadAddress, COMPLETED_NO);

    return parse_addresses(addrStart, addrEnd, keyStr.inout());
}

void
CorbalocURLScheme_impl::destroy()
    throw(SystemException)
{
    protocols_.clear();
    orbInstance_ = ORBInstance::_nil();
}

void
CorbalocURLScheme_impl::add_protocol(CorbalocProtocol_ptr protocol)
    throw(CorbalocURLScheme::ProtocolAlreadyExists,
          SystemException)
{
    String_var name = protocol -> name();
    if(protocols_.containsKey(name))
        throw ProtocolAlreadyExists();

    protocols_.put(name, CorbalocProtocol::_duplicate(protocol));
}

CorbalocProtocol_ptr
CorbalocURLScheme_impl::find_protocol(const char* name)
    throw(SystemException)
{
    CorbalocProtocol_var protocol;
    protocols_.get(name, protocol);
    return protocol._retn();
}

// ----------------------------------------------------------------------
// CorbanameURLScheme_impl constructor/destructor
// ----------------------------------------------------------------------

CorbanameURLScheme_impl::CorbanameURLScheme_impl(
    ORB_ptr orb,
    URLRegistry_ptr registry)
    : orb_(ORB::_duplicate(orb))
{
    URLScheme_var scheme = registry -> find_scheme("corbaloc");
    corbaloc_ = CorbalocURLScheme::_narrow(scheme);
}

CorbanameURLScheme_impl::~CorbanameURLScheme_impl()
{
}

// ----------------------------------------------------------------------
// CorbanameURLScheme_impl public member implementation
// ----------------------------------------------------------------------

char*
CorbanameURLScheme_impl::name()
    throw(SystemException)
{
    return string_dup("corbaname");
}

Object_ptr
CorbanameURLScheme_impl::parse_url(const char* url)
    throw(SystemException)
{
    //
    // Get the object key
    //
    String_var keyStr;
    const char* cp = strchr(url, '/');
    const char* fragmentStart = strchr(url, '#');
    if(cp != 0 && fragmentStart == 0)
    {
        //
        // e.g., corbaname::localhost:5000/blah
        //
        keyStr = string_dup(cp + 1);
    }
    else if(cp == 0 || fragmentStart - 1 == cp || fragmentStart < cp)
    {
        //
        // e.g., corbaname::localhost:5000
        //       corbaname::localhost:5000/#foo
        //       corbaname::localhost:5000#foo/bar
        //
        keyStr = string_dup("NameService");
    }
    else
    {
        ULong len = fragmentStart - cp - 1;
        keyStr = string_alloc(len);
        strncpy(keyStr.inout(), cp + 1, len);
        keyStr[len] = '\0';
    }

    //
    // Get start and end of protocol address(es)
    //
    const char* addrStart = url + 10; // skip "corbaname:"
    const char* addrEnd;

    if(addrStart == cp)
        throw BAD_PARAM("URL: no protocol address", MinorBadAddress,
                        COMPLETED_NO);

    if(cp == 0 && fragmentStart == 0)
        addrEnd = url + strlen(url) - 1;
    else if((cp != 0 && fragmentStart == 0) ||
            (cp != 0 && fragmentStart != 0 && cp < fragmentStart))
        addrEnd = cp - 1;
    else
        addrEnd = fragmentStart - 1;

    ULong addrLen = addrEnd - addrStart + 1;

    //
    // Create a corbaloc URL
    //
    String_var corbaloc = string_alloc(
        9 +             /* "corbaloc:" */
        addrLen +       /* address(es) */
        1 +             /* slash */
        strlen(keyStr)  /* key */
    );
    strcpy(corbaloc.inout(), "corbaloc:");
    strncpy(corbaloc.inout() + 9, addrStart, addrLen);
    strncpy(corbaloc.inout() + 9 + addrLen, "/", 1);
    strcpy(corbaloc.inout() + 9 + addrLen + 1, keyStr);

    //
    // Create object reference for the naming context
    //
    Object_var nc = corbaloc_ -> parse_url(corbaloc);

    // 
    // If there is no URL fragment "#.....", or the stringified
    // name is empty, then the URL refers to the naming context
    // itself
    // 
    if(fragmentStart == 0 || *(fragmentStart + 1) == '\0')
        return nc._retn();

    //
    // Make a DII invocation on the Naming Service to resolve the
    // specified context
    //
    try
    {
        //
        // Create typecodes for Name and NameComponent
        //
        StructMemberSeq contents;
        contents.length(2);
        contents[0].name = string_dup("id");
        contents[0].type = TypeCodeFactory::createStringTC(0);
        contents[1].name = string_dup("kind");
        contents[1].type = TypeCodeFactory::createStringTC(0);
        TypeCode_var tcNameComponent =
            TypeCodeFactory::createStructTC(
                "IDL:omg.org/CosNaming/NameComponent:1.0",
                "NameComponent", contents);

        TypeCode_var tcName =
            TypeCodeFactory::createSequenceTC(0, tcNameComponent);

        //
        // Parse path (remove URL escapes first) and create
        // NameComponent sequence
        //
        String_var fragment = string_dup(fragmentStart + 1);
        const char* fragmentEnd = UnescapeURL(fragment.inout());
        ULong fragmentLen = fragmentEnd - fragment.inout() + 1;
        OCI::ObjectKey name(fragmentLen, fragmentLen, (Octet*)fragment.in(),
                            false);
        CORBANameParser parser(name);
        if(!parser.isValid())
            throw BAD_PARAM("URL: invalid stringified name", 
                            MinorBadSchemeSpecificPart, COMPLETED_NO);

        const StrSeq< int >& content = parser.getContents();
        assert(content.length() % 2 == 0);

        Object_var dynAnyFactoryObj =
            orb_ -> resolve_initial_references("DynAnyFactory");
        DynamicAny::DynAnyFactory_var dynAnyFactory =
            DynamicAny::DynAnyFactory::_narrow(dynAnyFactoryObj);

        DynamicAny::AnySeq as; // Don't use "AnySeq"
        for(ULong i = 0 ; i < content.length() ; i += 2)
        {
            //
            // Create the DynStruct containing the id and kind fields
            //
            DynamicAny::DynAny_var dynAny = dynAnyFactory ->
                create_dyn_any_from_type_code(tcNameComponent);
            DynamicAny::DynStruct_var name = 
                DynamicAny::DynStruct::_narrow(dynAny);
            name -> insert_string(content[i]);
            name -> next();
            name -> insert_string(content[i + 1]);

            Any_var nany = name -> to_any();
            name -> destroy();

            ULong asLength = as.length();
            as.length(asLength + 1);
            as[asLength] = nany;
        }

        //
        // Create the Name
        //
        DynamicAny::DynAny_var dynAny =
            dynAnyFactory -> create_dyn_any_from_type_code(tcName);
        DynamicAny::DynSequence_var seq = 
            DynamicAny::DynSequence::_narrow(dynAny);
        seq -> set_length(as.length());
        seq -> set_elements(as);
        Any_var any = seq -> to_any();
        seq -> destroy();

        //
        // Create the DII request
        //
        Request_var request = nc -> _request("resolve");

        //
        // Copy in the arguments
        //
        Any& arg = request -> add_in_arg();
        arg = any;
        request -> set_return_type(_tc_Object);

        //
        // Invoke the request
        //
        request -> invoke();

        //
        // Return the result if there was no exception
        //
        if(request -> env() -> exception() == 0)
        {
            Object_ptr obj;
            request -> return_value() >>= obj;

            return Object::_duplicate(obj);
        }
    }
    catch(const SystemException&)
    {
        // Fall through
    }
    catch(const UserException&)
    {
        // Fall through
    }

    throw BAD_PARAM("URL: corbaname evaluation error", MinorOther,
                    COMPLETED_NO);

    return Object::_nil(); // Some compilers need this
}

void
CorbanameURLScheme_impl::destroy()
    throw(SystemException)
{
    orb_ = ORB::_nil();
    corbaloc_ = CorbalocURLScheme::_nil();
}
